-- import Data.Map (Map)
-- import qualified Data.Map as DM

import Data.Map

-- NOTE: if keys are Int, use Data.IntMap instead.

m0 = empty

-- Or: m1 == insert "foo" 3 m0
m1 = singleton "foo" 3
-- Easy element access; exception if missing.
three = m1!"foo"

-- Add "bar",1: (+) is not called
m2 = insertWith (+) "bar" 1 m1

-- "bar" value is now 1+3 = 4
m3 = insertWith (+) "bar" 3 m2

-- "bar" value is now 3+1+4 = 8
m4 = insertWithKey (\k v1 v2 -> length k + v1 + v2) "bar" 1 m3

-- Remove foo
m5 = delete "foo" m4

-- Double value of key "bar" (no effect if no "bar"). "bar" now 16
m6 = adjust (*2) "bar" m5

-- Value of "bar" gets length of "bar" added, i.e. 19
m7 = adjustWithKey (\k v -> length k + v) "bar" m6

-- m8 == fromList [("bar",19),("foo",6)]
m8 = insert "foo" 6 m7

incOrDel x = if rem x 2 == 0 then Just (x+1) else Nothing

-- m9 == fromList [("bar",19),("foo",7)]
m9  = update incOrDel "foo" m8
-- m9b == fromList [("foo",6)]
m9b = update incOrDel "bar" m8

keepIfBig k x = if x >= 3 * length k then Just x else Nothing

-- Keeps "bar", as 19 >= 3*3
m10 = updateWithKey keepIfBig "bar" m9
-- Drops "foo", as 7 < 3*3
m10b = updateWithKey keepIfBig "foo" m9

incIfBig k x = if x >= 3 * length k then Just (x+1) else Nothing

-- Update and THEN lookup...
-- (Just 20,fromList [("bar",20),("foo",7)])
p1 = updateLookupWithKey incIfBig "bar" m10
-- (Just 7,fromList [("bar",19)])
p2 = updateLookupWithKey incIfBig "foo" m10
-- (Nothing,fromList [("bar",19),("foo",7)])
p3 = updateLookupWithKey incIfBig "baz" m10

incIfPresent v = 
  case v of 
    Nothing -> Nothing
    Just x -> Just (x+1)

-- fromList [("bar",19),("foo",8)]
m11 = alter incIfPresent "foo" m10

-- m2 is bar 1, foo 3
-- m3 is bar 4, foo 3
-- Left-biased union: has ("bar",1)
m12 = union m2 m3

-- bar 5, foo 6
m13 = unionWith (+) m2 m3

addIfFoo k x y = if k == "foo" then x+y else x-y

-- bar -3, foo 6, because we add only when key is "foo"
m14 = unionWithKey addIfFoo m2 m3

-- fromList [("bar",1),("foo",3)]
m15 = unions [m1,m2,m3]

-- fromList [("bar",5),("foo",9)]
m16 = unionsWith (+) [m1,m2,m3]

-- takes difference on keys, even if values differ...
m17 = difference m2 m1

keepIfNoSmaller x y = if x >= y then Just x else Nothing
-- Keeps both keys, because 3 >= 3 for "foo", and "bar" not in m1
m18 = differenceWith keepIfNoSmaller m2 m1

keepIfBiggerOrFoo k x y = if (k == "foo" || x > y) then Just x else Nothing
-- Keeps both keys, because "foo" kept anyway and "bar" not in m1
m19 = differenceWithKey keepIfBiggerOrFoo m2 m1

-- bar=1, foo=3; i.e. keep values from m2
m20 = intersection m2 m3

-- bar=5, foo=6; i.e. when key is in both, add values
m21 = intersectionWith (+) m2 m3

-- ("bar",-3),("foo",6): we add for foo and subtract for anything else
m22 = intersectionWithKey (\k x y -> if k == "foo" then x+y else x-y) m2 m3

-- Add 1 to every value; "map" is ambiguous with Data.List.map
-- ("bar",-2),("foo",7)
m23 = Data.Map.map (+1) m22

-- Add length of key to every value
m24 = mapWithKey (\k x -> length k + x) m23

-- Start with 1; "bar" first; its new value is x+y, and we pass x*y=-2
-- on to foo, whose new value is -2+7=5, and spit out -2*7 = -14.
-- Note: x is accumulated value, y is value of key; and same ordering
-- on output pair.
-- (-14,fromList [("bar",-1),("foo",5)])
p4 = mapAccum (\x y -> (x*y,x+y)) 1 m23

-- Find sum of all (original) values and increment each value:
p5 = mapAccum (\x y -> (x+y,y+1)) 0 m23

-- Find sum of all (original) values and increase each value by key length:
p6 = mapAccumWithKey (\x k y -> (x+y,y+length k)) 0 m23

-- Reverse every key, leave values alone.
-- ("oof",10),("rab",1)
-- Note: when f is monotonic, mapKeysMonotonic is probably more efficient
m25 = mapKeys reverse m24

-- Replace each key with its length; keep later when same key
-- (3,1) only
m26 = mapKeys length m25
-- (3,11) because we use (+) to combine 1 and 10 on key 3
m27 = mapKeysWith (+) length m25

-- Better way to do the above ...
sum0 = fold (+) 0 m24
-- To list the keys:
list0 = foldWithKey (\k v lst -> (k:lst)) [] m24

-- Conversion:
-- List of values
nums1 = elems m24
-- List of keys
list1 = keys m24
-- Set of keys
set1 = keysSet m24
-- Pair list of keys and values
plist1 = assocs m24

-- Lists. Vanilla conversions. For each of these,
-- there is a "fromAsc" or "toAsc" version, also
-- fromDistinctAscList. Predictions not checked!
list24 = toList m24
m28 = fromList (("baz",6):list24)

-- With function to combine values of same key:
m29 = fromListWith (+) [("bar",1),("bar",2)]

-- Ditto, depending on key too:
conv k x y = if k == "bar" then x+y else x*y
-- fromList [("bar",3),("foo",2)]
m30 = fromListWithKey conv [("bar",1),("bar",2),("foo",1),("foo",2)]

-- Derive submap by filtering
-- Keeps "foo",2 only
m31 = Data.Map.filter even m30
-- Ditto, looking at the key too:
keepPre k _ = k <= "e"
m32 = filterWithKey keepPre m30
-- Split a map into two:
(m30even,m30odd) = partition even m30
-- Can look at the key too:
(m30pre,m30post) = partitionWithKey keepPre m30

-- Just keep the even values, and double those up
dblEven v = if even v then Just (v,v+1) else Nothing
m33 = mapMaybe dblEven m30

squarePre k v = if k <= "e" then Just (v*v) else Nothing
m34 = mapMaybeWithKey squarePre m30

-- Split a map into two depending on values...
eitherEven v = if even v then Left v else Right v
(m30l,m30r) = mapEither eitherEven m30
-- ... or keys too
eitherFoo k v = if k == "foo" then Left v else Right v
(m30l2,m30r2) = mapEitherWithKey eitherFoo m30

-- Pre "e" and Post "e" keys
(m30l3,m30r3) = split "e" m30
(m30l4,val,m30r4) = splitLookup "bar" m30

tv1 = isSubmapOf m29 m30

-- Find indices of keys
idx1 = lookupIndex "baz" m30 :: Maybe Int
idx2 = lookupIndex "foo" m30 :: Maybe Int

-- Element indexed 1 (foo,2)
e1 = elemAt 1 m30

-- Min and max pairs of map
pair1 = findMin m30
pair2 = findMax m30

-- Delete (and find) min. Can do max too.
m35 = deleteMin m30
(minPair,m36) = deleteFindMin m30

-- Increment value of minimal key
m37 = updateMin (\x -> Just (x+1)) m30
m38 = updateMinWithKey (\k x -> Just (x+1)) m30

-- minView: rest of map, and minimal pair, in monad.
-- Note: documentation is out of date!
mx1 = minView m30 :: Maybe (Map [Char] Integer, ([Char], Integer))
mx2 = minView empty :: Maybe (Map [Char] Integer, ([Char], Integer))

-- run putStrLn on these trees ...
tree1 = showTree m30
-- elemShower produces a string from a key and a value
elemShower k v = k ++ " maps to " ++ (show v)
tree2 = showTreeWith elemShower False True m30
