module Main where

import qualified Data.ByteString.Char8 as B
import Control.Monad
import Char
import List

lif :: IO [B.ByteString]
lif = (liftM B.lines) (B.readFile "alirow.hs")

line :: B.ByteString
line = B.pack "A C+CT"

data Alirow = Alirow B.ByteString B.ByteString
  deriving Eq

ar = Alirow (B.pack "ACGA") (B.pack "5IY5")

instance Show Alirow where
  show = showAlirow

showAlirow ar =
  concatMap (\(a,b) -> [a,b]) (alirowToPairs ar)

alirowToPairs (Alirow nucs quals) = (B.zip nucs quals)

stringToAlirow str =
  let (str1,str2) = listToPair str
  in Alirow (B.pack str1) (B.pack str2)

-- not relevant here!
listToPairs [] = []
listToPairs (a:(b:t)) = (a,b):(listToPairs t)

listToPair [] = ([],[])
listToPair (a:(b:t)) =
  let (x,y) = listToPair t
  in (a:x,b:y)

isValidAlirow (Alirow nucs quals) =
  B.length nucs == B.length quals &&
  B.all isValidNucleotide nucs &&
  B.all isValidQuality quals &&
  eqTogether ' ' (B.zip nucs quals)

isValidNucleotide :: Char -> Bool
isValidNucleotide c = 
  c `elem` " -_=" ||
  (c >= '0' && c <= '9') ||
  let u = toUpper c
  in u `elem` "ACGTN"

isValidQuality ' ' = True
isValidQuality q = inRange (ord q) 33 126

inRange v lo hi = v >= lo && v <= hi

-- useful somewhere...
sameLength [] [] = True
sameLength (_:x) (_:y) = sameLength x y
sameLength _ _ = False

-- True if all pairs are either (v,v) or have no v.
eqTogether :: (Eq t) => t -> [(t, t)] -> Bool
eqTogether v pairs =
  allSameValues (v ==) pairs

-- True if for every pair (x,y) in pairs,
-- pred x == pred y.
allSameValues pred pairs =
  let pred2 (x,y) = pred x == pred y
  in all pred2 pairs

qualityValue ' ' = 94
qualityValue c | isValidQuality c = ord c - ord '!'

-- pairs is list of (nuc,qc) values, e.g. output
-- of alirowToPairs
consensusPair :: [(Char,Char)] -> (Char,Int)
consensusPair = consensusPair1 . qualitySumsOnPairs

consensusPair1 [pair] = pair
consensusPair1 ((v1,q1):(_,q2):_) = (v1,q1-q2)

qualitySums = qualitySumsOnPairs . alirowToPairs

qualitySumsOnPairs pairs =
  sortBy higherSnd $ qualitySums1 pairs []

higherSnd (_,x) (_,y) = compare y x

qualitySums1 [] alist = alist
qualitySums1 ((nuc,qc):pairs) alist1 =
  let q = qualityValue qc
  in qualitySums1 pairs (alistOp alist1 nuc (+q) q)

alistOp [] key _ dft = [(key,dft)]
alistOp ((key1,val1):rest) key pred _ | key1 == key =
  (key,pred val1):rest
alistOp (h:t) key pred dft =
  h:(alistOp t key pred dft)

------------------

data NucVal = A | C | G | T | Gap
  deriving (Eq,Ord,Show,Enum,Bounded)

-- Number of "nucleotide" values: allow gap too. Most convenient to
-- have it as a Double.
nNucVals :: Double
nNucVals = fromInteger $ toInteger $ (fromEnum (maxBound :: NucVal)) - (fromEnum (minBound :: NucVal)) + 1

-- Three ways to describe a nucleotide prob dist.
data NucDist = NoInfo              -- uniform distrib
             | NucQual NucVal Int -- Int is quality, >= 0
             | NucProbs Double Double Double Double -- Probs of A,C,G,T; any shortfall is gap
  deriving (Eq,Ord,Show)

minQualVal = 0
maxQualVal = 94

-- Given a NucDist and a value, return the probability of that value
getNucProb :: NucDist -> NucVal -> Double
getNucProb NoInfo _ = 1/nNucVals
getNucProb (NucQual nv nq) v =
    let eProb = toErrorProb nq
    in if (v == nv) then 1-eProb else eProb/(nNucVals-1)
getNucProb (NucProbs p _ _ _) A = p
getNucProb (NucProbs _ p _ _) C = p
getNucProb (NucProbs _ _ p _) G = p
getNucProb (NucProbs _ _ _ p) T = p
getNucProb (NucProbs pA pC pG pT) Gap = 1-(pA+pC+pG+pT)

maxErrorProb = 1-1/nNucVals :: Double

-- Quality value to error prob
toErrorProb :: Int -> Double
toErrorProb nq = 
    if nq <= minQualVal
    then maxErrorProb
    else if nq >= maxQualVal
    then 0
    else exp $ -0.1*(fromInteger $ toInteger nq)*(log 10)

-- Error prob to quality value
toQualVal :: Double -> Int
toQualVal ep =
    if ep <= 0
    then maxQualVal
    else if ep >= 1
    then minQualVal
    else round $ -10 * (log ep) / (log 10)

-- Canonicalise NucDist to NucQual form
toNucQual NoInfo = NucQual A 0
toNucQual (NucProbs pA pC pG pT) = 
    listToNucQual1 [(A,pA),(C,pC),(G,pG),(T,pT),(Gap,1-(pA+pC+pG+pT))]
toNucQual nd = nd

listToNucQual1 alist =
    let (v,p):_ = sortBy higherSnd alist
    in NucQual v (toQualVal (1-p))

-- Canonicalise NucDist to NucProbs form
toNucProbs NoInfo = 
    let p = 1/nNucVals
    in NucProbs p p p p
toNucProbs (NucQual nv nq) =
    let ep1 = toErrorProb nq
        ep = ep1/(nNucVals-1)
        tp = 1-ep1
    in case nv of
         A -> NucProbs tp ep ep ep
         C -> NucProbs ep tp ep ep
         G -> NucProbs ep ep tp ep
         T -> NucProbs ep ep ep tp
         Gap -> NucProbs ep ep ep ep
toNucProbs np = np

-- Given a NucDist, extract the most likely value ...

getNDVal nd =
    let (NucQual v _) = toNucQual nd
    in v

-- ... and its quality
getNDQual nd =
    let (NucQual _ q) = toNucQual nd
    in q

-- Combine two NucDists.
-- First two clauses are for efficiency
ndConsensus NoInfo nd = nd
ndConsensus nd NoInfo = nd
ndConsensus nd1 nd2 =
    let NucQual v1 q1 = toNucQual nd1
        NucQual v2 q2 = toNucQual nd2
    in if v1 == v2 
       then NucQual v1 (max q1 q2)
       else if q1 > q2
            then NucQual v1 (q1-q2)
            else NucQual v2 (q2-q1)

stringToNucDist "  "    = NoInfo
stringToNucDist [vc,qc] = NucQual (charToNucVal vc) (charToQualVal qc)

charToNucVal 'A' = A
charToNucVal 'C' = C
charToNucVal 'G' = G
charToNucVal 'T' = T
charToNucVal 'a' = A
charToNucVal 'c' = C
charToNucVal 'g' = G
charToNucVal 't' = T
charToNucVal '-' = Gap

charToQualVal ' ' = maxQualVal
charToQualVal c = min maxQualVal $ max minQualVal $ (ord c) - qualBaseVal

nucDistToString NoInfo = "  "
nucDistToString nd =
    let NucQual v q = toNucQual nd
    in [nucValToChar v, qualValToChar q]

nucValToChar v =
    if v == Gap 
    then '-' 
    else head $ show v

qualBaseChar = '!'
qualBaseVal = ord qualBaseChar              

qualValToChar q =
  if q >= maxQualVal
  then ' '
  else if q <= minQualVal
  then qualBaseChar
  else chr $ q+qualBaseVal
