module Utility where

import Data.Map as DM
import Data.Array as DA
import Data.List as DL
import Control.Monad as CM

-------------------------------------------------------------------------

fileToList f = CM.liftM lines $ readFile f

-------------------------------------------------------------------------

fileToLOL f = CM.liftM (Prelude.map words) $ fileToList f

-------------------------------------------------------------------------

fileToMap :: FilePath -> IO (Map String String)
fileToMap f = CM.liftM lolToMap $ fileToLOL f

lolToMap :: (Ord k) => [[k]] -> Map k k
lolToMap lol = 
  let lol2  = Prelude.filter (lengthGE 2) lol
      pairs = Prelude.map listToPair lol2
  in  DM.fromList pairs

lengthGE 0 _     = True
lengthGE l (_:t) = l>0 && lengthGE (l-1) t
lengthGE _ _     = False

-- Allow more than two elements in the list; ignore 3rd and subsequent.
listToPair (x:y:_) = (x,y)

-------------------------------------------------------------------------

fileToArray :: FilePath -> IO (Array Int String)
fileToArray f = do CM.liftM (listToArray 1) (fileToList f)

listToArray lo lst = DA.listArray (indexPair lst lo) lst

indexPair lst lo = (lo,lastIndex lst lo)

-- i is index of first member...
lastIndex [] i = i-1
lastIndex (_:t) i = lastIndex t (i+1)

-------------------------------------------------------------------------

fileToMatrix :: FilePath -> IO (Array (Int, Int) String)
fileToMatrix f = CM.liftM (lolToMatrix (1,1)) (fileToLOL f)

-- Assumes every line has the same number of words.
lolToMatrix :: (Int,Int) -> [[a]] -> Array (Int, Int) a
lolToMatrix loPair lol =
  let iLOL   = indexLOL loPair lol
      inds   = indexPair2 loPair lol
  in DA.array inds iLOL

indexPair2 (lo1,lo2) lol =
  let hi1 = lastIndex lol lo1
      hi2 = lastIndex (head lol) lo2
  in ((lo1,lo2),(hi1,hi2))

-- And for good measure:
indexPair3 (lo1,lo2,lo3) lolol =
  let hi1 = lastIndex lolol lo1
      lol = head lolol
      hi2 = lastIndex lol lo2
      hi3 = lastIndex (head lol) lo3
  in ((lo1,lo2,lo3),(hi1,hi2,hi3))

--indexLOL _ [] = []
--indexLOL pr (h:t) = (indexLOL1 pr h) ++ (indexLOL (applyFst (+1) pr) t)

--indexLOL1 _ [] = []
--indexLOL1 pr (h:t) = (pr,h) : (indexLOL1 (applySnd (+1) pr) t)

-- Works, but not so nice:
-- indexLOL1Fold i1 i2 lst = 
--   let func j2 x = (j2+1,((i1,j2),x))
--   in snd $ DL.mapAccumL func i2 lst

-- e.g.: indexLOLGen (3,4) ["abc","pqr"]
--       [((3,4),'a'),((3,5),'b'),((3,6),'c'),((4,4),'p'),((4,5),'q'),((4,6),'r')]

indexLOL =  applyGen indexLOL1  (++) (applyFst (+1)) []

indexLOL1 = applyGen (curry id) (:)  (applySnd (+1)) []

-----------------------------------------------------------------------------

-- Functions for modifying part of a pair:

applyFst f = applyBoth f id

applySnd g = applyBoth id g

applyBoth f g (i,j) = (f i,g j)

-- Generalization of mapAccumR. Args are:
--   transform: given acc and head of list,
--                return item to be included (somehow) in result
--   combine:   combines result of transformer and
--                result of applyGen on tail; e.g. (++)
--   advance:   advances state to new state
--   base:      value for end of list
--   state:     aux info to pass along
--   list:      list to consume

applyGen _ _ _ base _ [] = base
applyGen transform combine advance base state (h:t) = 
  let hStuff   = transform state h 
      newState = advance state
      tStuff   = applyGen transform combine advance base newState t
  in combine hStuff tStuff

-------------------------------------------------------------------------

-- We should do this with Data.ByteString instead ...

fastaFileToMap f = 
  CM.liftM (fastaLinesToMap "" "" DM.empty) (fileToList f)

fastaLinesToMap key buf map [] = 
  insertIfKey key buf map
fastaLinesToMap key buf map (('>':rest):lns) =
  let newKey = (head $ words rest)
      newMap = (insertIfKey key buf map)
  in fastaLinesToMap newKey "" newMap lns
fastaLinesToMap key buf map (line:lns) =
  fastaLinesToMap key (buf ++ line) map lns

insertIfKey "" _ = id
insertIfKey key val = DM.insert key val

