module Utility where

import qualified Data.Map as DM
import qualified System.IO as IO
import qualified Text.Printf as TP
import qualified Control.Monad as CM
import Perl2HS as PH

-- Given a file, return string-to-string hash.
readSSFile = readAndParseToMap keyAndVal

-- Given a file, return string-to-list-of-strings hash.
readSLSFile = readAndParseToMap keyAndVals

keyAndVal (k:(v:_)) = Just (k,v)
keyAndVal _ = Nothing

keyAndVals (k:v) = Just (k,v)
keyAndVals _ = Nothing

readAndParseToMap splitter f = 
  do s <- IO.readFile f
     return $ foldr (parseLine splitter) DM.empty (lines s)
    
parseLine splitter s m =
  case (splitter . words . takeWhile (not . (==) '#')) s
  of Just (k,v) -> DM.insert k v m
     Nothing -> m

-- Given a file, return list of chunks (successive
-- lists of lines, perhaps omitting some)
splitLinesInFile splitter f =
  do s <- IO.readFile f
     return $ splitLines splitter (lines s)

splitLines splitter [] = []
splitLines splitter lns =
  case splitter lns
  of Just (chunk,rest) -> chunk:(splitLines splitter rest)
     Nothing -> splitLines splitter $ tail lns

splitToEach (h:t) = Just ([h],t)

readFastq = splitLinesInFile splitToFastq

splitToFastq lns = Just (splitAt 4 lns)

splitToFasta (h:t) =
    let isFastaAnnoLine line = (line /= []) && (head line) == '>'
    in if isFastaAnnoLine h
       then let (seqLines,rest) = break isFastaAnnoLine t
            in Just (h:seqLines,rest)
       else Nothing -- discard first line



(?) test (p,q) = if test then p else q

-- Return longest tail of arg2 whose elements that all satisfy f
takeTailWhile f = reverse . takeWhile f . reverse

-- Return shortest init of arg2 such that all the rest satisfy f
dropTailWhile f = reverse . dropWhile f . reverse

-- (##) f g x y = f (g x y)

findWithDefaultGen def =
  PH.fromString ## DM.findWithDefault (PH.toString def)

-- Fractional (Float/Double) quotient of any two Integrals.
fdiv a b = fromInteger (toInteger a) / fromInteger (toInteger b)

ifDo test act1 act2 =
  do val <- test
     val ? (act1,act2)

doPair fx fy (x,y) =
  do {fx x; fy y}

leadingZeros :: Int -> Int -> String
leadingZeros n v =
  let fmt = TP.printf "%%%dd" n -- or: "%" ++ (PH.toString n) ++ "d"
   in subst ' ' '0' $ TP.printf fmt v

subst v1 v2 = map (\x -> if x == v1 then v2 else x)

linesInFile = (CM.liftM lines) . IO.readFile

existenceOK yesList noList = 
   allM existenceOK1 pairs
   where pairs = (map (\x -> (x,False)) noList) ++ (map (\x -> (x,True)) yesList)

existenceOK1 :: (String,Bool) -> IO Bool
existenceOK1 (name,want) = do
  sz <- ms name
  if (sz > 0) == want
   then return True
   else do
     let msg = if want then "Missing or empty" else "Already exists"
     putStrLn $ msg ++ ": " ++ name
     return False

allM :: (Monad m) => (t -> m Bool) -> [t] -> m Bool
allM _ [] = return True
allM act (h:t) = do
  ok <- act h
  if ok
   then allM act t
   else return False

anyM :: (Monad m) => (t -> m Bool) -> [t] -> m Bool
anyM _ [] = return False
anyM act (h:t) = do
  ok <- act h
  if ok
   then return True
   else anyM act t

mayRead s =
  case (readsPrec 0 s) of
    [(val,"")] -> Just val
    _          -> Nothing

justList :: [Maybe a] -> [a]
justList [] = []
justList (Nothing:t) = justList t
justList (Just h:t)  = h : justList t
         
