module Main where

import System (getArgs)
import System.IO
import System.Directory
import System.Posix.Files
import System.FilePath.Posix
import IO
import Data.List
import qualified Data.Char as DC
import qualified Data.Map as DM
import Text.Printf
-- Data.ByteString is Word8 i.e. binary. Could use
-- Data.ByteString.Char8 for char mode.
import qualified Data.ByteString as BS
import qualified Perl2HS as PH

---- Constant part ----

-- Default (maximum) length of a chunk
defaultChunkLength = 1538330
-- USB disc (MP3 player)
usbDisc = "/media/disk/radio"

---- Imperative part ----

-- Switches: -n numberOfChunks or -s chunkSize; -d tgtDir

main = do args <- getArgs
          let (swMap,rest,errs) = PH.getopt "n:s:" args
          if errs == []
            then do mapM_ (splitFile swMap) rest
            else do mapM_ printf errs

-- getArgs >>= mapM_ splitFile

splitFile swMap file = 
  ".mp3" `isSuffixOf` file ?
    (splitFileIfNeeded swMap file,
     printf "%s: does not end in .mp3\n" file)

splitFileIfNeeded swMap file =
  do let top1 = DM.findWithDefault usbDisc 'd' swMap
     exist <- fileExist top1 
     let top = exist ? (top1 , "SPLIT")
     let (dir,b) = destinationComponents top file
     ifDo (fileExist dir)
          (printf "%s: already exists\n" dir)
          (splitFileReally swMap file b dir)

splitFileReally swMap file b dir =
  do createDirectoryIfMissing True dir
     fsz <- PH.getFileSize file
     let chsz = decideChunkSize swMap fsz
     let pfx = dir ++ "/" ++ b ++ "_"
     BS.readFile file >>= splitContents chsz pfx 0
     printf "%s: created\n" dir
   
splitContents chsz pfx idx bytes = 
  if bytes == BS.empty
    then return ()
    else doPair
           (BS.writeFile $ targetMP3File pfx idx)
           (splitContents chsz pfx $ succ idx)
           (BS.splitAt chsz bytes)

---- Functional part ----

{- swMap may have key 'n' giving desired number of chunks,
   nChunks is that if it exists, or 's' for chunk size.
   Otherwise:
     fsz is file size
     nchDft is number of chunks implied by default chunk
       length, rounded up so chunks are no bigger than that.
     Return chunk size determined by fsz and nChunks.
-}

decideChunkSize swMap fsz =
  let chunkSize = findWithDefaultGen defaultChunkLength 's' swMap
      nchDft = ceiling (fdiv fsz chunkSize)
      nChunks = findWithDefaultGen nchDft 'n' swMap
   in ceiling (fdiv fsz nChunks)

destinationComponents top file =
  let base = takeBaseName file
      bPfx = preDigitPrefix base
      dir = top ++ "/" ++ bPfx ++ "/" ++ base
   in (dir,base)

preDigitPrefix str =
  let pfx1 = takeWhile (not . DC.isDigit) str
   in dropTailWhile (== '_') pfx1

-- Not needed here ...
-- buildPath = PH.intercalate "/"

targetMP3File pfx idx = pfx ++ (leadingZeros 2 idx) ++ ".mp3"

---- Utility part ----

(?) test (p,q) = if test then p else q

-- Return longest tail of arg2 whose elements that all satisfy f
takeTailWhile f = reverse . takeWhile f . reverse

-- Return shortest init of arg2 such that all the rest satisfy f
dropTailWhile f = reverse . dropWhile f . reverse

(##) f g x y = f (g x y) 

findWithDefaultGen def =
  PH.fromString ## DM.findWithDefault (PH.toString def)

-- Fractional (Float/Double) quotient of any two Integrals.
fdiv a b = fromInteger (toInteger a) / fromInteger (toInteger b)
             
ifDo test act1 act2 =
  do val <- test
     val ? (act1,act2)

doPair fx fy (x,y) = 
  do {fx x; fy y}

leadingZeros :: Int -> Int -> String
leadingZeros n v =
  let fmt = printf "%%%dd" n -- or: "%" ++ (PH.toString n) ++ "d"
   in subst ' ' '0' $ printf fmt v

subst v1 v2 = map (\x -> if x == v1 then v2 else x)

     