module Main where

import qualified Data.ByteString.Char8 as B
import qualified Data.Map as M
import System (getArgs)
import System.Console.GetOpt
import Char
import Data.List
import Data.Ord
import Control.Monad
import Text.Printf

-- http://www.haskell.org/haskellwiki/Tutorials/Programming_Haskell/Introduction

-- Compile with e.g.
--   ghc -o vocab vocab.hs

-- Impure bit

main = do args <- getArgs
          let (ty,isRev,isProp,files) = getVocabOpts args 
          lines <- linesInFiles files
          let voc = getVocab1 ty isRev lines
          let fac = (if isProp then 1.0 / sumFunc snd voc else 1.0) :: Float
          mapM_ (printPair isRev fac) voc
          return ()

-- Given a list of files, return a list of all the lines (bytestrings) in them.
linesInFiles :: [FilePath] -> IO [B.ByteString]
linesInFiles []    = (liftM B.lines) B.getContents
linesInFiles files = (liftM concat) (mapM linesInFile files)

-- Given a file, return a list of the lines (bytestrings) in it.
linesInFile :: FilePath -> IO [B.ByteString]
linesInFile f = (liftM B.lines) (B.readFile f)

-- Print w and c, with the first (Bool) arg determining whether w comes first.
-- If second arg is 1, print without decimal point.
-- c is forced to have same type as f, i.e. Float, so we print with %.0f.

printPair False 1 (w,c) = printf "%.0f %s\n" c w
printPair True  1 (w,c) = printf "%s %.0f\n" w c
printPair False f (w,c) = printf "%8.6f %s\n" (c*f) w
printPair True  f (w,c) = printf "%s %8.6f\n" w (c*f)

---------------------------------- Pure from here --------------------------------

-- Option processing

optSpec = [Option "c" [] (NoArg 'c') "characters",
           Option "l" [] (NoArg 'l') "lines",
           Option "w" [] (NoArg 'w') "words",
           Option "r" [] (NoArg 'r') "reverse",
           Option "p" [] (NoArg 'p') "proportion"]

getVocabOpts :: [String] -> (Char, Bool, Bool, [String])
getVocabOpts args =
  let (opts,files,_) = getOpt RequireOrder optSpec args in
    (itemType opts, -- 'c', 'l' or (default) 'w'
     elem 'r' opts, -- whether to reverse
     elem 'p' opts, -- whether to do proportions
     files)         -- return the files

itemType :: [Char] -> Char
itemType chars = findFirstOr "clw" chars 'w'

-- General-purpose code

-- Turn first element of pair from bytestring to string,
-- leaving second unchanged.
fstToString :: (B.ByteString, t) -> ([Char], t)
fstToString (bs,n) = (B.unpack bs,n)

-- Apply f to everything in the list and sum the results.
sumFunc :: (Num b) => (a -> b) -> [a] -> b
sumFunc f = sum . (map f)

-- Return Just c where c is the first element of lst that
-- occurs in vals; or Nothing if no such.
findFirst :: (Eq a) => [a] -> [a] -> Maybe a
findFirst vals lst = find ((flip elem) vals) lst

-- Like find first, but no Maybe; return default instead.
findFirstOr :: (Eq t) => [t] -> [t] -> t -> t
findFirstOr vals lst dft = justOr (findFirst vals lst) dft

-- Whether a character is visible (printable and not whitespace)
isVisible :: Char -> Bool
isVisible c = isPrint c && not (isSpace c)

-- Un-Maybe, using default d for Nothing.
justOr :: Maybe t -> t -> t
justOr (Just x) _ = x
justOr _ d = d

-- List of something.
singleton :: t -> [t]
singleton x = [x] 

-- Map func to the LIST of items returned by filtering cond on 
-- the members of bs.
mapIf :: (Char -> Bool) -> (Char -> b) -> B.ByteString -> [b]
mapIf cond func bs =
  map func (B.unpack (B.filter cond bs))

-- Vocab-specific code.

-- A vocabulary is a map from bytestrings to floats.
type Vocab = M.Map B.ByteString Float

-- Get count of word in vocab map, with 0 as default.
getCount :: B.ByteString -> Vocab -> Float
getCount = M.findWithDefault 0

-- Given an item type ('c','w' or 'l'), whether to sort by key
-- as opposed to count, and a list of lines, get the vocab as a
-- list of (ByteString,Count) tuples.
getVocab :: Char -> Bool -> [B.ByteString] -> [(B.ByteString, Float)]
getVocab ty isRev lines = sortBy (vocabSorter isRev) (M.toList (createVocab ty lines))

-- As getVocab, but convert the ByteStrings to [Char]s.
getVocab1 :: Char -> Bool -> [B.ByteString] -> [([Char], Float)]
getVocab1 ty isRev lines = map fstToString $ getVocab ty isRev lines

-- Vocab-item sorting function. If isRev is false, sort by decreasing ("flip")
-- count ("snd"); else sort by key.
vocabSorter :: (Ord a, Ord b) => Bool -> (a, b) -> (a, b) -> Ordering
vocabSorter False = flip $ comparing snd
vocabSorter True  = comparing fst

-- Given an item type and a list of lines, create a vocab map.
createVocab :: Char -> [B.ByteString] -> Vocab
createVocab ty = foldl (addToVocab (splitter ty)) M.empty

-- Depending on the item type, return a function that splits a bytestring
-- (line) into a list of items.
splitter :: Char -> B.ByteString -> [B.ByteString]
splitter 'w' = B.words
splitter 'l' = singleton
splitter 'c' = mapIf isVisible B.singleton

-- Given a function splitting a line into bytestrings, a vocab, and a
-- line to split, return a new vocab with the bytestring counts incremented.
addToVocab :: (t -> [B.ByteString]) -> Vocab -> t -> Vocab
addToVocab spl voc line =
   foldl addWordToVocab voc (spl line)

-- Given a vocab and a bytestring, return an augmented bytestring from
-- incrementing the count for that word.
addWordToVocab :: Vocab -> B.ByteString -> Vocab
addWordToVocab voc word =
  M.insert word (getCount word voc + 1) voc
