package G1K::Calibrate;

use Exporter;
@ISA = qw(Exporter);

@EXPORT = qw(sampleFastq calibrate);

use G1K::G1K;
use Utility;
use Bsub;
use File::Basename;
use Cwd;

use strict; 

my $BIN = $ENV{G1K} . "/bin";
my $MAP2QMAP = "$BIN/map2qMapFile.pl";
my $SNP126 = "$ENV{G1K}/misc/dbSNP/snp126.snp";

# g1k.pl sam command
sub sampleFastq {
    my ($anaType,$proj,$indiv,$lib,$lane) = @_;
    # Find the source (DATA/...) and target (v1/...) directories.
    my ($srcD,$tgtD) = G1K::G1K::srcAndTgtDirs(@_);
    # Bail out if this is a bad lane.
    return unless (G1K::RunMaq::inLaneList($tgtD));
    # See if we've already got a quality map file. If so, we don't
    # do sampling, even if there's no sample dir. This allows us
    # to delete sample dirs when we're done.
    return if (-s "$tgtD/qualmapBayesian.txt");
    # Check we can locate the fastq file to recalibrate from.
    my $src = G1K::G1K::laneFastqLink($srcD);
    return unless ($src);
    # We'll try and make $tgt from $src.
    my $tgt = "$tgtD/sample/30m.fastq";
    my $ms = makeStatus([$src],[$tgt]);
    return unless ($ms > 0 && stamp($tgt));
    my $fh = openToRead($src);
    # nrSum will be undefined if no fastqcheck; then take the first
    # (good) nWanted reads, and hope they're typical of the rest.
    my ($nrSum,$rl) = fastqReadCount($src);
    my $nrWanted; # number of reads we want
    if (defined $nrSum) {
	# $nrWanted is 30M/readLength, but if that's more than nrSum,
	# we can only have nrSum. NOTE: 30M/readLength is chosen so that
	# we have an upper limit on the time taken by g1k.pl mca,
	# regardless of read length. But it might be better to have a
	# fixed number of reads, as that would tend to equalise the
	# count per cell used for recalibration.
	$nrWanted = min($nrSum,30000000/$rl);
	my $pc = $nrWanted*100/$nrSum; # percent we'll sample:
	report(sprintf("%s: creating with %d%%",$tgt,$pc));
    } else {
	# we'll take the first nrWanted = 30M/readLength
	$nrSum = $nrWanted = 30000000/$rl;
	report(sprintf("%s: creating with first %d good reads",$tgt,$nrSum));
    }
    my $gh = openToWrite($tgt);
    while ($nrSum >= 0) {
	# At any point in this loop, $nrSum is the number of reads left
	# in the file, and nrWanted is the number we still want, so
	# this definition of $want samples uniformly. If $nrSum was
	# initially undefined, $nrSum <= $nrWanted throughout and
	# $want will always be 1.
	my $want = (rand()*$nrSum <= $nrWanted);
	if ($want) {
	    # Read the fastq entry (4 lines)
	    my @quad;
	    foreach my $i (0 .. 3) {
		$quad[$i] = <$fh>;
	    }
	    # Read is good if the first 36 bases contains neither N
	    # nor a poly-A. But for paired-end we should probably use
	    # both halves. Occasionally this leads to problems when
	    # the first run was good but the second bad.
	    my $sq = substr($quad[1],0,36);
	    unless ($sq =~ /N/ || $sq =~ /AAAAAAAAAAAAAAA/) {
		print $gh @quad;
		$nrWanted --;
	    }
	} else {
	    # Read and discard 4 lines.
	    foreach my $i (0 .. 3) {
		<$fh>;
	    }
	}
	$nrSum --;
    }
    unstamp($tgt);
    report();
}

# Return the number of reads in a fastq file or files and the read (pair) length.
sub fastqReadCount {
    my @files = @_;
    my ($sum,$rl);
    foreach my $f (@files) {
	$f = readlink($f) if (-l $f);
	my $fc; # find the fastqcheck file corresponding to a fastq
	foreach my $suf (qw(fastq fastq.gz fq.gz)) {
	    if ($f =~ /$suf$/) {
		($fc = $f) =~ s/$suf$/fastqcheck/;
		last if (-s $fc);
	    }
	}
	if ($fc) {
	    # If we got it, read off the number of reads directly:
	    # first field of first line
	    my $fh = openToRead($fc);
	    my $line = <$fh>;
	    my ($nr) = split(" ",$line);
	    $sum += $nr;
	}
	if (!$rl) {
	    # Otherwise, get the read length from the first read in the
	    # file itself, and leave $sum unspecified.
	    my $fh = openToRead($f);
	    <$fh>;
	    chomp (my $line = <$fh>);
	    $rl = length($line);
	}
    }
    return ($sum,$rl);
}

# g1k.pl cal ...
# Use the output of g1k.pl mca (i.e. sample/aln.map) to derive a quality
# map file qualmapBayesian.txt, then apply that to the uncalibrated fastq
# file to derive recalibrated.fastq.gz.

sub calibrate {
    my ($anaType,$proj,$indiv,$lib,$lane) = @_;
    # Find source (DATA) and target (v1) dirs as usual, and return if this
    # is a bad lane.
    my ($srcD,$dir) = G1K::G1K::srcAndTgtDirs(@_);
    return unless (G1K::RunMaq::inLaneList($dir));
    # Input files needed
    my $src = "$dir/sample/aln.map";
    my $qmf = "$dir/qualmapBayesian.txt"; # or just qualmap.txt when fix done...
    my $ref = "$dir/ref.bfa";
    # We want to create this:
    my $tgt = "$dir/recalibrated.fastq.gz";
    # If we've already got the qualmap and there's a recalibrated fastq
    # file in the repository that's newer than the qualmap, assume we
    # don't need to do anything.
    if (-s $qmf && ! -s $tgt) {
	my $existing = G1K::RunMaq::chooseRecalibratedFastq($srcD,$dir);
	if (newer($existing,$qmf)) {
	    report("$dir: not creating recalibrated.fastq.gz because $existing is newer than qualmapBayesian.txt",1);
	    return;
	}
    }
    # Link to the right reference if required. It ought to be in the
    # sample directory.
    if (! -s $ref && -s "$dir/sample/ref.bfa") {
	link("$dir/sample/ref.bfa",$ref);
    }
    # If $qmf exists, we only require that; otherwise, we require $src and $ref.
    my $pReq = -s $qmf ? [$qmf] : [$src,$ref];
    my $ms = makeStatus($pReq,[$tgt]);
    if ($ms > 0 && stamp($tgt)) {
	my @cmds;
	# If $qmf doesn't exist, we have to create it.
	unless (-s $qmf) {
	    # Check we mapped enough reads to make it possible to create a
	    # reasonably accurate map.
	    if (enoughReadsMapped($src,1)) {
		my (undef,$isSingle) = G1K::G1K::getReadLength($srcD);
		my $single = $isSingle ? '-s' : '';
		# "-p $SNP126" masks out dbSNP positions.
		# Create qualmapLH.txt, then fix the recalibration with makeQualitiesBayesian.pl.
		# We ought to integrate the fix into $MAP2QMAP really.
		my $qmlh = "$dir/qualmapLH.txt";
		my $mqb = "$ENV{G1K}/bin/makeQualitiesBayesian.pl";
		unless (-s $qmlh) {
		    push @cmds,"$MAP2QMAP $single -p $SNP126 $src $ref > $qmlh";
		}
		push @cmds,"$mqb $qmlh > $qmf";
	    }
	}
	# If we either plan to make $qmf or already have it, we'll then create $tgt.
	if (@cmds || -s $qmf) {
	    my $fqLink = G1K::G1K::laneFastqLink($srcD);
	    die("$srcD: cannot find .fastq, .fastq.gz or .fq.gz link") unless ($fqLink);
	    # Apply $qmf to the uncalibrated data to create calibrated (and then remove the touchfile)
	    my $aqm = "$ENV{G1K}/bin/applyQualityMap.pl";
	    push @cmds,"$aqm $fqLink $qmf $tgt";
	    # Submit the job...
	    my $id = bsub($tgt,join('; ',@cmds));
	    report("submitted job $id for $tgt",1);
	} else {
	    # If we don't do anything, e.g. because not enough reads mapped, just unstamp.
	    unstamp($tgt);
	}
    }
}

sub enoughReadsMapped {
    my ($map,$rep) = @_;
    # Look for mapped-read counts in the aln.stat file corresponding to $map.
    (my $stat = $map) =~ s/map$/stat/;
    if (-s $stat) {
	my $n = 0;
	my $fh = openToRead($stat);
	while (<$fh>) {
	    if (/mapped [SP]E reads: (\d+) /) {
		$n += $1;
	    }
	}
	# Assume we need at least 100,000 reads for recalibration to be accurate.
	if ($n >= 100000) {
	    return 1;
	} else {
	    report("$map: only $n reads mapped, too few to recalibrate",1) if ($rep);
	    return 0;
	}
    } else {
	report("$map: aln.stat file missing or empty, not recalibrating",1) if ($rep);
	return 0;
    }
}

1;
