package Alignment;

use Carp;

use strict;

# Format of line: one of
# ALIGNMENT  739   Y55-10a02.p1k chr13        1      883    625419    624524   C     883 96.04 1061
# ALIGNMENT  739   Y55-10a02.p1k chr13      883        1    625524    624419   C     883 96.04 1061
#             0       1            2          3       4         5       6      7      8      9   10
# And we may have at the end:
#   status (11)
#   cluster number (12)
#   cigar (13)
#   loTrim (14)
#   hiTrim (15)
# And added later:
#   querySeq (16)
#   queryQual (17)
#   sbjctSeq (18)
#   sbjctQual (19)
# cigarString (20) [for mapping to underlying chr, NOT inside the alignment!]
# underlyingStart (21), for use in CreateReadLayout
# dangleDirection (22), L or R (or undef)
# special (23), reason for "special acceptance" on extra iteration

my %CIRC = (chrm => 85779);

# Allow several transposons (not present in current strain) to be spanned, effectively...
my $MAXIMUM_READ_SPAN = 25000;

sub new {
    my ($class,@a) = @_;
    my $pOM;
    if ($a[0] =~ /^ALIGNMENT/) {
	my $line = $a[0];
	$pOM = $a[1]; # may be undef
	my $force = $a[2];
	@a = split(/\s+/,$line);
	shift @a; # to discard "ALIGNMENT" marker
	# Separate out start point and cigar string
	my @startAndCig;
	if ($a[11] eq '@') {
	    my $start = $a[12];
	    my $cig = join(" ",@a[13..$#a]);
	    foreach my $k (11 .. $#a) {
		$a[$k] = undef;
	    }
	    $a[20] = $cig;
	    $a[21] = $start;
	}
	# separate out "special" suffix, which should
	# start lower-case.
	if ($a[10] =~ /^(.*)\|([a-z].*)$/) {
	    ($a[10],$a[23]) = ($1,$2);
	}
	if ($a[10] =~ /^(.*)\|(.*),(.*)$/) {
	    # separate out the "trim"
	    ($a[10],$a[14],$a[15]) = ($1,$2,$3);
	} elsif ($a[10] =~ /^(.*)\|([LR])$/) {
	    # separate out the dangle direction
	    ($a[10],$a[22]) = ($1,$2);
	}
	# Assume we only need these checks when we're constructing an alignment
	# from an "ALIGNMENT" line rather than a preexisting array...
	mayReorderCoordinates(\@a,$line);
	$a[2] = normalizeChromosomeName($a[2]); # just in case
	unless ($force) {
	    return undef if (isRedundantRange($a[2],$a[5],$a[6],$pOM));
	}
    }
    unless (defined $a[20]) {
	$a[20] = decideCigarString($pOM,$a[2],$a[5],$a[6]);
    }
    $a[5] = translateOffset($a[2],$a[5],$pOM);
    $a[6] = translateOffset($a[2],$a[6],$pOM);
    my $self = bless \@a,$class;
    return $self;
}

sub mayReorderCoordinates {
    my ($pA,$line) = @_;
    # Swap rlo and rhi if orientation is C
    if ($pA->[7] eq 'C') {
	my $tmp = $pA->[3];
	$pA->[3] = $pA->[4];
	$pA->[4] = $tmp;
    }
    # Check ordering is now OK. If score is zero, we may
    # have either a3=a4+1 or a5=a6+1 (not both). Otherwise,
    # we need a3<=a4 and a5<=a6.
    if ($pA->[0] > 0) {
	croakIfGreater($pA->[3],$pA->[4],$line);
	croakIfGreater($pA->[5],$pA->[6],$line);
    } elsif ($pA->[3] >= $pA->[4]+1) {
	croakIfGreater($pA->[3],$pA->[4]+1,$line);
	croakIfGreater($pA->[5],$pA->[6],$line);
    } elsif ($pA->[5] == $pA->[6]+1) {
	croakIfGreater($pA->[3],$pA->[4],$line);
	croakIfGreater($pA->[5],$pA->[6]+1,$line);
    }
}

sub croakIfGreater {
    my ($x,$y,$line) = @_;
    if ($x > $y) {
	confess("Bad coordinate ordering: $x > $y in $line");
    }
}

sub normalizeChromosomeName {
    my ($chr) = @_;
    $chr =~ s/.*\|(.*)\|.*/\1/;
    $chr =~ s/.*contig=([a-z0-9]*).*/\1/;
    return $chr;
}

sub chr {
    my ($self,$remove) = @_;
    my $chr = $self->[2];
    if ($remove) {
	my @a = split(/\./,$chr);
	if (@a == 1) {
	    return $chr;
	} elsif ($remove == 2) {
	    return $a[1]; # chr name only, no strain name or aux ID
	} else {
	    return "$a[0].$a[1]"; # remove aux ID
	}
    } else {
	return $chr;
    }
}

sub isRedundantRange {
    my ($chr,$clo,$chi,$pOM) = @_;
    # Can only be redundant if part of an aux sequence ...
    # And we insist on at least a 10-base overlap of some break point.
    return ($pOM && $chr =~ /\..*\./) ? $pOM->failsToSpanBreak($chr,$clo-1,$chi-1,10) : 0;
}

# clo always < chi EXCEPT when zero score

sub clo {
    my ($self,$level) = @_;
    return $self->[5][$level];
}

sub chi {
    my ($self,$level) = @_;
    return $self->[6][$level];
}

sub decideCigarString {
    my ($pOM,$chr,$lo,$hi) = @_;
    if ($pOM) {
	return join(" ",reduceCigarString($pOM->cigarString($chr,$lo,$hi)));
    } else {
	return 'M' . ($hi-$lo+1);
    }
}

sub reduceCigarString {
    my (@cs) = @_;
    # Remove deletions from either end
    while (@cs && $cs[0] =~ /^D/) {
	shift @cs;
    }
    while (@cs && $cs[-1] =~ /^D/) {
	pop @cs;
    }
    my @ans;
    for (my $i=0; $i<=$#cs; $i++) {
	if (@ans && substr($ans[-1],0,1) eq substr($cs[$i],0,1)) {
	    my $sum = substr($ans[-1],1) + substr($cs[$i],1);
	    substr($ans[-1],1) = $sum;
	} else {
	    push @ans,$cs[$i];
	}
    }
    return @ans;
}

# Unwise to use this to set!
# NOTE: the cigar string ignores trims, but is modified when trims
# are applied.

sub cigarString {
    my ($self,$val) = @_;
    return setOrGet($self,20,$val);
}

sub setOrGet {
    my ($self,$i,$val) = @_;
    $self->[$i] = $val if (defined $val);
    return $self->[$i];
}

# Returns [offInAux,offInStr,offInGlobal], truncated at
# end if mapping not available.
sub translateOffset {
    my ($chr,$off,$pOM) = @_;
    my @comps = split(/\./,$chr);
    my @ans;
    if (@comps == 3) {
	# it's an aux sequence
	$ans[0] = $off;
	if ($pOM) {
	    my $off1 = $pOM->resolution($chr,$off);
	    if (defined $off1) {
		$ans[1] = $off1;
		my $off2 = $pOM->resolution("$comps[0].$comps[1]",$off1);
		if (defined $off2) {
		    $ans[2] = $off2;
		} else {
		    # Or maybe it's not so serious...
		    die("Cannot translate $chr $off");
		}
	    }
	}
    } elsif (@comps == 2) {
	# it's a strain sequence
	@ans = ($off,$off);
	if ($pOM) {
	    my $off2 = $pOM->resolution($chr,$off);
	    if (defined $off2) {
		$ans[2] = $off2;
	    } else {
		1;
	    }
	}
    } else {
	# it's a global sequence
	@ans = ($off,$off,$off);
    }
    return \@ans;
}

sub addCigar {
    my ($self,$fh) = @_;
    my $chi = $self->chi;
    my ($qseq,$sseq);
    while (1) {
	<$fh>; # blank
	chomp (my $qLine = <$fh>);
	my @a = split;
	$qseq .= $a[2];
	<$fh>;
	chomp (my $sLine = <$fh>);
	@a = split;
	$sseq .= $a[2];
	last if ($a[3] == $chi); # which it should!
    }
    my $kMax = length($sseq)-1;
    my $k=0;
    # State says which of query or subject has non-gap, or 'm' if both do:
    my ($qStart,$sStart,$state,$len,@cigar);
    while ($k <= $kMax) {
	my $st = 'm';
	if (substr($qseq,$k,1) eq '-') {
	    $st = 's';
	} else {
	    $st = 'q';
	}
	if ($st eq $state) {
	    $len++;
	} else {
	    push @cigar,[$len,$state] if ($len);
	    $len = 0;
	    $state = $st;
	}
    }
    push @cigar,[$len,$state] if ($len);
    $self->[13] = \@cigar;
}

# alignment length

sub alen {
    my ($self) = @_;
    return $self->[8];
}

# Apply trim if any

sub applyTrim {
    my ($self) = @_;
    my $lt = $self->loTrim;
    return unless (defined $lt);
    my $ht = $self->hiTrim;
    my $dALen = 0;
    $dALen += applyTrim1($self,$lt,1,1);
    $dALen += applyTrim1($self,$ht,-1,-1);
    $self->[14] = $self->[15] = undef;
    $self->[0] -= ($lt+$ht); # assume one point per base matched
    $self->[8] -= $dALen; # alignment length
}

sub loTrim {
    my ($self,$val) = @_;
    return setOrGet($self,14,$val);
}

sub hiTrim {
    my ($self,$val) = @_;
    return setOrGet($self,15,$val);
}

# cEnd and rEnd: 1 means low end, -1 means high end.
sub applyTrim1 {
    my ($self,$n,$cEnd,$rEnd) = @_;
    return unless ($n > 0);
    my $nq = $n; # guess
    if ($self->sbjctSeq) {
	# No. of alignment positions to remove from $self->sbjctSeq to get $n nucleotides:
	my $len = removeForNucleotides($n,$self->sbjctSeq,$cEnd);
	(my $remQS = getFromEnd($self->querySeq,$len,$rEnd)) =~ tr/-//;
	$nq = length($remQS); # proper value -- usually same as $n
	foreach my $i (16 .. 19) {
	    my $end = $i <= 17 ? $rEnd : $cEnd;
	    $self->[$i] = removeAtEnd($self->[$i],$len,$end) if ($self->[$i]);
	}
    }
    my $deltaC = $n*$cEnd;
    my $k = $cEnd > 0 ? 5 : 6; # clo or chi (array)
    foreach my $i (0 .. $#{$self->[$k]}) {
	$self->[$k][$i] += $deltaC;
    }
    if ($cEnd == 1) {
	# underlyingStart:
	$self->[21] += runThroughCigarString($deltaC,$self->cigarString);
    }
    trimCigarString($self,$deltaC);
    my $deltaR = $nq*$rEnd;
    $self->[$rEnd > 0 ? 3 : 4] += $deltaR; # rlo or rhi (single number)
    return $nq;
}

sub sbjctSeq {
    my ($self,$val) = @_;
    return setOrGet($self,18,$val);
}
	

# $end is 1 for beginning, -1 for end.
sub removeForNucleotides {
    my ($n,$seq,$end) = @_;
    my $len = 0;
    my $pos = $end > 0 ? 0 : 1;
    while ($n > 0) {
	$n -- if (substr($seq,$pos,1) ne '-');
	$pos += $end;
	$len ++;
    }
    return $len;
}

sub getFromEnd {
    my ($str,$len,$end) = @_;
    return ($end > 0) ? substr($str,0,$len) : substr($str,length($str)-$len);
}

sub querySeq {
    my ($self,$val) = @_;
    return setOrGet($self,16,$val);
}

sub removeAtEnd {
    my ($str,$len,$end) = @_;
    return ($end > 0) ? substr($str,$len) : substr($str,0,length($str)-$len);
}

sub runThroughCigarString {
    my ($count,$cs) = @_;
    my $todo = $count;
    my @items = split(" ",$cs);
    my $ans = 0;
    while (@items) {
	my $item = shift @items;
	my $len = substr($item,1);
	if ($item =~ /^D/) {
	    $ans += $len;
	} elsif ($len >= $todo) {
	    return $ans+$todo;
	} else {
	    $ans += $len;
	    $todo -= $len;
	}
    }
    confess("runThroughCigarString: failed on $count and $cs");
}

sub trimCigarString {
    my ($self,$delta) = @_;
    my @ops = split(" ",$self->cigarString());
    if ($delta > 0) {
	# move left end inwards
	while ($delta > 0 && @ops) {
	    if ($ops[0] =~ /^D/) {
		shift @ops;
	    } else {
		my $len = substr($ops[0],1);
		if ($len >= $delta) {
		    $ops[0] = 'M' . ($len-$delta);
		    last;
		} else {
		    shift @ops;
		    $delta -= $len;
		}
	    }
	}
    } else {
	# move right end inwards
	while ($delta < 0 && @ops) {
	    if ($ops[-1] =~ /^D/) {
		pop @ops;
	    } else {
		my $len = substr($ops[-1],1);
		if ($len >= -$delta) {
		    $ops[-1] = 'M' . ($len+$delta);
		    last;
		} else {
		    pop @ops;
		    $delta += $len;
		}
	    }
	}
	
    }
    # Remove any D's from extreme ends
    while (@ops && $ops[0] =~ /^D/) {
	shift @ops;
    }
    while (@ops && $ops[-1] =~ /^D/) {
	pop @ops;
    }
    $self->cigarString(join(" ",@ops));
}

sub chir {
    my ($self,$level) = @_;
    if ($self->fc eq 'F') {
	return $self->chi($level);
    } else {
	return $self->clo($level);
    }
}

# fc should be F or C

sub fc {
    my ($self) = @_;
    return $self->[7];
}

# Clone is strain-place

sub clone {
    my ($self) = @_;
    my $name = $self->[1];
    return substr($name,0,rindex($name,'.'));
}

# Name is strain-place.[pq]1k, we hope 

sub name {
    my ($self) = @_;
    return $self->[1];
}

# clor corresponds to rlo, chir to rhi

sub clor {
    my ($self,$level) = @_;
    if ($self->fc eq 'F') {
	return $self->clo($level);
    } else {
	return $self->chi($level);
    }
}

sub clusnum {
    my ($self) = @_;
    return $self->[12];
}

sub compareInRead {
    my ($a1,$a2) = @_;
    return ($a1->rlo <=> $a2->rlo) ||
	   ($a1->rhi <=> $a2->rhi) ||
	   ($a1->clo(1) <=> $a2->clo(1)) ||
	   ($a1->chi(1) <=> $a2->chi(1));
}

# rlo always < rhi EXCEPT when zero score

sub rlo {
    my ($self) = @_;
    return $self->[3];
}

sub rhi {
    my ($self) = @_;
    return $self->[4];
}
	
# --------------------------

# Return 0 if not compatible i.e. capable of going into same cluster;
# -1 if a1 < a2 in cluster; 1 if a2 > a1.

sub compatibility {
    my ($a1,$a2) = @_;

    # Must be for same clone and chr if they're to be compatible:
    return 0 unless ($a1->chr(1) eq $a2->chr(1) && $a1->clone eq $a2->clone);

    if ($a1->name eq $a2->name) {
	# Same read: must be same orientation, i.e. we don't allow
	# inversions (yet...). Hand to compatible1 in the order in
	# which they occur in read.
	if ($a1->fc ne $a2->fc) {
	    return 0; # different orientation: reject at present
	} elsif (compareInRead($a1,$a2) <= 0) {
	    # a1 before a2 in read. If fail compatible1, fail here;
	    # else return cluster order, i.e. -1 if F (because then
	    # a1 < a2 in cluster order) else 1.
	    return !compatible1($a1,$a2) ? 0 : ($a1->fc eq 'F') ?  -1 : 1;
	} else {
	    # As above, but reversed.
	    return !compatible1($a2,$a1) ? 0 : ($a1->fc eq 'F') ?  1 : -1;
	}

    } else {

	# Mate reads: must be opposite orientation. If so,
	# hand to compatible2 in f,c order.
	if ($a1->fc eq $a2->fc) {
	    return 0;
	} elsif ($a1->fc eq 'F') {
	    return compatible2($a1,$a2) ? -1 : 0;
	} else {
	    return compatible2($a2,$a1) ?  1 : 0;
	}

    }
}

# Whether two alis of same read, same chr, same ori are compatible.
# We know/assume a1 before (or equal to) a2 in read.

# If overlap is too large, put in separate clusters; then the
# inserted material (which is duplicate) should be inserted OK.

sub compatible1 {
    my ($a1,$a2) = @_;

    my $minRLen = min($a1->rhi-$a1->rlo+1,$a2->rhi-$a2->rlo+1);
    # Maximum overlap we will allow
    my $maxOverlap = min(25,$minRLen/2);

    my $rOver = $a1->rhi - $a2->rlo + 1;
    return 0 if ($rOver > $maxOverlap); # too much overlap -> fail.

    # We also have too much overlap if $rOver is more than half the
    # length of either alignment.

    # Sort a1 and a2 to be in the order they _ought_ to be (for
    # compatibility) in the chromosome.
    my ($a1c,$a2c) = orderInChromosome($a1,$a2);

    # Fail if wrong order, or too much overlap, in chromosome.
    # $d is size of gap in chr sequence.
    my $d = $a2c->clo(1) - $a1c->chi(1);
    # Allow up to 10000 gap:
    return $d >= -$maxOverlap && $d <= 10000;

}

sub min {
    my ($a,$b) = @_;
    if (!defined $a) {
	return $b;
    } elsif (!defined $b) {
	return $a;
    } elsif ($a < $b) {
	return $a;
    } else {
	return $b;
    }
}

# Old clause at end of compatible1, allowing duplications:

#    if ($a1->rlo == $a2->rlo && $a1->rhi == $a2->rhi) {
#	# Duplication in chr: OK if gap is no bigger than aligned
#	# length (cautious), except this can't happen because $rOver
#	# will have been too big.
#	return ($a2c->chi-$a1c->clo <= 3*$a1c->alen);
#    } elsif ($a1c->clo == $a2c->clo && $a1c->chi == $a2c->chi) {
#	# Duplication in read; we know a1 < a2 in read; OK
#	# if (again) gap is no bigger than aligned length.
#	return ($a2->rhi-$a1->rlo <= 3*$a1->alen);
#    } elsif ($a2->rlo-$a1->rhi <= -$maxOverlap) {
#	# Too much overlap when not duplication (repeats $rOver test!)
#	return 0;
#    } else {
#	my $d = $a2c->clo - $a1c->chi;
#	return $d >= -$maxOverlap && $d <= 10000;
#    }

sub orderInChromosome {
    my ($a1,$a2) = @_;
    # Assume both same fc
    if ($a1->fc eq 'F') {
	return ($a1,$a2);
    } else {
	return ($a2,$a1);
    }
}

# Whether two alis, one from each mate, are consistent. For simplicity,
# we don't allow any overlap at all in the chromosome.

sub compatible2 {
    my ($af,$ac) = @_;
    my $dOuter = $ac->chi(1) - $af->clo(1) + 1;
    my $dInner = $ac->clo(1) - $af->chi(1) - 1;
    return ($dOuter <= $MAXIMUM_READ_SPAN && $dInner >= 0);
}

sub dangleDirection {
    my ($self,$val) = @_;
    return setOrGet($self,22,$val);
}

sub decrementOffsets {
    my ($self) = @_;
    foreach my $i (3 .. 6) {
	$self->[$i] --;
    }
}

sub getCigar {
    my ($self) = @_;
    return $self->[13];
}

# Assumes seqs, quals and cigar string are present.

# A match string is of the form
#   0 underlyingStart
#   1 m|d|I
#   2 length of this match
#   3 length of alignment REMAINING against underlying sequence (
#       initially sum of M and D numbers in cigar)
#   4 readName:rlo:rhi (describing this alignment not just this match)
#   5 nucleotides ("length" of them, except just 1 if it's a "d" string)
#   6 qualities   (ditto)
#   7 alignment id (added later)
#   8 special (if any; also added later)

sub getMatchStrings {
    my ($self) = @_;
    my $type = $self->getType();
    if ($type eq 'ALI') {
	return getNormalMatchStrings($self);
    } else {
	my @ans;
	if ($type ne 'DEL') { # INS or SUB
	    my $seq = $self->querySeq;
	    push @ans,sprintf("%8d I %5d %5d %s:%d:%d%s %s %s",
			      $self->underlyingStart,
			      length($seq),
			      0,
			      $self->name,$self->rloc,$self->rhic,
			      prefixWith('|',$self->dangleDirection),
			      $seq,
			      $self->queryQual);
	}
	if ($type ne 'INS') { # DEL or SUB
	    my $len = $self->chi(-1)-$self->clo(-1)+1;
	    push @ans,sprintf("%8d d %5d %5d %s:%d:%d = ~",
			      $self->underlyingStart,
			      $len,$len,
			      $self->name,$self->rloc,$self->rhic);
	}
	return @ans;
    }
}

# Type is INS for insertion in read
#         DEL for deletion in read
#         DUP for duplication in read
#         ALI for normal (positive-score) alignment
#         SUB for zero-score, non-indel highly-variable region

sub getType {
    my ($self) = @_;
    if ($self->clo == $self->chi+1) {
	return 'INS';
    } elsif ($self->rlo == $self->rhi+1) {
	return ($self->clo > $self->chi) ? 'DUP' : 'DEL';
    } elsif ($self->perci > 0) {
	return 'ALI';
    } else {
	return 'SUB';
    }
}

# percent identity

sub perci {
    my ($self) = @_;
    return $self->[9];
}

sub getNormalMatchStrings {
    my ($self) = @_;
    my @ans;
    my @ops = split(" ",$self->cigarString());
    # Delete initial M or D, then we can just treat them as lengths
    # Also sum these; we decrease sumL as we go, to be the number of
    # existing alignment rows in the layout we still need for this alignment.
    my $sumL = 0;
    foreach my $k (0 .. $#ops) {
	$ops[$k] = substr($ops[$k],1);
	$sumL += $ops[$k];
    }
    my $name = sprintf("%s:%d:%d",$self->name,$self->rloc,$self->rhic);
    my $cOff = $self->underlyingStart;
    my $qSeq = $self->querySeq;
    my $qQual = $self->queryQual;
    my $sSeq = $self->sbjctSeq;
    my $opOff = 0; # offset in current (match) op
    my @ans;
    while (@ops) {
	# We can assume $ops[0] is always a match, and $ops[1] if present is a delete.
	my $len = min(nextChunkSize($qSeq,$sSeq), $ops[0]-$opOff);
	if ($len > 0) {
	    push @ans,sprintf("%8d m %5d %5d %s %s %s",
			      $cOff,$len,$sumL,$name,
			      substr($qSeq,0,$len),substr($qQual,0,$len));
	    $sSeq = substr($sSeq,$len);
	    $qSeq = substr($qSeq,$len);
	    $qQual = substr($qQual,$len);
	    $opOff += $len;
	    $cOff += $len;
	    $sumL -= $len;
	}
	if ($opOff == $ops[0]) {
	    shift @ops;
	    if (@ops) {
		my $lg = shift @ops;
		push @ans,sprintf("%8d d %5d %5d %s - %s",
				  $cOff,$lg,$sumL,$name,
				  substr($qQual,0,1));
		$opOff = 0;
		$cOff += $lg;
		$sumL -= $lg;
	    } else {
		last;
	    }
	}
	if ($qSeq =~ /^(-+)/) {
	    my $lg = min(length($1), $ops[0]-$opOff);
	    push @ans,sprintf("%8d d %5d %5d %s - %s",
			      $cOff,$lg,$sumL,$name,
			      substr($qQual,0,1));
	    $sSeq = substr($sSeq,$lg);
	    $qSeq = substr($qSeq,$lg);
	    $qQual = substr($qQual,$lg); # because it's been expanded for gaps
	    $cOff += $lg;
	    $sumL -= $lg;
	    $opOff += $lg;
	} elsif ($sSeq =~ /^(-+)/) {
	    my $lg = length($1);
	    push @ans,sprintf("%8d I %5d %5d %s %s %s",
			      $cOff,$lg,$sumL,$name,
			      substr($qSeq,0,$lg),substr($qQual,0,$lg));
	    $sSeq = substr($sSeq,$lg);
	    $qSeq = substr($qSeq,$lg);
	    $qQual = substr($qQual,$lg);
	}
    }
    return mergeAdjacentDeletions(@ans);
}

# rloc corresponds to clo, rhic to chi

sub rloc {
    my ($self) = @_;
    if ($self->fc eq 'F') {
	return $self->rlo;
    } else {
	return $self->rhi;
    }
}

sub rhic {
    my ($self) = @_;
    if ($self->fc eq 'F') {
	return $self->rhi;
    } else {
	return $self->rlo;
    }
}

sub underlyingStart {
    my ($self,$val) = @_;
    return setOrGet($self,21,$val);
}   

sub queryQual {
    my ($self,$val) = @_;
    return setOrGet($self,17,$val);
}

sub nextChunkSize {
    my ($seq1,$seq2) = @_;
    my $i1 = index($seq1,'-');
    my $i2 = index($seq2,'-');
    if ($i1 < 0) {
	return ($i2 < 0) ? length($seq1) : $i2;
    } elsif ($i2 < 0 || $i1 <= $i2) {
	return $i1;
    } else {
	return $i2;
    }
}

sub mergeAdjacentDeletions {
    my (@lst) = @_;
    my @ans;
    my $i=0;
    while ($i <= $#lst) {
	my @msi = split(" ",$lst[$i]);
	if ($msi[1] eq 'd') {
	    my $j=$i+1;
	    while ($j <= $#lst) {
		my @msj = split(" ",$lst[$j]);
		if ($msj[1] eq 'd') {
		    $msi[2] += $msj[2];
		    $j ++;
		} else {
		    last;
		}
	    }
	    # We don't need to adjust the names (with their read offsets)
	    # as they ought to be the same for adjacent deletions.
	    push @ans,sprintf("%8d %s %5d %5d %s %s %s",@msi);
	    $i = $j;
	} else {
	    push @ans,$lst[$i];
	    $i++;
	}
    }
    # Get rid of deletions at extreme ends
    while (@ans && substr($ans[0],9,1) eq 'd') {
	shift @ans;
    }
    while (@ans && substr($ans[-1],9,1) eq 'd') {
	pop @ans;
    }
    return @ans;
}

sub prefixWith {
    my ($pfx,$val) = @_;
    if (defined $val) {
	return "$pfx$val";
    } else {
	return undef;
    }
}

sub place {
    my ($self) = @_;
    my $name = $self->[1];
    if ($name =~ /-(\d+[a-z]\d+)\./) {
	return $1;
    }
}

# pq should be p or q

sub pq {
    my ($self) = @_;
    return substr($self->[1],-3,1);
}

sub reduceChrName {
    my ($chr) = @_;
    if ($chr =~ /\.\d\d\d$/) {
	return substr($chr,length($chr)-4);
    } else {
	my $pos = index($chr,'.');
	if ($pos > 0) {
	    return substr($chr,$pos+1);
	} else {
	    return $chr;
	}
    }
}

# read length

sub rlen {
    my ($self) = @_;
    return $self->[10];
}

sub sbjctQual {
    my ($self,$val) = @_;
    return setOrGet($self,19,$val);
}

sub score {
    my ($self) = @_;
    return $self->[0];
}

# shared score

sub sharedScore {
    my ($self,$val) = @_;
    return setOrGet($self,13,$val);
}

sub special {
    my ($self,$val) = @_;
    return setOrGet($self,23,$val);
}

# Status and cluster number will only exist when we've read
# from alilines_status.txt ...

sub status {
    my ($self) = @_;
    return $self->[11];
}

sub strain {
    my ($self) = @_;
    my $name = $self->[1];
    return substr($name,0,rindex($name,'-'));
}

# Attempt to reproduce ssahaSNP format (new format, with rlo > rhi for C alignments).

sub toString {
    my ($self) = @_;
    my $s = $self->score;
    $s .= ' ' x (4-length($s)) if (length($s) < 4);
    my $special = $self->special;
    $special = "|$special" if ($special);
    return sprintf("ALIGNMENT  %4s  %s %s %8d %8d %9d %9d   %s %7d %5.2f %d%s%s",
		   $s,
		   $self->name,
		   $self->chr,
		   $self->rloc,
		   $self->rhic,
		   $self->clo,
		   $self->chi,
		   $self->fc,
		   $self->alen,
		   $self->perci,
		   $self->rlen,
		   $self->trimString,
		   $special
	           );
}

sub trimString {
    my ($self) = @_;
    my $lo = $self->loTrim;
    my $hi = $self->hiTrim;
    if ($lo || $hi) {
	return sprintf("|%d,%d",$lo,$hi);
    } else {
	my $d = $self->dangleDirection;
	if ($d) {
	    return "|$d";
	} else {
	    return undef;
	}
    }
}    

1;
