package SGRP::AlignmentClashRates;

use Exporter;
@ISA = qw(Exporter);

@EXPORT = qw(alignmentClashRates);

use SGRP::Variables;

use Utility;
use File::Basename;

use Carp;

use strict;

my $MIN_SNP_SEPARATION = 10;

my $HIGH_COVERAGE_THRESHOLD = 20;

sub alignmentClashRates {
    my ($f,$g) = @_;
    unless (isComplete($f)) {
	report("$g: cannot create because $f is not ready",1);
	return;
    }
    return unless (stamp($g));
    report("$g: creating");
    my $fh = openToRead($f);
    my $gh = openToWrite($g);
    my %posSet;
    my (%num,%den,%den2,%quals,%lo,%hi,%next);
    my $aOff = 0;
    my $chr = basename(dirname($f));
    my @readAt;
    my @isExtraAt;
    my %inHighCov;
    while (<$fh>) {
	chomp;
	if (/^>REF/) {
	    # nothing
	} elsif (/^>(\S+) (\d+) (\S+)/) {
	    my ($ali,$pos,$iter) = ($1,$2,$3);
	    $isExtraAt[$pos] = $iter =~ /^[a-z]/;
	    # If there's already something at pos, print its result.
	    if (defined $readAt[$pos]) {
		my $r = $readAt[$pos];
		my $str = substr($r,0,index($r,'-'));
		my $hc = defined $inHighCov{$r};
		printResult($chr,$pos,$hc,$isExtraAt[$pos],\%num,\%den,\%den2,\%quals,\%lo,\%hi,\%next,$posSet{$str},$gh);
		delete $inHighCov{$r}; # save space
	    }
	    my $str = substr($ali,0,index($ali,'-'));
	    $posSet{$str}{$pos} = $ali;
	    $readAt[$pos] = $ali;
	} else {
	    $aOff ++;
	    my $len = length($_);
	    my $posMax = $len/2-2;
	    foreach my $str (keys %posSet) {
		my $pH = $posSet{$str};
		my @posList = keys %$pH;
		my $pos1;
		my $nExtra;
		while (@posList) {
		    $pos1 = shift @posList;
		    my $col1 = 2*($pos1+1);
		    if ($pos1 > $posMax || substr($_,$col1,1) eq ' ') {
			my $r = $readAt[$pos1];
			my $hc = defined $inHighCov{$r};
			printResult($chr,$pos1,$hc,$isExtraAt[$pos1],\%num,\%den,\%den2,\%quals,\%lo,\%hi,\%next,$pH,$gh);
			delete $inHighCov{$r};
			$readAt[$pos1] = undef;
			# Exclude high coverage from comparisons...should exclude rDNA
		    } elsif (@posList >= $HIGH_COVERAGE_THRESHOLD) {
			unless (defined $nExtra) {
			    $nExtra = 0;
			    foreach my $pos (@posList) {
				$nExtra ++ if ($isExtraAt[$pos]);
			    }
			}
			# DON'T mark as high-coverage if this is a real alignment and there
			# are too few other real ones.
			unless (!$isExtraAt[$pos1] && scalar(@posList)-$nExtra < $HIGH_COVERAGE_THRESHOLD) {
			    $inHighCov{$pH->{$pos1}} = 1;
			}
		    } else {
			comparePositions($_,$pos1,$posMax,$pH,$col1,\@posList,\@isExtraAt,\%num,\%den,\%den2,
					 \%quals,$aOff,\%lo,\%hi,\%next);
		    }
		}
	    }
	}
    }
    unstamp($g);
    report();
}

sub printResult {
    my ($chr,$pos,$hc,$isExtra,$pNum,$pDen,$pDen2,$pQuals,$pLo,$pHi,$pNext,$pH,$gh) = @_;
    my $r = $pH->{$pos} || die("Missing r for pH of $pos");
    if ($hc || $pDen->{$r}) {
	my $n = $pNum->{$r};
	my $d = $pDen->{$r};
	my $d2 = $pDen2->{$r};
	my $sd = $d2 ? sqrt($d2-$d) : 0;
	# We subtract 199054 as that's the score we get from a double 56, and happens
	# to be the same order as the number of reads we get over the whole
	# genome.
	my $z = $sd > 0 ? max(($n-$d-199054)/$sd,0) : 0;
	# We (currently) only care about positive z ...
	if ($hc || $z > 0) {
	    my @quals;
	    if (defined $pQuals->{$r}) {
		@quals = sort {$b <=> $a} @{$pQuals->{$r}};
	    }
	    printf($gh "%9.3f %10d %5d %5d %35s %8s %d %d",$z,$n,$d,
		   scalar(@quals),$r,$chr,$isExtra,$hc);
	    if (defined $pLo->{$r}) {
		printf($gh " %7d %7d   ",$pLo->{$r},$pHi->{$r});
	    } else {
		printf($gh " %7s %7s   ",'','');
	    }
	    for (my $i=0; $i<10 && $i<=$#quals; $i++) {
		printf($gh " %2d",$quals[$i]);
	    }
	    print $gh "\n";
	}
    }
    delete $pH->{$pos};
    delete $pNum->{$r};
    delete $pDen->{$r};
    delete $pNext->{$r};
}

sub comparePositions {
    my ($line,$pos1,$posMax,$pH,$col1,$pPosList,$pIsExtraAt,$pNum,$pDen,$pDen2,$pQuals,$off,$pLo,$pHi,$pNext) = @_;
    my $v1 = substr($line,$col1,1);
    $v1 =~ tr/a-z/A-Z/;
    $v1 =~ s/[_=0-9]/-/g;
    my $qc1 = substr($line,$col1+1,1);
    my $isExtra1 = $pIsExtraAt->[$pos1];
    foreach my $pos2 (@$pPosList) {
	my $isExtra2 = $pIsExtraAt->[$pos2];
	next if ($isExtra1 && $isExtra2); # no interest if both are extra
    	my $col2 = 2*($pos2+1);
	next if ($pos2 > $posMax || substr($line,$col2,1) eq ' ');
	my $v2 = substr($line,$col2,1);
	$v2 =~ tr/a-z/A-Z/;
	$v2 =~ s/[_=0-9]/-/g;
	my $qc2 = substr($line,$col2+1,1);
	my ($r1,$r2) = sort($pH->{$pos1},$pH->{$pos2});
	my ($leave1,$leave2);
	if ($r1 eq $pH->{$pos1}) {
	    $leave1 = $isExtra2 && !$isExtra1;
	    $leave2 = $isExtra1 && !$isExtra2;
	} else {
	    $leave2 = $isExtra2 && !$isExtra1;
	    $leave1 = $isExtra1 && !$isExtra2;
	}
	# Don't compare if $off < next for this read pair
	my $nxt = $pNext->{$r1}{$r2};
	next if (defined $nxt && $nxt > $off);
	# First moment += pe * 1/pe = 1
	$pDen->{$r1} ++ unless ($leave1);
	$pDen->{$r2} ++ unless ($leave2);
	# We flatten all qualities to 56, as we don't believe anything higher...
	my $q1 = min(56,$qc1 eq ' ' ? 127 : ord($qc1)-33);
	my $q2 = min(56,$qc2 eq ' ' ? 127 : ord($qc2)-33);
	my $pe1 = exp(-0.1*$q1*log(10));
	my $pe2 = exp(-0.1*$q2*log(10));
	my $pe = $pe1 + $pe2*(1-$pe1);
	my $inv = 1/$pe;
        # Second moment += pe * (1/pe^2) = 1/pe
	$pDen2->{$r1} += $inv unless ($leave1);
	$pDen2->{$r2} += $inv unless ($leave2);
	if ($v1 ne $v2) {
	    $pNum->{$r1} += $inv unless ($leave1);
	    $pNum->{$r2} += $inv unless ($leave2);
	    $pNext->{$r1}{$r2} = $off + $MIN_SNP_SEPARATION;
	    my $q = min($q1,$q2);
	    if ($q >= 0) {
		my @rList;
		push @rList,$r1 unless ($leave1);
		push @rList,$r2 unless ($leave2);
		foreach my $r (@rList) {
		    push @{$pQuals->{$r}},$q;
		    $pLo->{$r} = $off unless (defined $pLo->{$r});
		    $pHi->{$r} = $off;
		}
	    }
	}
    }
}

1;
