package AlignmentCluster;

use SGRP::SequencedRegionSet;
use Utility;

use Carp;

use strict;

# Amount by which a break is allowed to fall inside a sequenced region
# without it mattering.

my $BISR_LEEWAY = 0; # was 10, but we deal with this in
                     # StrainAssemblies::printSequencesRegions now.

my $MIN_INSERTION_LENGTH = 30; # or we get a suspicious surplus less than this

# @self array:
#   0 list of alignments
#   1 score, without any extra for insertions at either end
#   2 status
#   3 watermarks
#   4 extra score for insertions at either end

sub new {
    my ($class,$ali) = @_;
    my $self = bless [],$class;
    $self->[0] = [$ali];
    $self->[1] = $ali->sharedScore;
    return $self;
}

# -----------

# Return 1 iff compatible.
# Add score, but subtract 20 unless p..q

sub add {
    my ($self,$ali) = @_;
    my $sz = $self->size;
    my $a;
    for (my $k=0; $k < $sz; $k++) {
	$a = $self->get($k);
	my $c = $a->compatibility($ali);
	if ($c > 0) {
	    $self->[0] = [@{$self->[0]}[0..$k-1],
			  $ali,
			  @{$self->[0]}[$k..$sz-1]];
	    $self->[1] += $ali->sharedScore - ($a->pq eq $ali->pq ? 20 : 0);
	    return 1;
	} elsif ($c == 0) {
	    return 0;
	}
    }

    # we put it right at the end.
    push @{$self->[0]},$ali;
    $self->[1] += $ali->sharedScore - ($a->pq eq $ali->pq ? 20 : 0);
    return 1;
}

sub size {
    my ($self) =@_;
    return scalar(@{$self->[0]});
}

sub get {
    my ($self,$idx) =@_;
    return $self->[0][$idx];
}

sub adjustScoreForDangles {
    my ($self,$pDang) = @_;
    my $chr = $self->chr(1);
    my $pH = $pDang->{$chr};
    return unless ($pH);
    my $pLeft = $pH->{L};
    my $pRight = $pH->{R};
    foreach my $pAli ($self->getAll) { # NOT getAllExtended!
	my $clo = $pAli->clo(1);
	my $chi = $pAli->chi(1);
	foreach my $off (@$pLeft) {
	    my $d = $clo-$off;
	    if (abs($d) <= 10) {
		if ($d < 0 && !$pAli->loTrim) {
		    $pAli->loTrim(-$d);
		    # NOT SAFE to actually apply the trim,
                    # as this is called before the creation
                    # of ssahaSNP.pass.gz.
		    #$pAli->applyTrim();
		}
		$self->[4] = 100; # INSERTION-EXTENSION BONUS
		#printf("Bonus at $off L for %s\n",$pAli->toString());
		return 1;
	    }
	}
	foreach my $off (@$pRight) {
	    my $d = $off-$chi;
	    if (abs($d) <= 10) {
		if ($d < 0 && !$pAli->hiTrim) {
		    $pAli->hiTrim(-$d);
		    #$pAli->applyTrim();
		}
		$self->[4] = 100; # INSERTION-EXTENSION BONUS
		#printf("Bonus at $off R for %s\n",$pAli->toString());
		return 1;
	    }
	}
    }
    return 0;
}

sub chr {
    my ($self,$level) =@_;
    return $self->[0][0]->chr($level);
}

sub getAll {
    my ($self) = @_;
    return @{$self->[0]};
}

sub clo {
    my ($self,$level) =@_;
    return $self->[0][0]->clo($level);
}

sub chi {
    my ($self,$level) =@_;
    return $self->[0][-1]->chi($level);
}

sub clone {
    my ($self) =@_;
    return $self->[0][0]->clone;
}

# "Effectively identical" if same spans of read, scores and %ident.
sub effectivelyIdentical {
    my ($self,$pClus) = @_;
    return 0 unless ($self->chr eq $pClus->chr);
    my $pMat1 = $self->spanAndScoreMatrix;
    my $pMat2 = $pClus->spanAndScoreMatrix;
    return 0 unless (scalar(@$pMat1) == scalar(@$pMat2));
    foreach my $i (0 .. $#$pMat1) {
	foreach my $j (0 .. $#{$pMat1->[$i]}) {
	    return 0 unless ($pMat1->[$i][$j] == $pMat2->[$i][$j]);
	}
    }
    return 1;
}

sub spanAndScoreMatrix {
    my ($self) = @_;
    my @mat;
    foreach my $pA ($self->getAll) {
	push @mat,[$pA->rlo,$pA->rhi,$pA->score,$pA->perci];
    }
    return \@mat;
}

sub rlo {
    my ($self,$pq) = @_;
    my $min;
    foreach my $aa (@{$self->[0]}) {
	if ($aa->pq eq $pq) {
	    $min = min($min,$aa->rlo);
	}
    }
    return $min;
}

sub rhi {
    my ($self,$pq) = @_;
    my $max;
    foreach my $aa (@{$self->[0]}) {
	if ($aa->pq eq $pq) {
	    $max = max($max,$aa->rhi);
	}
    }
    return $max;
}

sub score {
    my ($self) =@_;
    return $self->[1]+$self->[4];
}

sub ends {
    my ($self) = @_;
    my ($p,$q);
    foreach my $pA (@{$self->[0]}) {
	my $pq = $pA->pq;
	if ($pq eq 'p') {
	    $p = 1;
	} else {
	    $q = 1;
	}
	return 'pq' if ($p && $q);
    }
    return $p ? 'p' : 'q';
}

sub extraAlignments {
    my ($self,$pAli1,$pAli2) = @_;
    return () unless ($pAli1 || $pAli2);
    my $pq1 = $pAli1 && $pAli1->pq;
    my $pq2 = $pAli2 && $pAli2->pq;
    my @extras;
    if ($pq1 eq $pq2) {
	# Substitution (highly-variable region), or if the
	# two alignments juxtapose in the read, then a deletion.
	push @extras,$self->interExtra($pAli1,$pAli2);
    } else {
	# Insertion in read (to end of its reliable region)
	push @extras,$self->rightExtra($pAli1) if ($pAli1);
	push @extras,$self->leftExtra($pAli2) if ($pAli2);
    }
    foreach my $pA (@extras) {
	if ($pA->rlo < 0 || $pA->rhi < 0) {
	    confess(sprintf("Negative read offset in %s\n  pAli1 = %s\n  pAli2 = %s",
			    $pA->toString,$pAli1->toString,$pAli2->toString));
	}
    }
    return @extras;
}

sub interExtra {
    my ($self,$pAli1,$pAli2) = @_;

    my $rlo = $pAli1->rhi+1; # low end of new alignment
    my $rhi = $pAli2->rlo-1; # high end of new alignment

    my $fc = $pAli1->fc; # should both be same
    my ($clo,$chi);
    if ($fc eq 'F') {
	$clo = $pAli1->chir(1)+1;
	$chi = $pAli2->clor(1)-1;
    } else {
	$clo = $pAli2->clor(1)+1;
	$chi = $pAli1->chir(1)-1;
    }

    my ($d1,$d2)=(0,0);
    # We trim in the case of chi < clo+1 too because then we'll get
    # an insertion (in the read) of the overlap material, which is what
    # we want for a duplication.
    if ($rhi < $rlo-1 || $chi < $clo-1) {
	# need to locate the new alignment
	my $d = max($rlo-1-$rhi,$clo-1-$chi);
	$d1 = int($d/2);
	$rlo -= $d1;
	$clo -= $d1;
	$d2 = $d-$d1;
	$rhi += $d2; # and now $rhi = $rlo-1
	$chi += $d2; # and ditto for chi
	# "lo" and "hi" for trimming refer to chromosome orientation,
	# so we swap here in C case:
	if ($fc eq 'F') {
	    $pAli1->hiTrim($d1);
	    $pAli2->loTrim($d2);
	} else {
	    $pAli1->loTrim($d1);
	    $pAli2->hiTrim($d2);
	}
    }
    if ($rlo > $rhi && $clo > $chi) {
	# then nothing to delete/insert
	return ();
    } else {
	# WAS chr(1) below ... debug on NCYC361-28h19.p1k and REF.chr01.510
	my @a = (0, $pAli1->name, $pAli1->chr(1), $rlo, $rhi,
		 $clo, $chi, $fc, $rhi-$rlo+1, 0.0, $pAli1->rlen); 
	return (new Alignment(@a));
    }
}

# fc is the F/C value of the p reads in the cluster
# (or what would be the value if there were any).

sub fc {
    my ($self) = @_;
    my $pA = $self->[0][0];
    my $fc = $pA->fc;
    if ($pA->pq eq 'q') {
	$fc =~ tr/FC/CF/;
    }
    return $fc;
}

# rightExtra is when we have an insertion to the right (in the read) of pAli,
# i.e. in the inner part of the read. We have a size threshold of 20 here
# (compared to 15 in leftExtra) because when both were 15, of 126 insertions
# length <20, 120 were from rightExtra (and for >20, the distribution looks
# nice and flat).

# If F, chi = pAli->chi, clo = chi+1 (like left C)
# If C, clo = pAli->clo, chi = clo-1 (like left F)

# At this

sub rightExtra {
    my ($self,$pAli) = @_;
    my $pq = $pAli->pq;
    my $pPairs = $self->[3]{$pq};
    return () unless ($pPairs);
    my $rhi = $pAli->rhi;
    my $wmhi = $pPairs->[-1][1];
    if ($wmhi > $rhi) {
	my $fc = $pAli->fc;
	my $chi = $fc eq 'F' ? $pAli->chi(1) : $pAli->clo(1)-1;
	my $clo = $chi+1;
	my $len = $wmhi-$rhi;
	if ($len >= $MIN_INSERTION_LENGTH) {
	    my @a = (0, $pAli->name, $pAli->chr(1), $rhi+1, $wmhi,
		     $clo,$chi,$fc,$len,0.0,$pAli->rlen);
	    my $ali = new Alignment(@a);
	    $ali->dangleDirection($fc eq 'F' ? 'R' : 'L');
	    return $ali;
	}
    }
    return ();
}

sub len {
    my ($self,$lev) = @_;
    return $self->chi($lev)-$self->clo($lev)+1;
}

# leftExtra is when we have an insertion to the left (in the read) of pAli.

# If F, clo = pAli->clo, chi = clo-1, i.e. step back before starting pAli.
# If C, chi = pAli->chi, clo = chi+1

sub leftExtra {
    my ($self,$pAli) = @_;
    my $pq = $pAli->pq;
    my $pPairs = $self->[3]{$pq};
    return () unless ($pPairs);
    my $rlo = $pAli->rlo;
    my $wmlo = $pPairs->[0][0];
    if ($wmlo < $rlo) {
	my $fc = $pAli->fc;
	my $clo = $fc eq 'F' ? $pAli->clo(1) : $pAli->chi(1)+1;
	my $chi = $clo-1;
	my $len = $rlo-$wmlo;
	if ($len >= $MIN_INSERTION_LENGTH) {
	    my @a = (0, $pAli->name, $pAli->chr(1), $wmlo, $rlo-1,
		     $clo,$chi,$fc,$len,0.0,$pAli->rlen);
	    my $ali = new Alignment(@a);
	    $ali->dangleDirection($fc eq 'F' ? 'L' : 'R');
	    return $ali;
	}
    }
    return ();
}

sub getAllExtended {
    my ($self) = @_;
    my @aList = $self->getAll();
    # Add in extra "alignments" for the high-quality missing bits.
    # Sort according to the READ coordinates.
    my @sorted = sortByAlignment(@aList);
    my @extended;
    my $pPrev;
    foreach my $pAli (@sorted,undef) {
	push @extended,$self->extraAlignments($pPrev,$pAli);
	push @extended,$pAli if ($pAli);
	$pPrev = $pAli;
    }
    return @extended;
}

# We don't use the usual "sort" function here because of an obscure
# bug in perl 5.8.6 :-(...

sub sortByAlignment {
    my @aList = @_;
    my (@pList,@qList);
    foreach my $pA (@aList) {
	if ($pA->pq eq 'p') {
	    @pList = insertByRlo($pA,@pList);
	} else {
	    @qList = insertByRlo($pA,@qList);
	}
    }
    return (@pList,@qList);
}

sub insertByRlo {
    my ($pA,@list) = @_;
    my @ans;
    while (@list) {
	if ($pA->rlo <= $list[0]->rlo) {
	    push @ans,$pA;
	    push @ans,@list;
	    return @ans;
	} else {
	    push @ans,shift @list;
	}
    }
    return (@ans,$pA);
}

sub hasBreakInSequencedRegion {
    my ($self,$pSRS) = @_;
    return undef unless $pSRS;
    # Easiest (though not so efficient) to work from getAllExtended,
    # which will be ordered by READ coordinates.
    my @aList = $self->getAllExtended;
    # Take care of simple clusters:
    return undef if (@aList == 1 || (@aList == 2 && $aList[0]->pq ne $aList[1]->pq));
    my $iLastReal;
    foreach my $i (0 .. $#aList) {
	my $pAli = $aList[$i];
	# We only check for real (not "extra") alignments because (1) extra ones
	# don't have clor(2) or chir(2) set (they were created without $pOM) and
	# (2) their end points should be adjacent to end points of real alignments,
	# so we'll get the right answer anyway.
	if ($pAli->score > 0) {
	    # Look for break at rlo, or spanning across a sequenced region:
	    if ($i > 0 && $pAli->pq eq $aList[$i-1]->pq) {
		return 1 if ($pSRS->inSequencedRegion($pAli->chr(2),$pAli->clor(2),$BISR_LEEWAY));
		if (defined $iLastReal) {
		    return 2 if ($pSRS->overlapsSequencedRegion($pAli->chr(2),$aList[$iLastReal]->chir(2),
								$pAli->clor(2),$BISR_LEEWAY));
		}
	    }
	    # Look for break at rhi:
	    if ($i < $#aList && $pAli->pq eq $aList[$i+1]->pq) {
		return 1 if ($pSRS->inSequencedRegion($pAli->chr(2),$pAli->chir(2),$BISR_LEEWAY));
	    }
	    $iLastReal = $i;
	}
    }
    return undef;
}

# Return true if some (non empty) alignment in the cluster has a sharedScore
# that is at least $prop (e.g. 0.2) times its raw score. This is meant to
# rule out transposon-only clusters (as opposed to clusters that have one end
# in a transposon and the other on dry land).

sub hasRelativelyUniqueAlignment {
    my ($self,$prop) = @_;
    foreach my $pA ($self->getAll) {
	return 1 if ($pA->sharedScore >= $prop*$pA->score);
    }
    return 0;
}

# The "score" of an insertion (to be added to cluster score but not
# to the alignment itself) is that of the alignment it proceeds from,
# proportional to its length (but we don't allow it to be more than that
# of the alignment itself). That is, we assume the extension has
# the same quality as the bit we matched.

sub incrementInsertionScore {
    my ($self,$pAli,$len) = @_;
    my $aLen = $pAli->alen;
    my $aScore = $pAli->score;
    my $prop = $len > $aLen ? 1 : $len/$aLen;
    my $iScore = int(0.5+$aScore*$prop);
    $self->[4] += $iScore;
}

sub mostlyOverlaps {
    my ($self,$type,$pRegs) = @_;
    my ($chr,$lo,$hi,$len) = ($self->chr(2),$self->clo(2),$self->chi(2),$self->len);
    return undef unless defined ($pRegs->{$chr});
    foreach my $pR (@{$pRegs->{$chr}}) {
	my ($ty,$rlo,$rhi) = @$pR;
	if ($ty eq $type) {
	    my $overlap = min($rhi,$hi)-max($rlo,$lo)+1;
	    if ($overlap > 0) {
		return (2*$overlap > $len);
	    }
	}
    }
    return undef;
}    

sub noSingleGoodAlignment {
    my ($self) = @_;
    foreach my $pA ($self->getAll) {
	if ($pA->score >= 150 && $pA->perci >= 80) {
	    return 0;
	}
    }
    return 1;
}

sub proportionSpannedBy {
    my ($self,$pC) = @_;
    my (@pSpan,@qSpan);
    foreach my $pA (@{$self->[0]}) {
	if ($pA->pq eq 'p') {
	    foreach my $k ($pA->rlo .. $pA->rhi) {
		$pSpan[$k] = 1;
	    }
	} else {
	    foreach my $k ($pA->rlo .. $pA->rhi) {
		$qSpan[$k] = 1;
	    }
	}
    }
    foreach my $pA ($pC->getAll) {
	if ($pA->pq eq 'p') {
	    foreach my $k ($pA->rlo .. $pA->rhi) {
		$pSpan[$k] = 2 if ($pSpan[$k] == 1);
	    }
	} else {
	    foreach my $k ($pA->rlo .. $pA->rhi) {
		$qSpan[$k] = 2 if ($qSpan[$k] == 1);
	    }
	}
    }
    my @n;
    foreach my $k (1 .. $#pSpan) {
	$n[$pSpan[$k]] ++ if ($pSpan[$k]);
    }
    foreach my $k (1 .. $#qSpan) {
	$n[$qSpan[$k]] ++ if ($qSpan[$k]);
    }
    unless ($n[1]+$n[2]) {
	1;
    }
    return $n[2]/($n[1]+$n[2]);
}

sub saveWatermarks {
    my ($self,$pWM) = @_;
    my $clone = $self->clone;
    foreach my $pq (qw(p q)) {
	my $pA = $pWM->{"$clone.${pq}1k"};
	if ($pA && @$pA) {
	    $self->[3]{$pq} = $pA;
	}
    }
}

sub splitPQ {
    my ($self) = @_;
    my ($pClus,$qClus);
    foreach my $pA (@{$self->[0]}) {
	if ($pA->pq eq 'p') {
	    if ($pClus) {
		$pClus->add($pA);
	    } else {
		$pClus = new AlignmentCluster($pA);
	    }
	} elsif ($qClus) {
	    $qClus->add($pA);
	} else {
	    $qClus = new AlignmentCluster($pA);
	}
    }
    return ($pClus,$qClus);
}

# For colouringOfClone:
my %STATUS_LEVEL = qw(accept 3 accept_chi 3 accept_tel 3 accept_trans 3 accept_ambiguous 3
		      accept_manual 3
		      reject_low 2 
		      reject_close 2 accept_close 2
		      reject_vc 2 reject_chi 2 reject_insertion 2
		      reject_bad_break 2 reject_bisr 2 reject_mlt_bisr 2 reject_poor 2
                      accept_bisr 2 accept_late 3 accept_chi_late 3 accept_insertion 2
		      ambiguous_trans 5 ambiguous_tel 4
		      ambiguous_trans_ua 5 manual 3 ambiguous 3 reject_ncb 2 accept_ncb 2);

# For browser display: same except accepts are 6 not 3:
my %STATUS_CODE = qw(accept 6 accept_chi 6 accept_tel 6 accept_trans 6 accept_ambiguous 6
		     accept_manual 6
		     reject_low 2 
		     reject_close 2 
		     reject_vc 2 reject_chi 2 reject_insertion 2
		     reject_bad_break 2 reject_bisr 2 reject_mlt_bisr 2 reject_poor 2
		     ambiguous_trans 5 ambiguous_tel 4
		     ambiguous_trans_ua 5 manual 3 ambiguous 3 reject_ncb 2 accept_ncb 2);

sub status {
    my ($self,$st) = @_;
    if (defined $st) {
	confess("Undefined status: $st") unless (defined $STATUS_LEVEL{$st});
	$self->[2] = $st;
    }
    return $self->[2];
}

sub statusCode {
    my ($self) = @_;
    return $STATUS_CODE{$self->status};
}

sub statusLevel {
    my ($self) = @_;
    return $STATUS_LEVEL{$self->status};
}
    
1;
