package SGRP::AssessPlates;

use Exporter;
@ISA = qw(Exporter);

@EXPORT = qw(assessPlates);

# All output goes to read/$STRAIN/$SPECIES/plates:

# (1) assessInsertsForStrain:
#
# For each of the specified strains, look at clip.wm and
# $sp/alilines.all.txt and create $sp/insert_assessment.txt, giving
# statistics on each insert. Useful to see how well things are aligning
# and, in particular, whether the pair ends seem to be consistent with
# each other.
# 
# Also create $sp/insert_summary.txt: for each plate, number of
# reads that get each of fates 0 to 5 listed below. Mishandled
# plates can often be spotted by a high value for fates 3 (for
# plate identity confusion) and 4 (for fumbling of individual plates).
#
# (2) printPlateMatrices:
#
# Outputs plateMatrices.txt: diagrams of plates, where each position has
# 
# ##  0 space neither end present
# ##  1   .   one end present
# ##  2   :   bad (both reads present but one or both bad)
# ##  3   ;   unaligned (good reads but not alignments for both ends)
# ##  4   *   inconsistent (alignments but none consistent)
# ##  5   M   matched: some consistent alignments
#
# (3) assessInsertOffsets:
#
# Creates insert_offset_assessment.txt; mainly useful as input to (4) below.
# Looks for cases where p and q are inconsistent but either p or q is
# consistent with q or p on a related well. The relationships considered are:
#   -- same position, up to 4 plates away
#   -- same plate, within 4 positions in row and/or column direction
#   -- same plate, same position, but rotated
# Each output line is of the form
#   strain plate col row di,dj,dk,r
# where col and row are both numbers (1 to 16 and 1 to 24), di is change
# to plate number, dj is change to col, dk is change to row, and r is
# whether rotated. All zeros means self-match, i.e. consistent.
#
# (4) plotInsertOffsets:
#
# (green lines) or the diagonally opposite well (red lines, for plate rotation).
# Creates a .png file for plates that show at least $MIN_WEIRD (25 is about right)
# such positions. These files are a more informative version of the ASCII
# diagrams in (2), but are only produced for the most suspicious plates.
 
use SGRP::Variables;
use Utility;
use Carp;
use File::Basename;
use GD::Simple;
use GD;
use strict;

my $GRADE_CHARS = ' .:;*M';

my $MIN_WEIRD = 25;

sub assessPlates {
    my ($species,@strains) = @_;
    unless (@strains > 0) {
	confess("Usage: $0 aspl species strain1 strain2 ...");
    }
    foreach my $str (@strains) {
	assessInsertsForStrain($species,$str);
	printPlateMatrices($species,$str);
	my $ioa = assessInsertOffsets($species,$str);
	plotInsertOffsets($ioa);
    }
}

sub assessInsertsForStrain {
    my ($species,$str) = @_;
    my $strDir = "$ROOT/read/$str"; 
    my $spDir = "$strDir/$species"; 
    my $ass = "$spDir/plates/insert_assessment.txt";
    assessInserts("$strDir/clip.wm",
		  "$spDir/alilines.all.txt",
		  $ass);
    my $sum = "$spDir/plates/insert_summary.txt";
    if (stamp($sum)) {
	report("creating $sum");
	summarizeAssessment($str,$ass,$sum);
	report();
	unstamp($sum,1);
    }
}

sub assessInserts {
    my ($c,$a,$g) = @_;
    if (! -s $c) {
	report("missing or empty: $c",1);
	return;
    } elsif (! -s $a) {
	report("missing or empty: $a",1);
	return;
    } elsif (!stamp($g)) {
	return;
    }
    report("creating $g");
    my $pWM = readWMFile($c);
    my $pAli = readAlilinesFile($a);
    my $gh = openToWrite($g);
    print $gh <<'EOF';
# Produced by sgrp.pl aspl. Fields are:
#  1 Insert name
#  2 (Clipped) p read length (0 if not in clip.fastq)
#  3 Good p read length (quality >= 20 throughout)
#  4 Number of alignments for p read
#  5 Number of such alignments consistent with a q alignment
#  6 (Clipped) q read length (0 if not in clip.fastq)
#  7 Good q read length (quality >= 20 throughout)
#  8 Number of alignments for q read
#  9 Number of such alignments consistent with a p alignment
# 10 Type: 0 missing at BOTH ends (only in summary file)
#          1 incomplete (missing reads at one or both ends)
#          2 bad (both reads present but one or both bad)
#          3 unaligned (good reads but not alignments for both ends)
#          4 inconsistent (alignments but none consistent)
#          5 some consistent alignments
EOF
    assessInserts1($pWM,$pAli,$gh);
    $gh->close();
    report();
    unstamp($g);
}

# Result of this: $pWM->{insertName}{pq} = wmNums
sub readWMFile {
    my ($f) = @_;
    my $fh = openToRead($f);
    my %wm;
    while (<$fh>) {
	chomp;
	my ($read,@nums) = split;
	my ($pre,$pq,$post) = ($read =~ /^(.*)\.([pq])1k(.*)$/);
	$pre =~ s/\/.*//;
	my $ins = $pre; # "$pre$post";
	$wm{$ins}{$pq} = \@nums;
    }
    return \%wm;
}

sub readAlilinesFile {
    my ($f) = @_;
    my $fh = openToRead($f);
    my %ali;
    while (<$fh>) {
	chomp;
	my (undef,$score,$read,@rest) = split;
	my ($pre,$pq,$post) = ($read =~ /^(.*)\.([pq])1k(.*)$/);
	# DISCARD any "/foo" at the end, as these break read pairings
	$pre =~ s/\/.*//;
	my $ins = $pre; # "$pre$post";
	unshift @rest,$score;
	push @{$ali{$ins}{$pq}},\@rest;
    }
    return \%ali;
}

sub assessInserts1 {
    my ($pWM,$pAli,$gh) = @_;
    foreach my $ins (sort keys %$pWM) {
	print $gh $ins;
	my $pWM1 = $pWM->{$ins};
	my $pAli1 = $pAli->{$ins};
	my ($pAli1p,$pAli1q);
	if (defined $pAli1) {
	    $pAli1p = $pAli1->{p};
	    $pAli1q = $pAli1->{q};
	}
	# Type 1: only got one side
	if (!defined $pWM1->{p}) {
	    printf($gh " %4d %4d %3d %3d",0,0,0,0);
	    printSingletonData($pWM1->{q},$pAli1q,$gh);
	    print $gh " 1";
	} elsif (!defined $pWM1->{q}) {
	    printSingletonData($pWM1->{p},$pAli1p,$gh);
	    printf($gh " %4d %4d %3d %3d",0,0,0,0);
	    print $gh " 1";
	} else {
	    printPairData($pWM1->{p},$pWM1->{q},$pAli1p,$pAli1q,$gh);
	}
	print $gh "\n";
    }
}

sub printSingletonData {
    my ($pWM,$pAli,$gh) = @_;
    printLengthData($pWM,$gh);
    printf($gh " %3d %3d",defined $pAli ? scalar(@$pAli) : 0,0);
}

sub printLengthData {
    my ($pWM,$gh) = @_;
    my $len = $pWM->[0];
    my $goodLen = $#$pWM ? $pWM->[2]-$pWM->[1]+1 : 0;
    printf($gh " %4d %4d",$len,$goodLen);
}

sub printPairData {
    my ($pWMp,$pWMq,$pAlip,$pAliq,$gh) = @_;
    my $asp = initialAlignmentString($pAlip);
    my $asq = initialAlignmentString($pAliq);
    # Type is 2 if not both GOOD reads, else 3.
    my $type = ($#$pWMp > 0 && $#$pWMq > 0) ? 3 : 2;
    if ($pAlip && $pAliq) {
	$type = 4;
	foreach my $i (0 .. $#$pAlip) {
	    foreach my $j (0 .. $#$pAliq) {
		if (consistentAlignmentsX($pAlip->[$i],$pAliq->[$j])) {
		    substr($asp,$i,1) = substr($asq,$j,1) = '1';
		    $type = 5;
		}
	    }
	}
    }
    printLengthData($pWMp,$gh);
    printAlignmentSummary($asp,$gh);
    printLengthData($pWMq,$gh);
    printAlignmentSummary($asq,$gh);
    print $gh " $type";
}

sub initialAlignmentString {
    my ($pAli) = @_;
    return defined $pAli ? ('0' x @$pAli) : '';
}

# Fields of $pAN are: score chr rLo rHi cL cR FC ...
sub consistentAlignmentsX {
    my ($pA1,$pA2) = @_;
    return if ($pA1->[6] eq $pA2->[6]); # same orientation no good
    return unless ($pA1->[1] eq $pA2->[1]); # diff chrs no good
    if ($pA1->[6] eq 'F') {
	return consistentAlignmentsX1($pA1,$pA2);
    } else {
	return consistentAlignmentsX1($pA2,$pA1);
    }
}

sub consistentAlignmentsX1 {
    my ($pAF,$pAC) = @_;
    my $len = $pAC->[4] - $pAF->[4] + 1;
    return ($len >= 1000 && $len <= 30000);
}

sub printAlignmentSummary {
    my ($as,$gh) = @_;
    my $sum = alignmentSummary($as);
    print $gh $sum;
}

sub alignmentSummary {
    my ($as) = @_;
    my $n = length($as);
    $as =~ s/0//g;
    return sprintf(" %3d %3d",$n,length($as));
}

sub summarizeAssessment {
    my ($str,$f,$g) = @_;
    my $fh = openToRead($f);
    my @count;
    while (<$fh>) {
	next if (/^\#/);
	chomp;
	my ($read,@a) = split;
	my $val = $a[-1];
	my ($plate) = ($read =~ /$str\-(\d+)[a-z]/);
	if (defined $plate) {
	    $count[$plate][$val] ++;
	}
    }
    my $gh = openToWrite($g);
    foreach my $plate (0 .. $#count) {
	my $pC = $count[$plate];
	next unless ($pC);
	$pC->[0] = 384;
	foreach my $k (1 .. 5) {
	    $pC->[0] -= $pC->[$k];
	}
	printf($gh "%10s %2d",$str,$plate);
	foreach my $k (0 .. 5) {
	    printf($gh " %3d",$pC->[$k]);
	}
	print $gh "\n";
    }
    $gh->close();
}

sub printPlateMatrices {
    my ($sp,$str) = @_;
    my $spDir = "$ROOT/read/$str/$sp";
    my $f = "$spDir/plates/insert_assessment.txt";
    my $g = "$spDir/plates/plateMatrices.txt";
    if (! -s $f) {
	report("$f: missing or empty",1);
	return;
    } elsif (! stamp($g)) {
	return;
    }
    report("$g: creating");
    my $ls = length($str);
    my $nsp = 19-$ls;
    my $nspLeft = int($nsp/2);
    my $nspRight = $nsp-$nspLeft;
    my $fh = openToRead($f);
    my $gh = openToWrite($g);
    print $gh `grep '^\#\#' $0`;
    my ($wMin,$wMax,@val);
    while (<$fh>) {
	chomp;
	my ($read,@a) = split;
	if ($read =~ /-([0-9]+)([a-z])([0-9]+)/) {
	    my $w = $1;
	    my $x = ord($2) - ord('a') + 1;
	    my $y = $3;
	    my $v = $a[-1];
	    $val[$w][$x][$y] = $v;
	    $wMin = $w unless (defined $wMin && $wMin < $w);
	    $wMax = $w if ($wMax < $w);
	}
    }
    
    my @chars = split(//,$GRADE_CHARS);

    print $gh "\n";
    foreach my $w ($wMin .. $wMax) {
	my @sum = 0;
	foreach my $x (1 .. 16) {
	    foreach my $y (1 .. 24) {
		$sum[$val[$w][$x][$y]] ++;
	    }
	}
	printf($gh "%10s %2d counts:",$str,$w);
	my $tot = 0;
	foreach my $k (0 .. 5) {
	    printf($gh " %3d",$sum[$k]);
	    $tot += $sum[$k];
	}
	if ($tot == 384) {
	    print $gh "\n";
	} else {
	    printf($gh " SUMS TO %3d\n",$tot);
	}
    }
    
    for (my $w0 = $wMin; $w0 <= $wMax; $w0 += 4) {
	print $gh "\n";
	foreach my $w ($w0 .. $w0+3) {
	    last if ($w > $wMax);
	    printf($gh "%s%s plate %2d%s",' ' x $nspLeft,$str,$w,' ' x $nspRight);
	}
	print $gh "\n";
	foreach my $x (1 .. 16) {
	    foreach my $w ($w0 .. $w0+3) {
		last if ($w > $wMax);
		print $gh chr(ord('a')-1+$x) . " ";
		foreach my $y (1 .. 24) {
		    print $gh $chars[$val[$w][$x][$y]];
		}
		print $gh "  ";
	    }
	    print $gh "\n";
	}
    }
    report();
    unstamp($g);
}
	
sub assessInsertOffsets {
    my ($sp,$str) = @_;
    my $f = "$ROOT/read/$str/$sp/alilines.all.txt";
    my $g = "$ROOT/read/$str/$sp/plates/insert_offset_assessment.txt";
    if (-s $f && stamp($g)) {
	report("creating $g");
	assessInsertOffsets1($str,$f,$g);
	report();
	unstamp($g);
    }
    return $g;
}

sub assessInsertOffsets1 {
    my ($str,$f,$g) = @_;
    my $fh = openToRead($f);
    my $gh = openToWrite($g);
    my ($di,$dj,$dk) = (4,4,4);
    my @ali;
    my $maxDist = $str eq 'YGPM' ? 30000 : 15000;
    while (<$fh>) {
	chomp;
	my (undef,$score,$read,$chr,undef,undef,$c1,$c2,$fc) = split;
	$read =~ s/\/[^\.]*//; # discard /a etc
	my ($plate,$row,$col,$pq) = ($read =~ /.*-(\d+)([a-p])(\d+)\.([pq])1k/);
	if ($plate) {
	    $row = ord($row)-ord('a')+1; # row 1 .. 16
	    $pq = $pq eq 'p' ? 0 : 1; # p->0, q->1
	    # Only allow one match per read, for simplicity:
	    unless (defined $ali[$plate][$row][$col][$pq]) {
		#print "$plate $row $col $pq <- $_\n";
		$ali[$plate][$row][$col][$pq] = [$chr,$c1,$fc];
	    }
	}
    }
    
    foreach my $i (1 .. $#ali) {
	my $pA = $ali[$i];
	foreach my $j (1 .. 16) {
	    my $pB = $pA->[$j];
	    foreach my $k (1 .. 24) {
		my $pAli = $pB->[$k][0];
		next unless (defined $pAli);
		foreach my $ii (max($i-$di,1) .. min($i+$di,$#ali)) {
		    my @jList = ($j);
		    my @kList = ($k);
		    if ($ii == $i) {
			@jList = (max($j-$dj,1) .. min($j+$dj,16));
			@kList = (max($k-$dk,1) .. min($k+$dk,24));
		    }
		    foreach my $jj (@jList) {
			foreach my $kk (@kList) {
			    my @rList = (0); # whether to rotate plate
			    unless ($ii || $jj || $kk) {
				push @rList,1;
			    }
			    foreach my $r (@rList) {
				my $qAli;
				if ($r == 0) {
				    $qAli = $ali[$ii][$jj][$kk][1];
				} else {
				    # Don't do this if already done without rotation
				    next if (member(17-$jj,@jList) &&
					     member(25-$kk,@kList));
				    $qAli = $ali[$ii][17-$jj][25-$kk][1];
				}
				if ($qAli && consistentAlignments($pAli,$qAli,$maxDist)) {
				    printf($gh "%s %2d %2d %2d %d,%d,%d,%d\n",
					   $str,$i,$j,$k,
					   $ii-$i,$jj-$j,$kk-$k,$r);
				}
			    }
			}
		    }
		}
	    }
	}
    }
}

sub member {
    my ($x,@list) = @_;
    foreach my $y (@list) {
	return 1 if ($y eq $x);
    }
    return 0;
}

sub consistentAlignments {
    my ($pAli,$qAli,$maxDist) = @_;
    return if ($pAli->[2] eq $qAli->[2]); # same orientation no good
    return unless ($pAli->[0] eq $qAli->[0]); # diff chrs no good
    if ($pAli->[2] eq 'F') {
	return consistentAlignments1($pAli,$qAli,$maxDist);
    } else {
	return consistentAlignments1($qAli,$pAli,$maxDist);
    }
}
	
sub consistentAlignments1 {
    my ($pAF,$pAC,$maxDist) = @_;
    my $len = $pAC->[1] - $pAF->[1] + 1;
    return ($len >= 1000 && $len <= $maxDist);
}
	
sub plotInsertOffsets {
    my ($f) = @_;

    my $dir = dirname($f);

    my $pVal = readOffsetAssessments($f);
    foreach my $str (sort keys %$pVal) {
	my $pA = $pVal->{$str};
	foreach my $plate (1 .. $#$pA) {
	    next unless (defined $pA->[$plate]);
	    my $base = sprintf("%s.%2d",$str,$plate);
	    $base =~ tr/ /0/;
	    my $g = "$dir/$base.png";
	    unless (-s $g) {
		my $pB = $pA->[$plate];
		my $nWeird = 0;
		my %cType;
		foreach my $row (1 .. 16) {
		    foreach my $col (1 .. 24) {
			my $type = $pB->[$row][$col];
			if (defined $type && $type ne '0,0,0,0') {
			    $nWeird ++;
			    $cType{$type} ++;
			}
		    }
		}
		if ($nWeird >= $MIN_WEIRD) {
		    mayPlotInsertOffsetsFor(\%cType,$pB,$g);
		}
	    }
	}
    }
}

sub readOffsetAssessments {
    my ($f) = @_;
    my $fh = openToRead($f);
    my %val;
    my $showRed = 1;
    while (<$fh>) {
	chomp;
	unless (/^\#/) {
	    my ($str,$plate,$row,$col,$type) = split;
	    if ($type eq '0,0,0,0' || !defined $val{$str}[$plate][$row][$col]) {
		if ($showRed || $type =~ /0$/) {
		    $val{$str}[$plate][$row][$col] = $type;
		}
	    }
	}
    }
    return \%val;
}

sub mayPlotInsertOffsetsFor {
    my ($pCType,$pB,$g) = @_;

    my $cellSize = 25;
  
    my $im = new GD::Simple(17*$cellSize,25*$cellSize);
    $im->fgcolor('white');
    # We don't care about rare types of apparent mishandling
    # (reduces noise):
    foreach my $type (keys %$pCType) {
	delete $pCType->{$type} if ($pCType->{$type} <= 2);
    }
    my @types = sort {$pCType->{$b} <=> $pCType->{$a}} keys %$pCType;
    createBlackRectangles($pB,$cellSize,$im);
    $im->bgcolor('white');
    my $nWeird = createBlueRectangles($pB,$cellSize,$im);
    $nWeird += createLines($pB,$cellSize,$im);
    if ($nWeird >= $MIN_WEIRD) {
	report("writing $g with $nWeird items",1);
	reallyPlotInsertOffsets($im,$cellSize,$g);
    }
}

# Print black rectangles for self-consistency

sub createBlackRectangles {
    my ($pB,$cellSize,$im) = @_;
    foreach my $row (1 .. 16) {
	foreach my $col (1 .. 24) {
	    my $type = $pB->[$row][$col];
	    if ($type eq '0,0,0,0') {
		$im->bgcolor('black');
		$im->rectangle($row*$cellSize,
			       $col*$cellSize,
			       ($row+1)*$cellSize,
			       ($col+1)*$cellSize);
	    }
	}
    }
}

# Blue for p match to q on previous plate, light blue for p match to q on next plate
sub createBlueRectangles {
    my ($pB,$cellSize,$im) = @_;
    my $nWeird = 0;
    foreach my $row (1 .. 16) {
	foreach my $col (1 .. 24) {
	    my $type = $pB->[$row][$col];
	    next unless (defined $type);
	    next if ($type eq '0,0,0,0');  # already done, above
	    if ($type eq '-1,0,0,0') {
		$im->bgcolor('blue');
		$im->rectangle($row*$cellSize,
			       $col*$cellSize,
			       ($row+1)*$cellSize,
			       ($col+1)*$cellSize);				   
		$nWeird ++;
	    } elsif ($type eq '1,0,0,0') {
		$im->bgcolor('lightblue');
		$im->rectangle($row*$cellSize,
			       $col*$cellSize,
			       ($row+1)*$cellSize,
			       ($col+1)*$cellSize);
		
		$nWeird ++;
	    }
	}
    }
    return $nWeird;
}

sub createLines {
    my ($pB,$cellSize,$im) = @_;
    my $nWeird = 0;
    foreach my $row (1 .. 16) {
	foreach my $col (1 .. 24) {
	    my $type = $pB->[$row][$col];
	    next unless (defined $type);
	    next if ($type =~ /0,0,0$/);  # already done, above
	    # di for plate, dj for row, dk for col:
	    my ($di,$dj,$dk,$r) = split(/,/,$type);
	    my $row2 = $row+$dj;
	    my $col2 = $col+$dk;
	    if ($r == 1) {
		$row2 = 17-$row2;
		$col2 = 25-$col2;
	    }
	    unless ($pB->[$row2][$col2] eq '0,0,0,0') {
		my $dj2 = $row2-$row;
		my $dk2 = $col2-$col;
		$dj2 -= 0.2*sign($dj2);
		$dk2 -= 0.2*sign($dk2);
		$im->penSize(2-$r); # 2 for green, 1 for red
		# Only kink for green lines ...
		drawKink($im,
			 ($row+0.5)*$cellSize,($col+0.5)*$cellSize,
			 ($row+0.5+$dj2)*$cellSize,($col+0.5+$dk2)*$cellSize,
			 $r == 0 ? 'green' : 'red',
			 $r == 0 ? 0.1 : 0);
		$nWeird ++;
	    }
	}
    }
    return $nWeird;
}

sub sign {
    my ($x) = @_;
    return $x <=> 0;
}

sub drawKink {
    my ($im,$x1,$y1,$x2,$y2,$color,$curve) = @_;
    my $xd = $x2-$x1;
    my $yd = $y2-$y1;
    # sep is distance from x1 to x2
    my $sep = sqrt($xd*$xd+$yd*$yd);
    my $a = $curve*$sep;
    my $dx = $curve*($y2-$y1);
    my $dy = $curve*($x1-$x2);
    my $xm = ($x1+$x2)/2 + $dx;
    my $ym = ($y1+$y2)/2 + $dy;
    $im->fgcolor($color);
    $im->moveTo($x1,$y1);
    $im->ellipse(5,5);
    $im->lineTo($xm,$ym);
    $im->lineTo($x2,$y2);
    $im->fgcolor('white');
}

sub reallyPlotInsertOffsets {
    my ($im,$cellSize,$g) = @_;
    $im->fgcolor('black');
    foreach my $row (1 .. 16) {
	$im->moveTo(($row+0.4)*$cellSize,0.75*$cellSize);
	$im->string(chr(ord('a')+$row-1));
    }
    foreach my $col (1 .. 24) {
	$im->moveTo($cellSize*0.4,($col+0.75)*$cellSize);
	$im->string($col);
    }
    my $gh = openToWrite($g);
    binmode $gh;
    print $gh $im->png;
    $gh->close();
}

1;
