package SGRP::BuildFastqAlignments;

use Exporter;
@ISA = qw(Exporter);

@EXPORT = qw(buildFastqAlignments);

use SGRP::Variables;

use Utility;
use File::Basename;

use Carp;

use strict;

# Create e.g. $SGR/assembly/CERE/cere/methc/preimp/BC187.fastq.gz and ...postimp...
#
# Output files are in fastq format. 
#
# Each header line (after the "@" or "+") consists
# of:
#   chunk name: strain.chromosome.nnn for nnn=1,2,3 ...
#   chromosome, low offset and high offset for strain
#   direction of alignment: "+" or "-"
#   chromosome, low offset and high offset for reference
# Where we have an insertion in the strain, we will have high offset = low offset - 1
# in reference; and vice versa for an insertion in the reference (deletion in the strain).
# In the latter case, the sequence and quality strings are of zero length.
# Where we have neither an insertion nor a deletion, high-low should be the same for
# strain and reference, i.e. nucleotides align one-to-one.
#
# In a sequence line (one after "@" line), the meaning of lower case is different
# for preimp and postimp.
#   Preimp: nucleotide fails NQS, _or_ no nucleotide present (no coverage), so copied
#     from reference
#   Postimp: nucleotide was imputed (missing in pre-imputation).
#
# In a quality line, '!' is used in preimp when there was no nucleotide present in
# the strain. '!' can also occur as a real quality character, but a long string of '!'s
# almost certainly means the former case.

# Max size of any chunk, for easy reading:
my $MAX_CHUNK_SIZE = 1000;

sub buildFastqAlignments {
    my ($species,$filtMeth,$strainSet,@strains) = @_;
    unless (@strains) {
	confess("Usage: $0 bfqa species filtMeth strainSet str1 str2 ...");
    }
    #print STDERR "bfqa should no longer be needed!\n";
    my ($iter,$dir) = getCurrentIteration($species,$filtMeth,$strainSet,'p?');
    my $tdir = "$ROOT/assembly/$strainSet/$species/$filtMeth";
    foreach my $str (@strains) {
	buildFastqAlignmentFile($dir,$str,'sorted',$tdir,'preimp');
	buildFastqAlignmentFile($dir,$str,'imputed_y',$tdir,'postimp');
    }
}

sub buildFastqAlignmentFile {
    my ($dir,$str,$type,$tdir,$tsub) = @_;
    my $tgt1 = "$tdir/$tsub/$str.fastq.gz";
    my $tgt2 = "$tdir/$tsub/$str.fa.gz";
    if (stamp($tgt1)) {
	report("creating $tgt1 and $tgt2");
	my $gh1 = openToWrite($tgt1);
	my $gh2 = openToWrite($tgt2);
	my @pairs;
	foreach my $f (glob("$dir/p?/*/$type.gz")) {
	    my $chr = basename(dirname($f));
	    push @pairs,[$chr,$f];
	}
	# Sort by chr name ...
	foreach my $pPair (sort {$a->[0] cmp $b->[0]} @pairs) {
	    my ($chr,$f) = @$pPair;
	    my $fh = openToRead($f);
	    addFastqAlignments($fh,$str,$chr,$gh1,$gh2);
	    $fh->close();
	}
	$gh1->close();
	$gh2->close();
	report();
	unstamp($tgt1);
    }
}

sub addFastqAlignments {
    my ($fh,$str,$chr,$gh1,$gh2) = @_;
    chomp (my $hdr = <$fh>);
    my @strains = split(" ",substr($hdr,1));
    my $ks;
    foreach my $k (0 .. $#strains) {
	if ($strains[$k] eq $str) {
	    addFastqAlignments1($fh,$str,$k,$chr,$gh1,$gh2);
	    return;
	}
    }
}

sub addFastqAlignments1 {
    my ($fh,$str,$ks,$chr,$gh1,$gh2) = @_;
    my ($roff,$soff) = (1,1);
    my $ksv = 2*$ks+2;
    my $ksq = $ksv+1;
    my ($vbuf,$qbuf,$rlo,$slo,$rGoing,$sGoing);
    my $nChunk = 1;
    my $sSeq;
    while (<$fh>) {
	chomp;
	my $rv = substr($_,0,1); # reference value
	$rv =~ s/[0-9_]/-/;
	my $sv = substr($_,$ksv,1); # strain value
	my $sq; # strain quality
	if ($sv eq ' ') {
	    # Must be sorted.gz file; guess sv is same as rv,
	    # set quality to lowest ("!")
	    $sv = $rv;
	    $sv =~ tr/ACGT/acgt/; # lower-case; certainly fails NQS!
	    $sq = '!';
	} else {
	    $sv =~ s/[0-9_]/-/;
	    $sq = substr($_,$ksq,1);
	}
	my $rGoingHere = ($rv ne '-');
	my $sGoingHere = ($sv ne '-');
	# If gap in both, ignore the line ...
	next unless ($rGoingHere || $sGoingHere);
	unless ($rGoingHere == $rGoing && $sGoingHere == $sGoing && $soff-$slo < $MAX_CHUNK_SIZE) {
	    printFastqParagraph($vbuf,$qbuf,$rlo,$roff-1,$slo,$soff-1,$chr,"$str.$chr.$nChunk",$gh1);
	    $sSeq .= $vbuf;
	    $vbuf = $qbuf = '';
	    $rlo = $roff;
	    $slo = $soff;
	    $rGoing = $rGoingHere;
	    $sGoing = $sGoingHere;
	    $nChunk ++;
	}
	if ($sGoing) {
	    $vbuf .= $sv;
	    # ~ is highest visible qual value:
	    $qbuf .= ($sq eq ' ' ? '~' : $sq);
	    $soff ++;
	}
	$roff ++ if ($rGoing);
    }
    printFastqParagraph($vbuf,$qbuf,$rlo,$roff-1,$slo,$soff-1,$chr,"$str.$chr.$nChunk",$gh1);
    $sSeq .= $vbuf;
    print $gh2 ">$chr\n";
    my $kMax = int(length($sSeq)/60 - 1)*60;
    for (my $k=0; $k<=$kMax; $k+=60) {
	print $gh2 substr($sSeq,$k,60) . "\n";
    }
    print $gh2 substr($sSeq,$kMax+60) . "\n";
}

sub printFastqParagraph {
    my ($vbuf,$qbuf,$rlo,$rhi,$slo,$shi,$chr,$id,$gh1) = @_;
    return unless (defined $rlo); # start of chromosome
    my $info = sprintf("%s %s %d %d + %s %d %d",
		       $id,$chr,$slo,$shi,$chr,$rlo,$rhi);
    printf($gh1 "@%s\n%s\n",$info,$vbuf);
    printf($gh1 "+%s\n%s\n",$info,$qbuf);
}
	
1;
