package SGRP::BuildGbrowse;

use Exporter;
@ISA = qw(Exporter);

@EXPORT = qw(buildGbrowse combineGBDBFiles);

use SGRP::Variables;
use SGRP::SGRP;

use SGRP::BuildGbrowseConfig;
use SGRP::BuildGbrowseReference;
use SGRP::BuildGbrowseGFF;
use SGRP::BuildGbrowseSNP;
#use SGRP::BuildGbrowseBreak;

use Utility;
use File::Basename;
use Carp;

use strict;

# gff and fa files are put in $TODIR/gbdb/$species/$method2.
# 
# Config file is $SGR/gbdb/gbrowse.conf/$species.$method2.conf;
# this needs to be copied into ~/myperl/conf/gbrowse.conf
# (or that could be a symlink).
# 
# To start apache:
# 
# ~/myperl/bin/apachectl -d ~/myperl -k start
# 
# Then e.g.
#
# http://localhost:8080/cgi-bin/gbrowse/cere.methp/?name=chr01

# NOTE: the @strains here are only used for incrementing the config file
# and building data files for alignments. For SNPs, we check which
# _chromosomes_ need doing and aren't already being done, then for each
# of those, take the strain list from the top of the sorted/imputed file.

# CASING RULES: file names are lower case; contig names are lower case;
# "types" are upper (or mixed) case, except for species abbreviations
# "cere" and "para".

sub buildGbrowse {
    my ($sp,$meth,$strSet,@strains) = @_;

    unless (@strains) {
	confess("Usage: e.g. $0 bugb cere methc CERE str1 str2 ...");
    }

    my ($eDir,$fDir,$tgtDir);
    if ($RELEASE_DIR) {
	$eDir = $fDir = "$RELEASE_DIR/$sp/match";
	$tgtDir = "$RELEASE_DIR/$sp/gbdb";
	# Create alicoords if absent...we ought to have copied them when creating
	# the release dir, but we didn't (DMC, 15-8-08, last-minute fix)
	foreach my $d (glob("$eDir/*")) {
	    my $src = "$d/layout.gz";
	    my $tgt = "$d/alicoords.gz";
	    if (-s $src && stamp($tgt)) {
		report("$tgt: creating");
		SGRP::AlignmentCoords::alignmentCoords2($src,$tgt);
		report();
		unstamp($tgt);
	    }
	}
    } else {
	(undef,$eDir) = getCurrentIteration($sp,$meth,$strSet,'p?');
	$fDir = previousDirectory($eDir);
	foreach my $f ("$eDir/isExtraIteration","$fDir/isFinalIteration") {
	    unless (-f $f) {
		report("$f: missing",1);
		return;
	    }
	}
	$tgtDir = "$eDir/gbdb";
    }

    my $doSNPs = buildGbrowseConfigFile($sp,$meth,$strSet,$tgtDir);

    # pRefMap is from eDir, but it refers to the coords in fDir.
    my $gmFile = findGenomeMapFile($sp,'REF',$eDir);
    my $pRefMap = new OffsetMap('?',$gmFile);
    unless ($pRefMap) {
	die("$gmFile: missing or empty");
    }
    my $pRefRegs = $pRefMap->alignmentRegions(5,1000);

    # pDangles is from fDir.
    my $pDangles = readDangles($fDir);

    # Build genome-level stuff if needed
    buildGbrowseReferenceFiles($eDir,$pRefMap,$pRefRegs,$sp,$tgtDir);

    unless (-s "$tgtDir/breaks.gff") {
	buildGbrowseLARAlignmentFile($pRefMap,$pRefMap,$pRefRegs,$pDangles,
				     'REF','REF',"$tgtDir/ref_breaks.gff");
    }

    # Build file for species alignments if needed
    my $selfGFF = "$tgtDir/smatch.gff";
    my $selfDir = $RELEASE_DIR ? "$RELEASE_DIR/$sp/ref/self" : "$ROOT/ref/$sp/self";
    buildGbrowseSpeciesAlignmentFile($selfDir,$selfGFF);

    # Build files for alignments (for specified strain(s)).
    foreach my $str (@strains) {
	# $str Avoiding initial digit (by prefix S when necessary),
	# and its lower case version. We used strA as a field in
	# GFF files, and lstrA to construct filenames (and it
	# shouldn't be passed out of this function). We need to pass
	# str as well as strA to look up the info we want.
	my ($strA,$lstrA) = avoidInitialDigit($str);
	unless (-s "$tgtDir/matches.gff") { # produced by cmgb
	    my $tgt1 = "$tgtDir/${lstrA}_matches.gff";
	    buildGbrowseGFFFiles($pRefMap,$pRefRegs,$eDir,$str,$strA,$tgt1);
	}
	unless (-s "$tgtDir/level.gff") { # produced by cmgb
	    my $lev = "$tgtDir/${lstrA}_level.gff";
	    buildGBrowseLevelFile($pRefMap,$pRefRegs,$fDir,$str,$strA,$lev);
	}
	unless (-s "$tgtDir/breaks.gff") { # produced by cmgb
	    my $brk = "$tgtDir/${lstrA}_breaks.gff";
	    my $pStrMap = new OffsetMap('?',findGenomeMapFile($sp,$str,$eDir));
	    # Hand fDir to the break code, as we'll look in imputed_y.gz.
	    buildGbrowseLARAlignmentFile($pStrMap,$pRefMap,$pRefRegs,$pDangles,$str,$strA,$brk);
	}
    }
    # Do SNPs, if needed and not already done
    if ($doSNPs) {
	buildGbrowseSNPFiles($doSNPs,$pRefMap,$pRefRegs,$sp,$fDir,$tgtDir);
    }

}

sub findGenomeMapFile {
    my ($sp,$str,$eDir) = @_;
    return $RELEASE_DIR 
	? "$RELEASE_DIR/$sp/strains/$str/assembly/genome.map"
	: "$eDir/assembly/$str/genome.map";
}

sub readDangles {
    my ($dir) = @_;
    my %dangles;
    my $pat = $RELEASE_DIR ? "$dir/*/dangles.txt" : "$dir/p?/*/dangles_y.txt";
    foreach my $f (glob($pat)) {
	my $chr = basename(dirname($f));
	my %h;
	my $fh = openToRead($f);
	while (<$fh>) {
	    chomp;
	    my ($off,$lr) = split;
	    $h{$off} = $lr;
	}
	$dangles{$chr} = \%h;
    }
    return \%dangles;
}

sub buildGBrowseLevelFile {
    my ($pRefMap,$pRefRegs,$fDir,$str,$strA,$tgt) = @_;
    return unless (stamp($tgt));
    report("writing $tgt");
    my $gh = openToWrite($tgt);
    my $pat = $RELEASE_DIR ? "$fDir/${CHR_PATTERN}/layout.gz" : "$fDir/p?/${CHR_PATTERN}/layout_y.gz";
    foreach my $f (glob($pat)) {
	my $chr = basename(dirname($f));
	my $pH = levelsFromLayout($f,$str);
	my $prev = 0;
	my $lev = 0;
	my ($levMin,$levMax);
	foreach my $off (sort {$a <=> $b} keys %$pH) {
	    my $levMid = int(($levMin+$levMax)/2);
	    my $newLev = $lev+$pH->{$off};
	    $levMin = min($levMin,$newLev);
	    $levMax = max($levMax,$newLev);
	    # To try and save bandwidth on displaying high-coverage regions,
	    # we merge together regions of up to 50 bases with less than
	    # 2% difference in coverage.
	    if ($off-$prev > 50 || $levMin < 0.98*$levMax) {
		printGFFToRegions($pRefMap,$pRefRegs,$gh,$chr,$strA,'level',$prev,$off-1,$levMid,
				  undef,undef,"$strA level");
		$prev = $off;
		$levMin = $levMax = $lev = $newLev;
	    } else {
		$lev = $newLev;
	    }
	}
    }
    unstamp($tgt);
    report();
}

sub levelsFromLayout {
    my ($f,$str) = @_;
    my %dLevel;
    my $fh = openToRead($f);
    my $off = 0;
    my %present; # 2 if present for our strain, 1 if hard gap for our strain, undef else
    while (<$fh>) {
	if (/^>/) {
	    if (/^>(\S+)-\S+ (\d+) (\S+)/) {
		my ($st,$col,$iter) = ($1,2*($2+1),$3);
		if ($present{$col} == 2) {
		    $dLevel{$off} --;
		}
		delete $present{$col};
		if ($st eq $str && $iter =~ /^\d+$/) {
		    $present{$col} = 1;
		}
	    }
	} else {
	    $off ++;
	    my $dl = 0;
	    foreach my $col (keys %present) {
		my $v = substr($_,$col,1);
		if ($v =~ /[-A-Za-z]/) { # gap is OK!
		    if ($present{$col} == 1) {
			$present{$col} = 2;
			$dl ++;
		    }
		} elsif ($present{$col} == 2) {
		    $present{$col} = 1;
		    $dl --;
		}
	    }
	    $dLevel{$off} = $dl if ($dl);
	}
    }
    return \%dLevel;
}

sub combineGBDBFiles {
    if (@_ != 1) {
	confess("Usage: sgrp.pl gbdbDir");
    }
    my ($dir) = @_;
    unless (-d $dir) {
	confess("Cannot find directory $dir");
    }
    chdir($dir);
    combineGBDBFiles1($dir);
}

sub combineGBDBFiles1 {
    my ($dir) = @_;
    my (@snps1,@snps2,@snpdens,@level,@matches,@breaks);
    foreach my $f (qw(snps.gff snpdens.gff level.gff matches.gff breaks.gff)) {
	if (-s $f) {
	    confess("Already exists: $dir/$f");
	}
    }
    foreach my $d (glob("*_snps")) {
	foreach my $f (glob("$d/*.gff")) {
	    if ($f =~ /_snps\/.*gff$/) {
		push @snps2,$f;
	    }
	}
    }
    foreach my $f (glob("*.gff")) {
	if ($f =~ /_snps.gff$/) {
	    push @snps1,$f;
	} elsif ($f =~ /_snpdens.gff$/) {
	    push @snpdens,$f;
	} elsif ($f =~ /_level.gff$/) {
	    push @level,$f;
	} elsif ($f =~ /_matches.gff$/) {
	    push @matches,$f;
	} elsif ($f =~ /_breaks.gff$/) {
	    push @breaks,$f;
	} elsif ($f ne "cere.gff" 
		 && $f ne "para.gff" 
		 && $f ne 'contigs.gff' 
		 && $f ne 'smatch.gff') {
	    confess("Unrecognized GFF file: $dir/$f");
	}
    }
    catAndDelete("snps.gff",@snps1,@snps2);
    # Delete directories for the snps files
    foreach my $d (glob("*_snps")) {
	rmdir($d);
    }
    catAndDelete("snpdens.gff",@snpdens);
    catAndDelete("level.gff",@level);
    catAndDelete("matches.gff",@matches);
    catAndDelete("breaks.gff",@breaks);
}
    
sub catAndDelete {
    my ($tgt,@srcs) = @_;
    report(sprintf("creating %s from %d files",$tgt,scalar(@srcs)));
    my $gh = openToWrite($tgt);
    foreach my $f (@srcs) {
	my $fh = openToRead($f);
	while (<$fh>) {
	    print $gh $_;
	}
	$fh->close();
	unlink($f);
    }
    report();
}


sub alignmentPresent {
    my ($pA,$pBList) = @_;
    foreach my $pB (@$pBList) {
	return 1 if alignmentsEquivalent($pA,$pB);
    }
    return 0;
}

sub alignmentsEquivalent {
    my ($pA,$pB) = @_;
    foreach my $k (0 .. $#$pA) {
	return 0 unless ($pA->[$k] eq $pB->[$k]);
    }
    return 1;
}

sub consistentClusterPosition {
    my ($pClus,$qClusters) = @_;
    foreach my $k (0 .. $#$qClusters) {
	my $qClus = $qClusters->[$k];
	if ($qClus) {
	    if ($pClus->[0][6] eq 'F') {
		return $k if (consistentClusters($pClus,$qClus));
	    } else {
		return $k if (consistentClusters($qClus,$pClus));
	    }
	}
    }
    return -1; # no result
}

sub consistentClusters {
    my ($fClus,$cClus) = @_;
    return 0 unless ($cClus->[0][6] eq 'C');
    my $fHSP = $fClus->[-1];
    my $cHSP = $cClus->[0];
    return 0 unless ($fHSP->[0] eq $cHSP->[0]);
    my $d = $cHSP->[2] - $fHSP->[1];
    return ($d > 0 && $d <= 20000);
}

sub createSNPSpec {
    my ($seq,$qual) = @_;
    my $spec;
    if ($seq =~ /^-+$/) {
	$spec = $seq;
    } elsif ($seq =~ /^ +$/) {
	$spec = '?';
    } else {
	my @sa = split(//,$seq);
	my @qa = split(//,$qual);
	$spec = $seq;
	$spec =~ s/-//g;
	my $sep = ',';
	foreach my $i (0 .. $#qa) {
	    unless ($sa[$i] eq '-') {
		my $n = $qa[$i] eq ' ' ? 94 : (ord($qa[$i])-33);
		$n = int($n/10); # $n means P(error) <= 10^-$n, etc
		$spec .= "$sep$n";
		$sep = ''; # run multiple quality digits together
	    }
	}
    }
    return $spec;
}

# The score of a SNP (used to choose its colour) is one plus the maximum
# quality digit for any position where the imputed strain differs from the
# reference, or 0 if there is no such position.

sub differenceScore {
    my ($s1,$s2,$q) = @_;
    my $score = 0;
    my $kMax = length($s2)-1;
    $s2 =~ tr/a-z/A-Z/; # should now be canonical
    for (my $k=0; $k<=$kMax; $k++) {
	unless (substr($s1,$k,1) eq substr($s2,$k,1)) {
	    $score = max($score,1+substr($q,$k,1));
	}
    }
    return $score;
}

sub maxOffsetWanted {
    my ($chr,$pUseChr) = @_;
    return undef unless ($pUseChr);
    return undef if ($pUseChr->{$chr} eq 'ALL');
    my $max;
    foreach my $pA (@{$pUseChr->{$chr}}) {
	$max = max($max,$pA->[1]);
    }
    return $max;
}

# $pAlis1: accepted alignments
# $pAlis2: all (sensible) alignments

# Return a version of $pAlis2 with each alignment tagged (in field 7) with
# whether accepted or not.

sub mergeAlilines {
    my ($pAlis1,$pAlis2) = @_;
    my %alis;
    foreach my $r (keys %$pAlis2) {
	my @lst1;
	if (defined $pAlis1->{$r}) {
	    # Start with the accepted ones, and mark them as such (field 7)
	    @lst1 = @{$pAlis1->{$r}};
	    foreach my $pA (@lst1) {
		$pA->[7] = 1;
	    }
	}
	my @lst = @lst1;
	# Now add the rest, unless already present as accepted:
	if ($pAlis2->{$r}) {
	    my @lst2 = @{$pAlis2->{$r}};
	    foreach my $pA (@lst2) {
		unless (alignmentPresent($pA,\@lst1)) {
		    push @lst,$pA;
		}
	    }
	}
	$alis{$r} = \@lst;
    }
    return \%alis;
}

1;
