package SGRP::BuildGbrowseConfig;

use Exporter;
@ISA = qw(Exporter);

@EXPORT = qw(buildGbrowseConfigFile);

use Utility;
use File::Basename;
use Carp;

use strict;

# 0 means no SNP stuff
# 1 means SNP density only
# 2 means overall SNP track too
# 3 means per-strain SNP tracks too -- doesn't work; if we want
#   more detail, run imputed2html.pl and copy the resulting snps dirs over.

my $DO_SNPS = 2; # was 1 before we submitted the paper...

my $BGSNPSUB = <<\END;
sub { my $q = $_[0]->attributes('quality');
      $q =~ tr/A-Z/a-z/;
      $q eq 'confirmed' ? 'black' : 'imputed' ? 'gray' : 'white';
    }
END

my $FGSNPSUB = <<\END;
sub { my $s = $_[0]->score;
      qw(undef brown gray blue blue orange green red)[$s];
    }
END

my $HTSNPSUB = <<\END;
sub { my $s = $_[0]->score;
      $s >= 4 ? 15 : 10;
    }
END

chomp $BGSNPSUB;
chomp $FGSNPSUB;
chomp $HTSNPSUB;

# -------------------------------------------------------------------

# Gbrowse configuration file

# Build the configuration file: first the general part, then a track
# for each strain.

sub buildGbrowseConfigFile {
    my ($sp,$meth,$strSet,$tgtDir) = @_;
    my $conf = "$tgtDir/$sp.$meth.conf";
    if (stamp($conf)) {
	my $pStrains = SGRP::SGRP::getStrains($strSet);
	# Sort them so they go in alphabetic order in the conf file
	@$pStrains = sort @$pStrains;
	my $gh = openToWrite($conf);
	report("writing $conf");
	writeGeneralConfiguration($gh,$sp,$meth);
	writeTrackConfiguration($gh,$sp,$meth,$pStrains);
	$gh->close();
	report();
	unstamp($conf,1);
    }
    return $DO_SNPS;
}

# NOTE: to run on deskpro, we want "<" options in:
# 4,5c4,6
# < db_args           = -adaptor berkeleydb
# <                   -dir     '/nfs/team71/phd/dmc/sgroot/gbdb/cere/methc'
# ---
# > db_args           = -adaptor dbi::mysql
# >                   -dsn gbrowse_${sp}_dmc;hostname=webdbsrv
# >                   -user gbrowsero
# 16,18c17,19
# < stylesheet        = /gbrowse/gbrowse.css
# < buttons           = /gbrowse/images/buttons
# < tmpimages         = /gbrowse/tmp
# ---
# > stylesheet        = /gbrowse/css/gbrowse.css
# > buttons           = /gbrowse/gfx/buttons
# > tmpimages         = /tmp
#
# and copy cere.methc.conf to ~/myperl/conf/gbrowse.conf.
# All the above is done by makeBerkeleyDB.pl cere|para.
#
# Then we have to do (on deskpro):
#  apachectl start
#
# then browse to localhost:8080/cgi-bin/gbrowse/cere/

# NOTE: a [name] (header of data type) can't contain a hyphen,
# apparently! Hence "REFLAR".

# ?Temporarily omitted from the general tracks:
# 
# [PQ boundary]
# feature       = pqboundary
# glyph         = diamond
# point         = 1
# orient        = N
# key           = P:Q boundaries

sub writeGeneralConfiguration {
    my ($gh,$sp,$meth) = @_;
    my %full = qw(cere cerevisiae para paradoxus);
    # Example for cere shows a nice Ty insert in REF and S288c and YGPM only
    my %example = (cere => "chr04:1206300..1213600",
		   para => "chr04:206300..213600"); # need to check!
    my $time = localtime(time); # as scalar
    my @lt = localtime(time); # as array
    my $date = sprintf("%d/%2d/%4d",$lt[3],$lt[4]+1,$lt[5]+1900);
    $date =~ tr/ /0/;
    my $home = $ENV{HOME};
    my $meth2 = ($meth eq 'methc') ? '' : " $meth";
    #my $indelsub = 'sub { my $name = $_[0]->display_name; "?name=$name;class=contig" }';
    my $linksub = 'sub { my $name = $_[0]->display_name; $name =~ s/,.*//; "?name=$name;class=contig" }';
    print $gh <<END;
[GENERAL]
description       = SGRP$meth2 database for S $full{$sp}, created $date
db_adaptor        = Bio::DB::GFF
db_args           = -adaptor dbi::mysql
	  	    -dsn gbrowse_${sp}_dmc;hostname=webdbsrv
                    -user gbrowsero
aggregators       = match
                    processed_transcript
                    profile{level}
category          = Other
reference class   = contig
examples          = $example{$sp}
automatic classes = contig alignment
footer            = <hr><pre>\$Id: sgrp/$sp.$meth.conf, $time, dmc\@sanger.ac.uk \$</pre>
image widths      = 450 640 800 1024
default width     = 800
stylesheet        = /gbrowse/css/gbrowse.css
buttons           = /gbrowse/gfx/buttons
tmpimages         = /tmp
max segment       = 50000
default segment   = 5000
zoom levels       = 100 200 1000 2000 5000 10000 20000 40000 50000
overview bgcolor  = lightgrey
detailed bgcolor  = lightgoldenrodyellow
key bgcolor       = beige
label density     = 50
bump density      = 300
request timeout   = 120

[TRACK DEFAULTS]
glyph             = generic
height            = 10
bgcolor           = lightgrey
fgcolor           = black
font2color        = blue
label density     = 50
bump density      = 300
link              = AUTO

[Transcripts]
feature      = processed_transcript
glyph        = processed_transcript
bgcolor      = peachpuff
description  = 1
key          = Protein-coding genes

[Transposable element]
feature	      = transposable_element:SGD
stranded      = 1
key	      = Transposable elements
bgcolor	      = black

[DNA]
glyph          = dna
global feature = 1
height         = 40
do_gc          = 1
fgcolor        = red
axis_color     = blue
strand         = both
key            = DNA/GC Content

[Translation]
glyph          = translation
global feature = 1
height         = 40
fgcolor        = purple
start_codons   = 0
stop_codons    = 1
translation    = 6frame
key            = 6-frame translation

[REF deletions]
feature       = indel:REF
glyph         = generic
stranded      = 0
bgcolor       = black
height        = 8
link          = $linksub
key           = REF deletions

END
   if ($DO_SNPS) {
      print $gh <<END;
[snpdens]
feature       = profile:sd
glyph         = xyplot
graph_type    = histogram
height        = 50
scale         = left
key           = SNP density

END
  }

   if ($DO_SNPS > 1) {
      print $gh <<END;
[SNP]
feature       = SNP:any
glyph         = lightning
ref_strand    = 1
fgcolor       = $FGSNPSUB
bgcolor       = $BGSNPSUB
height        = $HTSNPSUB
key           = All SNPs

END

  }
   # Subs for species match:
   my $bgsub = 'sub { my $s = $_[0]->score; $s == 100 ? "red" : $s >= 99 ? "orangered" : $s >= 95 ? "darkorange" : ';
   $bgsub .= '$s >= 90 ? "orange" : $s >= 80 ? "gold" : "yellow" }';
   my $linksub = 'sub { my $name = $_[0]->display_name; $name =~ s/,.*//; "?name=$name;class=contig" }';
   foreach my $sp2 (qw(cere para)) {
      my $full = $sp2 eq 'cere' ? 'cerevisiae' : 'paradoxus';
      print $gh <<END;
[S $sp2 match]
feature       = smatch:$sp2
category      = Across strains
glyph         = segments
stranded      = 1
bgcolor       = $bgsub
group_pattern = /\.[pq]1k.*\$/
height        = 4
key           = S $full alignments
END
      if ($sp eq $sp2) {
	  print $gh "link          = $linksub\n\n";
      } else {
          print $gh "link          = \n\n";
      }
  }
}

sub writeTrackConfiguration {
    my ($gh,$sp,$meth,$pStrains) = @_;
    foreach my $str (@$pStrains) {
	writeMatchTrackConfiguration($gh,$sp,$meth,$str);
    }
}

# Score to colour mapping: see colouringOfClone
# $_[0] in callbacks is a Bio::DB::GFF::Feature...
# Colour palette is in Generic-Genome-Browser-1.64/docs/Gbrowse_Color_Palette.pdf
# Note that it's "gray" but "lightgrey" ...

# To group tracks by content instead of strain, change category to:
# Alignments, Coverage, Alignment breaks, SNPs, SNP Density

sub writeMatchTrackConfiguration {
    my ($gh,$sp,$meth,$str) = @_;
    # Ensure $strA doesn't start with digit:
    my ($strA) = avoidInitialDigit($str);
    my $bgsub = 'sub { my $s = $_[0]->score; $s <= 1 ? qw(white red)[$s] : qw(lightgreen blue burlywood brown blue burlywood brown)[int($s/2-1)] }';
    my $fgsub = 'sub { $_[0]->score % 2 ? "red" : "black" }';
    my $htsub = 'sub { ($_[0]->score < 10) ? 4 : 6 }';
    my $linksub = 'sub { my $name = $_[0]->display_name; $name =~ s/\.[pq]1k.*//; "?name=$name;class=contig" }';
    my $indelsub = 'sub { my $name = $_[0]->display_name; "?name=$name;class=contig" }';
    #my $snpsub = 'sub { $_[0]->score > 1 ? "red" : $_[0]->score > 0 ? "hotpink" : "gray" }';
    my $bgbreaksub = 'sub { $_[0]->type > 2 ? "red" : "pink" }';
    my $oribreaksub = 'sub { qw(W E W E N)[$_[0]->score-1] }';
    my $bgindelsub = 'sub { qw(black pink red)[$_[0]->score] }';
    print $gh <<END;
[$str match]
feature       = match:$strA
category      = $str
glyph         = segments
stranded      = 1
bgcolor       = $bgsub
fgcolor       = $fgsub
group_pattern = /\.[pq]1k.*\$/
height        = $htsub
link          = $linksub
key           = $str alignments

[$str level]
feature       = profile:$strA
category      = $str
glyph         = xyplot
graph_type    = histogram
height        = 50
scale         = left
key           = $str level

[${str} indels]
feature       = indel:${strA}
category      = $str
glyph         = generic
bgcolor       = $bgindelsub
key           = $str indels
link          = $indelsub

END
   if ($DO_SNPS > 2) {
       my $snpsub = 'sub { $_[0]->score > 1 ? "red" : $_[0]->score > 0 ? "hotpink" : "gray" }';
       print $gh <<END;
[$str SNP]
feature       = SNP:$strA
category      = $str
glyph         = lightning
ref_strand    = 1
fgcolor       = $FGSNPSUB
bgcolor       = $BGSNPSUB
height        = $HTSNPSUB
key           = $str SNPs

END
   }
}

# Removed from the above:
#[$str snpdens]
#feature       = profile:sd
#category      = $str
#glyph         = xyplot
#graph_type    = histogram
#height        = 50
#scale         = left
#key           = $str SNP density

1;
