package SGRP::BuildGbrowseGFF;

use Exporter;
@ISA = qw(Exporter);

@EXPORT = qw(buildGbrowseGFFFiles printGFFToRegions $CHR_PATTERN buildGbrowseLARAlignmentFile);

use SGRP::Alignment;
use SGRP::AlignmentCluster;
use SGRP::Variables;

use Utility;
use File::Basename;
use Carp;

use strict;

# Min length of external novel sequence to display
my $MIN_NOVEL_LENGTH = 20;

my $MAX_IGNORABLE_ALIGNMENT_SIZE = 4;

use vars qw($CHR_PATTERN);
$CHR_PATTERN = '*'; # 'chr01' or '*'; chr subdir names to build from

sub buildGbrowseGFFFiles {
    my ($pMap,$pRegs,$iDir,$str,$strA,$tgt) = @_;
    # Match (alignment) file
    if (stamp($tgt)) {
	buildGbrowseMatchFile($str,$strA,$iDir,$pMap,$pRegs,$tgt);
	unstamp($tgt);
    }
	
}
	
sub buildGbrowseMatchFile {
    my ($str,$strA,$iDir,$pMap,$pRegs,$tgt) = @_;

    my $pIsGoodRead = getGoodReads($str,dirname($iDir));
    my %alis;
    my %clash;
    my $dpat = $RELEASE_DIR 
	? "$iDir/${CHR_PATTERN}"
	: "$iDir/p?/${CHR_PATTERN}";
    my @files = glob("$dpat/alicoords.gz");
    unless (@files) {
	die("No matches for $dpat/alicoords.gz");
    }
    my $bailOut;
    foreach my $f (@files) {
	unless (isComplete($f)) {
	    report("$f: not complete",1);
	    $bailOut = 1;
	}
    }
    die("Bailing out") if ($bailOut);
    foreach my $f (@files) {
	# acr.txt should have been produced by crlo, as should alicoords.gz
	my $acrBase = $RELEASE_DIR ? 'alignmentClashRates.txt' : 'acr.txt';
	readClashingAlignments(dirname($f) . "/$acrBase",$str,\%clash);
	my $chr = basename(dirname($f));
	readAlicoords($f,$str,$pIsGoodRead,\%alis);
    }
    report("$tgt: writing");
    my $gh = openToWrite($tgt);
    foreach my $cid (sort keys %alis) {
	printCloneGFF($gh,$str,$strA,$cid,$alis{$cid},\%clash,$pMap,$pRegs);
    }
    $gh->close();
    report();
  
}

sub getGoodReads {
    my ($str,$spDir) = @_;
    my $f = $RELEASE_DIR ? "$spDir/strains/$str/renamed.wm" : "$ENV{SGR}/read/$str/renamed.wm";
    my $fh = openToRead($f);
    my %isGoodRead;
    while (<$fh>) {
	chomp;
	my ($r,undef,@rest) = split;
	$isGoodRead{$r} = 1 if (@rest);
    }
    return \%isGoodRead;
}

sub readClashingAlignments {
    my ($f,$str,$pClash) = @_;
    if (-s $f) {
	report("$f: reading");
	my $fh = openToRead($f);
	while (<$fh>) {
	    chomp;
	    if (/ ${str}-/) {
		my @a = split;
		if ($a[0] >= 3) {
		    my $name = $a[4];
		    $name =~ s/\|.*//;
		    $pClash->{$name} = 1;
		}
	    }
	}
	report();
    } else {
	report("$f: missing or empty",1);
    }
}

sub readAlicoords {
    my ($f,$str,$pIsGoodRead,$pAlis) = @_;
    report("$f: reading");
    my $fh = openToRead($f);
    while (<$fh>) {
	chomp;
	my ($read,$chr,undef,$iter,@rest) = split;
	my ($clone,$pq) = ($read =~ /^$str-(.*)\.([pq])1k$/);
	my $qp = ($pq eq 'p') ? 'q' : 'p';
	my $hasMate = defined $pIsGoodRead->{"${str}-${clone}.${qp}1k"} ? 1 : 0;
	my $isReal = ($iter =~ /^\d+$/);
	if ($clone) {
	    # Each quad is for one alignment: clo rlo chi rhi, where
	    # rlo > rhi for C alignment. We ignore alignments of length < 5
	    # (which shouldn't really be there) as they mess up the display.
	    my @quads;
	    for (my $i=0; $i<$#rest; $i+=4) {
		if ($rest[$i+2] - $rest[$i] >= $MAX_IGNORABLE_ALIGNMENT_SIZE) {
		    push @quads,[@rest[$i..$i+3]];
		}
		#$level{$rest[$i]} ++;
		#$level{$rest[$i+2]+1} --;
	    }
	    # Each splitQuad is a list of quads for one (split) alignment for this read.
	    push @{$pAlis->{$clone}{$pq}},splitQuads($chr,$iter,$hasMate,\@quads);
	}
    }
    report();
}

sub splitQuads {
    my ($chr,$iter,$hasMate,$pQuads) = @_;
    my $i=0;
    my @ans;
    while ($i <= $#$pQuads) {
	my $dir = quadDirection($pQuads->[$i]);
	my $j = $i;
	while ($j < $#$pQuads && quadDirection($pQuads->[$j+1]) == $dir && $pQuads->[$j+1][1] == $pQuads->[$j][3]+$dir) {
	    $j++;
	}
	push @ans,{chr => $chr, iter => $iter, quads => [@{$pQuads}[$i..$j]], hasMate => $hasMate};
	$i = $j+1;
    }
    return @ans;
}

sub quadDirection {
    my ($pQuad) = @_;
    my (undef,$rlo,undef,$rhi) = @$pQuad;
    return $rlo < $rhi ? 1 : -1;
}

# "Score" determines colour of alignment. It should show
# how the alignment was placed. Where two numbers are shown,
# second is for clash score >= 3 (alignmentClashRates.pl),
# which should be shown with red border.

#  0     white        poor quality unplaced (read display only)
#  1     red          extra, bisr
#  2,3   lightgreen   extra, close (NOTE change of colour)
#  4,5   blue         extra-chi_late
#  6,7   burlywood    extra-late, no mate
#  8,9   brown        extra-late, mate consistent
#  10,11 blue         mate not placed consistently
#  12,13 burlywood    mate does not exist
#  14,15 brown        mate placed consistently

# Show 9 and below as thin, 10 and above as thick.

sub printCloneGFF {
    my ($gh,$str,$strA,$cid,$pqAlis,$pClash,$pMap,$pRegs) = @_;
    my @alis = ($pqAlis->{p},$pqAlis->{q});
    # 2 for consistent mate, 1 for inconsistent, within each ali
    decideHasMate(@alis);
    foreach my $i (0,1) {
	my $pAli = $alis[$i];
	if ($pAli) {
	    my $pq = $i ? 'q' : 'p';
	    my $r = "${str}-${cid}.${pq}1k";
	    # Each $pH is a (split) alignment
	    foreach my $j (0 .. $#$pAli) {
		my $pH = $pAli->[$j];
		my $chr = $pH->{chr};
		my $iter = $pH->{iter};
		my $pQuads = $pH->{quads};
		my $hasMate = $pH->{hasMate};
		my $fc = quadDirection($pQuads->[0]) > 0 ? 'F' : 'C';
		# Alignment identifier:
		my $aid = "${strA}-$cid.${pq}1k";
		$aid .= "." . ($j+1) if ($#$pAli);
		printGFFToRegions($pMap,$pRegs,
		                  $gh,$chr,$strA,'match',
				  $pQuads->[0][0],$pQuads->[-1][2],
				  undef,$fc,undef,
				  "alignment $aid");
		my $score = decideScoreForColour($iter,$hasMate);
		my @recs;
		foreach my $pQuad (@$pQuads) {
		    my $scoreToUse = $score;
		    $scoreToUse ++ if ($score >= 2 && isClashingAlignment($r,$pQuad,$pClash));
		    push @recs,printGFFToRegions($pMap,$pRegs,
						 $gh,$chr,$strA,'HSP',
						 $pQuad->[0],$pQuad->[2],
						 $scoreToUse,$fc,undef,
						 "alignment $aid",1);
		}
		printConflatedRecords($gh,$chr,$strA,'HSP',"alignment $aid",\@recs);
	    }
	}
    }
}

sub decideHasMate {
    my ($pAli,$qAli) = @_;
    # No mates if only one set exists:
    return unless ($pAli && $qAli);
    # 2 means consistent mates; 1 means inconsistent.
    # We could look this up from clustering, but it's probably
    # easier to work it out here.
    foreach my $pH (@$pAli) {
	foreach my $qH (@$qAli) {
	    my $pDir = quadDirection($pH->{quads}[0]);
	    my $qDir = quadDirection($qH->{quads}[0]);
	    if ($pDir > $qDir) {
		next if decideHasMate1($pH,$qH);
	    } else { 
		next if decideHasMate1($qH,$pH);
	    }
	}
    }
}

sub decideHasMate1 {
    my ($fH,$cH) = @_;
    return unless ($fH->{chr} eq $cH->{chr});
    my ($flo,undef,$fhi) = @{$fH->{quads}[0]};
    my ($clo,undef,$chi) = @{$cH->{quads}[-1]};
    # Assume consistent if separation is between 0 and 25000
    # in the _alignment_ ...
    if ($clo >= $flo && $chi >= $fhi && $fhi-$clo <= 25000) {
	$fH->{hasMate} = $cH->{hasMate} = 2;
	return 1;
    } else {
	return 0;
    }
}

sub printGFFToRegions {
    my ($pMap,$pRegsAll,$gh,$chr,$strA,$type,$lo,$hi,$score,$fc,$phase,$anno,$noPrint,$widen) = @_;
    my $rchr = "REF.$chr";
    my $pRegs = $pRegsAll->{$rchr};
    # Back translate to reference chromosome if pMap is defined. We don't print yet, just
    # return the coords, so we can splice together afterwards
    my @recs;
    my @larInts = larIntersections($pRegs,$lo,$hi);
    my ($refLo,$refHi);
    if (defined $pMap) {
	$refLo = $pMap->backTranslation($rchr,$lo,2)-$widen;
	$refHi = $pMap->backTranslation($rchr,$hi,1)+$widen;
	if ($refHi >= $refLo) {
	    # Don't print single-base level changes
	    unless ($refLo == $refHi && $type eq 'level') {
		if ($noPrint) {
		    push @recs,[$refLo,$refHi,$score,$fc];
		} else {
		    printGFFWorkaround($gh,$chr,$strA,$type,$refLo,$refHi,$score,$fc,$phase,$anno);
		}
	    }
	}
    }
    # See if we intersect any LARs. We hope there are few enough that simple
    # linear search is efficient enough. We print this stuff if $noPrint is <=1,
    # but not 2.
    if ($noPrint <= 1) {
	foreach my $pInt (@larInts) {
	    my ($regId,$regLo,$regHi) = @$pInt;
	    my $name = larName($chr,$regId);
	    printGFFWorkaround($gh,$name,$strA,$type,$regLo,$regHi,$score,$fc,$phase,$anno);
	}
    }
    return @recs;
}

sub larIntersections {
    my ($pRegs,$lo,$hi,$isRef) = @_;
    my @ans;
    #my $i0 = $isRef ? 2 : 0;
    #my $i1 = $isRef ? 3 : 1;
    my $i2 = $isRef ? 0 : 2;
    my $i3 = $isRef ? 1 : 3;
    foreach my $i (0 .. $#$pRegs) {
	# pReg is refLo refHi aliLo aliHi
	my $pReg = $pRegs->[$i];
	if ($pReg->[$i2] > $hi) {
	    return @ans;
	}
	if ($pReg->[$i3] >= $lo) {
	    my $liLo = 1+max($lo-$pReg->[$i2],0);
	    my $liHi = 1+min($hi-$pReg->[$i2],$pReg->[$i3]-$pReg->[$i2]);
	    push @ans,[$i,$liLo,$liHi,$pReg->[$i2]];
	}
    }
    return @ans;
}

sub larName {
    my ($chr,$i) = @_;
    my $name = sprintf("%s_lar%3d",$chr,$i);
    $name =~ tr/ /0/;
    return $name;
}

sub decideScoreForColour {
    my ($iter,$mate) = @_;
    if ($iter eq 'bisr') {
	return 1;
    } elsif ($iter eq 'close') {
	return 2;
    } elsif ($iter eq 'chi_late') {
	return 4;
    } elsif ($iter eq 'late') {
	return ($mate == 2) ? 8 : 6;
    } elsif ($mate == 2) {
	return 14;
    } elsif ($mate == 1) {
	return 10;
    } else {
	return 12;
    }
}

sub isClashingAlignment {
    my ($r,$pQuad,$pClash) = @_;
    my (undef,$rlo,undef,$rhi) = @$pQuad;
    return defined $pClash->{"$r:$rlo:$rhi"};
}

sub conflateRecords {
    my ($pRecs) = @_;
    my @ans;
    foreach my $pA (@$pRecs) {
	# Allow up to a 4 bp gap -- cf
	if (@ans && $ans[-1][1] >= $pA->[0]-$MAX_IGNORABLE_ALIGNMENT_SIZE) {
	    $ans[-1][1] = $pA->[1];
	    $ans[-1][2] = max($ans[-1][2],$pA->[2]);
	    $ans[-1][3] = $ans[-1][3] || $pA->[3]; # fc
	} else {
	    push @ans,$pA; # the default
	}
    }
    return @ans;
}

sub buildGbrowseLARAlignmentFile {
    my ($pStrMap,$pRefMap,$pRefRegsAll,$pDanglesAll,$str,$strA,$g) = @_;
    return unless (stamp($g));
    report("$g: writing");
    my $gh = openToWrite($g);
    my $type = $strA; # was strA_indel
    foreach my $rchr (sort keys %$pRefRegsAll) {
	(my $chr = $rchr) =~ s/^REF.//;
	my $pDangle = $pDanglesAll->{$chr}; # dangles are in ali coords ...
	my $pStrPairs = $pStrMap->alignedSegments("$str.$chr",1);
	# Match for str over the whole chr:
	#printGFFToRegions($pRefMap,$pRefRegsAll,$gh,$chr,$type,'match',
			  #$pStrPairs->[0][0],$pStrPairs->[-1][1],undef,undef,undef,'MUTUAL');
	my $pRefPairs = $pRefMap->alignedSegments("REF.$chr",1);
	my $iLAR = 0;
	my $pRefRegList = $pRefRegsAll->{$rchr};
	my $pRefRegPrev = [0,0,0,0];
	foreach my $iLAR (0 .. $#$pRefRegList) {
	    my $pRefReg = $pRefRegList->[$iLAR];
	    my $pRefRegInter = intermediateRefReg($pRefRegPrev,$pRefReg);
	    printPresenceLines($pRefMap,$pRefRegsAll,$gh,$chr,$type,$pRefPairs,$pStrPairs,$pDangle,$pRefRegInter);
	    printPresenceLines($pRefMap,$pRefRegsAll,$gh,$chr,$type,$pRefPairs,$pStrPairs,$pDangle,$pRefReg,$iLAR);
	    $pRefRegPrev = $pRefReg;
	}
    }
    unstamp($g);
    report();
}

sub intermediateRefReg {
    my ($pRefReg1,$pRefReg2) = @_;
    my ($refLo1,$refHi1,$aliLo1,$aliHi1) = @$pRefReg1;
    my ($refLo2,$refHi2,$aliLo2,$aliHi2) = @$pRefReg2;
    return [$refHi1+1,$refLo2-1,$aliHi1+1,$aliLo2-1];
}

sub printPresenceLines {
    my ($pRefMap,$pRefRegsAll,$gh,$chr,$type,$pRefPairs,$pStrPairs,$pDangle,$pRefReg,$iLAR) = @_;
    my ($name,$annoRefToLAR,$annoLARToRef);
    my ($refLo,$refHi,$aliLo,$aliHi) = @$pRefReg; # The LAR, or inter-LAR region.
    my $noPrintHSP = 2;
    if (defined $iLAR) {
	$name = larName($chr,$iLAR);
	$annoRefToLAR = sprintf("alignment %s",$name);
	$annoLARToRef = sprintf("alignment %s:%d..%d",$chr,$refLo,$refHi);
	$noPrintHSP = 1;
    }
    my @recs = ();
    while (@$pRefPairs && $pRefPairs->[0][1] < $aliLo) {
	shift @$pRefPairs;
    }
    while (@$pStrPairs && $pStrPairs->[0][1] < $aliLo) {
	shift @$pStrPairs;
    }
    #printGFF($gh,$chr,$type,'match',$refLo,$refHi,undef,undef,undef,$annoRefToLAR);
    #printGFF($gh,$name,$type,'match',1,$aliHi-$aliLo+1,undef,undef,undef,$annoLARToRef) if (defined $name);

    # First deal with deletions in the strain, which will be deletions in the LAR (if any) and may
    # also be deletions in ref. Deletions have score zero.
    my @recs;
    foreach my $pPair (deletedRegions($aliLo,$aliHi,$pStrPairs)) {
	push @recs,printGFFToRegions($pRefMap,$pRefRegsAll,$gh,$chr,$type,'indel',
				     @$pPair,0,undef,undef,$annoLARToRef,$noPrintHSP);
    }
    printConflatedRecords($gh,$chr,$type,'indel',$annoRefToLAR,\@recs);

    # Now deal with insertions in the strain.
    @recs = ();
    foreach my $pStrPair (@$pStrPairs) {
	my ($aLo,$aHi) = @$pStrPair;
	last if ($aLo > $aliHi); # outside the LAR
	$aLo = max($aLo,$aliLo);
	$aHi = min($aHi,$aliHi);
	while (@$pRefPairs && $pRefPairs->[0][1] < $aLo) {
	    shift @$pRefPairs;
	}
	foreach my $pTup (insertionsFromStrainRegion($aLo,$aHi,$pRefPairs,$pDangle)) {
	    my ($aLo1,$aHi1,$score,$fc) = @$pTup;
	    # Widen ref regions by 5 here to get them printed...
	    push @recs,printGFFToRegions($pRefMap,$pRefRegsAll,$gh,$chr,$type,'indel',
					 $aLo1,$aHi1,$score,$fc,undef,$annoLARToRef,$noPrintHSP,5);
	}
    }
    printConflatedRecords($gh,$chr,$type,'indel',$annoRefToLAR,\@recs);
}

sub deletedRegions {
    my ($aliLo,$aliHi,$pPairs) = @_;
    unless (@$pPairs) {
	return ([$aliLo,$aliHi]);
    }
    my @ans;
    if ($pPairs->[0][0] > $aliLo) {
	push @ans,regionMoreThan(10,$aliLo,$pPairs->[0][0]-1);
    }
    foreach my $i (1 .. $#$pPairs) {
	return @ans if ($pPairs->[$i][0] > $aliHi);
	push @ans,regionMoreThan(10,$pPairs->[$i-1][1]+1,$pPairs->[$i][0]-1);
    }
    if ($pPairs->[-1][1] < $aliHi) {
	push @ans,regionMoreThan(10,$pPairs->[-1][1]+1,$aliHi);
    }
    return @ans;
}    

sub regionMoreThan {
    my ($lim,$lo,$hi) = @_;
    return ($hi-$lo >= $lim) ? ([$lo,$hi]) : ();
}
	
sub printConflatedRecords {
    my ($gh,$chr,$type,$cat,$anno,$pRecs) = @_;
    foreach my $pA (conflateRecords($pRecs)) {
	my ($refLo,$refHi,$score,$fc) = @$pA;
	printGFFWorkaround($gh,$chr,$type,$cat,$refLo,$refHi,$score,$fc,undef,$anno);
    }
}

sub printGFFWorkaround {
    my ($gh,$chr,$type,$cat,$refLo,$refHi,$score,$fc,$phase,$anno) = @_;
    if ($cat eq 'matchHSP') {
	# To deal with breaks properly; but should not currently happen ...
	printGFF($gh,$chr,$type,'match',$refLo,$refHi,undef,undef,undef,$anno);
	printGFF($gh,$chr,$type,'HSP',$refLo,$refHi,$score,$fc,$phase,$anno) if (defined $anno);
    } else {
	printGFF($gh,$chr,$type,$cat,$refLo,$refHi,$score,$fc,$phase,$anno);
    }
}

# Split [$aLo,$aHi], which are in alignment coords, into subregions that reflect
# whether ref has them too.
sub insertionsFromStrainRegion {
    my ($aLo,$aHi,$pRefPairs,$pDangle) = @_;
    my @tups;
    my $k=0; # index into $pRefPairs
    while ($aLo <= $aHi) {
	if ($k > $#$pRefPairs) {
	    # No more ref regions, so aLo to aHi must be insertion.
	    if ($aHi-$aLo>=10) {
		push @tups,[$aLo,$aHi,2,decideDangle($aLo,$aHi,$pDangle)];
	    }
	    last;
	}
	my ($rLo,$rHi) = @{$pRefPairs->[$k]}; # ref region in ali coords
	if ($aLo < $rLo) {
	    # Insertion to just before rLo.
	    my $top = min($rLo-1,$aHi);
	    if ($top-$aLo>=10) {
		push @tups,[$aLo,$top,2,decideDangle($aLo,$top,$pDangle)];
	    }
	    $aLo = $top+1;
	} elsif ($rHi < $aHi) {
	    # Non-insertion up to rHi, then increment $k.
	    #push @tups,[$aLo,$rHi,1,undef];
	    $aLo = $rHi+1;
	    $k ++;
	} else {
	    # Whole thing is non-insertion.
	    #push @tups,[$aLo,$aHi,1,undef];
	    last;
	}
    }
    return @tups;
}

sub decideDangle {
    my ($lo,$hi,$pDangle) = @_;
    if ($pDangle->{$lo} eq 'L') {
	return 'C'; # left-pointing
    } elsif ($pDangle->{$hi} eq 'R') {
	return 'F'; # right-pointing
    } else {
	return undef;
    }
}

1;
