package SGRP::BuildGbrowseReference;

use Exporter;
@ISA = qw(Exporter);

@EXPORT = qw(buildGbrowseReferenceFiles buildGbrowseSpeciesAlignmentFile);

use SGRP::Variables;
use SGRP::BuildGbrowseGFF; # for printGFFToRegions

use Utility;
use File::Basename;
use Carp;

use strict;

# -------------------------------------------------------------------

# Reference genome stuff

sub buildGbrowseReferenceFiles {
    my ($eDir,$pMap,$pRegs,$sp,$tgtDir) = @_;
    my $spFA = "$tgtDir/$sp.fa";
    my $refDir = $RELEASE_DIR ? "$RELEASE_DIR/$sp/ref" : "$ROOT/ref/$sp";
    copyReferenceFasta($refDir,$spFA);
    buildGbrowseContigFile($pRegs,"$refDir/genome.sz","$tgtDir/contigs.gff");
    my $spGFF = "$tgtDir/$sp.gff";
    my $ggff = $RELEASE_DIR ? "$RELEASE_DIR/$sp/ref/genomeAligned.gff" : "$eDir/assembly/genome.gff";
    buildGbrowseReferenceFile($pMap,$pRegs,$ggff,$spGFF);
}

sub copyReferenceFasta {
    my ($srcDir,$tgt) = @_;
    if (stamp($tgt)) {
	my $src;
	foreach my $f ($tgt,"$srcDir/genome.fa") {
	    if (-s $f) {
		$src = $f;
		last;
	    }
	}
	unless ($src) {
	    print STDERR "WARNING: cannot find genome.fa in $srcDir";
	    return;
	}
	report("writing $tgt");
	copyReferenceFasta1($src,$tgt);
	report();
	unstamp($tgt,1);
    }
}

sub copyReferenceFasta1 {
    my ($src,$tgt) = @_;
    my %useChr;
    my $fh = openToRead($src);
    my $gh = openToWrite($tgt);
    my $active;
    my %useChr;
    while (<$fh>) {
	if (/^>(\S*)/) {
	    chomp (my $chr = $1);
	    $active = $chr !~ /^vc_/;
	    s/\[.*//; # throw away [...] stuff in para.fa
	}
	print $gh $_ if ($active);
    }
    $gh->close();
}

sub buildGbrowseContigFile {
    my ($pRegs,$src,$tgt) = @_;
    if (stamp($tgt)) {
	my $fh = openToRead($src);
	report("writing $tgt");
	my $gh = openToWrite($tgt);
	while (<$fh>) {
	    chomp;
	    my ($id,$sz) = split;
	    unless ($id =~ /^vc_/ || $id eq 'ALL') {
		printGFF($gh,$id,'SGRP','contig',1,$sz,undef,undef,undef,"contig $id");
	    }
	}
	foreach my $rchr (keys %$pRegs) {
	    my $pA = $pRegs->{$rchr};
	    my $chr = $rchr;
	    $chr =~ s/^REF\.//;
	    foreach my $i (0 .. $#$pA) {
		my ($refLo,$refHi,$aliLo,$aliHi) = @{$pA->[$i]};
		my $name = SGRP::BuildGbrowseGFF::larName($chr,$i);
		printGFF($gh,$name,'sgrp','contig',1,$aliHi-$aliLo+1,undef,undef,undef,
			 "contig $name; refregion $refLo..$refHi");
	    }
	}		
	$gh->close();
	unstamp($tgt,1); # not clear why this fails!
	report();
    }
}

# This builds (pruned) cere.gff or in principle para.gff.

sub buildGbrowseReferenceFile {
    my ($pMap,$pRegs,$src,$tgt) = @_;
    my %trans;
    unless (-s $src) {
	# This will happen for paradoxus:
	print "WARNING: missing or empty: $src\n";
	touch($tgt) unless (-f $tgt); # make sure it's there; empty is better than missing
	return;
    }
    return unless (stamp($tgt));
    my %wantKey = map {$_ => 1} qw(gene mRNA 5'-UTR 3'-UTR CDS transposable_element);
    my $fh = openToRead($src);
    my $gh = openToWrite($tgt);
    while (<$fh>) {
	s/^chrM/chrm/; # lower-case mitochondrion name
	my @a = split;
	if ($wantKey{$a[2]}) {
	    $a[-1] =~ s/Note=/Description=/; # rename, or it's shown in the display
	    # Get attributes into right format:
	    $a[-1] =~ s/%20/ /g;
	    $a[-1] =~ s/%2C/,/g;
	    $a[-1] =~ s/%3B/,/g; # should be ";" but that would confuse things
	    $a[-1] =~ tr/=/ /;
	    # If we have a "gene" attribute, set the "ID" attribute to be
	    # that. (Maybe a bit dangerous in general ...)
	    if ($a[-1] =~ /;gene ([^;]+);/) {
		my $niceID = $1;
		$a[-1] =~ s/ID [^;]+/ID $niceID/;
		$a[-1] =~ s/Parent [^;]+/Parent $niceID/;
	    }
	    printGFFToRegions($pMap,$pRegs,$gh,@a);
	}
    }
    $gh->close();
    unstamp($tgt);
}

sub buildGbrowseSpeciesAlignmentFile {
    my ($selfDir,$tgtFile) = @_;
    return unless (stamp($tgtFile));
    report("writing $tgtFile");
    my $gh = openToWrite($tgtFile);
    foreach my $sp2 (qw(cere para)) {
	foreach my $f (glob("$selfDir/*.$sp2.ali")) {
	    my $fh = openToRead($f);
	    buildGBrowseSpeciesAlignmentFile1($fh,$sp2,$gh);
	}
    }
    unstamp($tgtFile);
    report();
}

sub buildGBrowseSpeciesAlignmentFile1 {
    my ($fh,$sp2,$gh) = @_;
    while (<$fh>) {
	chomp;
	my @a = split;
	printGFF(
	    $gh,
	    $a[2], # from chr
	    $sp2,'smatch', # smatch means species-match
	    min($a[4],$a[5]), # extent of match
	    max($a[4],$a[5]), # ditto
	    $a[10], # percent ident
	    $a[8], # FC i.e. +-
	    undef,
	    sprintf("alignment %s:%d..%d, %s%%",
		    $a[3], # to chr
		    $a[6],$a[7],$a[10]));
    }
}

1;
