package SGRP::BuildGbrowseSNP;

use Exporter;
@ISA = qw(Exporter);

@EXPORT = qw(buildGbrowseSNPFiles);

use Utility;
use SGRP::BuildGbrowseGFF; # for printGFFToRegions and $CHR_PATTERN
use SGRP::Variables;
use File::Basename;
use Carp;

use strict;

# For SNP file construction:
my %SIMPLIFIED = qw(A A C C G G T T - -
		    a A c C g G t T 0 - 1 - 2 - 3 - 4 - 5 - 6 - 7 - 8 - 9 -);

my %TRANSLATION = ( TTT => 'Phe', TTC => 'Phe', TTA => 'Leu', TTG => 'Leu',
		    TCT => 'Ser', TCC => 'Ser', TCA => 'Ser', TCG => 'Ser', 
		    TAT => 'Tyr', TAC => 'Tyr', TAA => 'STP', TAG => 'STP', 
		    TGT => 'Cys', TGC => 'Cys', TGA => 'STP', TGG => 'Trp', 
		    CTT => 'Leu', CTC => 'Leu', CTA => 'Leu', CTG => 'Leu', 
		    CCT => 'Pro', CCC => 'Pro', CCA => 'Pro', CCG => 'Pro', 
		    CAT => 'His', CAC => 'His', CAA => 'Gln', CAG => 'Gln', 
		    CGT => 'Arg', CGC => 'Arg', CGA => 'Arg', CGG => 'Arg', 
		    ATT => 'Ile', ATC => 'Ile', ATA => 'Ile', ATG => 'Met',
		    ACT => 'Thr', ACC => 'Thr', ACA => 'Thr', ACG => 'Thr',
		    AAT => 'Asn', AAC => 'Asn', AAA => 'Lys', AAG => 'Lys', 
		    AGT => 'Ser', AGC => 'Ser', AGA => 'Arg', AGG => 'Arg',
		    GTT => 'Val', GTC => 'Val', GTA => 'Val', GTG => 'Val', 
		    GCT => 'Ala', GCC => 'Ala', GCA => 'Ala', GCG => 'Ala', 
		    GAT => 'Asp', GAC => 'Asp', GAA => 'Glu', GAG => 'Glu', 
		    GGT => 'Gly', GGC => 'Gly', GGA => 'Gly', GGG => 'Gly');

sub buildGbrowseSNPFiles {
    my ($doSNPs,$pMap,$pRegs,$sp,$fDir,$tgtDir) = @_;
    my $pat = $RELEASE_DIR ? "$fDir/${CHR_PATTERN}" : "$fDir/p?/${CHR_PATTERN}";
    my ($pCDS,$pFA);
    # Go through all the chromosome (or contig) dirs containing 
    # sorted.gz and imputed.gz files, and build SNP gff files from them.
    # Do them smallest first, to make testing easier.
    my $suf = 'z'; # or could be y ...
    my $sBase = $RELEASE_DIR ? 'sequenced.gz' : "sorted_${suf}.gz";
    my $iBase = $RELEASE_DIR ? 'imputed.gz' : "imputed_${suf}.gz";
    foreach my $srcDir (sortDirsByIncreasingFileSize($pat,$sBase)) {
	my $sFile = "$srcDir/$sBase";
	my $iFile = "$srcDir/$iBase";
	if (-s $sFile && -s $iFile) {
	    my $chr = basename($srcDir);
	    my $tgt1 = "$tgtDir/${chr}_snps.gff";
	    if (stamp($tgt1)) {
		report("processing SNPs (for any strain) in $srcDir/{$sBase,$iBase}.gz");
		# Only do now, to save time if not needed:
		unless ($pCDS) {
		    $pCDS = readCDSLines("$tgtDir/$sp.gff");
		    $pFA = readFALines("$tgtDir/$sp.fa");
		}
		extendGbrowseSNPFiles($doSNPs,$pMap,$pRegs,$sp,$chr,$pFA->{$chr},$pCDS->{$chr},$sFile,$iFile,$tgtDir,$tgt1);
		report();
		unstamp($tgt1,1);
	    }
	}
    }
}

sub sortDirsByIncreasingFileSize {
    my ($pat,$base) = @_;
    my %size;
    foreach my $srcDir (glob($pat)) {
	my $sz = -s "$srcDir/$base";
	$size{$srcDir} = $sz if ($sz);
    }
    return (sort {$size{$a} <=> $size{$b}} keys %size);
}

sub readCDSLines {
    my ($f) = @_;
    my %cds;
    my $fh = openToRead($f);
    while (<$fh>) {
	chomp;
	my ($chr,undef,$type,$lo,$hi,undef,$strand,$phase) = split;
	next unless ($type eq 'CDS');
	push @{$cds{$chr}},[$lo,$hi,$strand,$phase,getGeneName($_)];
    }
    while (my ($chr,$pA) = each %cds) {
	my @sorted = sort {$a->[0] <=> $b->[0] || $a->[1] <=> $b->[1]} @$pA;
	$cds{$chr} = \@sorted;
    }
    return \%cds;
}

sub getGeneName {
    my ($line) = @_;
    foreach my $key (qw(gene Name)) {
	if ($line =~ /${key}[= ]([^;]+)/) {
   	    return $1;
        }
    }
    return undef;
}

sub readFALines {
    my ($f) = @_;
    my %fa;
    my $fh = openToRead($f);
    my ($chr,$seq);
    while (<$fh>) {
	chomp;
	if (/^>(\S+)/) {
	    $fa{$chr} = $seq if ($seq);
	    $chr = $1;
	    $seq = "";
	} else {
	    $seq .= $_;
	}
    }
    $fa{$chr} = $seq if ($seq);
    return \%fa;
}

sub extendGbrowseSNPFiles {
    my ($doSNPs,$pMap,$pRegs,$sp,$chr,$refSeq,$pCDS,$sFile,$iFile,$tgtDir,$tgt1) = @_;
    my $gh;
    $gh = openToWrite($tgt1) if ($doSNPs > 1);
    my @snpCount; # index is offset//100
    my $sh = openToRead($sFile);
    my $ih = openToRead($iFile);
    my $pSStrains = strainsFromHeader($sh);
    my $pIStrains = strainsFromHeader($ih);
    my $off = 1; # offset in ALIGNMENT coords
    chomp (my $sPrev = <$sh>);
    chomp (my $iPrev = <$ih>);
    my $refLen = length($refSeq);
    my $pStrainMap = strainMap($pSStrains,$pIStrains);
    while (1) {
	my ($psSNPs,$piSNPs);
	my $delta; # number of lines read
	my $r = substr($sPrev,0,1); # ref value
	$r =~ s/[0-9]/-/;
	($delta,$iPrev,$piSNPs) = readSNPs($iPrev,$ih);
	last unless ($iPrev);
	(undef,$sPrev,$psSNPs) = readSNPs($sPrev,$sh,$delta);
	last unless ($sPrev);
	$psSNPs = remapPsSNPs($psSNPs,$pStrainMap);
	my $ctx;
	# Offset is from 1 ...
	if ($off < 3) {
	    $ctx = ('N' x (3-$off)) . substr($refSeq,0,$off+2);
	} elsif ($refLen-$off < 3) {
	    $ctx = substr($refSeq,$off-3) . ('N' x ($off+2-$refLen));
	} else {
	    $ctx = substr($refSeq,$off-3,5);
	}
	my $real = writeSNPMarker($doSNPs,$pMap,$pRegs,$sp,$chr,$ctx,$pCDS,$r,$psSNPs,$piSNPs,$off,$gh,
				  $pIStrains);
	$snpCount[int($off/100)] ++ if ($real);
	$off += $delta;
    }
    # Output SNP densities .. $i is strain index ...
    my $tgt2 = "$tgtDir/${chr}_snpdens.gff";
    $gh = openToWrite($tgt2);
    foreach my $j (2 .. $#snpCount-2) {
	my $den = findSNPDensity(\@snpCount,$j);
	printGFFToRegions($pMap,$pRegs,$gh,$chr,'sd','level',
			  $j*100-49,$j*100+50,$den,undef,undef,"sd level");
    }
}

sub strainsFromHeader {
    my ($fh) = @_;
    my $hdr = <$fh>;
    chomp $hdr;
    $hdr =~ tr/A-Z/a-z/;
    my @strains = split(/ /,substr($hdr,1));
    return \@strains;
}

sub strainMap {
    my ($pSub,$pSuper) = @_;
    my %pos;
    foreach my $i (0 .. $#$pSuper) {
	$pos{$pSuper->[$i]} = $i;
    }
    my @map;
    foreach my $j (0 .. $#$pSub) {
	$map[$j] = $pos{$pSub->[$j]};
    }
    return \@map;
}

# $line is the line (for the next reference nucleotide) left over from
# the last line-read from $fh while we were searching for insertions.

# snpDescriptions returns a list (one element per non-reference strain)
# of tuples of the form [varies,seq,qual] for the current segment,
# where varies=1 if the seq varies from the reference, seq is the 
# sequence (nucleotide posn and 0 or more insertion positions) for
# the strain and qual is the quality scores.

sub readSNPs {
    my ($line,$fh,$toGet) = @_;
    my ($next,$lMax,$pLines) = readSegment($line,$fh,$toGet);
    my $pVaries = variationColumns($lMax,$pLines);
    return (scalar(@$pLines),$next,snpDescriptions($lMax,$pLines,$pVaries));
}

sub readSegment {
    my ($line,$fh,$toGet) = @_;
    # @lines will become $line plus any following insertion lines.
    my @lines = ($line);
    my $next;
    my $lMax = length($line); # max length of the lines in @lines
    while (<$fh>) {
	chomp;
	s/^[a-z]/-/; # turn lower case for ref into gap (a few cases in imp!)
	my $done;
	if (defined $toGet) {
	    $toGet --;
	    $done = ($toGet == 0);
	} else {
	    $done = /^[^-0-9=]/;
	}
	if ($done) {
	    # We'll hand this back to be the next $line on the next call:
	    $next = $_;
	    last;
	} else {
	    push @lines,$_;
	    $lMax = max($lMax,length($_));
	}
    }
    return ($next,$lMax,\@lines);
}

sub variationColumns {
    my ($lMax,$pLines) = @_;
    my @varies;
    # refVal on each line is always upper case or '-'. If ref value
    # is such, we use that; else we use the majority value, breaking
    # ties by first occurrence.
    my @refVal;
    foreach my $i (0 .. $#$pLines) {
	$refVal[$i] = decideReferenceValue($pLines->[$i]);
    }
    my $isPoly;
    for (my $k=2; $k<$lMax; $k+=2) {
	foreach my $i (0 .. $#$pLines) {
	    my $r = $refVal[$i];
	    if ($r) {
		my $line = $pLines->[$i];
		my $v = substr($line,$k,1);
		unless ($v eq ' ' || $v eq '=' || $r eq $SIMPLIFIED{$v}) {
		    $varies[$k] = $r; # store ref val here for snpDescriptions
		    $isPoly = 1;
		    last;
		}
	    }
	}
    }
    return $isPoly && \@varies;
}

sub decideReferenceValue {
    my ($line) = @_;
    my %count;
    my $len = length($line);
    for (my $k=0; $k<$len; $k+=2) {
	my $v = substr($line,$k,1);
	if ($v =~ /[-A-Z]/) {
	    return $v if ($k == 0); # reference is always preferred
	    $count{$v} ++;
	}
    }
    return undef unless (%count); # might occasionally happen with v lo quality line
    # Otherwise, choose highest scorer...
    my ($max) = sort {$b <=> $a} (values %count);
    my @cands;
    foreach my $v (keys %count) {
	push @cands,$v if ($count{$v} == $max);
    }
    return $cands[0] if (@cands == 1); # unique highest
    # Else return first-occurring highest scorer
    for (my $k=2; $k<$len; $k+=2) {
	my $v = substr($line,$k,1);
	return $v if ($count{$v} == $max);
    }
}

sub snpDescriptions {
    my ($lMax,$pLines,$pVaries) = @_;
    return undef unless ($pVaries);
    my @snps;
    for (my $k=2; $k<$lMax; $k+=2) {
	my ($seq,$qual);
	foreach my $line (@$pLines) {
	    $seq .= substr($line,$k,1);
	    $qual .= substr($line,$k+1,1);
	}
	# Change numbers to hyphens, but leave lower-cased nucleotides alone
	$seq =~ tr/0-9/-/;
	# Cut down trailing sequences of hyphens
	if ($seq =~ /^(.+)-+$/) {
	    $seq = $1;
	    $qual = substr($qual,0,length($seq));
	}
	push @snps,[$pVaries->[$k],$seq,$qual];
    }
    return \@snps;
}

sub remapPsSNPs {
    my ($psSNPs,$pStrainMap) = @_;
    return undef unless (defined @$psSNPs);
    my @ans;
    foreach my $i (0 .. $#$pStrainMap) {
	$ans[$pStrainMap->[$i]] = $psSNPs->[$i];
    }
    return \@ans;
}

sub writeSNPMarker {
    my ($doSNPs,$pMap,$pRegs,$sp,$chr,$ctx,$pCDS,$r,$psSNPs,$piSNPs,$off,$gh,$pStrains) = @_;
    return undef unless ($psSNPs);
    @$piSNPs = () unless ($piSNPs);
    # Record what SNPs (or polymorphic sequences) we see...
    my @tups;
    # $i will be strain index (for non-reference strains)
    my $iMax = max($#$psSNPs,$#$piSNPs);
    $r = undef if ($r eq '=');
    my $real;
    for (my $i=0; $i<=$iMax; $i++) {
	recordTuple(0,$i,\@tups,$psSNPs->[$i]);
	my $r1 = recordTuple(2,$i,\@tups,$piSNPs->[$i]);
	if ($r1) {
	    $real = 1;
	    $r = $r1 unless (defined $r);
	}
    }
    if ($real && $doSNPs > 1) {
	my $pCDSHere = findSpanningCDS($pMap,$chr,$off,$pCDS);
	writeSNPMarker1($pMap,$pRegs,$sp,$ctx,\@tups,$chr,$pCDSHere,$off,$r,$gh,$pStrains);
    }
    return $real;
}

sub recordTuple {
    my ($imp02,$str,$pTups,$pTrip) = @_;
    return unless (defined @$pTrip);
    my ($vary,$seq,$qual) = @$pTrip;
    if ($vary) {
	$pTups->[$str][$imp02] = $seq;
	$pTups->[$str][$imp02+1] = $qual;
    }
    return $vary; # the ref or consensus value ...
}

sub findSpanningCDS {
    my ($pMap,$chr,$off,$pCDS) = @_;
    my $roff1 = $pMap->backTranslation("REF.$chr",$off,1);
    my $roff2 = $pMap->backTranslation("REF.$chr",$off,2);
    if ($roff1 == $roff2) { # i.e. if it's in ref
	foreach my $pA (@$pCDS) {
	    # If we wanted to spot splice site deletions, we'd have to
	    # extend by 2 bases each side here for certain exons only...
	    if ($pA->[0] <= $roff1 && $pA->[1] >= $roff2) {
		return $pA;
	    }
	}
    }
    return undef; # intergenic or large insertion into gene
}

# Each SNP is displayed as an allele tower. Top is reference value;
# then we have up to two alternative values, ordered by
# differenceScore (see below).

# $pCDS is [lo,hi,strand,phase] for the CDS spanning this offset, if any.
# $ctx is the five ref nucleotides centered on $off (padded with N's at ends of chr).

sub writeSNPMarker1 {
    my ($pMap,$pRegs,$sp,$ctx,$pTups,$chr,$pCDS,$off,$r,$gh,$pStrains) = @_;
    my ($real,%scoreAll,@score,@strSeq);
    # $str is INDEX into strain list ...
    my $title = snpTitle($pMap,$chr,$off);
    return unless ($title);
    my $refSpec =  "$title; MAJ $r";
    # Make the main SNP entry
    my $spec = $refSpec;
    my ($sScoreMax,$iScoreMax) = (0,0);
    my ($sStrandAny,$iStrandAny);
    my ($codonPos,%codonItem);

    foreach my $str (0 .. $#$pTups) {

	next unless ($pTups->[$str]);

	my ($sseq,$squal,$iseq,$iqual) = @{$pTups->[$str]};
	$sseq = undef if ($sseq eq $r);
	$iseq = undef if ($iseq eq $r);
	my ($sScore,$sStrand,@sArr,$iScore,$iStrand,@iArr);
	($sScore,$sStrand,@sArr) = snpTypeScore($sseq,$ctx,$off,$pCDS) if ($sseq);
	($iScore,$iStrand,@iArr) = snpTypeScore($iseq,$ctx,$off,$pCDS) if ($iseq);

	my $outScore = $iScore || $sScore;
	$sScoreMax = max($sScoreMax,$sScore);
	$iScoreMax = max($iScoreMax,$iScore);
	$sStrandAny = $sStrand if ($sStrand);
	$iStrandAny = $iStrand if ($iStrand);
    }

    my $outScore = $iScoreMax || $sScoreMax;
    my $strand = $iStrandAny || $sStrandAny;
    if (defined $codonPos) {
	my $trio = 'nnn';
	substr($trio,$codonPos,1) = 'N';
	my $strand = 
	$spec .= "; Codon-position $trio in $pCDS->[4] on $strand strand";
    }
    foreach my $item (sort keys %codonItem) {
	$spec .= "; Codon-change $item";
    }
    # "Imputed" should never be used here, as the SNP has to be sequenced somewhere!
    $spec .= sprintf("; Quality %s",
		     $iScoreMax ? ($sScoreMax ? 'Confirmed' : 'Imputed') : 'Discounted');
    $spec .= sprintf("; Effect %s",snpTypeScoreKeyword($outScore), $pCDS);
    my (undef,$off1,$off2) = split(/-/,$title);
    my $off12 = $off2 ? "$off1.$off2" : $off1;
    # Choice of file is based on ALIGNMENT offset; name of SNP is based
    # on REF offset.
    my $sev = sprintf("%7d",$off);
    $sev =~ tr/ /0/;
    # Run imputed2html.pl to create the /snps/ ... files...
    $spec .= sprintf("; Context <a href=\"/snps/%s/%s/%s/%s000.html#o%s\">here</a>",
		     $sp,$chr,substr($sev,0,2),substr($sev,2,2),$off12);

    printGFFToRegions($pMap,$pRegs,$gh,$chr,'ANY','SNP',$off,$off,$outScore,$strand,undef,$spec);
}

# $off is in alignment coords. Return undef if negative length...
sub snpTitle {
    my ($pMap,$chr,$off) = @_;
    # Find offset in ref coords; go to left of gap
    my $off1 = $pMap->backTranslation("REF.$chr",$off,1);
    # Find where left-of-gap comes to in alignment coords
    my $off2 = $pMap->translation("REF.$chr",$off1);
    $chr =~ s/^chr//;
    $chr =~ s/^c//;
    $chr =~ s/^scplasm/pl/;
    # Prefix SNP titles with s, so we don't start them with digits:
    if ($off == $off2) {
	return sprintf("SNP s%s-%d",$chr,$off1);
    } else {
	return sprintf("SNP s%s-%d-%d",$chr,$off1,$off-$off2);
    }
}

sub snpTypeScore {
    my ($seq,$ctx,$off,$pCDS) = @_;
    if ($pCDS) {
	my ($refCodon,$strCodon,$strand,$pos012) = findCodons($seq,$ctx,$off,$pCDS);
	my $refAA = $TRANSLATION{$refCodon} || '***';
	my $strAA = $TRANSLATION{$strCodon} || '***';
	my $strStop = ($strAA eq 'STP');
	unless ($strStop) {
	    for (my $k=3; $k<=length($strAA)-3; $k+=3) {
		if ($TRANSLATION{substr($strCodon,$k,3)} eq 'STP') {
		    $strStop = 1;
		    last;
		}
	    }
	}
	my $score;
	if ($refAA eq $strAA) {
	    $score = 2; # synonymous
	} elsif ($strStop) {
	    $score = 7; # stop codon creation (or insertion)
	} elsif ($refAA eq 'STP') {
	    $score = 6; # stop codon deletion
	} elsif ($strAA eq '***') {
	    $score = length($strCodon)%3 ? 5 : 3; # indel, or nonsyn if divides by 3
	    # We don't bother with start-deletion (score 4) for now ...
	} else {
	    $score = 3;
	}
	return ($score,$strand,$pos012,$refCodon,$strCodon);
    } else {
	return (1); # non-coding
    }
}

sub findCodons {
    my ($seq,$ctx,$off,$pCDS) = @_;
    # Phase is no. of nucs in the codon already used up, i.e.
    # phase 1 means exon starts with last 2 nucs of a codon.
    my ($lo,$hi,$strand,$phase) = @$pCDS;
    $seq =~ tr/acgt/ACGT/;
    $ctx =~ tr/acgt/ACGT/;
    my $rel;
    if ($strand eq '+') {
	$rel = $off-$lo;
    } else {
	$rel = $hi-$off;
	$ctx =~ tr/ACGT/TGCA/;
	$ctx = reverse($ctx);
	$seq =~ tr/ACGT/TGCA/;
	$seq = reverse($seq);
    }
    # Position of $off in the codon: 0 means first position
    my $pos = ($rel+$phase)%3;
    # Position in $ctx that we want to extract from
    my $ctxPos = 2-$pos;
    my $refCodon = substr($ctx,$ctxPos,3);
    my $strCodon = $refCodon;
    substr($strCodon,$pos,1) = $seq;
    return ($refCodon,$strCodon,$strand,$pos);
}

sub snpTypeScoreKeyword {
    my ($score) = @_;
    my @a = (undef,'Non-coding','Synonymous','Non-synonymous','Start codon deletion','Indel',
	     'Stop codon deletion','Stop codon creation');
    return $a[$score];
}

sub findSNPDensity {
    my ($pC,$jTgt) = @_;
    my $sum = 0;
    my @pascal = (1,4,6,4,1);
    foreach my $d (0 .. 4) {
	my $j = $jTgt-2+$d;
	$sum += $pC->[$j]*$pascal[$d];
    }
    return $sum/1600;
}

1;
