package SGRP::CombineLayoutByStrain;

use Exporter;
@ISA = qw(Exporter);

@EXPORT = qw(combineLayoutByStrain);

use SGRP::Variables;

use Utility;
use File::Basename;

use Carp;

use strict;

my $LOG2 = log(2);
my $LOG3 = log(3);
my $LOG10 = log(10);

my $FLOOR = -log(0.75); #-10*log(0.75)/log(10);

my $ZERO = ord('!');

my $KEEP_IN_SYNC = 2; # 2: don't even introduce new ">" lines.

my $IS_ALL_SOLEXA;

# From rqpw.gz, creates:
#
# clbs.gz
# clbs_clashes.gz
# clbs_clash_stats.gz

# Each line in clbs_clashes.gz is of the form
#
# nuc1 ev1 nuc2 ev2 ... : off.suboff strain refVal
#
# where at the specified offset (chromosomal off, plus suboff'th gap) the specified
# strain has possible values nucI each with quality scores evI, and the reference
# is refVal. We have a clash if the nucI's disagree with each other and/or with
# refVal.

sub combineLayoutByStrain {
    my ($species,$filtMeth,$strainSet,@parts) = @_;
    unless (@parts) {
	confess("Usage: $0 clbs species filtMeth strainSet part1 part2 ...");
    }
    my ($iter,$dir) = getCurrentIteration($species,$filtMeth,$strainSet,'p?');
    $IS_ALL_SOLEXA = -f ("$dir/isAllSolexa"); # for imputation paper
    foreach my $part (@parts) {
	$part = "p$part" if ($part =~ /^[0-9]$/);
	foreach my $chrDir (glob("$dir/$part/*")) {
	    combineLayoutByStrain1($chrDir);
	}
    }
}

sub combineLayoutByStrain1 {
    my ($dir) = @_;
    my $src = "$dir/rqpw.gz";
    return unless (isComplete($src,1));
    my $tgt = "$dir/clbs.gz";
    if (stamp($tgt)) {
	my $tgtCla = "$dir/clbs_clashes.gz"; # h
	my $tgtCls = "$dir/clbs_clash_stats.gz"; # s
	report("creating $tgt",1);
	combineLayoutByStrain2($src,$tgt,$tgtCla,$tgtCls);
	unstamp($tgt);
	report("created $tgt",1);
    }
}

sub combineLayoutByStrain2 {
    my ($f,$g,$h,$s) = @_;
    my $fh = openToRead($f);  # rqpw.gz
    my $gh = openToWrite($g); # clbs.gz
    # For debugging:
    #$h =~ s/\.gz$//;
    my $hh = openToWrite($h); # clbs_clashes.gz
    #autoflush $hh 1;
    my (@inLayout,@outLayout,%outPosition);
    # Offset in alignment, from 1, counting all positions.
    my $off = 0;
    # Offset in reference, from 1, counting ref upper-case nucs only
    my $offr = 0; 
    my %clashCount;
    # Don't risk memory overflow:
    my $pClashCount = (-s $f < 100000000) ? \%clashCount : undef;
    while (<$fh>) {
	chomp;
	if (/^>REF/) {
	    print $gh "$_\n" if ($gh);
	} elsif (/^>(\S+) (\d+)/) {
	    my ($strain,$posn) = ($1,$2);
	    $strain =~ s/-.*//; # forget the read ID
	    $strain =~ s/YGPM.*/YGPM/; ## special case for dodgy YGPM ones
	    $inLayout[$posn] = $strain;
	    unless (defined $outPosition{$strain}) {
		push @outLayout,$strain;
		$outPosition{$strain} = $#outLayout;
		printf($gh ">%s %d\n",$strain,$#outLayout) if ($gh);
	    } elsif ($KEEP_IN_SYNC && $gh) {
		printf($gh ">%s\n"); # to keep lines in sync
	    }
	} else {
	    # it's a nucleotide line ... first increment offsets
	    if ($off % 100000 == 0) {
		report("$f: done $off lines",1);
	    }
	    $off ++; # so off is one-offset (it was initialized to 0 above)
	    $offr ++ if (/^[A-Z]/);
	    # Combine data for $gh, and print clashes to $hh as side effect
	    my $out = combineLine($_,\@inLayout,\@outLayout,$off,$offr,$hh,$pClashCount);
	    unless ($out eq '>' && !$KEEP_IN_SYNC) {
		print $gh $out . "\n" if ($gh);
	    }
	}
    }
    $gh->close() if (defined $gh);
    if ($pClashCount) {
	my $sh = openToWrite($s); # clbs_clash_stats.gz
	foreach my $str (sort keys %clashCount) {
	    foreach my $nqs (0,1) {
		my $pH = $clashCount{$str}[$nqs];
		if (defined %$pH) {
		    foreach my $pat (sort keys %$pH) {
			printf($sh "%s %d %s :: %d %d\n",
			       $str,$nqs,$pat,$pH->{$pat}[0],$pH->{$pat}[1]);
		    }
		}
	    }
	}
	$sh->close();
    }
}

sub combineLine {
    my ($line,$pInLay,$pOutLay,$off,$offr,$hh,$pClashCount) = @_;
    # No change if no strains (efficiency):
    return $line unless ($line =~ /.../);
    # Reference part and strain part
    my $ref = substr($line,0,2);
    $line = substr($line,2);
    # Quality list for each value (values are cased and include '-' and '=')
    my %evidence;
    my $outLine = $ref;
    my %somePair;
    # Process every pair in the input layout
    foreach my $k (0 .. $#$pInLay) {
	my $str = $pInLay->[$k];
	my $pair = substr($line,$k*2,2);
	# We may be off the end:
	if ($pair) {
	    if ($pair eq '-' || $pair eq '=') {
		$pair .= ' '; # end-of-line case
	    }
	    # Replace n with highly-uncertain arbitrary nucleotide,
	    # so we don't get a space there; it should be just as
	    # easy to correct it as impute it later.
	    if ($pair =~ /^[nN]/) {
		$pair = 'a!';
	    }
	    unless ($pair =~ /^[ n]/) {
		augmentEvidence($str,$pair,\%evidence,\%somePair);
	    }
	}
    }
    my $refVal = substr($ref,0,1);
    my $refVal2 = $refVal;
    $refVal2 =~ tr/ACGTNacgtn/acgtnACGTN/; # other casing
    my $hasNonGap = ($refVal ne '-' && $refVal ne '=');
    foreach my $k (0 .. $#$pOutLay) {
	my $str = $pOutLay->[$k];
	if (!defined $evidence{$str}) {
	    $outLine .= '  ';
	} else {
	    my $pair = $IS_ALL_SOLEXA ?
		decidePairSolexa($str,$ref,\%evidence,\%somePair) :
		decidePair($str,$ref,\%evidence,\%somePair);
	    $outLine .= $pair;
	    $hasNonGap ||= ($pair !~ /^[-=]/);
	}
	my $pH = $evidence{$str};
	delete $pH->{'-'}; # not of interest for "clashes";
	delete $pH->{'='}; # not of interest for "clashes";
	my @vals = keys %$pH;
	foreach my $v (@vals) {
	    # Need to sort the quality values before printing,
	    # and also replace by concatenated string.
	    $pH->{$v} = join("",sort {$b cmp $a} @{$pH->{$v}});
	}
	@vals = sort {$pH->{$b} cmp $pH->{$a}} @vals;
	# Print a "clash" if there are multiple values OR a disagreement
	# with the reference value. Currently we don't count a disagreement
	# if the refVal is a gap.
	if (@vals > 1 ||
	    (@vals && $refVal ne '-' && $refVal ne '=' && 
	     $vals[0] ne $refVal && $vals[0] ne $refVal2)) {
	    my $sp = '';
	    foreach my $v (@vals) {
		printf($hh "%s%s %s",$sp,$v,$pH->{$v});
		$sp = ' ';
	    }
	    print $hh " : $off:$offr $str $refVal\n";
	}
	# Add to clash count: except when ref is '-' or '=', or no vals, store under
	#  (1) strain
	#  (2) whether NQS or not
	#  (3) sequence of qual lists, separated by ' '
	#  (4) whether agrees with ref or not
	if ($pClashCount && @vals && $refVal ne '-' && $refVal ne '=') {
	    my $isNQS = ($vals[0] =~ /[ACGTN]/) ? 1 : 0;
	    my $pat;
	    foreach my $v (@vals) {
		$pat .= ' ' . $pH->{$v};
	    }
	    my $agree = ($vals[0] eq $refVal || $vals[0] eq $refVal2) ? 1 : 0;
	    $pClashCount->{$str}[$isNQS]{substr($pat,1)}[$agree] ++;
	}
    }
    if ($hasNonGap || $KEEP_IN_SYNC >= 2) { # $outLine =~ /[A-Z]/) {
	return $outLine;
    } else {
	return ">"; # to keep things in sync when all gaps
    }
}

sub augmentEvidence {
    my ($str,$pair,$pEvidence,$pSomePair) = @_;
    my ($nuc,$qs) = split(//,$pair);
    push @{$pEvidence->{$str}{$nuc}},$qs;
    if ($pSomePair->{$str} && $nuc ne '-' && $nuc ne '=') {
	$pSomePair->{$str} = 'XX';
    } else {
	$pSomePair->{$str} = $pair;
    }
}

sub decidePair {
    my ($str,$ref,$pEvidence,$pSomePair) = @_;
    my $pE = $pEvidence->{$str};
    # Score per nucleotide/gap.
    my %score;
    foreach my $nuc (keys %$pE) {
	my $failsNQS = ($nuc =~ /[_a-z]/);
	foreach my $qc (@{$pE->{$nuc}}) {
	    my $v = $nuc;
	    if ($failsNQS) {
		$v =~ tr/_a-z/-A-Z/;
		# '>' is for quality 29, which is upper bound
		# on effective non-NQS qualities.
		$qc = '>' if ($qc gt '>');
	    }
	    $score{$v} += ord($qc)-$ZERO;
	}
    }
    # Don't let the two kinds of gap disagree with each other:
    # turn the weaker into the stronger
    if ($score{'='} && $score{'-'}) {
	if ($score{'='} >= $score{'-'}) {
	    $score{'='} += $score{'-'};
	    delete $score{'-'};
	} else {
	    $score{'-'} += $score{'='};
	    delete $score{'='};
	}
    }
    my ($bWin,$bNext) = sort {$score{$b} <=> $score{$a} || $a eq substr($ref,0,1)} keys %score;
    my $rawScore = $score{$bWin};
    if (defined $bNext) {
	$rawScore -= $score{$bNext};
    }
    # Flatten (normalize) the raw score; see 0124 and sgrp/misc/quality_fitting
    my $flatScore = 55*(1-exp(-0.029*$rawScore));
    if (defined $bNext && (defined $score{'-'} || defined $score{'='})) {
	$flatScore *= 0.7;
    }
    # "max(0," is required because for unknown reasons we occasionally
    # get $ch lt $ZERO otherwise.
    my $ch = chr($ZERO + max(0,int(0.5+$flatScore)));
    # Can restore -NQS (lower case) if there was no +NQS evidence for this nucleotide,
    # but it's probably cleaner to reintroduce it at the end.
    #unless (defined $pE->{$bWin}) {
	#$bWin =~ tr/-A-Z/_a-z/;
    #}
    return $bWin . $ch;
}

sub decidePairSolexa {
    my ($str,$ref,$pEvidence,$pSomePair) = @_;
    my $pE = $pEvidence->{$str};
    # Sum of qualities per nucleotide/gap
    my %score;
    # Qualities per nucleotide/gap.
    my %qualList;
    foreach my $nuc (keys %$pE) {
	my $failsNQS = ($nuc =~ /[_a-z]/);
	foreach my $qc (@{$pE->{$nuc}}) {
	    my $v = $nuc;
	    if ($failsNQS) {
		$v =~ tr/_a-z/-A-Z/;
		# '>' is for quality 29, which is upper bound
		# on effective non-NQS qualities.
		$qc = '>' if ($qc gt '>');
	    }
	    my $q = rd($qc)-$ZERO;
	    push @{$qualList{$v}},$q;
	    $score{$v} += $q;
	}
    }
    # Don't let the two kinds of gap disagree with each other:
    # turn the weaker into the stronger
    if ($score{'='} && $score{'-'}) {
	if ($score{'='} >= $score{'-'}) {
	    $score{'='} += $score{'-'};
	    push @{$qualList{'='}},@{$qualList{'-'}};
	    delete $score{'-'};
	    delete $qualList{'-'};
	} else {
	    $score{'-'} += $score{'='};
	    push @{$qualList{'-'}},@{$qualList{'='}};
	    delete $score{'='};
	    delete $qualList{'='};
	}
    }
    my ($bWin,$bNext) = sort {$score{$b} <=> $score{$a} || $a eq substr($ref,0,1)} keys %score;
    my $pWinQuals = [sort {$b <=> $a} @{$qualList{$bWin}}];
    my $pNextQuals = defined $bNext ? [sort {$b <=> $a} @{$qualList{$bNext}}] : [];
    return $bWin . maqQuality($pWinQuals,$pNextQuals);
}

sub maqQuality {
    my ($pQualsA,$pQualsB) = @_;
    my (@epa,@epb);
    foreach my $q (@$pQualsA) {
	push @epa,exp(-log(10)*0.1*$q);
    }
    foreach my $q (@$pQualsB) {
	push @epb,exp(-log(10)*0.1*$q);
    }
    my $ep = maqConsensus(\@epa,\@epb);
    my $q = int(-log($ep)*10/log(10));
    return $q >= 94 ? ' ' : chr($q+33);
}

sub maqConsensus {
    my ($pEPA,$pEPB) = @_;
    my $postA = findAlphaNK($pEPA,$pEPB);
    my $postB = findAlphaNK($pEPB,$pEPA);
    return $postA/($postA+$postB);
}

sub findAlphaNK {
    my ($pEPA,$pEPB) = @_;
    if (@$pEPA) {
	my $eb = epsilonBar($pEPA);
	my $pABI = alphaBarI($pEPA,$pEPB,$eb);
	my $pBBI = betaBarI($pABI);
	my $cNK = cNK($pBBI,$eb);
	return alphaNK($cNK,$pEPA);
    } else {
	return 1;
    }
}

sub epsilonBar {
    my ($pEPA) = @_;
    my ($num,$den) = (0,0);
    my $fi = 1;
    foreach my $i (0 .. $#$pEPA) {
	$num += $fi*log($pEPA->[$i]);
	$den += $fi;
	$fi *= 0.8;
    }
    return exp($num/$den);
}
    
sub alphaBarI {
    my ($pEPA,$pEPB,$eb) = @_;
    my @ans;
    my $n = scalar(@$pEPA)+scalar(@$pEPB);
    my $term;
    foreach my $i (0 .. scalar(@$pEPA)) {
	if ($i == 0) {
	    $term = power(1-$eb,$n);
	} else {
	    $term *= (($n-$i+1)/$i) * ($eb/(1-$eb));
	}
	push @ans,$term;
    }
    return \@ans;
}

sub betaBarI {
    my ($pAlphaBarI) = @_;
    my ($num,$den) = (1,1);
    my @ans;
    foreach my $i (0 .. $#$pAlphaBarI) {
	$num -= $pAlphaBarI->[$i];
	$den -= $pAlphaBarI->[$i-1] if ($i > 0);
	push @ans,($num <= 0 ? 0 : $num/$den);
    }
    return \@ans;
}

sub cNK {
    my ($pBetaBarI,$eb) = @_;
    my @ans;
    my $prod = 1;
    my $fi = 1;
    foreach my $i (0 .. $#$pBetaBarI-1) {
	$fi *= 0.8;
	$prod *= power($pBetaBarI->[$i]/$eb,$fi);
    }
    return (1-power($pBetaBarI->[-1],$fi))*$prod;
}

sub power {
    my ($x,$n) = @_;
    if ($x <= 0) {
	return 0;
    } else {
	return exp($n*log($x));
    }
}

sub alphaNK {
    my ($cNK,$pEPA) = @_;
    my $prod = 1;
    my $fi = 1;
    foreach my $i (0 .. $#$pEPA) {
	$prod *= $pEPA->[$i]*$fi;
	$fi *= 0.8;
    }
    return $cNK*$prod;
}


1;
