package SGRP::CorrectNames;

# Create renamed.fastq and renamed.wm.

use Exporter;
@ISA = qw(Exporter);

@EXPORT = qw(correctNames);

use SGRP::Variables;
use SGRP::SsahaSNP;
use Utility;
use File::Basename;

use Carp;

use strict;

# Strict seems better (fewer false-positive "novel" regions).

my $STRICT_WATERMARKS = 1;

sub correctNames {
    my (@strains) = @_;
    unless (@strains) {
	confess("Usage: $0 crnm strain1 strain2 ...");
    }
    foreach my $str (@strains) {
	correctNames1($str);
    }
}

sub correctNames1 {
    my ($str) = @_;
    my $tgtDir = "$ROOT/read/$str";
    my $src = "$tgtDir/clip.fastq";
    unless (-s $src) {
	report("missing or empty: $src",1);
	return;
    }

    my $cor = "$tgtDir/name_corrections.txt";
    my $cwm = "$tgtDir/clip.wm";
    createWatermarkFile($src,$cwm);
    createReadNameCorrections($str,$cwm,$cor);

    my $tgt = "$tgtDir/renamed_withvec.fastq";
    if (stamp($tgt)) {
	report("creating $tgt");
	correctNames2($src,$cor,$tgt);
	report();
	unstamp($tgt);
    }
    
    my $tgt0 = "$tgtDir/renamed.fastq";
    if (stamp($tgt0)) {
	removeVectorSequences($tgt,$tgt0);
	unstamp($tgt0);
    }
    

    my $tgt1 = "$tgtDir/renamed.0.fastq";
    if (stamp($tgt1)) {
	report("splitting $tgt0");
	splitFastqFile($tgt0);
	report();
	unstamp($tgt1);
    }

    my $tgt2 = "$tgtDir/renamed.wm";
    createWatermarkFile($tgt0,$tgt2);

    return 1;
}	

sub createWatermarkFile {
    my ($src,$tgt) = @_;
    if (stamp($tgt)) {
	report("creating $tgt");
	if ($STRICT_WATERMARKS) {
	    createWatermarkFile2($src,$tgt);
	} else {
	    createWatermarkFile1($src,$tgt);
	}
	report();
	unstamp($tgt);
    }
}

# Copied over from old PruneAlignments.pl and was used for methp; this
# version is more restrictive, i.e. judges more regions as "poor",
# which seems to be right.

sub createWatermarkFile2 {
    my ($src,$tgt) = @_;

    my $th = 20; # min quality value to count as good
    my $qmin = chr(ord('!')+$th);

    # A "good" region is at least 40 bases long, and is such that 
    # every subregion of 20 bases contained in or overlapping with
    # the good region has at least 14 good positions.

    # Each line in $tgt is for one read. It contains the read name
    # and length, then 0 or more pairs of one-offsets each indicating
    # a good region.

    # In YGPM and perhaps other strains, the first 100 bases often
    # have high quality values but don't get alignments, so we
    # don't bother with "good" regions in [1,100].

    my $minsz = 40;
    my $himin = 100;
    my $srsz = 20;
    my $srmin = 14;

    my $read;
    my $fh = openToRead($src);
    my $gh = openToWrite($tgt);
    while (<$fh>) {
	chomp;
	if (/^\+(\S+) /) {
	    $read = $1;
	} elsif ($read) {
	    printWatermarks2($minsz,$himin,$srsz,$srmin,$qmin,$gh,$read,$_);
	    $read = undef;
	}
    }
}

sub printWatermarks2 {
    my ($minsz,$himin,$srsz,$srmin,$qmin,$gh,$read,$line) = @_;
    printf($gh "%s %4d",$read,length($line));
    my @a = split(//,$line);
    my @cum;
    my $nbad = 0;
    my @bad;
    foreach my $q (@a) {
	$nbad ++ if ($q lt $qmin);
	push @cum,$nbad;
	if ($#cum >= $srsz && $cum[$#cum] - $cum[$#cum-$srsz] > $srmin) {
	    foreach my $k ($#cum-$srsz+1 .. $#cum) {
		$bad[$k] = 1;
	    }
	}
    }
    # YGPM is weird, so we don't trust the first 100 bases ...
    my $lo= ($read =~ /^YGPM/) ? 100 : 0;
    while ($lo <= $#a-$minsz+1) {
	my $hi=$lo;
	$hi ++ while ($hi < $#a && !$bad[$hi]);
	if ($hi >= $lo+$minsz && $hi >= $himin) {
	    # output one-offset ...
	    printf($gh " %4d %4d",$lo+1,$hi+1);
	}
	$lo = $hi+$srmin+1;
    }
    print $gh "\n";
}

# Copied from watermarks.pl; theoretically looks better, but
# in practice it allows too much unaligned material as "good"
# and therefore potentially novel.

sub createWatermarkFile1 {
    my ($src,$tgt) = @_;

    my $th = 20; # min quality value to count as good
    my $qmin = chr(ord('!')+$th);

    # A "good" region is such that every subregion of 20 bases
    # has at least 14 good positions.

    my $srsz = 20;
    my $srmin = 14;

    my $read;
    my $fh = openToRead($src);
    my $gh = openToWrite($tgt);
    while (<$fh>) {
	chomp;
	if (/^\+(\S+) /) {
	    $read = $1;
	} elsif ($read) {
	    printWatermarks1($read,$_,$qmin,$srsz,$srmin,$gh);
	    $read = undef;
	}
    }
}

sub printWatermarks1 {
    my ($read,$line,$qmin,$srsz,$srmin,$gh) = @_;
    printf($gh "%s %4d",$read,length($line));
    my @a = split(//,$line);
    my @cum;
    my $nok = 0;
    foreach my $q (@a) {
	$nok ++ if ($q ge $qmin);
	push @cum,$nok;
    }
    my $lowWater = getLowWatermark(\@cum,1,$srsz,$srmin);
    if (defined $lowWater) {
	my $highWater = getHighWatermark(\@cum,-1);
	if (defined $highWater) {
	    printf($gh " %4d %4d",$lowWater+1,$highWater+1);
	}
    }
    print $gh "\n";
}

sub getLowWatermark {
    my ($pC,$d,$srsz,$srmin) = @_;
    my $nok = 0;
    my $init = 1;
    for (my $k = int($#$pC/2)-$srsz; $k>=0; $k--) {
	if ($pC->[$k+$srsz]-$pC->[$k] < $srmin) {
	    return $init ? undef : $k+1;
	}
	$init = 0;
    }
    return 0;
}

sub getHighWatermark {
    my ($pC,$d,$srsz,$srmin) = @_;
    my $nok = 0;
    my $init = 1;
    for (my $k = int($#$pC/2)+$srsz; $k<=$#$pC; $k++) {
	if ($pC->[$k]-$pC->[$k-$srsz] < $srmin) {
	    return $init ? undef : $k-1;
	}
	$init = 0;
    }
    return $#$pC;
}

sub createReadNameCorrections {
    my ($str,$src,$tgt) = @_;
    return unless (stamp($tgt));
    report("creating $tgt");
    my $fh = openToRead($src);
    my %longestGood;
    my %slashed;
    while (<$fh>) {
	chomp;
	my ($r,undef,@nums) = split;
	my $len = 0;
	for (my $k=0; $k<$#nums; $k+=2) {
	    $len = max($len,$nums[$k+1]-$nums[$k]+1);
	}
	$longestGood{$r} = $len;
	if ($r =~ /^(.*)\/.*\.([pq]1k)/) {
	    push @{$slashed{"$1.$2"}},$r;
	}
    }
    my %corr;
    # First set up the systematic plate errors; see th/0927
    my @names = keys %longestGood;
    if ($str eq 'BC187') {
	foreach my $name (grep /-31[a-z].*p1k/, @names) {
	    ($corr{$name} = $name) =~ s/-31/-32/;
	}
	foreach my $name (grep /-32[a-z].*p1k/, @names) {
	    ($corr{$name} = $name) =~ s/-32/-33/;
	}
    } elsif ($str eq 'YPS606') {
	foreach my $name (grep /-29[a-z].*p1k/, @names) {
	    ($corr{$name} = $name) =~ s/-29/-30/;
	}
	foreach my $name (grep /-30[a-z].*p1k/, @names) {
	    ($corr{$name} = $name) =~ s/-30/-29/;
	}
    } elsif ($str eq 'UWOPS03_461_4') {
	# Plates 14 to 20 have p reads for the plate below
	foreach my $name (@names) {
	    my ($pre,$n,$post) = ($name =~ /^(.*_4-)(\d+)([a-z].*p1k)$/);
	    if ($post && $n >= 14 && $n <= 20) {
		$corr{$name} = sprintf("%s%d%s",$pre,$n-1,$post);
	    }
	}
    } elsif ($str eq 'Q59_1') {
	# Plate 24 p reads match plate 25 q reads
	foreach my $name (grep /-24[a-z].*p1k/, @names) {
	    ($corr{$name} = $name) =~ s/-24/-25/;
	}
    }
    # See below for defns of area and transform codes...
    my %transform =
	(CBS5829 => {19 => [2,1], 28 => [2,1], 64 => [2,2]},
	 DBVPG6044 => {2 => [5,1]},
	 DBVPG6765 => {3 => [5,2], 66 => [2,1], 78 => [2,1]},
	 N_43 => {4 => [2,2]},
	 SK1 => {3 => [5,2], 
		 30 => [2,1], 50 => [2,1], 77 => [2,1],
		 19 => [2,2], 39 => [2,2]},
	 W303 => {61 => [4,2]},
	 Y55 => {15 => [2,1], 60 => [2,1], 71 => [2,1]},
	 YPS138 => {1 => [2,2]},
	 UWOPS91_917_1 => {29 => [3,3], 50 => [2,1], 51 => [2,1]},
	 '273614X' => {38 => [2,1]});
    transformPlates($transform{$str},\@names,\%corr);
    foreach my $r (keys %slashed) {
	my @sorted = sort {$longestGood{$b} <=> $longestGood{$a}} ($r,@{$slashed{$r}});
	my $winner = shift @sorted;
	if ($winner ne $r) {
	    $corr{$winner} = $r;
	    $corr{$r} = "";
	}
	foreach my $rr (@sorted) {
	    $corr{$rr} = ""; # don't want these ones
	}
    }
    # If A is corrected to B, make sure B is corrected to null if nothing else
    my @vals = values %corr;
    foreach my $v (@vals) {
	if (!defined $corr{$v} && defined $longestGood{$v}) {
	    $corr{$v} = '';
	}
    }
    my $gh = openToWrite($tgt);
    foreach my $r (sort keys %corr) {
	my $c = $corr{$r};
	if ($c) {
	    print $gh "$r $c\n";
	} else {
	    print $gh "$r\n";
	}
    }
    $gh->close();
    unstamp($tgt);
    report();
}

# Pairs are [area,trans], where:
#  area 1 = 1-16 for all cols, and whole of cols g to p
#  area 2 = 13-26 for all cols, and 9-26 for a to f
#  area 3 = whole plate
#  area 4 = all of rows 1 and 2, and 3 and 4 from c upwards
#  area 5 = 1-16 for all cols, and 1-20 for g to p
#
#  trans 1 = rows 1,2,5,6 etc: row += 2
#            rows 3,4,7,8 etc: row -= 2, col -= 2
#  trans 2 = rows 1,2,5,6 etc: row += 2, col += 2
#            rows 3,4,7,8 etc: row -= 2
#  trans 3 = rotate: row -> 25-row, col -> 17-col

sub transformPlates {
    my ($pH,$pNames,$pCorr) = @_;
    return unless ($pH);
    while (my ($plate,$pA) = each %$pH) {
	my ($area,$trans) = @$pA;
	my $pat = sprintf("-%s[a-z].*p1k",$plate);
	foreach my $name (grep /$pat/, @$pNames) {
	    my ($a,$n) = ($name =~ /([a-z])(\d\d)/);
	    my $inArea;
	    if ($area == 1) {
		$inArea = ($n <= 16 || ($n <= 26 && $a ge 'g'));
	    } elsif ($area == 2) {
		$inArea = ($n >= 13 || ($n >= 9 && $a le 'f'));
	    } elsif ($area == 3) {
		$inArea = 1;
	    } elsif ($area == 4) {
		$inArea = $a ge 'c' ? ($n <= 4) : ($n <= 2);
	    } elsif ($area == 5) {
		$inArea = ($n <= 16 || ($n <= 20 && $a ge 'g'));
	    } else {
		confess("Bad area number: $area for $name");
	    }
	    if ($inArea) {
		my ($aa,$nn) = ($a,$n);
		if ($trans == 1) {
		    if (($n-1) % 4 < 2) {
			$nn += 2;
		    } else {
			$nn -= 2;
			$aa = chr(ord($aa)-2);
		    }
		} elsif ($trans == 2) {
		    if (($n-1) % 4 < 2) {
			$nn += 2;
			$aa = chr(ord($aa)+2);
		    } else {
			$nn -= 2;
		    }
		} elsif ($trans == 3) {
		    $nn = 25-$nn;
		    # 96+(17-(a-96)) = 209-a
		    $aa = chr(209-ord($aa));
		} else {
		    confess("Bad trans number: $trans for $name");
		}
		# Keep valid letters, so as not to waste reads:
		if ($aa eq '`') {
		    $aa = 't';
		} elsif ($aa eq '_') {
		    $aa = 's';
		}
		# Keep nn = 25 or 26 from addition; "+28" means -1 -> 27 and 0 -> 28
		if ($nn < 1) {
		    $nn += 28;
		};
		# Condition should always be satisfied ...
		if ($aa =~ /[a-t]/ && $nn >= 1 && $nn <= 28) {
		    $nn = "0$nn" if ($nn !~ /../); # restore leading zero if necessary
		    ($pCorr->{$name} = $name) =~ s/$a$n/$aa$nn/;
		} else {
		    $pCorr->{$name} = "";
		}
	    }
	}
    }
}

sub correctNames2 {
    my ($src,$cor,$tgt) = @_;
    my $pCor = readCorrections($cor);
    my $fh = openToRead($src);
    my $gh = openToWrite($tgt);
    while (<$fh>) {
	if (/^([\@\+])(\S+)(.*)/) {
	    my $pfx = $1;
	    my $old = $2;
	    my $rest = $3;
	    if ($old =~ /23a01/) {
		1;
	    }
	    if ($pCor && defined $pCor->{$old}) {
		my $new = $pCor->{$old};
		if ($new) {
		    print $gh "$pfx$new$rest\n";
		    my $line = <$fh>;
		    print $gh $line;
		} else {
		    <$fh>; # gobble up data line
		}
	    } else {
		print $gh $_;
		my $line = <$fh>;
		print $gh $line;
	    }
	} else {
	    chomp;
	    confess("Bad line in $src: $_");
	}
    }
    $gh->close();
}

sub readCorrections {
    my ($f) = @_;
    return unless (-s $f);
    my $fh = openToRead($f);
    my %cor;
    while (<$fh>) {
	chomp;
	my ($old,$new) = split;
	$cor{$old} = $new || 0;
    }
    return \%cor;
}

#sub correctWatermarkFile {
#    my ($src,$cor,$tgt) = @_;
#    my $pCor = readCorrections($cor);
#    my $fh = openToRead($src);
#    my $gh = openToWrite($tgt);
#    while (<$fh>) {
#	chomp;
#	my ($r,@nums) = split;
#	if (defined $pCor->{$r}) {
#	    print $gh "$pCor->{$r} @nums\n" if ($pCor->{$r});
#	} else {
#	    print $gh "$_\n";
#	}
#    }
#    $gh->close();
#}

# src: renamed_withvec.fastq
# tgt: renamed.fast

sub removeVectorSequences {
    my ($src,$tgt) = @_;
    my $ssFile = runSsahaSNPForVectors($src);
    my $pH = findVectorRegions($ssFile);
    #unlink($ssFile);
    report("creating $tgt");
    pruneFastqForVectors($src,$pH,$tgt);
    report();
}

sub runSsahaSNPForVectors {
    my ($src) = @_;
    my ($bin,$fixFlag) = findSsahaSNPBinary();
    my $vecFA = "$ROOT/ref/vect/vect.fa";
    my $dir = dirname($src);
    my $out = "$dir/ssVect.alilines.txt";
    my $err = "$dir/ssVect.err";
    unless (-s $out) {
	my $cmd = 
	    "$bin $src $vecFA -diff 2000 -NQS 0 " .
	    "-quality 0 -tags 1 -align 1 -score 20 -seeds 2 -cut 50000 $fixFlag";
	report("creating $out");
	`$cmd 2>$err | grep '^ALIGNMENT' > $out; sleep 10\n`;
	report();
	if (-s $err) {
	    my $eh = openToRead($err);
	    while (<$eh>) {
		unless (/^Warning/) {
		    die("Error in file $err: $_");
		}
	    }
	}
    }
    return $out;
}

sub findVectorRegions {
    my ($f) = @_;
    my $fh = openToRead($f);
    my %isVec;
    while (<$fh>) {
	if (/^ALIGNMENT/) {
	    chomp;
	    my @a = split;
	    my $rlen = $a[-1];
	    my (undef,undef,$read,undef,$v1,$v2) = @a;
	    my ($lo,$hi) = sort {$a <=> $b} ($v1,$v2);
	    #my $vlen = $hi-$lo+1;
	    # We don't use a vector match if it's further from
	    # both ends than its own length, because it's unlikely
	    # to be real...but we're being cautious, so we will
	    # use such. In fact there's huge variation in
	    # the number of vector alignments by strain, so they
	    # probably are mostly genuine. (0 for YPS128, 1908
	    # for S288c, 9873 for YS2).
	    #unless ($lo-1 > $vlen && $hi+$vlen < $rlen) {
	    push @{$isVec{$read}},[$lo,$hi];
	    #}
	}
    }
    return \%isVec;
}

sub pruneFastqForVectors {
    my ($src,$pH,$tgt) = @_;
    my $fh = openToRead($src);
    my $gh = openToWrite($tgt);
    while (1) {
	my @quad;
	foreach my $n (1 .. 4) {
	    my $line = <$fh>;
	    return unless ($line);
	    push @quad,$line;
	}
	my ($read) = split(" ",substr($quad[0],1));
	if (defined $pH->{$read}) {
	    pruneFastqForVectors1($pH->{$read},$gh,@quad);
	} else {
	    print $gh @quad;
	}
    }
}

sub pruneFastqForVectors1 {
    my ($pRList,$gh,@quad) = @_;
    my ($read,undef,$start,undef,$end) = split(" ",substr($quad[0],1));
    my ($len,$lo,$hi) = regionToKeep($pRList,$end-$start+1);
    if ($len) {
	foreach my $i (1,3) {
	    chomp $quad[$i];
	}
	printf($gh "@%s bases %d to %d\n",$read,1,$len);
	printf($gh "%s\n",substr($quad[1],$lo-1,$len));
	printf($gh "+%s bases %d to %d\n",$read,1,$len);
	printf($gh "%s\n",substr($quad[3],$lo-1,$len));
    }
}

# We keep the longest subregion of 1..$end that isn't covered by
# anything in $pRList, as long as it's length 30 or more.

sub regionToKeep {
    my ($pRList,$end) = @_;
    my @cands;
    my $pos = 1;
    my @sorted = sort {$a->[0] <=> $b->[0]} @$pRList;
    while (@sorted) {
	my $pDel = shift @sorted;
	my ($lo,$hi) = @$pDel;
	# We only want to keep regions of 30 or more.
	my $len = $lo-$pos;
	if ($len >= 30) {
	    push @cands,[$len,$pos,$lo-1];
	}
	$pos = max($pos,$hi+1);
    }
    my $len = $end-$pos+1;
    if ($len >= 30) {
	push @cands,[$len,$pos,$end];
    }
    if (@cands) {
	@cands = sort {$b->[0] <=> $a->[0]} @cands;
	return @{$cands[0]};
    } else {
	return undef;
    }
}

sub splitFastqFile {
    my ($f) = @_;
    my $fh = openToRead($f);
    my @lines = <$fh>;
    my $n = (scalar @lines)/4;
    my $m = 4*int(($n+9)/10);
    my $b = $f;
    $b =~ s/\.fastq$//;
    # Cycle round the parts so we distribute artificial
    # reads evenly
    my @out;
    foreach my $p (0 .. 9) {
	push @out,openToWrite("$b.$p.fastq");
    }
    my $p=0;
    foreach my $k (0 .. $n-1) {
	my $i = $k*4;
	my $gh = $out[$p];
	foreach my $j (0 .. 3) {
	    print $gh $lines[$i+$j];
	}
	$p = ($p < 9) ? $p+1 : 0;
    }
}

1;

