package SGRP::CreateReadLayout;

use Exporter;
@ISA = qw(Exporter);

@EXPORT = qw(createReadLayout);

use SGRP::MatchStrings;
use SGRP::Variables;
use SGRP::AlignmentCoords;
use SGRP::AlignmentClashRates;
use SGRP::SortLayoutFiles; # for realignBuffer (and setDNABitsAndSymbols)

use Utility;
use File::Basename;

use Carp;

use strict;

#my $NHN = 0;

# NOTE on gap qualities: look for "GAP QUALITY" comments
# for places to alter.

# Fields of match string:
#   0 start
#   1 IMD
#   2 len
#   3 str
#   4 nucs
#   5 quals
#   6 aliId

my $DEBUG = 0;
my $DEBUG_MATCHES = 0;

my $DO_REALIGNMENT = 1; # usually we do

sub createReadLayout {
    my ($species,$filtMeth,$strainSet,@parts) = @_;
    unless (@parts) {
	confess("Usage: $0 crlo species filtMeth strainSet part1 part2 ...");
    }
    setDNABitsAndSymbols(); # for realignment
    my ($iter,$dir) = getCurrentIteration($species,$filtMeth,$strainSet,'p?');
    my $prev = previousDirectory($dir);
    my $isExtra = $IS_FINAL_ITERATION eq 'EXTRA';
    # When $isExtra, this will only work after we create layout_z.gz
    # post-Solexa with createLayoutZ.pl ...
    my $layoutBase = $isExtra ? 'layout_z.gz' : 'layout.gz';
    # Dir containing genome .fa files (broken up into "parts"); we only
    # use it on iteration zero.
    my $isInit = !defined $prev;
    my $gDir = "$ROOT/ref/$species" if ($isInit);
    foreach my $part (@parts) {
	$part = "p$part" if ($part =~ /^[0-9]$/);
	foreach my $chrDir (glob("$dir/$part/*")) {
	    my $chr = basename($chrDir);
	    # This takes too long on rDNA on final iter; but actually OK
	    # if we use the "long" queue.
	    #$DO_REALIGNMENT = 0 if ($chr eq 'chr12' && $IS_FINAL_ITERATION);
	    # And doing it on "extra" messes up the refGapPair being a space,
	    # and hence mapping coords...
	    $DO_REALIGNMENT = 0 if ($IS_FINAL_ITERATION eq 'EXTRA');
	    my ($inFile,$pRefRegs);
	    if ($isInit) {
		# go for "_extended" version if it exists ...
		my $inFileER = "$gDir/$part/${chr}_extended_recased.fa";
		if (-s $inFileER) {
		    $inFile = $inFileER;
		} else {
		    # can be OK ...
		    $inFile = "$gDir/$part/$chr.fa";
		}
	    } else {
		# Look for successive previous dirs ...
		my $prev1 = $prev;
		while ($prev1) {
		    my $lay = "$prev1/$part/$chr/$layoutBase";
		    if (-s $lay) {
			$inFile = $lay;
			last;
		    }
		    $prev1 = previousDirectory($prev1);
		}
	    }
	    createReadLayout1($chr,$chrDir,$isInit,$inFile,$iter);
	    if ($isExtra) {
		# Both of these are to enable GFF file building:
		alignmentCoords1("$chrDir/layout.gz","$chrDir/alicoords.gz");
		alignmentClashRates("$chrDir/layout.gz","$chrDir/acr.txt");
	    }
	}
    }
}

sub createReadLayout1 {
    my ($chr,$dir,$isInit,$inFile,$iter) = @_;

    my $aFile = "$dir/matches.gz";
    my $qFile = "$dir/qualities.gz";

    return unless (isComplete($aFile,1));
    return unless (isComplete($qFile,1));

    my $outFile1 = "$dir/layout1.gz";
    my $outFile = "$dir/layout.gz";

    return unless (stamp($outFile));

    my $msFile = "$dir/mstrings.gz";
    unless (-s $msFile) {
	report("creating $msFile");
	alignmentsToMatchStrings($chr,$aFile,$qFile,$msFile);
	report();
    }

    unless (-s $outFile1) {
	report("creating $outFile1");
	print "\n" if ($DEBUG >= 1);
	my $mh = openToRead($msFile);
	my $fh = openToRead($inFile); # chr file layout file
	my $gh = openToWrite($outFile1);
	outputAlignments($isInit,$iter,$fh,$mh,$gh);
	$gh->close();
	report();
    }
    
    my $acFile = "$dir/alicoords1.gz";
    alignmentCoords1($outFile1,$acFile);

    removeIsolatedRegions($dir,$acFile);

    unlink($outFile1);

    unstamp($outFile);
}

# $isInit: whether first round or not
# $iter: iteration number, for printing in ">" lines
# $fh < chr.fa (if isInit) else old layout.gz (or layout_y.gz)
# $mh < mstrings.gz
# $gh > layout.gz

sub outputAlignments {
    my ($isInit,$iter,$fh,$mh,$gh) = @_;

    # Maintain $chrOff so that $outBuf[$k] corresponds to
    # position $k+$chrOff in chromosome. (Chromosome posns from 1).
    my $chrOff = 1;
    # $outBuf[$k] is an ARRAY REF. Last element is the line for position $k+$chrOff in the input alignment;
    # preceding ones are ">str posn" layout directives.
    my @outBuf;

    # Alignment-id for each position; actually we use the NAME of the alignment here,
    # to allow looking at the read name.
    my @aliId;

    # initialize saved match string
    my $pMSSaved = readMatchString($mh);

    while ($pMSSaved) {

	# Read match-strings: MD and I sets separately.
	my ($pMSListI,$pMSListMD);
	($pMSListI,$pMSListMD,$pMSSaved) = readMatchStrings($pMSSaved,$mh);

	# Extend outBuf up to just after the end of longest current match
	my $start = msListStart($pMSListI,$pMSListMD);
	my $lim = msListLimit($start,$pMSListMD,$chrOff);
	extendOutputBuffer($isInit,$fh,$lim,\@outBuf);

	# Print outBuf up to just short of $pMS; return new $chrOff.
	my $llp;
	($chrOff,$llp) = printUpToAlignment(\@outBuf,$chrOff,$isInit,$start,$gh);

	# Incorporate matches, returning new start-of-alignment lines WITH NEWLINES
	my @soaLines;
	# We sort by alignment id number so that earlier alignments (already running)
	# get a chance to re-reserve their columns before new ones come and pinch
	# them.
	foreach my $pMS (sort {$a->[-1] <=> $b->[-1]} @$pMSListMD) {
	    push @soaLines,incorporateMatch($iter,\@outBuf,\@aliId,$pMS,$chrOff);
	}
	# Print new insertions
	if (@$pMSListI) {
	    # @outBuf could be empty at the end of the chromosome...
	    printNewInsertions($iter,$llp,@outBuf && $outBuf[0][-1],\@aliId,$pMSListI,\@soaLines,$gh);
	}
	# Print the start-of-alignment lines just created
	foreach my $pA (@soaLines) {
	    my ($name,$k,$iter,undef,$done) = @$pA;
	    unless ($done) {
		printf($gh ">%s %d %s\n",$name,$k,$iter);
	    }
	}
    }

    # Print out tail end of chromosome
    printTailEndOfChromosome($isInit,$chrOff,$fh,\@outBuf,$gh);

}

sub readMatchString {
    my ($mh) = @_;
    my $line = <$mh> || return undef;
    chomp $line;
    return [split(" ",$line)];
}

sub readMatchStrings {
    my ($pSaved,$mh) = @_;
    my (@ansI,@ansMD);
    if ($pSaved->[1] eq 'I') {
	push @ansI,$pSaved;
    } else {
	push @ansMD,$pSaved;
    }
    my $pos = $pSaved->[0];
    while (1) {
	my $pMS = readMatchString($mh);
	if ($pMS && $pMS->[0] == $pos) {
	    if ($pMS->[1] eq 'I') {
		push @ansI,$pMS;
	    } else {
		push @ansMD,$pMS;
	    }
	} else {
	    return (\@ansI,\@ansMD,$pMS);
	}
    }
}    

sub msListStart {
    my ($pMSListI,$pMSListMD) = @_;
    if (@$pMSListI) {
	return $pMSListI->[0][0];
    } else {
	return $pMSListMD->[0][0];
    }
}

sub msListLimit {
    my ($start,$pMSListMD,$chrOff) = @_;
    my $lim = $start - $chrOff;
    $lim += maxMSLength($pMSListMD) if (@$pMSListMD);
    return $lim;
}

# Position 3 of match string is reqLen
sub maxMSLength {
    my ($pMSList) = @_;
    my $ans = $pMSList->[0][3];
    for (my $i=1; $i<=$#$pMSList; $i++) {
	$ans = max($ans,$pMSList->[$i][3]);
    }
    #print "MML = $ans\n";
    return $ans;
}

sub extendOutputBuffer {
    my ($isInit,$fh,$newTop,$pOutBuf) = @_;
    return if ($newTop < -1); # already long enough
    if ($isInit) {
	while (<$fh>) {
	    unless (/^>/) {
		chomp;
		foreach my $v (split(//,$_)) {
		    my $q = $v ge 'a' ? '!' : ' ';
		    push @$pOutBuf,["$v$q"];
		}
		return if ($#$pOutBuf >= $newTop);
	    }
	}
    } else {
	my $nToGet = $newTop-$#$pOutBuf;
	while ($nToGet > 0) {
	    # @buf is next sequence line and any preceding ">" lines
	    my @buf;
	    while (1) {
		chomp ($_ = <$fh>);
		push @buf,$_;
		unless (/^>/) {
		    push @$pOutBuf,\@buf;
		    last;
		}
	    }
	    $nToGet --;
	}
    }
}
		
# Print the stuff that's now complete, indicating gaps according
# to @$pLayout.

# $lim is limit up to (and not including) which it is now safe to print:

sub printUpToAlignment {

    my ($pOutBuf,$chrOff,$isInit,$lim,$gh) = @_;

    my $llp;
    my $isExtra = $IS_FINAL_ITERATION eq 'EXTRA';

    while ($chrOff < $lim) {

	# Get the line from the output buffer.
	my $pA = @$pOutBuf ? shift @$pOutBuf : die("empty pOutBuf");
	if ($isInit && $chrOff % 100 == 0) {
	    print $gh ">REF $chrOff\n";
	}
	foreach my $line (@$pA) {
	    print $gh "$line\n" if ($isExtra || hasNucleotides($line));
	}
	$llp = $pA->[-1];
	$chrOff ++;
	print "13 CHROFF $chrOff\n" if ($DEBUG >= 2);
    }
    return ($chrOff,$llp);
}

sub hasNucleotides {
    my ($line) = @_;
    #$NHN ++;
    return 1 if (substr($line,0,1) eq '>'); # always keep annotation lines!
    my $iMax = length($line)-2;
    for (my $i=0; $i<=$iMax; $i+=2) {
	return 1 if (substr($line,$i,1) =~ /[ACGTNacgtn]/);
    }
    #print "No nucleotides at NHN=$NHN, line $line\n";
    return 0;
}

# $chrOff is just for debugging ...
sub incorporateMatch {
    my ($iter,$pOutBuf,$pAliId,$pMS,$chrOff) = @_;
    my (undef,$type,$len,$reqLen,$name,$nucs,$quals,$id,$special) = @$pMS;
    my @soaLines;
    #printf("Incorporating %s\n",substr(join(" ",@$pMS),0,100));
    #foreach my $k (0 .. $#$pAliId) {
	#if ($pAliId->[$k]) {
	#    printf("%d:%d ",$k,$pAliId->[$k]);
	#}
    #}
    my ($k,$isNew) = firstFreePosition($pOutBuf,$pAliId,"$name/$id",$reqLen);
    #print "FFP $id -> $k $isNew\n";
    my $col = ($k+1)*2;
    my $reqd = $col+2;
    if ($isNew) {
	# Record SOA line with its id and 0 meaning "not printed yet"
	push @soaLines,[$name,$k,$special || $iter,$id,0];
    }
    my $isDel = $type eq 'd';
    my ($nAgr,$nDis,$vSeq,$rSeq);
    for (my $i=0; $i<$len; $i++) {
	my $rv = substr($pOutBuf->[$i][-1],0,1);
	my $pad = $reqd - length($pOutBuf->[$i][-1]);
	if ($pad > 0) {
	    if (defined $pOutBuf->[$i] && $#{$pOutBuf->[$i]} >= 0) {
		$pOutBuf->[$i][-1] .= ' ' x $pad;
	    } else {
		1;
	    }
	}
	my $ii = $isDel ? 0 : $i; # del only has one instance of each "nucleotide" and quality
	my $v = substr($nucs,$ii,1);
	my $q = substr($quals,$ii,1);
	if ($isDel && $len > 20) {
	    $v = '='; # insist on hard gap if long deletion
	}
	substr($pOutBuf->[$i][-1],$col,2) = "$v$q";
	if ($DEBUG_MATCHES) {
	    if (!$isDel && $q ge '+') {
		$v =~ tr/acgt/ACGT/;
		$rv =~ tr/acgt/ACGT/;
		if ($v =~ /[ACGT]/ && $rv =~ /[ACGT]/) {
		    if ($v eq $rv) {
			$nAgr++;
		    } else {
			$nDis++;
		    }
		}
	    } else {
		$v =~ tr/ACGT/acgt/;
		$rv =~ tr/ACGT/acgt/;
	    }
	    $vSeq .= $v;
	    $rSeq .= $rv;
	}
    }
    if ($DEBUG_MATCHES) {
	printf("%d:%d in %s\n",$nAgr,$nDis,join(" ",@{$pMS}[0..3,6]));
	my $kMax = length($vSeq)-1;
	for (my $k=0; $k<=$kMax; $k+=60) {
	    printf("   %s\n   %s\n",substr($vSeq,$k,60),substr($rSeq,$k,60));
	}
	if ($nDis >= $nAgr && $nDis >= 10) {
	    die();
	}
    }
    return @soaLines;
}

# See also printNewInsertions ...
sub firstFreePosition {
    my ($pOutBuf,$pAliId,$id,$len) = @_;
    my $got=0;
    my $idR = substr($id,0,index($id,':'));
    foreach my $k (0 .. $#$pAliId) {
	my $idk = $pAliId->[$k];
	if ($idk eq $id) {
	    # then already saved for this alignment
	    return ($k,0);
	}
    }
    foreach my $k (0 .. $#$pAliId) {
	my $idk = $pAliId->[$k];
	if (substr($idk,0,index($idk,':')) eq $idR) {
	    # taking over from earlier segment of same read
	    # BUT we have to be sure columns are free!
	    my $col = ($k+1)*2;
	    if (columnIsFreeThroughout($col,$pOutBuf,0,$len-1)) {
		$pAliId->[$k] = $id;
		return ($k,1);
	    }
	}
    }
    # Try for a new place to put it ...
    my $k = 0;
  LOOP:
    while (1) {
	my $col = ($k+1)*2;
	# Check first and (for efficiency) last
	if (columnIsFreeThroughout($col,$pOutBuf,0,$len-1)) {
	    $pAliId->[$k] = $id;
	    return ($k,1);
	}
	$k ++;
    }
    $pAliId->[$k] = $id;
    return ($k+1,1);
}

# Check all positions, starting with last for efficiency
sub columnIsFreeThroughout {
    my ($col,$pOutBuf,$lo,$hi) = @_;
    foreach my $i ($hi,$lo..$hi-1) {
	return 0 unless (columnIsFree($col,$pOutBuf,$i));
    }
    return 1;
}

sub columnIsFree {
    my ($col,$pOutBuf,$off) = @_;
    return 1 if ($#$pOutBuf < $off);
    return 1 if (length($pOutBuf->[$off][-1]) <= $col);
    return (substr($pOutBuf->[$off][-1],$col,2) eq '  ');
}

sub printNewInsertions {
    my ($iter,$llp,$lnp,$pAliId,$pMSList,$pSoaLines,$gh) = @_;
    # Get longest first:
    my @msList = reverse(@$pMSList);
    my $maxLen = $msList[0][2];
    my $maxSeq = $msList[0][5];
    my $refGapPair = pairToInsert($maxLen,$llp,$lnp,0);
    if ($IS_FINAL_ITERATION eq 'EXTRA') {
	# We have space for ref to show it's an "extra" insertion only.
	$refGapPair = '  ';
    } elsif ($refGapPair =~ /^ /) {
	# Start of chromosome (or anywhere): must have '= ' not '  ' for ref
	# when not "extra".
	$refGapPair = '= ';
    }
    my $lenL = length($llp);
    my $lenN = length($lnp);
    # tried: !defined $llp ? $lenN : !defined $lnp ? $lenL : min($lenL,$lenN);
    my $minLineLen = max($lenL,$lenN);
    my $k=0; # position number
    my @msMap;
  LOOP:
    foreach my $pMS (@$pMSList) {
	my $id = sprintf("%s/%s",$pMS->[4],$pMS->[7]); # WAS: 7, until 3-9-07
	# See also firstFreePosition here ...
	foreach my $k (0 .. $#$pAliId) {
	    if ($pAliId->[$k] eq $id) {
		$msMap[$k] = $pMS;
		# Print relevant SOA line if present, so it comes BEFORE the insertion
		foreach my $pA (@$pSoaLines) {
		    if ($pA->[3] == $pMS->[7]) {
			printf($gh ">%s %d %s\n",@{$pA}[0..2]);
			$pA->[4] = 1; # mark as done
			last;
		    }
		}
		next LOOP;
	    }
	}
	# We don't have an existing aliId ...
	my $idR = substr($id,0,index($id,':'));
	while (1) {
	    my $idk = $pAliId->[$k];
	    last if (substr($idk,0,index($idk,':')) eq $idR);
	    my $col = ($k+1)*2;
	    if ($col < $lenL && substr($llp,$col,2) ne '  ' && 
		$col < $lenN && substr($lnp,$col,2) ne '  ') {
		$k++;
	    } else {
		last;
	    }
	}
	printf($gh ">%s %d %s\n",$pMS->[4],$k,$pMS->[8] || $iter);
	$msMap[$k] = $pMS;
	$pAliId->[$k] = $id;
	$k ++;
    }
    my $jMax = max($#msMap,$minLineLen/2-2);
    my @buf;
    my $doRealign = $DO_REALIGNMENT;
    foreach my $i (0 .. $maxLen-1) { # i is row ...
	my $ghLine = $refGapPair;
	#for (my $j=0; $j<=$#msMap; $j++) {
	for (my $j=0; $j<=$jMax; $j++) {
	    my $pMS = $msMap[$j];
	    my $pair;
	    if (defined $pMS && $pMS->[2] > $i) {
		# sequence is there, and long enough:
		$pair = substr($pMS->[5],$i,1) . substr($pMS->[6],$i,1);
	    } else {
		$pair = pairToInsert($maxLen,$llp,$lnp,($j+1)*2);
	    }
	    $ghLine .= $pair;
	}
	if (hasNucleotides($ghLine)) {
	    if ($doRealign) {
		push @buf,$ghLine;
	    } else {
		print $gh "$ghLine\n";
	    }
	}
    }
    realignBuffer(\@buf,$gh) if ($doRealign); # in SortLayoutFiles.pm
}

sub pairToInsert {
    my ($maxLen,$llp,$lnp,$col) = @_;
    my $prevPair = (length($llp) <= $col) ? undef : substr($llp,$col,2);
    my $nextPair = (length($lnp) <= $col) ? undef : substr($lnp,$col,2);
    return '  ' unless (defined $prevPair && $prevPair ne '  ' && defined $nextPair && $nextPair ne '  ');
    # If hard gap on both sides, return hard gap
    if ($maxLen >= 10 ||
	(substr($prevPair,0,1) eq '=' && substr($nextPair,0,1) eq '=')) {
	return '=' . minQual($prevPair,$nextPair);
    } elsif (substr($prevPair,0,1) eq ' ' && substr($nextPair,0,1) eq ' ') {
	return '  ';
    } else {
	return '-' . minQual($prevPair,$nextPair);
    }
}

sub minQual {
    my ($pair1,$pair2) = @_;
    my $qc1 = substr($pair1,1,1) || ' ';
    my $qc2 = substr($pair2,1,1) || ' ';
    return ($qc1 eq ' ') ? $qc2 : ($qc2 eq ' ' || $qc1 le $qc2) ? $qc1 : $qc2;
}

sub printTailEndOfChromosome {
    my ($isInit,$chrOff,$fh,$pOutBuf,$gh) = @_;
    foreach my $pA (@$pOutBuf) {
	foreach my $line (@$pA) {
	    print $gh "$line\n" if (hasNucleotides($line));
	}
    }
    if ($isInit) {
	while (<$fh>) {
	    chomp;
	    foreach my $v (split(//,$_)) {
		if ($chrOff % 100 == 0) {
		    printf($gh ">REF %d\n",$chrOff);
		}
		my $q = $v ge 'a' ? '!' : ' ';
		print $gh "$v$q\n";
		$chrOff ++;
	    }
	}
    } else {
	while (<$fh>) {
	    print $gh $_ if (hasNucleotides($_));
	}
    }
}

sub removeIsolatedRegions {
    my ($d,$acFile) = @_;
    report("$d/layout.gz: creating");
    my $pRem = regionsToRemove($acFile);
    # Write first before it all gets discarded by removeIsolated1 ...
    my $gh = openToWrite("$d/removed.txt");
    foreach my $pA (@$pRem) {
	printf($gh "%7d %7d %s\n",@$pA);
    }
    $gh->close();
    removeIsolated1("$d/layout1.gz",$pRem,"$d/layout.gz");
    report();
}

# Remove regions of length 20 or less, and the gaps leading to them.

sub regionsToRemove {
    my ($f) = @_; # alicoords1.gz
    my @rem; # each elt is [lo,hi,read]
    my $fh = openToRead($f);
    while (<$fh>) {
	chomp;
	my ($read,undef,undef,undef,@quads) = split;
	next if ($read eq 'REF');
	# Quads to spans: take 0th and 2nd of each quad,
	my @spans;
	for (my $i=0; $i<$#quads; $i+=4) {
	    push @spans,$quads[$i],$quads[$i+2];
	}
	# Forward pass
	my $spanned = 0;
	my $i = 0;
	while ($spanned < 20 && $i < $#spans) {
	    $spanned += $spans[$i+1]-$spans[$i]+1;
	    $i += 2;
	}
	$i -= 2;
	if ($i > 0) {
	    #push @{$rem{$read}},[$spans[0],$spans[$i]-1,'L'];
	    push @rem,[$spans[0],$spans[$i]-1,$read];
	}
	# Backward pass
	my $spanned = 0;
	my $i = $#spans-1;
	while ($spanned < 20 && $i >= 0) {
	    $spanned += $spans[$i+1]-$spans[$i]+1;
	    $i -= 2;
	}
	$i += 2;
	if ($i < $#spans-1) {
	    #push @{$rem{$read}},[$spans[$i+1]+1,$spans[-1],'R'];
	    push @rem,[$spans[$i+1]+1,$spans[-1],$read];
	}
    }
    @rem = sort {$a->[0] <=> $b->[0] || $a->[1] <=> $b->[1]} @rem;
    return \@rem; # \%rem;
}
	
sub removeIsolated1 {
    my ($f,$pRem,$g) = @_;
    my $fh = openToRead($f);
    my $gh = openToWrite($g);
    my $ln = 0;
    my %posnOf; # pos of each read
    my %annoAt; # annotation of each position (until printed)
    my @readAt; # CURRENT read (if any) at each position
    while (<$fh>) {
	chomp;
	if (/^>REF /) {
	    print $gh "$_\n";
	} elsif (/^>([^ :]+)\S+ (\d+) /) {
	    my ($read,$pos) = ($1,$2);
	    # Remember (current) vertical position for this read
	    if (defined $posnOf{$read}) {
		my $p = $posnOf{$read};
		if ($readAt[$p] eq $read) { # it's _currently_ there
		    $readAt[$p] = undef;
		}
	    }
	    if (defined $readAt[$pos]) {
		my $r = $readAt[$pos];
		$posnOf{$r} = undef;
	    }
	    $posnOf{$read} = $pos;
	    $readAt[$pos] = $read;
	    # Remember annotation for this position
	    $annoAt{$pos} = $_;
	} else {
	    $ln ++;
	    while (@$pRem && $pRem->[0][1] < $ln) {
		shift @$pRem; # for efficiency...
	    }
	    my $len = length($_);
	    # Overwrite pRem positions with spaces
	    foreach my $pA (@$pRem) {
		last if ($pA->[0] > $ln);
		if ($pA->[1] >= $ln) {
		    my $p = $posnOf{$pA->[2]};
		    if (defined $p) {
			my $col = 2*($p+1);
			if ($col < $len) {
			    substr($_,$col,2) = '  ';
			}
		    } # else don't worry...
		}
	    }
	    # For each position with a waiting annotation, write and
	    # discard the annotation as long as the data is non-space.
	    foreach my $p (keys %annoAt) {
		my $col = 2*($p+1);
		if ($col < $len && substr($_,$col,2) ne '  ') {
		    print $gh "$annoAt{$p}\n";
		    delete $annoAt{$p};
		}
	    }
	    # Print the (edited) data line.
	    print $gh "$_\n";
	}
    }
}


1;
