package SGRP::ExtendReference;

use Exporter;
@ISA = qw(Exporter);

@EXPORT = qw(extendReference);

use SGRP::Variables;

use Utility;
use File::Basename;

use strict;

my $SSAHA2 = '/usr/local/ensembl/bin/ssaha2';

my $MAX_SLACK = 10;
my $MIN_SCORE = 100;
my $MIN_CHUNK = 50;
my $MAX_SHORTFALL = 50;
# Minimum gap between any pair of proposed transposons. There is a
# preponderance of gaps up to about 50. rd suggests using 5 or 10.
my $MIN_TRANSPOSON_GAP_SIZE = 10;
my $MIN_TRANSPOSON_PROPOSAL_COUNT = 2;

my $TEL_SZ = 20000;

# augmented.fa (produced by "sgrp.pl self") gives REF.chrNN sequences (the
# standard reference) and mappings from REF.chrNN.KKK to those (and implicitly
# to the backbone which is identical to REF.chrNN). It is "augmented" only
# with aux sequences for the deletion of repeats.

# extended.{fa,map} are created here. The sequences are named EXT.chrNN and
# are longer than REF.chrNN because they are extended by inserting transposon
# sequences (and perhaps one day other things). extended.map contains:
#   -- "EXT.chrNN 0 0 L" for every extended chromosome chrNN of length L
#   -- "EXT.chrNN.KKK ..." lines lifted over from the corresponding REF.chrNN.KKK
#      in augmented.map
#   -- "EXT.chrNN.MMM ..." lines, where MMM >= 500, for auxiliary sequences
#      formed by deleting selected extensions from EXT.chrNN.
# And extended.fa contains sequences for all these.

# We split extended.fa into pN/chrNN_extended.fa.
# genome.map maps REF onto EXT, and is used only to create pN/chrNN_extended_recased.fa
# from pN/chrNN_extended.fa. The recased file is then used to create the (iteration-zero)
# version of chrNN/layout.gz

sub extendReference {
    my ($species,@strains) = @_;
    unless (@strains > 0) {
	die("Usage: $0 extd species str1 str2...");
    }
    my $refDir = "$ROOT/ref/$species";
    my $teDir = "$refDir/te";
    my $pSz = findChromosomeSizes($species);
    my @ptbList;
    foreach my $str (@strains) {
	unless (-f "$teDir/$str.ptbs") {
	    my $teOut = runSsaha2AgainstTransposons($teDir,$str);
	    if ($teOut) {
		push @ptbList,findPossibleTransposonBorders($teDir,$teOut,$pSz,$str);
	    }
	}
    }
    (my $set = $species) =~ tr/a-z/A-Z/; # cere to CERE, etc
    my $pAlias = readAliases("$ROOT/aliases/strainSets.txt");
    my $pAllStrains = $pAlias->{$set};
    unless (defined @$pAllStrains) {
	die("Cannot expand strain-set alias $set");
    }
    my $ptiFile1 = findPossibleTransposonInsertions($teDir,$species,$set,$pAllStrains);
    if ($ptiFile1) {
	proposeTransposonInsertions($ptiFile1,$teDir,$pSz,$species,$set);
	extendReferenceWithTransposons($teDir,$set,$refDir);
    }
}

sub findChromosomeSizes {
    my ($sp) = @_;
    my %chrSz;
    my $fh = openToRead("$ROOT/ref/$sp/genome.sz");
    while (<$fh>) {
	chomp;
	my ($chr,$sz) = split;
	$chrSz{$chr} = $sz;
    }
    return \%chrSz;
}

sub runSsaha2AgainstTransposons {
    my ($teDir,$str) = @_;
    my $fa = "$teDir/te.fa";
    my $fq = "$ROOT/read/$str/renamed.fastq";
    my $g = "$teDir/$str.te.out";
    foreach my $f ($fa,$fq) {
	if (! -f $f) { # empty is OK!
	    report("missing: $f",1);
	    return;
	}
    }
    if (stamp($g)) {
	report("creating $g");
	`$SSAHA2 $fq $fa -tags 1 > $g`;
	report();
	unstamp($g);
    }
    return $g;
}

sub findPossibleTransposonBorders {
    my ($teDir,$ssaha2Out,$pSz,$str) = @_;
    my (@regs,$oldRead,$oldRLen);
    my $g = "$teDir/$str.ptbs";
    return $g unless (stamp($g));
    report("creating $g");
    my $fh = openToRead($ssaha2Out);
    my $gh = openToWrite($g);
    while (<$fh>) {
	chomp;
	next unless (/^ALIGNMENT/);
	my (undef,$score,$read,$elt,$rlo,$rhi,$clo,$chi,$fc,$mlen,$perci,$rlen) = split;
	# Ignore if telomeric element
	my ($teChr,$teLo,$teHi) = split(/\./,$elt);
	next if ($teLo <= $TEL_SZ || $teHi >= $pSz->{$teChr}-$TEL_SZ);
	unless ($read eq $oldRead) {
	    processRegions($oldRead,$oldRLen,$gh,@regs);
	    $oldRead = $read;
	    $oldRLen = $rlen;
	    @regs = ();
	}
	next unless ($score >= $MIN_SCORE);
	($rlo,$rhi) = sort {$a <=> $b} ($rlo,$rhi);
	($clo,$chi) = sort {$a <=> $b} ($clo,$chi);
	my $eLen = elementLength($elt);
	my $sb;
	# Test is to stop us suggesting repeated transposons in the aug sequence
	if ($eLen <= 7000) {
	    $sb = suggestsBorder($rlo,$rhi,$rlen,$eLen,$clo,$chi,$fc);
	}
	if ($sb) {
	    push @regs,[$rlo,$rhi,$sb,$elt,$eLen,$fc];
	} else {
	    push @regs,[$rlo,$rhi];
	}
    }
    processRegions($oldRead,$oldRLen,$gh,@regs);
    $gh->close();
    unstamp($g);
    report();
    return $g;
}

# Find the union of the "r" parts of the regions.
# Each region is of the form ($rlo,$rhi,$sb,$elt,$eLen,$fc)

sub processRegions {
    my ($read,$rLen,$gh,@regs) = @_;
    return unless (@regs);
    # Longer elements first, for printing choice
    @regs = sort {$b->[4] <=> $a->[4]} @regs;
    my ($lo,$hi) = @{$regs[0]};
    foreach my $i (1 .. $#regs) {
	my $pR = $regs[$i];
	$lo = min($lo,$pR->[0]);
	$hi = max($hi,$pR->[1]);
    }
    foreach my $pR (@regs) {
	if ($pR->[0] == $lo && $pR->[2] eq 'L') {
	    printf($gh "%s %4d %4d %s %s\n",$read,1,$lo-1,$pR->[3],$pR->[5]);
	    # We return rather than trying for an "R" as well because if we
	    # get an "L", it's overwhelmingly likely that any apparently
	    # non-te at the R end is poor sequence.
	    return;
	}
    }
    foreach my $pR (@regs) {
	if ($pR->[1] == $hi && $pR->[2] eq 'R') {
	    printf($gh "%s %4d %4d %s %s\n",$read,$hi+1,$rLen,$pR->[3],$pR->[5]);
	    return;
	}
    }
}

sub elementLength {
    my ($elt) = @_;
    my (undef,$lo,$hi) = split(/\./,$elt);
    return $hi-$lo+1;
}

sub suggestsBorder {
    my ($rlo,$rhi,$rlen,$eLen,$clo,$chi,$fc) = @_;
    if ($fc eq 'F') {
	# Border to LEFT of this match if rlo and clo both small
	return 'L' if ($rlo > $MIN_CHUNK && $clo <= $MAX_SLACK);
	return 'R' if ($rhi <= $rlen-$MIN_CHUNK && $chi >= $eLen-$MAX_SLACK);
    } else {
	return 'L' if ($rlo > $MIN_CHUNK && $chi >= $eLen-$MAX_SLACK);
	return 'R' if ($rhi <= $rlen-$MIN_CHUNK && $clo <= $MAX_SLACK);
    }
}

# Need to have already done alignments against ORIGINAL genome.fa file here,
# hence "_notext". (Augmented is OK too).

sub findPossibleTransposonInsertions {
    my ($teDir,$species,$set,$pAllStrains) = @_;
    my $g = "$teDir/$set.ptis1";
    return $g unless (stamp($g));
    report("creating $g");
    my $gh = openToPipe("sort > $g");
    my $ans = $g;
    foreach my $s (sort @$pAllStrains) {
	my $aFile = "$ROOT/read/$s/${species}_notext/alilines.all.txt";
	my $pFile = "$teDir/$s.ptbs";
	if (! -s $aFile) {
	    report("missing or empty: $aFile",1);
	    $ans = undef;
	    last;
	} elsif (! -f $pFile) {
	    report("missing: $pFile",1);
	    $ans = undef;
	    last;
	} else {
	    possibleTransposonInsertions($aFile,$pFile,$gh);
	}
    }
    unstamp($g);
    report();
    return $ans;
}

sub possibleTransposonInsertions {
    my ($aFile,$pFile,$gh) = @_;
    my %ptb;
    my $fh = openToRead($pFile);
    while (<$fh>) {
	chomp;
	my ($read,@reg) = split;
	$ptb{$read} = \@reg;
    }
    my $fh = openToRead($aFile);
    my ($curRead,$curTh);
    $curTh = $MIN_SCORE;
    while (<$fh>) {
	chomp;
	my (undef,$score,$read,$chr,$rlo,$rhi,$clo,$chi,$fc) = split;
	# No aux sequences and no chrm or scplasm1:
	next unless (defined $ptb{$read} && $chr =~ /chr..$/);
	unless ($read eq $curRead) {
	    $curRead = $read;
	    $curTh = max($MIN_SCORE,$score-$MAX_SHORTFALL);
	}
	next if ($score < $curTh);
	($rlo,$rhi) = sort {$a <=> $b} ($rlo,$rhi);
	($clo,$chi) = sort {$a <=> $b} ($clo,$chi);
	my ($regLo,$regHi,$teId,$teFC) = @{$ptb{$read}};
	my ($off,$dir);
	if ($regLo == 1) {
	    my $d = $regHi-$rhi;
	    next if (abs($d) > $MAX_SLACK);
	    $off = ($fc eq 'F') ? $chi+$d : $clo-1-$d;
	    $dir = 'R';
	} else {
	    my $d = $regLo-$rlo;
	    next if (abs($d) > $MAX_SLACK);
	    $off = ($fc eq 'F') ? $clo-1+$d : $chi-$d;
	    $dir = 'L';
	}
	printf($gh "%s %7d %s %s %s %s\n",$chr,$off,$dir,$teId,$fc eq $teFC ? 'F' : 'C',$read);
    }
}

sub proposeTransposonInsertions {
    my ($inFile,$teDir,$pSz,$species,$set) = @_;
    my $outFile = "$teDir/$set.ptis2";
    return unless (stamp($outFile));
    report("creating $outFile");
    my $pOri = readAliases("$teDir/te.ori");
    my %count;
    my %count2;
    my %byStrain;
    my $pExist = existingTransposonBoundaries($teDir);
    my $fh = openToRead($inFile);
    my $gh = openToPipe("sort > $outFile");
    while (<$fh>) {
	chomp;
	my ($chr,$off,undef,$teId,$fc,$read) = split;
	$chr =~ s/^REF\.//;
	my ($str) = split(/-/,$read);
	#unless (closeToExistingTransposon($chr,$off,$pExist)) {
	$count{$chr}{$off} ++;
	$count2{$chr}{$off}{"$teId $fc"} ++;
	#}
    }
    foreach my $chr (sort keys %count) {
	my @offs = sort {$a <=> $b} keys %{$count{$chr}};
	my (%nFrom,%maxCount,%maxMap,%hasOri);
	for (my $i=0; $i<=$#offs; $i++) {
	    my $oi = $offs[$i];
	    for (my $j=$i; $j<=$#offs && $offs[$j] < $oi+$MAX_SLACK; $j++) {
		my $oj = $offs[$j];
		my $c = $count{$chr}{$oj};
		$nFrom{$oi} += $c;
		foreach my $ori (absoluteOrientations($count2{$chr}{$oj},$pOri)) {
		    $hasOri{$oi}{$ori} = 1;
		}
		if ($c > $maxCount{$oi}) {
		    $maxCount{$oi} = $c;
		    $maxMap{$oi} = $oj;
		}
	    }
	}
	@offs = sort {$nFrom{$b} <=> $nFrom{$a}} @offs;
	foreach my $off (@offs) {
	    if ($nFrom{$off} >= $MIN_TRANSPOSON_PROPOSAL_COUNT) {
		my $off2 = $maxMap{$off};
		if ($off2 >= $TEL_SZ && $off2 <= $pSz->{$chr}-$TEL_SZ) {
		    my $pC2 = $count2{$chr}{$off2};
		    my @keys = sort {$pC2->{$b} <=> $pC2->{$a}} keys %$pC2;
		    printf($gh "%s %7d %s %3d %3d\n",
			   $chr,$off2,$keys[0],$nFrom{$off});
		    my @oris = keys %{$hasOri{$off}};
		    foreach my $off3 (@offs) {
			if (abs($off3-$off) < $MIN_TRANSPOSON_GAP_SIZE) {
			    foreach my $ori (@oris) {
				delete $hasOri{$off3}{$ori};
			    }
			    unless (keys %{$hasOri{$off3}}) {
				delete $nFrom{$off3};
			    }
			}
		    }
		}
	    }
	}
    }
    unstamp($outFile);
    report();
}

sub existingTransposonBoundaries {
    my ($teDir) = @_;
    my $fh = openToRead("$teDir/te.fa");
    my %exist;
    while (<$fh>) {
	if (/^>(\S+)/) {
	    my ($chr,$lo,$hi) = split(/\./,$1);
	    push @{$exist{$chr}},($lo,$hi);
	}
    }
    foreach my $chr (keys %exist) {
	@{$exist{$chr}} = sort {$a <=> $b} @{$exist{$chr}};
    }
    return \%exist;
}

sub absoluteOrientations {
    my ($pH,$pOri) = @_;
    my %isOri;
    foreach my $str (keys %$pH) {
	my ($te,$ori1) = split(" ",$str);
	my $ori2 = $pOri->{$te};
	my $ori = $ori1 eq $ori2 ? 'F' : 'C';
	$isOri{$ori} = 1;
    }
    return keys %isOri;
}

# Files:
#
# genome.fa     original reference sequence
# augmented.fa  genome.fa augmented with aux sequences for deletions from genome.fa
# augmented.map map file to go with the above
# extended.fa   augmented.fa with extra transposons inserted into sequences and
#               extra aux sequences for omitting those
# extended.map  map file to go with the above: lifted-over augmented.map
# genome.map    map file to show how genome.fa maps onto extended.fa

sub extendReferenceWithTransposons {
    my ($teDir,$set,$refDir) = @_;
    my $extFA = "$refDir/extended.fa";
    extendWithTransposons("$refDir/augmented.fa",
			  "$refDir/augmented.map",
			  "$teDir/te.fa",
			  "$teDir/$set.ptis2",
			  $extFA,
			  "$refDir/extended.map",
			  "$refDir/genome.map");
    report("splitting and recasing $extFA");
    SGRP::MakeGenomes::splitGenome($extFA,$refDir,'_extended');
    readAndRecaseExtended($refDir);
    report();
}

sub extendWithTransposons {
    my ($inFA,$inMap,$teFA,$teSpec,$extFA,$extMap,$genMap) = @_;
    if (stamp($extFA)) {
	report("creating $extFA etc");
	my $pSpec = readTESpecs($teSpec);
	my $pSeqs = readFasta($teFA);
	extendWithTransposons1($inFA,$inMap,$pSpec,$pSeqs,
			       $extFA,$extMap,$genMap);
	unstamp($extFA);
	report();
    }
}

sub readTESpecs {
    my ($f) = @_;
    my $fh = openToRead($f);
    my %spec;
    while (<$fh>) {
	chomp;
	my ($chr,$off,$te,$fc) = split;
	my (undef,$teLo,$teHi) = split(/\./,$te);
	# We don't subtract 1 here, even though .map files are
	# zero-offset, so this means "insert transposon BEFORE
	# current position".
	$spec{$chr}{$off} = [$te,$fc,$teHi-$teLo+1];
    }
    return \%spec;
}

sub readFasta {
    my ($f) = @_;
    my $fh = openToRead($f);
    my %seq;
    my ($chr,$buf);
    while (<$fh>) {
	chomp;
	if (/^>(\S+)/) {
	    $seq{$chr} = $buf if ($buf);
	    $buf = "";
	    $chr = $1;
	} else {
	    $buf .= $_;
	}
    }
    $seq{$chr} = $buf if ($buf);
    return \%seq;
}

sub extendWithTransposons1 {
    my ($augFA,$augMap,$pSpec,$pSeqs,$extFA,$extMap,$genMap) = @_;
    my $mfh = openToRead($augMap);
    my $mghGen = openToWrite($genMap);
    my $mghExt = openToWrite($extMap);
    my $afh = openToRead($augFA);
    chomp (my $afhLine = <$afh>);
    my $agh = openToWrite($extFA);
    my ($oldChr,$pChrMap);
    while (<$mfh>) {
	chomp;
	my $mapLine = $_;
	# $chr will be REF... at this stage
	my ($chr,$lo1,$lo2,$chrLen) = split(" ",$mapLine);
	my $inSeq;
	unless ($chr eq $oldChr) {
	    # Get header and sequence for new chr or chunk
	    $oldChr = $chr;
	    if ($afhLine eq ">$chr") {
		($inSeq,$afhLine) = readSequence($afh);
		#print $agh ">$chr\n";
	    }
	    #print "PROCESSING $chr\n";
	}
	(my $chrBase = $chr) =~ s/^REF\.//;
	my $pH = $pSpec->{$chrBase};
	if (defined $pH) {
	    $pChrMap = printExtendedSequenceAndMaps($chr,$chrLen,$inSeq,
						    $pSeqs,$pH,$agh,$mghExt,$mghGen);
	} else {
	    if ($chr =~ /chr\d\d\.\d\d\d$/) {
		liftoverMapline($mapLine,$pChrMap,$mghExt); # aux sequence
	    } else {
		print $mghGen "$mapLine\n"; # transposon-less chromosome, e.g. chrm
		$mapLine =~ s/^REF/EXT/;
		print $mghExt "$mapLine\n";
	    }
	    if ($inSeq) {
		#print "SIMPLE: $chrBase\n";
		# Copy the sequence...
		printFasta("EXT.$chrBase",$inSeq,$agh);
		$inSeq = undef;
	    }
	}
    }
    return $extFA;
}

sub readSequence {
    my ($fh) = @_;
    my $seq;
    while (<$fh>) {
	chomp;
	if (/^>/) {
	    return ($seq,$_);
	} else {
	    $seq .= $_;
	}
    }
    return ($seq,undef);
}

sub printExtendedSequenceAndMaps {
    my ($chr,$chrLen,$inSeq,$pSeqs,$pH,$agh,$mghExt,$mghGen) = @_;

    $chr =~ s/REF/EXT/;

    # Each table element is [refOff,teId,teFC,teLen,$extOff]
    # where extOff is offset of START of this transposon
    my @table;
    my $extra = 0;
    foreach my $off (sort {$a <=> $b} keys %$pH) {
	push @table,[$off,@{$pH->{$off}},$off+$extra];
	$extra += $table[-1][3];
    }

    # Map from original reference to extended sequence, and print to genome.map
    my $pChrMap = referenceToExtendedMap($chr,\@table,$chrLen,$mghGen);

    # Define and print the extended sequence
    my $extSeq = printExtendedSequence($chr,\@table,$inSeq,$pSeqs,$agh);
    # Print a map line for it
    my $len = length($extSeq);
    print $mghExt "$chr 0 0 $len\n";

    # Find all the possible aux sequences. We assume we have to omit a
    # CONTIGUOUS range of insertions each time, because each insertion will
    # be >> 1200 bp.
    my $pDelRanges = findDeletionRanges(\@table);
    printExtendedAuxSequences($chr,$pDelRanges,$extSeq,\@table,$pChrMap,$agh,$mghExt);

    return $pChrMap;
}

sub referenceToExtendedMap {
    my ($chr,$pTable,$chrLen,$gh) = @_;
    $chr =~ s/^EXT/REF/;
    my @chrMap;
    my $extra = 0;
    my $prev = 0;
    foreach my $i (0 .. $#$pTable+1) {
	my $here = $i <= $#$pTable ? $pTable->[$i][0] : $chrLen;
	push @chrMap,[$prev,$prev+$extra,$here-$prev];
	if ($i <= $#$pTable) {
	    $prev = $here;
	    $extra += $pTable->[$i][3];
	}
    }
    foreach my $pA (@chrMap) {
	printf($gh "%s %d %d %d\n",$chr,@$pA);
    }
    return \@chrMap;
}

sub printExtendedSequence {
    my ($chr,$pTable,$inSeq,$pSeqs,$gh) = @_;
    my $extSeq;
    my $prev = 0;
    foreach my $i (0 .. $#$pTable) {
	my ($here,$te,$fc,$len) = @{$pTable->[$i]};
	$extSeq .= substr($inSeq,$prev,$here-$prev);
	$extSeq .= extractSequence($pSeqs,$te,$fc,$len);
	$prev = $here;
    }
    $extSeq .= substr($inSeq,$prev);
    printFasta($chr,$extSeq,$gh);
    return $extSeq;
}

sub extractSequence {
    my ($pSeqs,$te,$fc,$len) = @_;
    my $seq = $pSeqs->{$te};
    unless (length($seq) == $len) {
	die("Length mismatch");
    }
    if ($fc eq 'C') {
	$seq = reverse($seq);
	$seq =~ tr/ACGT/TGCA/;
    }
    return $seq;
}

sub printFasta {
    my ($key,$seq,$gh) = @_;
    print $gh ">$key\n" if ($key);
    my $iMax = length($seq)-1;
    for (my $i=0; $i<=$iMax; $i+=60) {
	print $gh substr($seq,$i,60) . "\n";
    }
}

sub findDeletionRanges {
    my ($pTable) = @_;
    my @delRanges;
    foreach my $i (0 .. $#$pTable) {
	foreach my $j ($i .. $#$pTable) {
	    last if ($pTable->[$j][0]-$pTable->[$i][0] > 2400);
	    push @delRanges,[$i,$j];
	}
    }
    return \@delRanges;
}

sub printExtendedAuxSequences {
    my ($chr,$pDelRanges,$extSeq,$pTable,$pChrMap,$agh,$mgh) = @_;
    my $nAux = 500; # Assume at most 500 aux sequences per chr already!
    foreach my $pRange (@$pDelRanges) {
	# Cook up an aux sequence for omitting insertions in each set
	my ($i,$j) = @$pRange;
	# Initial segment
	my $start = $pTable->[$i][4]-1200;
	my $auxSeq = substr($extSeq,$start,1200);
	my $auxOff = 0;
	my @auxMap = [$auxOff,$start,1200];
	$auxOff += 1200;
	# Internal segments: k to k+1
	foreach my $k ($i .. $j-1) {
	    # Offset of one after END of transposon:
	    my $start = $pTable->[$k][4]+$pTable->[$k][3]+1;
	    my $len = $pTable->[$k+1][4] - $start;
	    $auxSeq .= substr($extSeq,$start,$len);
	    push @auxMap,[$auxOff,$start,$len];
	    $auxOff += $len;
	}
	# Final segment
	my $start = $pTable->[$j][4]+$pTable->[$j][3]+1;
	$auxSeq .= substr($extSeq,$start,1200);
	push @auxMap,[$auxOff,$start,1200];
	# Print auxSeq and auxMap
	# Identifier for the aux that skips the insertion: EXT.chr01.500 etc
	my $key = sprintf("%s.%3d",$chr,$nAux);
	$key =~ tr/ /0/;
	printFasta($key,$auxSeq,$agh);
	foreach my $pA (@auxMap) {
	    #liftoverMapline("$key @$pA",$pChrMap,$mgh);
	    printf($mgh "%s %d %d %d\n",$key,@$pA);
	}
	$nAux ++;
    }
}

# $line will be from an aux sequence,
# e.g. EXT.chr01.500 0 136284 1200
# mapping from EXT.chr01.500 to REF.chr01
# $pA will be e.g. 139134 156450 41715
# mapping from REF.chr01 to EXT.chr01.

#  AUX:  [ ---------- )
#       lo1   len
#        |
#  REF: lo2          lim 
#           [ ----------------- )
#          mLo1     mLen      mLim
#           |
#  EXT:     [ ----------------- )
#          mLo2

sub liftoverMapline {
    my ($line,$pMap,$gh) = @_;

    # $lo2 is lo in REF
    my ($chr,$lo1,$lo2,$len) = split(" ",$line);
    $chr =~ s/^REF/EXT/;
    # limit (one past the end) in REF
    my $lim = $lo2+$len;
	
    if (defined @$pMap) {
	foreach my $pA (@$pMap) {
	    
	    # $mLo1 is lo in REF, $mLo2 is lo in EXT
	    my ($mLo1,$mLo2,$mLen) = @$pA;
	    last if ($mLo1 >= $lim); # gone past $line
	    my $mLim = $mLo1+$mLen; # in REF
	    next if ($mLim <= $lo2); # not at $line yet
	    
	    # print a line for the overlap ...
	    my $delta = $mLo2-$mLo1; # to lift REF to EXT
	    # start of mapped region in AUX coords:
	    my $oLo1 = $lo1 + max(0,$mLo1-$lo2);
	    # start of mapped region in EXT coords:
	    my $oLo2 = $delta + max($lo2,$mLo1);
	    my $oLim = $delta + min($lim,$mLim);
	    # length of mapped region:
	    my $oLen = $oLim - $oLo2;
	    print $gh "$chr $oLo1 $oLo2 $oLen\n";
	}
    } else {
	# e.g. transposon-less paradoxus ...
	print $gh "$chr $lo1 $lo2 $len\n"; # just copy, but REF->EXT
    }
}

sub readAndRecaseExtended {
    my ($refDir) = @_;
    my $pMap = readReferenceRegions("$refDir/genome.map");
    foreach my $f (glob("$refDir/p?/*_extended.fa")) {
	my $chr = basename($f,'_extended.fa');
	readAndRecaseExtended1($f,$pMap);
    }
}

sub readReferenceRegions {
    my ($f) = @_;
    my %map;
    my $fh = openToRead($f);
    while (<$fh>) {
	chomp;
	my ($chr,$from,$to,$len) = split;
	$chr =~ s/^REF\.//;
	# Inclusive regions, offset 0:
	push @{$map{$chr}},[$to,$to+$len-1];
    }
    return \%map;
}

sub readAndRecaseExtended1 {
    my ($f,$pMap) = @_;
    my $g = $f;
    $g =~ s/extended.fa$/extended_recased.fa/;
    moveAside($g) if (-s $g);
    #return if (-s $g);
    my $pSeq = readFastaFile($f);
    foreach my $chr (keys %$pSeq) {
	my $pMapChr = $pMap->{$chr} || die("No pMap found for $chr, file $f");
	my $seq = $pSeq->{$chr};
	# Lower-case regions not in $pMapChr
	my $hi = $pMapChr->[0][0];
	substr($seq,0,$hi) =~ tr/A-Z/a-z/ if ($hi > 0);
	foreach my $i (1 .. $#$pMapChr) {
	    my $lo = $pMapChr->[$i-1][1];
	    my $hi = $pMapChr->[$i][0];
	    substr($seq,$lo+1,$hi-$lo-1) =~ tr/A-Z/a-z/;
	}
	substr($seq,$pMapChr->[-1][1]+1) =~ tr/A-Z/a-z/;
	$pSeq->{$chr} = $seq;
    }
    writeFastaFile($pSeq,$g);
}

1;
