package SGRP::ExtractSNPs;

use Exporter;
@ISA = qw(Exporter);

@EXPORT = qw(extractSNPsPreImp
	     extractSNPRates
	     prepareMargaritaInput
	     extractSNPsPostImp);

## sgrp.pl esnp -> extractSNPsPreImp
## sgrp.pl esnr -> extractSNPRates
## sgrp.pl isnp -> extractSNPsPostImp
##
##   and see also
##
## sgrp.pl pmar -> SGRP::PrepareMargarita::prepareMargaritaInput
## sgrp.pl rsnp -> SGRP::ReportSNPs::reportSNPs

# NOTE: there are two different SNP extraction processes here, one
# for reporting (esnp) and one for Margarita (pmar). They could be integrated,
# but the esnp process also prepares SNP counts used to adjust error rates
# when the Margarita input is created, so we may as well leave things alone.

# ALSO NOTE: offsets are from 1 for all SNP reports; digit after the "."
# in clbs_clashes.gz is 0 if reference is a nucleotide, then 1, 2, ... for
# any following insertions in strains. In snp_report.gz (from sorted.gz)
# we report only true SNPs (nucleotide->nucleotide, no gaps) and of
# those, only the ones that pass NQS. In isnp_report.gz (from imputed.gz)
# we again only report true SNPs, but this time don't insist on NQS
# (lower case means "filled in/corrected" rather than "fails NQS" in
# imputed files).

use SGRP::Variables;

use Utility;
use File::Basename;

use Carp;

use strict;

##
## ---------------------------------------------------------------------
## sgrp.pl ensp cere methc CERE 0 1 2 ...
## ---------------------------------------------------------------------
##
## extractSNPsPreImp
##   -> extractSNPsPreImp1
##        SRC: .../chrNN/sorted.gz
##        -> createSNPCountFile
##             TGT: .../chrNN/ssnp_counts.txt [for esnr]
##        TGT: .../chrNN/ssnp_rates.txt [for pmar]
##        -> reportSNPs
##             SRC: .../chrNN/sorted.gz
##             SRC: .../chrNN/clbs_clashes.gz
##             TGT: .../chrNN/ssnp_report.gz [for rsnp]

# See 0807 for reasoning:
my $MIN_RELIABLE_QUALITY = 45;

sub extractSNPsPreImp {
    my ($species,$filtMeth,$strainSet,@parts) = @_;
    unless (@parts) {
	confess("Usage: $0 esnp species filtMeth strainSet part1 part2 ...");
    }
    my ($iter,$dir) = getCurrentIteration($species,$filtMeth,$strainSet,'p?');
    foreach my $part (@parts) {
	$part = "p$part" if ($part =~ /^[0-9]$/);
	foreach my $chrDir (glob("$dir/$part/*")) {
	    extractSNPsPreImp1($chrDir);
	}
    }
}

sub extractSNPsPreImp1 {
    my ($dir) = @_;
    my $src;
    foreach my $suf (qw(z y m)) {
	$src = $suf eq 'm' ? "$dir/sorted.gz" : "$dir/sorted_$suf.gz";
	last if (isComplete($src,1));
    }
    return unless (isComplete($src,1));
    my $snpRepFile = "$dir/ssnp_report.gz";
    my $snpCountFile = "$dir/ssnp_counts.txt";
    if (stamp($snpCountFile)) {
	report("creating $snpCountFile");
	createSNPCountFile($src,$snpCountFile);
	report();
	unstamp($snpCountFile);
    }
    # Extract chr-specific rates (as opposed to esnr which
    # creates rates for whole genome)
    extractSNPRates1("$dir/ssnp_rates.txt",$dir);
    my $cla = "$dir/clbs_clashes.gz";
    if (-s $cla && stamp($snpRepFile)) {
	report("creating $snpRepFile");
	reportSNPs($src,$cla,$snpRepFile);
	report();
	unstamp($snpRepFile);
    }
}

# $tgt is used for adjusting error probs on SNPs before handing
# to mgr_in. It's specific for a single chr/contig.

sub createSNPCountFile {
    my ($src,$tgt) = @_;
    my $fh = openToRead($src);
    $_ = <$fh>;
    chomp;
    s/^>//;
    my @strains = split;
    # count[strainIdx][quality][0|1]
    my @count;
    my $zero = ord('!');
    my $off;
    while (<$fh>) {
	chomp;
	my $ref = substr($_,0,1);
	if ($ref =~ /^[ACGT]$/) {
	    for (my $k=0; $k<=$#strains; $k++) {
		my $kk = ($k+1)*2;
		my $val = substr($_,$kk,1);
		$val =~ tr/a-z/A-Z/;
		if ($val =~ /^[ACGT]$/) {
		    my $q = ord(substr($_,$kk+1,1))-$zero;
		    my $diff = $val ne $ref;
		    $count[$k][$q][$diff] ++;
		}
	    }
	    $off ++;
	}
    }
    my $gh = openToWrite($tgt);
    foreach my $k (0 .. $#strains) {
	my $pCount = $count[$k];
	my ($bigK,$sumTop,$sumBot);
	for (my $q=$MIN_RELIABLE_QUALITY; $q<=$#$pCount; $q++) {
	    if (defined $pCount->[$q]) {
		my @n = @{$pCount->[$q]};
		my $nq = $n[0]+$n[1];
		my $eq = exp(-log(10)*$q/10);
		$bigK += $n[1];
		$sumTop += $nq*$eq;
		$sumBot += $nq*(1-4*$eq/3);
	    }
	}
	printf($gh "%s %d %5.3f %5.3f\n",$strains[$k],$bigK,$sumTop,$sumBot);
    }
    $gh->close();
}

sub extractSNPRates1 {
    my ($tgt,@srcDirs) = @_;
    return unless (stamp($tgt));
    report("creating $tgt");
    my (%bigK,%sumTop,%sumBot);
    foreach my $part (0 .. 9) {
	foreach my $chrDir (@srcDirs) {
	    my $f = "$chrDir/ssnp_counts.txt";
	    if (-s $f) {
		my $fh = openToRead($f);
		while (<$fh>) {
		    chomp;
		    my ($str,$k,$st,$sb) = split;
		    $bigK{$str} += $k;
		    $sumTop{$str} += $st;
		    $sumBot{$str} += $sb;
		}
	    } elsif (-d $chrDir) {
		warn("Cannot find $f");
	    }
	}
    }
    my $gh = openToWrite($tgt);
    foreach my $str (sort keys %bigK) {
	my $rate = $sumBot{$str} ? max(0,($bigK{$str}-$sumTop{$str})/$sumBot{$str}) : 0;
	printf($gh "%s %10.8f %d %5.3f %5.3d\n",
	       $str,$rate,$bigK{$str},$sumTop{$str},$sumBot{$str});
    }
    $gh->close();
    unstamp($tgt);
    report();
}

# reportSNPs is called from both extractSNPsPreImp1 and
# extractSNPsPostImp1.

# The $cla file (and in consequence M/N vote figures) should only be
# there for normal SNPs, not imputed.

# $matchCount[$k][$i][$j]:
#   $k is strain index (zero when strain is ref)
#   $i is 0 when ref is nucleotide, 1 when ref is gap
#   $j ditto for strain, but 2 when strain is absent

# Half-window size, and full-window size.
my $HWSZ = 15;
my $HWSZ2 = $HWSZ*2;
# For printing:
my $PHWSZ = 10;

# Good nucleotide (in the strict sense) values.
my %GOOD = map {$_ => 1} qw(A C G T);

sub reportSNPs {
    my ($src,$cla,$tgt) = @_;
    my @matchCount;
    my $fh = openToRead($src); # sorted.gz or imputed.gz
    my $pRatio = readRatios($cla) if (defined $cla); # clbs_clashes.gz
    my $gh = openToWrite($tgt);
    my $chr = basename(dirname($src));
    $_ = <$fh>;
    chomp;
    s/^>//;
    my @strains = split;
    my @buf;
    for (my $k=0; $k<2*$HWSZ; $k++) {
	$_ = <$fh>;
	chomp;
	push @buf,$_;
	augmentMatchCount($_,\@matchCount);
    }
    # off1 will be offset from 1; off2 is 0 for no gap, N for Nth gap,
    # as in CombineLayoutByStrain.pm. offA is offset in alignment coords, from 1.
    my ($off1,$off2,$offA) = (0,0,0);
    foreach my $k (0 .. $HWSZ-1) {
	if ($buf[$k] =~ /^[-0-9=]/) {
	    $off2 ++;
	} else {
	    $off1 ++;
	    $off2 = 0;
	}
	$offA ++;
    }
    while (<$fh>) {
	chomp;
	push @buf,$_;
	augmentMatchCount($_,\@matchCount);
	if ($buf[$HWSZ] =~ /^[-0-9=]/) {
	    $off2 ++;
	} else {
	    $off1 ++;
	    $off2 = 0;
	}
	$offA ++;
	spotSNPs($chr,$off1,$offA,\@buf,\@strains,$pRatio->{$off1}[$off2],$gh);
	shift @buf;
    }
    unshift @strains,'REF'; # for reporting match counts
    reportMatchCounts(\@matchCount,\@strains,$gh);
    $gh->close();
}

sub readRatios {
    my ($f) = @_;
    my $fh = openToRead($f);
    my %ratio;
    while (<$fh>) {
	chomp;
	my ($pre,$post) = (/^(.*) : (.*)/);
	my @a = split(/ /,$pre);
	my ($off,$strain,$rBase) = split(/ /,$post);
	my ($off1,$off2) = split(/\./,$off);
	my $sBase = $a[0];
	$sBase =~ tr/_a-z/-A-Z/;
	$sBase =~ s/\d/-/;
	$rBase =~ s/\d/-/;
	# If equal, we had clashes between the strain reads but decided
	# on the reference nucleotide, so no SNP will result.
	unless ($sBase eq $rBase) {
	    my $num = length($a[1]);
	    my $den = $num;
	    for (my $k=3; $k<=$#a; $k+= 2) {
		$den += length($a[$k]);
	    }
	    $ratio{$off1}[$off2]{$strain} = "$num/$den";
	}
    }
    return \%ratio;
}

# Indices for pMC:
#
# First index: ref
#   0 for nucleotide, 1 for gap
# Second index: strain
#   0 for nucleotide, 1 for gap, 2 for space
# Third index:
#   0 for strain upper case (or space), 1 for strain lower case
# Fourth index:
#   0 if ref and strain agree when recased, 1 if they don't.

sub augmentMatchCount {
    my ($line,$pMC) = @_;
    my $ksMax = length($line)/2-1;
    my $kks=0;
    my $rv = substr($_,0,1);
    $rv =~ tr/=_acgt/--ACGT/;
    my $i = ($rv eq '-') ? 1 : 0; # first index
    for (my $ks=0; $ks<=$ksMax; $ks++) {
	my $v = substr($_,$kks,1);
	my $k = ($v =~ /[_acgt]/) ? 1 : 0; # third index
	$v =~ tr/=_acgt/--ACGT/;
	my $j = ($v eq ' ') ? 2 : ($v eq '-' ? 1 : 0); # second index
	my $l = ($v eq $rv) ? 0 : 1; # fourth index
	$pMC->[$ks][$i][$j][$k][$l] ++;
	$kks += 2;
    }
}

sub spotSNPs {
    my ($chr,$off1,$offA,$pBuf,$pStrains,$pRat,$gh) = @_;
    my $tgt = $pBuf->[$HWSZ];
    # k is 0 for ref, 1 to kMax for strains, so we use pStrains->[$k-1]
    my $kMax = length($tgt)/2-1;
    my $rv = substr($tgt,0,1); # ref value
    $rv =~ tr/acgt/ACGT/; # ref value can be lower-case for imputed!
    # We don't count indels as SNPs (if we did, we'd have to update
    # CombineLayoutByStrain.pm too)
    if ($GOOD{$rv}) {
	for (my $k=1; $k<=$kMax; $k++) {
	    my $k2=$k*2;
	    my $v = substr($tgt,$k2,1);
	    # For imputed case, we want ALL snps, not just upper-case.
	    # For sorted (pre-imputed) we just want the ones that pass NQS.
	    $v =~ tr/acgt/ACGT/ unless ($pRat);
	    if ($GOOD{$v} && $v ne $rv) {
		my $strain = $pStrains->[$k-1];
		my $rat = $pRat->{$strain} if (defined $pRat);
		reportSNP($chr,$off1,$offA,$strain,$rat,$k2,$pBuf,$gh);
	    }
	}
    }
}

sub reportSNP {
    my ($chr,$off,$offA,$str,$rat,$k2,$pBuf,$gh) = @_;
    my $rseq = getSequence($pBuf,0,1);
    my $sseq = getSequence($pBuf,$k2,1);
    my $qseq = getSequence($pBuf,$k2+1,0);
    ($rseq,$sseq,$qseq) = removeDoubleGaps($rseq,$sseq,$qseq);
    printf($gh "%s %s %d\n",$rseq,$chr,$off);
    my $qc = substr($qseq,$PHWSZ,1);
    my $qan = ($qc eq ' ' ? 127 : ord($qc)) - ord('!');
    printf($gh "%s %s %d %s %s>%s %d %s\n",
	   $sseq,$chr,$offA,$str,
	   substr($rseq,$PHWSZ,1),
	   substr($sseq,$PHWSZ,1),
	   $qan,
	   $rat);
    printf($gh "%s\n\n",$qseq);
}

sub getSequence {
    my ($pBuf,$k2,$digs2Gaps) = @_;
    my $ans;
    foreach my $k (0 .. $HWSZ2) {
	$ans .= substr($pBuf->[$k],$k2,1) || ' ';
    }
    if ($digs2Gaps) {
	$ans =~ s/\d/-/g;
    }
    return $ans;
}

sub removeDoubleGaps {
    my ($rseq,$sseq,$qseq) = @_;
    my $mid = $HWSZ;
    my ($rseqL,$sseqL,$qseqL) = 
	removeDoubleGaps1(-1,substr($rseq,0,$mid),substr($sseq,0,$mid),substr($qseq,0,$mid));
    my ($rseqR,$sseqR,$qseqR) =
	removeDoubleGaps1(1,substr($rseq,$mid+1),substr($sseq,$mid+1),substr($qseq,$mid+1));
    $rseq = sprintf("%s%s%s",$rseqL,substr($rseq,$mid,1),$rseqR);
    $sseq = sprintf("%s%s%s",$sseqL,substr($sseq,$mid,1),$sseqR);
    $qseq = sprintf("%s%s%s",$qseqL,substr($qseq,$mid,1),$qseqR);
    return ($rseq,$sseq,$qseq);
}

sub removeDoubleGaps1 {
    my ($dirn,@lines) = @_;
    if ($dirn < 0) {
	foreach my $k (0 .. $#lines) {
	    $lines[$k] = reverse($lines[$k]);
	}
    }
    foreach my $k (0,1) {
	$lines[$k] =~ s/[0-9]/-/g; # easiest for understanding reports; rline and sline only
    }
    my @ans = removeDoubleGaps2(@lines);
    foreach my $k (0 .. $#ans) {
	$ans[$k] = substr($ans[$k],0,$PHWSZ);
    }
    if ($dirn < 0) {
	foreach my $k (0 .. $#ans) {
	    $ans[$k] = reverse($ans[$k]);
	}
    }
    return @ans;
}

sub removeDoubleGaps2 {
    my ($rLine,$sLine,$qLine) = @_;
    my $nToRm = $HWSZ-$PHWSZ;
  LOOP:
    foreach my $k (0 .. $PHWSZ-1) {
	while (substr($rLine,$k,1) eq '-' && substr($sLine,$k,1) eq '-') {
	    substr($rLine,$k) = substr($rLine,$k+1);
	    substr($sLine,$k) = substr($sLine,$k+1);
	    substr($qLine,$k) = substr($qLine,$k+1);
	    $nToRm --;
	    last LOOP if ($nToRm == 0);
	}
    }
    return ($rLine,$sLine,$qLine);
}

sub reportMatchCounts {
    my ($pMC,$pStr,$gh) = @_;
    print $gh "\nMatchCounts\n";
    foreach my $ks (0 .. $#$pStr) {
	printf($gh "%11s",$pStr->[$ks]);
	foreach my $i (0,1) {
	    foreach my $j (0,1,2) {
		foreach my $k (0,1) {
		    foreach my $l (0,1) {
			printf($gh " %d",$pMC->[$ks][$i][$j][$k][$l]);
		    }
		}
	    }
	}
	print $gh "\n";
    }
}

sub countImputedSNPs {
    my ($f,$g) = @_;
    my ($nSites,$nSNPs) = (0,0);
    my %snps;
    my $fh = openToRead($f);
    my $hdr = <$fh>;
    chomp $hdr;
    my @strains = ('AAREF',split(/ /,substr($hdr,1)));
    my @count;
    while (<$fh>) {
	chomp;
	s/(.)./\1/g; # get rid of quality values
	tr/a-z/A-Z/; # upper case everything
	my $iMax = length($_)-1;
	my @iList;
	my @cList;
	for (my $i=0; $i<=$iMax; $i++) {
	    my $ci = substr($_,$i,1);
	    if ($ci =~ /[ACGT]/) {
		push @iList,$i;
		push @cList,$ci;
	    }
	}
	my $foundSNP = 0;
	while (@iList > 1) {
	    my $i = shift @iList;
	    my $ci = shift @cList;
	    for (my $jj=0; $jj<=$#iList; $jj++) {
		my $j = $iList[$jj];
		my $cj = $cList[$jj];
		unless ($cj eq $ci) {
		    $count[$i][$j] ++;
		    $foundSNP = 1;
		}
	    }
	}
	$nSNPs ++ if ($foundSNP);
	$nSites ++;
    }
    for (my $i = 0; $i <= $#count; $i++) {
	my $si = $strains[$i];
	my $pC = $count[$i];
	for (my $j = $i+1; $j <= $#$pC; $j++) {
	    my $sj = $strains[$j];
	    if ($si lt $sj) {
		$snps{$si}{$sj} += $pC->[$j]; 
	    } else {
		$snps{$sj}{$si} += $pC->[$j]; 
	    }
	}
    }
    my $gh = openToWrite($g);
    print $gh "N_SITES $nSites N_SNPS $nSNPs\n";
    foreach my $si (sort keys %snps) {
	my $pH = $snps{$si};
	foreach my $sj (sort keys %$pH) {
	    printf($gh "%s %s %d\n",$si,$sj,$pH->{$sj});
	}
    }
    $gh->close();
}

##
## ---------------------------------------------------------------------
## sgrp.pl esnr cere methc CERE
## ---------------------------------------------------------------------
##
## extractSNPRates
##    SRC: p?/chr*/ssnp_counts.txt [from esnp]
##    TGT: ssnp_rates.txt          [for things like
##                                  NJ trees and heat plots; pmar uses
##                                  chr-specific file].

sub extractSNPRates {
    unless (@_ == 3) {
	confess("Usage: $0 esnr species filtMeth strainSet");
    }
    my ($species,$filtMeth,$strainSet) = @_;
    my ($iter,$dir) = getCurrentIteration($species,$filtMeth,$strainSet,'p?');
    my $tgt = "$dir/ssnp_rates.txt";
    extractSNPRates1($tgt,glob("$dir/p[0-9]/*"));
}

##
## ---------------------------------------------------------------------
## sgrp.pl isnp cere methc CERE 0 1 2 ...
## ---------------------------------------------------------------------
##
## extractSNPsPostImp
##   -> extractSNPsPostImp1
##        SRC: .../chrNN/bl_imputed.gz [from impu]
##        -> reportSNPs
##             TGT: .../chrNN/isnp_report.gz [for esnr]
##        -> countImputedSNPs
##             TGT: .../chrNN/isnp_counts.txt [for esnr]

sub extractSNPsPostImp {
    my ($species,$filtMeth,$strainSet,@parts) = @_;
    unless (@parts) {
	confess("Usage: $0 isnp species filtMeth strainSet part1 part2 ...");
    }
    my ($iter,$dir) = getCurrentIteration($species,$filtMeth,$strainSet,'p?');
    foreach my $part (@parts) {
	$part = "p$part" if ($part =~ /^[0-9]$/);
	foreach my $chrDir (glob("$dir/$part/*")) {
	    extractSNPsPostImp1($chrDir);
	}
    }
}

sub extractSNPsPostImp1 {
    my ($dir) = @_;
    my $imp;
    foreach my $suf (qw(z y m)) {
	$imp = "$dir/imputed_$suf.gz";
	last if (-s $imp);
    }
    my $isnpRepFile = "$dir/isnp_report.gz";
    my $isnpCountFile = "$dir/isnp_counts.txt";
    if (-s $imp) {
	if (stamp($isnpRepFile)) {
	    report("creating $isnpRepFile");
	    reportSNPs($imp,undef,$isnpRepFile);
	    report();
	    unstamp($isnpRepFile);
	}
	if (stamp($isnpCountFile)) {
	    report("creating $isnpCountFile");
	    countImputedSNPs($imp,$isnpCountFile);
	    report();
	    unstamp($isnpCountFile);
	}
    }
}

1;
