package SGRP::Filter;

use Exporter;
@ISA = qw(Exporter);

@EXPORT = qw(readSizes applySelectionFile);

sub readSizes {
    my ($sp) = @_;
    my $szFile = "$REF/$sp/genome.sz";
    createSizeFile("$REF/$sp/genome.fa",$szFile) unless (-s $szFile);
    my %sz;
    my $fh = openToRead($szFile);
    while (<$fh>) {
	chomp;
	my ($chr,$n) = split;
	$sz{$chr} = $n;
    }
    return \%sz;
}

sub createSizeFile {
    my ($f,$g) = @_;
    my $fh = openToRead($f);
    my $gh = openToWrite($g);
    my ($chr,$sz,$sum);
    while (<$fh>) {
	chomp;
	if (/^>(\S+)/) {
	    if (defined $chr) {
		print $gh "$chr $sz\n";
		$sum += $sz unless ($chr =~ /^vc_/);
	    }
	    $chr = $1;
	    $sz = 0;
	} else {
	    $sz += length($_);
	}
    }
    if (defined $chr) {
	print $gh "$chr $sz\n";
	$sum += $sz unless ($chr =~ /^vc_/);
	print $gh "ALL $sum\n";
    }
}

sub applySelectionFile {
    my ($src,$pOM,$filt,$zsFile,$tgt,$introZS) = @_;
    my (%survivor,%alis,%trim);
    unless (-s $filt) {
	# Save time, and prevent wrong behaviour ...
	my $gh = openToWrite($tgt);
	$gh->close();
	return;
    }
    my $fh = openToRead($filt); # e.g. alilines.txt
    my @zeroScorers;
    my $aliLine;
    while (<$fh>) {
	if (/^ALIGNMENT/) {
	    # Remove and keep any "trim" annotation
	    chomp;
	    my ($aliLine,$tr) = split(/\|/,$_);
	    $aliLine .= "\n"; # restore the newline
	    $trim{$aliLine} = $tr if ($tr);
	    my @a = split(" ",$aliLine);
	    if ($a[1] > 0) {
		#$survivor{join(" ",@a)} = 1;
		$survivor{$aliLine} = 1;
	    } elsif ($introZS) {
		push @zeroScorers,$aliLine;
	    }
	}
    }
    getZeroScorerSequences(\@zeroScorers,$pOM,$zsFile,\%alis);
    my $fh = openToRead($src);
    my ($chr,$lo,@buf);
    while (<$fh>) {
	if (/^ALIGNMENT/ || /^Process/ || /^====/ || /^\#\# /) {
	    if (defined $chr && defined $lo) {
		if (isSurvivor(\@buf,\%survivor)) {
		    my @buf2 = @buf; # copy
		    push @{$alis{$chr}{$lo}},\@buf2;
		}
		#$nPushed ++;
		@buf = ();
		$chr = $lo = undef;
	    }
	    if (/^ALIGNMENT/) {
		my $pA = new Alignment($_,$pOM,1);
		$chr = $pA->chr(1);
		# See MergeMatches::mungeLine too for the same ordering...
		$lo = $pA->clo(1);
		$lo || die("Cannot find clo(1) from $_");
		# Adjust for trim, which will only be present in
		# alilines, not $fh:
		if (defined $trim{$_}) {
		    my ($tLo) = split(/,/,$trim{$_});
		    $lo += $tLo;
		}
	    } elsif (/^\#\# (\S+):/) {
		$chr = $1; # shouldn't happen?
	    }
	}
	if (defined $chr) {
	    push @buf,$_;
	}
    }
    if (defined $chr) {
	if (isSurvivor(\@buf,\%survivor)) {
	    push @{$alis{$chr}{$lo}},\@buf;
	}
    }
    foreach my $line (keys %survivor) {
	print STDERR "WARNING: did not find $line";
    }
    die("Unused alignment lines") if (keys %survivor);
    my $gh = openToWrite($tgt);
    foreach my $chr (sort keys %alis) {
	my $pH = $alis{$chr};
	foreach my $lo (sort {$a <=> $b} keys %$pH) {
	    my $pA = $pH->{$lo};
	    foreach my $pAli (@$pA) {
		my ($head,@body) = @$pAli;
		if (defined $trim{$head}) {
		    my $tr = $trim{$head};
		    chomp $head;
		    print $gh "$head|$tr\n";
		} else {
		    print $gh $head;
		}
		foreach my $line (@body) {
		    print $gh $line;
		}
	    }
	}
    }
    $gh->close();
}

sub getZeroScorerSequences {
    my ($pZeroScorers,$pOM,$zsFile,$pAlis) = @_;
    my %wanted;
    foreach my $line (@$pZeroScorers) {
	my @arr = split(" ",$line); # OK - we don't need a full Alignment object here
	my ($rlo,$rhi) = ($arr[4],$arr[5]);
	# We want the sequence (from the read) for this alignment
	# if it covers a non-empty piece of the read
	my $want = $arr[8] eq 'F' ? $rhi >= $rlo : $rlo >= $rhi;
	$wanted{$arr[2]} = 1 if ($want);
    }
    # Fill sequence of those reads for which we want it
    my %sequence;
    my $fh = openToRead($zsFile); # renamed.fastq file
    while (<$fh>) {
	chomp;
	if (substr($_,0,1) eq '@') {
	    s/ .*//;
	    my $name = substr($_,1);
	    if ($wanted{$name}) {
		chomp ($sequence{$name} = <$fh>);
	    }
	}
    }
    # Set $pAlis for those zero-scoring alignments for which we collected a sequence
    foreach my $line (@$pZeroScorers) {
	my $pA = new Alignment($line,$pOM,1);
	my $lo = $pA->clo(1);
	if ($pA->chi(1) < $lo) {
	    $lo -= 0.5; # to ensure inserts come early
	}
    	push @{$pAlis->{$pA->chr(1)}{$lo}},[$line,constructZSAlignment($pA,$sequence{$pA->name})];
    }
}

sub constructZSAlignment {
    my ($pA,$fullSeq) = @_;
    my @lines = ("\n");
    if ($fullSeq) {
	my $seq = substr($fullSeq,$pA->rlo-1,$pA->rhi-$pA->rlo+1);
	my ($off,$dirn);
	if ($pA->fc eq 'F') {
	    $off = $pA->rlo;
	    $dirn = 1;
	} else {
	    $seq = reverse($seq);
	    $seq =~ tr/ACGT/TGCA/;
	    $off = $pA->rhi;
	    $dirn = -1;
	}
	while ($seq) {
	    my $head;
	    if (length($seq) > 60) {
		$head = substr($seq,0,60);
		$seq = substr($seq,60);
	    } else {
		$head = $seq;
		$seq = "";
	    }
	    # Simulate Ssaha output ...
	    my $len = length($head);
	    push @lines,sprintf("Query %10d %s %d\n",$off,$head,$off+($len-1)*$dirn);
	    $off += $len*$dirn;
	    push @lines,sprintf("%16s %s\n",' ','-' x $len);
	    push @lines,sprintf("Sbjct %10d %s %d\n",0,'-' x $len,0);
	    push @lines,"\n";
	}
    }
    return @lines;
}
    

sub isSurvivor {
    my ($pBuf,$pSurv) = @_;
    return 1 unless (defined %$pSurv); # none defined -> use all
    foreach my $line (@$pBuf) {
	#my @a = split(" ",$line);
	#my $normLine = join(" ",@a);
	if (defined $pSurv->{$line}) {
	    delete $pSurv->{$line}; # prevent it being used again!
	    return 1;
	}
    }
    return 0;
}

sub alignmentIsPrior {
    my ($pAli,$chr,$lo) = @_;
    my $d = $pAli->[3] cmp $chr || $pAli->[6] <=> $lo;
    return $d <= 0;
}    

# If there are 10 names and $th is 0.5, $nth is 4,
# so the partitioning is le name versus gt name.

sub findPropName {
    my ($th,$pRC) = @_;
    my @keys = sort keys %$pRC;
    my $nth = int(-0.5+$th*scalar(@keys));
    $nth = 0 if ($nth < 0);
    return $keys[$nth];
}    

use SGRP::Variables;
use SGRP::Alignment;
use SGRP::SGRP;

use Utility;
use File::Basename;

use Carp;

use strict;

# pParams should be a hash with:
#   matchEnds: 0: do not look for matching ends, just print top scoring match(es)
#                 for each read independently
#              1: must match ends when both p and q present, else do as for 0
#              2: reject all when either p or q is missing
#              Default: 0
#   diff:      shortfall allowed from max [combined] score for read to go through;
#              negative means discard all when tie for top score
#              Default: 0
#   minIL:     min insert length required for ends to match (for matchEnds>0)
#              Default: 1500
#   maxIL:     max insert length required for ends to match (ditto)
#              Default: 15000
#   maxCov:    coverage multiplier at which to stop reading
#              Default: infinite
#   probUse:   probability that we use any given read. Default: 1.
#   maxCount:  max no. of alignments for any one read. Default: unlimited.
#   useAbove:  if e.g. 0.7, use the 30% of reads with alphabetically latest names
#              (this is done per strain, so should be arbitrary).
#   useBelow:  ditto but 70% of earliest.

sub makeSelectionFile {
    my ($species,$src,$tgt,$pParams,$pTE) = @_;
    my $fh = openToRead($src);
    my $maxCov = getHash($pParams,'maxCov');
    my $maxNucs;
    if (defined $maxCov) {
	my $pSz = readSizes($species);
	$maxNucs = $pSz->{ALL} * $maxCov;
    }
    my $probUse = getHash($pParams,'probUse');
    my $maxCount = getHash($pParams,'maxCount');
    my $useAbove = getHash($pParams,'useAbove');
    my $useBelow = getHash($pParams,'useBelow');
    my @aliLines = <$fh>; # all alilines in file
    my %readCount;
    foreach my $line (@aliLines) {
	my (undef,undef,$name) = split(/\s+/,$line);
	$readCount{$name} ++;
    }
    if (defined $maxCount) {
	foreach my $name (keys %readCount) {
	    if ($readCount{$name} > $maxCount) {
		delete $readCount{$name};
	    }
	}
    }
    my $maxName = findPropName($useBelow,\%readCount) if (defined $useBelow);
    my $minName = findPropName($useAbove,\%readCount) if (defined $useAbove);
    my %lines;
    foreach my $line (@aliLines) {
	next if (defined $probUse && rand() > $probUse);
	my (undef,$score,$name,$chr,$lo,$hi,$clo,$chi) = split(/\s+/,$line);
	next if (defined $maxCount && !defined $readCount{$name});
	next if (segmentFails($minName,$maxName,$name));
	next if (badSutureMatch($chr,$clo,$chi));
	# Only allow maxNucs bases, by decreasing it:
	if (defined $maxNucs) {
	    $maxNucs += $lo-$hi-1;
	    last if ($maxNucs < 0);
	}
	# $bit is 0 for p, 1 for q
	my ($ins,$bit) = parseName($name);
	push @{$lines{$ins}[$bit]{$score}},$line;
    }
    my $gh = openToWrite($tgt);
    foreach my $ins (sort keys %lines) {
	printWinners($gh,$pParams,$pTE,@{$lines{$ins}});
    }
    $gh->close();
}

sub segmentFails {
    my ($minName,$maxName,$name) = @_;
    return 1 if (defined $minName && $name le $minName);
    return 1 if (defined $maxName && $name gt $maxName);
    return 0;
}

# lo and hi in this file are offset-one, as are the chromosomal
# coordinates in the ALIGNMENT lines, so we leave things as they
# are.
sub readTranselFile {
    my ($f) = @_;
    my $fh = openToRead($f);
    my %te;
    while (<$fh>) {
	chomp;
	next if (/^\#/);
	my ($chr,$lo,$hi,$name) = split;
	push @{$te{$chr}},[$lo,$hi,$name];
    }
    return \%te;
}

1;
