package SGRP::Imputer;

use Exporter;
@ISA = qw(Exporter);

@EXPORT = qw(runImputer runStaticImputer runImputerForHardGaps);

use SGRP::Variables;
use SGRP::PrepareImputer;
use SGRP::Margarita; # for markerCount
use Utility;
use File::Basename;

use Carp;

use strict;

my @PE;

# If set, don't output lines consisting only of gaps when creating imputed_y.gz.
my $IGNORE_GAP_LINES = 1; # 0;

# Normally 1, but need 0 to train quality estimations

my $WIPE_UNLIKELY_INDELS = 1;

my $ZERO = ord('!');

# Calibrated from comparing S288c and REF, and requiring roughly
# equal numbers of gaps in each in imputed_y.gz over all autosomes.
my $BIAS_TO_ABSENCE = 2;

sub runImputer {
    runImputer1('impu',@_);
}

# cmd will be impu (MF method), impx (for hard gaps) or stat (static).

sub runImputer1 {
    my ($cmd,$species,$filtMeth,$strainSet,@parts) = @_;
    unless (@parts) {
	confess("Usage: $0 $cmd species filtMeth strainSet part1 part2 ...");
    }
    my ($iter,$dir) = getCurrentIteration($species,$filtMeth,$strainSet,'p?');
    my $starBranchLength = ($cmd eq 'stat') ? getStarBranchLength($dir) : undef;
    foreach my $part (@parts) {
	$part = "p$part" if ($part =~ /^[0-9]$/);
	# C++ or D (branch length) version:
	if ($cmd eq 'impu' || $cmd eq 'impx') {
	    my $suf = $cmd eq 'impu' ? 'm' : 'x';
	    # Prepare first...
	    if ($suf eq 'x') {
		foreach my $af (filesFromSmallest(glob("$dir/$part/*/args_bl"))) {
		    unless (-f "$af.touch") {
			my $d = dirname($af);
			prepareImputer1($suf,$d);
		    }
		}
	    }
	    foreach my $af (filesFromSmallest(glob("$dir/$part/*/args_bl"))) {
		unless (-f "$af.touch") {
		    my $d = dirname($af);
		    my $sufToUse = $suf;
		    my $tgt = "$dir/imputed_$sufToUse.gz";
		    unless (-s $tgt) {
			if ($suf eq 'm') {
			    $sufToUse = (markerCount("$d/mgr_in") == 0) ? 'x' : $suf;
			    prepareImputer1($sufToUse,$d) if ($sufToUse eq 'x');
			}
		    }
		    runImputer2($sufToUse,$d);
		}
	    }
	} else {
	    my $blFile = "$dir/snpMatrix/sorted.bl";
	    if (-s $blFile && ! -f "$blFile.touch") {
		foreach my $d (glob("$dir/$part/*")) {
		    if (-s "$d/sorted.gz") {
			foreach my $suf (@STATIC_IMPUTATION_SUFS) {
			    unless (-s "$d/imputed_$suf.gz") {
				prepareImputer1($suf,$d,$blFile);
				runImputer2($suf,$d,$starBranchLength);
			    }
			}
		    } else {
			report("$d/sorted.gz: missing or empty",1);
		    }
		}
	    } else {
		report("$blFile: missing or incomplete",1);
	    }
	}
    }
}

sub getStarBranchLength {
    my ($dir) = @_;
    my $fh = openToRead("$dir/snpMatrix/sorted.q40");
    my @lines = <$fh>;
    my $dim = scalar(@lines);
    my $sum = 0;
    foreach my $i (0 .. $#lines-1) {
	my $line = $lines[$i];
	chomp $line;
	my @a = split(" ",$line);
	foreach my $v (@a[$i+1 .. $#a]) {
	    $sum += $v;
	}
    }
    # We don't halve the result because distance between any two nodes
    # is twice the branch length, but we only counted 1/2 dim (dim-1)
    # distances, so the 2 and the 1/2 cancel.
    return $sum/($dim*($dim-1));
}

# Directory and file structure, for imputation suffix S in
# @STATIC_IMPUTATION_SUFS
#
# Inputs:
#
#   impS/iNNN/imp_in            -- a piece of sorted.gz; same for all methods so no _S
#             args_bl
#             mgr_in
#
# Outputs:
#
#   impS/iNNN/imputed         -- imputed version of imp_in            
#   impS/iNNN/imputed.touch   -- touch file
#
# $dir is imp/iNNN
# $starBranchLength is only needed for $suf eq 'd'

sub runImputer2 {
    my ($suf,$dir,$starBranchLength) = @_;
    # x is for dual (hard gaps)
    my $impDir = "$dir/imp" . ($suf eq 'x' ? 'm' : $suf);
    if (! -d $impDir) {
	report("missing or empty: $impDir",1);
	return;
    }
    # Overall target file:
    my $tgt = "$dir/imputed_$suf.gz";
    if (-s $tgt && !($suf eq 'x' && ! -s "$dir/imputed_y.gz")) {
	report("$tgt: already exists",1);
	return;
    }
    my $bin = "$HOME/src/impute/d/impute";
    unless (-x $bin) {
	confess("Missing or not executable: $bin");
    }
    my $outPfx = $suf eq 'x' ? 'dual' : 'imputed';
    foreach my $subD (glob("$impDir/i[0-9][0-9][0-9]")) {
	my $subT = "$subD/$outPfx";
	if (stamp($subT)) {
	    report("creating $subT");
	    if ($suf eq 'd') {
		runDumbImputer($subD,$starBranchLength);
	    } else {
		# Max arg index is non-zero only for Margarita:
		my $acMax = $suf eq 'm' || $suf eq 'x' ? $ARG_COUNT-1 : 0;
		my $inPfx = $suf eq 'x' ? 'dual' : 'imp';
		if (-s "$subD/${inPfx}_in") {
		    my $cmd = "cd $subD; $bin -l 0 -h $acMax -L $outPfx.log -a ${inPfx}_in -o $outPfx mgr_in args_bl >$outPfx.out 2>$outPfx.err";
		    # Try it three times, give up if stays empty ...
		    foreach my $k (0 .. 2) {
			if ($k) {
			    print STDERR "RETRYING: $cmd\n";
			}
			`$cmd`;
			if (-s $subT) {
			    last;
			} else {
			    print STDERR "FAILED: $cmd\n";
			    sleep(10); # pause to see if nasty condition goes away
			}		
		    }
		}
	    }
	    report();
	    unstamp($subT,1);
	    unlink($subT) unless (-s $subT); # another job might manage it...
	}
    }
    # Try and do final copy
    tryFinalCopy($dir,$impDir,$outPfx,$suf,$tgt);
}

sub runDumbImputer {
    my ($subD,$bl) = @_;
    setPEArray(); # to save time
    my $fh = openToRead("$subD/imp_in");
    my $gh = openToWrite("$subD/imputed");
    my $hdr = <$fh>;
    print $gh $hdr;
    chomp $hdr;
    my $ns = scalar(split(" ",$hdr))+1;
    while (<$fh>) {
	chomp;
	my $line = runDumbImputer1($_,$bl,$ns);
	print $gh "$line\n";
    }
}

sub setPEArray {
    unless (@PE) {
	my $k = -log(10)/10;
	for (my $q=0; $q<=94; $q++) {
	    $PE[$q] = exp($k*$q);
	}
    }
}

sub runDumbImputer1 {
    my ($line,$bl,$ns) = @_;
    my $nExtra = $ns*2 - length($line);
    if ($nExtra > 0) {
	$line .= ' ' x $nExtra;
    }
    my $iMax = $ns-1;
    my (@vals,@quals,%isVal);
    for (my $i=0; $i<=$iMax; $i++) {
	my $ii = $i*2;
	my $v = substr($line,$ii,1);
	unless ($v eq ' ') {
	    $v =~ s/[_=]/-/;
	    $v =~ tr/acgt/ACGT/;
	    my $qc = substr($line,$ii+1,1);
	    my $q = ($qc eq ' ') ? 94 : (ord($qc)-33);
	    $vals[$i] = $v;
	    $quals[$i] = $q;
	    $isVal{$v} = 1;
	}
    }
    my @vSet = keys %isVal;
    my $n = scalar(@vSet);
    if ($n == 0) {
	# All spaces: bogus location, from all hard gaps,
	# so value doesn't matter; return unchanged.
	return $line;
    } elsif ($n == 1) {
	# apparently monomorphic: easy and quick
	return ($vSet[0] . ' ') x $ns;
    }
    my $pHub = hubProbabilityDistribution(\@vals,\@quals,\@vSet,$bl,$n);
    my ($ans,$fill);
    foreach my $i (0 .. $iMax) {
	my $v = $vals[$i];
	if (defined $v) {
	    $ans .= dumbImputedPair1($v,$PE[$quals[$i]],$pHub,$n,$bl);
	} else {
	    if (!defined $fill) {
		$fill = dumbImputedPair2($pHub,$n,$bl);
	    }
	    $ans .= $fill;
	}
    }
    return $ans;
}

sub hubProbabilityDistribution {
    my ($pVals,$pQuals,$pVSet,$bl,$n) = @_;
    my (@pu,@lor);
    foreach my $i (0 .. $#$pQuals) {
	if (defined $pQuals->[$i]) {
	    $pu[$i] = min(1,$PE[$pQuals->[$i]] + $bl) * ($n-1)/$n;
	    $lor[$i] = log(1/$pu[$i]-1);
	}
    }
    my %cost;
    foreach my $i (0 .. $#lor) {
	if (defined $lor[$i]) {
	    my $v = $pVals->[$i];
	    foreach my $h (@$pVSet) {
		unless ($h eq $v) {
		    $cost{$h} += $lor[$i];
		}
	    }
	}
    }
    return costsToProbs(\%cost);
}
    

sub costsToProbs {
    my ($pCost) = @_;
    my $min;
    foreach my $cost (values %$pCost) {
	$min = min($min,$cost);
    }
    my $sumProbs;
    foreach my $cost (values %$pCost) {
	$sumProbs += exp($min-$cost);
    }
    if ($sumProbs == 0) {
	return undef;
    } else {
	my $norm = 1/$sumProbs;
	my %prob;
	foreach my $key (keys %$pCost) {
	    $prob{$key} = exp($min-$pCost->{$key})*$norm;
	}
	return \%prob;
    }
}

sub dumbImputedPair1 {
    my ($v,$pe,$pHub,$n,$bl) = @_;
    my %prob;
    my @set = keys %$pHub;
    my $tot;
    foreach my $w (@set) {
	my $pvw = ($v eq $w) ? 1-$pe : $pe/($n-1);
	my $sum;
	foreach my $h (@set) {
	    my $pwh = ($w eq $h) ? 1-$bl : $bl/($n-1);
	    my $ph = $pHub->{$h};
	    $sum += $pvw*$pwh*$ph;
	}
	$prob{$w} = $sum;
	$tot += $sum;
    }
    @set = sort {$prob{$b} <=> $prob{$a}} @set;
    my $winner = $set[0];
    my $pePost = 1-$prob{$winner}/$tot;
    my $qPost = $pePost > 0 ? -10*log($pePost)/log(10) : 95;
    if ($winner ne $v) {
	$winner =~ tr/ACGT/acgt/;
    }
    return $winner . ($qPost > 94 ? ' ' : chr($qPost+33));
}

sub dumbImputedPair2 {
    my ($pHub,$n,$bl) = @_;
    my %prob;
    my @set = keys %$pHub;
    my $tot;
    foreach my $w (@set) {
	my $sum;
	foreach my $h (@set) {
	    my $pwh = ($w eq $h) ? 1-$bl : $bl/($n-1);
	    my $ph = $pHub->{$h};
	    $sum += $pwh*$ph;
	}
	$prob{$w} = $sum;
	$tot += $sum;
    }
    @set = sort {$prob{$b} <=> $prob{$a}} @set;
    my $winner = $set[0];
    my $pePost = 1-$prob{$winner}/$tot;
    my $qPost = -10*log($pePost)/log(10);
    $winner =~ tr/ACGT/acgt/;
    return $winner . ($qPost > 94 ? ' ' : chr($qPost+33));
}

sub tryFinalCopy {
    my ($dir,$impDir,$base,$suf,$tgt) = @_;
    my @subs;
    foreach my $subD (glob("$impDir/i[0-9][0-9][0-9]")) {
	my $subT = "$subD/$base";
	unless (-s $subT && ! -f "$subT.touch") {
	    report("returning because $subT not ready",1);
	    return;
	}
	push @subs,$subT;
    }
    if (stamp($tgt)) {
	report("creating $tgt");
	if ($suf eq 'x') {
	    copyDeletingNonInitialHeaders("$dir/sorted_ac_regions.txt",$tgt,@subs);
	} else {
	    copyPreservingHardGaps("$dir/sorted.gz",$tgt,@subs);
	    #unlink($impin);
	}
	unstamp($tgt);
	report();
    } else {
	report("$tgt already (being) created",1);
    }
    if ($suf eq 'x') {
	joinImputedMAndXToY($dir);
	createDanglesFromLayout("$dir/layout_y.gz","$dir/dangles_y.txt");
    }
}

# As well as deleting non-initial headers, we also make sure presence/absence
# is averaged across the whole of each region with unchanging hard gaps.

sub copyDeletingNonInitialHeaders {
    my ($regFile,$tgt,@inFiles) = @_;
    my $pRegs = readRegions($regFile);
    my $gh = openToWrite($tgt);
    my $isInit = 1;
    my $off = 0;
    my @buf;
    foreach my $f (@inFiles) {
	my $fh = openToRead($f);
	my $hdr = <$fh>;
	if ($isInit) {
	    print $gh $hdr;
	    $isInit = 0;
	}
	while (<$fh>) {
	    chomp;
	    push @buf,$_;
	    $off ++;
	    if ($off == $pRegs->[0]) {
		averagePresence(\@buf); # the key operation
		foreach my $line (@buf) {
		    print $gh "$line\n";
		}
		@buf = ();
		shift @$pRegs;
	    }
	}
    }
    averagePresence(\@buf); # the key operation
    foreach my $line (@buf) {
	print $gh "$line\n";
    }
    if (@$pRegs) {
	die("Still some pRegs left: @{$pRegs->[0]}");
    }	
}

sub readRegions {
    my ($f) = @_;
    my $fh = openToRead($f);
    my @regs;
    while (<$fh>) {
	chomp;
	my (undef,$hi) = split;
	push @regs,$hi; 
    }
    return \@regs;
}

sub averagePresence {
    my ($pBuf) = @_;
    my $len = length($pBuf->[0]);
    for (my $i=0; $i<$len; $i+=2) {
	my (%sum,$n);
	foreach my $line (@$pBuf) {
	    my $v = substr($line,$i,1);
	    $v =~ tr/ac/AC/;
	    if ($v eq 'A' || $v eq 'C') {
		my $qc = substr($line,$i+1,1);
		$sum{$v} += $qc eq ' ' ? 94 : (ord($qc)-33);
	    } elsif ($v ne 'T' && $v ne 't') {
		die("Bad value $v at col $i of $line");
	    }
	}
	my $diff = $sum{C} - $sum{A};
	my $count = scalar(@$pBuf);
	my $val;
	if ($diff > 0) {
	    $val = 'C';
	} else {
	    $val = 'A';
	    $diff = -$diff;
	}
	my $q = int(0.5+$diff/$count);
	my $qc = $q > 93 ? ' ' : chr($q+33);
	my $pair = "$val$qc";
	foreach my $line (@$pBuf) {
	    my $v = substr($line,$i,1);
	    unless ($v eq 'T' || $v eq 't') {
		substr($line,$i,2) = $pair;
	    }
	}
    }
}
   
sub copyPreservingHardGaps {
    my ($sorted,$tgt,@inFiles) = @_;
    my $sh = openToRead($sorted);
    my $gh = openToWrite($tgt);
    <$sh>;
    my $bh = openToRead(shift @inFiles);
    my $hdr = <$bh>;
    print $gh $hdr;
    while (1) {
	while (<$bh>) {
	    chomp;
	    chomp (my $sLine = <$sh>);
	    my $outLine = copyPreservingHardGaps1($_,$sLine) . "\n";
	    print $gh $outLine;
	}
	last unless (@inFiles);
	$bh = openToRead(shift @inFiles);
	<$bh>; # discard header this time
    }
}

# Where we have '=' in sLine, copy it into line, overwriting whatever
# was imputed there. We also change digits (representing distinguished
# soft gaps) to - or _ depending on whether they are present in sorted.
# And, we replace zeroth-column lower-case values by hard gaps; and when we had
# only such a value, we copy it to all other positions (overriding
# what the imputer would have done on the basis of garbage input).
# Thus column zero of the imputed file now represents the REFERENCE
# sequence, not the extended backbone (which is copied across to the strains
# where it may apply).

sub copyPreservingHardGaps1 {
    my ($line,$sLine) = @_;
    my $len = length($sLine);
    my $len2 = length($line);
    if ($sLine =~ /^[a-z]. *$/) {
	# repeat hard gap for ref, then e.g. "a!" across all strains
	$line = '= ' . substr($sLine,0,2) x ($len2/2-1);
    } else {
	for (my $i=0; $i<$len; $i+=2) {
	    if (substr($sLine,$i,1) eq '=') {
		substr($line,$i,2) = substr($sLine,$i,2);
	    } elsif ($i == 0 && substr($sLine,$i+1,1) eq '!') {
		substr($line,$i,2) = '= ';
	    }
	}
	for (my $i=0; $i<$len2; $i+=2) {
	    if (substr($line,$i,1) =~ /[0-9]/) {
		substr($line,$i,1) = ($i >= $len || substr($sLine,$i,1) eq ' ') ? '_' : '-';
	    }
	}
    }
    # Ensure ref nucleotides, including N's, are unchanged; and ensure
    # no nucleotides, even lower case, are sneaked in.
    if ($sLine =~ /^[A-Z]/ || ($sLine =~ /^[0-9]/ && $line =~ /^[A-Za-z]/)) {
	substr($line,0,2) = substr($sLine,0,2);
    }
    return $line;
}

sub joinImputedMAndXToY {
    my ($d) = @_;
    my $fm = "$d/imputed_m.gz";
    my $fx = "$d/imputed_x.gz";
    my $g = "$d/imputed_y.gz";
    my $sf = "$d/sorted.gz";
    my $lf = "$d/layout.gz";
    my $sg = "$d/sorted_y.gz";
    # NOTE: layout_y.gz does not reproduce all the reference nucleotides.
    # It's really only intended for finding the ali coordinates
    # at which alignments start and finish (and it should be accurate
    # for those).
    my $lg = "$d/layout_y.gz";
    report("$g etc: creating");
    foreach my $f ($fm,$fx,$sf) { # don't check $lf because of Solexa incorporation round
	unless (-s $f && ! -f "$f.touch") {
	    report("$f: missing, empty or incomplete",1);
	    return;
	}
    }
    unless (stamp($g)) {
	report("$g: cannot stamp",1);
	return;
    }
    my $pWipes = $WIPE_UNLIKELY_INDELS ? decideIndelWipes($sf) : undef;
    my $mh = openToRead($fm); # imputed_m.gz
    my $xh = openToRead($fx); # imputed_x.gz
    my $sh = openToRead($sf); # sorted.gz
    my $lh = -s $lf ? openToRead($lf) : undef; # layout.gz
    my $gh; # imputed_y.gz
    my $mHdr = <$mh>;
    my $xHdr = <$xh>;
    if ($mHdr eq $xHdr) {
	$gh = openToWrite($g);
	print $gh $mHdr;
    } else {
	report("headers differ in $d/imputed_{m,x}.gz",1);
	return;
    }
    my $hh = openToWrite($sg); # sorted_y.gz
    my $hdr = <$sh>;
    print $hh $hdr;
    my $ah = openToWrite($lg); # layout_y.gz
    my $pMap = defined $lh ? readSorted12Map("$d/sorted12.map") : undef;
    joinImputedMAndXToY1($mh,$xh,$sh,$lh,$pWipes,$pMap,$gh,$hh,$ah);
    unstamp($g);
    report();
}

# Run on sorted file.

# Calculate s = sum of qual scores for variant (gap or nucleotide,
# whichever has smaller sum) divided by total no. of strains
# sequenced at this location. Reset quality of each strain with
# the variant as follows, depending on indel length:
#
# Len=1 => q = 2.13354 s - 3.18592
# Len=2 => q = 2.68949 s - 0.43062
# Len>2 => q = 2.68067 s (we force const to zero here)
#
# Curves are plotted in indelEvidencePost.pl.
#
# When q >= 3.01 we keep the value but take min of existing
# quality and 'q' value specified here. 
# However, q < 3.01 means p(error) > 0.5, so in that case,
# we switch over to the other value, with new quality q' that
# is the complement of the old:
#
# q  = -10 log_10 p = -10 ln p / ln 10
# q' = -10 log_10 (1-p) = -10 ln (1-p) / ln 10
# where
# p = exp(-(0.1 ln 10) q)

sub decideIndelWipes {
    my ($sf) = @_;
    my $pGNScores = gapAndNucleotideScores($sf);
    # Go through all the gnScores deciding what we actually will wipe. @wipe is
    # a version of @gnScores cleaned up to allow regions of >1 offset.
    my @wipe;
    my $i=0;
    while ($i <= $#$pGNScores) {
	my ($type,$len,$bCount,$maxScore) = gapAndNucleotideRegionScore($i,$pGNScores);
	# Now we've found a region satisfying the conditions.
	my $lb = $len-$bCount; # effective length: length minus all-blank count
	if ($lb > 0) {
	    # Effective quality for keeping the minority allele ('G' or 'N').
	    my $qEff = max(0,effectiveQuality($lb,$maxScore));
	    # Remember if p(should wipe) >= 0.01.
	    if ($qEff < 20) {
		foreach my $k ($i .. $i+$len-1) {
		    # Skip over blank sites...no need to wipe them
		    unless ($pGNScores->[$k][1] eq 'B') {
			push @wipe,[$pGNScores->[$k][0],$type,$qEff];
		    }
		}
	    }
	}
	$i += $len;
    }
    return \@wipe;
}

sub gapAndNucleotideScores {
    my ($sf) = @_;
    my $sh = openToRead($sf);
    <$sh>; # discard header
    my $off = 0; # offset into file: header line is zero
    # Evidence for wiping at each offset ...
    my @gnScores;
    while (<$sh>) {
	$off ++;
	chomp;
	# Sum of (non-hard) gap quality scores and nucleotide quality scores,
	# and number of strains contributing
	my ($gs,$ns,$c) = gapAndNucleotideQualitySums($_);
	if (defined $gs) { # then it's an interesting line, even if $ns is 0
	    # Evidence for wiping at this offset...
	    push @gnScores,indelWipeSummary($off,$gs,$ns,$c);
	}
    }
    return \@gnScores;
}

sub gapAndNucleotideQualitySums {
    my ($line) = @_;
    my $iMax = length($line)-2;
    my ($gs,$ns,$c);
    for (my $i=0; $i<=$iMax; $i+=2) {
	my $v = substr($line,$i,1);
	unless ($v eq ' ' || $v eq '=') {
	    my $qc = substr($line,$i+1,1);
	    my $q = $qc eq ' ' ? 94 : ord($qc)-33;
	    if ($v > 0) { # gaps are represented as numbers in sorted.gz
		$gs += $q;
	    } else {
		$ns += $q;
	    }
	    $c ++;
	}
    }
    return ($gs,$ns,$c);
}

sub indelWipeSummary {
    my ($off,$gs,$ns,$c) = @_;
    if (!defined $ns) {
	return [$off,'B']; # 'blank' line, i.e. gaps only
    } elsif ($gs < $ns) {
	return [$off,'G',$gs/$c]; # gaps are weaker
    } else {
	return [$off,'N',$ns/$c]; # nucs are weaker
    }
}

sub gapAndNucleotideRegionScore {
    my ($i,$pGNScores) = @_;
    # B (blank), G (wipe gap) or N (wipe nuc); G and N will override B.
    my $type = $pGNScores->[$i][1];
    my $len = 1; # length of the region to wipe in
    my $bCount = $type eq 'B' ? 1 : 0;
    # Max of any of the gnScores in the region...
    my $maxScore = $pGNScores->[$i][2];
    # Find max $len such that gnScores indices $i to $i+$len-1 inclusive are in same region,
    # and types are all G or all N, allowing B as honorary G or N ...
    my $j; # assign to $i+$len:
    while ($j = $i+$len &&
	   $j <= $#$pGNScores && 
	   $pGNScores->[$j][0] == $pGNScores->[$i][0]+$len) {
	# if above succeeded, $i to $j inclusive are contiguous
	my $typeHere = $pGNScores->[$j][1];
	# Don't include $i+$len in the region if the type is incompatible.
	last if (($type eq 'G' && $typeHere eq 'N') || ($type eq 'N' && $typeHere eq 'G'));
	$maxScore = max($maxScore,$pGNScores->[$j][2]);
	if ($typeHere == 'B') {
	    $bCount ++;
	} else {
	    $type = $typeHere if ($type eq 'B');
	}
	$len++;
    }
    return ($type,$len,$bCount,$maxScore)
}

sub effectiveQuality {
    my ($lb,$s) = @_;
    if ($lb == 1) {
	return 2.13354 * $s - 3.18592;
    } elsif ($lb == 2) {
	return 2.68949 * $s - 0.43062;
    } else {
	return 2.68067 * $s;
    }
}

sub readSorted12Map {
    my ($f) = @_;
    my $fh = openToRead($f);
    my @map;
    while (<$fh>) {
	chomp;
	my @a = split;
	push @map,\@a;
    }
    return \@map;
}

sub joinImputedMAndXToY1 {
    my ($mh,$xh,$sh,$lh,$pWipes,$pMap,$gh,$hh,$ah) = @_;

    my $off = 0;
    my $loff = 0; # layout.gz offset
    my $lastRealLayoutLine;
    while (1) {
	my $mLine = <$mh>;
	my $xLine = <$xh>;
	my $sLine = <$sh>; # should really check existence here too...
	$off ++;
	if (!$mLine) {
	    if ($xLine) {
		print "WARNING: more lines in imputed_x than imputed_m\n";
	    }
	    last;
	} elsif (!$xLine) {
	    print "WARNING: more lines in imputed_m than imputed_x\n";
	    last;
	}
	chomp $mLine;
	chomp $xLine;
	my $mLen = length($mLine);
	my $xLen = length($xLine);
	unless ($mLen == $xLen) {
	    print "WARNING: lengths differ\n$mLen: $mLine\n$xLen: $xLine\n";
	    last;
	}
	my $lLine; # layout line (non-annotation)

	my $rv = substr($sLine,0,1);
	if (defined $lh) {
	    my $loffTgt = layoutOffsetTarget($off,$pMap);
	    while ($loff < $loffTgt) {
		while (my $lLineHere = <$lh>) {
		    if ($lLineHere =~ /^>/) {
			print $ah $lLineHere; # note: these will be out of sync with others ...
		    } else {
			$loff ++;
			# Prefer to have an lLine that matches the sorted value.
			$lLine = $lLineHere unless (substr($lLine,0,1) eq $rv && substr($lLineHere,0,1) ne $rv);
			last;
		    }
		}
	    }
	    if (defined $lLine) {
		$lastRealLayoutLine = $lLine;
	    } else {
		$lLine = $lastRealLayoutLine;
		chomp $lLine;
		my $len = length($lLine);
		for (my $i=0; $i<$len; $i+=2) {
		    if (substr($lLine,$i,1) ne ' ') {
			substr($lLine,$i,2) = '- ';
		    }
		}
		$lLine .= "\n";
	    }
	}
	$mLine = applyIndelWipe($mLine,$off,$pWipes) if ($pWipes);
	my $gLine = substr($mLine,0,2); # merged line; copy ref value
	substr($gLine,0,1) =~ s/[0-9]/-/; # reformat gap character
	my $hasNuc = ($gLine =~ /^[A-Z]/);
	for (my $i=2; $i<$mLen; $i+=2) {
	    #my ($xv,$xq) = biasedACAndQuality(substr($xLine,$i,2));
	    my ($xv,$xq) = split("",substr($xLine,$i,2));
	    if ($xv eq 'A' || $xv eq 'a') {
		$gLine .= "=$xq"; # gap
	    } elsif ($xv eq 'T' || $xv eq 't') {
		$gLine .= "N$xq"; # diverged, so don't-know
	    } else {
		my $mv = substr($mLine,$i,1);
		my $mq = substr($mLine,$i+1,1);
		my $q = ($mq eq ' ') ? $xq : ($xq eq ' ' || $mq lt $xq) ? $mq : $xq;
		$gLine .= "$mv$q"; # nucleotide, if $mv is one
		$hasNuc = 1 unless ($hasNuc || $mv !~ /[A-Za-z]/);
	    }
	}
	unless ($IGNORE_GAP_LINES && !$hasNuc) {
	    print $gh "$gLine\n";
	    print $hh $sLine; # copy sorted line if we're copying here
	    print $ah $lLine if (defined $lLine); # ditto for layout line
	}
    }
}

sub layoutOffsetTarget {
    my ($off,$pMap) = @_;
    while ($#$pMap > 0 && $pMap->[1][1] <= $off) {
	shift @$pMap;
    }
    return $pMap->[0][0] - $pMap->[0][1] + $off;
}

sub applyIndelWipe {
    my ($line,$off,$pWipes) = @_;
    if (@$pWipes && $pWipes->[0][0] == $off) {
	# Locations of given type ('G' or 'N') have effective
	# quality qEff. If that's < 3.01, we overturn the value
	# and use the complementary quality.
	my (undef,$type,$qEff) = @{shift @$pWipes};
	# Find the replacement value
	my $len = length($line);
	my ($repVal,$repQual);
	if ($qEff < 3.01) {
	    $repVal = replacementValue($line,$type);
	    $repQual = complementaryQuality($qEff);
	} else {
	    $repQual = $qEff;
	}
	$repQual = int($repQual+0.5); # nearest int value
	my $repQC = $repQual > 126 ? ' ' : chr(33+$repQual);
	# We don't alter the ref nucleotide here ...
	for (my $i=2; $i<$len; $i+=2) {
	    my $v = substr($line,$i,1);
	    next if ($v eq ' ' || $v eq '=');
	    my $vType = ($v =~ /[A-Za-z]/) ? 'N' : 'G';
	    if ($vType eq $type) {
		substr($line,$i,1) = $repVal if (defined $repVal);
		substr($line,$i+1,1) = minQC($repQC,substr($line,$i+1,1));
	    }
	}
    }
    return $line;
}

sub replacementValue {
    my ($line,$type) = @_;
    my %sumQ;
    my $len = length($line);
    for (my $i=0; $i<$len; $i+=2) {
	my $v = substr($line,$i,1);
	next if ($v eq ' ' || $v eq '=');
	# If we're wiping gaps, we want $v to be a nucleotide, and VV.
	if (($type eq 'G') eq ($v =~ /[A-Za-z]/)) {
	    $v =~ tr/a-z/A-Z/;
	    my $qc = substr($line,$i+1,1);
	    my $q = ($qc eq ' ') ? 94 : (ord($qc)-33);
	    $sumQ{$v} += $q;
	}
    }
    # Highest-scoring value of desired type:
    my ($val) = sort {$sumQ{$b} <=> $sumQ{$a}} keys %sumQ;
    $val = '=' unless (defined $val); # must have something!
    return $val;
}

sub complementaryQuality {
    my ($q) = @_;
    if ($q <= 0) {
	return 127;
    } else {
	my $log10 = log(10);
	my $p = exp(-0.1 * $log10 * $q);
	return -10 * log(1-$p)/$log10;
    }
}

sub minQC {
    my ($qc1,$qc2) = @_;
    return $qc1 eq ' ' ? $qc2 :
	$qc2 eq ' ' ? $qc1 :
	$qc1 lt $qc2 ? $qc1 : $qc2;
}

sub createDanglesFromLayout {
    my ($f,$dFile) = @_;
    my $fh = openToRead($f);
    my $dh = openToWrite($dFile);
    my $off = 0; # offset from 1, i.e. exclude ">"
    my %rightDangle;
    report("$dFile: creating");
    while (<$fh>) {
	chomp;
	$off ++ unless (/^>/);
	if (/^>(\S+) (\d+) /) {
	    my ($r,$col) = ($1,($2+1)*2);
	    if ($r =~ /L$/) {
		printf($dh "%d L\n",$off+1);
	    } elsif ($r =~ /R$/) {
		my ($pos1,$pos2) = ($r =~ /:(\d+):(\d+)/);
		my $len = $pos2>$pos1 ? $pos2-$pos1+1 : $pos1-$pos2+1;
		$rightDangle{$col} = $len;
	    }
	} else {
	    foreach my $col (keys %rightDangle) {
		if (substr($_,$col,1) =~ /[A-Za-z]/) {
		    $rightDangle{$col} --;
		    if ($rightDangle{$col} == 0) {
			print $dh "$off R\n";
			delete $rightDangle{$col};
		    }
		}
	    }
	}
    }
    report();
}

# For paper:

sub runStaticImputer {
    runImputer1('stat',@_);
}

sub runImputerForHardGaps {
    runImputer1('impx',@_);
}


# Bias qualities and, if necessary, A/C so we get approx
# equal numbers of A's in non-refs as in refs.
sub biasedACAndQuality {
    my ($pair) = @_;
    my $xv = substr($pair,0,1);
    $xv =~ tr/ac/AC/;
    my $xq = substr($pair,1,1);
    unless ($xq eq ' ') {
	my $q = ord($xq)-$ZERO;
	if ($xv eq 'A') {
	    $q += $BIAS_TO_ABSENCE;
	    $xq = ($q < 94) ? chr($q+$ZERO) : ' ';
	} elsif ($q <= $BIAS_TO_ABSENCE) {
	    $q = $BIAS_TO_ABSENCE-$q;
	    $xv = 'A'; # SWITCH from C to A -- the point of this
	    $xq = chr($ZERO+$q);
	} else {
	    $q -= $BIAS_TO_ABSENCE;
	    $xq = chr($q+$ZERO);
	}
    }
    return ($xv,$xq);
}

1;
