package SGRP::MakeFastq;

use Exporter;
@ISA = qw(Exporter);

@EXPORT = qw(makeFastq);

use SGRP::Variables;
use Utility;
use File::Basename;

use Carp;

use strict;

my $ARCH = arch();

my $ZEMIN = '/nfs/team71/psg/zn1/src/process';
my $Z_ARCH = '/linux-64' if ($ARCH eq 'ia64');

my $DIR2FASTQ = "$ZEMIN/dir2fastq/many${Z_ARCH}/dir2fastq";
my $CLIP = "$ZEMIN/clip${Z_ARCH}/ssaha_clip"; # but we may not use it ... see below
my $SCREEN = "$ZEMIN/screen${Z_ARCH}/ssaha_screen";

my $CONTAM = '/nfs/infodata8/data/jcm/mouseWGS/saveStuff/masterContamMerge';

# If data has been processed by ASP, that's better than ssahaClip. 18-12-2006.

# ...however, ASP (as I am accessing it) only provides passed reads, and does
# not clip anything out of those reads. 9-10-2007.

my $USE_SSAHA_CLIP = 1;

sub makeFastq {
    my (@names) = @_;
    unless (@names) {
	confess("Usage (on cbi not farm): $0 mkfq strain1 strain2 ...");
    }
    my $pAlias = readAliases("$ALIASES/strains.txt");
    # Mapping from strain names to trace dirs (old system only).
    my $pLoc = readAliases("$ALIASES/traceDirs.txt");
    foreach my $name (@names) {
	my $wtsiName = reverseAlias($name,$pAlias);
	makeCloneDir($wtsiName,$name,$pLoc->{$name},$pAlias);
    }
}

sub reverseAlias {
    my ($name,$pAlias) = @_;
    while (my ($k,$v) = each %$pAlias) {
	return $k if ($v eq $name);
    }
    return $name;
}

sub makeCloneDir {
    my ($wtsiName,$realName,$traceDir,$pAlias) = @_;
    my $tgtDir = "$ROOT/read/$realName";
    my $movedAside;
    if (-d $tgtDir && !$USE_SSAHA_CLIP) {
	$movedAside = moveAside($tgtDir);
    }
    my $clip = "$tgtDir/clip.fastq";
    return unless (stamp($clip));
    mkdir_p($tgtDir);
    my $got; # whether we created something
    my @srcDirs;
    if ($traceDir) {
	@srcDirs = glob("$traceDir/[0-9][0-9][0-9][0-9]");
    }
    my $isComplex = (scalar(@srcDirs) > 0);
    unless ($isComplex) {
	@srcDirs = ($traceDir || $wtsiName);
    }
    my @clipped;
    foreach my $srcDir (@srcDirs) {
	my $idx = $isComplex ? basename($srcDir) : '0001';
	my $tgt = "$tgtDir/$idx.3.fastq";
	if (stamp($tgt)) { # should always succeed when not doing ssahaClip ...
	    my $grabbed = "$tgtDir/${idx}.1.fastq";
	    unless (-s $grabbed) {
		if ($traceDir) {
		    doDir2Fastq($srcDir,$grabbed);
		} else {
		    return unless doProj2Fastq($wtsiName,$tgtDir,$grabbed,$movedAside);
		}
	    }
	    if (-s $grabbed) {
		my $screened = doScreen($grabbed);
		push @clipped,($USE_SSAHA_CLIP ? doClip($screened) : $screened);
		$got = 1;
	    }
	    unstamp($tgt,1);
	} else {
	    $got = 1;
	    push @clipped,$tgt if (-s $tgt);
	}
    }
    if ($got) {
	createClipFile($clip,$pAlias,@clipped);
	#removeDebris($tgtDir);
    }
    unstamp($clip,1);
    return $got;
}	

sub doDir2Fastq {
    my ($srcDir,$tgt) = @_;
    report("$tgt: creating from $srcDir");
    my $log = "$tgt.log";
    my $err = "$tgt.err";
    `$DIR2FASTQ $tgt $srcDir >$log 2>$err`;
    report();
}

sub doProj2Fastq {
    my ($name,$dir,$tgt,$movedAsideDir) = @_;
    my $fofn = $tgt;
    $fofn =~ s/fastq$/fofn/;
    # We use PASS not PROJ here so we only get the reads
    # that pass ASP
    if (-s $fofn) {
	rename($fofn,"$fofn.orig");
    }
    report("creating $fofn with ts_group2fofn $name PASS");
    `ts_group2fofn $name PASS >$fofn 2>$fofn.err`;
    report();
    if (-s $fofn) {
	my $fofnOrig;
	$fofnOrig = "$movedAsideDir/" . basename($fofn) if ($movedAsideDir);
	if ($fofnOrig && identicalFiles($fofnOrig,$fofn)) {
	    report("no change in $fofn",1);
	    `rm -rf $dir`;
	    rename($movedAsideDir,$dir);
	    return 0;
	} else {
	    my $fofnAll = "$fofn.all";
	    report("creating $fofnAll with ts_group2fofn $name PROJ");
	    `ts_group2fofn $name PROJ >$fofnAll 2>$fofnAll.err`;
	    report();
	    report("creating $tgt with fofn2fastq");
	    `fofn2fastq $fofn > $tgt`;
	    report();
	    return 1;
	}
    } else {
	report("no reads found, removing $dir",1);
	`rm -rf $dir`;
	return 0;
    }
}

# Zemin's instructions:
#
# step 1
# /nfs/team71/psg/zn1/src/process/screen/linux-64/ssaha_screen
# -number 4000000
# /nfs/infodata8/data/jcm/mouseWGS/saveStuff/masterContamMerge
# yeast_clean1 yeast.fastq > screen.out
# 
# This produces three files:
# yeast_clean1_000.fastq (most clean reads)
# yeast_clean1_ctm.fastq (reads with one vector segment removed, but may
# contain vectors in the other parts of the read)
# screen.out (output file about contamination)
# (-number 4000000) makes sure the output reads are in one file. In our
# assembly, we produce a number of small output fastq files.
# 
# step 2
# /nfs/team71/psg/zn1/src/process/screen/linux-64/ssaha_screen
# /nfs/infodata8/data/jcm/mouseWGS/saveStuff/masterContamMerge yeast_clean2
# yeast_clean1_ctm.fastq > screen_ctm.out
# 
# step 3
# cat yeast_clean1_000.fastq yeast_clean2_000.fastq yeast_clean2_ctm.fastq >
# yeast_clean.fastq
# 
# This produces a cleaned fastq file yeast_clean.fastq

sub doScreen {
    my ($src) = @_;
    my $tgt = successorFastqFile($src);
    if (-s $tgt) {
	#report("$tgt: already exists",1);
    } elsif (! -s $src) {
	confess("Missing or empty: $src");
    } else {
	my $base = $src;
	$base =~ s/\.fastq$//;
	my $key1 = "${base}-screen1";
	my $key2 = "${base}-screen2";
	my $cmd = "$SCREEN -number 4000000 $CONTAM";

	report("${key1}_{000,ctm}.fastq: creating from $src");
	`$cmd $key1 $src > $key1.out`;
	report();

	report("${key2}_{000,ctm}.fastq: creating from ${key1}_ctm.fastq");
	`$cmd $key2 ${key1}_ctm.fastq > $key2.out`;
	report();

	report("$tgt: creating by concatenation");
	my $files = "${key1}_000.fastq ${key2}_000.fastq ${key2}_ctm.fastq";
	`cat $files > $tgt; rm $files`;
	report();
    }
    return $tgt;
}

sub successorFastqFile {
    my ($src) = @_;
    if ($src =~ /^(.*)\.(\d+)\.fastq$/) {
	return sprintf("%s.%d.fastq",$1,$2+1);
    } else {
	confess("no successor for $src");
    }
}

sub doClip {
    my ($src) = @_;
    my $tgt = successorFastqFile($src);
    if (-s $tgt) {
	#report("$tgt: already exists",1);
    } elsif (! -s $src) {
	confess("Missing or empty: $src");
    } else {
	report("$tgt: creating from $src");
	`$CLIP -qual 20 $src $tgt >$tgt.log 2>$tgt.err`;
	report();
    }
    return $tgt;
}

sub createClipFile {
    my ($tgt,$pAlias,@srcList) = @_;
    report("$tgt: creating");
    my %quad;
    foreach my $src (@srcList) {
	my $fh = openToRead($src);
	while (<$fh>) {
	    if (/^\@(\S+)/) {
		my $name = $1;
		# If we've run ssahaClip, this part will be present,
		# otherwise probably not.
		my ($lo,$hi) = (/bases (\d+) to (\d+)/);
		my $canon = canonicalReadName($name,$pAlias);
		my @a = $_;
		foreach my $k (1 .. 3) {
		    my $line = <$fh>;
		    push @a,$line;
		}
		# Fill in "bases i to j" when not present
		unless ($hi) {
		    foreach my $k (0,2) {
			chomp $a[$k];
			$a[$k] .= sprintf(" bases 1 to %d\n",length($a[1])-1);
		    }
		}
		substr($a[0],1,length($name)) = $canon;
		substr($a[2],1,length($name)) = $canon;
		$quad{$canon} = \@a;
	    } else {
		confess("Bad source line in $src: $_");
	    }
	}
    }
    my $gh = openToWrite($tgt);
    foreach my $name (sort keys %quad) {
	foreach my $line (@{$quad{$name}}) {
	    print $gh $line;
	}
    }
    $gh->close();
    report();
    return $tgt;
}

sub canonicalReadName {
    my ($name,$pAlias) = @_;
    # Fix YGPM nasty
    if ($name =~ /^YGPM/ && $name !~ /^YGPM-/) {
	$name = 'YGPM-' . substr($name,4);
    }
    my ($str,$read,$pq,$suf) = ($name =~ /^(.+)-(.+)\.([pq]1k)(.*)$/);
    $str = alias($str,$pAlias);
    $read .= "/$suf" if ($suf);
    return sprintf("%s-%s.%s",$str,$read,$pq);
}

sub removeDebris {
    my ($dir) = @_;
    foreach my $f (glob('0???.*')) {
	unlink($f);
    }
}

1;
