package SGRP::MakeGenomes;

use Exporter;
@ISA = qw(Exporter);

@EXPORT = qw(makeGenomes splitGenome);

use SGRP::Variables;

use Utility;
use File::Basename;

use Carp;

use strict;

sub makeGenomes {
    my @spList = @_;
    foreach my $sp (@spList) {
	my $gdir = "$REF/$sp";
	my $orig = "$gdir/genome_orig.gff";
	my $fa = "$gdir/genome.fa";
	my $gff = "$gdir/genome.gff";
	my $sp = basename($gdir);
	my $pAlias = readAliases("$ALIASES/${sp}_chr.txt");
	unless (newer($fa,$orig)) {
	    my $szFile = "$gdir/genome.sz";
	    makeGenome($fa,$szFile,$pAlias,$orig);
	               # We no longer add in vector here:
		       #glob("$REF/extras/*.fa"));
	}
	unless (newer($gff,$orig) || ! -f $orig) {
	    makeGFF($gff,$pAlias,$orig);
	}
	unless (newer("$gdir/p0",$fa)) {
	    report("splitting $fa");
	    splitGenome($fa,$gdir);
	    report();
	}
	my $gff = "$gdir/genome.gff";
	my $syn = "$gdir/4waysyn.txt";
	if (-s $gff && ! -s $syn) {
	    report("creating $syn");
	    findSynonymousPositions($gff,$fa,$syn);
	    report();
	}
    }
}

# > $REF/*/genome.fa
# < $ALIASES/${sp}_chr.txt
# < $REF/*/genome_orig.gff
# < $REF/extras/*.fa -- in principle tho not currently

sub makeGenome {
    my ($tgt,$szFile,$pAlias,@srcs) = @_;
    report("creating $tgt and $szFile");
    my %seq;
    foreach my $src (@srcs) {
	my $name;
	my $fh = openToRead($src);
	# This should bypass everything before the first ">" line.
	while (<$fh>) {
	    chomp;
	    if (/^>(\S+)/) {
		$name = canonicaliseContigName($1,$pAlias);
	    } elsif (defined $name) {
		$seq{$name} .= $_;
	    }
	}
    }
    my $gh = openToWrite($tgt);
    my $sh = openToWrite($szFile);
    foreach my $name (sort keys %seq) {
	printf($gh ">%s\n",$name);
	my $len = length($seq{$name});
	print $sh "$name $len\n";
	for (my $i=0; $i<$len; $i+=60) {
	    print $gh substr($seq{$name},$i,60) . "\n";
	}
    }
    $gh->close();
    report();
}

sub canonicaliseContigName {
    my ($name,$pAlias) = @_;
    $name =~ s/\[.*//; # chop off [stuff...]
    $name = alias($name,$pAlias); # apply specific alias
    $name =~ tr/A-Z/a-z/; # lower-case
    if ($name =~ /^(.)(\d+)$/) { # c3 -> c003, etc, for paradoxus
	my ($pfx,$idx) = ($1,$2);
	$name = sprintf("%s%3d",$pfx,$idx);
	$name =~ tr/ /0/;
    }
    return $name;
}

sub makeGFF {
    my ($tgt,$pAlias,$src) = @_;
    report("creating $tgt");
    my $fh = openToRead($src);
    my $gh = openToWrite($tgt);
    while (<$fh>) {
	last if (/^\#\#FASTA/);
	next if (/^\#/);
	chomp;
	my @a = split;
	my $name1 = $a[0];
	my $name2 = alias($name1,$pAlias);
	unless ($name2 eq $name1) {
	    $a[0] = $name2;
	    $a[-1] =~ s/^ID=${name1}/ID=$name2/;
	}
	print $gh join("\t",@a) . "\n";
    }
    report();
}

# < $REF/*/genome.fa
# > $REF/*/p?/$chr$suf.fa

sub splitGenome {
    my ($fa,$dir,$suf) = @_;
    my $fh = openToRead($fa);
    my $gh;
    while (<$fh>) {
	if (/^>.*\.\d\d\d$/) {
	    # Don't output anything when there's a .NNN suffix (for an aux sequence);
	    # these files are only used by CreateLayout.
	    $gh = undef;
	} elsif (/^>(\S*)/) {
	    my $chr = $1;
	    my $p = substr($chr,-1);
	    $p = 0 unless ($p =~ /^[0-9]$/);
	    $chr =~ s/^EXT\.//;
	    $chr =~ s/^REF\.//;
	    $gh = openToWrite("$dir/p$p/$chr$suf.fa");
	    print $gh ">$chr\n";
	} elsif ($gh) {
	    print $gh $_;
	}
    }
}	

# Prints all offsets that are inside a non-dubious coding region on at
# least one strand, and are four-way synonymous in whichever strands
# they are in such a region for.

# < $REF/*/genome.gff
# > $REF/*/4waysyn.txt

sub findSynonymousPositions {
    my ($gff,$fa,$syn) = @_;
    my %synPos;
    my $fh = openToRead($gff);
    while (<$fh>) {
	chomp;
	my ($chr,undef,$key,$lo,$hi,undef,$ori,$phase) = split;
	next unless ($key eq 'CDS');
	next if (/Dubious/); # don't accept "dubious" genes
	my @list;
	if ($ori eq '+') {
	    for (my $k=$lo+2-$phase; $k<=$hi; $k+=3) {
		$synPos{$chr}{$ori}{$k} = 1;
	    }
	} elsif ($ori eq '-') {
	    for (my $k=$hi+$phase-2; $k>=$lo; $k-=3) {
		$synPos{$chr}{$ori}{$k} = 1;
	    }
	}
    }
    my $fh = openToRead($fa);
    my ($chr,$seq);
    while (<$fh>) {
	chomp;
	if (/^>(\S+)/) {
	    processChr($chr,$seq,\%synPos) if ($chr);
	    $chr = $1;
	    $seq = "";
	} else {
	    $seq .= $_;
	}
    }
    processChr($chr,$seq,\%synPos) if ($chr);
    
    my $gh = openToWrite($syn);
    foreach my $chr (sort keys %synPos) {
	my @list = sort {$a <=> $b} (keys %{$synPos{$chr}{'+'}}, keys %{$synPos{$chr}{'-'}});
	$chr =~ tr/A-Z/a-z/;
	print $gh "$chr @list\n";
    }
    $gh->close();
}

sub processChr {
    my ($chr,$seq,$pSynPos) = @_;
    my $pH = $pSynPos->{$chr};
    return unless ($pH);
    my %is_syn12 = map {$_ => 1} qw(TC CT CC CG AC GT GC GG);
    foreach my $ori (qw(+ -)) {
	my @list = sort {$a <=> $b} keys %{$pH->{$ori}};
	foreach my $k (@list) {
	    my $codon12;
	    if ($ori eq '+') {
		$codon12 = substr($seq,$k-2,2);
	    } else {
		$codon12 = reverse(substr($seq,$k,2));
		$codon12 =~ tr/ACGT/TGCA/;
	    }
	    unless ($is_syn12{$codon12}) {
		# Delete in BOTH directions.
		delete $pH->{'+'}{$k};
		delete $pH->{'-'}{$k};
	    }
	}
    }
}


1;
