package SGRP::Margarita;

use Exporter;
@ISA = qw(Exporter);

@EXPORT = qw(runMargarita checkMargaritaOutputs markerCount);

use SGRP::Variables;
use SGRP::FindBranchLengths;
use SGRP::PrepareImputer;

use Utility;
use Sys::Hostname;
use File::Basename;

use Carp;

use strict;

# Current Margarita seems to have constant memory use, and CPU time (seconds)
# satisfying roughly
#
#   Time = (Markers * Haplotypes / 33) ^ 2.52
#
#

# Number of hours we would like to aim for:
my $MARGARITA_HOURS = 1;

# "Padding" (markers to repeat) from neighbouring chunks:
my $MARKER_OVERLAP_SIZE = 20; 

#my $MARKER_CHUNK_VOLUME = 35*sqrt(35)*14000;

my $MARGARITA_LITE = 0;

my $LIBBOOST_TEST_FILE = '/usr/lib/libboost_program_options.so';

sub runMargarita {
    my ($species,$filtMeth,$strainSet,@parts) = @_;
    unless (@parts) {
	confess("Usage: $0 marg species filtMeth strainSet part1 part2 ...");
    }
    checkBoostWillWork();
    my ($iter,$dir) = getCurrentIteration($species,$filtMeth,$strainSet,'p?');
    foreach my $part (@parts) {
	$part = "p$part" if ($part =~ /^[0-9]$/);
	# "[cs]" is to catch chr* and scplasm1 dirs only, not vc_.
	foreach my $f (filesFromSmallest(glob("$dir/$part/[cs]*/mgr_in"))) {
	    runMargarita1($f);
	}
    }
}

sub checkBoostWillWork {
    my $arl = absoluteReadLink($LIBBOOST_TEST_FILE);
    unless (-s $arl) {
	die(sprintf("Missing: %s (from %s) on host %s",
		    $arl,$LIBBOOST_TEST_FILE,hostname()));
    }
}

sub runMargarita1 {
    my ($src) = @_;
    return unless (isComplete($src,1));
    if (markerCount($src) == 0) {
	# nothing to do; will eventually call smim instead
	report("$src: no markers, nothing to do",1);
	return;
    }
    my $dir = dirname($src);
    unless (-x $MJM_JAVA) {
	confess("Missing or not executable: $MJM_JAVA");
    }
    my $jar;
    my $jarName = $MARGARITA_LITE ? 'MargaritaLite' : 'Margarita';
    foreach my $v qw(LUS HOME) {
	my $jar1 = $ENV{$v} . "/$jarName.jar";
	if (-s $jar1) {
	    $jar = $jar1;
	    last;
	}
    }
    unless ($jar) {
	confess("Cannot find $jarName.jar in LUS or HOME");
    }
    my $jcmd = "$MJM_JAVA -Xmx1500M -jar $jar";
    my @subs = glob("$dir/[0-9][0-9][0-9]");
    # Make them if needed ...
    unless (@subs) {
	@subs = splitMargaritaInput($src);
    }
    my @blFiles;
    foreach my $subd (@subs) {
	runMargarita2($jcmd,$subd);
    }
    mayConcatenateARGsBLFilesMain("$dir/args_bl",@subs);
}

sub markerCount {
    my ($f) = @_;
    my $fh = openToRead($f);
    my $line = <$fh>;
    chomp $line;
    my (undef,$nm) = split(" ",$line);
    return $nm;
}

sub splitMargaritaInput {
    my ($src) = @_;
    my $dir = dirname($src);
    my $fh = openToRead($src);
    $_ = <$fh>;
    chomp;
    # Numbers of strains and markers:
    my ($ns,$nm) = split;
    my (@offsets,@patterns);
    for (my $k=0; $k<$nm; $k++) {
	push @offsets,scalar(<$fh>);
    }
    chomp @offsets;
    @patterns = <$fh>; # rest of file
    $fh->close();
    my @ans;
    my $markerChunkSize = findMarkerChunkSize($ns);
    my $isCircular = isCircularChromosome(basename($dir));
    my $nChunks = int(($nm-1)/$markerChunkSize)+1;
    # Effective maximum chunk size: the size of all chunks except
    # perhaps the last.
    my $effectiveMCS = int(($nm-1)/$nChunks)+1;
    my $emcsWithPadding = $effectiveMCS+2*$MARKER_OVERLAP_SIZE;
    report("splitting $nm markers into $nChunks chunks of padded size $emcsWithPadding");
    for (my $i=0; $i<$nm; $i += $effectiveMCS) {
	my $subd = sprintf("$dir/m%3d",$i/$effectiveMCS);
	$subd =~ tr/ /0/;
	push @ans,$subd;
	my $tgt = "$subd/mgr_in";
	unless (-s $tgt) {
	    my $gh = openToWrite($tgt);
	    my $kLim = $i+$effectiveMCS;
	    $kLim = $nm if ($nm < $kLim);
	    my $padBelow = ($i>0 || $isCircular);
	    my $padAbove = ($i+$effectiveMCS<$nm || $isCircular);
	    # max and min here are to avoid going off the end of the marker set ...
	    my $kMinO =  $padBelow ? $i-$MARKER_OVERLAP_SIZE : $i;
	    my $kLimO =  $padAbove ? $kLim+$MARKER_OVERLAP_SIZE : $kLim;
	    printf($gh "%d %d\n",$ns,$kLimO-$kMinO);
	    my @pats;
	    # Print offsets:
	    if ($padBelow) {
		for (my $k=$kMinO; $k<$i; $k++) {
		    my $kk = $k;
		    my $adj = 0;
		    if ($isCircular) {
			my $approxSz = $offsets[-1]+$offsets[0];
			while ($kk < 0) {
			    $kk += scalar(@offsets);
			    $adj -= $approxSz;
			}
		    }
		    my $off = $offsets[$kk] + $adj;
		    push @pats,$patterns[$kk];
		    print $gh "$off\n";
		}
            }
	    for (my $k=$i; $k<$kLim; $k++) {
		print $gh $offsets[$k] . "\n";
		push @pats,$patterns[$k];
	    }
	    if ($padAbove) {
		for (my $k=$kLim; $k<$kLimO; $k++) {
		    my $kk = $k;
		    my $adj = 0;
		    if ($isCircular) {
			my $approxSz = $offsets[-1]+$offsets[0];
			while ($kk > $#offsets) {
			    $kk -= scalar(@offsets);
			    $adj += $approxSz;
			}
		    }
		    my $off = $offsets[$kk] + $adj;
		    push @pats,$patterns[$kk];
		    print $gh "$off\n";
		}
	    }
	    # Print patterns ...
	    print $gh @pats;
	    $gh->close();
	}
    }
    report();
    return @ans;
}


sub findMarkerChunkSize {
    my ($nHaps) = @_;
    my $secs = $MARGARITA_HOURS*3600;
    my $area = exp(log($secs)/2.52)*6600;
    my $nMarkers = int($area/$nHaps);
    # We insist on having at least twice as many markers as haplotypes to give
    # Margarita a chance of finding the structure; even this may not get satisfied
    # as the chromosome may be shorter than required ...
    if ($nMarkers < 2*$nHaps) {
	my $fac = 2*$nHaps/$nMarkers;
	my $estHours = $MARGARITA_HOURS * exp(log($fac)*2.52);
	printf(STDERR 
	       "WARNING: increasing nMarkers from %d to %d, so each ARG will take %3.2f hours not %3.2f hours\n",
	       $nMarkers,2*$nHaps,$estHours,$MARGARITA_HOURS);
	$nMarkers - 2*$nHaps;
    }
    return $nMarkers - 2*$MARKER_OVERLAP_SIZE;
}


# The circular chromosomes are currently the ones that
# are NOT of the form c.*\d, i.e. chrm and scplasm1.
# This can easily change!

sub isCircularChromosome {
    my ($name) = @_;
    return $name !~ /^c.*[0-9]$/;
}

sub runMargarita2 {
    my ($jcmd,$subd) = @_;
    # If it's all done, the arg_NNN.gz files may not be there ...
    unless ($MARGARITA_LITE) {
	return if (-s "$subd/args_bl" || -f "$subd/args_bl.touch");
    }
    my $mgrIn = "$subd/mgr_in";
    # Could have -qualoff as an extra switch here, or -alpha 20. The latter is much
    # better on Fregene data!
    my $extraFlags = "-alpha 20";
    my $mgrKey = 'mgr';
    if ($MARGARITA_LITE) {
	$mgrIn = ensureMargaritaLiteInput($subd,$mgrIn);
	$extraFlags = ' 1 -args'; # could have -trees too
	$mgrKey = 'mgrLite';
    }
    report("Margarita call: $jcmd",1);
    my @srcs;
    for (my $k=0; $k<$ARG_COUNT; $k++) {
	my $key = sprintf("%s/arg%3d",$subd,$k);
	$key =~ tr/ /0/;
	my $bl = "${key}_bl";
	push @srcs,$bl;
	my $toStamp = $MARGARITA_LITE ? "${key}.mgrLite.gz" : $bl;
	if (stamp($toStamp)) {
	    # We don't compress this file because sometimes we get garbage...
	    my $mgrout = "$key.$mgrKey";
	    my $mgrerr = "$key.$mgrKey.err";
	    unless (-s $mgrout) {
		report("creating $mgrout with $extraFlags");
		# "1" is number of ARGs to produce.
		# Removing %ARGS helps compatibility in lite case.
		my $mcmd = "$jcmd $mgrIn 1 $extraFlags 2>$mgrerr > $mgrout";
		# Try it a few times
		my $nTimes = 3;
		foreach my $k (1 .. $nTimes) {
		    `$mcmd`;
		    checkMargaritaOutput($mgrout,$mgrerr);
		    last unless (-s $mgrerr);
		    report();
		    unless ($k == $nTimes) {
			report("retrying after error in Margarita output: $mgrerr");
			foreach my $f (glob("$key*")) {
			    moveAside($f);
			}
		    }
		}
		if (-s $mgrerr) {
		    confess("REPEATED ERROR in Margarita output: $mgrerr");
		}
		report();
		# Empty error file can be deleted ...
		unlink($mgrerr);
	    }
	    unless ($MARGARITA_LITE) {
		findBranchLengths($key);
	    }
	    #unlink($mgrout);
	    unstamp($toStamp,1);
	}
    }
    unless ($MARGARITA_LITE) {
	foreach my $src (@srcs) {
	    unless (-s $src && ! -f "$src.touch") {
		my $b = basename($src);
		# Don't concatenate if not ready ...
		report("$subd/args_bl: not ready to concatenate because of $b",1);
		return;
	    }
	}
	concatenateArgsBLFilesSub("$subd/args_bl",@srcs);
    }
}

sub ensureMargaritaLiteInput {
    my ($dir,$src) = @_;
    my $tgt = "$dir/mgr_lite_in";
    unless (-s $tgt) {
	my $fh = openToRead($src);
	my $gh = openToWrite($tgt);
	chomp ($_ = <$fh>);
	my ($ns,$nm) = split;
	print $gh "$ns 0 $nm\n";
	for (my $k=0; $k<$nm; $k++) {
	    $_ = <$fh>;
	    print $gh $_;
	}
	my @pat;
	while (<$fh>) {
	    chomp;
	    s/(.)./\1/g; # discard quality scores
	    tr/a-z/A-Z/; # upper case all
	    my $vRef = substr($_,0,1);
	    my $iMax = length($_)-1;
	    for (my $i=0;$i<=$iMax;$i++) {
		# Slight approximation here: we take all non-ref
		# values as equivalent to each other.
		my $v = substr($_,$i,1);
		$pat[$i] .= $v eq ' ' ? 'M' : $v eq $vRef ? 0 : 1;
	    }
	}
	foreach my $hap (@pat) {
	    print $gh "$hap\n";
	}
	$gh->close();
    }
    return $tgt;
}

# Return values: 0 is OK, 1 is bad.

sub checkMargaritaOutput {
    my ($f,$errFile) = @_;
    my $fh = openToRead($f);
    my @lines = <$fh>;
    foreach my $k (0 .. $#lines) {
	chomp $lines[$k];
    }
    my ($nt,@rest) = split(" ",$lines[2]);
    if (scalar(@lines) < 4) {
	appendMessage($errFile,sprintf("Truncated file %s: only %d lines\n",$f,scalar(@lines)));
	return 1;
    }
    unless ($lines[0] eq '%ARGINFERENCE' && 
	    $lines[1] eq 'SEQS SNPS MUTS COAS RECS GECS TRCS SECS HEURP' &&
	    $nt > 0 && scalar(@rest) == 8 &&
	    $lines[3] eq 'ARG 0') {
	appendMessage($errFile,sprintf("Bad header in %s:\n%s\n",$f,join("\n",@lines)));
	return 1;
    }
    my %expect = map {$_ => 1} (0 .. $nt-1);
    @lines = @lines[4..$#lines];
    foreach my $k (0 .. $#lines) {
	my ($t,$op,@rest) = split(" ",$lines[$k]);
	unless ($t == $k) {
	    appendMessage($errFile,"Bad line in $f: $lines[$k]");
	    return 1;
	}
	my (@lo,@hi);
	if ($op eq 'mu' || $op eq 'er') {
	    @lo = ($rest[0]);
	    @hi = ($rest[1]);
	} elsif ($op eq 'co') {
	    @lo = ($rest[0],$rest[1]);
	    @hi = ($rest[2]);
	} elsif ($op eq 're') {
	    @lo = ($rest[0]);
	    @hi = ($rest[1],$rest[2]);
	}
	foreach my $id (@lo) {
	    if ($expect{$id}) {
		delete $expect{$id};
	    } else {
		appendMessage($errFile,"Bad line in $f: unexpected id $id in $lines[$k]");
		return 1;
	    }
	}
	foreach my $id (@hi) {
	    if ($expect{$id}) {
		appendMessage($errFile,"Bad line in $f: already expected id $id in $lines[$k]");
		return 1;
	    } else {
		$expect{$id} = 1;
	    }
	}
    }
    my @rem = sort {$a <=> $b} keys %expect;
    # We expect exactly 1 node (the top one) to remain
    if (@rem != 1) {
	appendMessage($errFile,"$f apparently truncated: still expecting @rem");
	return 1;
    } else {
	return 0;
    }
}

sub appendMessage {
    my ($g,$msg) = @_;
    if (defined $g) {
	my $gh = openToAppend($g);
	print $gh $msg;
	$gh->close();
    } else {
	print "$msg";
    }
}

sub concatenateArgsBLFilesSub {
    my ($tgt,@srcs) = @_;
    if (stamp($tgt)) {
	concatenateArgsBLFilesSub1($tgt,'',9,1,@srcs);
	concatenateArgsBLFilesSub1($tgt,'.log',4,0,@srcs);
	unstamp($tgt,1);
    }
}

sub concatenateArgsBLFilesSub1 {
    my ($tgtKey,$suf,$pos,$sort,@srcs) = @_;
    my $tgt = "$tgtKey$suf";
    report("creating $tgt");
    my $gh = $sort ? openToPipe("sort > $tgt") : openToWrite($tgt);
    foreach my $k (0 .. $#srcs) {
	my $f = "$srcs[$k]$suf";
	my $fh = openToRead($f);
	my $kk = sprintf("%3d",$k);
	my $ln=0;
	while (<$fh>) {
	    $ln ++;
	    if (length($_) < $pos+3) {
		printf(STDERR 
		       "WARNING: line %d of %s has length only %d; discarding it\n",
		       $ln,$f,length($_));
	    } else {
		substr($_,$pos,3) = $kk;
		print $gh $_;
	    }
	}
	#unlink($f);
    }
    $gh->close();
    report();
}

sub mayConcatenateARGsBLFilesMain {
    my ($tgt,@dirs) = @_;
    return if (-s $tgt || -f "$tgt.touch");
    foreach my $d (@dirs) {
	my $src = "$d/args_bl";
	return unless (-s $src && ! -f "$src.touch");
    }
    if (stamp($tgt)) {
	report("creating $tgt");
	concatenateARGsBLFilesMain($tgt,@dirs);
	unstamp($tgt,1);
	report();
	# Save having to do the prim step:
	prepareImputer1('m',dirname($tgt));
    }
}

# $dir/mgr_in contains: strainCount markerCount, then markerCount offsets
# $dir/args_bl contains: lo ArgNo hi stuff  lines, where lo and hi are
# marker indices.

sub concatenateARGsBLFilesMain {
    my ($tgt,@dirs) = @_;
    my $isCircular = isCircularChromosome(basename(dirname($tgt)));
    my $lowWater = 0;
    my $gh = openToPipe("sort > $tgt");
    for (my $i=0; $i<=$#dirs; $i++) {
	my $dir = $dirs[$i];
	my $padBelow = $isCircular || $i>0;
	my $padAbove = $isCircular || $i<$#dirs;
	my $nm = readMarkerCount("$dir/mgr_in");
	my $fh = openToRead("$dir/args_bl");
	my $drop = $MARKER_OVERLAP_SIZE*$padBelow;
	while (<$fh>) {
	    chomp;
	    my ($lo,$na,$hi,@rest) = split;
	    $lo = max($lo,$MARKER_OVERLAP_SIZE) if ($padBelow);
	    $hi = min($hi,$nm-1-$MARKER_OVERLAP_SIZE) if ($padAbove);
	    if ($lo <= $hi) {
		$lo += $lowWater-$drop;
		$hi += $lowWater-$drop;
		printf($gh "%6d %6d %6d %s\n",$lo,$na,$hi,join(" ",@rest));
	    }
	}
	$lowWater += $nm - $MARKER_OVERLAP_SIZE*($padBelow+$padAbove);
    }
}

sub readMarkerCount {
    my ($f) = @_;
    my $fh = openToRead($f);
    $_ = <$fh>;
    chomp;
    my @a = split;
    return $a[1];
}

sub checkMargaritaOutputs {
    my ($species,$filtMeth,$strainSet,@parts) = @_;
    unless (@parts) {
	confess("Usage: $0 cmar species filtMeth strainSet part1 part2 ...");
    }
    my ($iter,$dir) = getCurrentIteration($species,$filtMeth,$strainSet,'p?');
    my @bad;
    foreach my $p (@parts) {
	foreach my $d (glob("$dir/p$p/*/m???")) {
	    foreach my $n (0 .. $ARG_COUNT-1) {
		my $f = sprintf("%s/arg%3d.mgr",$d,$n);
		$f =~ tr/ /0/;
		if (-f $f) {
		    push @bad,$f if checkMargaritaOutput($f);
		} else {
		    print "Missing file: $f\n";
		    push @bad,$f;
		}
	    }
	}
    }
    if (@bad) {
	my %suggested;
	print "Suggested deletions:\n";
	foreach my $f (@bad) {
	    my $pat = $f;
	    $pat =~ s/\.mgr$/*/;
	    if (glob($pat)) {
		print "rm $pat\n";
	    }
	    my $d = dirname($f);
	    foreach my $abl ("$d/args_bl",dirname($d) . "/args_bl") {
		if (-s $abl && !$suggested{$abl}) {
		    print "rm $abl\n";
		    $suggested{$abl} = 1;
		}
	    }
	}
    } else {
	print "All Margarita files OK\n";
    }
}
	
1;
