package SGRP::MatchStrings;

use Exporter;
@ISA = qw(Exporter);

@EXPORT = qw(alignmentsToMatchStrings);

use Utility;
use SGRP::Alignment;
use Carp;

use strict;

# Input: chr name, matches.gz and qualities.gz files
# Output: match strings

my $DEBUG = 0;

my $NQS_THRESHOLD = 23;

sub alignmentsToMatchStrings {
    my ($chr,$aliFile,$qualFile,$outFile) = @_;
    my $ah = openToRead($aliFile);
    chomp (my $aSaved = <$ah>); # initialize
    my $qh = openToRead($qualFile);
    #my $gh;
    #if (1) {
	#$outFile =~ s/\.gz$//;
	#$gh = openToWrite($outFile);
	#autoflush $gh 1;
    #} else {
    my $gh = openToPipe("sort | gzip -c > $outFile");
    #}
    my $pAli;
    #my %isInsertionPosition;
    my $aliId = 1;
    while (1) {
	$pAli = readAlignment($ah,\$aSaved,$qh,$chr); # from crlo
	last unless ($pAli);
	$pAli->applyTrim();
	my $special = $pAli->special;
	$special = " $special" if ($special);
	foreach my $ms ($pAli->getMatchStrings()) {
	    # Don't print low-quality insertions in "extra" round at the end:
	    my @a;
	    unless ($special && 
		    (@a = split(" ",$ms)) &&
		    $a[1] eq 'I' && $a[2] < 3 && qualitySum($a[6]) < 30) {
		print $gh "$ms $aliId$special\n";
	    }
	}
	$aliId ++;
    }
    $gh->close();
}

sub readAlignment {
    my ($ah,$pASaved,$qh,$chrRequired) = @_;

    my %ali;

    # Record from the ALIGNMENT line $$pASaved itself.
    return undef unless (defined $$pASaved); # end of file
    my $pAli = new Alignment($$pASaved);
    if ($DEBUG >= 1) {
	printf("> $$pASaved\n");
	printf("Starting to process: %s\n",$pAli->toString());
    }
    # We decrement all offsets by 1 to get zero-offset for both query and sbjct...
    # let's see if we can avoid this ...
    # $pAli->decrementOffsets;
    $$pASaved = undef;

    # Prepare to read clone and reference sequences:
    my ($cloneSeq,$refSeq);
    
    # Get the quality string from the qualities.gz file,
    # chopping off the front and/or reversing if required:
    my $qData = getQualityString($pAli,$qh);

    # Build up cloneSeq and refSeq from the alignment file,
    # and set the new $$pASaved;
    my ($cloneSeq,$refSeq) = readAlignmentSequences($ah,$pASaved);
    
    # Introduce gaps into qData where needed, yielding cloneQual
    my $cloneQual = insertGapQualities($qData,$cloneSeq,$pAli->name);

    # Lower-case positions in $cloneSeq that do not meet HighQuality.
    $cloneSeq = caseByHighQuality($cloneSeq,$refSeq,$cloneQual);

    $pAli->querySeq($cloneSeq);
    $pAli->queryQual($cloneQual);
    $pAli->sbjctSeq($refSeq);

    # Shorten sequences and adjust coords if necessary; then we can forget
    # about trimming.
    if ($DEBUG >= 1) {
	printf("Before trimming: %s\n",$pAli->toString());
    }
    $pAli->applyTrim();

    if ($DEBUG >= 1) {
	printf("After  trimming: %s\n",$pAli->toString());
    }

    return $pAli;

}

sub getQualityString {
    my ($pAli,$qh) = @_;

    # Get the quality line, and check it exists and is for the right clone:
    my $qLine = <$qh>;
    unless ($qLine) {
	printf(STDERR "WARNING: expected qLine for %s, found EOF\n",$pAli->name);
	return undef;
    }
    chomp $qLine;
    my ($qName,$qData) = split(/ /,$qLine);
    unless ($qName eq $pAli->name) {
	confess(sprintf("Match clone %s but Quality clone %s",$pAli->name,$qName));
    }

    # Chop off the initial and final parts of the quality string that didn't get aligned,
    # and reverse it if required:
    $qData = substr($qData,0,$pAli->rhi); # keep the first rhi characters
    $qData = substr($qData,$pAli->rlo-1); # and delete the first rlo-1
    if ($pAli->fc eq 'C') {
	$qData = reverse($qData);
    }

    return $qData;
}

sub readAlignmentSequences {
    my ($ah,$pASaved) = @_;
    my ($cloneSeq,$refSeq);
    while (<$ah>) {
	chomp;
	my ($key,undef,$seq) = split;
	$key =~ s/:$//;
	if ($key eq 'Query') {
	    $cloneSeq .= $seq;
	} elsif ($key eq 'Sbjct') {
	    $refSeq .= $seq;
	} elsif ($key eq 'ALIGNMENT') {
	    $$pASaved = $_;
	    last;
	}
    }
    return ($cloneSeq,$refSeq);
}

# $qData has no gaps; create $cloneQual by introducing gap qualities
# into it according to where the gaps are in $cloneSeq.

sub insertGapQualities {
    my ($qData,$cloneSeq,$name) = @_;
    my $qOff = 0;
    my $cloneQual;
    my $qdl = length($qData);
    for (my $aOff = 0; $aOff < length($cloneSeq); $aOff++) {
	if (substr($cloneSeq,$aOff,1) eq '-') {
	    # GAP QUALITY: we could have ':' here if we want a fixed
	    # value.
	    if ($qOff >= $qdl) {
		print STDERR "WARNING: qOff $qOff, qdl $qdl for clone $name\n";
		# and then the below will fail ...
	    }
	    $cloneQual .= minimumNucleotideQuality(substr($qData,$qOff-1,2));
	} else {
	    # We could have a ' ' where $q is repeated, but this
	    # only reduces compressed space by 15% or so (it's 6 times
	    # higher than without any qualities at all...)
	    my $q = substr($qData,$qOff,1);
	    $cloneQual .= $q;
	    $qOff ++;
	}
    }
    return $cloneQual;
}

sub minimumNucleotideQuality {
    my ($pair) = @_;
    my ($q1,$q2) = (substr($pair,0,1),substr($pair,1,1));
    return ($q1 lt $q2 ? $q1 : $q2);
}    
	
sub caseByHighQuality {
    my ($cloneSeq,$refSeq,$cloneQual) = @_;
    my @ca = split(//,$cloneSeq);
    my @ra = split(//,$refSeq);
    my @qa = split(//,$cloneQual);
    my @aa;
    foreach my $k (0 .. $#ca) {
	my $cn = $ca[$k];
	if ($cn eq 'N') {
	    $cn = 'n'; # N is always low-quality
	} elsif ($cn ne '-') { # can't lower-case a gap!
	    my $rn = $ra[$k];
	    my $q = $qa[$k];
	    if ($k < 5 || $k > $#ca-5 || $q lt $NQS_THRESHOLD) {
		$cn =~ tr/ACGT/acgt/;
	    } else {
		my $nMis = $ca[$k] eq $ra[$k] ? 0 : 1;
		$nMis = countMismatches(\@ca,\@ra,\@qa,$k,-1,$nMis);
		if ($nMis > 1) {
		    $cn =~ tr/ACGT/acgt/;
		} else {
		    $nMis = countMismatches(\@ca,\@ra,\@qa,$k,1,$nMis);
		    if ($nMis > 1) {
			$cn =~ tr/ACGT/acgt/;
		    }
		}
	    }
	}
	$aa[$k] = $cn;
    }
    return join("",@aa);
}

# Return 99 if we get a poor-quality base, as this will cause
# a thumbs-down just like nMis = 2.
sub countMismatches {
    my ($pCA,$pRA,$pQA,$k,$d,$nMis) = @_;
    my $nToDo = 5;
    while ($nToDo > 0) {
	$k += $d;
	return 99 if ($k < 0 || $k > $#$pCA);
	if ($pCA->[$k] eq '-') {
	    if ($pRA->[$k] eq '-') {
		next;
	    } elsif ($nMis) {
		return 2;
	    } else {
		$nMis = 1;
	    }
	} elsif ($pQA->[$k] lt '0') {
	    return 99; # poor-quality neighbour
	} elsif ($pCA->[$k] ne $pRA->[$k]) {
	    if ($nMis) {
		return 2;
	    } else {
		$nMis = 1;
	    }
	}
	$nToDo --;
    }
    return $nMis;
}

sub qualitySum {
    my ($lst) = @_;
    my $sum = 0;
    foreach my $c (split("",$lst)) {
	$sum += ord($c)-33;
    }
    return $sum;
}
