package SGRP::MergeMatches;

use Exporter;
@ISA = qw(Exporter);

@EXPORT = qw(mergeMatches);

use SGRP::SelectNOAs;
use SGRP::Variables;

use Utility;
use File::Basename;

use Carp;

use strict;

sub mergeMatches {
    my ($species,$paramKey,$strainSet,@parts) = @_;

    unless (@parts) {
	confess("Usage: $0 merg species filtParamName strainSetName 0 1 2 ...");
    }

    # We're going to create the align directory, which is first in
    # version 000, so undef here...
    my ($iter,$tgtDir) = getCurrentIteration($species,$paramKey,$strainSet,undef);
    #if (-d "$tgtDir/p0") {
	#report("already exists: $tgtDir/p0",1);
	#return;
    #}

    my $pStrains = SGRP::SGRP::getStrains($strainSet);

    foreach my $p (@parts) {

	# Get offset map and input alignment dirs
	my ($pOM,$pSrcDirs) = collectOffsetMapAndAlignmentDirs($p,$species,$paramKey,$strainSet,$pStrains,$iter,$tgtDir);

	my ($pMatchFiles,$pClipFiles) = collectMatchAndClipFiles($pStrains,$pSrcDirs);
	
	foreach my $chr (getChromosomeNames($pOM,$species,$p)) {
	    my $chrDir = "$tgtDir/p$p/$chr";
	    if (-s "$chrDir/qualities.gz") {
		print "Already done: $chrDir\n";
		next;
	    }
	    # Get non-overlapping alignments (main piece of work)
	    #report("deciding alignments to accept for p$p/$chr");
	    my $pAccept = decideAlignmentsToAccept($species,$paramKey,$pStrains,$pOM,$pSrcDirs,$chr,$chrDir);
	    if ($pAccept) {
		# Create match and qualities files
		my $matchFile = mergeAlignmentsToMatchFile($pMatchFiles,$pAccept,$pOM,$chrDir);
		createQualitiesFile($pClipFiles,$matchFile,$chrDir);
	    }
	}

    }
}

sub collectOffsetMapAndAlignmentDirs {
    my ($p,$species,$paramKey,$strainSet,$pStrains,$iter,$tgtDir) = @_;
    my (@srcDirs,@gmFiles);
    foreach my $str (@$pStrains) {
	if ($iter) {
	    push @srcDirs,"$tgtDir/align/$str";
	    push @gmFiles,"$tgtDir/assembly/$str/genome.map";
	} else {
	    push @srcDirs,"$ROOT/read/$str/$species/$paramKey";
	}
    }
    unless ($iter) {
	my $key = ($species eq 'para') ? 'augmented' : 'extended';
	push @gmFiles,"$ROOT/ref/$species/$key.map";
    }
    #report("reading offset map files");
    my $pOM = new OffsetMap($p,@gmFiles);
    #report();
    return ($pOM,\@srcDirs);
}

sub collectMatchAndClipFiles {
    my ($pStrains,$pSrcDirs) = @_;
    my (@matchFiles,@clipFiles);
    foreach my $k (0 .. $#$pStrains) {
	my $str = $pStrains->[$k];
	my $srcDir = $pSrcDirs->[$k];
	# This will contain all the reads that passed single-strain nola
	# filtering, not just those in pAccept, which are multi-strain.
	my $mf = "$srcDir/ssahaSNP.pass.gz";
	if (-s $mf) {
	    push @matchFiles,$mf;
	} else {
	    confess("Missing or empty: $mf");
	}
	my $cf = "$ROOT/read/$str/renamed.fastq";
	if (-s $cf) {
	    push @clipFiles,$cf;
	} else {
	    confess("Missing or empty: $cf");
	}
    }
    return (\@matchFiles,\@clipFiles);
}

sub getChromosomeNames {
    my ($pOM,$species,$p) = @_;
    if (defined $pOM) {
	return $pOM->getChromosomeNames; # non-initial iterations
    }
    my $fh = openToRead("$ROOT/ref/$species/genome.sz");
    my @names;
    while (<$fh>) {
	chomp;
	my ($name) = split;
	unless ($name eq 'ALL' || $name =~ /^vc_/) {
	    if (substr($name,-1) == $p) { # == so that letters match 0
		push @names,$name;
	    }
	}
    }
    return @names;
}

sub decideAlignmentsToAccept {
    my ($species,$paramKey,$pStrains,$pOM,$pSrcDirs,$chr,$tgtDir) = @_;
    my %accept;
    my $noaFile = "$tgtDir/alilines.nola.txt";
    my $logFile = "$tgtDir/alilines.nola.log";
    if (stamp($noaFile)) {
        my @dirs;
        selectNonOverlappingAlignments($chr,$pSrcDirs,$pOM,2,$noaFile,$logFile);
	unstamp($noaFile,1);
    }
    if (-s $noaFile) {
	my $fh = openToRead($noaFile);
	while (<$fh>) {
	    $accept{$_} = 1; # UNchomped line
	}
	return \%accept;
    } elsif ($IS_FINAL_ITERATION) {
	# either final or extra ...
	linkBeforeSorted($tgtDir);
	return undef;
    } else {
	report("MISSING OR EMPTY: $noaFile, so removing its directory",1);
	`rm -rf $tgtDir`; # here's how we terminate iteration of a chromosome ...
	return undef;
    }
}

sub linkBeforeSorted {
    my ($tgtDir) = @_;
    `rm -rf $tgtDir`;
    my @a = split(/\//,$tgtDir);
    $a[-3] = '???';
    my @earlier = glob(join("/",@a));
    unless (@earlier) {
	report("MISSING OR EMPTY: $tgtDir/alilines.nola.txt",1);
	die("But no previous version exists");
    }
    my $srcDir = $earlier[-1];
    my @aa = split(/\//,$srcDir);
    my $trio = $aa[-3];
    my $srcSorted = "$srcDir/sorted.gz";
    my $srcLayout = "$srcDir/layout.gz";
    mkdir_p($tgtDir);
    foreach my $f (glob("$srcDir/*")) {
	next if (-d $f);
	if ($f ne $srcSorted && newer($srcSorted,$f) && !newer($srcLayout,$f)) {
	    my $b = basename($f);
	    my $g = "$tgtDir/$b";
	    if (-s $g) {
		moveAside($g);
	    } else {
		unlink($g);
	    }
	    link($f,$g);
	    report("Linking $g from $trio",1);
	}
    }
}
	
    

sub mergeAlignmentsToMatchFile {
    my ($pMatchFiles,$pAccept,$pOM,$tgtDir) = @_;
    my $matchFile = "$tgtDir/matches.gz";
    if (stamp($matchFile)) {
	report("creating $matchFile");
	mergeAlignments($pMatchFiles,$pAccept,$pOM,$matchFile);
	report();
	unstamp($matchFile);
    }
    return $matchFile;
}

sub mergeAlignments {
    my ($pMatchFiles,$pAccept,$pOM,$tgt) = @_;
    my $gh = openToWrite($tgt);
    my @in;
    foreach my $f (@$pMatchFiles) {
	my $fh = openToRead($f);
	while (<$fh>) {
	    last if (/^ALIGNMENT /);
	}
	push @in,mungeLine($_,$pOM,$fh);
    }
    my $active = 0;
  LOOP:
    while (@in) {
	@in = sort {$a->[0] cmp $b->[0] || $a->[1] <=> $b->[1]} @in;
	$active = $pAccept->{$in[0][2]}; # the alignment line itself
	if ($active) {
	    chomp (my $line = $in[0][2]);
	    my $utlo = $in[0][3];
	    my $cs = $in[0][4];
	    print $gh "$line @ $utlo $cs\n";
	}
	my $fh = $in[0][5];
	while (<$fh>) {
	    if (/^ALIGNMENT/) {
		$in[0] = mungeLine($_,$pOM,$fh);
		next LOOP;
	    } else {
		print $gh $_ if ($active);
	    }
	}
	# If we get here, $fh is exhausted
	shift @in;
    }
    $gh->close();
}

sub mungeLine {
    my ($line,$pOM,$fh) = @_;
    my $pA = new Alignment($line,$pOM,1);
    # See Filter::getZeroScorerSequences for this sort order too:
    my $untrimmedLo = $pA->clo(2);
    my $lo = ($pA->chi(2) < $untrimmedLo) ? $untrimmedLo-0.5 : $untrimmedLo;
    #my $pAns = [$pA->chr(2),$lo+$pA->loTrim,$line,$fh];
    # Note: the start point $lo+$pA->loTrim, which will be printed after
    # the "@", is trimmed, but the cigar string isn't.
    my $pAns = [$pA->chr(2),$lo+$pA->loTrim,$line,$untrimmedLo,$pA->cigarString,$fh];
    return $pAns;
}

sub createQualitiesFile {
    my ($pClipFiles,$matchFile,$tgtDir) = @_;
    my $qualFile = "$tgtDir/qualities.gz";
    if (-s $matchFile && ! -f "$matchFile.touch" && stamp($qualFile)) {
	report("creating $qualFile");
	sortQualities($pClipFiles,$matchFile,$qualFile);
	report();
	unstamp($qualFile);
    }
    return $qualFile;
}

sub sortQualities {
    my ($pClipFiles,$matchFile,$qualFile) = @_;
    my $end = ($qualFile =~ /gz$/) ? "| gzip -c > $qualFile" : "> $qualFile";
    my $gh = openToPipe("sort -n | sed 's/^[0-9]* //' $end");
    my $n=0;
    my %posn;
    my $fh = openToRead($matchFile);
    while (<$fh>) {
	if (/ALIGNMENT/) {
	    my (undef,undef,$clone) = split;
	    push @{$posn{$clone}},$n;
	    $n++;
	}
    }
    foreach my $f (@$pClipFiles) {
	next unless (-s $f);
	my $fh = openToRead($f);
	my $clone;
	while (<$fh>) {
	    if ($clone) {
		chomp;
		if (defined $posn{$clone}) {
		    foreach my $n (@{$posn{$clone}}) {
			printf($gh "%d %s %s\n",$n,$clone,$_);
		    }
		    $posn{$clone} = undef;
		}
	    }
	    if (/^\+(.*) bases/) {
		$clone = $1;
	    } else {
		$clone = undef;
	    }
	}
    }
    $gh->close();
}

1;
