package SGRP::PrepareImputer;

use Exporter;
@ISA = qw(Exporter);

@EXPORT = qw(prepareImputer prepareImputer1);

use SGRP::Variables;

use Utility;
use File::Basename;

use Carp;

use strict;

my $IS_CROSS_SPECIES;

my $ZERO = ord('!');

# prim command: prepare for Margarita imputation, so 'm' suffix. Normally
# called from Margarita.

sub prepareImputer {
    my ($species,$filtMeth,$strainSet,@parts) = @_;
    unless (@parts) {
	confess("Usage: $0 prim species filtMeth strainSet part1 part2 ...");
    }
    # This should really be passed down as a parameter...
    $IS_CROSS_SPECIES = $species eq 'cere' && $strainSet =~ /^PAR/;
    my ($iter,$dir) = getCurrentIteration($species,$filtMeth,$strainSet,'p?');
    foreach my $part (@parts) {
	$part = "p$part" if ($part =~ /^[0-9]$/);
	foreach my $sf (filesFromSmallest(glob("$dir/$part/*/sorted.gz"))) {
	    foreach my $suf ('m','x',@STATIC_IMPUTATION_SUFS) {
		prepareImputer1($suf,dirname($sf));
	    }
	}
    }
}

sub prepareImputer1 {
    my ($suf,$dir,$species) = @_;
    my $tgt = "$dir/imputed_$suf.gz";
    return if (-s $tgt);
    if ($suf eq 'x') {
	my $ac = "$dir/sorted_ac.gz";
	if (stamp($ac)) {
	    createACFile($dir);
	    unstamp($ac);
	}
	if (-f "$ac.touch") {
	    report("$ac being created -- not doing impx",1);
	} elsif (stamp("$dir/impm/i000/dual_in")) {
	    prepareImputerForHardGaps($dir);
	    unstamp("$dir/impm/i000/dual_in");
	}
    } else {
	prepareImputer2($suf,$dir);
	# Ensure dual_in files are created too if final iter
	if ($suf eq 'm' && $IS_FINAL_ITERATION) {
	    prepareImputer1('x',$dir);
	}
    }
}

sub createACFile {
    my ($d) = @_;
    my ($ac1,$acRegs) = createACFile1($d);
    my $ac = "$d/sorted_ac.gz";
    my $acRegs2 = "$d/sorted_ac_nondivquals.txt";
    createACFile2($ac1,$acRegs,$ac,$acRegs2);
}

sub createACFile1 {
    my ($d) = @_;
    my $f = "$d/sorted.gz";
    my $g = "$d/sorted_ac1.gz";
    my $h = "$d/sorted_ac_regions.txt";
    return ($g,$h) if (-s $g && -s $h); # already done first step
    my $pRBP = getRepeatBoundaryPositions($d);
    my ($ntLo,$ntHi);
    if ($IS_CROSS_SPECIES) {
	($ntLo,$ntHi) = nonTelomereRange(30000,$f);
    }
    my $fh = openToRead($f);
    my $hdr = <$fh>;
    report("$g: creating");
    my $gh = openToWrite($g);
    #autoflush $gh 1; # for debugging only
    print $gh $hdr;
    chomp $hdr;
    my $hh = openToWrite($h);
    my @strains = split(" ",substr($hdr,1));
    my $nCols = 2*(scalar(@strains)+1);
    my $off = 0; # offset in file: first line (header) is zero
    my $k = 0; # current offset in $pRegs
    my ($patStart,$pat,@buf);
    my $rOff = 0;
    while (<$fh>) {
	chomp;
	$off ++;
	$rOff ++ if (/^[A-Z]/);
	my $line;
	if (allBlankOrHardGap($_)) {
	    # Everything is space or hard gap: we _have_ to say all A's,
	    # as nucleotide imputation will yield nothing!
	    $line = 'A ' x ($nCols/2);
	} else {
	    my %seen;
	    # True if we're in non-telomere region and nothing sequenced:
	    my $blankApartFromRef;
	    if (defined $ntLo && defined $ntHi && $off >= $ntLo && $off <= $ntHi) {
		$blankApartFromRef = ($_ !~ /..[^ ]/);
	    }
	    for (my $i=0; $i<$nCols; $i+=2) {
		my $v = substr($_,$i,1);
		my $ac;
		if ($v eq ' ' || $v eq '' || !defined $v) {
		    $ac = $blankApartFromRef ? 'A' : ' ';
		} elsif ($v eq '=' || ($i == 0 && substr($_,1,1) eq '!')) {
		    $ac = 'A';
		} else {
		    $ac = 'C';
		}
		$line .= "$ac ";
		$seen{$ac} = 1 if ($ac ne ' ');
	    }
	    # If all same, fill in with that value, as it's the only
	    # possible one.
	    if (0) {
		my @seen = keys %seen;
		if (@seen == 1) {
		    $line = "$seen[0] " x ($nCols/2);
		}
	    }
	}
	push @buf,$line;
	if (/^[A-Z]/ && atRepeatBoundaryPosition($rOff,$pRBP)) {
	    spreadPresenceConstraints(\@buf) if (@buf > 1);
	    ($patStart,$pat) = printBufferAndUpdatePat(\@buf,$patStart,$pat,$off,$gh,$hh);
	    @buf = ();
	}
    }
    spreadPresenceConstraints(\@buf) if (@buf > 1);
    printBufferAndUpdatePat(\@buf,$patStart,$pat,$off,$gh,$hh,1);
    report();
    return ($g,$h);
}

sub getRepeatBoundaryPositions {
    my ($d) = @_;
    my @comps = split(/\//,$d);
    my $chr = $comps[-1];
    my $sp = $comps[-5];
    # Created with repeatBoundaryPositions.pl *.cere.ali in "self" dir...
    my $rbpFile = "$ROOT/ref/$sp/self/repeatBoundaryPositions.txt";
    return unless (-s $rbpFile);
    my @a;
    my $fh = openToRead($rbpFile);
    while (<$fh>) {
	chomp;
	my ($c,$off) = split;
	if ($c eq $chr) {
	    push @a,$off;
	}
    }
    return @a ? \@a : undef;
}

sub nonTelomereRange {
    my ($len,$f) = @_;
    # Only for linear chromosomes:
    return undef unless (basename(dirname($f)) =~ /^chr..$/);
    my ($ntLo,$ntHi);
    my $fh = openToRead($f);
    <$fh>;
    my $seq;
    while (<$fh>) {
	chomp;
	$seq .= substr($_,0,1);
    }
    my $nToGet = $len;
    my $iMax = length($seq)-1;
    for (my $i=0; $i<=$iMax; $i++) {
	if (substr($seq,$i,1) =~ /^[A-Z]/) {
	    $nToGet --;
	    if ($nToGet == 0) {
		$ntLo = $i+1;
		last;
	    }
	}
    }
    my $nToGet = $len;
    for (my $i=$iMax; $i>=0; $i--) {
	if (substr($seq,$i,1) =~ /^[A-Z]/) {
	    $nToGet --;
	    if ($nToGet == 0) {
		$ntHi = $i-1;
		last;
	    }
	}
    }
    return ($ntLo,$ntHi);
}

sub allBlankOrHardGap {
    my ($line) = @_;
    my $len = length($line);
    # Note: we don't count ref-quality = '!' as a hard gap here, because we
    # can use the (lower-case) nucleotide value to impute from.
    for (my $i=0; $i<$len; $i+=2) {
	my $v = substr($line,$i,1);
	return 0 unless ($v eq ' ' || $v eq '=');
    }
    return 1;
}

sub atRepeatBoundaryPosition {
    my ($off,$pRBP) = @_;
    return 1 unless (defined @$pRBP); # always OK if no boundaries defined
    if ($pRBP->[0] == $off) {
	shift @$pRBP;
	return 1;
    } else {
	return 0;
    }
}

sub spreadPresenceConstraints {
    my ($pBuf) = @_;
    return;
    my $len = scalar(@$pBuf);
    my $iMax = length($pBuf->[0])-2;
    my $th = $len*0.1;
    for (my $i=0; $i<=$iMax; $i+=2) {
	my $sumA = 0;
	my $sumC = 0;
	foreach my $line (@$pBuf) {
	    if ($line =~ /^C/) { 
		$sumC ++ if (substr($line,$i,1) eq 'C');
	    } elsif ($line =~ /^A/) {
		$sumA ++ if (substr($line,$i,1) eq 'A');
	    }
	}
	if (max($sumA,$sumC) < $th) {
	    # too uncertain; will impute
	} elsif ($sumA >= $sumC) { # bias towards A!
	    foreach my $line (@$pBuf) {
		substr($line,$i,1) =~ tr/ /A/;
	    }
	} elsif ($sumC >= $sumA*2) {
	    foreach my $line (@$pBuf) {
		substr($line,$i,1) =~ tr/ /C/;
	    }
	}
    }
}

sub printBufferAndUpdatePat {
    my ($pBuf,$patStart,$pat,$off,$gh,$hh,$final) = @_;
    foreach my $i (0 .. $#$pBuf) {
	my $line = $pBuf->[$i];
	print $gh "$line\n";
	unless ($line eq $pat) {
	    my $offI = $off+$i-$#$pBuf;
	    if (defined $pat) {
		printf($hh "%d %d %s\n",$patStart,$offI-1,$pat);
	    }
	    $patStart = $offI;
	    $pat = $line;
	}
    }
    if ($final && defined $pat) {
	printf($hh "%d %d %s\n",$patStart,$off,$pat);
    }
    return ($patStart,$pat);
}

sub prepareImputerForHardGaps {
    my ($d) = @_;
    my $f = "$d/sorted_ac.gz";
    if (! -s $f) {
	report("cannot run impx: missing $f",1);
	return;
    } elsif (! -d "$d/impm") {
	report("cannot run impx: missing $d/impm",1);
	return;
    }
    my $fh = openToRead($f);
    <$fh>; # discard header; we'll get them from the $if files
    report("$d/impm/iNNN/dual_in: creating");
    my $ni = 0;
    my ($ih,$gh);
    while (<$fh>) {
	my $sLine = $_;
	unless (<$ih>) {
	    # We only use the imp_in files to get the right number
	    # of lines per dual_in file; we don't want the contents.
	    my $if = sprintf("%s/impm/i%3d/imp_in",$d,$ni);
	    $if =~ tr/ /0/;
	    my $g = sprintf("%s/impm/i%3d/dual_in",$d,$ni);
	    $g =~ tr/ /0/;
	    unless (-s $if) {
		print "missing $if ...";
		report();
		print "Input line was:\n$_";
		return;
	    }
	    if (-s $g) {
		print "already got $g ...";
		report();
		return;
	    }
	    $ih = openToRead($if);
	    $gh = openToWrite($g);
	    my $hdr = <$ih>;
	    print $gh $hdr;
	    $ni ++;
	    print ".";
	    <$ih>; # data line that we need
	}
	print $gh $sLine;
    }
    report();
}

sub prepareImputer2 {
    my ($suf,$dir) = @_;
    my ($treeFile,$mgi) = locateTreeAndMGIFiles($suf,$dir);
    my $sor = "$dir/sorted.gz";
    my $iDir = "$dir/imp$suf";
    if (-d $iDir || -f "$iDir.touch") {
	report(sprintf("already exists: %s[.touch]",$iDir),1);
	return;
	# wait for completion...
	#for (my $n=0; $n<12; $n++) {
	#    return unless (-f "$iDir.touch");
	#    sleep 10;
	#}
    } else {
	foreach my $f ($treeFile,$sor,$mgi) {
	    if (! $f) {
		# undef is OK
	    } elsif (! -s $f) {
		report("missing or empty: $f",1);
		return;
	    } elsif (-f "$f.touch") {
		report("not finished yet: $f",1);
		return;
	    }
	}
	unless (isComplete($treeFile)) {
	    report("incomplete: $treeFile",1);
	    return;
	}
	stamp($iDir);
	createImpDirectory($suf,$sor,$mgi,$treeFile,$iDir);
	unstamp($iDir,1);
    }
}

sub locateTreeAndMGIFiles {
    my ($suf,$dir) = @_;
    if ($suf eq 'm' || $suf eq 'x') {
	return ("$dir/args_bl","$dir/mgr_in");
    } elsif ($suf eq 'g' || $suf eq 'd') {
	# Genomic extent (truly static)
	my $dd = dirname(dirname($dir));
	return ("$dd/snpMatrix/sorted.bl",undef);
    } elsif ($suf eq 'c') {
	# Chromosome, i.e. "infinite" window width
	return ("$dir/windowed_inf_bl","$dir/markers_inf")
    } elsif ($suf eq 'v') {
	return ("$dir/windowed_500_bl","$dir/markers_500")
    } elsif ($suf eq 'w') {
	return ("$dir/windowed_200_bl","$dir/markers_200")
    } elsif ($suf eq 'h') {
	return ("$dir/windowed_100_bl","$dir/markers_100")
    } elsif ($suf eq 'f') {
	return ("$dir/windowed_50_bl","$dir/markers_50")
    } elsif ($suf eq 'e') {
	return ("$dir/windowed_20_bl","$dir/markers_20")
    } elsif ($suf eq 't') {
	return ("$dir/windowed_10_bl","$dir/markers_10");
    } else {
	confess "Unexpected suffix for locateTreeAndMGIFiles: $suf";
    }
}

sub createImpDirectory {
    my ($suf,$sor,$mgi,$treeFile,$iDir) = @_;
    report("$iDir: creating");
    my $pMCounts = createImpDirectoryInputFiles($sor,$mgi,$iDir);
    report();
    report("$iDir/iNNN/args_bl files: creating");
    if ($suf eq 'g' || $suf eq 'd' || $suf eq 'c') {
	copyAndEditArgsBLLine($treeFile,$iDir);
    } else {
	writeArgsBLFiles($pMCounts,$treeFile,$iDir);
    }
    report();
}

sub createImpDirectoryInputFiles {
    my ($sor,$mgi,$iDir) = @_;
    # $mgi is undef for 'g' and 'd', and will have widely varying chunk
    # sizes for other ops (m,h,t)...
    my ($nIndivs,$pMarg) = readMarkerOffsets($sor,$mgi);
    my $fh = openToRead($sor);
    my $hdr = <$fh>; # keep newline in header
    my $idx = 0; # chunk index
    # Lines (chromosome positions) per chunk; should give 20m or so per job
    my $pChunkStarts = decideChunkStarts($nIndivs,$pMarg);
    # Offset in alignment (from 0, because that's what $mgi has)
    my $off = 0;
    # Number of lines done in current chunk
    my $nWritten = 0;
    my $gh;
    # Per-chunk marker counts
    my @mCounts;
    # Size of current chunk
    my $chunkSize;
    while (<$fh>) {
	if ($nWritten == 0) {
	    (my $trio = sprintf("i%3d",$idx)) =~ tr/ /0/;
	    # Start new imp_in file (a piece of sorted.gz)
	    $gh = openToWrite("$iDir/$trio/imp_in");
	    print '.';
	    print $gh $hdr;
	    $chunkSize = $pChunkStarts->[$idx+1] - $pChunkStarts->[$idx];
	    my $hiWater = $off+$chunkSize-1; # max offset we want for this chunk
	    push @mCounts,writeMargaritaChunkFile($nIndivs,$pMarg,$off,$hiWater,"$iDir/$trio/mgr_in");
	    $idx ++;
	}
	# Deal with this alignment line: turn hard gaps and N's into spaces; and turn zeroth-column lower case items
	# into spaces too UNLESS that would destroy all the information.
	chomp;
	my $len = length($_);
	my $hasInfo;
	for (my $i=0; $i<$len; $i+=2) {
	    my $v = substr($_,$i,1);
	    if ($v eq '=' || $v eq 'N') {
		substr($_,$i,2) = '  ';
	    } elsif (!$hasInfo && $i > 0 && substr($_,$i,1) ne ' ') {
		$hasInfo = 1;
	    }
	}
	# Safe to destroy the ref value as we have others
	if ($hasInfo && substr($_,1,1) eq '!') {
	    substr($_,0,2) = '  ';
	}
	print $gh "$_\n";
	# Increment off and nWritten
	$off ++;
	$nWritten ++;
	if ($nWritten == $chunkSize) {
	    $gh->close();
	    $nWritten = 0; # which will trigger new one being created
	}
    }
    $gh->close() if ($gh);
    return \@mCounts;
}

sub readMarkerOffsets {
    my ($sor,$mgi) = @_;
    if (defined $mgi) {
	my $fh = openToRead($mgi);
	chomp (my $line = <$fh>);
	my ($nIndivs,$nMarkers) = split(" ",$line);
	my @offs;
	foreach my $n (1 .. $nMarkers) {
	    my $line = <$fh>;
	    push @offs,$line;
	}
	chomp @offs;
	return ($nIndivs,\@offs);
    } else {
	# get nIndivs from $sor, and return no offsets
	my $fh = openToRead($sor);
	my $hdr = <$fh>;
	my @a = split(" ",$hdr);
	return (scalar(@a)+1,undef);
    }
}

# We need to ensure we get at least one marker in each file, so put
# the chunk boundaries at inter-marker positions. This also ensures
# we switch ARGs halfway between markers, as we want.

sub decideChunkStarts {
    my ($nIndivs,$pMarg) = @_;
    my $nlPerChunk = 1+int((500000-1)/($nIndivs-1));
    if ($pMarg) {
	# Try and adjust to get equal sized chunks
	my $estSz = $pMarg->[0]+$pMarg->[-1];
	my $nChunks = int(1+$estSz/$nlPerChunk);
	$nlPerChunk = int($estSz/$nChunks);
    }
    my @starts;
    my $pos = 0;
    if ($pMarg) {
	my $m = 0;
      LOOP:
	while (1) {
	    push @starts,$pos;
	    last if ($nlPerChunk == 0);
	    # Usually just advance by one chunk, but occasionally
	    # more than one, to ensure a marker inside the chunk.
	    while (1) {
		$pos += $nlPerChunk;
		last if ($pos > $pMarg->[$m]);
	    }
	    # Adjust $pos to get it halfway between two markers,
	    while ($m <= $#$pMarg) {
		last LOOP if ($m == $#$pMarg);
		if ($pMarg->[$m+1] > $pos) {
		    $pos = int(($pMarg->[$m]+$pMarg->[$m+1])/2);
		    $m++;
		    last;
		}
		$m++;
	    }
	}
	push @starts,$starts[-1]+10000000; # something really big just in case
   } else {
       # Create more chunks than we'll need; OK, as we'll run out of sorted.gz lines
       while ($pos < 10000000) {
	   push @starts,$pos;
	   $pos += $nlPerChunk;
       }
    }
    return \@starts;
}

sub writeMargaritaChunkFile {
    my ($nIndivs,$pMarg,$delta,$hiWater,$file) = @_;
    my $mh = openToWrite($file);
    my $mCount = 0;
    while ($mCount <= $#$pMarg && $pMarg->[$mCount] <= $hiWater) {
	$mCount ++;
    }
    if ($mCount == 0) {
	# Keep imputer happy with one marker
	print $mh "$nIndivs 1\n0\n";
	return 0;
    } else {
	print $mh "$nIndivs $mCount\n";
	my $mc = $mCount;
	while ($mCount > 0) {
	    printf($mh "%d\n",(shift @$pMarg) - $delta);
	    $mCount --;
	}
	$mh->close();
	return $mc; # No. of markers that went into this file
    }
}

sub copyAndEditArgsBLLine {
    my ($treeFile,$iDir) = @_;
    my $fh = openToRead($treeFile);
    my $line = <$fh>;
    chomp $line;
    my @a = split(" ",$line);
    # Prevent zero branch lengths:
    for (my $i=4; $i<=$#a; $i+=2) {
	if ($a[$i] == 0) {
	    $a[$i] = 0.000001;
	}
    }
    foreach my $d (glob("$iDir/i???")) {
	# Get $nm from mgr_in file:
	my $mh = openToRead("$d/mgr_in");
	my $mLine = <$mh>;
	chomp $mLine;
	my (undef,$nm) = split(" ",$mLine);
	# Write args_bl line
	my $gh = openToWrite("$d/args_bl");
	$a[2] = max($nm-1,0); # max, to make sure it works with no markers
	print $gh "@a\n";
    }
}

sub writeArgsBLFiles {
    my ($pMCounts,$src,$dir) = @_;
    my $fh = openToRead($src);
    my @out; # output stream for each chunk
    my @delta = (0); # marker-offset-difference for each chunk AND ONE MORE off end
    my $del = 0;
    foreach my $c (@$pMCounts) {
	$del += $c;
	push @delta,$del;
    }
    my $chMin = 0; # idx of lowest chunk still active
    while (<$fh>) {
	chomp;
	my ($mLo,$argID,$mHi,@rest) = split;
	while ($mLo >= $delta[$chMin+1] && $out[$chMin]) {
	    $out[$chMin]->close();
	    $chMin ++;
	}
	my $ch = $chMin;
	while ($delta[$ch] <= $mHi) {
	    my $mLoRel = max(0,$mLo-$delta[$ch]);
	    my $mHiRel = max($mLoRel,min($delta[$ch+1]-1,$mHi)-$delta[$ch]);
	    if ($ch > $#out) {
		$out[$ch] = createArgsBLFile($dir,$ch);
		print '.';
	    }
	    my $gh = $out[$ch];
            # Prevent zero branch lengths:
	    for (my $i=1; $i<=$#rest; $i+=2) {
		if ($rest[$i] == 0) {
		    $rest[$i] = 0.000001;
		}
	    }
	    printf($gh "%6d %6d %6d %s\n",$mLoRel,$argID,$mHiRel,join(" ",@rest));
	    $ch ++;
	}
    }
}

sub createArgsBLFile {
    my ($dir,$n) = @_;
    (my $trio = sprintf("i%3d",$n)) =~ tr/ /0/;
    my $g = "$dir/$trio/args_bl";
    return openToPipe("sort > $g");
}

sub spreadPresenceConstraintsOriginal {
    my ($pBuf) = @_;
    my $lenC = 0;
    foreach my $line (@$pBuf) {
	$lenC ++ if ($line =~ /^C/);
    }
    return if ($lenC <= 1); # nothing to do
    my $iMax = length($pBuf->[0])-2;
    my $th = $lenC*0.01;
    for (my $i=0; $i<=$iMax; $i+=2) {
	my $sum = 0;
	foreach my $line (@$pBuf) {
	    if ($line =~ /^C/ && substr($line,$i,1) eq 'C') {
		$sum ++;
		if ($sum >= $th) {
		    foreach my $line2 (@$pBuf) {
			substr($line2,$i,1) =~ tr/ /C/ if ($line2 =~ /^C/);
		    }
		    last;
		}
	    }
	}
    }
}

sub createACFile2 {
    my ($f,$regFile,$g,$regFile2) = @_;
    # Decide species name ... easier to do it here than pass it
    # into all calls
    my $sp = ($f =~ /cere/) ? 'cere' : 'para';
    my $fh = openToRead($regFile);
    my $gh = openToWrite($regFile2);
    my (@low,@high,@mat,@pats,$ns);
    report("$g: creating");
    while (<$fh>) {
	chomp;
	my ($lo,$hi,$pat) = split(" ",$_,3);
	push @pats,$pat;
	my @pres;
	my $iMax = length($pat)-2;
	for (my $i=2; $i<=$iMax; $i+=2) {
	    push @pres,(substr($pat,$i,1) =~ /[AC]/ ? 1 : 0);
	}
	push @low,$lo;
	push @high,$hi;
	push @mat,\@pres;
	$ns = max($ns,scalar(@pres));
    }
    # Find median gap lengths
    my @medLen;
    my $sumMedLen;
    foreach my $j (0 .. $ns-1) {
	my $gapLen = 0;
	my @gapLens;
	foreach my $i (0 .. $#mat) {
	    if (!$mat[$i][$j]) {
		$gapLen += $high[$i]-$low[$i]+1;
	    } elsif ($gapLen) {
		push @gapLens,$gapLen;
	    }
	}
	push @gapLens,$gapLen if ($gapLen);
	@gapLens = sort {$a <=> $b} @gapLens;
	$medLen[$j] = $gapLens[int($#gapLens/2)];
	$sumMedLen += $medLen[$j];
    }
    my $avMedLen = $sumMedLen/$ns;
    my @probDiv; # prob of divergence, indexed on region
    foreach my $i (0 .. $#mat) {
	my (@nBefore,@nAfter);
	foreach my $j (0 .. $ns) {
	    unless ($mat[$i][$j]) {
		$nBefore[$j] = countAbsences(\@mat,\@low,$i,$j,-1);
		$nAfter[$j] = countAbsences(\@mat,\@high,$i,$j,1);
	    }
	}
	my $sz = $high[$i]-$low[$i]+1;
	my $pat = $pats[$i];
	my $uniqueAC = findUniqueAC($pat);
	foreach my $j (0 .. $ns-1) {
	    unless ($mat[$i][$j]) {
		my $a = maximumArea(\@nBefore,\@nAfter,$j,$sz);
		# Adjust area using median gap length
		$a *= $avMedLen/$medLen[$j];
		my $p = areaToDivergenceProb($a,$sp);
		if ($p) {
		    $probDiv[$i][$j] = $p;
		    my $col = 2*($j+1);
		    if ($p >= 0.5) {
			substr($pat,$col,1) = 'T';
			my $qc = errorProbToQualChar(1-$p);
			unless ($qc eq ' ') {
			    substr($pat,$col+1,1) = $qc;
			}
		    } elsif (defined $uniqueAC) {
			# We'd like to do this even if both A and C
			# are present, but we can't say "A or C"
			# as a value ...
			substr($pat,$col,1) = $uniqueAC;
			my $qc = errorProbToQualChar($p);
			unless ($qc eq ' ') {
			    substr($pat,$col+1,1) = $qc;
			}
		    }
		}
	    }
	}
	$pats[$i] = $pat;
	print $gh "$low[$i] $high[$i] $pat\n";
    }
    # Now apply provDivs to AC data itself, to create sorted_ac.gz.
    my $fh = openToRead($f);
    my $gh = openToWrite($g);
    my $off = 0;
    my $idx = 0;
    my $pat = $pats[$idx];
    while (<$fh>) {
	unless (/^>/) {
	    $off ++;
	    if ($high[$idx] < $off) {
		$idx ++;
	    }
	    $_ = $pats[$idx] . "\n";
	}
	print $gh $_;
    }
    report();
}

sub findUniqueAC {
    my ($pat) = @_;
    if ($pat =~ /A/) {
	if ($pat =~ /C/) {
	    return undef;
	} else {
	    return 'A';
	}
    } else {
	return 'C';
    }
}

sub errorProbToQualChar {
    my ($ep) = @_;
    return ' ' if ($ep <= 0);
    my $q = -log($ep)*10/log(10);
    $q = int($q+$ZERO);
    return $q >= 127 ? ' ' : chr($q);
}

# See estimateDivergenceProbabilities.pl for reasoning here ...
# undef means zero.

sub areaToDivergenceProb {
    my ($a,$sp) = @_;
    my $p;
    if ($sp eq 'cere') {
	if ($a <= 2000) {
	    return undef;
	} elsif ($a <= 21000) {
	    $p=0.2*($a-2000)/19000;
	} elsif ($a <= 26000) {
	    $p=0.95;
	} else {
	    $p=1;
	}
    } else {
	if ($a <= 1000) {
	    return undef;
	} elsif ($a <= 15000) {
	    $p=$a/15000;
	} else {
	    $p=1;
	}
    }
    # Allow imputation to have a say ...
    return $p*0.99;
}

sub countAbsences {
    my ($pMat,$pOff,$i,$j,$di) = @_;
    my $k = $i+$di;
    while (1) {
	last if ($k < 0 || $k > $#$pMat || $pMat->[$k][$j]);
	$k += $di;
    }
    $k -= $di;
    my $len = $pOff->[$k]-$pOff->[$i];
    return $len*$di;
}
	
sub maximumArea {
    my ($pBef,$pAft,$j,$sz) = @_;
    # Try various lBef and lAft thresholds;
    # 0 < lBef <= $pBef->[$j] and 0 < lAft <= $pAft->[$j] to qualify.
    my %isBef = (0,1);
    my $nBJ = $pBef->[$j];
    foreach my $v (@$pBef) {
	$isBef{$v} = 1 if ($v > 0 && $v <= $nBJ);
    }
    my %isAft = (0,1);
    my $nAJ = $pAft->[$j];
    foreach my $v (@$pAft) {
	$isAft{$v} = 1 if ($v > 0 && $v <= $nAJ);
    }
    my ($maxArea,$bestB,$bestA);
    foreach my $b (keys %isBef) {
	foreach my $a (keys %isAft) {
	    my $qual;
	    foreach my $j (0 .. $#$pBef) {
		$qual ++ if ($pBef->[$j] >= $b && $pAft->[$j] >= $a);
	    }
	    my $area = $qual*($a+$b+$sz);
	    if ($area > $maxArea) {
		$maxArea = $area;
		$bestB = $b;
		$bestA = $a;
	    }
	}
    }
    return $maxArea;
}


1;
