package SGRP::PrepareMargarita;

use Exporter;
@ISA = qw(Exporter);

# prepareMargaritaInput2 is for standalone use

@EXPORT = qw(prepareMargaritaInput prepareMargaritaInput2);

use SGRP::Variables;
use SGRP::SNPMatrix; # in case we need to call addToSNPMatrix
use Utility;
use Carp;

use strict;

##
## ---------------------------------------------------------------------
## sgrp.pl pmar cere methc CERE 0 1 2 ...
## ---------------------------------------------------------------------
##
## prepareMargaritaInput
##   # SRC: snp_rates.txt [from esnr] -- no longer used
##   -> prepareMargaritaInput1
##        -> prepareMargaritaInput2
##           SRC: .../chrNN/sorted.gz
##           TGT: .../chrNN/mgr_in      [Margarita input; used by marg]

# For quality conversions:
my $ZERO = ord('!');

# Whether we try to do adjustment for relatedness ...
my $DO_ADJUSTMENT = 0;

## We only give SNPs to Margarita when (1) the sum of quality scores for
## values that differ from the reference is >= 40 (this saves time;
## if it's <40 Margarita should always decide they're sequencing errors),
## AND (2) they are not indel-internal.

# Sum-of-qualities threshold for handing to Margarita. Should be
# _independent_ of SNP rate.
my $SOQ_THRESHOLD = 40;

# Try to weed out poor-quality apparently-polymorphic positions.
#
# On 200-individual Fregene data:
#
#      QP     FN rate    FP rate
#      10      2.98%      4.18%
#      15      4.55%      0.81%
#      20      6.38%      0.16%
#      25      8.96%      0.03%
#
# For 35-individual Fregene data:
#
#      QP     FN rate    FP rate
#       5      1.78%      4.97%
#      10      4.98%      0.74%
#      15      7.63%      0.18%
#      20     10.70%      0.05%
#
# For 12 individuals:
#
#      QP     FN rate    FP rate
#       5      3.76%      3.56%
#      10      8.00%      0.37%
#      15     11.89%      0.08%
#      20     16.27%      0.00%
#
# Since FP's are more serious (they will mislead Margarita), we choose QP=20,
# corresponding to P(error) = 0.01, for 200 indivs, and QP=15 for 35 and 12.
# See setQualityPenalty below.

my $QUALITY_PENALTY;

sub prepareMargaritaInput {
    my ($species,$filtMeth,$strainSet,@parts) = @_;
    unless (@parts) {
	confess("Usage: $0 pmar species filtMeth strainSet part1 part2 ...");
    }
    # Try to reflect SNP rate in the data in question ...
    if ($strainSet =~ /FREG/) {
	$SOQ_THRESHOLD = 40;
    } else {
	$SOQ_THRESHOLD = 25;
    }
    my ($iter,$dir) = getCurrentIteration($species,$filtMeth,$strainSet,'p?');
    foreach my $part (@parts) {
	$part = "p$part" if ($part =~ /^[0-9]$/);
	foreach my $chrDir (glob("$dir/$part/*")) {
	    # Use a chr-specific one now ...
	    my $pSNPRate;
	    # $pSNPRate = readSNPRates("$chrDir/snp_rates.txt");
	    prepareMargaritaInput1($chrDir,$pSNPRate);
	}
    }
}

sub prepareMargaritaInput1 {
    my ($dir,$pSNPRate) = @_;
    my $nm = prepareMargaritaInput2("$dir/sorted.gz","$dir/mgr_in",$pSNPRate);
    #prepareMargaritaInput2("$dir/adjusted.gz","$dir/mgr_in_adj",$pSNPRate);
    if (defined $nm && $nm == 0) {
	# Then we should call smat
	addToSNPMatrix($dir,1);
    }
}

# Return no. of markers if mgr_in is created, or -1 if not.
sub prepareMargaritaInput2 {
    my ($src,$mgrInFile,$pSNPRate) = @_;
    return -1 unless (isComplete($src,1));
    if (stamp($mgrInFile)) {
	report("creating $mgrInFile");
	my $nm = prepareMargaritaInput3($src,$pSNPRate,$mgrInFile);
	report();
	unstamp($mgrInFile);
	return $nm;
    } else {
	return -1;
    }
}

sub prepareMargaritaInput3 {
    my ($srcFile,$pSNPRate,$mjmFile) = @_;
    return unless (isComplete($srcFile,1));
    my $fh = openToRead($srcFile);
    # Store the lines
    my @lines = <$fh>;
    chomp @lines;
    my @strains = split(/ /,substr($lines[0],1));
    setQualityPenalty(scalar(@strains));
    my ($maxRate,@rate);
    if ($pSNPRate) {
	foreach my $k (0 .. $#strains) {
	    my $r = $rate[$k+1] = $pSNPRate->{$strains[$k]};
	    $maxRate = $r if ($maxRate < $r);
	}
	# Normalize rates so adjustment has right properties
	if ($maxRate) {
	    foreach my $k (1 .. $#rate) {
		$rate[$k] /= $maxRate;
	    }
	}
    }
    my $ns = scalar(@strains)+1; # +1 for reference
    # Find the polymorphic offsets.
    my (@polyOffs,@polyLines);
    my $pRateToUse = $maxRate ? \@rate : undef;
    foreach my $off (1 .. $#lines) {
	my $line = hideHardGaps($lines[$off]);
	if (effectivelyPolymorphicLine(\@lines,$off,$pRateToUse)) {
	    push @polyOffs,$off;
	    push @polyLines,$line;
	} 
	#$off ++;
    }
    my $gh = openToWrite($mjmFile);
    printf($gh "%d %d\n",$ns,scalar(@polyOffs));
    # Print the positions of the polymorphisms
    foreach my $off (@polyOffs) {
	print $gh "$off\n";
    }
    # Print the data, ensuring each line is the right length (extend
    # with spaces where needed), and adjusting probabilities to
    # reflect SNP rates (all these lines are polymorphic).
    my $tgtLen = $ns*2;
    foreach my $k (0 .. $#polyOffs) {
	my $off = $polyOffs[$k];
	my $line = $polyLines[$k];
	my $len = length($line);
	if ($len > $tgtLen) {
	    confess("Line is length $len, max is $tgtLen: '$line'");
	} elsif ($len < $tgtLen) {
	    $line .= ' ' x ($tgtLen-$len);
	}
	print $gh "$line\n";
    }
    $gh->close();
    return scalar(@polyOffs); # number of markers
}

sub setQualityPenalty {
    my ($ns) = @_;
    if ($ns <= 35) {
	$QUALITY_PENALTY = 15;
    } elsif ($ns >= 200) {
	$QUALITY_PENALTY = 20;
    } else {
	$QUALITY_PENALTY = int($ns/33)+14;
    }
}

# For input to Margarita, treat hard gaps as if they were missing.
# Then we will impute (for all strains) what _would_ be the value
# if there was something there. Just need to make sure we don't
# overwrite hard gaps on output.

sub hideHardGaps {
    my ($line) = @_;
    my $len = length($line);
    for (my $i=0; $i<$len; $i+= 2) {
	if (substr($line,$i,1) eq '=') {
	    substr($line,$i,2) = '  ';
	}
    }
    return $line;
}

# $pLines: all the lines in the sorted.gz file
# $off: offset we're looking at
# $pRate: SNP rate (wrt reference) for each strain, normalized so maximum is 1

sub effectivelyPolymorphicLine {
    my ($pLines,$off,$pRate) = @_;
    my $line = $pLines->[$off];
    my $ref = substr($line,0,1);
    return 0 if ($ref eq 'N' || $ref eq 'n');
    my $len = length($line);
    # Context:
    my ($prev,$next);
    $prev = $pLines->[$off-1] if ($off > 1);
    $next = $pLines->[$off+1] if ($off < $#$pLines);
    # Each SNP value has a list of its k-values (offsets of the nucleotide
    # in the line):
    my %snpKList;
    # Sum of qualities of SNP positions, subtracting $QUALITY_PENALTY from each.
    my %sumQ;
    # Allele frequency; minor (freq of second-placed) needs to be >1 to be
    # worth handing to Margarita.
    my %af;
    for (my $k=0; $k<$len; $k+= 2) {
	my $c = substr($line,$k,1);
	if ($c ne ' ' && $c ne 'N' && $c ne 'n') {
	    $c =~ tr/acgt_/ACGT-/; # "_" and "-" shouldn't occur, though ...
	    unless (isInternalIndel($ref,$c,$k,$prev,$next)) {
		# SNP to $c at position $k.
		push @{$snpKList{$c}},$k;
		my $qc = substr($line,$k+1,1);
		my $q = (($qc eq ' ') ? 127 : ord($qc)) - $ZERO;
		my $effectiveQ = $q - $QUALITY_PENALTY;
		if ($effectiveQ > 0) {
		    $sumQ{$c} += $effectiveQ;
		    $af{$c} ++;
		}
	    }
	}
    }
    # If we have just one SNP and it's got quality less than $SOQ_THRESHOLD, it'll almost certainly
    # be treated as a read error, so we won't hand it to Margarita at all (saves time).
    # And the same if we have several whose sum of q scores is less than $SOQ_THRESHOLD. If we
    # only have 1 SNP position, even if very strong, it won't affect the ARGs so no point handing it on.
    if (%snpKList) {
	my @sorted = sort {$af{$b} <=> $af{$a}} keys %af;
	my $minor = $sorted[1]; # second-placed value
	if ($minor && $af{$minor} > 1 && $sumQ{$minor} >= $SOQ_THRESHOLD) {
	    return 1;
	}
    }
    return 0;
}

sub isInternalIndel {
    my ($ref,$c,$k,$prev,$next) = @_;
    return 0 unless ($prev && $next);
    if ($c > 0 || $c eq '=') {
	# $c is the indel
	return (substr($prev,$k,1) eq $c
		&& substr($next,$k,1) eq $c);
    } elsif ($ref > 0 || $ref eq '=') {
	# $ref is the indel
	return (substr($prev,0,1) eq $ref
		&& substr($next,0,1) eq $ref);
    }
    return 0;
}

sub adjustForSNPRate {
    my ($val,$q,$ref,$rate) = @_;
    my $eq = exp(-log(10)*$q/10);
    $eq = 0.75 if ($eq > 0.75);
    my $adjEQ = $eq/($rate*(1-$eq)+$eq);
    if ($adjEQ > 0.5) {
	$val = $ref;
	$adjEQ = 1-$adjEQ;
    }
    my $code = $ZERO+int(-10*log($adjEQ)/log(10));
    my $qc = ($code > 127) ? ' ' : chr($code);
    return "$val$qc";
}

sub readSNPRates {
    my ($rateFile) = @_;
    my %snpRate;
    if (-s $rateFile) {
	fileToHash($rateFile,\%snpRate);
    } else {
	confess("Missing or empty: $rateFile");
    }
    return \%snpRate;
}

1;
