=head1 NAME

SGRP::PruneAlignments::pruneAlignments

=head1 SYNOPSIS

pruneAlignments species method str1 str2 ...

=head1 DESCRIPTION

For each strain str, reads $ROOT/read/str/species/alilines.txt and
clusters the alignments of each read into
$ROOT/read/str/species/methc/clusters.txt, putting the accepted reads
in $ROOT/read/str/species/methc/alilines.txt.

Allowed values of "method" are "methc" and "methca". In the latter case only,
"ambiguous" and "manual" alignments are accepted (this is done for Leo).

In the clusters file, per-read outputs are separated by
newlines. Clusters are named c0, c1 ...  with c0 being the top
scorer. Shortfall is min(score(c0)-score(ci)) over ci's scoring
_strictly_ less than c0; we assume identical score means identical
regions, e.g. rDNA.

Cluster is printed as header line then a line for each alignment. Header
is e.g.

Y55-31o13 c0 (45,0) chr02:494..4210 (3717) TM 990 IDX 99 SF 0.67

with these meanings:

Y55-31o13           Clone name
c0                  Cluster ID
(45,0)              Cluster score and shortfall
chr02:494..4210     Chr region covered by cluster
(3717)              Length of that region
TM 990              No. of positions at which cluster "touches maximum" 
                    (see below)
IDX 99              Index (from 1) in this file, i.e. 99th most ambiguous clone
SF 0.67             Shortfall for this clone.

IDX and SF are only given for c0 as they apply to the whole clone.

A cluster "touches maximum" at all positions in the clone that are spanned
by that cluster and not by any higher-scoring cluster. 

Alignments in a cluster look like this:

  F p   38 < 701 :    1  798 :     494    1284 :  798  96.99  798
  C q    6 < 110 :  649  458 :    4033    4210 :  192  89.58  697

with these meanings:

F             Forward alignment (C is complement)
p             p or q read
38 < 701      Shared score and raw score (see below)
1 798         Span of read scanned by alignment (1st >= 2nd for C alignment)   
494 1284      Span of chr spanned by alignment (1st always <= 2nd)
798           Alignment length (aLen) in read
96.99         Percent identity
798           Length of read

The "shared score" of an alignment is the raw (ssaha) score * rSum / aLen,
where rSum = sum over all positions i of alignment of 1/Ni, where Ni alignments
cover position i. This penalizes repeats/transposons and means the top-scoring
cluster c0 is more likely to be the "anchor" location (beside which the
transposon has inserted).

The score of a cluster is the sum of the shared scores of its
Salignments, minus 20 for every alignment after the first in a single
read. So if a cluster has p alignments scoring 50,60,70 and one q
alignment scoring 40, cluster score is 50+60+70+80-2*20.

=head1 AUTHOR

David Carter, C<dmc@sanger.ac.uk>

=head1 SEE ALSO

       

=cut

package SGRP::PruneAlignments;

use Exporter;
@ISA = qw(Exporter);

@EXPORT = qw(pruneAlignments);

use SGRP::Variables;
use SGRP::SGRP;
use SGRP::Alignment;
use SGRP::AlignmentCluster;
use SGRP::SelectNOAs;
use SGRP::Filter;
use SGRP::OffsetMap;
use SGRP::SequencedRegionSet;

use Utility;
use File::Basename;
use Carp;

use strict;

sub pruneAlignments {
    my ($species,$meth,$strainSet,@strains) = @_;
    unless (@strains) {
	die("Usage: $0 alip species filtMeth strainSet strain1 strain2 ...");
    }
    my ($iter,$iDir) = getCurrentIteration($species,$meth,$strainSet,'align'); 
    ensureSsahaSNPDone($iter,$iDir,$species,$meth,$strainSet,@strains);
    my $pRegs = readDifficultRegions($species,$iter,$iDir);
    foreach my $str (@strains) {
	# Dir to get ssahaSNP alignments (from ssnp command) from
	my $srcDir = $iDir ? "$iDir/align/$str" : "$ROOT/read/$str/$species";
	if (-d $srcDir) {
	    # Dir to get renamed.wm and renamed.fastq files from
	    my $rnDir = "$ROOT/read/$str";
	    # Possible dir for genome.{map,seqr} files
	    my $genDir = $iDir ? "$iDir/assembly/$str" : undef;
	    # Dir to put pruned stuff in
	    my $tgtDir = $iDir ? "$iDir/align/$str" : "$srcDir/$meth";
	    my ($pOM,$alPass) = pruneAlignments1($species,$meth,$srcDir,$rnDir,
						 $genDir,$tgtDir,$pRegs);
	    if ($alPass) {
		# We only insist on "nice" alignments for the 'FINAL'
		# iteration, not the 'EXTRA' one.
		filterForFinalIteration($alPass) if ($IS_FINAL_ITERATION eq 'FINAL');
		pruneAlignments2($alPass,"$rnDir/renamed.fastq",
				 "$srcDir/ssahaSNP.all.gz",$tgtDir,$pOM);
	    }
	} else {
	    die("no directory $srcDir");
	}
    }
}

sub ensureSsahaSNPDone {
    my ($iter,$iDir,$species,$meth,$strainSet,@strains) = @_;
    foreach my $str (@strains) {
	my $srcDir = $iDir ? "$iDir/align/$str" : "$ROOT/read/$str/$species";
	my $src = "$srcDir/ssahaSNP.all.gz";
	my $srcAL = "$srcDir/alilines.all.txt";
	# Catch up if needed:
	unless (-s $src) {
	    foreach my $p (0 .. 9) {
		SGRP::SGRP::sgrpCommand({},'ssnp',$species,$meth,$strainSet,$p,@strains);
	      }
	}
	unless (-s $srcAL) {
	    createAlilinesFile($src,$srcAL);
	}
    }
}

sub createAlilinesFile {
    my ($src,$tgt) = @_;
    my $fh = openToRead($src);
    my $gh = openToWrite($tgt);
    while (<$fh>) {
        print $gh $_ if (/^ALIGNMENT/);
    }
    $fh->close();
    $gh->close();
}

sub readDifficultRegions {
    my ($sp,$iter,$iDir) = @_;
    my %regs;
    my $f;
    my $pOM;
    if ($iter > 0) {
	# Best, because it will contain transposon copies:
	$f = "$iDir/assembly/genome.gff";
    } else {
	$f = "$ROOT/ref/$sp/extended.gff";
	unless (-s $f) {
	    $f = "$ROOT/ref/$sp/genome.gff";
	    my $mapFile = chooseMapFile($iDir,$sp);
	    $pOM = new OffsetMap('?',$mapFile); # will be null if mapFile missing
	}
    }
    # genome.gff etc doesn't have to exist ...
    if (-s $f) {
	my $fh = openToRead($f);
	while (<$fh>) {
	    chomp;
	    my ($chr,undef,$type,$lo,$hi) = split;
	    $chr = 'chrm' if ($chr eq 'chrM');
	    $lo = $pOM->translation("EXT.$chr",$lo) if ($pOM);
	    $hi = $pOM->translation("EXT.$chr",$hi) if ($pOM);
	    if (!defined $lo || !defined $hi) {
		# nothing
	    } elsif ($type eq 'chromosome' && $chr ne 'chrm') {
		push @{$regs{$chr}},['subtelomere',$lo,$lo+29999];
		push @{$regs{$chr}},['subtelomere',$hi-29999,$hi];
	    } elsif ($type eq 'centromere') {
		# For now, just the (120bp or so) centromere itself:
		push @{$regs{$chr}},[$type,$lo,$hi];
	    } elsif ($type eq 'transposable_element') {
		push @{$regs{$chr}},[$type,$lo,$hi];
	    }
	}
	foreach my $chr (keys %regs) {
	    my @a = sort {$a->[1] <=> $b->[1] || $a->[2] <=> $b->[2]} @{$regs{$chr}};
	    $regs{$chr} = \@a;
	}
    }
    return \%regs;
}

sub chooseMapFile {
    my ($iDir,$sp) = @_;
    return "$iDir/assembly/REF/genome.map" if (defined $iDir);
    foreach my $b (qw(extended augmented)) {
	my $f = "$ROOT/ref/$sp/$b.map";
	return $f if (-s $f);
    }
    die("chooseMapFile: failed for iDir=$iDir, sp=$sp");
}

sub pruneAlignments1 {
    my ($species,$meth,$srcDir,$rnDir,$genDir,$tgtDir,$pRegs) = @_;
    # Possible genome.map file
    my $mapFile = $genDir ? "$genDir/genome.map" : chooseMapFile(undef,$species);
    # Possible genome.seqr file
    my $seqrFile = $genDir ? "$genDir/genome.seqr" : undef;
    # Possible genome.dangles file
    my $dangFile = $genDir ? "$genDir/genome.dangles" : undef;
    # We need mainToo=1 so that we can use $pSRS which needs level 2
    my $pOM = new OffsetMap('?',$mapFile); # may be null
    my $pSRS = new SequencedRegionSet($seqrFile); # may also be null
    my $pDang = readDangles($dangFile) if ($dangFile);
    return ($pOM,clusterAlignments($meth,$rnDir,$srcDir,
				   $pOM,$pSRS,$pDang,$tgtDir,$pRegs));
}

sub readDangles {
    my ($f) = @_;
    my $fh = openToRead($f);
    my %dang;
    while (<$fh>) {
	chomp;
	unless (/^#/) {
	    my ($chr,$off,$lr) = split;
	    push @{$dang{$chr}{$lr}},$off;
	}
    }
    return \%dang;
}

sub clusterAlignments {
    my ($meth,$wmDir,$srcDir,$pOM,$pSRS,$pDang,$tgtDir,$pRegs) = @_;

    my $f = "$srcDir/alilines.all.txt";
    unless (-f $f) {
	report("missing: $f",1);
	return;
    }
    my $g = "$tgtDir/clusters.txt";
    my $h1 = "$tgtDir/alilines.pass.txt";
    my $h2 = "$tgtDir/alilines.all.status.txt";
    my $bad;
    foreach my $gh ($g,$h1,$h2) {
	if (-s $gh) {
	    report("already exists: $gh",1);
	    $bad = 1;
	}
    }
    return $h1 if ($bad);
    report("reading $f");
    my $gh = *STDOUT;
    my $hh1;
    my $hh2 = *STDOUT;
    if ($g) {
	$gh = openToWrite($g);
	$hh1 = openToWrite($h1);
	$hh2 = openToWrite($h2);
    }
    # We only check that $f has any contents here, after we've
    # created alilines.pass.txt etc. We need these to exist (even if
    # empty) for merg to work.
    return unless (-s $f);
    my $pAlis = readAlignments($f,$pOM,$hh2);
    report();
    report("setting shared scores");
    setSharedScores($pAlis);
    report();

    # Get watermarks
    report("reading watermarks");
    my $pWM = readWatermarksFromFile("$wmDir/renamed.wm");
    report();

    report("finding clusters");
    my (%clusters,%shortfall,%reported);
    foreach my $clone (sort keys %$pAlis) {
	my $pAC = $pAlis->{$clone};
	my (@cList,$threshold);
	foreach my $chr (sort keys %$pAC) {
	    push @cList,clusterAlignments1($clone,$chr,$pWM,$pDang,$pAC->{$chr});
	}
	if (@cList) {
	    my @sorted = sort {$b->score <=> $a->score} @cList;
	    $clusters{$clone} = \@sorted;
	    $shortfall{$clone} = findShortfall(\@sorted);
	}
    }
    report();

    # Evaluate the clusters for each clone, i.e. assign a status
    # to each cluster (accept, reject_close etc)
    report("evaluating clusters");
    foreach my $clone (sort keys %clusters) {
	my $pMas = maximumAvailableScore($clone,$pWM);
	evaluateClusters($clusters{$clone},$pMas,$pRegs);
	$shortfall{$clone} = findShortfall($clusters{$clone});
    }
    # Reject clusters with a break in a sequenced region, or that score
    # much less than such a cluster.
    rejectBISRClusters(\%clusters,$pSRS);
    # Reject accepted clusters involving breaks spanned by reads accepted
    # in THIS iteration (23-1-2007)
    rejectClustersWithUnlikelyBreaks(\%clusters);
    # Reject chimeric clusters for same clone that overlap
    rejectClashingChimericClusters(\%clusters);
    report();
    # Print the clusters
    if ($g) {
	report("writing files in $tgtDir");
    }
    my $idx=0;
    foreach my $clone (sort keys %clusters) {
	printAlignments($pWM,$clusters{$clone},$shortfall{$clone},
			++$idx,$gh,$hh1,$hh2,$meth);
    }
    if ($g) {
	$gh->close();
	$hh1->close();
	$hh2->close();
	report();
    }
    return $h1;
}

sub readAlignments {
    my ($f,$pOM,$lh) = @_;
    my $fh = openToRead($f);
    my %alis;
    my @prevs;
    while (<$fh>) {
	#print;
	chomp;
	# Will be null if it's redundant (fails to span break)
	my $pA = new Alignment($_,$pOM);
	if (!$pA) {
	    printf($lh "$_ reject_ftsb\n");
	} elsif (equivalentAlignments($pA,\@prevs)) {
	    printf($lh "$_ reject_equiv\n");
	} else {
	    # report to $lh when clustered etc ...
	    push @{$alis{$pA->clone}{$pA->chr(1)}},$pA;
	    # remember this alignment as a previous one (for this
	    # clone); scores always decrease, so we can use score
	    # as a criterion for throwing away previous @prevs
	    if (@prevs 
		&& $prevs[0]->clone eq $pA->clone
		&& $prevs[0]->score eq $pA->score) {
		push @prevs,$pA;
	    } else {
		@prevs = ($pA);
	    }
	}
    }
    $fh->close();
    return \%alis;
}

sub equivalentAlignments {
    my ($pA1,$pAPrevs) = @_;
    foreach my $pA2 (@$pAPrevs) {
	return 1 if 
	    $pA1->score == $pA2->score &&
	    $pA1->name eq $pA2->name &&
	    $pA1->rlo == $pA2->rlo &&
	    $pA1->rhi == $pA2->rhi &&
	    $pA1->chr(1) eq $pA2->chr(1) &&
	    $pA1->clo(1) == $pA2->clo(1) &&
	    $pA1->chi(1) == $pA2->chi(1);
    }
    return 0;
}
	

sub setSharedScores {
    my ($pAlis) = @_;
    foreach my $pH (values %$pAlis) {
	setSharedScores1($pH);
    }
}

sub setSharedScores1 {
    my ($pH) = @_;
    my %depth = (p => [], q => []);
    my %seen;
    foreach my $pA (values %$pH) {
	foreach my $aa (@$pA) {
	    # Only count apparently identical alignments (to exact repeats
	    # in genome) once
	    my $key = sprintf("%d %d %d %s",$aa->rlo,$aa->rhi,$aa->score,$aa->pq);
	    unless ($seen{$key}) {
		$seen{$key} = 1;
		my $pD = $depth{$aa->pq};
		foreach my $k ($aa->rlo .. $aa->rhi) {
		    $pD->[$k] ++;
		}
	    }
	}
    }
    foreach my $pA (values %$pH) {
	foreach my $aa (@$pA) {
	    my $pD = $depth{$aa->pq};
	    my $sum=0;
	    foreach my $k ($aa->rlo .. $aa->rhi) {
		unless ($pD->[$k]) {
		    1;
		}
		$sum += 1/$pD->[$k];
	    }
	    $aa->sharedScore($aa->score*$sum/($aa->rhi-$aa->rlo+1));
	}
    }
}    

sub readWatermarksFromFile {
    my ($f) = @_;
    my $fh = openToRead($f);
    my ($pWM,undef) = readWatermarks($fh);
    return $pWM;
}

sub readWatermarks {
    my ($fh) = @_;
    my %wm;
    my %len;
    while (<$fh>) {
        chomp;
        my ($read,$l,@nums) = split;
        my @ans;
        for (my $k=0; $k<=$#nums; $k+=2) {
            push @ans,[$nums[$k],$nums[$k+1]];
        }
        if ($read) { # $pCorr->{$read} might be present but value undef ...
            $wm{$read} = \@ans;
            $len{$read} = $l;
        }
    }
    return (\%wm,\%len);
}

sub clusterAlignments1 {
    my ($clone,$chr,$pWM,$pDang,$pAC) = @_;

    my @cList = ();
    # This seems (after some trial and error) to be the best
    # order to make sensible clusters without taking the 2^N
    # route of creating all possible clusters.
    my $pSorted = sortByShortfallThenPosition($pAC);
    foreach my $ali (@$pSorted) {
	my $done = 0;
	foreach my $c (@cList) {
	    #last if ($done = $c->add($ali));
	    # Try to add to ALL clusters; so one alignment can be
	    # in several.
	    $done = 1 if ($c->add($ali));
	}
	unless ($done) {
	    push @cList,new AlignmentCluster($ali);
	}
    }
    foreach my $c (@cList) {
	$c->saveWatermarks($pWM);
	$c->adjustScoreForDangles($pDang);
    }
    return @cList;
}

sub sortByShortfallThenPosition {
    my ($pAlis) = @_;
    my %max;
    foreach my $pA (@$pAlis) {
	my $pq = $pA->pq;
	$max{$pq} = max($max{$pq},$pA->score);
    }
    my @sorted = sort {$max{$a->pq}-$a->score <=> $max{$b->pq}-$b->score 
		       || $a->clo(1) <=> $b->clo(1)} 
                      @$pAlis;
    return \@sorted;
}

sub findShortfall {
    my ($pSorted) = @_;
    my $max = $pSorted->[0]->score;
    foreach my $k (1 .. $#$pSorted) {
	my $s = $pSorted->[1]->score;
	return $max-$s if ($s<$max);
    }
    return $max;
}

sub maximumAvailableScore {
    my ($clone,$pWM) = @_;
    # Remove any .[pq]1k from clone: we don't allow chimeric ones
    $clone =~ s/\.[pq]1k$//;
    my %mas;
    foreach my $pq qw(p q) {
	my $sum = 0;
	my $read = "$clone.${pq}1k";
	if (defined $pWM->{$read}) {
	    foreach my $pA (@{$pWM->{$read}}) {
		$sum += $pA->[1]-$pA->[0]+1;
	    }
	}
	$mas{$pq} = $sum;
    }
    return \%mas;
}

sub evaluateClusters {
    my ($pClusters,$pMas,$pRegs) = @_;
    rejectVectorClusters($pClusters);
    my $chi = isChimericClusters($pClusters);
    rejectLowScoringClusters($pClusters,$pMas,$chi);
    rejectTransposonClusters($pClusters,$pRegs);
    rejectSubtelomericClusters($pClusters,$pRegs);
    rejectDodgyInsertionClusters($pClusters);
    evaluateClustersFinal($pClusters,$chi);
}

sub rejectVectorClusters {
    my ($pClusters) = @_;
    foreach my $pC (@$pClusters) {
	if ($pC->chr =~ /^vc_/) {
	    $pC->status('reject_vc');
	}
    }
}

sub isChimericClusters {
    my ($pClusters) = @_;
    my ($topPScore,$topQScore,$topPQScore);
    foreach my $k (0 .. $#$pClusters) {
	my $pC = $pClusters->[$k];
        my $s = $pC->score;
	my $e = $pC->ends;
	if ($e eq 'pq') {
	    $topPQScore = max($topPQScore,$s);
	} elsif ($e eq 'p') {
	    $topPScore = max($topPScore,$s);
	} else {
	    $topQScore = max($topQScore,$s);
	}
    }
    return ($topPScore > $topPQScore && $topQScore > $topPQScore);
}

sub rejectLowScoringClusters {
    my  ($pClusters,$pMas,$chi) = @_;
    # Do separately if chimeric ...
    if ($chi) {
	foreach my $pq (qw(p q)) {
	    rejectLowScoringClusters1(clustersForEnd($pq,$pClusters),$pMas->{$pq});
	}
    } else {
	rejectLowScoringClusters1($pClusters,$pMas->{p}+$pMas->{q});
    }
}

# "Low scoring" is defined as 25 or more below the winner in shared-score.
# (Was 10 until 4-9-07, but many below 25 look dubious).

sub rejectLowScoringClusters1 {
    my ($pClusters,$mas) = @_;
    return unless (@$pClusters);
    my $sTop = $pClusters->[0]->score;
    my $ambiguityShortfall = 25;
    # Low-score threshold:
    my $th = $sTop - $ambiguityShortfall;
    # Say all are ambiguous ("manual") if first and second clusters
    # are within $ambiguityShortfall of each other.
    # Score of "second" is 0 if no second; and if first and second clusters
    # are effectively identical, e.g. in rDNA region, take "second" as first
    # one that isn't identical.
    my $allAmbig;
    # hasRelativelyUniqueAlignment: some alignment in the cluster has a
    # sharedScore that is at least 0.2 times its raw score. To rule out
    # transposons.
    if ($pClusters->[0]->hasRelativelyUniqueAlignment(0.2)) {
	my $sNext = 0;
	foreach my $i (1 .. $#$pClusters) {
	    unless ($pClusters->[0]->effectivelyIdentical($pClusters->[$i])) {
		$sNext = $pClusters->[$i]->score;
		last;
	    }
	}
	$allAmbig = $sNext >= $th;
    } else {
	$allAmbig = 1;
    }
    foreach my $pC (@$pClusters) {
	if ($pC->score < $th || $pC->chr =~ /^vc_/) { # exclude vectors here!
	    $pC->status('reject_low');
	} elsif ($allAmbig) {
	    $pC->status('reject_close');
	}
    }
}

sub clustersForEnd {
    my ($pq,$pClusters) = @_;
    my @ans;
    foreach my $pC (@$pClusters) {
	push @ans,$pC if ($pC->ends eq $pq);
    }
    return \@ans;
}

sub rejectTransposonClusters {
    my ($pClusters,$pRegs) = @_;
    my @trans;
    my ($maxScore,$nWithMax);
    foreach my $pC (@$pClusters) {
	my $d = $pC->score - $maxScore;
	if ($d > 0) {
	    $maxScore = $pC->score;
	    $nWithMax = 1;
	} elsif ($d == 0) {
	    $nWithMax ++;
	}
    }
    foreach my $pC (@$pClusters) {
	if ($pC->status =~ /^reject/ || $nWithMax > 1) {
	    if ($pC->mostlyOverlaps('transposable_element',$pRegs)) {
		push @trans,$pC;
		$pC->status('ambiguous_trans'); # even if already rejected
	    }
	}
    }
    return unless (@trans);
    foreach my $pC (@$pClusters) {
	if ($pC->status =~ /^reject/ || $nWithMax > 1) {
	    foreach my $pCT (@trans) {
		if ($pC->proportionSpannedBy($pCT) >= 0.5) {
		    $pC->status('ambiguous_trans_ua'); # even if already rejected
		    last;
		}
	    }
	}
    }    
}

sub rejectSubtelomericClusters {
    my ($pClusters,$pRegs) = @_;
    return if (@$pClusters == 1);
    my $th = $pClusters->[1]->score + 100;
    foreach my $pC (@$pClusters) {
	if ($pC->status =~ /^reject/) {
	    if ($pC->score < $th && $pC->mostlyOverlaps('subtelomere',$pRegs)) {
		$pC->status('ambiguous_tel'); # even if already rejected
	    }
	}
    }
}

sub rejectDodgyInsertionClusters {
    my ($pClusters) = @_;
    my $rejd;
    foreach my $pC (@$pClusters) {
	unless ($pC->status =~ /^reject/) {
	    if ($rejd || isDodgyInsertionCluster($pC)) {
		$pC->status("reject_insertion");
		$rejd = 1; # and reject all following ones too
	    }
	}
    }
}

# A "dodgy insertion cluster" is one that does not provide
# strong enough support for the insertion that it contains.
# By strong enough support we mean:
#   1  it must not be a singleton (we need the mate to position
#      it reliably in the genome). If the insertion is real,
#      another read (facing the other way) will provide this
#      extension, we hope. BUT we allow singletons if
#      the (real) alignments in our end are strong.
#   2  there must be only one insertion involved (over both ends)
#   3  it must have exactly one or (actual) alignment in the
#      end with the insertion, or two to be closed, and exactly
#      one at the other end (if present).
# If it fails 2 or 3, it's likely there's something complex
# going on, so it's better to reject the read pair (even though
# it does appear to align round here) than risk creating wrong
# sequence. Note that all of 1, 2 and 3 may come good on future
# iterations as extensions from other reads turn insertions on
# this iteration into matches.

sub isDodgyInsertionCluster {
    my ($pC) = @_;
    my @all = $pC->getAllExtended;
    my %alis;
    my $nIns;
    my $ipq;
    foreach my $pAli (@all) {
	my $pq = $pAli->pq;
	push @{$alis{$pq}},$pAli;
	if ($pAli->getType eq 'INS') {
	    $ipq = $pq;
	    # More than one insertion (in same end or different)
	    # -> not reliable
	    return 1 if ($nIns);
	    $nIns ++;
	}
    }
    # No insertions -> not relevant
    return 0 unless ($nIns);
    # When we get here, we know we have exactly one insertion,
    # and it's at end $ipq.
    my $iqp = $ipq eq 'p' ? 'q' : 'p';
    # To be strong enough, we need an insertion plus a good
    # alignment (or two) on one side, and a single alignment on the other.
    my $nipq = scalar(@{$alis{$ipq}});
    # FORMERLY: we didn't allow == 3
    return 1 unless ($nipq == 2 || $nipq == 3); # dodgy if this end has double insn
    my $th;
    if (defined $alis{$iqp}) {
	# Have mate: OK if mate is simple.
	return 1 if (scalar(@{$alis{$iqp}}) > 1); # dodgy if other end exists and is complex
	$th = 100; # just need a bit aligned on this side. FORMERLY 400.
    } else {
	$th = 400; # need more aligned on this side if no mate. FORMERLY always dodgy here.
    }
    # No mate. Check whether the alignment(s) beside the insertion are
    # strong: all perci must be >= 90, and sum of lengths >= $th.
    my $sumL = 0;
    foreach my $pAli (@{$alis{$ipq}}) {
	if ($pAli->getType eq 'ALI') {
	    # then cluster is dodgy if this alignment is too weak:
	    return 1 if ($pAli->perci < 90);
	    $sumL += $pAli->chi-$pAli->clo+1;
	}
    }
    return ($sumL < $th); # dodgy if too little aligned
}

sub evaluateClustersFinal {
    my ($pClusters,$chi) = @_;
    my %nok;
    my %maxScore;
    foreach my $pC (@$pClusters) {
	my $e = $chi ? $pC->ends : 0;
	unless ($pC->status =~ /^reject/ || $pC->status =~ /^ambiguous/) {
	    $nok{$e}{$pC->score} = 1;
	}
	$maxScore{$e} = max($maxScore{$e},$pC->score);
    }
    return unless (%nok);
    # Re-assess chimeric: not chimeric if one read has turned out
    # to be transposon or telomere.
    my $chi2 = $chi && scalar(keys %{$nok{p}}) >= 1 
	            && scalar(keys %{$nok{q}}) >= 1;
    foreach my $e (keys %nok) {
	my @scores = keys %{$nok{$e}};
	my $n = scalar(@scores);
	my $val = ($n > 1) ? 'ambiguous' : $chi2 ? 'accept_chi' : 'accept';
	foreach my $pC (@$pClusters) {
	    my $ec = $chi ? $pC->ends : 0;
	    if (!$pC->status && $ec eq $e) {
		if ($pC->score < $maxScore{$e}) {
		    $pC->status('reject_low');
		} elsif ($val eq 'ambiguous' && $pC->chr(2) ne 'chrm') {
		    $pC->status('manual');
		} elsif ($val eq 'accept' && notClearlyBestCluster($pC,$pClusters)) {
		    $pC->status('reject_ncb');
		} else {
		    $pC->status($val);
		}
	    }
	}
    }
}

sub notClearlyBestCluster {
    my ($pC,$pClusters) = @_;
    my ($sp,$sq) = bestPAndQScores($pC);
    my $score = $sp+$sq;
    my ($pMax,$qMax);
    foreach my $pC2 (@$pClusters) {
	unless ($pC2 == $pC) {
	    my ($sp2,$sq2) = bestPAndQScores($pC2);
	    $pMax = max($pMax,$sp2);
	    $qMax = max($qMax,$sq2);
	}
    }
    # Because we can have normal+tranposon here and not count as
    # chimeric in $chi2 sense, we have to make playing field level...
    if (!defined $sq) {
	return ($pMax > $sp);
    } elsif (!defined $sp) {
	return ($qMax > $sq);
    } else {
	return ($pMax+$qMax > $sp+$sq); # we allow joint best, for rDNA stuff
    }
}

sub bestPAndQScores {
    my ($pC) = @_;
    my ($pMax,$qMax);
    foreach my $pA ($pC->getAll) {
	my $s = $pA->score; # raw score, not shared
	if ($pA->pq eq 'p') {
	    $pMax = max($pMax,$s);
	} else {
	    $qMax = max($qMax,$s);
	}
    }
    return ($pMax,$qMax);
}

# reject_bisr: cluster has a break in a region that was sequenced
# (i.e. had an alignment accepted) in an EARLIER run, or jumps
# across such a region.

sub rejectBISRClusters {
    my ($pClusters,$pSRS) = @_;
    while (my ($clone,$pA) = each %$pClusters) {
	my $threshold;
	foreach my $pC (@$pA) {
	    if ($pC->hasBreakInSequencedRegion($pSRS)) {
		$threshold = max($threshold,$pC->score-100);
		if ($pC->status !~ /^reject/) {
		    $pC->status('reject_bisr');
		}
	    }
	}
	if (defined $threshold) {
	    foreach my $pC (@$pA) {
		if ($pC->score < $threshold && $pC->status !~ /^reject/) {
		    # mlt_bisr: reject because scores much less than another that
		    # had break in sequenced region
		    $pC->status('reject_mlt_bisr');
		}
	    }
	}
    }
}

# reject_bad_break: cluster has a break in a region that another
# (currently accepted) cluster in the CURRENT set spans across.

sub rejectClustersWithUnlikelyBreaks {
    my ($pClusters) = @_;
    # First mark out all the UNaligned regions in current clusters
    my $pUnaligned = findUnalignedRegions($pClusters);

    # Then reject accepted clusters with breaks in aligned regions.
    # Rejecting the whole cluster is too drastic; the most we should
    # do is make it chimeric and reject just the half with the break,
    # but no time now.

    while (my ($clone,$pA) = each %$pClusters) {
      CLUSTER:
	foreach my $pAC (@$pA) {
	    if ($pAC->status =~ /^accept/) {
		my @alis = $pAC->getAll;
		my $chr = $pAC->chr(1);
		for (my $i=0; $i < $#alis; $i++) {
		    my $pALo = $alis[$i];
		    my $pAHi = $alis[$i+1];
		    if ($pALo->fc eq $pAHi->fc &&
			!isWithinUnalignedRegion($pUnaligned->{$chr},$pALo->chi(1)+1,
						 $pAHi->clo(1)-1)) {
			$pAC->status('reject_bad_break');
			next CLUSTER;
		    }
		}
		# reject_poor if no alignment has score >= 150 and %id >= 80 (currently)
		if ($pAC->noSingleGoodAlignment) {
		    $pAC->status('reject_poor');
		}
	    }
	}
    }
}

sub findUnalignedRegions {
    my ($pClusters) = @_;
    my %unaligned;
    foreach my $pA (values %$pClusters) {
	foreach my $pAC (@$pA) {
	    if ($pAC->status =~ /^accept/) {
		my @alis = $pAC->getAll;
		foreach my $pSpan (alignmentSpansAllowingForOverlaps(\@alis)) {
		    reduceUnalignedRegions($pSpan,\%unaligned);
		}
	    }
	}
    }
    return \%unaligned;
}

sub alignmentSpansAllowingForOverlaps {
    my ($pAlis) = @_;
    my @spans;
    foreach my $pAli (@$pAlis) {
	push @spans,[$pAli->chr(1),$pAli->clo(1),$pAli->chi(1)];
    }
    @spans = sort {$a->[0] cmp $b->[0] || $a->[1] <=> $b->[1] || $a->[2] <=> $b->[2]} @spans;
    my @spans2;
    foreach my $pSpan (@spans) {
	if (@spans2 && 
	    $pSpan->[0] eq $spans2[-1][0] &&
	    $pSpan->[1] <= $spans2[-1][2]) {
	    # Make new span covering just the overlap
	    my $tmp = $spans2[-1][2];
	    $spans2[-1][2] = $pSpan->[1]-1;
	    $pSpan->[1] = $tmp+1;
	    # The middle bit:
	    if ($pSpan->[1]+1 <= $tmp) {
		push @spans2,[$pSpan->[0],$pSpan->[1]+1,$tmp-1];
	    }
	}
	push @spans2,$pSpan;
    }
    return @spans2;
}

sub reduceUnalignedRegions {
    my ($pSpan,$pUnaligned) = @_;
    my ($chr,$clo,$chi) = @$pSpan;
    unless (defined $pUnaligned->{$chr}) {
	$pUnaligned->{$chr}{0} = '1000000000';
    }
    reduceUnalignedRegions1($pUnaligned->{$chr},$clo,$chi);
}

sub reduceUnalignedRegions1 {
    my ($pUna,$lo,$hi) = @_;
    return if ($lo >= $hi); # trivial region
    while (my ($ulo,$uhi) = each %$pUna) {
	if ($lo <= $ulo) {
	    if ($hi >= $ulo) {
		delete $pUna->{$ulo};
		if ($hi < $uhi) {
		    $pUna->{$hi+1} = $uhi;
		}
	    }
	} elsif ($lo <= $uhi) {
	    $pUna->{$ulo} = $lo-1;
	    if ($hi < $uhi) {
		$pUna->{$hi+1} = $uhi;
	    }
	}
    }
}

sub isWithinUnalignedRegion {
    my ($pUna,$lo,$hi) = @_;
    #while (my ($ulo,$uhi) = each %$pUna) {
    foreach my $ulo (keys %$pUna) {
	# We allow a leeway of 10 nucleotides because of alignment arbitrarinesses...
	# See also $BISR_LEEWAY in AlignmentCluster.pm.
	if ($lo >= $ulo-10 && $hi <= $pUna->{$ulo}+10) {
	    return 1;
	}
    }
    return 0;
}

sub rejectClashingChimericClusters {
    my ($pClusters) = @_;
    foreach my $pA (values %$pClusters) {
	for (my $i = 0; $i < $#$pA; $i ++) {
	    my $pCI = $pA->[$i];
	    next unless ($pCI->status eq 'accept_chi');
	    for (my $j = $i+1; $j <= $#$pA; $j ++) {
		my $pCJ = $pA->[$j];
		next unless ($pCJ->status eq 'accept_chi');
		next unless ($pCI->chr(1) eq $pCJ->chr(1));
		# Now see if they're within 10k of each other
		my $d = max($pCI->clo(1)-$pCJ->chi(1), $pCJ->clo(1)-$pCI->chi(1));
		if ($d < 10000) {
		    # then we don't want to keep them both ... may make an
		    # arbitrary choice here ...
		    if ($pCI->score < $pCJ->score) {
			$pCI->status('reject_chi');
		    } else {
			$pCJ->status('reject_chi');
		    }
		}
	    }
	}
    }
}

# $gh is clusters.txt
# $hh1 is alilines.pass.txt
# $hh2 is alilines.all.status.txt

sub printAlignments {
    my ($pWM,$pCList,$sf,$idx,$gh,$hh1,$hh2,$meth) = @_;

    my @sorted = @$pCList;

    my @pq = qw(P Q);

    my %max = (p => [], q => []); # max score along p and q

    my $k=0;
    my $smax = $sorted[0]->score;
    my $sfs = sprintf("IDX %d SF %4.2f",$idx,$sf);
    my %hList;
    foreach my $c (@sorted) {
	my @alis = $c->getAllExtended();
	my $tm = touchesMax(\@alis,$c->score(),\%max);
	# Report at top level if all chrs are same, but 1
	# if we have mix of base and aug.
	my $lev = 0;
	my $chr0 = $alis[0]->chr(0);
	foreach my $i (1 .. $#alis) {
	    unless ($alis[$i]->chr(0) eq $chr0) {
		$lev = 1;
		last;
	    }
	}
	printf($gh "%s c%d %s (%d,%d) %s:%d..%d (%d) TM %d %s\n",
	       $c->clone,$k,$c->status,$c->score,$smax-$c->score,
	       $c->chr($lev),$c->clo($lev),$c->chi($lev),$c->len($lev),$tm,$sfs);
	$sfs='';
	my $st = $c->status;
	my $special; # for read names, on 'EXTRA' iteration
	if ($IS_FINAL_ITERATION eq 'EXTRA') {
	    if ($st eq 'reject_close' || 
		$st eq 'reject_bad_break' ||
		$st eq 'reject_insertion' || 
		$st eq 'reject_ncb' || 
		$st eq 'reject_bisr') {
		$st =~ s/^reject/accept/;
		$st =~ s/bad_break/bisr/; # small class into large...
	    } elsif ($st =~ /^accept/) {
		$st .= '_late';
	    }
	    $c->status($st);
	    $special = substr($st,7) if ($st =~ /^accept_/);
	}
	# $acc means accept on this iteration, subject to
	# successful merging.
	my $acc = ($st =~ /^accept/);
	foreach my $pA (@alis) { 
	    $pA->special($special) if ($special);
	    my @rlh;
	    if ($pA->fc eq 'F') {
		@rlh = ($pA->rlo,$pA->rhi);
	    } else {
		@rlh = ($pA->rhi,$pA->rlo);
	    }
	    printf($gh "  %s %s %4d <%4d : %4d %4d : %7d %7d : %4d %6.2f %4d\n",
		   $pA->fc,$pA->pq,$pA->sharedScore,$pA->score,
		   @rlh,
		   $pA->clo,$pA->chi,
		   $pA->alen,$pA->perci,$pA->rlen);
	    if ($hh1) {
		push @{$hList{$pA->pq}{$pA->score}},[$pA,$st,$acc,$k];
	    }
	}	
	$k++;
    }
    print $gh "\n";

    my %haveAccept;
    my @isAmbi;
    my %ambiOccs;
    my %stStatus;
    foreach my $pq (keys %hList) {
	foreach my $s (sort {$b <=> $a} keys %{$hList{$pq}}) {
	    foreach my $pTup (@{$hList{$pq}{$s}}) {
		my ($pA,$status,$acc,$k) = @$pTup;
		my $st = $pA->toString();
		if ($acc) {
		    unless ($haveAccept{$st}) {
			$haveAccept{$st} = 1;
			print $hh1 "$st\n";
		    }
		} elsif ($status =~ /^ambiguous/ || $status =~ /^manual/) {
		    $ambiOccs{$st}[$k] = 1;
		    $isAmbi[$k] = 1;
		}
		#print $hh2 "$st $status $k\n";
		$stStatus{$st}{$k} = $status;
	    }
	}
    }
    # Also print to $hh1 every alignment that occurred in
    # all "ambiguous" and "manual" clusters (when none accepted
    # so far). This means we _can_ end up accepting an "extra" (ins or del)
    # alignment when we reject an (ambiguous) alignment beside it! But
    # such cases should be rare.
    unless (%haveAccept) {
	my @ambi = 0;
	foreach my $k (0 .. $#isAmbi) {
	    push @ambi,$k if ($isAmbi[$k]);
	}
	foreach my $st (sort keys %ambiOccs) {
	    my $ok = 1; # ok means "occurs in _all_ ambiguous clusters"
	    foreach my $k (@ambi) {
		unless ($ambiOccs{$st}[$k]) {
		    $ok = 0;
		    last;
		}
	    }
	    if ($ok) {
		print $hh1 "$st\n";
		foreach my $k (@ambi) {
		    my $s = $stStatus{$st}{$k};
		    # manual -> accept_manual (never occurs?)
		    # ambiguous_REASON -> accept_REASON (tel, trans, ...)
		    unless ($s =~ /^accept/) {
			$s =~ s/^ambiguous_//;
			$s = "accept_$s";
			$stStatus{$st}{$k} = $s;
		    }
		}
	    }
	}
    }
    while (my ($st,$pH) = each %stStatus) {
	print $hh2 $st;
	my %vals;
	foreach my $k (sort {$a <=> $b} keys %$pH) {
	    push @{$vals{$pH->{$k}}},$k;
	}
	foreach my $s (sort keys %vals) {
	    printf($hh2 " %s %s",$s,join(",",@{$vals{$s}}));
	}
	print $hh2 "\n";
    }
}

sub touchesMax {
    my ($pAlis,$score,$pMax) = @_;
    my $tm = 0;
    foreach my $pA (@$pAlis) {
	my $pM = $pMax->{$pA->pq};
	if ($pA->rlo < 0 || $pA->rhi < 0) {
	    confess("Negative read offset in " . $pA->toString);
	}
	foreach my $k ($pA->rlo .. $pA->rhi) {
	    if ($score >= $pM->[$k]) {
		$pM->[$k] = $score;
		$tm ++;
	    }
	}
    }
    return $tm;
}

sub filterForFinalIteration {
    my ($alPass) = @_;
    my $orig = "$alPass.orig";
    return unless (-s $alPass && ! -s $orig);
    rename($alPass,$orig);
    report("$alPass: filtering for final iteration");
    my $fh = openToRead($orig);
    my $gh = openToWrite($alPass);
    my $lf;
    if (basename($alPass) eq 'alilines.pass.txt') {
	$lf = dirname($alPass) . '/alilines.final.status.txt';
    } else {
	die("Bad alilines.pass.txt file name: $alPass");
    }
    my $lh = openToWrite($lf);
    my ($clone,$status,@lines);
    my $minPercId = 95;
    my $minScore = 400; # assumes we penalize mismatches enough...
    while (<$fh>) {
	chomp;
	my @a = split;
	my ($cl) = ($a[2] =~ /^(.*)\.[pq]1k/);
	if ($cl ne $clone) {
	    mayPrintLines(\@lines,$status,$gh,$lh);
	    $clone = $cl;
	    $status = 'accept';
	    @lines = ();
	}
	if ($a[1] == 0) {
	    $status = 'reject_final_indel';
	} elsif ($status ne 'accept') {
	    # keep existing rejection reason
	} elsif ($a[10] < $minPercId) {
	    $status = 'reject_final_percid';
	} elsif ($a[1] < $minScore) {
	    $status = 'reject_final_score'; # but non-zero
        }	    
	push @lines,$_;
    }
    mayPrintLines(\@lines,$status,$gh,$lh);
    $gh->close();
    $lh->close();
    report();
}

sub mayPrintLines {
    my ($pLines,$status,$gh,$lh) = @_;
    if ($status eq 'accept') {
	print $gh join("\n",@$pLines) . "\n";
    }
    foreach my $line (@$pLines) {
	print $lh "$line $status\n";
    }
}

# Create ssahaSNP.pass.gz from alilines.pass.txt

sub pruneAlignments2 {
    my ($alPass,$rnfqFile,$ssaFile,$tgtDir,$pOM) = @_;
    my $selFile = "$tgtDir/ssahaSNP.pass.gz";
    if (stamp($selFile)) {
	report("creating $selFile");
	applySelectionFile($ssaFile,
			   $pOM,
			   $alPass,
			   $rnfqFile,
			   $selFile,1);
	report();
	unstamp($selFile,1);
    }
}

# Run this in an align/$str directory

sub pruneAlignments1Test {
    my $root = $ENV{SGR};
    my $str= 'BC187';
    pruneAlignments1('methc','.',"$root/read/$str",
		     "../../assembly/$str",'tmpTgt');
}

sub reportRejection {
    my ($pC,$key,$gh,$pReported) = @_;
    foreach my $pA ($pC->getAll) {
	my $line = $pA->toString;
	$line =~ s/\|.*//; # get rid of trim
	unless ($pReported->{$line}) {
	    print $gh "$line $key\n";
	    $pReported->{$line} = 1;
	}
    }
}
	
1;
