package SGRP::RedoQualityPairwise;

use Exporter;
@ISA = qw(Exporter);

@EXPORT = qw(redoQualityPairwise);

use SGRP::Variables;

use Utility;
use File::Basename;

use Carp;

use strict;

my $NQS_THRESHOLD  = '8'; # i.e. 23 numerically.
my $NQS_THRESHOLD2 = '0'; # i.e. 15 numerically

sub redoQualityPairwise {
    my ($species,$filtMeth,$strainSet,@parts) = @_;
    unless (@parts) {
	confess("Usage: $0 rqpw species filtMeth strainSet part1 part2 ...");
    }
    my ($iter,$dir) = getCurrentIteration($species,$filtMeth,$strainSet,'p?');
    if ($IS_FINAL_ITERATION eq 'EXTRA') {
	# This will prevent anything further happening, too ...
	confess("$0 rqpw: returning because isExtraIteration");
    } else {
	foreach my $part (@parts) {
	    $part = "p$part" if ($part =~ /^[0-9]$/);
	    foreach my $chrDir (glob("$dir/$part/*")) {
		redoQualityPairwise1($chrDir);
	    }
	}
    }
}

sub redoQualityPairwise1 {
    my ($dir) = @_;
    my $src = "$dir/layout.gz";
    return unless (isComplete($src,1));
    my $tgt = "$dir/rqpw.gz";
    return unless (stamp($tgt));
    my $dFile = "$dir/dangles.txt";
    my $pPoor = poorlyAlignedReadsAndDangles($src,$dFile);
    report("creating $tgt");
    my $fh = openToRead($src);
    my $gh = openToWrite($tgt);
    my @buf;
    my @pre;
    my @pre2;
    my $off = 0;
    my $pSaved = [];
    my $ptr = 0;
    my %toZap;
    while (<$fh>) {
	chomp;
	if (/^>REF /) {
	    push @pre2,$_;
	} elsif (/^>(\S+) (\d+)/) {
	    my ($r,$idx) = ($1,$2);
	    my $col = 2*($idx+1);
	    if ($pPoor->{$r}) {
		$toZap{$col} = 1;
		push @pre2,'>'; # to keep line numbers; forget read
	    } else {
		delete $toZap{$col};
		push @pre2,$_;
	    }
	} else {
	    if (@pre2) {
		my @pre3 = @pre2; # copy
		push @pre,\@pre3;
		@pre2 = ();
	    } else {
		push @pre,undef;
	    }
	    # Zap relevant columns
	    my $cMax = length($_)-2;
	    foreach my $col (keys %toZap) {
		if ($col <= $cMax) {
		    substr($_,$col,2) = '  ';
		}
	    }
	    push @buf,$_;
	    if ($off == 1000) {
		foreach my $k (5 .. 500) {
		    adjustCasingPairwise(\@buf,$k);
		}
		foreach my $k (5 .. 500) {
		    printAndShift(\@buf,\@pre,$gh);
		}
	    } elsif ($off > 1000) {
		adjustCasingPairwise(\@buf,500);
		printAndShift(\@buf,\@pre,$gh);
	    }
	    $off ++;
	}
    }
    foreach my $k (501 .. 995) {
	adjustCasingPairwise(\@buf,$k);
    }
    while (@buf) {
	printAndShift(\@buf,\@pre,$gh);
    }
    $gh->close();
    report();
    unstamp($tgt);
}

# A poorly-aligned read is one with more than 1/3
# nucleotide disagreements with ref. We don't allow such reads
# into rqpw.gz; even if they're not bugs, they're not likely to
# be reliable.
#
# We also take advantage of this read-through of the file to
# create dangles.txt.

sub poorlyAlignedReadsAndDangles {
    my ($f,$dFile) = @_;
    my $fh = openToRead($f);
    my $dh = openToWrite($dFile);
    my (@name,@agr,@dis,@start);
    my $line = 0; # line number in file
    my $off = 0; # offset from 1, i.e. exclude ">"
    my %isPal;
    my %rightDangle;
    report("scanning $f for poorly-aligned reads and dangles");
    while (<$fh>) {
	chomp;
	$line ++;
	$off ++ unless (/^>/);
	if (/^>(\S+) (\d+) /) {
	    my ($r,$col) = ($1,($2+1)*2);
	    if (defined $name[$col]) {
		# We allow as many disagreements as agreements because
		# we could have multiple divergences from reference
		# over several iterations.
		if ($dis[$col] > max(10,$agr[$col])) {
		    $isPal{$name[$col]} = "$agr[$col] $dis[$col]";
		}
		$agr[$col]=$dis[$col]=0;
	    }
	    $name[$col] = $r;
	    $start[$col] = $line;
	    if ($r =~ /L$/) {
		printf($dh "%d L\n",$off+1);
	    } elsif ($r =~ /R$/) {
		my ($pos1,$pos2) = ($r =~ /:(\d+):(\d+)/);
		my $len = $pos2>$pos1 ? $pos2-$pos1+1 : $pos1-$pos2+1;
		$rightDangle{$col} = $len;
	    }
	} else {
	    foreach my $col (keys %rightDangle) {
		if (substr($_,$col,1) =~ /[A-Za-z]/) {
		    $rightDangle{$col} --;
		    if ($rightDangle{$col} == 0) {
			print $dh "$off R\n";
			delete $rightDangle{$col};
		    }
		}
	    }
	    my $max = length($_);
	    if (/^[ACGTacgt]/) {
		tr/a-z/A-Z/;
		my $r = substr($_,0,1);
		for (my $col=2; $col<$max; $col+=2) {
		    my $s = substr($_,$col,1);
		    if ($s =~ /[ACGT]/) { # upper cased already
			if ($s eq $r) {
			    $agr[$col] ++;
			} else {
			    $dis[$col] ++;
			}
		    }
		}
	    }
	}
    }
    foreach my $col (2 .. $#name) {
	if (defined $name[$col]) {
	    if ($dis[$col] > max(10,$agr[$col])) {
		$isPal{$name[$col]} = "$agr[$col] $dis[$col]";
	    }
	}
    }
    report();
    foreach my $r (sort keys %isPal) {
	print "  Poorly-aligned: $r ($isPal{$r})\n";
    }
    return \%isPal;
}

sub adjustCasingPairwise {
    my ($pBuf,$kTgt) = @_;
    my $line = $pBuf->[$kTgt];
    my $iMax = length($line)-2;
    # If there is only one strain, we can't rescue it any further
    # here because there are no other strains to compare it against -
    # only the ref sequence, which has already been done.
    return if ($iMax < 4);
    # Columns that can support others, and by the same token,
    # can themselves be improved:
    my @iList;
  SUPPORT:
    for (my $i=2; $i<=$iMax; $i+=2) {
	next if (substr($line,$i+1,1) lt $NQS_THRESHOLD2);
	foreach my $d (-1,1) {
	    my $nToDo = 5;
	    my $k = $kTgt+$d;
	    while ($nToDo > 0) {
		next SUPPORT if ($k < 0 || $k > $#$pBuf);
		my $c = substr($pBuf->[$k],$i,1);
		next SUPPORT if ($c eq ' ');
		if ($c ne '-' && $c ne '=') {
		    next SUPPORT if (substr($pBuf->[$k],$i+1,1) lt $NQS_THRESHOLD2);
		    $nToDo --;
		}
		$k += $kTgt;
	    }
	}
	push @iList,$i;
    }
    return unless (scalar(@iList) > 1);
  ILOOP:
    foreach my $i (@iList) {
	# Doesn't need raising:
	next if (substr($line,$i,1) =~ /[- ACGT]/);
	# Own quality too low:
	next if (substr($line,$i+1,1) lt $NQS_THRESHOLD);
	# It stands a chance; let's look for another column to support it.
      JLOOP:
	foreach my $j (@iList) {
	    next if ($j == $i || substr($line,$j,1) =~ /[- ]/);
	    # See if j can support i. Just need to check not more than
	    # one mismatch.
	    my $nMis = 0;
	    foreach my $d (-1,1) {
		my $nToDo = 5;
		my $k = $d > 0 ? $kTgt : $kTgt+$d;
		my $nMis = 0;
		while ($nToDo > 0) {
		    next JLOOP if ($k < 0 || $k > $#$pBuf);
		    my $ci = substr($pBuf->[$k],$i,1);
		    my $cj = substr($pBuf->[$k],$j,1);
		    unless (($ci eq '-' || $ci eq '=') && ($cj eq '-' || $cj eq '=')) {
			$ci =~ tr/ACGT/acgt/;
			$cj =~ tr/ACGT/acgt/;
			unless ($ci eq $cj) {
			    $nMis ++;
			    next JLOOP if ($nMis > 1);
			}
			$nToDo --;
		    }
		    $k += $d;
		}
	    }
	    # Yes, j can rescue i (and vice versa)
	    substr($line,$i,1) =~ tr/acgt/ACGT/;
	    substr($line,$j,1) =~ tr/acgt/ACGT/;
	}
    }
    $pBuf->[$kTgt] = $line; # put it back
}

sub printAndShift {
    my ($pBuf,$pPre,$gh) = @_;
    if (defined $pPre->[0]) {
	foreach my $line (@{$pPre->[0]}) {
	    print $gh "$line\n";
	}
    }
    shift @$pPre;
    print $gh "$pBuf->[0]\n";
    shift @$pBuf;
}
    


1;
