package SGRP::ReportSNPs;

use Exporter;
@ISA = qw(Exporter);

@EXPORT = qw(reportSNPs);

use SGRP::Variables;

use Utility;
use File::Basename;

use Carp;

use strict;

# Create $ROOT/report/[i]snps_cere_CERE__meth1_meth2_meth....txt.gz
# from [i]snp_report.gz files created by esnp and isnp.

# Window size
my $WSZ = 21;

sub reportSNPs {
    my ($species,$strainSet,@filtMeths) = @_;
    unless (@filtMeths) {
	confess("Usage: $0 rsnp species strainSet meth1 meth2 ...");
    }
    @filtMeths = sort @filtMeths;
    foreach my $type ('','i') {
	my $tgt = sprintf("%s/report/%ssnps_%s_%s_%s.txt.gz",
			  $ROOT,$type,$species,$strainSet,join("_",@filtMeths));
	if (stamp($tgt)) {
	    reportSNPs1($tgt,$type,$species,$strainSet,@filtMeths);
	    unstamp($tgt);
	}
    }
    foreach my $meth (@filtMeths) {
	my $iscFile = sprintf("%s/report/isnp_counts_%s_%s_%s.txt",
			      $ROOT,$species,$strainSet,$meth);
	if (stamp($iscFile)) {
	    my ($iter,$dir) = getCurrentIteration($species,$meth,$strainSet,'p?');
	    addISNPCounts($dir,$iscFile);
	    unstamp($iscFile);
	}
    }
}

sub reportSNPs1 {
    my ($tgt,$type,$species,$strainSet,@filtMeths) = @_;
    my $srcTop = sprintf("%s/match/%s/%s",$ROOT,$strainSet,$species);
    my %inFile;
    my $ok = 1;
    foreach my $meth (@filtMeths) {
	foreach my $p (0 .. 9) {
	    my $srcTop2 = "$srcTop/$meth/p$p";
	    foreach my $d (glob("$srcTop2/*")) {
		my $f = "$d/${type}snp_report.gz";
		# Record it to prevent re-reporting later:
		$inFile{basename($d)}{$meth} = $f;
		unless (-s $f) {
		    print STDERR "WARNING: missing or empty: $f\n";
		    $ok = 0;
		}
	    }
	}
    }
    foreach my $chr (sort keys %inFile) {
	foreach my $meth (@filtMeths) {
	    unless (defined $inFile{$chr}{$meth}) {
		my $part = substr($chr,-1);
		$part = 0 unless ($part > 0);
		print STDERR "WARNING: missing: $srcTop/$meth/p$part/$chr/${type}snp_report.gz\n";
	    }
	}
    }
    if ($ok || 1) {
	my $gh = openToWrite($tgt);
	my %pMatchCount;
	foreach my $chr (sort keys %inFile) {
	    report("doing $chr");
	    reportSNPs2($chr,$inFile{$chr},$gh,\%pMatchCount);
	    report();
	}
	reportMatchCounts(\%pMatchCount,$gh);
	$gh->close();
    } else {
	print STDERR "Cannot create $tgt because of errors\n";
    }
}

sub reportSNPs2 {
    my ($chr,$pH,$gh,$pMC) = @_;
    my (@in,@buf);
    my @meths;
    foreach my $meth (sort keys %$pH) {
	push @meths,$meth if (-s $pH->{$meth});
    }
    foreach my $meth (@meths) {
	my $fh = openToRead($pH->{$meth});
	push @in,$fh;
	push @buf,readSNP($fh);
    }
    while (1) {
	my (@kWin,$pBest);
	foreach my $k (0 .. $#buf) {
	    my $pSNP = $buf[$k];
	    my $cmp = compareSNPs($pSNP,$pBest);
	    if ($cmp < 0) {
		$pBest = $pSNP;
		@kWin = ($k);
	    } elsif ($cmp == 0) {
		$pBest = $pSNP;
		push @kWin,$k;
	    }
	}
	last unless (defined $pBest);
	# Replenish:
	my @methWin;
	foreach my $k (@kWin) {
	    push @methWin,$meths[$k];
	    $buf[$k] = readSNP($in[$k]);
	}
	my ($rLine,$sLine,$qLine) = formatSNPTrio($pBest->{vanilla});
	# Specify which method(s) if there were several possible:
	my $methStr = @meths > 1 ? (" " . join(",",@methWin)) : '';
	printf($gh "%s\n%s%s\n%s\n\n",
	       $rLine,
	       $sLine, $methStr,
	       $qLine);
    }
    combineMatchCounts(\@in,\@meths,$pMC);
}

sub readSNP {
    my ($fh) = @_;
    my %snp;
    $_ = <$fh>; # ref line
    return undef unless ($_); # eof
    return undef if (/MatchCounts/); # process later
    my $vanilla = $_; # UNCHOMPED
    $snp{refSeq} = substr($_,0,$WSZ);
    $_ = <$fh>; # str line
    return undef if (/MatchCounts/); # process later
    $vanilla .= $_; # add UNCHOMPED
    chomp;
    $snp{strSeq} = substr($_,0,$WSZ);
    ($snp{chr},$snp{off},$snp{str},$snp{trans},$snp{qual},$snp{vote}) 
	= split(/ /,substr($_,$WSZ+1));
    $_ = <$fh>; # qual seq line
    chomp;
    $vanilla .= $_; # add CHOMPED
    $snp{qualSeq} = $_;
    $snp{vanilla} = $vanilla;
    <$fh>; # discard blank line
    return \%snp;
}

sub compareSNPs {
    my ($pS1,$pS2) = @_;
    return 1 unless (defined $pS1);
    return -1 unless (defined $pS2);
    my $cmp = ($pS1->{off} <=> $pS2->{off}) 
	|| ($pS1->{str} cmp $pS2->{str})
	|| ($pS1->{trans} cmp $pS2->{trans});
    if ($cmp || ($pS1->{vanilla} eq $pS2->{vanilla})) {
	return $cmp;
    } else {
	return 1;
    }
}

sub formatSNPTrio {
    my ($lines) = @_;
    my ($rLine,$sLine,$qLine) = split(/\n/,$lines);
    # Easier to have conventional ("-") representation for gaps:
    substr($rLine,0,$WSZ) =~ s/\d/-/g;
    substr($sLine,0,$WSZ) =~ s/\d/-/g;
    return ($rLine,$sLine,$qLine);
}

sub combineMatchCounts {
    my ($pIn,$pMeths,$pMC) = @_;
    foreach my $i (0 .. $#$pIn) {
	my $fh = $pIn->[$i];
	my $meth = $pMeths->[$i];
	while (<$fh>) {
	    chomp;
	    s/^ *//;
	    my ($str,@vals) = split;
	    # Hack to make it come first alphabetically in sort below:
	    $str = ' REF' if ($str eq 'REF');
	    foreach my $k (0 .. $#vals) {
		$pMC->{$meth}{$str}[$k] += $vals[$k];
	    }
	}
    }
}

# Header triples are of the form [NG][NGSngs][+-], where:
#  -- first char is for ref being nucleotide or gap
#  -- second is for strain being nuc, gap or space, and upper/lower case
#     ("s" is actually impossible)
#  -- third is for ref and strain agreeing (+) or not (-) when casing ignored
#     (so some combinations are impossible, e.g. NG+, in which case we leave
#     out the +/- for the one that is possible).

sub reportMatchCounts {
    my ($pMC,$gh) = @_;
    foreach my $meth (sort keys %$pMC) {
	print $gh "\nMatchCounts for method $meth\n";
	my $pC = $pMC->{$meth};
	my $maxStrLen = 6;
	my @maxLen; # field lengths
	foreach my $str (keys %$pC) {
	    $maxStrLen = length($str) if ($maxStrLen < length($str));
	    foreach my $k (0 .. $#{$pC->{$str}}) {
		my $c = $pC->{$str}[$k];
		if ($c > 0) {
		    my $len = length($c);
		    $len = 3 if ($len < 3); # to allow for key
		    $maxLen[$k] = $len if ($len > $maxLen[$k]);
		}
	    }
	}
	printf($gh "%${maxStrLen}s","Strain");
	my @keys = qw(NN+ NN- Nn+ Nn- NG+ NG  Ng+ Ng  NS+ NS  Ns+ Ns-
		      GN+ GN  Gn+ Gn  GG+ GG- Gg+ Gg- GS+ GS  Gs+ Gs-);
	foreach my $k (0 .. 23) {
	    if ($maxLen[$k]) {
		printf($gh " %${maxLen[$k]}s",$keys[$k]);
	    }
	}
	print $gh "\n";
	foreach my $str (sort keys %$pC) {
	    printf($gh "%${maxStrLen}s",$str);
	    # 23 because indices are 2x3x2x2
	    foreach my $k (0 .. 23) {
		if ($maxLen[$k]) {
		    printf($gh " %${maxLen[$k]}d",$pC->{$str}[$k]);
		}
	    }
	    print $gh "\n";
	}
    }
}

sub addISNPCounts {
    my ($dir,$tgt) = @_;
    my ($sumSites,$sumSNPs) = (0,0);
    my %count;
    foreach my $pd (glob("$dir/p?")) {
	foreach my $chrd (glob("$pd/*")) {
	    my $src = "$chrd/isnp_counts.txt";
	    if (-s $src) {
		my $fh = openToRead($src);
		$_ = <$fh>;
		chomp;
		my (undef,$nSites,undef,$nSNPs) = split;
		$sumSites += $nSites;
		$sumSNPs += $nSNPs;
		while (<$fh>) {
		    chomp;
		    my ($s1,$s2,$n) = split;
		    $count{$s1}{$s2} += $n;
		}
	    }
	}
    }
    my $gh = openToWrite($tgt);
    print $gh "N_SITES $sumSites N_SNPS $sumSNPs\n";
    foreach my $s1 (sort keys %count) {
	my $pC = $count{$s1};
	foreach my $s2 (sort keys %$pC) {
	    printf($gh "%s %s %d\n",$s1,$s2,$pC->{$s2});
	}
    }
    $gh->close();
}

1;
