package SGRP::SGRP;

use Exporter;
@ISA = qw(Exporter);

@EXPORT = qw(sgrp sgrpCommand getStrains);

# This needs to be commented out when we run on cbi1 (and occasionally the farm) because GD.so not found there...
#use SGRP::AssessPlates;
use SGRP::AlignmentCoords;
use SGRP::BuildFastqAlignments;
use SGRP::BuildGbrowse;
use SGRP::CombineLayoutByStrain;
use SGRP::CorrectNames;
use SGRP::CreateReadLayout;
use SGRP::ExtendReference;
use SGRP::ExtractNovelSequence;
use SGRP::ExtractSNPs;
use SGRP::FindBranchLengths;
use SGRP::GenerateHypotheses;
use SGRP::Imputer;
use SGRP::MakeFastq;
use SGRP::MakeGenomes;
use SGRP::Margarita;
use SGRP::MergeMatches;
use SGRP::PrepareImputer;
use SGRP::PrepareMargarita;
use SGRP::PruneAlignments;
use SGRP::RedoQualityPairwise;
use SGRP::ReportSNPs;
use SGRP::RunFregeneExperiment;
use SGRP::SelfAlignment;
use SGRP::SNPMatrix;
use SGRP::SNPMatrixImputer;
use SGRP::SortLayoutFiles;
use SGRP::SplitAlignments;
use SGRP::StrainAssemblies;
use SGRP::SsahaSNP;
use SGRP::TranslateGenes;
use SGRP::Variables;

use Utility;
use Bsub;
use File::Basename;
use Getopt::Std;

use Carp;

use strict;

autoflush STDOUT 1;

##
##Sample commands:
##
##  sgrp.pl mkfa cere para ...
##  sgrp.pl mkfq BC187 ...
##
##  sgrp.pl self cere 0 1 2 ...
##  sgrp.pl extd cere str1 str2 ...
##
##  sgrp.pl crnm BC187 ... [create renamed.fastq from clip.fastq]
##
##  sgrp.pl ssnp cere methc CERE 0 BC187 ...
##  sgrp.pl aspl cere BC187 ...
##  sgrp.pl alip cere methc CERE BC187 ... [alignment pruning, calls ssnp]
##  sgrp.pl sali cere methc 0 9 methc09 BC187 ...
##
##  sgrp.pl merg cere methc CERE 0 1 2 ...
##
##  sgrp.pl proc cere methc CERE 0 1 2 ... [equiv to crlo ... impu]
##  sgrp.pl prc2 cere methc CERE 0 1 2 ... [equiv to crlo ... slof]
##    sgrp.pl crlo cere methc CERE 0 1 2 ...
##    sgrp.pl rqpw cere methc CERE 0 1 2 ...
##    sgrp.pl clbs cere methc CERE 0 1 2 ...
##    sgrp.pl slof cere methc CERE 0 1 2 ...
##    sgrp.pl esnp cere methc CERE 0 1 2 ...
##    sgrp.pl pmar cere methc CERE 0 1 2 ...
##    sgrp.pl marg cere methc CERE 0 1 2 ...
##    sgrp.pl prim cere methc CERE 0 1 2 ... [usu done by marg]
##    sgrp.pl impu cere methc CERE 0 1 2 ...
##
##  sgrp.pl stra cere methc CERE str1 str2 ...
##
##  sgrp.pl bugb cere methc CERE str1 str2 ...
##  sgrp.pl cmgb gbdbDir
##  sgrp.pl bfqa cere methc CERE str1 str2 ...
##
##  sgrp.pl alco cere methc CERE 0 1 2 ...
##  sgrp.pl xnov cere methc CERE 0 1 2 ...
##
##  sgrp.pl smat cere methc CERE 0 1 2 ...
##  sgrp.pl smim cere methc CERE 0 1 2 ...
##
##  sgrp.pl isnp cere methc CERE 0 1 2 ...
##  sgrp.pl rsnp cere CERE meth1 meth2 ...
##
##  sgrp.pl rfre cere methc CERE 12 1 [do not use -b!]
##
##  sgrp.pl tran cere methc CERE 0 1 2 ... 
##
## Overall ref:strain SNP rates; for matrix, use sgrp.pl smat
##  sgrp.pl esnr cere methc CERE

my %NFIXED = (
	      aspl => 1,
	      cmgb => 0,
	      crnm => 0,
	      extd => 1,
	      filt => 2,
	      mkfq => 0,
	      rfre => 6,
	      rsnp => -1,
	      sali => 5,
	      self => 1,
	      ssnp => 4
	      ); # and default is 3

my %CMD = (mkfa => \&makeGenomes,
	   mkfq => \&makeFastq,
	   self => \&selfAlignment,
	   extd => \&extendReference,
	   crnm => \&correctNames,
	   ssnp => \&ssahaSNP,
	   aspl => \&assessPlates,
	   alip => \&pruneAlignments,
	   sali => \&splitAlignments,
	   #hyps => \&generateHypotheses,
	   merg => \&mergeMatches,
	   crlo => \&createReadLayout,
	   alco => \&alignmentCoords,
	   rqpw => \&redoQualityPairwise,
	   clbs => \&combineLayoutByStrain,
	   slof => \&sortLayoutFiles,
	   esnp => \&extractSNPsPreImp,
	   esnr => \&extractSNPRates,
	   pmar => \&prepareMargaritaInput,
	   marg => \&runMargarita,
	   cmar => \&checkMargaritaOutputs,
	   #brln => \&findBranchLengths,
	   prim => \&prepareImputer,
	   impu => \&runImputer,
	   impx => \&runImputerForHardGaps,
	   stat => \&runStaticImputer, # for paper
	   smat => \&snpMatrix,
	   smim => \&runSNPMatrixImputer, # for speed
	   stra => \&strainAssemblies,
	   bugb => \&buildGbrowse,
	   cmgb => \&combineGBDBFiles,
	   bfqa => \&buildFastqAlignments,
	   rfre => \&runFregeneExperiment,
	   xnov => \&extractNovelSequence,
	   isnp => \&extractSNPsPostImp,
	   tran => \&translateGenes,
	   rsnp => \&reportSNPs);


# Commands in a set must all take the same arguments...
# We repeat from "marg" so we check skipped chromosomes again.
# We don't currently have esnp after slof, or isnp are impu,
# for pipeline efficiency

my %CMD_SET = (proc => 'crlo rqpw clbs slof esnp pmar marg prim impu',
	       # Proc with static imputation:
	       prcs => 'crlo rqpw clbs slof esnp pmar smat marg prim impu smim',
	       prc0 => 'rqpw clbs slof esnp pmar',
	       prc1 => 'crlo rqpw clbs slof pmar',
	       # Same but with alco:
	       #prc1a => 'crlo alco rqpw clbs slof pmar',
	       prc2 => 'marg prim',
	       prc3 => 'impu isnp',
	       prc1s => 'crlo alco rqpw clbs slof');

sub sgrp {
    my ($pOpt,@args) = @_;
    setVariables($pOpt->{r},$pOpt->{i});
    my %info = (nFixed => \%NFIXED,
		nFixedDefault => 3,
		command => \%CMD,
		commandSet => \%CMD_SET,
		decide => \&decideBsubParametersSGRP,
		rootDir => $ROOT,
		shellCommand => 'sgrp.pl',
		instructionModule => dirname($0) . "/SGRP/SGRP.pm",
		callback => \&sgrp);
    expandAndSubmit($pOpt,\%info,@args);
}

sub decideBsubParametersSGRP {
    my ($cmd,@args) = @_;
    my $req;
    if (substr($ROOT,0,8) eq '/lustre/') {
	$req = 'select[lustre]';
    }
    my $fixedSwitches = "-r $ROOT";
    if (defined $FIXED_ITERATION) {
	$fixedSwitches .= " -i $FIXED_ITERATION";
    }
    my $fixedCmdPfx = "sgrp.pl $fixedSwitches $cmd";
    my $queue;
    # "merg" needs to be large for CERE chr12
    if (($JAVA_IMPUTER && $cmd eq 'impu') || 
	($cmd eq 'merg' && ($args[2] eq 'CERE' || $args[2] eq 'PARA'))) {
	$queue = 'long'; # was: 'bigmem';
	my $type = ($req eq 'select[lustre]') ? 'lustre' : 'linux';
	# NOTE: we need =2000 and not ==2000 here!!
	my $mem = 7000;
	$req = sprintf('select[%s && mem>=%d] rusage[mem=%d]',$type,$mem,$mem);
    } elsif ($cmd eq 'self' || ($cmd eq 'ssnp' && $args[-1] eq 'YONG')) {
	my $mem = $SSAHA2_MEMORY + 100;
	$req = sprintf("select[mem>=%d] rusage[mem=%d]",$mem,$mem);
	$queue = 'long'; # was: 'bigmem';
    } elsif ($cmd eq 'ssnp' && $req =~ /lustre/) {
	# Adam has deleted the lin32 binary ...
	$req = 'select[type==X86_64 || type==LINUX64]';
    } elsif (($cmd =~ /^pr.*c/ || $cmd eq 'smat') && $args[2] =~ /^FREG[1-9]00/) {
	$queue = 'long';
    } elsif (($cmd eq 'prc1' || $cmd eq 'crlo') && ($args[2] eq 'CERE' || $args[2] eq 'PARA')) {
	# We really only need this on the "final" iteration ...
	$queue = 'long';
    }
    return ($fixedCmdPfx,$queue,$req);
}

sub getStrains {
    my ($setName) = @_;
    my $ssFile = "$ALIASES/strainSets.txt";
    $ssFile = "$ENV{HOME}/strainSets.txt" unless (-s $ssFile);
    my $pAlias = readAliases($ssFile);
    unless (-s $ssFile) {
	confess("missing or empty: $ssFile");
    }
    unless (defined $pAlias->{$setName}) {
	# See if setName itself appears among any of the aliases.
	# If so, return it as a singleton strain set.
	foreach my $pA (values %$pAlias) {
	    return [$setName] if (grep {$_ eq $setName} @$pA);
	}
    }
    unless (defined $pAlias->{$setName}) {
	confess("no values defined in $ssFile for: $setName");
    }
    return $pAlias->{$setName};
}

1;
