package SGRP::SNPMatrix;

=head1 NAME

snpMatrix.pl

=head1 SYNOPSIS

=head1 DESCRIPTION

Reads all type.gz files under "directory", where "type" is set to "sorted"
by default (in a variable below). Creates output files:

    outPfx.strains, containing the N strains encountered
    outPfx.qK, for integer K = 0, 10, 20, ...

Each outPfx.qK file contains an NxN matrix Rate, where Rate[I][J] is
the SNP rate between strains I and J with quality threshold K.  The
SNP rate is the proportion of such positions at which these nucleotide
values differ. K=40 seems to give the most informative rates.

Example of use on farm:

    cd $SGR/match/CERE/cere/methp
    bsub -o smat.bout -e smat.berr -R'select[lustre]' \
       snpMatrix.pl sorted . rates.sorted

Note: the strains are not ordered alphabetically, but by similarity.
We first look for the two most dissimilar strains (highest value in
the matrix), and start with one of them, then successively select
the unselected strain with the lowest dissimilarity value to any
so-far-selected one. These gives nice displays on Matlab, like this:

    ssh hgs2
    /hgs2/matlab/bin/matlab
    File>Import Data
    imagesc(mymatrix)

=head1 AUTHOR

David Carter, C<dmc@sanger.ac.uk> , August 2006

=head1 SEE ALSO

    matrix2Phylip.pl for further processing and display of these matrices
    combineArgsRelatedness.pl for equivalent based on ARG topology

=cut

use Exporter;
@ISA = qw(Exporter);

@EXPORT = qw(snpMatrix addToSNPMatrix);

use SGRP::Variables;
use SGRP::SGRP;

use Utility;
use File::Basename;
use File::Find;
use Cwd;
use Carp;

use strict;

my $BASE = 'sorted_z'; # change this here if required: sorted_z for final version of PALAS,
                       # but sorted for imputation paper

my $QUICKTREE = "$ENV{HOME}/software/quicktree_1.1/bin/quicktree";

my $BLOCK_SIZE = 25000; # was 5000 originally, but save file counts;
                        # 25k will do for wsz = 100k, for the paper.

# Should be 40 for ABI, 30 for Solexa (see th/0615);
# char is 'I' or '?' respectively.
my $QUAL_THRESHOLD_CHAR;
# For use in SNP rate calculation; see th/0615.
my $QUAL_THRESHOLD_B;

my %ERROR_PROB;

sub snpMatrix {
    my ($species,$filtMeth,$strainSet,@parts) = @_;
    unless (@parts) {
	confess("Usage: sgrp.pl smat cere methc CERE [0-9] ...");
    }
    ($QUAL_THRESHOLD_CHAR,$QUAL_THRESHOLD_B) = setQualThreshold($filtMeth,$strainSet);
    setErrorProbs();
    my ($iter,$topDir) = getCurrentIteration($species,$filtMeth,$strainSet,'p?');
    my $outPfx = "$topDir/snpMatrix/$BASE";
    my $pSynPos = readSynonymousPositions("$ROOT/ref/$species/4waysyn.txt");
    if (0 && -s "$outPfx.bl") {
	report("already exists: $outPfx.bl",1);
    } else {
	snpMatrixActive($topDir,$pSynPos,@parts);
	my ($pNum,$pDen) = snpMatrixInactive($topDir);
	snpMatrixFinal($strainSet,$pSynPos,$outPfx,$pNum,$pDen) if ($pNum);
    }
}

sub setErrorProbs {
    $ERROR_PROB{' '} = 0;
    foreach my $q (0 .. 94) {
	my $qc = chr($q+33);
	$ERROR_PROB{$qc} = exp(-$q*0.1*log(10));
    }
}

sub setQualThreshold {
    my ($filtMeth,$strainSet) = @_;
    if ($filtMeth eq 'noerrs') {
	return ('~',0); # actually anything should work for the char
    } else {
	my $qth = ($strainSet =~ /^FREG.*S$/) ? 30 : 40;
	my $pe = exp(-0.1*$qth*log(10));
	return (chr($qth+ord('!')), 2*$pe*(1-$pe));
    }
}

sub readSynonymousPositions {
    my ($f) = @_;
    return undef unless (-s $f);
    my $fh = openToRead($f);
    my %synPos;
    while (<$fh>) {
	chomp;
	my ($chr,@list) = split;
	$synPos{$chr} = \@list;
    }
    return \%synPos;
}

# Process actively for the part(s) we've been given: create the
# summary files if they don't exist.

sub snpMatrixActive {
    my ($topDir,$pSynPos,@parts) = @_;
    my %given = map {$_ => 1} @parts;
    foreach my $dir (glob("$topDir/p?")) {
	my $p = substr($dir,-1);
	print $p;
	my $active = $given{$p};
	foreach my $chrDir (glob("$dir/*")) {
	    addToSNPMatrix($chrDir,$pSynPos,$active);
	}
    }
    print "\n";
}

sub addToSNPMatrix {
    my ($dir,$pSynPos,$active,$pNum,$pDen) = @_;
    my $f = "$dir/$BASE.gz";
    my $g = "$dir/$BASE.smat";
    return 0 unless (-s $f && ! -f "$f.touch"); # not ready
    createSmatFile($dir,$f,$g,$pSynPos) if ($active);
    if (-s $g && ! -f "$g.touch") {
	addToSNPMatrix1($g,$pNum,$pDen) if ($pNum && $pDen);
	print ".";
	return 1; # success
    } else {
	print "!";
	return 0; # missed one
    }
}

# e.g. $f = sorted.gz -> $g = sorted.smat, via $dir/smat/ files

sub createSmatFile {
    my ($dir,$f,$g,$pSynPos) = @_;
    return if (-s $g); # may be done already

    # Create the smat/blockNNN files if necessary
    my $smatDir = "$dir/smat";
    mayPrepareSmatInputs($f,$smatDir);
    return if (-s $g); # may be done by now

    # Try to create all the counts files; return unless they're
    # all done by the time we finish (so only the last job does
    # the last stage).
    return unless (createSmatCountsFiles($smatDir,$pSynPos));

    if (stamp($g)) {
	createSmatCumulFiles($smatDir,$g);
	createWindowAndMarkerFiles($dir,$g);
	# Finally create $g: final cumulative-counts file is the one for the whole chr 
	my $ccFile = (glob("$smatDir/cumul???"))[-1];
	link($ccFile,$g); 
	unstamp($g);
    }
}

# smatDir may be in one of three states:
#  -- does not exist: create it, and the blockNNN files in it.
#  -- smatDir/creating.touch exists: wait up to 2 mins for it to disappear
#  -- smatDir exists, no creating.touch: assume block files exist (check first!)

sub mayPrepareSmatInputs {
    my ($f,$smatDir) = @_;
    my $creating = "$smatDir/creating";
    if (-f "$smatDir/block001") {
	return; # assume all done
    } elsif (-f "$creating.touch") {
	# Wait for 2 minutes max for the block files to be created
	for (my $n=0; $n<24 && -f "$creating.touch"; $n++) {
	    sleep(5);
	}
	die("still exists: $creating.touch") if (-f "$creating.touch");
    } elsif (stamp($creating)) {
	# We need to create the block files...
	report("creating $smatDir/blockNNN");
	prepareSmatInputs($f,$smatDir);
	report();
	unstamp($creating);
    } else {
	die("cannot stamp $creating");
    }
}

# Given e.g. $f = sorted.gz, create block files in $dir. This is fairly quick
# per block, but is best done by a single sequential process.

sub prepareSmatInputs {
    my ($f,$dir) = @_;
    my $fh = openToRead($f);
    my $hdr = <$fh>; 
    my $n=1; # block 1 is from vertices 0 to 1.
    my $gPrev;
    my ($g,$gh) = createBlockFile($dir,$n,$hdr);
    my $done=0;
    while (<$fh>) {
	print $gh $_;
	$done++;
	if ($done == $BLOCK_SIZE) {
	    $done = 0;
	    $n++;
	    $gPrev = $g;
	    ($g,$gh) = createBlockFile($dir,$n,$hdr);
	}
    }
    $gh->close();
    if ($done < $BLOCK_SIZE/2 && $gPrev) { 
        # then we want to concat the last $g to the
	# previous one (discarding the header).
	`tail +2 $g >> $gPrev`;
	mayUnlink($g);
	$n--;
    }
    return $n; # number of (non-empty) blocks created
}

# Coordinates of block, cumul and span files:
#
# With block size B (25k by default), vertex V is at position V*B.
# V=0,1,2,...VMax, where VMax = round(aliSz/B); the final vertex
# is at aliSz-1 (i.e. the last position) not V*B. That is, the
# final block is of length between B/2 and 3B/2, so that the blocks
# cover the whole chromosome and the final one is a sensible length.
#
# blockVVV is portion of sorted.gz between (V-1)B and B (V=1,2,...).
#
# countsVVV has counts between vertices V-1 and V (V=1,2,...).
#
# cumulVVV has counts between vertices 0 and V (V=0,2,...),
# so that cumul000 is empty (all zeros), and for V>0 and U=V-1,
# cumulVVV = cumulUUU + countsVVV.
#
# span_UUU_VVV.matrix is matrix between vertices U and V,
# calculated from cumulVVV - cumulUUU.
#
# Markers (to delimit where trees apply) are indexed according to
# window size W = 2BR where R must be an integer. Marker
# M is at position M*W/2 + W/4. Between markers M and M+1 we are
# closest to the middle of the window from M*W/2 to (M+2)*W/2 =
# MBR to (M+2)BR, i.e. we use span_MR_(M+2)R. The final window
# span is from nR to VMax, where n+2 = round(VMax/R). Max marker
# index mMax is therefore n+1, i.e. mMax = round(VMax/R)-1.
#
# If B = 25k, for W = 50k we have R=1, so:
#
# 0         25k        50k        5k        100k       125k
# V=0       V=1        V=2        V=3        V=4        V=5
# |----------|----------|----------|----------|----------|
#       |          |          |          |          |
#      M=0        M=1        M=2        M=3        M=4
#  <---- span_0_2 -----> <---- span_2_4 ----->
#             <---- span_1_3 -----> <---- span_3_5 ----->
#
# For W=100k we have R=2, so:
#
# 0         50k        00k      150k       200k       250k
# V=0       V=2        V=4        V=6       V=8       V=10
# |----------|----------|----------|----------|----------|
#       |          |          |          |          |
#      M=0        M=1        M=2        M=3        M=4
#  <---- span_0_20-----> <--- span_20_40 ---->
#             <--- span_10_30 ----> <--- span_30_50 ---->

# Span files created (and then perhaps mostly deleted) are, e.g.:
#
# outputSNPMatrices:
#   span_000_004.strains  strain names, ordered by relatedness
#   span_000_004.q40      rate matrix, organized by strain names
#   span_000_004.matrix   = paste q40 strains
# createQuicktreeInput:
#   span_000_004.alpha    strain count, then q40 rate matrix
#                           multiplied up by 1000, and with
#                           each row STARTING with strain name.
# quicktree program (on alpha file)
#   span_000_004.tree
# qtToBlTree (on tree file)
#   span_000_004.bl       strain names are alphabetical but with REF first


sub createBlockFile {
    my ($dir,$n,$hdr) = @_;
    my $g = "$dir/block" . trio($n);
    my $gh = openToWrite($g);
    print $gh $hdr;
    return ($g,$gh);
}

sub trio {
    my ($n) = @_;
    my $trio = sprintf("%3d",$n);
    $trio =~ tr/ /0/;
    return $trio;
}

sub mayUnlink {
    my ($f) = @_;
    #unlink $f;
}

sub createSmatCountsFiles {
    my ($smatDir,$pSynPos) = @_;
    # Collect block001 ... blockNNN
    my @blockFiles = glob("$smatDir/block???");
    my @countsFiles;
    my $chr = basename(dirname($smatDir));
    # Create (where necessary) counts001 .. countsNNN
    foreach my $inFile (@blockFiles) {
	my $outFile = $inFile;
	$outFile =~ s/block(...)$/counts\1/;
	my $pSynList = $pSynPos && $chr ? $pSynPos->{$chr} : undef;
	mayCreateSmatCountsFile($inFile,$outFile,$pSynList);
	push @countsFiles,$outFile;
    }
    # See if all countsNNN files are ready; only the last-to-finish job will
    # do anything from here.
    foreach my $f (@countsFiles) {
	unless (-f $f && ! -f "$f.touch") {
	    return 0; # not all done
	}
    }
    return 1; # all done
}

# Lines in ...smat file are of the form
#   strain1 strain2 numAll denAll numSyn denSyn
# We create these files separately to save time.
# The counts are for pairs of quality $QUAL_THRESHOLD or above.

sub mayCreateSmatCountsFile {
    my ($f,$g,$pSynList,$isAlicat) = @_;
    if (stamp($g)) {
	report("$g: creating");
	my ($pStrains,$pNum,$pDen) = countReliableVariations($f,$pSynList,$isAlicat);
	writeSmatCountsFile($pStrains,$pNum,$pDen,$g);
	unstamp($g);
	report();
    }
}

sub countReliableVariations {
    my ($f,$pSynList,$isAlicat) = @_;
    my $fh = openToRead($f);
    $_ = <$fh>;
    chomp;
    s/^>//;
    my @strains = split;
    unless ($isAlicat) {
	@strains = ('REF',@strains);
    }
    my $nLines = 0;
    my $off = 1; # nucleotide offset, from 1
    my $ao = 0; # alignment offset, from 0
    my (@num,@den);
    while (<$fh>) {
	$nLines ++;
	chomp;
	if ($isAlicat) {
	    # Change back to nuc-and-qual format
	    unless (substr($_,16,1) eq '|' && substr($_,-1,1) eq '|') {
		die("Bad line (not vanilla alicat.pl format) in $f:\n$_\n");
	    }
	    $_ = substr($_,17);
	    s/\|$//;
	}
	my ($pKList,$pVals,$pEps) = valuesForGoodQualities($_);
	my $isSyn;
	if (/^[ACGT]/) {
	    if (defined @$pSynList && $pSynList->[0] == $off) {
		$isSyn = 1;
		shift @$pSynList;
	    }
	    $off ++;
	}
	for (my $i=0; $i<$#$pKList; $i++) {
	    my $ki = $pKList->[$i];
	    for (my $j=$i+1; $j<=$#$pKList; $j++) {
		my $kj = $pKList->[$j];
		my $isSNP = ($pVals->[$i] ne $pVals->[$j]) ? 1 : 0;
		my $pe = $pEps->[$i]+$pEps->[$j]*(1-$pEps->[$i]/3);
		$num[$ki][$kj][0] += $isSNP - $pe;
		$den[$ki][$kj][0] ++;
		if ($isSyn) {
		    $num[$ki][$kj][1] += $isSNP - $pe;
		    $den[$ki][$kj][1] ++;
		}
	    }
	}
    }
    return (\@strains,\@num,\@den);
}

sub valuesForGoodQualities {
    my ($line) = @_;
    my $kMax = (length($line)-2)/2;
    my (@kList,@vals,@eps);
    for (my $k=0; $k<=$kMax; $k++) {
	my $kk = $k+$k;
	my $v = substr($line,$kk,1);
	$v =~ tr/acgt/ACGT/ if ($k>0);
	if ($v =~ /^[ACGT]$/) {
	    my $qc = substr($line,$kk+1,1);
	    if ($qc eq ' ' || $qc ge $QUAL_THRESHOLD_CHAR) {
		push @vals,$v;
		push @eps,$ERROR_PROB{$qc};
		push @kList,$k;
	    }
	}
    }
    return (\@kList,\@vals,\@eps);
}

sub writeSmatCountsFile {
    my ($pStrains,$pNum,$pDen,$g) = @_;
    my $gh = openToWrite($g);
    foreach my $ki (0 .. $#$pDen) {
	my $pNum1 = $pNum->[$ki];
	my $pDen1 = $pDen->[$ki];
	my $si = $pStrains->[$ki];
	foreach my $kj ($ki+1 .. $#$pDen1) {
	    my $pNA;
	    $pNA = $pNum1->[$kj] if (defined $pNum1 && defined $pNum1->[$kj]);
	    my $pND = $pDen1->[$kj];
	    if ($pND) {
		my $sj = $pStrains->[$kj];
		printf($gh "%s %s",$si,$sj);
		foreach my $n01 (0,1) {
		    printf($gh " %8.6f %d",
			   (defined $pNA ? $pNA->[$n01] : 0),
			   $pND->[$n01]);
		}
		print $gh "\n";
	    }
	}
    }
}

sub createSmatCumulFiles {
    my ($smatDir,$g) = @_;
    my %counts;
    report("creating $smatDir/cumulNNN");
    # Counts from 0 to 0 are zero, of course:
    touch("$smatDir/cumul000");
    # Do the non-trivial ones: 1 to Vmax:
    my @countsFiles = glob("$smatDir/counts???");
    foreach my $f (@countsFiles) {
	createCumulativeCounts($f,\%counts);
    }
    report();
    # May remove block and count files:
    mayUnlink(glob("$smatDir/block???"));
    mayUnlink(@countsFiles);
}

sub createCumulativeCounts {
    my ($f,$pCounts) = @_;
    my $g = $f;
    $g =~ s/counts(...)$/cumul\1/;
    if (-s $f) {
	my $fh = openToRead($f);
	while (<$fh>) {
	    chomp;
	    my ($s1,$s2,@nums) = split;
	    my $pC = $pCounts->{$s1}{$s2};
	    if (defined $pC) {
		foreach my $i (0 .. $#nums) {
		    $pC->[$i] += $nums[$i];
		}
	    } else {
		$pCounts->{$s1}{$s2} = \@nums;
	    }
	}
	$fh->close();
    }
    # We write separately to make sure we get everything in pCounts
    # even if it didn't appear in $f
    my $gh = openToWrite($g);
    foreach my $s1 (sort keys %$pCounts) {
	my $pH1 = $pCounts->{$s1};
	foreach my $s2 (sort keys %$pH1) {
	    my $pA = $pH1->{$s2};
	    if (defined $pA) {
		printf($gh "%s %s %s\n",
		       $s1,$s2,join(" ",@$pA));
	    }
	}
    }
    $gh->close();
}

sub createWindowAndMarkerFiles {
    my ($dir,$g) = @_;
    # Create matrices
    my $pStrains = allStrainNames("$dir/$BASE.gz");
    foreach my $wsz qw(100 inf 500 200 100 50) { # 20 10) {
	createWindowBLFile("$dir/smat",$pStrains,$wsz);
    }
}

sub allStrainNames {
    my ($f) = @_;
    my $fh = openToRead($f);
    my $hdr = <$fh>;
    chomp $hdr;
    return ['REF',split(" ",substr($hdr,1))];
}

# wsz is window size in kb, or 'inf' to span the whole
# chromosome.

sub createWindowBLFile {
    my ($smatDir,$pStrains,$wsz) = @_;
    my $dir = dirname($smatDir);
    my $g = "$dir/windowed_${wsz}_bl";
    return unless (stamp($g));
    report("creating $g");
    my $gh = openToWrite($g);

    # cumul000 to cumulNNN, for vertices 0 to N (cumul000 is empty).
    my @cumuls = glob("$smatDir/cumul???");
    my $vMax = $#cumuls;

    # $span is number of blocks spanned by a window ("2R" in comment
    # block above). $step is number of blocks between
    # each pair of adjacent markers, i.e. half the window size ("R" in
    # comment block above). $mMax is maximum marker value, for end of
    # last window. We'll have a bit of material before the first marker
    # and after the last; the imputer can cope with this correctly.
    my ($span,$step,$mMax);
    if ($wsz eq 'inf') {
	$span = $step = $vMax;
	$mMax = 1;
    } else {
	$span = $wsz*1000/$BLOCK_SIZE;
	if ($span > $vMax) {
	    # treat the same as inf:
	    $span = $step = $vMax;
	    $mMax = 1;
	} else {
	    $step = $span/2;
	    $mMax = int(0.5+$vMax/$step)-1;
	}
    }
    
    # Write the trees. mLo < mMax because mHi is mLo+1
    for (my $mLo = 0; $mLo < $mMax; $mLo ++) {
	my $vLo = $mLo*$step;
	my $vHi = ($mLo+1 == $mMax) ? $vMax : $vLo+$span;
	my @a = windowBLTreeArray($smatDir,$pStrains,$mLo,$vLo,$vHi);
	print $gh "@a\n";
    }
    $gh->close();
    my $mh = openToWrite("$dir/markers_${wsz}");
    printf($mh "%d %d\n",scalar(@$pStrains),$mMax+1);
    if ($mMax == 1) {
	my $offMax = $vMax*$BLOCK_SIZE;
	print $mh "0\n$offMax\n";
    } else {
	my $mult = $wsz*500;
	my $off = $wsz*250;
	for (my $m=0; $m<=$mMax; $m++) {
	    print $mh "$off\n";
	    $off += $mult;
	}
    }
    $mh->close();
    unstamp($g);
    report();
}

sub windowBLTreeArray {
    my ($smatDir,$pStrains,$mLo,$vLo,$vHi) = @_;
    my ($pNum,$pDen) = cumulDiffCounts($smatDir,$pStrains,$vLo,$vHi);
    my $pfx = sprintf("%s/span_%s_%s",$smatDir,trio($vLo),trio($vHi));
    outputSNPMatrices($pfx,0,$pNum,$pDen,$pStrains);
    createNJTree($pfx);
    # and re-read it...
    my $fh = openToRead("$pfx.bl");
    my $line = <$fh>;
    $fh->close();
    chomp $line;
    my @a = split(" ",$line);
    $a[0] = $mLo;
    $a[2] = $mLo+1;
    # Delete the alpha, bl and tree files; paste the q40 and strains files into one.
    # NOTE that q40 is an anachronism; the qual is $QUAL_THRESHOLD i.e. 30 for Solexa.
    `paste $pfx.q40 $pfx.strains > $pfx.matrix`;
    foreach my $suf (qw(alpha bl tree q40 strains)) {
	mayUnlink("$pfx.$suf");
    }
    return @a;
}

sub cumulDiffCounts {
    my ($smatDir,$pStrains,$vLo,$vHi) = @_;
    my ($pNumLo,$pDenLo) = readNumAndDen($smatDir,$vLo,$pStrains);
    my ($pNumHi,$pDenHi) = readNumAndDen($smatDir,$vHi,$pStrains);
    my (%num,%den);
    foreach my $s1 (keys %$pDenLo) {
	foreach my $s2 (keys %{$pDenLo->{$s1}}) {
	    $num{$s1}{$s2}[0] = $pNumHi->{$s1}{$s2} - $pNumLo->{$s1}{$s2};
	    $den{$s1}{$s2}[0] = $pDenHi->{$s1}{$s2} - $pDenLo->{$s1}{$s2};
	}
    }
    return (\%num,\%den);
}

sub readNumAndDen {
    my ($dir,$m,$pStrains) = @_;
    my $f = sprintf("%s/cumul%s",$dir,trio($m));
    my $fh = openToRead($f);
    my (%num,%den,%seen);
    while (<$fh>) {
	chomp;
	my ($s1,$s2,$nc,$dc) = split;
	$num{$s1}{$s2} = $nc;
	$den{$s1}{$s2} = $dc;
	$seen{$s1} = $seen{$s2} = 1;
    }
    foreach my $s (@$pStrains) {
	unless ($seen{$s}) {
	    foreach my $ss (@$pStrains) {
		unless ($ss eq $s) {
		    my ($s1,$s2) = sort byStrainName ($s,$ss);
		    $den{$s1}{$s2} = 0 unless (defined $den{$s1}{$s2});
		}
	    }
	}
    }
    return (\%num,\%den);
}

sub byStrainName {
    $a eq 'REF' ? -1 : $b eq 'REF' ? 1 : $a cmp $b;
}

sub outputSNPMatrices {
    my ($outPfx,$syn01,$pNum,$pDen,$pStrains,@dList) = @_;
    my $pRates = findSNPRateMatrix($pNum,$pDen,$pStrains,$syn01,$outPfx);
    my @sList = orderKeysByRelatedness($pRates);
    #report("writing to $outPfx files ...");
    my $gh = openToWrite("$outPfx.strains");
    print $gh join("\n",@sList) . "\n";
    $gh->close();
    
    my $gh = openToWrite("$outPfx.q40");
    foreach my $i (0 .. $#sList) {
	my $si = $sList[$i];
	foreach my $j (0 .. $#sList) {
	    my $sj = $sList[$j];
	    print $gh ' ' if ($j);
	    printf($gh "%12.10f",$pRates->{$si}{$sj});
	}
	print $gh "\n";
    }
    #report();
}

sub findSNPRateMatrix {
    my ($pNum,$pDen,$pStrains,$syn01,$outPfx) = @_;
    my %rate;
    my @sorted = sort {$a cmp $b} @$pStrains;
    my %notSeen = map {$_ => 1} @sorted;
    foreach my $si (keys %$pDen) {
	my $pNumI = $pNum->{$si};
	my $pDenI = $pDen->{$si};
	foreach my $sj (keys %$pDenI) {
	    my $r = findSNPRate(defined $pNumI ? $pNumI->{$sj} : undef,
				   $pDenI->{$sj},$syn01);
	    if (defined $r) {
		$rate{$si}{$sj} = $rate{$sj}{$si} = $r;
		delete $notSeen{$si};
		delete $notSeen{$sj};
	    }
	}
    }
    # Complete, for missing ones
    if (keys %notSeen) {
	fillInMissingRates($pStrains,\%rate,$outPfx);
    }
    return \%rate;
}

sub findSNPRate {
    my ($pNum,$pDen,$syn01) = @_;
    my $sumD = $pDen->[$syn01];
    if ($sumD) {
	my $sumN = defined $pNum ? $pNum->[$syn01] : 0;
	my $r = $sumN/$sumD;
	# See th/0615 and (superceding it) th/1022, th/1023
	#my $rate1 = ($r-$QUAL_THRESHOLD_B)/(1-2*$QUAL_THRESHOLD_B);
	#my $rate2 = $r*3/4;
	#return max(0.00001,max($rate1,$rate2));  #4
	#return max(0.0000001,$rate1);            #3
	#return max(0.00001,$r);                 #5
	return max(0.0000001,$r);
    } else {
	return undef;
    }
}

sub fillInMissingRates {
    my ($pStrains,$pRate,$outPfx) = @_;
    my ($count,$sum);
    foreach my $si (keys %$pRate) {
	my $pRI = $pRate->{$si};
	foreach my $sj (keys %$pRI) {
	    $count ++;
	    $sum += $pRI->{$sj};
	}
    }
    #if ($count == 0) {
	#confess("Zero count in fillInMissingRates for $outPfx...");
    #}
    my $meanRate = $count > 0 ? $sum/$count : 0.001; # just a guess...
    foreach my $i (0 .. $#$pStrains-1) {
	my $si = $pStrains->[$i];
	foreach my $j ($i+1 .. $#$pStrains) {
	    my $sj = $pStrains->[$j];
	    unless (defined $pRate->{$si}{$sj}) {
		$pRate->{$si}{$sj} = $pRate->{$sj}{$si} = $meanRate;
	    }
	}
    }
}
     
sub orderKeysByRelatedness {
    my ($pRel) = @_;
    my @todo = keys %$pRel;
    # Find the LEAST related (highest pRel value) pair of strains, and start
    # with one of them.
    my ($max,$maxi);
    foreach my $i (0 .. $#todo-1) {
	foreach my $j ($i+1 .. $#todo) {
	    my $r = $pRel->{$todo[$i]}{$todo[$j]};
	    if ($r > $max) {
		$max = $r;
		$maxi = $todo[$j] eq 'REF' ? $j : $i;
	    }
	}
    }
    my @ans = ($todo[$maxi]);
    @todo = (@todo[0 .. $maxi-1],@todo[$maxi+1 .. $#todo]);
    while (@todo) {
	my $si = $ans[-1];
	@todo = sort {$pRel->{$si}{$a} <=> $pRel->{$si}{$b}} @todo;
	push @ans,shift @todo;
    }
    return @ans;
}

sub createNJTree {
    my ($pfx) = @_;
    #report("creating $pfx.bl");
    # We change UWOPS to UW to avoid exceeding name length limit in quicktree.
    # Also multiply all values by 1000 to avoid underflow in quicktree, and back again afterwards.
    my $alpha = createQuicktreeInput($pfx);
    `$QUICKTREE -in m -out t $pfx.alpha > $pfx.tree`;
    qtToBlTree("$pfx.tree","$pfx.bl");
    #report();
}

sub createQuicktreeInput {
    my ($pfx) = @_;
    my $fh = openToRead("$pfx.strains");
    my @strains = <$fh>;
    my $out = "$pfx.alpha";
    my $gh = openToWrite($out);
    print $gh scalar(@strains) . "\n";
    $fh = openToRead("$pfx.q40");
    while (<$fh>) {
	chomp;
	my @row = split;
	my $str = shift @strains;
	chomp $str;
	$str =~ s/UWOPS/UW/;
	print $gh $str;
	foreach my $v (@row) {
	    printf($gh " %8.6f",$v*1000);
	}
	print $gh "\n";
    }
    $gh->close();
    return $out;
}

# Grammar of tree:
#
#   Tree -> Node ;
#   Node -> Name
#   Node -> ( Branch , Branch )
#   Branch -> Node : Length

sub qtToBlTree {
    my ($f,$g) = @_;
    my $fh = openToRead($f);
    my @lines = <$fh>;
    chomp @lines;
    my $treeString = join("",@lines);
    $treeString =~ s/UW/UWOPS/g; # see above
    # Remove final semi-colon
    $treeString =~ s/;$//;
    # Parse using grammar; we divide lengths by 1000 because
    # we multiplied by 1000 earlier. We also replace zeroes
    # by 0.0000001.
    my $pT = parseNodeString($treeString,1);
    my @names = sortedStrainNames($pT);
    my %nameIdx;
    foreach my $i (0 .. $#names) {
	$nameIdx{$names[$i]} = $i;
    }
    numberTreeNodes($pT,\%nameIdx,scalar(@names));
    my $gh = openToWrite($g);
    my $sbl = sumBranchLengths($pT);
    printf($gh "0 0 1000000000 %10.8f",$sbl);
    my (@mthr,@len);
    fillMotherAndLength($pT,\@mthr,\@len);
    foreach my $i (0 .. $#mthr) {
	printf($gh " %8.6f %d",$len[$i]/$sbl,$mthr[$i]);
    }
    print $gh "\n";
    $gh->close();
}

# parseNodeString returns a compacted tree. A non-terminal
# node (e.g. the whole true) is an array reference of the form
#   [ident, ldtrLength, ldtrNode, rdtrLength, rDtrNode]
# A terminal node is of the form
#   [ident, name].
# Initially, the idents are undef.

sub parseNodeString {
    my ($str,$isTop) = @_;
    if (substr($str,0,1) eq '(') {
	die "Bad tree string (brackets): $str" unless (substr($str,-1) eq ')');
	my $str1 = substr($str,1,length($str)-2);
	my $pos = topPosition($str1,',');
	my $strA = substr($str1,0,$pos);
	my $strB = substr($str1,$pos+1);
	if (!$isTop) {
	    return ['NT',
		    parseBranchString($strA),
		    parseBranchString($strB)];
	} else {
	    my $pos2 = topPosition($strB,',');
	    my $strB1 = substr($strB,0,$pos2);
	    my $strB2 = substr($strB,$pos2+1);
	    my $pT = resolveTopNode(parseBranchString($strA),
				    parseBranchString($strB1),
				    parseBranchString($strB2));
	    return compactTree($pT);
	}
    } else {
	return ['T',$str];
    }
}

sub topPosition {
    my ($str,$ch) = @_;
    my $brack = 0;
    my $iMax = length($str)-1;
    for (my $i=0; $i<=$iMax; $i++) {
	my $c = substr($str,$i,1);
	if ($c eq '(') {
	    $brack ++;
	} elsif ($c eq ')') {
	    $brack --;
	} elsif ($c eq $ch && $brack == 0) {
	    return $i;
	}
    }
    die("Bad tree string (no top '$ch'): $str");
}

sub parseBranchString {
    my ($str) = @_;
    if ($str =~ /^(.*):([0-9\.]+)$/) {
	my ($sub,$len) = ($1,$2);
	return ['BR',$len,parseNodeString($sub)];
    } else {
	die("Bad branch string: $str");
    }
}

sub resolveTopNode {
    my ($pB0,$pB1,$pB2) = sort {$b->[1] <=> $a->[1]} @_;
    my $h12 = ($pB1->[1]+$pB2->[1])/2;
    my $hR = $pB0->[1] - $h12;
    return ['NT',
	    ['BR',$pB0->[1]-$hR,$pB0->[2]],
	    ['BR',$hR,['NT',$pB1,$pB2]]];
}	    

sub compactTree {
    my ($pT) = @_;
    if (scalar(@$pT) == 3) {
	return
	    [undef,
	     max(0.0000001,$pT->[1][1]/1000), # undo earlier mult by 1000
	     compactTree($pT->[1][2]),
	     max(0.0000001,$pT->[2][1]/1000),
	     compactTree($pT->[2][2])];
    } else {
	return [undef,$pT->[1]];
    }
}

sub sortedStrainNames {
    my ($pT) = @_;
    my @names = treeStrainNames($pT);
    return sort byStrainName  @names;
}

sub treeStrainNames {
    my ($pT) = @_;
    if (@$pT == 2) { # terminal node
	return ($pT->[1]);
    } else {
	return (treeStrainNames($pT->[2]),treeStrainNames($pT->[4]));
    }
}

sub numberTreeNodes {
    my ($pT,$pIdx,$next) = @_;
    if (@$pT == 2) {
	$pT->[0] = $pIdx->{$pT->[1]};
	return $next;
    } else {
	my $next1 = numberTreeNodes($pT->[2],$pIdx,$next);
	my $next2 = numberTreeNodes($pT->[4],$pIdx,$next1);
	$pT->[0] = $next2;
	return $next2+1;
    }
}

sub sumBranchLengths {
    my ($pT) = @_;
    if (@$pT == 2) {
	return 0;
    } else {
	return 
	    $pT->[1]
	    +$pT->[3]
	    +sumBranchLengths($pT->[2])
	    +sumBranchLengths($pT->[4]);
    }
}

sub fillMotherAndLength {
    my ($pT,$pMthr,$pLen) = @_;
    return if (@$pT == 2);
    $pMthr->[$pT->[2][0]] = $pT->[0];
    $pMthr->[$pT->[4][0]] = $pT->[0];
    $pLen->[$pT->[2][0]] = $pT->[1];
    $pLen->[$pT->[4][0]] = $pT->[3];
    fillMotherAndLength($pT->[2],$pMthr,$pLen);
    fillMotherAndLength($pT->[4],$pMthr,$pLen);
}

sub addToSNPMatrix1 {
    my ($g,$pNum,$pDen) = @_;
    my $fh = openToRead($g);
    while (<$fh>) {
	chomp;
	my ($si,$sj,$num0,$den0,$num1,$den1) = split;
	$pNum->{$si}{$sj}[0] += $num0 if ($num0);
	if ($den0) {
	    $pDen->{$si}{$sj}[0] += $den0;
	} elsif (!defined $pDen->{$si}{$sj}[0]) {
	    # make sure it's defined (this case should be rare)
	    $pDen->{$si}{$sj}[0] = 0;
	}
	$pNum->{$si}{$sj}[1] += $num1 if ($num1);
	$pDen->{$si}{$sj}[1] += $den1 if ($den1);
    }
}

# Process inactively (just read summary files, don't create) over all
# parts

sub snpMatrixInactive {
    my ($topDir,$pSynPos) = @_;
    my $missedOne;
    my (%num,%den);
    foreach my $dir (glob("$topDir/p?")) {
	print substr($dir,-1);
	foreach my $sub (glob("$dir/*")) {
	    if (-s "$sub/$BASE.gz") {
		$missedOne = 1 unless (addToSNPMatrix($sub,undef,0,\%num,\%den));
	    }
	}
    }
    print "\n";
    if ($missedOne) {
	return undef;
    } else {
	return (\%num,\%den);
    }
}

sub snpMatrixFinal {
    my ($strainSet,$pSynPos,$outPfx,$pNum,$pDen) = @_;
    unless (-s "$outPfx.strains") {
	my $pStrains = getOrInferStrains($strainSet);
	my $pStrainsAug = ['REF',@$pStrains];
	outputSNPMatrices($outPfx,0,$pNum,$pDen,$pStrainsAug);
	if ($pSynPos) {
	    outputSNPMatrices("$outPfx.syn",1,$pNum,$pDen,$pStrainsAug);
	}
    }
    createNJTree($outPfx);
}

sub getOrInferStrains {
    my ($strainSet) = @_;
    if ($strainSet =~ /^FREG(\d+)_/) {
	my $n = $1;
	return [1 .. $n];
    } else {
	return SGRP::SGRP::getStrains($strainSet);
    }
}

# $f should be the output of alicat.pl (with no -s, -p or -t).
sub snpMatrixOneShot {
    my ($outPfx,@files) = @_;
    my (%num,%den);
    foreach my $f (@files) {
	my $c = "$f.counts";
	mayCreateSmatCountsFile($f,$c,undef,1);
	addToSNPMatrix1($c,\%num,\%den);
    }
    my $pStrains = allKeys(\%den);
    outputSNPMatrices($outPfx,0,\%num,\%den,$pStrains);
    createNJTree($outPfx); 
}

sub allKeys {
    my ($pH) = @_;
    my %isKey;
    foreach my $key1 (keys %$pH) {
	$isKey{$key1} = 1;
	foreach my $key2 (keys %{$pH->{$key1}}) {
	    $isKey{$key2} = 1;
	}
    }
    my @keys = sort keys %isKey;
    return \@keys;
}

sub readStrainsFromHeader {
    my ($f) = @_;
    my $fh = openToRead($f);
    my $hdr = <$fh>;
    chomp $hdr;
    my @strains = ('REF',split(" ",substr($hdr,1)));
    return \@strains;
}
	    
1;
