package SGRP::SNPMatrixImputer;

use Exporter;
@ISA = qw(Exporter);

@EXPORT = qw(runSNPMatrixImputer);

# Note: this imputes using the SNP matrix created by sgrp.pl smat, which
# must have finished for ALL parts before smim is run for ANY part, because
# it creates sorted.smat for the whole genome.

use SGRP::Variables;

use Utility;
use File::Basename;

use Carp;

use strict;

my $FAC = 10/log(0.1);

my $DATA_CHUNK_SIZE = 10000;

sub runSNPMatrixImputer {
    my ($species,$filtMeth,$strainSet,@parts) = @_;
    unless (@parts) {
	confess("Usage: $0 smim species filtMeth strainSet part1 part2 ...");
    }
    my $pErr = errorProbHash();
    my ($iter,$dir) = getCurrentIteration($species,$filtMeth,$strainSet,'p?');
    my $pMat = readSNPMatrix("$dir/snpMatrix/sorted");
    return unless ($pMat);
    foreach my $part (@parts) {
	$part = "p$part" if ($part =~ /^[0-9]$/);
	foreach my $af (filesFromSmallest(glob("$dir/$part/*/sorted.gz"))) {
	    unless (-f "$af.touch") {
		runSNPMatrixImputer1(dirname($af),$pMat,$pErr);
	    }
	}
    }
}

sub errorProbHash {
    my %h;
    my $ml = -log(10);
    foreach my $q (0 .. 93) {
	$h{chr(33+$q)} = exp($ml*$q/10);
    }
    $h{' '} = exp($ml*12.7);
    return \%h;
}

sub readSNPMatrix {
    my ($pfx) = @_;
    my $f = "$pfx.strains";
    unless (-s $f) {
	warn("Missing or empty: $f");
	return undef;
    }
    my $fh = openToRead($f);
    my @strains = <$fh>;
    chomp @strains;
    $fh = openToRead("$pfx.q40");
    my %mat;
    my $i = 0;
    while (<$fh>) {
	chomp;
	my @vals = split;
	foreach my $j (0 .. $#vals) {
	    $mat{$strains[$i]}{$strains[$j]} = $vals[$j];
	}
	$i ++;
    }
    return \%mat;
}

sub runSNPMatrixImputer1 {
    my ($dir,$pMat,$pErr) = @_;
    my $src = "$dir/sorted.gz";
    my $tgt = "$dir/imputedx.gz"; # "x" here is for "SNP-matriX"
    if (! -s $src) {
	report("missing or empty: $src",1);
	return;
    } elsif (-f "$src.touch") {
	report("not finished yet: $src",1);
	return;
    } elsif (-s $tgt) {
	report("already done: $tgt",1);
	return;
    } else {
	my $chunk = 0; # do in 10k chunks
	my $fh = openToRead($src);
	my $hdr = <$fh>;
	chomp $hdr;
	my @pieces;
	while (1) {
	    my $tgtForCh = sprintf("$dir/impm/imputedm_%3d",$chunk);
	    $tgtForCh =~ tr/ /0/;
	    my $pData = readDataChunk($fh);
	    last unless ($pData);
	    if (-s $tgt) {
		report("now present: $tgt",1);
		return;
	    }
	    if (stamp($tgtForCh)) {
		report("creating $tgtForCh");
		runSNPMatrixImputer2($hdr,$pData,$pMat,$pErr,$tgtForCh);
		unstamp($tgtForCh,1);
		report();
	    }
	    push @pieces,$tgtForCh;
	    $chunk ++;
	}
	# See if we've got everything; if so, try and create imputedm.gz
	foreach my $f (@pieces) {
	    if (-f "$f.touch" || ! -f $f) {
		report("not done with $f yet",1);
		return; # not ready
	    }
	}
	# We can create it ...
	if (stamp($tgt)) {
	    my $adj = "$dir/adjusted.gz";
	    report("creating $tgt and adjusted.gz");
	    my @list = allStrainList($pMat);
	    shift @list;
	    createStrainImputedAndAdjusted(\@pieces,$src,$tgt,$adj,join(" ",@list));
	    unstamp($tgt,1);
	    report();
	}
    }
}

sub readDataChunk {
    my ($fh) = @_;
    my $nToDo = $DATA_CHUNK_SIZE;
    my @lines;
    while (<$fh>) {
	push @lines,$_;
	$nToDo --;
	last unless ($nToDo);
    }
    if (@lines) {
	return \@lines;
    } else {
	return undef;
    }
}

sub runSNPMatrixImputer2 {
    my ($hdr,$pData,$pMat,$pErr,$tgtForCh) = @_;
    my $gh = openToWrite($tgtForCh);
    # Strains actually in the sorted.gz file:
    my @strains = ('REF',split(/ /,substr($hdr,1)));
    # Strains referenced in $pMat (should be a superset):
    my @allStrains = allStrainList($pMat);
    # Map from strains to allStrains:
    my @map;
    foreach my $i (0 .. $#strains) {
	my $si = $strains[$i];
	foreach my $j (0 .. $#allStrains) {
	    if ($allStrains[$j] eq $si) {
		$map[$i] = $j;
		last;
	    }
	}
    }
    # could do this earlier!
    my @mat;
    foreach my $i (0 .. $#allStrains) {
	my $si = $allStrains[$i];
	foreach my $j (0 .. $#allStrains) {
	    my $sj = $allStrains[$j];
	    $mat[$i][$j] = $pMat->{$si}{$sj};
	}
    }
    my $mappedLength = 2*scalar(@allStrains);
    foreach my $line (@$pData) {
	chomp $line;
	$line = mapLineToAllStrains($line,\@map,$mappedLength);
	printf($gh "%s\n",snpMatrixImputeLine($line,\@mat,$pErr));
    }
    $gh->close();
}

sub allStrainList {
    my ($pMat) = @_;
    return sort {$a eq 'REF' ? -1 : $b eq 'REF' ? 1 : $a cmp $b} keys %$pMat;
}

sub mapLineToAllStrains {
    my ($line,$pMap,$len) = @_;
    my $mapped = ' ' x $len;
    for (my $i=0; $i<=$#$pMap; $i++) {
	my $j = $pMap->[$i];
	substr($mapped,$j*2,2) = substr($line,$i*2,2);
    }
    return $mapped;
}
    

sub snpMatrixImputeLine {
    my ($line,$pMat,$pErr) = @_;
    my $iMax = length($line)/2 - 1;
    while ($iMax < $#$pMat) {
	$line .= '  ';
	$iMax ++;
    }
    my (@val,@pe);
    for (my $i=0; $i<=$iMax; $i++) {
	my $v = substr($line,$i*2,1);
	unless ($v eq ' ') {
	    if ($v > 0) {
		$v = '-';
	    } else {
		$v =~ tr/acgt/ACGT/;
	    }
	    $val[$i] = $v;
	    $pe[$i] = $pErr->{substr($line,$i*2+1,1)};
	}
    }
    my $ans;
    # $i is position to impute
    for (my $i=0; $i<=$iMax; $i++) {
	$ans .= snpMatrixImputeValueAndQuality($i,\@val,\@pe,$pMat,$iMax);
    }
    return $ans;
}

sub snpMatrixImputeValueAndQuality {
    my ($i,$pVal,$pE,$pMat,$iMax) = @_;
    my (%minPErr,%minPMut);
    for (my $j=0; $j<=$iMax; $j++) {
	if (my $v = $pVal->[$j]) {
	    my $pMut = $pMat->[$j][$i];
	    $minPMut{$v} = min($minPMut{$v},$pMut);
	    my $pErr = $pE->[$j];
	    $minPErr{$v} = min($minPErr{$v},$pErr);
	}
    }
    my %minPE;
    foreach my $v (keys %minPMut) {
	my $pMut = $minPMut{$v};
	my $pErr = $minPErr{$v};
	$minPE{$v} = $pMut + $pErr - $pMut*$pErr;
    }
    my @keys = sort {$minPE{$a} <=> $minPE{$b}} keys %minPE;
    my $v = $keys[0]; # winning value
    # Lower-case if different from input value, to show it's
    # imputed:
    my $p = $minPE{$v};
    $p /= $minPE{$keys[1]} if (@keys > 1);
    $v =~ tr/ACGT/acgt/ unless ($v eq $pVal->[$i]);
    my $qn = 33 + int(0.5+($p && $FAC*log($p)));
    return $v . ($qn > 126 ? ' ' : chr($qn));
}    

sub createStrainImputedAndAdjusted {
    my ($pPieces,$src,$tgt,$adj,$strains) = @_;
    my $fh = openToRead($src);
    my $hdr = <$fh>;
    my $gh = openToWrite($tgt);
    my $ah = openToWrite($adj);
    #print $gh $hdr;
    #print $ah $hdr;
    print $gh ">$strains\n";
    print $ah ">$strains\n";
    foreach my $f (@$pPieces) {
	my $ph = openToRead($f);
	while (<$ph>) {
	    print $gh $_;
	    my $sortedLine = <$fh>;
	    my $adjLine = createAdjustedLine($sortedLine,$_);
	    print $ah $adjLine;	    
	}
    }
    # Delete them all afterwards, in case we fall over while copying
    my $dir = dirname($pPieces->[0]);
    if ($dir && -d $dir) {
	`rm -rf $dir`;
    }
}

# We want a line (with \n at the end) that is the same as sLine but has
# quality AND NUCLEOTIDE characters (for present values only) from iLine
sub createAdjustedLine {
    my ($sLine,$iLine) = @_;
    my $aLine = $sLine;
    my $kMax = length($sLine)-2;
    my $gapChar;
    for (my $k=0; $k<=$kMax; $k+=2) {
	unless (substr($aLine,$k,1) eq ' ') {
	    my $pair = substr($iLine,$k,2);
	    # We must change '-' to some gap-char NUMBER, so let's use 1
	    # as a guess...
	    $pair =~ s/^-/1/;
	    substr($aLine,$k,2) = $pair;
	}
    }
    return $aLine;
}

1;
