package SGRP::SelectNOAs;

use Exporter;
@ISA = qw(Exporter);

@EXPORT = qw(selectNonOverlappingAlignments);

use SGRP::Alignment;
use SGRP::Variables;

use Utility;
use File::Basename;

use Carp;

use strict;

# If set, we don't even allow alignments of simple clusters to overlap.
my $STRICT_NON_OVERLAPS = 1;

my $LOG = 1;

sub selectNonOverlappingAlignments {
    my ($chr,$pSrcDirs,$pOM,$level,$noaFile,$logFile) = @_;
    if (-s $noaFile) {
	report("already exists: $noaFile",1);
    } else {
	report("reading strain-specific alilines and clusters files");
	my (%alis,%cScore,%break);
	foreach my $dir (@$pSrcDirs) {
	    readAlignments("$dir/alilines.pass.txt",$chr,$pOM,$level,\%alis,\%break);
	    readClusterScores("$dir/clusters.txt",\%cScore);
	    increaseAuxClusterScores(\%alis,\%cScore);
	}
	report();
	report("creating $noaFile");
	my $gh = openToWrite($noaFile);
	my $lh;
	if ($LOG) {
	    $lh = openToWrite($logFile);
	    autoflush $lh 1;
	}
	foreach my $chr (sort keys %alis) {
	    selectNonOverlappingAlignments1($chr,$alis{$chr},\%cScore,$break{$chr},
					    $level,$gh,$lh);
	}
	report();
    }
    return $noaFile;
}

# Pass $pAlis as an argument because we read into it from lots of
# different files.

sub readAlignments {
    my ($f,$chr,$pOM,$level,$pAlis,$pBreak) = @_;
    my $fh = openToRead($f);
    my %aliList; # alignments per read
    while (<$fh>) {
	#print;
	chomp;
	my $pA = new Alignment($_,$pOM);
	if ($pA) {
	    my $lo = $pA->clo($level);
	    if ($pA->clo($level) < $pA->chi($level)) {
		$lo -= 0.5;
	    }
	    if ($pA->chr(2) eq $chr) {
		push @{$pAlis->{$pA->chr($level)}{$lo}},$pA;
		push @{$aliList{$pA->name}},$pA;
	    }
	} else {
	    # OK, fails to span break though.
	    # die("Cannot form Alignment object from $_");
	}
    }
    foreach my $name (keys %aliList) {
	if ($#{$aliList{$name}} >= 1) {
	    my @arr = sort {$a->clo <=> $b->clo || $a->chi <=> $b->chi} @{$aliList{$name}};
	    my $chr = $arr[0]->chr($level);
	    for (my $i=0; $i<=$#arr; $i++) {
		my $pAli = $arr[$i];
		if ($pAli->score == 0) { # then it's an indel one
		    if ($i > 0) {
			my $bp = $pAli->clo($level);
			push @{$pBreak->{$chr}{$bp}},$name;
		    }
		    if ($i < $#arr) {
			my $bp = $pAli->chi($level);
			if ($bp > $pAli->clo($level) || $i == 0) {
			    # so this is a deletion; for an insertion, we only
			    # want one break, i.e. the one above, if we got it
			    # then (so we allow $i == 0 for the case when we didn't)
			    push @{$pBreak->{$chr}{$bp}},$name;
			}
		    }
		}
	    }
	}
    }
    
}

sub readClusterScores {
    my ($f,$pScore) = @_;
    my $fh = openToRead($f);
    my (%altScore,$clone,$isChimeric);
    while (<$fh>) {
	chomp;
	if (/ TM /) {
	    # then it's a status line ...
	    # p1k or q1k will be part of clone name.
	    if ($isChimeric) {
		$clone = undef;
	    }
	    my ($status,$scores);
	    ($clone,undef,$status,$scores) = split;
	    my $s;
	    if ($scores =~ /^\(-?(\d+),/) {
		$s = $1;
	    } else {
		confess("Bad line: $_");
	    }
	    if ($status =~ /_chi/) {
		$isChimeric = 1;
		# go to next line and look for p or q
		my $line = <$fh>;
		my $pq = substr($line,4,1);
		if ($pq eq 'p' || $pq eq 'q') {
		    $clone .= ".${pq}1k";
		} else {
		    confess("Bad line: $_");
		}
	    }
	    if ($status =~ /^accept/) {
		$pScore->{$clone} = $s;
	    } elsif ($s > $altScore{$clone}) {
		$altScore{$clone} = $s;
	    }
	} elsif ($clone && /^  [FC] ([pq]) /) {
	    #$npq{$1} ++;
	} elsif ($clone) {
	    $clone = undef;
	    $isChimeric = undef;
	}
    }
    # Use altScores where no accept found
    foreach my $clone (keys %altScore) {
	unless ($pScore->{$clone}) {
	    $pScore->{$clone} = $altScore{$clone};
	}
    }
}

# We want to encourage the acceptance of alignments to aux sequences
# so the deletions they imply get carried through (and can propagate
# to other strains).

sub increaseAuxClusterScores {
    my ($pAlis,$pCScore) = @_;
    my %toInc;
    foreach my $pH (values %$pAlis) {
	foreach my $pA (values %$pH) {
	  ALI:
	    foreach my $pAli (@$pA) {
		if ($pAli->chr =~ /\.\d\d\d$/) {
		    foreach my $id ($pAli->clone,$pAli->name) {
			if (defined $pCScore->{$id}) {
			    $toInc{$id} = 1;
			    next ALI;
			}
		    }
		    die("Cannot find $pCScore for " . $pAli->name);
		}
	    }
	}
    }
    foreach my $id (keys %toInc) {
	$pCScore->{$id} += 1000; # big amount, to really encourage it
    }
}

sub selectNonOverlappingAlignments1 {
    my ($chr,$pAlis,$pCScore,$pBreak,$level,$gh,$lh) = @_;
    # On final iteration, we DO allow overlaps, as there won't be another chance,
    # and the alignments have been pre-filtered to have no breaks etc.
    my $disallowOverlaps = $STRICT_NON_OVERLAPS && !$IS_FINAL_ITERATION;
    my @buf; # list of "current" alignments sorted by chi; elt 0 is 1 or 0 for whether accept
    my (%rejected,%isR);
    # List of alignment start points
    my @loList = sort {$a <=> $b} keys %$pAlis;
    # List of break points
    my @brList = sort {$a <=> $b} keys %$pBreak;
    # Allowance for slight alignment overruns
    my $allow = $disallowOverlaps ? 0 : 5;
    print $lh "# $chr: starting to select\n" if ($lh);
    # Go through @brList, marking reads with nearby breakpoints as inconsistent.
    # They're not really inconsistent, but if we accept several, we'll miss the
    # chance to align them with each other.
    foreach my $i (0 .. $#brList) {
	my $brI = $brList[$i];
	my @names1 = @{$pBreak->{$brI}};
	foreach my $j ($i .. $#brList) {
	    last if ($brList[$j]-$brI > $allow);
	    my @names2 = @{$pBreak->{$brList[$j]}};
	    foreach my $name1 (@names1) {
		unless (defined $rejected{$name1}) {
		    my ($r1,$s1) = clusterNameAndScore($name1,$pCScore);
		    foreach my $name2 (@names2) {
			my ($r2,$s2) = clusterNameAndScore($name2,$pCScore);
			recordRejection($r1,$s1,$r2,$s2,$brI,\%rejected,$lh);
		    }
		}
	    }
	}
	my @names1Pruned;
	foreach my $name1 (@names1) {
	    push @names1Pruned,$name1 unless (defined $rejected{$name1});
	}
	$pBreak->{$brI} = @names1Pruned ? \@names1Pruned : undef;
    }
    # Also mark inconsistency when one spans a region broken by another
    foreach my $lo (@loList) {
	# Discard all breaks before $lo
	while (@brList && $brList[0] <= $lo+$allow) {
	    shift @brList;
	}
	# $pA2 is an alignment starting at $lo
	foreach my $pA2 (@{$pAlis->{$lo}}) {
	    my $lo2 = $pA2->clo($level);
	    my $hi2 = $pA2->chi($level);
	    my ($r2,$s2) = clusterNameAndScore($pA2->name,$pCScore);
	    next if (defined $rejected{$r2});
	    foreach my $bp (@brList) {
		last if ($bp+$allow >= $hi2); # not relevant to this alignment
		foreach my $name1 (@{$pBreak->{$bp}}) {
		    # r1 has a break at bp and r2 spans bp, so they're inconsistent:
		    my ($r1,$s1) = clusterNameAndScore($name1,$pCScore);
		    recordRejection($r1,$s1,$r2,$s2,$bp,\%rejected,$lh);
		}
	    }
	    $isR{$r2} = 1;
	}
    }

    if ($disallowOverlaps) {
	foreach my $i (0 .. $#loList) {
	    my $loI = $loList[$i];
	    # $pAI is an alignment starting at $loI
	    foreach my $pAI (@{$pAlis->{$loI}}) {
		my $lo2I = $pAI->clo($level);
		my $hi2I = $pAI->chi($level);
		my ($rI,$sI) = clusterNameAndScore($pAI->name,$pCScore);
		my $strI = $rI;
		$strI =~ s/-.*//;
		next if (defined $rejected{$rI});
	      JLOOP:
		foreach my $j ($i .. $#loList) {
		    my $loJ = $loList[$j];
		    # $pAJ is an alignment starting at $loJ
		    foreach my $pAJ (@{$pAlis->{$loJ}}) {
			next if ($i == $j && $pAI >= $pAJ); # avoid self and double matches
			my $lo2J = $pAJ->clo($level);
			last JLOOP if ($lo2J > $hi2I);
			my $hi2J = $pAJ->chi($level);
			my ($rJ,$sJ) = clusterNameAndScore($pAJ->name,$pCScore);
			my $strJ = $rJ;
			$strJ =~ s/-.*//;
			if ($strI eq $strJ || ($disallowOverlaps && $STRICT_NON_OVERLAPS > 1)) {
			    # We only apply strict criterion when same strain, unless strictness >= 2
			    my $rej = recordRejection($rI,$sI,$rJ,$sJ,max($lo2I,$lo2J),\%rejected,$lh);
			    last JLOOP if ($rej == 1); # because $rI rejected, so go on to next $pAI
			}
		    }
		}
	    }
	}
    }

    my %accept;
    my @rList = keys %isR;
    foreach my $r1 (keys %isR) {
	unless (defined $rejected{$r1}) {
	    $accept{$r1} = 1;
	}
    }
    foreach my $lo (@loList) {
	foreach my $pA (@{$pAlis->{$lo}}) {
	    my ($clone,undef) = clusterNameAndScore($pA->name,$pCScore);
	    if ($accept{$clone}) {
		printf($gh "%s\n",$pA->toString);
	    }
	    #printf($gh "%d %s\n",$accept{$clone},$pA->[$#$pA]);
	}
    }
}

sub clusterNameAndScore {
    my ($name,$pCScore) = @_;
    unless (defined $pCScore->{$name}) {
	$name =~ s/\.[pq]1k$//;
    }
    return ($name,
	    $pCScore->{$name} || confess("No score for $name"));
}

sub recordRejection {
    my ($r1,$s1,$r2,$s2,$bp,$pRej,$lh) = @_;
    return if ($r1 eq $r2 || defined $pRej->{$r1} || defined $pRej->{$r2});
    my $d = $s1-$s2;
    # Break ties by alphabetical order:
    if ($d > 0 || ($d == 0 && $r1 lt $r2)) {
	$pRej->{$r2} = 1;
	print $lh "Rejecting $r2 because of $r1 (clashed at $bp)\n" if ($lh);
	return 2;
    } else {
	$pRej->{$r1} = 1;
	print $lh "Rejecting $r1 because of $r2 (clashed at $bp)\n" if ($lh);
	return 1;
    }
    return 0;
}

1;
