package SGRP::SelfAlignment;

use Exporter;
@ISA = qw(Exporter);

@EXPORT = qw(selfAlignment);

use SGRP::Variables;
use SGRP::Alignment;

use Utility;
use File::Basename;

use strict;

my $BIG_CHUNK_SIZE = 60000;
my $SMALL_CHUNK_SIZE = 30000;
# How close we can be to chunk boundary and survive
my $PADDING = 6;

# Was: /nfs/team71/psg/aws/bin/lin32_stable/ssaha2';
my $SSAHA2 = '/usr/local/ensembl/bin/ssaha2';

# For clustering aligned regions:
my $CAI_MIN_PERCI = 90; # min percentage identity for a match
my $CAI_MIN_REPFREQ = 2; # min frequency of a repeat to output
my $CAI_MAX_SLACK = 10; # max difference in start/end points to cluster

# Output: e.g. $SGR/ref/cere/self/chr01.{cere,para}.ali

sub selfAlignment {
    my ($species,@parts) = @_;
    unless (@parts > 0) {
	die("Usage: $0 self species part1 part2...");
    }
    my $refDir = "$ROOT/ref/$species";
    my $selfDir = "$refDir/self";
    my $refGenomeFa = "$ROOT/ref/$species/genome.fa";
    createSelfChoppedDir($refGenomeFa,$selfDir);
    my %isP = map {$_ => 1} @parts;
    foreach my $f (glob("$selfDir/chopped/*.fa")) {
	createAndPruneSelfAlignments($f,\%isP);
    }
    # Regions that align to other regions in the same genome,
    # and how many they align to.
    my $clusFile = clusterAlignedRegions($species,\%isP);
    if (defined $clusFile) {
	# Make an augmented assembly (same reference, but augmented
	# with possible deletions of cluster-regions) if possible and necessary.
	mayMakeAugmentedAssembly($refGenomeFa,$clusFile,$refDir);
	# Create transposable-element .fa file
	my $tefa = createTransposableElementFA($refDir);
	selfAlignTransposableElementFA($tefa) if ($tefa);
    }
}

# Create e.g. $SGR/ref/cere/self/chopped/chopped_$chr.fa

sub createSelfChoppedDir {
    my ($refGenomeFa,$selfDir) = @_;
    # Create self/chopped/ with query files for ssaha2 run.
    # "stamp" makes $selfDir if non-existent...
    if (! -d $selfDir && stamp("$selfDir/touch")) {
	chopGenome($refGenomeFa,"$selfDir/chopped");
	unstamp("$selfDir/touch",1);
    } else {
	# wait till all files are written...
	while (-f "$selfDir/touch.touch") {
	    sleep(5);
	}
    }
}

sub chopGenome {
    my ($refGenomeFA,$dir) = @_;
    my $fh = openToRead($refGenomeFA);
    my ($chr,$buf);
    while (<$fh>) {
	chomp;
	if (/^>/) {
	    chopAndPrint($chr,$buf,$dir) if ($buf);
	    $buf = "";
	    $chr = substr($_,1);
	} else {
	    $buf .= $_;
	}
    }
    chopAndPrint($chr,$buf,$dir) if ($buf);
}

sub chopAndPrint {
    my ($chr,$buf,$dir) = @_;
    my $gh = openToWrite("$dir/chopped_$chr.fa");
    my $len = length($buf);
    my $iMax;
    my $hscs = $SMALL_CHUNK_SIZE/2;
    for (my $i=0; $i<$len; $i+=$BIG_CHUNK_SIZE) {
	# Small chunk straddling $i
	if ($i > $hscs && $i < $len-$hscs) {
	    printChunk($chr,$buf,$i-$hscs,$SMALL_CHUNK_SIZE,$gh);
	}
	# Big chunk starting at $i
	printChunk($chr,$buf,$i,min($BIG_CHUNK_SIZE,$len-$i),$gh);
    }
}

sub printChunk {
    my ($chr,$buf,$i,$len,$gh) = @_;
    my $loc = sprintf("%s.%7d",$chr,$i); # $i is offset from start
    $loc =~ tr/ /0/;
    print $gh ">$loc\n";
    for (my $j=$i; $j<$i+$len; $j+=50) {
	print $gh substr($buf,$j,50) . "\n";
    }
}

# Input:   e.g. $SGR/ref/cere/self/chopped/chopped_chr01.fa
# Outputs: e.g. $SGR/ref/cere/self/chopped/chopped_chr01.cere.ali
#               $SGR/ref/cere/self/chr01.cere.ali (pruned version)

sub createAndPruneSelfAlignments {
    my ($f,$pWanted) = @_;
    my $p = substr($f,-4,1); # digit before the .fa
    $p = 0 unless ($p > 0);
    return unless $pWanted->{$p};
    foreach my $sp2 (qw(cere para)) {
	my $sbjct = "$ROOT/ref/$sp2/genome.fa";
	(my $g1 = $f) =~ s/\.fa$/.$sp2.ali/;
	(my $g = $g1) =~ s/chopped\/chopped_//;
	if (stamp($g1)) {
	    selfAlignFile($f,$sbjct,$g1);
	    unstamp($g1,1);
	}
	if (stamp($g)) {
	    pruneSelfAlignmentFile($g1,$g);
	    unstamp($g,1);
	}
    }
}

sub selfAlignFile {
    my ($chopped,$ref,$ali1,$ali) = @_;
    my $cmd = "$SSAHA2 $chopped $ref -memory $SSAHA2_MEMORY -align 0 -tags 1";
    report("running $cmd",1);
    report("writing $ali1");
    `$cmd > $ali1`;
    report();
}

sub pruneSelfAlignmentFile {
    my ($f,$g) = @_;
    report("writing $g");
    my $fh = openToRead($f);
    my $gh = openToWrite($g);
    my $hscs = $SMALL_CHUNK_SIZE/2;
    while (<$fh>) {
	chomp;
	my @a = split;
	next unless ($a[0] eq 'ALIGNMENT');
	my ($chr,$off) = split(/\./,$a[2]);
	$a[2] = $chr;
	my $pAli = new Alignment(@a[1..$#a]);
	# Discard if trivial (identity alignment)
	next if ($pAli->name eq $pAli->chr
		 && $pAli->fc eq 'F'
		 && $pAli->rlo+$off == $pAli->clo
		 && $pAli->rhi+$off == $pAli->chi);
	# Discard if goes too near non-terminal chunk boundary
	# Too near edge of chunk:
	next if ($off > 0 && $pAli->rlo < $PADDING);
	# Test for size: if it fails, we must be at a final chunk, so don't do this test
	if ($pAli->rlen == $BIG_CHUNK_SIZE || $pAli->rlen == $SMALL_CHUNK_SIZE) {
	    next if ($pAli->rlen-$pAli->rhi < $PADDING);
	}
	# For small chunk, must straddle middle, or we'll get same alignment from big:
	if ($pAli->rlen == $SMALL_CHUNK_SIZE) {
	    next unless ($pAli->rhi >= $hscs-$PADDING && $pAli->rlo <= $hscs+$PADDING);
	}
	# Switch to chromosomal offsets
	$a[4] += $off;
	$a[5] += $off;
	# Recreate $pAli with new offsets, and print
	$pAli = new Alignment(@a[1..$#a]);
	printf($gh "%s\n",$pAli->toString);
    }
    $gh->close();
    report();
}

# Creates (eventually) $ROOT/ref/$species/self/clusters.txt

sub clusterAlignedRegions {
    my ($species,$pWanted) = @_;
    my $selfDir = "$ROOT/ref/$species/self";
    my @aliFiles = glob("$selfDir/*.$species.ali");
    my @aliFilesToDo;
    foreach my $f (@aliFiles) {
	my $b = basename($f);
	my $p = substr($b,index($b,'.')-1,1);
	$p = 0 unless ($p > 0);
	push @aliFilesToDo,$f if ($pWanted->{$p});
    }
    my $pRegs = readRegions(\@aliFilesToDo);
    foreach my $chr (sort keys %$pRegs) {
	my $clusFile = "$selfDir/clusters.$chr.txt";
	if (stamp($clusFile)) {
	    report("creating $clusFile");
	    clusterAndPrint($chr,$pRegs->{$chr},$clusFile);
	    unstamp($clusFile,1);
	    report();
	}
    }
    foreach my $p (keys %$pWanted) {
	touch("$selfDir/doneCluster.$p");
    }
    my @clusFiles;
    foreach my $f (@aliFiles) {
	my $b = basename($f);
	my $chr = substr($b,0,index($b,'.'));
	my $clusFile = "$selfDir/clusters.$chr.txt";
	unless (-f $clusFile && ! -f "$clusFile.touch") {
	    report("returning because $clusFile is not ready",1);
	    return undef;
	}
	push @clusFiles,$clusFile;
    }
    # Return if not all parts done
    foreach my $p (0 .. 9) {
	unless (-f "$selfDir/doneCluster.$p") {
	    report("returning because (at least) cluster $p not finished",1);
	    return;
	}
    }
    my $clusFile = "$selfDir/clusters.txt";
    report("creating $clusFile");
    `cat @clusFiles > $clusFile`;
    #unstamp($clusFile);
    report();
    return $clusFile;
}

sub readRegions {
    my ($pFiles) = @_;
    my %regs;
    foreach my $f (@$pFiles) {
	my $fh = openToRead($f);
	while (<$fh>) {
	    chomp;
	    my @a = split;
	    if ($a[10] >= $CAI_MIN_PERCI) {
		my $chr = $a[2];
		my $lo = min($a[4],$a[5]);
		my $hi = max($a[4],$a[5]);
		push @{$regs{$chr}},[$lo,$hi,1];
		# Look for near-tandem repeat (to within 7k, i.e. to
		# include transposons)
		push @{$regs{$chr}},nearTandemRepeatTrios(@a);
	    }
	}
    }
    return \%regs;
}

sub nearTandemRepeatTrios {
    my (undef,undef,$chr1,$chr2,$lo1,$hi1,$lo2,$hi2,$fc) = @_;
    if ($fc eq 'F' && $chr1 eq $chr2) {
	my $d = $lo2-$hi1;
	if ($d > 0 && $d <= 7000) {
	    # Give "frequency" of $CAI_MIN_REPFREQ to ensure always used
	    return ([$lo1,$lo2-1,$CAI_MIN_REPFREQ],[$hi1+1,$hi2,$CAI_MIN_REPFREQ]);
	}
    }
    return ();
}    

sub clusterAndPrint {
    my ($chr,$pRegs,$g) = @_;
    my $pCRegs = clusterRegions($pRegs);
    my $gh = openToWrite($g);
    # Print results
    foreach my $pR (sort {$a->[0] <=> $b->[0] || $a->[1] <=> $b->[1]} @$pCRegs) { # ,@$pDRegs) {
	printf($gh "%s %7d %7d %3d %s\n",$chr,@$pR);
    }
}

sub clusterRegions {
    my ($pRegs) = @_;
    my @regs;
    foreach my $pA (sort {$a->[0] <=> $b->[0]} @$pRegs) {
	push @regs,[$pA->[0],$pA->[0],$pA->[1],$pA->[1],$pA->[2]];
    }
    while (1) {
	my ($dMin,$lolo,$lohi,$hilo,$hihi,$n,$iDel,$jDel) = ($CAI_MAX_SLACK+0.5);
	foreach my $i (0 .. $#regs-1) {
	    next unless (defined $regs[$i]);
	    my ($lolo0,$lohi0,$hilo0,$hihi0,$n0) = @{$regs[$i]};
	    foreach my $j ($i+1 .. $#regs) {
		next unless (defined $regs[$j]);
		my ($lolo1,$lohi1,$hilo1,$hihi1,$n1) = @{$regs[$j]};
		my $dLo = $lohi1-$lolo0;
		last if ($dLo >= $dMin);
		my $dHi = max(abs($hihi1-$hilo0),abs($hihi0-$hilo1));
		if ($dHi < $dMin) {
		    $dMin = max($dLo,$dHi);
		    $lolo = $lolo0; # must be = to: min($lolo0,$lolo1);
		    $lohi = max($lohi0,$lohi1);
		    $hilo = min($hilo0,$hilo1);
		    $hihi = max($hihi0,$hihi1);
		    $n = $n0+$n1;
		    $iDel = $i;
		    $jDel = $j;
		}
	    }
	}
	last unless (defined $lolo);
	$regs[$iDel] = [$lolo,$lohi,$hilo,$hihi,$n];
	$regs[$jDel] = undef;
    }
    my @finalRegs;
    foreach my $pR (@regs) {
	if (defined $pR && $pR->[4] >= $CAI_MIN_REPFREQ) {
	    my $lo = int(($pR->[0]+$pR->[1])/2);
	    my $hi = int(($pR->[2]+$pR->[3])/2);
	    push @finalRegs,[$lo,$hi,$pR->[4]];
	}
    }
    @finalRegs = sort {$a->[0] <=> $b->[0] || $a->[1] <=> $b->[1]} @finalRegs;
    return \@finalRegs;
}

sub mayMakeAugmentedAssembly {
    my ($refGenomeFa,$clusFile,$refDir) = @_;
    return unless ($clusFile);
    my $augFaFile = "$refDir/augmented.fa";
    my $augMapFile = "$refDir/augmented.map";
    if (stamp($augFaFile)) {
	report("creating $augFaFile and augmented.map");
	makeAugmentedAssembly($refGenomeFa,$clusFile,$augFaFile,$augMapFile);
	unstamp($augFaFile);
	report();
    }
}

# clusterFile should be output of clusteredAlignedRegions.pl 
# Sample call, used to create ref/cere/augmented.fa and augmented.map:
# SGRP::StrainAssemblies::referenceStrainAssembly(
#       'genome.fa','cere.clusters.txt','augmented.fa','augmented.map')
# where cere.clusters.txt is output of clusteredAlignedRegions.pl
#   running on ref/cere/self/*.cere.ali

sub makeAugmentedAssembly {
    my ($faFile,$clusterFile,$outFile,$mapFile) = @_;

    my (%seq,$chr,$buf);
    my $fh = openToRead($faFile);
    while (<$fh>) {
	chomp;
	if (/^>(\S+)/) {
	    $seq{$chr} = $buf if ($buf);
	    $buf = undef;
	    $chr = $1;
	} else {
	    $buf .= $_;
	}
    }
    $seq{$chr} = $buf if ($buf);

    my %hgr;
    my $fh = openToRead($clusterFile);
    while (<$fh>) {
	my ($chr,$lo,$hi) = split;
	push @{$hgr{$chr}},[$lo,$hi];
    }
    my $gh = openToWrite($outFile);
    my $mh = openToWrite($mapFile);
    foreach my $chr (sort keys %seq) {
	my $pSets = SGRP::StrainAssemblies::findAllHGRSets($chr,$hgr{$chr});
	SGRP::StrainAssemblies::printFormatted($gh,"REF.$chr",$seq{$chr});
	printf($mh "REF.%s 0 0 %d\n",$chr,length($seq{$chr}));
	SGRP::StrainAssemblies::writeAuxiliarySequences("REF.$chr",$pSets,$seq{$chr},undef,$gh,undef,$mh);
    }
}

sub createTransposableElementFA {
    my ($refDir) = @_;
    my $gffFile = "$refDir/genome.gff";
    my $faFile = "$refDir/genome.fa";
    my $teFile = "$refDir/te/te.fa";
    if (-s $gffFile && -s $faFile && stamp($teFile)) {
	report("creating $teFile");
	createSelectedElementFaFile('transposable_element',$gffFile,$faFile,$teFile);
	unstamp($teFile);
	report();
    }
    return -s $teFile ? $teFile : undef;
}

sub createSelectedElementFaFile {
    my ($key,$gff,$inFile,$outFile) = @_;
    my $fh = openToRead($gff);
    my %regs;
    while (<$fh>) {
	chomp;
	my ($chr,undef,$type,$lo,$hi) = split;
	if ($type eq $key) {
	    if (defined $regs{$chr} && $regs{$chr}[-1][1] >= $lo-1) {
		# allow for overlapping elements...
		$regs{$chr}[-1][1] = $hi;
	    } else {
		push @{$regs{$chr}},[$lo,$hi];
	    }
	}
    }
    my $fh = openToRead($inFile);
    my $gh = openToWrite($outFile);
    my ($chr,$seq);
    while (<$fh>) {
	chomp;
	if (/^>(\S+)/) {
	    outputChr($chr,$seq,$regs{$chr},$gh);
	    $chr = $1;
	    $seq = '';
	} else {
	    $seq .= $_;
	}
    }
    outputChr($chr,$seq,$regs{$chr},$gh);
}

sub outputChr {
    my ($chr,$seq,$pRegs,$gh) = @_;
    return unless ($chr);
    foreach my $pR (@$pRegs) {
	my ($lo,$hi) = @$pR;
	my $name = sprintf("%s.%7d.%7d",$chr,$lo,$hi);
	$name =~ tr/ /0/;
	print $gh ">$name\n";
	my $piece = substr($seq,$lo-1,$hi-$lo+1);
	printPiece($piece,$gh,60);
	substr($seq,$lo-1,$hi-$lo+1) =~ s/./N/g;
    }
}

sub printPiece {
    my ($seq,$gh,$ll) = @_;
    my $iMax = length($seq)-1;
    for (my $i=0; $i<=$iMax; $i+=$ll) {
	print $gh substr($seq,$i,$ll) . "\n";
    }
}

sub selfAlignTransposableElementFA {
    my ($f) = @_;
    my $g = $f;
    $g =~ s/\.fa$/.sa/;
    if (stamp($g)) {
	my $cmd = "$SSAHA2 $f $f -align 0 -tags 1 > $g";
	report("creating $g");
	`$cmd`;
	report();
	unstamp($g);
    }
    my $h = $g;
    $h =~ s/\.sa$/.ori/;
    if (stamp($h)) {
	report("creating $h");
	decideTransposonOrientations($g,$h);
	report();
	unstamp($h);
    }
    return $h;
}

sub decideTransposonOrientations {
    my ($f,$g) = @_;
    my %rel;
    my $fh = openToRead($f);
    my $gh = openToWrite($g);
    while (<$fh>) {
	chomp;
	my ($tag,$score,$name1,$name2,undef,undef,undef,undef,$fc,$aliLen,$perc) = split;
	next unless ($tag eq 'ALIGNMENT' && $perc >= 90 && $name1 lt $name2);
	my $len1 = lengthFromName($name1);
	my $len2 = lengthFromName($name2);
	if ($aliLen >= min($len1,$len2)*0.75) {
	    $rel{$name1}{$name2} = $fc;
	}
    }
    my %ori;
    my $changed = 1;
    my @keys = sort keys %rel;
    $ori{$keys[0]} = 'F'; # to get things started
    while ($changed) {
	$changed = 0;
	foreach my $name1 (@keys) {
	    my $pH = $rel{$name1};
	    foreach my $name2 (keys %$pH) {
		if (defined $ori{$name1} && !defined $ori{$name2}) {
		    $ori{$name2} = combineOrientations($ori{$name1},$pH->{$name2});
		    $changed = 1;
		} elsif (defined $ori{$name2} && !defined $ori{$name1}) {
		    $ori{$name1} = combineOrientations($ori{$name2},$pH->{$name2});
		    $changed = 1;
		}
	    }
	}
    }
    foreach my $name (sort keys %ori) {
	printf($gh "%s %s\n",$name,$ori{$name});
    }       
}

sub lengthFromName {
    my ($name) = @_;
    my (undef,$lo,$hi) = split(/\./,$name);
    return $hi-$lo+1;
}

sub combineOrientations {
    my ($x,$y) = @_;
    return $x eq $y ? 'F' : 'C';
}

1;
