package SequencedRegionSet;

use Utility;
use Carp;
use strict;

# The files should be *.seqr files ...

sub new {
    my ($class,@files) = @_;
    my %h;
    foreach my $file (@files) {
	if (defined $file) {
	    my $fh = openToRead($file);
	    while (<$fh>) {
		chomp;
		my ($seq,$lo,$hi) = split;
		push @{$h{$seq}},[$lo,$hi];
	    }
	}
    }
    if (%h) {
	my $self = bless \%h,$class;
	return $self;
    } else {
	return undef;
    }
}

sub inSequencedRegion {
    my ($self,$chr,$off,$leeway) = @_;
    my $pA = $self->{$chr};
    return undef unless ($pA);
    my $k = sequencedRegionIndex($pA,$off);
    return ($k >= 0 &&
	    $off >= $pA->[$k][0] + $leeway &&
	    $off <= $pA->[$k][1] - $leeway);
}

# Returns the index of the sequenced region starting at or
# most recently before $off (and possibly finishing before $off too).

sub sequencedRegionIndex {
    my ($pA,$off) = @_;
    my ($kLo,$kHi);
    # Deal with edge cases, because of sparsity in telomeres
    if ($off >= $pA->[-1][0]) {
	return $#$pA; # top end
    } elsif ($off < $pA->[0][0]) {
	return -1; # below bottom end
    } elsif ($off < $pA->[1][0]) {
	return 0;
    } else {
	# Iterate:
	$kLo = 1;
	$kHi = $#$pA;
	my $vLo = $pA->[$kLo][0];
	my $vHi = $pA->[$kHi][0];
	while ($kHi-$kLo > 1) {
	    my $k = max($kLo+1,min($kHi-1,$kLo + int(($kHi-$kLo)*($off-$vLo)/($vHi-$vLo))));
	    my $v = $pA->[$k][0];
	    if ($off < $v) {
		$kHi = $k;
		$vHi = $v;
	    } else {
		$kLo = $k;
		last if ($off == $v);
		$vLo = $v;
	    }
	}
	return $kLo;
    }
}

# If wantLen, returns the length of the overlap (0 if none); otherwise
# returns 1 iff there is an overlap.

sub overlapsSequencedRegion {
    my ($self,$chr,$off1,$off2,$leeway,$wantLen) = @_;
    my ($lo,$hi);
    if ($off1 < $off2) {
	$lo = $off1+$leeway;
	$hi = $off2-$leeway;
    } else {
	$lo = $off2+$leeway;
	$hi = $off1-$leeway;
    }
    return 0 if ($lo > $hi); # too small to bother with
    my $pA = $self->{$chr};
    return 0 unless ($pA); # no overlap if nothing sequenced for this chr
    my $kLo = sequencedRegionIndex($pA,$lo);
    my $kHi = sequencedRegionIndex($pA,$hi);
    #return 1 if (defined $kHi && !defined $kLo);
    if ($wantLen) {
	my $sum=0;
	for (my $k=max($kLo,0); $k<=$kHi; $k++) {
	    my ($regLo,$regHi) = @{$pA->[$k]};
	    my $bot = max($regLo,$lo);
	    my $top = min($regHi,$hi);
	    $sum += $top-$bot+1 if ($top >= $bot);
	}
	return $sum;
    } else {
	# For speed:
	return 1 if ($kLo < $kHi);
	# Can also return 1 if kLo and kHi are equal but region
	# ends between lo and hi.
	return ($pA->[$kLo][1] >= $lo && $pA->[$kLo][1] <= $hi);
    }
}

1;
