package SGRP::SsahaSNP;

use Exporter;
@ISA = qw(Exporter);

@EXPORT = qw(ssahaSNP findSsahaSNPBinary);

use SGRP::Variables;

use Utility;
use File::Basename;

use Carp;

use strict;

my $DIFF = 2000; # larger than any threshold we should want...

# 1 if using special "unstable" ssahaSNP, else 0:
my $UNSTABLE_SSAHA_SNP = 1;

# Min score after adjustment.
my $MINIMUM_ALIGNMENT_SCORE = 20;

sub ssahaSNP {
    my ($species,$filtMeth,$strainSet,$part,@strains) = @_;
    unless (@strains > 0) {
	confess("Usage: $0 ssnp species filtMeth strainSet part strain1 strain2 ...");
    }
    my ($ssahaSNP,$fixFlag) = findSsahaSNPBinary();
    my ($iter,$iDir) = getCurrentIteration($species,$filtMeth,$strainSet,'assembly');
    my $prevDir = previousDirectory($iDir);
    my $pUsed = $iter ? getUsedReads($prevDir) : undef;
    my $crossSpecies = 
	($strainSet =~ /^PAR/ && $species eq 'cere') ||
	($strainSet =~ /^CER/ && $species eq 'para');
    foreach my $str (@strains) {
	my $refGenomeFA = getRefGenomeFA($species,$str,$strainSet,$iter,$iDir);
	# Dir to get OLD fastq files from:
	my $fqDir = ($iter > 1) ? "$prevDir/align/$str" : "$ROOT/read/$str";
	my @tgtList;
	my $spDir = $iter ? "$iDir/align/$str" : "$fqDir/$species"; # Dir to put alignments in
	my $tgt = "$spDir/ssahaSNP.all.gz";
	unless (-s $tgt) {
	    foreach my $p (0 .. 9) {
		my $src = "$fqDir/renamed.$p.fastq";
		my $tgt = "$spDir/ssahaSNP.$p.out.gz";
		push @tgtList,$tgt;
		# Do ANY part for now, to ensure it all gets run. We don't test for
		# $src, in case the filtering has already been done.
		#if (-s $src) { # if ($p == $part && -s $src) {
		my $adj = !($crossSpecies || $str eq 'YONG');
		runSsahaSNP($spDir,$p,$src,$pUsed,$adj,
			    $refGenomeFA,$tgt,$ssahaSNP,$fixFlag,$refGenomeFA);
		#}
	    }
	    if (allFilesReady(@tgtList) && stamp($tgt)) {
		report("creating $tgt");
		`gunzip -c @tgtList | gzip -c > $tgt`;
		report();
		createAlignmentLinesFile($tgt);
		unstamp($tgt);
	    }
	}
	my $ali = "$spDir/alilines.all.txt";
	my $cov = "$spDir/alilines.coverage.txt";
	if ($iter == 0 && -s $ali && stamp($cov)) {
	    report("creating $cov");
	    createAlignmentCoverageFile($ali,$cov);
	    report();
	    unstamp($cov);
	}
    }
}

sub findSsahaSNPBinary {
    my $unst = $UNSTABLE_SSAHA_SNP;
    my $arch = arch();
    my $suf = $unst ? 'unstable' : 'stable';
    my $fix = $unst ? '-fix 1' : '';
    # Adam's home directory:
    my $binDir = '/nfs/team71/psg/aws/bin';
    $arch = 'lin32' if ($arch eq 'i686');
    $arch = 'i386' if ($arch eq 'x86_64');
    return ("$binDir/${arch}_${suf}/ssahaSNP",$fix);
}

sub getUsedReads {
    my ($dir) = @_;
    my %used;
    foreach my $f (glob("$dir/p?/*/alilines.nola.txt")) {
	my $fh = openToRead($f);
	while (<$fh>) {
	    chomp;
	    my (undef,undef,$read) = split;
	    $used{$read} = 1;
	}
    }
    return \%used;
}

sub getRefGenomeFA {
    my ($sp,$str,$strainSet,$iter,$iDir) = @_;
    my $fa;
    if ($iter) {
	$fa = "$iDir/assembly/$str/genome.fa";
    } else {
	# Try augmented first, or failing that, the original
	# genome it comes from. But if $strainSet is PAR* and
	# $sp is cere, we don't want extended. ... Actually we go
	# with extended anyway, to achieve consistency of coordinates
	# and because of obscure bug with using "augmented".
	foreach my $base (qw(extended augmented genome)) {
	    #unless ($base eq 'extended' && $sp eq 'cere' && $strainSet =~ /^PAR/) {
		$fa = "$ROOT/ref/$sp/$base.fa";
		last if (-s $fa);
	    #}
	}
    }
    if (-s $fa) {
	return $fa;
    } else {
	confess("Cannot find $fa");
    }
}

sub filterOutUsedReads {
    my ($src,$pUsed,$tgtDir) = @_;
    my $tgt = sprintf("%s/%s",$tgtDir,basename($src));
    return $tgt if (-f $tgt); # can be empty, so -f here
    return undef unless (-f $src); # ... and here
    # do the filtering
    if (stamp($tgt)) {
	report("creating $tgt");
	my $fh = openToRead($src);
	my $gh = openToWrite($tgt);
	my $active;
	while (<$fh>) {
	    if (/^@(\S+) .*bases/) {
		$active = !$pUsed->{$1};
	    }
	    print $gh $_ if ($active);
	}
	report();
	unstamp($tgt);
    }
    return $tgt;
}

sub runSsahaSNP {
    my ($spDir,$p,$src,$pUsed,$adjust,$refGenomeFA,$tgt,$ssahaSNP,$fixFlag,$refGenomeFA) = @_;
    return if (-s $tgt);
    mkdir_p($spDir);
    my $err = "$spDir/ssahaSNP.$p.err";
    if (stamp($tgt)) {
	$DIFF = 2000000 if ($src =~ /YONG/);
	if ($pUsed) {
	    $src = filterOutUsedReads($src,$pUsed,$spDir);
	}
	unless (-f $src) { # It should exist, but can be empty ...
	    print STDERR "Cannot find or create $src\n";
	    unstamp($tgt);
	    return;
	}
	my $cmd = 
	    "$ssahaSNP $src $refGenomeFA -diff $DIFF -NQS 0 -quality 0 -tags 1 -align 1";
	unless ($src =~ /YONG/) {
	    $cmd .=  " -score 20 -seeds 2 -cut 50000 $fixFlag";
	}
	if (-x $ssahaSNP) {
	    report("creating $tgt with\n  $cmd");
	    my $fh = openFromPipe($cmd);
	    my $gh = openToWrite($tgt);
	    adjustSsahaSNPScores($fh,$adjust,$gh);
	    # We put the 'sleep 10' in to ensure the gzip is allowed to finish.
	    #`$cmd | gzip -c >$tgt 2>$err; sleep 10\n`;
	    report();
	    unstamp($tgt);
	} else {
	    unstamp($tgt);
	    confess("Missing or not executable: $ssahaSNP");
	}
    }
}

sub adjustSsahaSNPScores {
    my ($fh,$adjust,$gh) = @_;
    my $active = 1;
    while (<$fh>) {
	if (/^ALIGNMENT/) {
	    my @a = split;
	    # We used to subtract twice the difference between the alignment length
	    # and the original score, but this is unreliable (tends to penalize
	    # short subject sequences).
	    #my $adj = $adjust ? 3*$a[1]-2*$a[9] : $a[1];
	    # Instead, we calculate from alignment length and %ident, so that
	    # 90% ident -> zero score.
	    my $adj = $adjust ? int(0.5+$a[9]*($a[10]/10-9)) : $a[1];
	    $active = ($adj >= $MINIMUM_ALIGNMENT_SCORE); # say 20
	    if ($active && $adjust) {
		# Score should be LEFT aligned!!
		my $decr = length($a[1])-length($adj);
		if ($decr > 0) {
		    my $padded = $adj . (' ' x $decr);
		    s/$a[1]/$padded/; # should work as early in line and no funny chars!
		} elsif ($decr < 0) {
		    my $padded = $a[1] . (' ' x (-$decr));
		    s/$padded/$adj/;
		} else {
		    s/$a[1]/$adj/;
		}
	    }
	} elsif (/^=/) {
	    $active = 1;
	} 
	if ($active) {
	    if (/^Query/) {
		my $qLine = $_;
		my $mLine = <$fh>; # "middle" line
		my $sLine = <$fh>;
		if ($sLine =~ /N/) {
		    ($qLine,$mLine,$sLine) = avoidAligningNs($qLine,$mLine,$sLine);
		}
		print $gh $qLine;
		print $gh $mLine;
		print $gh $sLine;
	    } else {
		print $gh $_;
	    }
	}
    }
}

sub avoidAligningNs {
    my ($qLine,$mLine,$sLine) = @_;
    my $iLim = length($sLine);
    my $i=0;
    while (1) {
	my $v;
	while ($i<$iLim && !(substr($sLine,$i,1) eq 'N' && substr($mLine,$i,1) eq '?')) {
	    $i++;
	}
	last if ($i == $iLim);
	my $j=$i+1;
	while ($j<$iLim && substr($sLine,$j,1) eq 'N' && substr($mLine,$j,1) eq '?') {
	    $j++;
	}
	# We have N from i to j-1 inclusive
	my $len = $j-$i;
	# If query already has a gap symbol at i-1, or sbjct has a gap at j,
	# put query gaps first; else put sbjct gaps first.
	# mLine is the same either way.
	# mLine: pre-N stuff, then 2 len gaps, then post-N
	$mLine = substr($mLine,0,$i) . ('--' x $len) . substr($mLine,$j);
	if (substr($qLine,$i-1,1) eq '-' || substr($sLine,$j,1) eq '-') {
	    # qLine: pre-N stuff, then gaps, then qLine's [i,j) sequence, then post-N stuff.
	    $qLine = substr($qLine,0,$i) . ('-' x $len) . substr($qLine,$i);
	    # sLine: pre-N stuff, then sLine's [i,j) sequence, then gaps, then post-N.
	    $sLine = substr($sLine,$0,$j) . ('-' x $len) . substr($sLine,$j);
	} else {
	    # qLine: pre-N stuff, then qLine's [i,j) sequence, then gaps, then post-N stuff.
	    $qLine = substr($qLine,0,$j) . ('-' x $len) . substr($qLine,$j);
	    # sLine: pre-N stuff, then gaps, then sLine's [i,j) sequence, then post-N.
	    $sLine = substr($sLine,$0,$i) . ('-' x $len) . substr($sLine,$i);
	}

	# Update $i to after the N sequence
	$i = $j+$len;
	$iLim += $len;
    }
    return ($qLine,$mLine,$sLine);
}
	     

sub createAlignmentLinesFile {
    my ($ssgz) = @_;
    my $dir = dirname($ssgz);
    my $ali = "$dir/alilines.all.txt";
    unless (-s $ali) {
	# The more usual scenario ...
	report("creating $ali");
	my $fh = openToRead($ssgz);
	my $gh = openToWrite($ali);
	while (<$fh>) {
	    if (/^ALIGNMENT/) {
		print $gh $_;
	    }
	}
	$gh->close();
	report();
    }
    return $ali;
}

sub createAlignmentCoverageFile {
    my ($aFile,$cFile) = @_;
    my $fh = openToRead($aFile);
    my ($read,@buf,%avCov,%totCov);
    while (<$fh>) {
	chomp;
	my @a = split;
	unless ($a[2] eq $read) {
	    augmentCoverage(\@buf,\%avCov,\%totCov) if (@buf);
	    $read = $a[2];
	    #last if ($read eq 'BC187-15p11.p1k'); # for speed
	    @buf = ();
	}
	push @buf,\@a;
    }
    augmentCoverage(\@buf,\%avCov,\%totCov) if (@buf);
    my $gh = openToWrite($cFile);
    foreach my $chr (sort keys %avCov) {
	my $pA = $avCov{$chr};
	my $pT = $totCov{$chr};
	my ($kLast,$oaLast,$otLast);
	foreach my $k (1 .. $#$pA) {
	    my $oa = sprintf("%3.1f",$pA->[$k]);
	    my $ot = sprintf("%3d",$pT->[$k]);
	    unless ($oa eq $oaLast && $ot eq $otLast) {
		if (defined $kLast) {
		    printf($gh "%s %7d %7d %3s %5s\n",
			   $chr,$kLast,$k-1,$otLast,$oaLast);
		}
		$kLast = $k;
		$oaLast = $oa;
		$otLast = $ot;
	    }
	}
	if (defined $kLast) {
	    printf($gh "%s %7d %7d %3s %5s\n",
		   $chr,$kLast,$#$pA,$otLast,$oaLast);
	}
    }
}

sub augmentCoverage {
    my ($pBuf,$pACov,$pTCov) = @_;
    # Find depth in read
    my @depth;
    foreach my $pA (@$pBuf) {
	my ($rlo,$rhi) = @{$pA}[4..5];
	if ($rlo <= $rhi) {
	    foreach my $k ($rlo .. $rhi) {
		$depth[$k] ++;
	    }
	} else {
	    foreach my $k ($rhi .. $rlo) {
		$depth[$k] ++;
	    }
	}
    }
    # Augment coverages
    foreach my $pA (@$pBuf) {
	my ($chr,$rlo,$rhi,$clo,$chi,$fc) = @{$pA}[3..8];
	if ($fc eq 'C') {
	    my $tmp = $rlo;
	    $rlo = $rhi;
	    $rhi = $tmp;
	}
	my $rate = ($rhi-$rlo+1)/($chi-$clo+1);
	foreach my $i ($clo .. $chi) {
	    my $k = int(0.5+$rlo+($i-$clo)*$rate);
	    $pACov->{$chr}[$i] += 1/$depth[$k];
	    $pTCov->{$chr}[$i] ++;
	}
    }
}

1;
