package SGRP::StrainAssemblies;

use Exporter;
@ISA = qw(Exporter);

@EXPORT = qw(strainAssemblies referenceStrainAssembly);

use SGRP::Variables;

use Utility;
use File::Basename;

use Carp;

use strict;

my $MAX_READ_LENGTH = 1200;

my $MAX_BUILDING_COUNT = 200; # to prevent massive explosions ...

my $MIN_CHUNK_LENGTH = 10;

# As proportion of originalGenomeSize * #strains
my $SEQUENCED_INCREASE_THRESHOLD = 0.0001;
# Hard-wire S cere for now:
my $REFERENCE_GENOME_SIZE = 12162911;

sub strainAssemblies {
    my ($species,$filtMeth,$strainSet,@strains) = @_;
    unless (@strains) {
	confess("Usage: $0 stra species filtMeth strainSet str1 str2 ...");
    }
    my ($iter,$srcDir) = getCurrentIteration($species,$filtMeth,$strainSet,'p?');
    my $topDir = substr($srcDir,0,length($srcDir)-4);
    my @subsToGet = getSubsToGet($topDir);
    my @iterDirs = reverse(glob("$topDir/[0-9][0-9][0-9]"));
    while (@iterDirs && basename($iterDirs[0]) > $iter) {
	shift @iterDirs;
    }
    # Here's where we move on to the next iteration ...
    (my $iter1 = sprintf("%3d",$iter+1)) =~ tr/ /0/;
    my $tgtDir = "$topDir/$iter1";
    foreach my $str (@strains) {
	createStrainAssembly(\@subsToGet,\@iterDirs,$tgtDir,$str);
	decideWhetherFinalIteration($tgtDir,$strainSet,@iterDirs) if ($str eq 'REF');
    }
}

sub getSubsToGet {
    my ($topDir) = @_;
    my @ans;
    my $dir000 = "$topDir/000";
    my $len = length($dir000)+1;
    foreach my $sub (glob("$dir000/p?/*")) {
	if (-s $sub) {
	    push @ans,substr($sub,$len);
	}
   }
    return @ans;
}

sub createStrainAssembly {
    my ($pSubsToGet,$pIterDirs,$tgtDir,$str) = @_;
    my $tDir = "$tgtDir/assembly/$str";
    my $fa = "$tDir/genome.fa";
    my $mf = "$tDir/genome.map";
    my $qf = "$tDir/genome.qual";
    my $sf = "$tDir/genome.seqr"; # sequenced-region file
    my $df = "$tDir/genome.dangles";
    my $af = "$tDir/genome.ali.gz"; # can get assembly.fa from this with alignmentToAssembly.pl
    my $mh;
    if (stamp($fa)) {
	report("creating files in $tDir");
	my $gh = openToWrite($fa);
	$mh = openToWrite($mf);
	my $qh = openToWrite($qf);
	my $sh = openToWrite($sf);
	my $dh = $str eq 'REF' ? undef : openToWrite($df);
	my $ah;
	# Only do this on final iteration: we only need it
	# then, and it sometimes makes gzip fall over with
	# "no space on device".
	if ($IS_FINAL_ITERATION && $str ne 'REF') {
	    $ah = openToWrite($af);
	}
	#foreach my $sub (glob("$srcDir/p[0-9]/*")) {
	foreach my $sub (locateSubDirs($pSubsToGet,$pIterDirs)) {
	    printSequencedRegions($sub,$str,$sh);
	    my $if;
	    my $isFinal = 1;
	    foreach my $suf (qw(z y m g h c)) {
		$if = "$sub/imputed_$suf.gz";
		last if (-s $if);
		$isFinal = 0 if ($suf eq 'm');
		$if = undef;
	    }
	    if ($if) {
		my $fh = openToRead($if);
		my ($idx,$ns) = getStrainIndex($str,$fh);
		if ($ns) {
		    my $chr = basename($sub);
		    # These dangles will be misaligned ...
		    #my $pDang = readDangles("$sub/dangles.txt");
		    augmentStrainAssembly($str,$chr,$isFinal,$fh,$idx,$ns,$gh,$mh,$qh,$dh,$ah);
		} else {
		    # We don't have to have every chromosome in every strain, e.g. scplasm1 ...
		    print STDERR "WARNING: cannot find strain $str in first line of $if\n";
		}
	    } else {
		print STDERR "WARNING: cannot find imputed_[mghc].gz in $sub\n";
	    }
	}
	# Copy the vectors over
	my $vfa = "$ROOT/ref/vect/vect.fa";
	my $vh = openToRead($vfa);
	while (<$vh>) {
	    print $gh $_;
	}
	report();
	unstamp($fa);
    }
    # Do GFF liftover if this is REF strain
    if ($str eq 'REF') {
	$mh->close() if ($mh); # as we will need to read this!
	liftoverGFF("$tgtDir/assembly");
    }
    # This relies on the above having already been done for REF, so
    # probably need to run 'stra' again on final iteration.
    liftoverGFF2("$tgtDir/assembly",$str);
}

sub locateSubDirs {
    my ($pSubsToGet,$pIterDirs) = @_;
    my @ans;
    foreach my $sub (@$pSubsToGet) {
	foreach my $dir (@$pIterDirs) {
	    my $d = "$dir/$sub";
	    if (-d $d) {
		push @ans,$d;
		last;
	    }
	}
    }
    return @ans;
}

sub printSequencedRegions {
    my ($d,$str,$gh) = @_;
    my $chr = basename($d);
    my @sufs = ('_z','_y','');
    my $fh;
    foreach my $suf (@sufs) {
	my $f = "$d/sorted$suf.gz"; # prefer sorted_z if available!
	if (-s $f) {
	    $fh = openToRead($f);
	    last;
	}
    }
    my $off = 0;
    my $hdr = <$fh>;
    chomp $hdr;
    my @strains = ('REF',split(" ",substr($hdr,1)));
    my $col;
    foreach my $i (0 .. $#strains) {
	if ($strains[$i] eq $str) {
	    $col = $i*2;
	    last;
	}
    }
    # If this strain isn't in the sorted file, none of this
    # chromosome was sequenced (e.g. scplasm1).
    return unless (defined $col);
    my ($srStart,$srEnd);
    while (<$fh>) {
	$off ++;
	my $v = substr($_,$col,1);
	if (defined $v && $v ne ' ') {
	    unless (defined $srStart) {
		$srStart = $off;
	    }
	    $srEnd = undef if (defined $srEnd); # after restart
	} elsif (defined $srStart) {
	    # Bridge gaps of up to 5 bases ...
	    if (defined $srEnd) {
		if ($off-$srEnd >= 5) {
		    print $gh "$chr $srStart $srEnd\n";
		    $srStart = $srEnd = undef;
		}
	    } else {
		$srEnd = $off-1;
	    }
	}
    }
    if (defined $srStart) {
	$srEnd = $off-1 unless (defined $srEnd);
	print $gh "$chr $srStart $srEnd\n";
    }
    $gh->flush;
}

sub getStrainIndex {
    my ($str,$fh) = @_;
    chomp (my $hdr = <$fh>);
    my @strains = split(" ",substr($hdr,1));
    if ($str eq 'REF') {
	return (0,scalar(@strains));
    } else {
	for (my $k=0; $k<=$#strains; $k++) {
	    # "$k == $#strains" test means we use the MRCA if wr
	    # can't find the strain itself.
	    if ($strains[$k] eq $str || $k == $#strains) {
		if ($strains[$k] eq 'MRCA') {
		    print STDERR "WARNING: using MRCA sequence in place of $str\n";
		}
		return ($k+1,scalar(@strains)); # index 1 for first strain, etc
	    }
	}
    }
    return 0; # means fail
}

# mapFile lines are of the form
#   seqName seqLo otherLo length
# where seqName is the sequence being defined
#       seqLo   is the starting point of a chunk (offset 0) in seqName
#       otherLo is the starting point of the corresponding chunk in "the other" sequence
#       length  is the length of the chunk (probably not needed, but keep it for now).
# NOTE that when seqName is the main sequence for the strain (strainName.chrName),
# the "other" sequence is the alignment in sorted.gz, imputed.ga etc. But when
# seqName is an auxiliary sequence (strainName.chrName.NNN) the "other" sequence
# is the one for the strain, NOT the alignment. This is possible because the main
# sequence is the maximal one, i.e. we never impute-across hard gaps from other strains.

sub augmentStrainAssembly {
    my ($str,$chr,$isFinal,$fh,$idx,$ns,$gh,$mh,$qh,$dh,$ah) = @_;
    my $col = $idx*2;
    my ($pHGRList,$mainSeq,$mainQuals) = deriveMainSequence($str,$ns,$chr,$col,$fh,$gh,$qh,$mh,$dh,$ah);
    # If $isFinal, i.e. imputed_$suf was suffix y or z, then we don't want any 
    # aux sequences as we've already decided which regions are included.
    unless ($isFinal) {
	# A valid set of HGRs for an aux sequence is any of length 1 or more
	# such that the sum of internal distances between them is <= $AUX_REGION_LENGTH.
	# We represent an HGRList as a list: [sumIntDist,HGR1,HGR2,...]
	my $pSets = findAllHGRSets($chr,$pHGRList);
	writeAuxiliarySequences("$str.$chr",$pSets,$mainSeq,$mainQuals,$gh,$qh,$mh);
    }
}

sub deriveMainSequence {
    my ($str,$ns,$chr,$col,$fh,$gh,$qh,$mh,$dh,$ah) = @_;
    my %seenHardGapRegion;
    print $gh ">$str.$chr\n";
    print $qh ">$str.$chr\n";
    print $ah ">$str.$chr\n" if ($ah);
    my $offStr = 1; # offsets from 1, as that's what ssahaSNP does
    my $offAli = 1;
    my $mhPfx = "$str.$chr $offStr $offAli";
    my $runLength = 0;
    my @hardGapStart;
    my ($seq,$quals);
    # Start and end of current or recently-finished sequenced region. If
    # neither is defined, we're in an unsequenced region of length > srSpan.
    # If both are defined, we finished one not less than srSpan positions ago.
    # If only srStart is defined, we're inside a sequenced region. We should
    # never have just srEnd defined.
    #  Note that srStart srEnd are in the _base_ (level 2) coordinate system.
    #my $srSpan = 10; # allowed gap between sequenced regions to count as one
    #my ($srStart,$srEnd);
    my $prev; # previous line
    while (<$fh>) {
	chomp;
	my $nuc = substr($_,$col,1);
	my $ref = substr($_,0,1);
	if ($ah && "$ref$nuc" =~ /[ACGTacgt]/) {
	    printf($ah "%s %s%s\n",$ref,$nuc,substr($_,$col+1,1));
	}
	my $isReal = ($nuc =~ /[-=ACGT]/); # i.e. upper-case, not imputed
	# Search for hard gap starts/ends; and deal with sequenced-region records.
	if ($isReal) {
	    #if (defined $srEnd) {
	    #	$srEnd = undef; # short span of unsequenced; ignore it.
	    #} elsif (!defined $srStart) {
	    #   $srStart = $offAli; # start new sequenced region here
	    #}
	    @hardGapStart = ();
	#} elsif (isGhostLine($_)) {
	#    # Line consisting of all gaps: just ignore it, don't increase counter.
	#    # This is NOT safe because it will mess up layout on the next round...
	#    next;
	} else {
	    #if (defined $srEnd) {
		#if ($offAli-$srEnd > $srSpan) {
		#    if ($srEnd-$srStart >= $srSpan) { # else trivial
		#	printf($sh "%s %d %d\n",$chr,$srStart,$srEnd);
		#    }
		#    $srStart = $srEnd = undef;
		#}
	    #} elsif (defined $srStart) {
	    #    $srEnd = $offAli-1;
	    #}
	    for (my $i=0; $i<$ns; $i++) {
		my $ii = $i*2;
		next if ($ii == $col); # ignore own hard gap
		my $hgs = $hardGapStart[$i];
		my $v = substr($_,$ii,1);
		# We let a lower-cased part start a "hard gap" in the reference, as this
		# will have arisen from an optionally inserted transposon.
		if ($v eq '=' || ($i == 0 && $v =~ /[a-z]/)) {
		    $hardGapStart[$i] = $offStr unless (defined $hgs);
		} elsif (defined $hgs) {
		    $seenHardGapRegion{$hgs}{$offStr-1} = 1;
		    $hardGapStart[$i] = undef;
		}
	    }
	}
	printDangle("$str.$chr",$prev,$_,$col,$offStr,$dh);
	if ($nuc eq '-' || $nuc eq '=' || $nuc eq '_') {
	    if ($mhPfx) {
		if ($runLength > 0) {
		    print $mh "$mhPfx $runLength\n";
		    #printRelevantDangles($mhPfx,$runLength,$pDang,$dh);
		    $runLength = 0;
		}
		$mhPfx = undef;
	    }
	} else {
	    if (!$mhPfx) {
		$mhPfx = "$str.$chr $offStr $offAli";
	    }
	    $nuc =~ tr/a-z/A-Z/;
	    print $gh $nuc;
	    my $qual = substr($_,$col+1,1);
	    print $qh $qual;
	    $seq .= $nuc;
	    $quals .= $qual;
	    if ($offStr%60 == 0) {
		print $gh "\n";
		print $qh "\n";
	    }
	    $offStr ++;
	    $runLength ++;
	}
	$offAli ++;
	$prev = $_;
    }
    # Finish off sequenced-region spans:
    #if (defined $srStart) {
	#$srEnd = $offAli unless (defined $srEnd);
	#if ($srEnd-$srStart >= $srSpan) {
	#    printf($sh "%s %d %d\n",$chr,$srStart,$srEnd);
	#}
    #}
    printDangle("$str.$chr",$prev,undef,$col,$offStr,$dh);
    if ($mhPfx && $runLength > 0) {
	print $mh "$mhPfx $runLength\n";
	#printRelevantDangles($mhPfx,$runLength,$pDang,$dh);
    }
    if ($offStr%60 != 1) {
	print $gh "\n";
	print $qh "\n";
    }
    for (my $i=0; $i<$ns; $i++) {
	my $ii = $i*2;
	next if ($ii == $col); # ignore own hard gap
	my $hgs = $hardGapStart[$i];
	if (defined $hgs) {
	    $seenHardGapRegion{$hgs}{$offAli-1} = 1;
	}
    }
    my @hardGapRegions;
    foreach my $lo (sort {$a <=> $b} keys %seenHardGapRegion) {
	foreach my $hi (sort {$a <=> $b} keys %{$seenHardGapRegion{$lo}}) {
	    if ($hi-$lo >= 6) { # exclude short regions for efficiency
		push @hardGapRegions,[$lo,$hi];
	    }
	}
    }
    return (\@hardGapRegions,$seq,$quals);
}

sub printDangle {
    my ($chr,$line0,$line1,$col,$off,$dh) = @_;
    return unless ($dh);
    # A absent (lower case, in imputed), N nucleotide or short gap, H hard gap
    my $s0 = valueType($line0,$col,'A');
    my $s1 = valueType($line1,$col,'A');
    return if ($s0 eq $s1); # for speed
    if ($s0 eq 'A') {
	# If any col goes from N to H, then left dangle
	if ($s1 eq 'N' && findTransition($line0,$line1,$col,'N','H')) {
	    printf($dh "%s %d L\n",$chr,$off+1); # off+1 for 1-offset
	}
    } elsif ($s0 eq 'N') {
	# If any col goes from H to N, then right dangle
	if ($s1 eq 'A' && findTransition($line0,$line1,$col,'H','N')) {
	    printf($dh "%s %d R\n",$chr,$off); # off for 1-offset on previous position
	}
    }
}
 
sub valueType {
    my ($line,$col,$dft) = @_;
    return $dft unless ($line);
    my $v = substr($line,$col,1);
    if ($v =~ /[a-z_]/) {
	return 'A'; # imputed, so absent in sequenced
    } elsif ($v eq '=') {
	return 'H'; # hard gap
    } else {
	return 'N'; # nucleotide or short gap
    }
}

sub findTransition {
    my ($line0,$line1,$strCol,$from,$to) = @_;
    my $len = length($line0);
    for (my $col=0; $col<$len; $col+=2) {
	if ($col != $strCol &&
	    valueType($line0,$col,'N') eq $from &&
	    valueType($line1,$col,'N') eq $to) {
	    return 1;
	}
    }
    return 0;
}

# $chr: chr name, for error/warning reporting only
# $pHGRList: list of hard-gap regions in form [lo,hi]

sub findAllHGRSets {
    my ($chr,$pHGRList) = @_;
    my (@building,@done);
    my $nRestricted = 0;
    foreach my $pHGR (@$pHGRList) {
	my @newBuilding = ([0,$pHGR]);
	foreach my $pSet (@building) {
	    my $sz = $pHGR->[0] - $pSet->[-1][1] - 1;
	    my $sid = $pSet->[0] + $sz;
	    if ($sz >= 0 && $sid < $MAX_READ_LENGTH) {
		# Keep pSet on the go in both forms
		push @newBuilding,$pSet;
		if ($sid >= 0) {
		    my @newSet = @$pSet;
		    push @newSet,$pHGR;
		    $newSet[0] = $sid;
		    push @newBuilding,\@newSet;
		}
	    } else {
		push @done,$pSet;
	    }
	}
	if (scalar(@newBuilding) > $MAX_BUILDING_COUNT) {
	    @building = (sort {hgrScore($b) <=> hgrScore($a)} @newBuilding)[0 .. $MAX_BUILDING_COUNT-1];
	    $nRestricted ++;
	} else {
	    @building = @newBuilding;
	}
	#printf("  BuildingCount %d\n",scalar(@building));
    }
    if ($nRestricted) {
	printf(STDERR "WARNING: restricted aux sequences for %s at %d choice points\n",$chr,$nRestricted);
    }
    push @done,@building;
    return \@done;
}

# "Score" of hgr set is total length / number of segments, i.e.
# we prefer less dense.
sub hgrScore {
    my ($pHGR) = @_;
    return ($pHGR->[-1][1]+1-$pHGR->[1][0]+2*$MAX_READ_LENGTH)/scalar(@$pHGR); 
}

sub writeAuxiliarySequences {
    my ($pfx,$pSets,$mainSeq,$mainQuals,$gh,$qh,$mh) = @_;
    foreach my $i (0 .. $#$pSets) {
	writeAuxiliarySequence($pfx,$i+1,$pSets->[$i],$mainSeq,$mainQuals,$gh,$qh,$mh);
    }
}

sub writeAuxiliarySequence {
    my ($pfx,$idx,$pSet,$mainSeq,$mainQuals,$gh,$qh,$mh) = @_;
    # Each pair is the inclusive range for a hard-gap region.
    my ($sid,@pairs) = @$pSet;
    my $pad = $MAX_READ_LENGTH - $sid;
    my $preLen = min($pairs[0][0],$pad);
    my ($seq,$quals);
    my $auxId = sprintf("%s.%3d",$pfx,$idx);
    $auxId =~ tr/ /0/;
    my $auxLen = 0;
    my @mapList;
    return unless addChunk($auxId,$mainSeq,$mainQuals,
			   $pairs[0][0]-$preLen,
			   $preLen,
			   \$seq,\$quals,\$auxLen,\@mapList);
    for (my $i=0; $i<$#pairs; $i++) {
	my $pPair1 = $pairs[$i];
	my $pPair2 = $pairs[$i+1];
	return unless addChunk($auxId,$mainSeq,$mainQuals,
			       $pPair1->[1]+1, # posn after previous HG region
			       $pPair2->[0]-$pPair1->[1]-1, # length of our chunk
			       \$seq,\$quals,\$auxLen,\@mapList);
    }
    return unless addChunk($auxId,$mainSeq,$mainQuals,
			   $pairs[-1][1]+1,
			   min(length($mainSeq)-$pairs[-1][1]-1,$pad),
			   \$seq,\$quals,\$auxLen,\@mapList);
    foreach my $line (@mapList) {
	print $mh $line;
    }
    printFormatted($gh,$auxId,$seq);
    printFormatted($qh,$auxId,$quals) if ($qh);
}

sub addChunk {
    my ($id,$mainSeq,$mainQuals,$chunkStart,$chunkLen,$pSeq,$pQuals,$pAuxLen,$pML) = @_;
    if ($chunkLen == 0) {
	return 1; # no-op but OK
    } elsif ($chunkLen < $MIN_CHUNK_LENGTH) {
	return 0; # bad set
    } else {
	$$pSeq .= substr($mainSeq,$chunkStart,$chunkLen);
	$$pQuals .= substr($mainQuals,$chunkStart,$chunkLen);
	push @$pML,"$id $$pAuxLen $chunkStart $chunkLen\n";
	$$pAuxLen += $chunkLen;
	return 1;
    }
}

sub printFormatted {
    my ($gh,$id,$seq) = @_;
    confess("No handle") unless ($gh);
    print $gh ">$id\n";
    my $len = length($seq);
    for (my $i=0; $i<$len; $i+=60) {
	print $gh substr($seq,$i,60) . "\n";
    }
}

sub liftoverGFF {
    my ($tgtDir) = @_;
    # set sp filt iter assembly
    my (undef,$iter,$filt,$sp,$set) = reverse(split(/\//,$tgtDir));
    my $srcGFF = "$ROOT/ref/$sp/genome.gff";
    my $ptisFile = "$ROOT/ref/$sp/te/$set.ptis2";
    my $mapFile = "$tgtDir/REF/genome.map";
    my $ok = 1;
    foreach my $f ($srcGFF,$mapFile) {
	unless (-s $f) {
	    report("missing or empty: $f",1);
	    $ok = 0;
	}
    }
    my $tgtGFF = "$tgtDir/genome.gff";
    if (-s $tgtGFF) {
	report("already exists: $tgtGFF",1);
    } elsif ($ok && stamp($tgtGFF)) {
	liftoverGFF1($srcGFF,$ptisFile,$mapFile,$tgtGFF);
	unstamp($tgtGFF);
    }
}

sub liftoverGFF1 {
    my ($srcGFF,$ptisFile,$mapFile,$tgtGFF) = @_;
    report("lifting over GFF: reading $ptisFile");
    my $pTI = readTransposonInsertions($ptisFile,$srcGFF);
    report();
    report("lifting over GFF: reading $mapFile");
    my $pMap = readMap($mapFile);
    report();
    report("lifting over GFF: writing $tgtGFF");
    my $fh = openToRead($srcGFF);
    my $gh = openToWrite($tgtGFF);
    #autoflush $gh 1;
    while (<$fh>) {
	#print "DOING $_";
	chomp;
	my @a = split;
	my ($chr,$db,undef,$lo,$hi) = @a;
	# Is it time to print an added TE?
	my $pTIC = $pTI->{$chr};
	while (defined $pTIC && @$pTIC && $pTIC->[0][0] < $lo) {
	    my $pElt = shift @$pTIC;
	    my ($off,$teFC,@teDef) = @$pElt;
	    if (defined $pMap->{$chr}) {
		# pMap->chr may be undefined, e.g. if chr is scplasm1
		# and we don't have one
		my @te = ($chr,$db,'transposable_element',
			  mapLocation($chr,$off,$pMap)+1,
			  mapLocation($chr,$off+2,$pMap)-2, # to avoid obscure bug
			  '.',
			  ($teFC eq 'F') eq ($teDef[6] eq '+') ? '+' : '-',
			  '.',
			  $teDef[-1]);
		print $gh join("\t",@te) . "\n";
	    }
	    #print "DONE @te\n";
	    #printf("pTIC has size %d\n",scalar(@$pTIC));
	}
	#print "Doing mapLocations\n";
	if (defined $pMap->{$chr}) {
	    $a[3] = mapLocation($chr,$lo,$pMap);
	    $a[4] = mapLocation($chr,$hi,$pMap);
	    print $gh join("\t",@a) . "\n";
	}
    }
    report();
}

sub readTransposonInsertions {
    my ($f,$srcGFF) = @_;
    my %ti;
    if (-s $f && -s $srcGFF) {
	my $fh = openToRead($srcGFF);
	my %teRef;
	while (<$fh>) {
	    chomp;
	    my @a = split;
	    if ($a[2] eq 'transposable_element') {
		$teRef{$a[0]}{$a[3]}{$a[4]} = \@a;
	    }
	}
	my $fh = openToRead($f);
	while (<$fh>) {
	    chomp;
	    my ($chr,$off,$te,$fc) = split;
	    my ($teChr,$teLo,$teHi) = split(/\./,$te);
	    my $pA = $teRef{$teChr}{$teLo+0}{$teHi+0};
	    $pA->[-1] =~ s/ID=Y/ID=COPY_Y/;
	    $pA->[-1] =~ s/Name=Y/Name=COPY_Y/;
	    if (defined @$pA) {
		push @{$ti{$chr}},[$off,$fc,@$pA];
	    } else {
		die("Cannot resolve $te");
	    }
	}
    }
    return \%ti;
}

sub readMap {
    my ($f) = @_;
    my $fh = openToRead($f);
    my %map;
    while (<$fh>) {
	chomp;
	my ($chr,$from,$to,$len) = split;
	if ($chr =~ /^REF\./) {
	    push @{$map{substr($chr,4)}},[$from,$to,$len];
	}
    }
    return \%map;
}

sub mapLocation {
    my ($chr,$off,$pMap) = @_;
    $chr =~ tr/A-Z/a-z/; # for chrm
    my $pA = $pMap->{$chr};
    # First check for off top end
    if ($off > $pA->[-1][0]) {
	return $pA->[-1][1]+$off-$pA->[-1][0];
    }
    # Then iterate
    my $kLo = 0; # lowest possible $k
    my $kHi = $#$pA+1; # ONE MORE THAN highest possible $k
    while (1) {
	my $k = int(($kLo+$kHi)/2);
	my $pAK = $pA->[$k];
	my $lo = $pAK->[0];
	my $hi = $lo + $pAK->[2] - 1;
	if ($lo > $off) {
	    $kHi = $k;
	} elsif ($hi < $off) {
	    $kLo = $k+1;
	} else {
	    return $pAK->[1]+$off-$lo;
	}
    }
}

sub liftoverGFF2 {
    my ($assemblyDir,$str) = @_;
    my $aliGFF = "$assemblyDir/genome.gff";
    my $tgtDir = "$assemblyDir/$str";
    my $tgtGFF = "$tgtDir/genome.gff";
    unless (-s $aliGFF && ! -f "$aliGFF.touch") {
	report("$tgtGFF: cannot create because $aliGFF not ready",1);
	return;
    }
    return unless (stamp($tgtGFF));
    liftoverGFF3($str,$aliGFF,"$tgtDir/genome.map",$tgtGFF);
    unstamp($tgtGFF);
}
    
sub liftoverGFF3 {
    my ($str,$aliGFF,$mapFile,$tgtGFF) = @_;
    my $pOM = new OffsetMap('?',$mapFile);
    report("$tgtGFF: creating");
    my $fh = openToRead($aliGFF);
    my $gh = openToWrite($tgtGFF);
    while (<$fh>) {
	chomp;
	my @a = split(" ",$_,9);
	my ($chr,undef,$type,$lo,$hi) = @a;
	$chr = 'chrm' if ($chr eq 'chrM');
	$a[3] = $pOM->backTranslation("$str.$chr",$lo);
	$a[4] = $pOM->backTranslation("$str.$chr",$hi);
	# Fudge because copied TE's come through with length 1 when back
	# translating to REF ...
	if ($a[3] < $a[4] || ($a[3] == $a[4] && $type ne 'transposable_element')) {
	    # If zero or negative extent, object does not exist
	    # in strain
	    print $gh join("\t",@a) . "\n";
	}
    }
    report();
}

sub decideWhetherFinalIteration {
    my ($tgtDir,$strainSet,$prev1Dir,$prev2Dir) = @_;
    if (-f "$prev1Dir/isFinalIteration") {
	# "Final" is always followed by "Extra".
	my $g = "$tgtDir/isExtraIteration";
	touch($g);
	print "Creating $g\n";
    } else {
	# For now:
	my $trio = substr($prev1Dir,-3);
	return if ($trio < 19); # iteration 20 is first that can be final
	my $nseq1 = readUnsequenced("$prev1Dir/sequenced.txt");
	my $nseq2 = readUnsequenced("$prev2Dir/sequenced.txt");
	return unless ($nseq1 && $nseq2);
	my $inc = $nseq1-$nseq2;
	my $nStr = strainSetSize($strainSet);
	my $th = $REFERENCE_GENOME_SIZE*$nStr*$SEQUENCED_INCREASE_THRESHOLD;
	printf("Increase in sequenced is %d-%d = %d; threshold is %d\n",
	       $nseq1,$nseq2,$inc,$th);
	my $g = "$tgtDir/isFinalIteration";
	if ($inc <= $th) {
	    print "Creating $g\n";
	    touch($g);
	} else {
	    print "Not creating $g\n";
	}
    }
}

sub readUnsequenced {
    my ($f) = @_;
    return undef unless (-s $f);
    my $fh = openToRead($f);
    my $line = <$fh>;
    chomp $line;
    return $line;
}

sub strainSetSize {
    my ($set) = @_;
    my $pAlias = readAliases("$ROOT/aliases/strainSets.txt");
    my $pAllStrains = $pAlias->{$set};
    return scalar(@$pAllStrains);
}

# For each existing set in $pBuilding:
#  -- find its SID if we add $pHGRList->[$i];
#       if OK, add extended version to $pBuilding unless overlap,
#       else add current version to $pHGRSets and
#       delete it from $pBuilding.
# Also form a new set from $pHGRList->[$i] and add it
#   to pBuilding.
# At end: add everything in $pBuilding to $pHGRSets

sub hgrSetOrderp {
    my $iMax = min($#$a,$#$b);
    foreach my $i (1 .. $iMax) {
	foreach my $j (0,1) {
	    my $cmp = $a->[$i][$j] <=> $b->[$i][$j];
	    return $cmp if ($cmp);
	}
    }
    return $#$a <=> $#$b;
}

sub isGhostLine {
    my ($line) = @_;
    my $iMax = length($line)-2;
    for (my $i=0; $i<=$iMax; $i+=2) {
	return 0 if (substr($line,$i,1) =~ /[A-Za-z]/);
    }
    return 1;
}

sub readDangles {
    my ($f) = @_;
    my @ans;
    my $fh = openToRead($f);
    while (<$fh>) {
	chomp;
	my @a = split;
	push @ans,\@a;
    }
    return \@ans;
}

1;
