package SGRP::TranslateGenes;

use Exporter;
@ISA = qw(Exporter);

@EXPORT = qw(translateGenes translateCodon);

use SGRP::Variables;
use SGRP::SGRP;

use Utility;
use File::Basename;
use Cwd;
use Carp;

use strict;

# NOTE: we use Z for stop ...

my %TRANSLATION_CHAR = ( TTT => 'F', TTC => 'F', TTA => 'L', TTG => 'L',
			 TCT => 'S', TCC => 'S', TCA => 'S', TCG => 'S',
			 TAT => 'Y', TAC => 'Y', TAA => 'Z', TAG => 'Z',
			 TGT => 'C', TGC => 'C', TGA => 'Z', TGG => 'W',
			 CTT => 'L', CTC => 'L', CTA => 'L', CTG => 'L',
			 CCT => 'P', CCC => 'P', CCA => 'P', CCG => 'P',
			 CAT => 'H', CAC => 'H', CAA => 'Q', CAG => 'Q',
			 CGT => 'R', CGC => 'R', CGA => 'R', CGG => 'R',
			 ATT => 'I', ATC => 'I', ATA => 'I', ATG => 'M',
			 ACT => 'T', ACC => 'T', ACA => 'T', ACG => 'T',
			 AAT => 'N', AAC => 'N', AAA => 'K', AAG => 'K',
			 AGT => 'S', AGC => 'S', AGA => 'R', AGG => 'R',
			 GTT => 'V', GTC => 'V', GTA => 'V', GTG => 'V',
			 GCT => 'A', GCC => 'A', GCA => 'A', GCG => 'A',
			 GAT => 'D', GAC => 'D', GAA => 'E', GAG => 'E',
			 GGT => 'G', GGC => 'G', GGA => 'G', GGG => 'G');

sub translateGenes {
    my ($species,$filtMeth,$strainSet,@parts) = @_;
    unless (@parts) {
	confess("Usage: sgrp.pl tran cere methc CERE [0-9] ...");
    }
    my ($iter,$dir) = getCurrentIteration($species,$filtMeth,$strainSet);
    my $gff = "$ROOT/ref/$species/genome.gff";
    unless (-s $gff) {
	report("$gff: missing or empty",1);
	return;
    }
    foreach my $p (@parts) {
	foreach my $d (glob("$dir/p$p/*")) {
	    foreach my $suf (qw(z y m)) {
		my $f = "$d/imputed_$suf.gz";
		if (-s $f) {
		    my $g = "$d/translations_$suf";
		    if (stamp($g)) {
			translateGenes1($f,$gff,$g);
			unstamp($g);
		    }
		    last;
		}
	    }
	}
    }
}

sub translateGenes1 {
    my ($f,$gff,$g) = @_;
    my $rf = mayRevcompFile($f);
    my ($pGenesFwd,$pGenesRev) = readGenes($gff,basename(dirname($f)));
    translateGenes2($f,$pGenesFwd,"${g}_fwd");
    translateGenes2($rf,$pGenesRev,"${g}_rev");
    report();
}    

sub mayRevcompFile {
    my ($f) = @_;
    my $d = dirname($f);
    my $b = basename($f);
    my $g = "$d/rev_$b";
    revcompFile($f,$g) unless (-s $g);
    return $g;
}

sub revcompFile {
    my ($f,$g) = @_;
    report("$g: creating");
    my $gh = openToWrite($g);
    my $fh = openToRead($f);
    my $hdr = <$fh>;
    print $gh $hdr;
    $fh->close();
    $fh = openFromPipe(($f =~ /\.gz$/) ? "gunzip -c $f | tac" : "tac $f");
    while (<$fh>) {
	unless ($_ eq $hdr) {
	    my $iLim = length($_);
	    for (my $i=0; $i<$iLim; $i+=2) {
		my $v = substr($_,$i,1);
		$v =~ tr/ACGTacgt/TGCAtgca/;
		substr($_,$i,1) = $v;
	    }
		print $gh $_;
	}
    }
    report();
}

sub readGenes {
    my ($f,$tgtChr) = @_;
    my $fh = openToRead($f);
    my $chrsz;
    my (@genesFwd,@genesRev,%geneLine);
    my %cdsList;
    while (<$fh>) {
	chomp;
	my @a = split;
	my ($chr,undef,$type,$lo,$hi,undef,$strand,$phase,$anno) = @a;
	if ($type eq 'chromosome') {
	    if ($chr eq $tgtChr) {
		$chrsz = $hi;
	    } elsif (@genesFwd) {
		last;
	    }
	} elsif ($chr ne $tgtChr) {
	    next;
	} elsif ($type eq 'gene') {
	    my ($name) = ($anno =~ /.*Name=([^;]+);/);
	    die("No name in $_") unless ($name);
	    $geneLine{$name} = $_;
	    if ($strand eq '+') {
		push @genesFwd,$name;
	    } else {
		unshift @genesRev,$name;
	    }
	} elsif ($type eq 'CDS') {
	    my ($name) = ($anno =~ /.*Name=([^;]+);/);
	    die("No name in $_") unless ($name);
	    if ($strand eq '+') {
		push @{$cdsList{$name}},[$lo,$hi];
	    } else {
		unshift @{$cdsList{$name}},[$chrsz+1-$hi,$chrsz+1-$lo];
	    }
	}
    }
    @genesFwd = sort {$cdsList{$a}[0][0] <=> $cdsList{$b}[0][0]} @genesFwd;
    @genesRev = sort {$cdsList{$a}[0][0] <=> $cdsList{$b}[0][0]} @genesRev;
    my @fwd;
    foreach my $name (@genesFwd) {
	push @fwd,[$geneLine{$name},@{$cdsList{$name}}];
    }
    my @rev;
    foreach my $name (@genesRev) {
	push @rev,[$geneLine{$name},@{$cdsList{$name}}];
    }
    return (\@fwd,\@rev);
}

sub translateGenes2 {
    my ($f,$pGeneList,$tgtDir) = @_;
    report("$tgtDir: creating");
    my $fh = openToRead($f);
    my $hdr = <$fh>;
    #print $gh $hdr;
    my $off = 1; # Reference offset of start of @buf
    my $nref = 0; # no. of reference positions currently in @buf
    my @buf;
    my %leftover;
    while (@$pGeneList) {
	my $pGene = shift @$pGeneList;
	my ($anno,@cdsPairs) = @$pGene;
	my $gh = openToWrite("$tgtDir/" . annotationName($anno) . ".gz");
	print $gh $hdr; # strain list is always needed
	print $gh ">$anno\n";
	my $lo = $cdsPairs[0][0];
	my $hi = $cdsPairs[-1][1];
	# Extend @buf to go as far as $hi
	my $toGet = 1+$hi-($off+$nref);
	while ($toGet > 0) {
	    my $line = <$fh>;
	    # Change N's to ref values for this purpose
	    push @buf,nToRefValue($line);
	    if ($line =~ /^[A-Z]/) {
		$toGet --;
		$nref ++;
	    }
	}
	# Get rid of elements of @buf before $lo
	while ($off < $lo) {
	    if (substr($buf[0],0,1) =~ /^[A-Z]/) {
		$off ++;
		$nref --;
	    }
	    shift @buf;
	}
	# Now buf should be long enough, and should start at beginning
	# of gene.
	my @geneLines;
	my @geneBuf;
	my $k=0; # offset in @buf
	my $kr=0; # REFERENCE offset in @buf
	foreach my $pPair (@cdsPairs) {
	    #print $gh ">CDS @$pPair\n";
	    my ($relLo,$relHi) = ($pPair->[0]-$off,$pPair->[1]-$off);
	    while ($kr <= $relHi) {
		die("Overrun") if ($k > $#buf);
		push @geneBuf,$buf[$k] if ($kr >= $relLo);
		$kr ++ if ($buf[$k] =~ /^[A-Z]/);
		$k ++;
	    }
	}
	# And translate.
	my $k=0;
	while ($k < $#geneBuf) {
	    my $l = nextStartPosition($k,\@geneBuf);
	    last unless (defined $l);
	    my @codonLines = @geneBuf[$k .. $l-1];
	    my $outputLine;
	    my $iLim = length($geneBuf[$k])-1; # -1 because no chomping
	    for (my $i=0; $i<$iLim; $i+=2) {
		$outputLine .= translation(\@codonLines,$i);
		$outputLine .= minQual(\@codonLines,$i+1);
	    }
	    print $gh "$outputLine\n";
	    $k = $l;
	}
	$gh->close();
    }
    report();
}

sub nToRefValue {
    my ($line) = @_;
    my $iLim = length($line);
    my $r = substr($line,0,1);
    for (my $i=2; $i<$iLim; $i+=2) {
	substr($line,$i,1) =~ s/N/$r/;
    }
    return $line;
}

sub annotationName {
    my ($line) = @_;
    foreach my $key (qw(gene Name ID)) {
	if ($line =~ /$key=([^;]+);/) {
	    my $name = $1;
	    $name =~ s/[^-\.A-Za-z0-9]/_/g; # for safety in making file names
	    return $name;
	}
    }
    die("No gene=, Name= or ID= found in: $line");
}

sub nextStartPosition {
    my ($k,$pGeneBuf) = @_;
    my $nSeen = 0; # number seen in ref
    my $k0 = $k;
    while (1) {
	$nSeen ++ if ($pGeneBuf->[$k] =~ /^[A-Z]/);
	$k ++;
	#die("nextStartPosition: overrun at $k") if ($k > $#$pGeneBuf+1);
	return undef if ($k > $#$pGeneBuf+1); # allow overruns; needed for buggy paradoxus GFF
	return $k if ($nSeen == 3);
	if ($k-$k0 >= 3 && $nSeen == 0) {
	    # Check whether 3 seen in _any_ strain
	    my $iLim = length($pGeneBuf->[$k0]);
	    for (my $i=2; $i<$iLim; $i+=2) {
		my $iSeen = 0;
		foreach my $j ($k0 .. $k-1) {
		    $iSeen ++ if (substr($pGeneBuf->[$j],$i,1) =~ /[A-Z]/);
		}
		return $k if ($iSeen == 3);
	    }
	}
    }
}

sub translation {
    my ($pCodonLines,$i) = @_;
    my ($nucs0,$nucs1,$nucs2);
    foreach my $line (@$pCodonLines) {
	my $r = substr($line,0,1);
	my $v = substr($line,$i,1);
	if ($r =~ /^[A-Z]/) {
	    $nucs0 .= $r;
	    $v =~ tr/a-z/A-Z/;
	    if ($v =~ /[ACGT]/) {
		$nucs1 .= $v;
		$nucs2 .= $v;
	    } else {
		$nucs2 .= $r;
	    }
	} elsif ($v =~ /^[A-Z]/) {
	    $nucs1 .= $v;
	    $nucs2 .= $v;
	} # else leave it: gap in both
    }
    my $tr;
    my $l1 = length($nucs1);
    my $l2 = length($nucs2);
    if (length($nucs1) == 3) {
	$tr = translateCodon($nucs1);
    } elsif (length($nucs2) == 3) {
	$tr = translateCodon($nucs2);
	$tr =~ tr/A-Z/a-z/;
    } elsif (length($nucs0) == 3 && (isSublist($nucs1,$nucs0) || isSublist($nucs0,$nucs1))) {
	$tr = translateCodon($nucs0);
	$tr =~ tr/A-Z/a-z/;
    } elsif ($l1 < 3) {
	$tr = '-'; # too few for a codon
    } else {
	$tr = '+'; # too many for a codon
    }
    return $tr;
}

sub translateCodon {
    my ($nucs) = @_;
    if ($nucs =~ /N/) {
	return 'X';
    } else {
	return $TRANSLATION_CHAR{$nucs} || die("No translation for $nucs");
    }
}

sub isSublist {
    my ($list1,$list2) = @_;
    my $len1 = length($list1);
    my $len2 = length($list2);
    my $k1 = 0;
    my $k2 = 0;
    while (1) {
	return 1 if ($k1 == $len1);
	return 0 if ($k2 == $len2);
	if (substr($list1,$k1,1) eq substr($list2,$k2,1)) {
	    $k1 ++;
	}
	$k2 ++;
    }
}

sub minQual {
    my ($pCodonLines,$i) = @_;
    my $min = ' ';
    foreach my $line (@$pCodonLines) {
	my $qc = substr($line,$i,1);
	unless ($qc eq ' ') {
	    if ($min eq ' ' || $min gt $qc) {
		$min = $qc;
	    }
	}
    }
    return $min;
}
	
1;
