#!/usr/bin/perl

# This assumes "split" lines (-s option).

use Getopt::Std;
use FileHandle;

my %OPT;
getopts('c:n:o:',\%OPT); # chr name, no. of input lines (interleaved) to merge, output file

use strict;

alicat2html($OPT{c},$OPT{n},$OPT{o},@ARGV);

sub alicat2html {
    my ($chr,$nMerge,$outFile,@inFiles) = @_;

    my @colors = qw(red green blue orange brown);
    
    my %changeColour = qw(X inter S syn N nonsyn P indel T stop);

    my $gh = $outFile ? new FileHandle(">$outFile") : *STDOUT;

    printHeader($gh,$chr);

    my $pNames;
    my $intab = 0;
    my $nSince = -1;
    @inFiles = ('-') unless (@inFiles);
    foreach my $f (@inFiles) {
	my $fh = $f eq '-' ? *STDIN : new FileHandle($f);
	while (<$fh>) {
	    chomp;
	    if (/^\s*</) {
		# Assume it's HTML already; pass it on (after the header)
		print $gh "$_\n";
	    } elsif (/^>/) {
		# Assume it's a strain list; will print in fives
		$pNames = shortenNames('Consen',split(" ",substr($_,1)));
	    } else {
		$nSince ++;
		my $off = substr($_,0,11);
		$off =~ s/ //g;
		my ($off1,$off2) = split(/\./,$off);
		if ($nSince == 0) { # i.e. initially
		    printStrains($off1,$pNames,$gh);
		    print $gh "<table>\n";
		}

		my @lines = $_;
		if ($nMerge > 1) {
		    foreach my $n (2 .. $nMerge) {
			chomp (my $line = <$fh>);
			push @lines,$line;
		    }
		}
		
		$_ = $lines[-1];
		
		my $cons = substr($_,12,3);
		if (/\| [ A-Z] [+-] [ACGTI]/) {
		    print $gh "</table>\n";
		    if ($nSince >= 15) {
			printStrains($off1,$pNames,$gh);
			$nSince = 0;
		    }
		    print $gh "<table>\n";
		}
		my $col;
		my $offBr = $off2 ? "$off1<br></br>.$off2" : $off;
		my $idx = index($_,'| |');
		if (substr($cons,0,1) eq '[') {
		    my $change = substr($_,2*$idx-12,1);
		    $col = $changeColour{$change};
		    printf($gh "<tr id=o%s><td class=\"lpad %s\">%s</td>",
			   $off,$col,$offBr);
		} else {
		    printf($gh "<tr><td class=lpad>%s</td>",$off);
		}
		printf($gh "<td>%s</td>",substr($cons,1,1));

		my @tails;
		foreach my $line (@lines) {
		    push @tails,substr($line,17);
		}
		printSplitLines(\@tails,$col,$gh);
		print $gh "</tr>";
		print $gh "\n";
	    }
	}
	print $gh "</table>\n";
	
	print $gh "</body>\n";
	print $gh "</html>\n";
    }
}

# Used to have:  html  { font-size:10px; }

sub printHeader {
    my ($gh,$chr) = @_;
    print $gh <<\END;
<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.1//EN"
    "http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd">
<html  xmlns="http://www.w3.org/1999/xhtml" >
<head>
  <title>
END
    print $gh "Chunk of $chr\n"; # should give extent really !!
    print $gh <<\END;
  </title>
  <style> 

    table {border:thin solid black;
           font-family:courier; 
           margin-bottom:0.2em;
           border-collapse:collapse; }

    td {border:thin solid black;}

    .inter  {background-color:#C0C0C0}
    .syn    {background-color:#00FF00}
    .nonsyn {background-color:#00FFFF}
    .indel  {background-color:#FFCC66}
    .stop   {background-color:#FFC8C8}

    .pad {padding:0.5em}

    .lpad {padding:0.5em; width:5em}

    .twoch {width:1.31em}

    .cramped {line-height:75%; color:brown}

  </style>
</head>
<body>
END
}

sub shortenNames {
    my (@names) = @_;
    my @short;
    foreach my $str (@names) {
	$str =~ s/_//g;
	if (length($str) > 6) {
	    $str = substr($str,0,3) . substr($str,-3);
	} else {
	    $str .= ' ' x (7-length($str));
	}
	push @short,$str;
    }
    return \@short;
}

sub printStrains {
    my ($off,$pNames,$gh) = @_;
    print $gh "<table class=cramped><tr>";
    if (defined $OPT{c}) {
	my $chr = $OPT{c};
	if (length($chr) < length($off)) {
	    $chr .= ' ' x (length($off)-length($chr));
	}
	print $gh "<td class=lpad>$chr</td>";
    } else {
	print $gh "<td class=lpad>$off</td>";
    }
    foreach my $i (0 .. $#$pNames) {
	if ($i%5 == 2) {
	    print $gh "<td></td>";
	}
	my $class = ($i < 2 || $i == $#$pNames) ? '' : ' class=twoch';
	printf($gh "<td%s>%s</td>",$class,join("<br></br>",split("",$pNames->[$i])));
    }
    print $gh "</tr></table>\n";
}
    
  
sub printSplitLines {
    my ($pData,$col,$gh) = @_;
    my @idx;
    foreach my $line (@$pData) {
	push @idx,index($line,'| |');
    }
    my (@nucs,@quals);
    foreach my $i (0 .. $#$pData) {
	my $data = $pData->[$i];
	push @nucs,substr($data,0,$idx[$i]);
	my $qstr = substr($data,$idx[$i]+3,$idx[$i]);
	$qstr =~ s/\|$//;
	push @quals,$qstr;
    }
    # Imputed is LAST item in list; get anno from that.
    my $anno = substr($pData->[-1],2*$idx[-1]+7);
    my $r = substr($nucs[-1],0,1);
    $r =~ tr/a-z/A-Z/;
    $r =~ s/[_=]/-/g;
    my $len = length($nucs[-1]);
    my (%color,$nColor);
    for (my $i=0; $i<$len; $i++) {
	my ($v,$q);
	foreach my $j (0 .. $#nucs) {
	    $v .= substr($nucs[$j],$i,1);
	    $q .= substr($quals[$j],$i,1);
	}
	my $vLast = substr($v,-1,1);
	$v =~ tr/ /./;
	substr($q,-1,1) =~ s/ /9/; # last one (e.g. imputed) must be present
	$q =~ tr/ /./; # others (e.g. sorted) can be absent
	$vLast=~ tr/a-z/A-Z/;
	$vLast =~ s/[_=]/-/g;
	if ($vLast eq $r || !defined $r || !defined $col) {
	    print $gh "<td>";
	} else {
	    print $gh "<td class=$col>";
	}
	if ($i == 0) {
	    print $gh substr($v,-1,1);
	} else {
	    print $gh $v;
	    print $gh "<br></br>$q</td>";
	}
	if ($i%5 == 0 && $i+2<$len) {
	    print $gh "<td></td>";
	}
    }
    if ($anno) {
	print $gh "<td rowspan=3 class=pad>$anno</td>";
    }
}

