#!/usr/bin/env perl

use SGRP::OffsetMap;
use Utility;
use File::Basename;
use strict;

# Arg: dir, and strain(s) to process

alicoords2GFF(@ARGV);

sub alicoords2GFF {
    my ($top,@strains) = @_;
    my $pMap = new OffsetMap('?',"$top/assembly/REF/genome.map");
    my $pRegs = $pMap->alignmentRegions(10,1000);
    foreach my $str (@strains) {
	alicoords2gffStr($top,$str,$pMap,$pRegs);
    }
}

sub alicoords2gffStr {
    my ($top,$str,$pMap,$pRegs) = @_;
    my $pIsGoodRead = getGoodReads($str);
    my %alis;
    my %clash;
    foreach my $f (glob("$top/p?/*/alicoords.gz")) {
	# acr.txt should have been produced using alignmentClashRates.pl ...
	readClashingAlignments(dirname($f) . "/acr.txt",$str,\%clash);
	readAlicoords($f,$str,$pIsGoodRead,\%alis);
    }
    my $lstr = strainNameForGFF($str);
    my $g = "$top/gff/$str/match.gff";
    report("$g: writing");
    my $gh = openToWrite($g);
    foreach my $cid (sort keys %alis) {
	printCloneGFF($gh,$str,$lstr,$cid,$alis{$cid},\%clash,$pMap,$pRegs);
    }
    $gh->close();
    report();
}

sub getGoodReads {
    my ($str) = @_;
    my $fh = openToRead("$ENV{SGR}/read/$str/renamed.wm");
    my %isGoodRead;
    while (<$fh>) {
	chomp;
	my ($r,undef,@rest) = split;
	$isGoodRead{$r} = 1 if (@rest);
    }
    return \%isGoodRead;
}

sub readClashingAlignments {
    my ($f,$str,$pClash) = @_;
    if (-s $f) {
	report("$f: reading");
	my $fh = openToRead($f);
	while (<$fh>) {
	    chomp;
	    if (/ ${str}-/) {
		my @a = split;
		if ($a[0] >= 3) {
		    my $name = $a[4];
		    $name =~ s/\|.*//;
		    $pClash->{$name} = 1;
		}
	    }
	}
	report();
    } else {
	report("$f: missing or empty",1);
    }
}

sub strainNameForGFF {
    my ($str) = @_;
    $str =~ tr/a-z/A-Z/;
    $str =~ s/^(\d)/s\1/;
    return $str;
}

sub readAlicoords {
    my ($f,$str,$pIsGoodRead,$pAlis) = @_;
    report("$f: reading");
    my $fh = openToRead($f);
    while (<$fh>) {
	chomp;
	my ($read,$chr,undef,$iter,@rest) = split;
	my ($clone,$pq) = ($read =~ /^$str-(.*)\.([pq])1k$/);
	my $qp = ($pq eq 'p') ? 'q' : 'p';
	my $hasMate = defined $pIsGoodRead->{"${str}-${clone}.${qp}1k"} ? 1 : 0;
	if ($clone) {
	    # Each quad is for one alignment: clo rlo chi rhi, where
	    # rlo > rhi for C alignment
	    my @quads;
	    for (my $i=0; $i<$#rest; $i+=4) {
		push @quads,[@rest[$i..$i+3]];
	    }
	    # Each splitQuad is a list of quads for one (split) alignment for this read.
	    push @{$pAlis->{$clone}{$pq}},splitQuads($chr,$iter,$hasMate,\@quads);
	}
    }
    report();
}

sub splitQuads {
    my ($chr,$iter,$hasMate,$pQuads) = @_;
    my $i=0;
    my @ans;
    while ($i <= $#$pQuads) {
	my $dir = quadDirection($pQuads->[$i]);
	my $j = $i;
	while ($j < $#$pQuads && quadDirection($pQuads->[$j+1]) == $dir && $pQuads->[$j+1][1] == $pQuads->[$j][3]+$dir) {
	    $j++;
	}
	push @ans,{chr => $chr, iter => $iter, quads => [@{$pQuads}[$i..$j]], hasMate => $hasMate};
	$i = $j+1;
    }
    return @ans;
}

sub quadDirection {
    my ($pQuad) = @_;
    my (undef,$rlo,undef,$rhi) = @$pQuad;
    return $rlo < $rhi ? 1 : -1;
}

# "Score" determines colour of alignment. It should show
# how the alignment was placed. Where two numbers are shown,
# second is for clash score >= 3 (alignmentClashRates.pl),
# which should be shown with red border.

#  0     white        poor quality unplaced (read display only)
#  1     red          extra, bisr
#  2,3   lightgreen   extra, close (NOTE change of colour)
#  4,5   blue         extra-chi_late
#  6,7   burlywood    extra-late, no mate
#  8,9   brown        extra-late, mate consistent
#  10,11 blue         mate not placed consistently
#  12,13 burlywood    mate does not exist
#  14,15 brown        mate placed consistently

# Show 9 and below as thin, 10 and above as thick.

sub printCloneGFF {
    my ($gh,$str,$lstr,$cid,$pqAlis,$pClash,$pMap,$pRegs) = @_;
    my @alis = ($pqAlis->{p},$pqAlis->{q});
    # 2 for consistent mate, 1 for inconsistent, within each ali
    decideHasMate(@alis);
    foreach my $i (0,1) {
	my $pAli = $alis[$i];
	if ($pAli) {
	    my $pq = $i ? 'q' : 'p';
	    my $r = "${str}-${cid}.${pq}1k";
	    # Each $pH is a (split) alignment
	    foreach my $j (0 .. $#$pAli) {
		my $pH = $pAli->[$j];
		my $chr = $pH->{chr};
		my $iter = $pH->{iter};
		my $pQuads = $pH->{quads};
		my $hasMate = $pH->{hasMate};
		my $fc = quadDirection($pQuads->[0]) > 0 ? 'F' : 'C';
		# Alignment identifier:
		my $aid = "${lstr}-$cid.${pq}1k";
		$aid .= "." . ($j+1) if ($#$pAli);
		printGFFToRegions($pMap,$pRegs,
		                  $gh,$chr,$lstr,'match',
				  $pQuads->[0][0],$pQuads->[-1][2],
				  undef,$fc,undef,
				  "alignment $aid",1);
		my $score = decideScoreForColour($iter,$hasMate);
		my @recs;
		foreach my $pQuad (@$pQuads) {
		    my $scoreToUse = $score;
		    $scoreToUse ++ if ($score >= 2 && isClashingAlignment($r,$pQuad,$pClash));
		    push @recs,printGFFToRegions($pMap,$pRegs,
						  $gh,$chr,$lstr,'HSP',
						  $pQuad->[0],$pQuad->[2],
						  undef,$fc,$scoreToUse,
						  "alignment $aid",0);
		}
		foreach my $pA (conflateRecords(@recs)) {
		    my ($refLo,$refHi,$scoreToUse) = @$pA;
		    printGFF($gh,$chr,$lstr,'HSP',$refLo,$refHi,undef,$fc,$scoreToUse,"alignment $aid");
		}
	    }
	}
    }
}

sub printGFFToRegions {
    my ($pMap,$pRegs,$gh,$chr,$lstr,$type,$lo,$hi,$foo,$fc,$score,$anno,$doPrint) = @_;
    # Back translate to reference chromosome. We don't print yet, just
    # return the coords, so we can splice together afterwards
    my $rchr = "REF.$chr";
    my $refLo = $pMap->backTranslation($rchr,$lo,2);
    my $refHi = $pMap->backTranslation($rchr,$hi,1);
    my @recs;
    if ($refHi >= $refLo) {
	if ($doPrint) {
	    printGFF($gh,$chr,$lstr,$type,$refLo,$refHi,$foo,$fc,$score,$anno);
	} else {
	    push @recs,[$refLo,$refHi,$score];
	}
    }
    # See if we intersect any LARs. We hope there are few enough that simple
    # linear search is efficient enough. We print this stuff regardless of
    # $doPrint.
    foreach my $pInt (larIntersections($pRegs->{$rchr},$lo,$hi)) {
	my ($regId,$regLo,$regHi) = @$pInt;
	my $name = sprintf("%s.lar%3d",$chr,$regId);
	$name =~ tr/ /0/;
	printGFF($gh,$name,$lstr,$type,$regLo,$regHi,$foo,$fc,$score,$anno);
    }
    return @recs;
}

sub conflateRecords {
    my (@recs) = @_;
    my @ans;
    foreach my $pA (@recs) {
	if (@ans && $ans[-1][1] == $pA->[0]-1) {
	    $ans[-1][1] = $pA->[1];
	    $ans[-1][2] = max($ans[-1][2],$pA->[2]);
	} else {
	    push @ans,$pA;
	}
    }
    return @ans;
}

sub larIntersections {
    my ($pRegs,$lo,$hi) = @_;
    my @ans;
    foreach my $i (0 .. $#$pRegs) {
	my $pReg = $pRegs->[$i];
	return @ans if ($pReg->[2] > $hi);
	if ($pReg->[3] >= $lo) {
	    push @ans,[$i,max(1,1+$lo-$pReg->[2]),1+min($pReg->[3],$hi)-$pReg->[2]];
	}
    }
    return @ans;
}

sub printGFF {
    my ($gh,@arr) = @_;
    if (@arr > 9) {
	confess("Overlong GFF line: @arr");
    }
    unless (defined $gh) {
	confess("Undefined output handle for @arr");
    }
    foreach my $i (0 .. 8) {
	if (!defined $arr[$i]) {
	    $arr[$i] = '.';
	} elsif ($arr[$i] eq 'F') {
	    $arr[$i] = '+';
	} elsif ($arr[$i] eq 'C') {
	    $arr[$i] = '-';
	}
    }
    # Ensure 2nd field doesn't start with a digit
    #$arr[1] = avoidInitialDigit($arr[1]);
    print $gh join("\t",@arr) . "\n";
}

sub decideHasMate {
    my ($pAli,$qAli) = @_;
    # No mates if only one set exists:
    return unless ($pAli && $qAli);
    # 2 means consistent mates; 1 means inconsistent.
    # We could look this up from clustering, but it's probably
    # easier to work it out here.
    foreach my $pH (@$pAli) {
	foreach my $qH (@$qAli) {
	    my $pDir = quadDirection($pH->{quads}[0]);
	    my $qDir = quadDirection($qH->{quads}[0]);
	    if ($pDir > $qDir) {
		next if decideHasMate1($pH,$qH);
	    } else { 
		next if decideHasMate1($qH,$pH);
	    }
	}
    }
}

sub decideHasMate1 {
    my ($fH,$cH) = @_;
    return unless ($fH->{chr} eq $cH->{chr});
    my ($flo,undef,$fhi) = @{$fH->{quads}[0]};
    my ($clo,undef,$chi) = @{$cH->{quads}[-1]};
    # Assume consistent if separation is between 0 and 25000
    # in the _alignment_ ...
    if ($clo >= $flo && $chi >= $fhi && $fhi-$clo <= 25000) {
	$fH->{hasMate} = $cH->{hasMate} = 2;
	return 1;
    } else {
	return 0;
    }
}

sub decideScoreForColour {
    my ($iter,$mate) = @_;
    if ($iter eq 'bisr') {
	return 1;
    } elsif ($iter eq 'close') {
	return 2;
    } elsif ($iter eq 'chi_late') {
	return 4;
    } elsif ($iter eq 'late') {
	return ($mate == 2) ? 8 : 6;
    } elsif ($mate == 2) {
	return 14;
    } elsif ($mate == 1) {
	return 10;
    } else {
	return 12;
    }
}

sub isClashingAlignment {
    my ($r,$pQuad,$pClash) = @_;
    my (undef,$rlo,undef,$rhi) = @$pQuad;
    return defined $pClash->{"$r:$rlo:$rhi"};
}

