#!/usr/bin/env perl

use Getopt::Std;
use File::Basename;
use FileHandle;
use strict;

# Arguments: a list of directories each containing a file called
# "sequenced.gz" (or the value of "-i"). Creates a version of each
# file with NQS applied: a nucleotide or short gap symbol is upper
# case if it passes NQS, lower case otherwise (where lower case "-" is
# "_"). A base or short-gap symbol meets the NQS if its quality score
# is >= 20 ("-q" switch), and the 5 ("-s" switch) bases (not gaps) to
# either side all have quality scores >= 15 ("-n" switch). The output
# file goes in the same directory as the input and has base name
# sequenced_nqs_{-q}_{-n}_{-s}.gz, or as specified by "-o".

autoflush STDOUT 1;

my %OPT = qw(q 15 n 10 s 5 i sequenced.gz);
getopts('q:n:s:i:o:',\%OPT);

# Example:
#
# cd cere/match
# applyNQS.pl 15 10 5 chr*

if (@ARGV) {
    $OPT{o} = sprintf("sequenced_nqs_%d_%d_%d.gz",$OPT{q},$OPT{n},$OPT{s}) unless ($OPT{o});
    applyNQS($OPT{q},$OPT{n},$OPT{s},$OPT{i},$OPT{o},@ARGV);
} else {
    print STDERR <<EOF;
Usage: $0 [switches] dir1 dir2 ...
Switches:
   -q minOwnQual (default 15)
   -n minNbrQual (default 10)
   -s flankSize  (default 5)
   -i inBase     (default sequenced.gz)
   -o outBase    (default sequenced_nqs_{minOwnQual}_{minNbrQual}_{flankSize}.gz
EOF
}

sub applyNQS {
    my ($ownTh,$nbrTh,$flankSize,$inBase,$outBase,@dirs) = @_;
    foreach my $d (@dirs) {
	my $f = "$d/sequenced.gz";
	my $g = sprintf("%s/sequenced_nqs_%d_%d_%d.gz",
			$d,$ownTh,$nbrTh,$flankSize);
	if (-s $f && ! -s $g) {
	    print "$g: creating ...";
	    applyNQS($ownTh,$nbrTh,$flankSize,$f,$g);
	    print "done\n";
	} elsif (-s $g) {
	    print "$g: already exists\n";
	} else {
	    print "$f: not found\n";
	}
    }
}

sub applyNQS1 {
    my ($ownTh,$nbrTh,$flankSize,$f,$g) = @_;
    my $ownCh = chr($ownTh+33);
    my $nbrCh = chr($nbrTh+33);
    my $fh = new FileHandle($f =~ /\.gz$/ ? "gunzip -c $f |" : $f);
    my $gh = new FileHandle($g =~ /\.gz$/ ? "| gzip -c > $g" : ">$g");
    my $hdr = <$fh>;
    print $gh $hdr;
    my @lines = <$fh>;
    for (my $i=0; $i<=$#lines; $i++) {
	print $gh applyNQSToLine($ownCh,$nbrCh,$flankSize,\@lines,$i);
    }
}

sub applyNQSToLine {
    my ($ownCh,$nbrCh,$flankSize,$pLines,$i) = @_;
    my $line = $pLines->[$i];
    chomp $line;
    my $kMax = length($line)-2;
    for (my $k=2; $k<=$kMax; $k+=2) {
	my $v = substr($line,$k,1);
	if ($v =~ /[A-Z0-9]/) {
	    my $qc = substr($line,$k+1,1);
	    if ($qc lt $ownCh 
		|| !goodQualityFlank($nbrCh,$flankSize,$pLines,$k,$i,-1)
		|| !goodQualityFlank($nbrCh,$flankSize,$pLines,$k,$i,1)) {
		substr($line,$k,1) =~ tr/A-Z0-9/a-z__________/;
	    } else {
		substr($line,$k,1) =~ s/[0-9]/-/;
	    }
	}
    }
    return $line . "\n";
}

sub goodQualityFlank {
    my ($nbrCh,$flankSize,$pLines,$k,$i,$d) = @_;
    my $j = $i+$d;
    my $toDo = $flankSize;
    while ($toDo > 0) {
	my $v = substr($pLines->[$j],$k,1);
	return 0 if ($v eq ' ');
	if ($v =~ /[A-Z]/) {
	    my $qc = substr($pLines->[$j],$k+1,1);
	    return 0 if ($qc lt $nbrCh);
	    $toDo --;
	}
	$j += $d;
	return 0 if ($j < 0 || $j > $#$pLines);
    }
    return 1;
}
