#!/usr/bin/env perl

# Input: blast output file(s) for the same query sequence
# Switches:
#   -o outputFile (else to standard output)
#   -q fastqDir (to find quality values).
# Output: layout-style alignment file that can be handed to alicat.pl -L.
#
# Example:
#    blast2ali.pl -o cup1tmp -q cere/strains CUP1-1.fa.out
# To print just the alignments for Y55 reads (see "perldoc alicat.pl"):
#    alicat.pl -L -i Y55 -sp cup1tmp

use Utility;
use Getopt::Std;

use strict;

my %OPT;

getopts('o:q:v',\%OPT);

my (@anno,$name,$stats,$q1,$qseq,$q2,$pos,$dirn,@mat,@ref);
$pos = -1;
while (<>) {
    chomp;
    if (/^>(\S+)/) {
	$name = $1;
    } elsif (/^ (Score|Identities) = /) {
	my @a = split;
	foreach my $i (1 .. $#a-1) {
	    if ($a[$i] eq '=') {
		$stats .= $a[$i+1] . ",";
	    }
	}
    } elsif (/Plus Strand/) {
	$dirn = 1;
    } elsif (/Minus Strand/) {
	$dirn = -1;
    } elsif (/^Query: /) {
	(undef,$q1,$qseq,$q2) = split;
    } elsif (/^Sbjct: /) {
	my (undef,$s1,$sseq,$s2) = split;
	if ($name) {
	    $pos ++;
	    $stats =~ s/,,+/,/g;
	    $stats =~ s/,$//;
	    $anno[$pos] = [$name,$s1,undef,$pos,$stats];
	    $stats = "";
	    $name = undef;
	}
	$anno[$pos][2] = $s2; # always, so we get final value
	# Fill in relevant col of @mat
	my @qList = split(//,$qseq);
	my @sList = split(//,$sseq);
	while (@qList) {
	    my $qv = shift @qList;
	    my $sv = shift @sList;
	    if ($dirn > 0) {
		if ($qv eq '-') {
		    $mat[$q1-1][$pos] .= $sv;
		} else {
		    $mat[$q1][$pos] .= $sv;
		    $ref[$q1] = $qv;
		    $q1 ++;
		}
	    } else {
		$sv =~ tr/ACGT/TGCA/;
		if ($qv eq '-') {
		    $mat[$q1][$pos] = $sv . $mat[$q1][$pos];
		} else {
		    $mat[$q1][$pos] = $sv . $mat[$q1][$pos];
		    $ref[$q1] = $qv;
		    $q1 --;
		}
	    }
	}
    }
}

my @qualString;
my (%wantR,%wantS,@colDirection);
if ($OPT{q}) {
    foreach my $j (0 .. $#anno) {
	my ($r,$s1,$s2,$pos,$stats) = @{$anno[$j]};
	$wantR{$r} = $j;
	my ($s) = split(/-/,$r);
	$wantS{$s} ++;
	$colDirection[$j] = $stats =~ /Plus/ ? 1 : -1;
    }
    foreach my $s (sort keys %wantS) {
	my $f = "$OPT{q}/$s/renamed.fastq";
	report("$f: reading") if ($OPT{v});
	my $fh = openToRead($f);
	my $want;
	my $todo = $wantS{$s};
	while ($todo) {
	    my $hdr = <$fh>;
	    my ($r) = split(" ",substr($hdr,1));
	    <$fh>;
	    <$fh>;
	    my $qLine = <$fh>;
	    if (defined $wantR{$r}) {
		chomp $qLine;
		$qualString[$wantR{$r}] = $qLine;
		$todo --;
	    }
	}
	report() if ($OPT{v});
    }
}	    
    
my $iMin = 1;
while (!$mat[$iMin]) {
    $iMin ++;
}

my $gh = openToWrite($OPT{o});

my @readOff;
foreach my $i ($iMin .. $#mat) {
    my $pA = $mat[$i];
    if ($pA) {
	foreach my $j (0 .. $#$pA) {
	    if ($pA->[$j]) {
		my $pAn = $anno[$j];
		if ($pAn) {
		    printf($gh ">%s:%d:%d %d %s\n",@$pAn);
		    $readOff[$j] = $colDirection[$j] > 0 ? $pAn->[1] : $pAn->[2];
		    $anno[$j] = undef;
		}
	    }
	}
    }
    print $gh ">REF $iMin\n" if ($i == $iMin);
    printf($gh "%s ",$ref[$i]);
    if ($pA) {
	my @rest;
	foreach my $j (0 .. $#$pA) {
	    if ($pA->[$j]) {
		my $v = substr($pA->[$j],0,1);
		my $q;
		if ($v eq '-') {
		    $q = ' ';
		} else {
		    $q = substr($qualString[$j],$readOff[$j]-1,1);
		    $readOff[$j] += $colDirection[$j];
		}
		print $gh $v . $q;
		my @tail = split(//,substr($pA->[$j],1));
		foreach my $i (0 .. $#tail) {
		    $rest[$i][$j] = $tail[$i];
		}
	    }
	}
	foreach my $i (0 .. $#rest) {
	    print $gh "\n- ";
	    foreach my $j (0 .. $#$pA) {
		if ($rest[$i][$j]) {
		    my $q = substr($qualString[$j],$readOff[$j]-1,1);
		    $readOff[$j] += $colDirection[$j];
		    print $gh $rest[$i][$j] . $q;
		} else {
		    print $gh "- ";
		}
	    }
	}
    }
    print $gh "\n";
}
