#!/usr/bin/env perl

use strict;

my $DEBUG = 1;

my $MIN_TO_KEEP = 10;

blastSummary();

sub blastSummary {
    my ($qchr,$schr,$score,$qmin,$qmax,$smin,$smax,$fc,$percid,$idx);
    
    printf("%5s %10s %7s %7s %10s %7s %7s %2s %6s %4s %s\n",
	   "Score","QSeq","QMin","QMax","SSeq","SMin","SMax","FC","PercId","Idx","A");
    
    my @arr;
    while (<>) {
	chomp;
	if (/^Query=\s+(\S+)/) {
	    $qchr = $1;
	} elsif (/^>(\S+)/) {
	    $schr = $1;
	} elsif (/Plus Strand HSPs/) {
	    $fc = 'F';
	} elsif (/Minus Strand HSPs/) {
	    $fc = 'C';
	} elsif (/ Score = (\d+) /) {
	if (defined $score) {
	    $idx ++;
	    push @arr,[$score,$qchr,$qmin,$qmax,$schr,$smin,$smax,$fc,$percid,$idx];
	    $qmin=$qmax=$smin=$smax=$percid=undef;
	}
	$score = $1;
	$score /= 2 if ($fc eq 'C'); # best option?
	} elsif (/ Identities = (\d+)\/(\d+) /) {
	    $percid = 100*$1/$2;
	} elsif (/^Query:/) {
	    my (undef,$q1,undef,$q2) = split;
	    if ($fc eq 'F') {
		$qmin = $q1 unless (defined $qmin);
		$qmax = $q2;
	    } else {
		$qmax = $q1 unless (defined $qmax);
		$qmin = $q2;
	    }
	} elsif (/^Sbjct:/) {
	    my (undef,$s1,undef,$s2) = split;
	    $smin = $s1 unless (defined $smin);
	    $smax = $s2;
	}
    }
    if (defined $score) {
	$idx ++;
	push @arr,[$score,$qchr,$qmin,$qmax,$schr,$smin,$smax,$fc,$percid,$idx];
    }
    # $qTaken{$name} is a list (in increasing order) of regions of $name already
    # allocated. $sTaken likewise for subject sequences.
    # $qTgt{$name} is subject sequence onto which query sequence $name maps (one only!)
    # $accepted{$qName} is a list of alignments already accepted for
    # those sequences.
    my (%qTaken,%sTaken,%qTgt,%accepted);
    foreach my $pA (sort {$b->[0] <=> $a->[0]} @arr) {
	my ($score,$qchr,$qmin,$qmax,$schr,$smin,$smax,$fc,$percid,$idx) = @$pA;
	# subrange of qmin..qmax and smin..smax that we accept; or undef if reject.
	my $pAcc;
	unless (defined $qTgt{$qchr} && $qTgt{$qchr} ne $schr) {
	    $pAcc = shouldAccept($pA,\%qTaken,\%sTaken,\%accepted);
	    $qTgt{$qchr} = $schr;
	}
	printf("%5d %10s %7d %7d %10s %7d %7d %2s %6.2f %4d ",
	       $score,$qchr,$qmin,$qmax,$schr,$smin,$smax,$fc,$percid,$idx);
	if ($pAcc) {
	    printf(" 1 %7d %7d %7d %7d\n",@$pAcc);
	} else {
	    print(" 0\n");
	}
    }
}

# Return 0 to reject, 1 to accept with pruning, 2 to accept without pruning.

sub max {
    my ($x,$y) = @_;
    return ($x > $y) ? $x : $y;
}

sub shouldAccept {
    my ($pAli,$pQTaken,$pSTaken,$pAccepted) = @_;
    my (undef,$qchr,$qmin,$qmax,$schr,$smin,$smax,$fc) = @$pAli;
    # First see if any material from the alignments survives:
    my ($qminKeep,$qmaxKeep) = rangeToKeep($qmin,$qmax,$pQTaken->{$qchr});
    return undef unless ($qmaxKeep-$qminKeep+1 >= max($MIN_TO_KEEP,($qmax-$qmin+1)/20));
    my ($sminKeep,$smaxKeep) = rangeToKeep($smin,$smax,$pSTaken->{$schr});
    return undef unless ($smaxKeep-$sminKeep+1 >= max($MIN_TO_KEEP,($smax-$smin+1)/20));
    # Accepted alignments between this qchr and (because qTgt is a function) schr:
    my $pA = $pAccepted->{$qchr};
    if (defined $pA) {
	# Get the accepted regions on either side of the current one in subject coords,
	# and check query coords order is consistent.
	my ($pALeft,$pARight);
	foreach my $i (0 .. $#$pA) {
	    if ($pA->[$i][2] > $smaxKeep) {
		$pALeft = $pA->[$i-1] if ($i>0);
		$pARight = $pA->[$i];
		last;
	    }
	}
	unless (defined $pARight) {
	    $pALeft = $pA->[-1];
	}
	# Now $pALeft and/or $pARight are defined. Our "keep" coords must be
	# consistent wrt at least one of them -- or both if their orientations
	# are the same.
	my ($okL,$okR);
	if (defined $pALeft) {
	    if ($pALeft->[4] eq 'F') {
		$okL = $qminKeep > $pALeft->[1];
	    } else {
		$okL = $qmaxKeep < $pALeft->[0];
	    }
	}
	if (defined $pARight) {
	    if ($pARight->[4] eq 'F') {
		$okR = $qmaxKeep < $pARight->[0];
	    } else {
		$okR = $qminKeep > $pARight->[1];
	    }
	}
	my $ok;
	if (defined $pALeft && defined $pARight && $pALeft->[4] eq $pARight->[4]) {
	    $ok = $okL && $okR;
	} else {
	    $ok = $okL || $okR;
	}
	return undef unless ($ok);
	    
    }
    my $pNew = [$qminKeep,$qmaxKeep,$sminKeep,$smaxKeep,$fc];
    if (defined $pAccepted->{$qchr}) {
	my @sorted = sort {$a->[2] <=> $b->[2]} (@{$pAccepted->{$qchr}},$pNew);
	$pAccepted->{$qchr} = \@sorted;
    } else {
	$pAccepted->{$qchr} = [$pNew];
    }
    takeRange($qchr,$qminKeep,$qmaxKeep,$pQTaken);
    takeRange($schr,$sminKeep,$smaxKeep,$pSTaken);
    return [$qminKeep,$qmaxKeep,$sminKeep,$smaxKeep];
}

sub takeRange {
    my ($chr,$min,$max,$pH) = @_;
    unless (defined $pH->{$chr}) {
	$pH->{$chr} = [[$min,$max]];
	return;
    }
    $pH->{$chr} = addRange($min,$max,$pH->{$chr});
}

sub addRange {
    my ($min,$max,$pA) = @_;
    my @sorted = sort {$a->[0] <=> $b->[0]} ([$min,$max],@$pA);
    my @ans;
    foreach my $pPair (@sorted) {
	if (@ans && $ans[-1][1]+1 >= $pPair->[0]) {
	    $ans[-1][1] = $pPair->[1]; # merge
	} else {
	    push @ans,$pPair;
	}
    }
    return \@ans;
}

sub rangeToKeep {
    my ($min,$max,$pTaken) = @_;
    my ($lo,$hi) = sort {$a <=> $b} ($min,$max);
    if (defined @$pTaken) {
	foreach my $pA (@$pTaken) {
	    my ($lo2,$hi2) = @$pA;
	    # Finish if we've gone past our range
	    last if ($lo2 > $hi);
	    # Do nothing if we haven't got to it yet
	    next if ($hi2 < $lo);
	    # There is some overlap.
	    my $dLeft = $lo2-$lo;
	    my $dRight = $hi-$hi2;
	    if ($dLeft > $dRight) {
		if ($dLeft < $MIN_TO_KEEP) {
		    return undef;
		} else {
		    $hi = $lo2-1;
		}
	    } else {
		if ($dRight < $MIN_TO_KEEP) {
		    return undef;
		} else {
		    $lo = $hi2+1;
		}
	    }
	}
    }
    return ($lo,$hi);
}
