#!/usr/bin/env perl

use Utility;
use strict;

buildAssembly();

sub buildAssembly {

    # Change these as required. Probably run in $SGR/ref/para.
    #my $insFile = '1106_para_assembly';
    #my $faFile = 'ref4600_shred_lon.fasta';
    #my $supFile = 'ref4600_shred_lon.supercontigs';

    my $insFile = '080520_para_assembly';
    # cat ref4600_shred_lon.fasta kell_notcov.fasta > 
    #    ref4600_and_kellis_notcov.fasta
    # cat ref4600_shred_lon.supercontigs kell_notcov.supercontigs >
    #    ref4600_and_kellis_notcov.supercontigs

    my $faFile = 'ref4600_and_kellis_notcov.fasta';
    my $supFile = 'ref4600_and_kellis_notcov.supercontigs';

    # rDNA is chr12 451419 to 468930 according to self-matching, 
    # In cere dir:
    # (echo '>rDNA'; extractRegion.pl -c chr12 -f 451419 -t 468930 -l 60 genome.fa) > rdna.fa

    my $rDNAFile = 'rdna.fa';
    
    my $outFile = 'genome.fa';
    my $logFile = 'genome.log';
    my $szFile = 'genome.sz';

    if (-s $outFile) {
	die("Already exists: $outFile");
    }
    foreach my $f ($insFile,$supFile,$faFile,$rDNAFile) {
	unless (-s $f) {
	    die("Missing or empty: $f");
	}
    }
    
    my $pIns = readInstructionFile($insFile);
    my $pSupers = readSupercontigsFile($supFile);
    my $pSeq = readFastaFile($faFile);
    $pSeq->{rDNA} = readFastaFile($rDNAFile)->{rDNA};
    
    my $gh = openToWrite($outFile);
    my $lh = openToWrite($logFile);
    my $sh = openToWrite($szFile);
    foreach my $chr (sort keys %$pIns) {
	buildChromosome($chr,$pIns->{$chr},$pSupers,$pSeq,$gh,$lh,$sh);
    }

}

sub readInstructionFile {
    my ($f) = @_;
    my $fh = openToRead($f);
    my %items;
    while (<$fh>) {
	chomp;
	if (/^chr/) {
	    my ($chr,$eq,@stuff) = split;
	    if ($eq eq '=') {
		my @a;
		for (my $i=0; $i<$#stuff; $i+=2) {
		    my $gapSz = $stuff[$i];
		    my $item = $stuff[$i+1];
		    if ($gapSz =~ /^\d+$/) {
			push @a,$gapSz;
		    } else {
			die("Bad gap size $gapSz in line $_");
		    }
		    my $pItem = parseItem($item);
		    if ($pItem) {
			push @a,$pItem;
		    } else {
			die("Bad item $item in line $_");
		    }
		}
		push @a,$stuff[-1]; # final number
		$items{$chr} = \@a;
	    }
	}
    }
    return \%items;
}

sub parseItem {
    my ($item) = @_;
    if ($item eq 'rDNA') {
	return {name => 'rDNA', spec => $item};
    }
    my ($most,$fc,@rest) = split(/\//,$item);
    return undef if (@rest || ($fc ne 'F' && $fc ne 'C'));
    my ($trio,$more) = ($most =~ /^S(\d\d\d)(.*)$/);
    return undef unless (defined $trio);
    my %ans = (name => "Supercontig_0000$trio", fc => $fc, spec => $item);
    my ($lo,$hi) = (1,-1); # -1 means "as high as possible"
    if ($more) {
	$more =~ s/^\.//;
	my ($pre,$post) = split(/-/,$more);
	if (defined $post) {
	    ($lo,$hi) = ($pre,$post);
	} else {
	    my ($n,$offLo,$offHi) = ($more =~ /^(\d+):(\d+)\.\.(\d+)$/);
	    $lo = "$n:$offLo";
	    $hi = "$n:$offHi";
	}
    }
    $ans{lo} = $lo;
    $ans{hi} = $hi;
    return \%ans;
}
    
sub readSupercontigsFile {
    my ($f) = @_;
    my $fh = openToRead($f);
    my ($sup,@conts,%ans);
    while (<$fh>) {
	if (/^supercontig (\S+)$/) {
	    if (defined $sup) {
		my @copy = @conts;
		$ans{$sup} = \@copy;
	    }
	    @conts = ();
	    $sup = $1;
	} elsif (/^contig (\S+)$/) {
	    push @conts,$1;
	}
    }
    if (defined $sup) {
	$ans{$sup} = \@conts;
    }
    return \%ans;
}

sub buildChromosome {
    my ($chr,$pA,$pSupers,$pSeq,$gh,$lh,$sh) = @_;
    print $gh ">$chr\n";
    my $buf;
    for (my $i=0; $i < $#$pA; $i+=2) {
	my $gapSz = $pA->[$i];
	my $pItem = $pA->[$i+1];
	if ($i > 0 || $gapSz > 0) {
	    $buf .= 'N' x 100;
	}
	my $name = $pItem->{name} || die "No name";
	if ($name eq 'rDNA') {
	    $buf .= $pSeq->{rDNA};
	} else {
	    my $lo = $pItem->{lo} || die "No lo";
	    my $hi = $pItem->{hi} || die "No hi";
	    my $fc = $pItem->{fc} || die "No fc";
	    my $pConts = $pSupers->{$name} || die "Unknown supercontig $name";
	    my $seq = getSupercontigSequence($chr,$name,$lo,$hi,$pConts,$pSeq,$lh);
	    if ($fc eq 'C') {
		$seq = reverse($seq);
		$seq =~ tr/ACGT/TGCA/;
	    }
	    printf($lh "%s: %7d for %s\n",$chr,length($seq),$pItem->{spec});
	    $buf .= $seq;
	}
    }
    if ($pA->[-1] > 0) {
	$buf .= 'N' x 100;
    }
    my $jLim = length($buf);
    for (my $j=0; $j < $jLim; $j+=60) {
	print $gh substr($buf,$j,60) . "\n";
    }
    printf($lh "%s: size %7d\n",$chr,$jLim);
    printf($sh "%s %d\n",$chr,$jLim);
}

sub getSupercontigSequence {
    my ($chr,$scName,$lo,$hi,$pConts,$pSeq,$lh) = @_;
    my $seq;
    my ($iLo,$offLo) = split(/:/,$lo);
    my ($iHi,$offHi) = split(/:/,$hi);
    $iLo --; # to make them zero-offset
    if ($iHi < 0) {
	$iHi = $#$pConts;
    } else {
	$iHi --;
    }
    for (my $i=$iLo; $i<=$iHi; $i++) {
	$seq .= 'N' x 50 if ($i>0);
	my $contSeq = $pSeq->{$pConts->[$i]} || die "Unknown contig $pConts->[$i]";
	if ($i == $iHi && defined $offHi) {
	    $contSeq = substr($contSeq,0,$offHi);
	}
	if ($i == $iLo && defined $offLo) {
	    $contSeq = substr($contSeq,$offLo-1);
	}
	printf($lh "%s: %s %s to %s: %s has length %d\n",$chr,$scName,
	       $lo,$hi < 0 ? scalar(@$pConts) : $hi,
	       $pConts->[$i],length($contSeq));
	$seq .= $contSeq;
    }
    return $seq;
}

