#!/usr/bin/env perl

# Input: self-self alignment, e.g. chr01.cere.ali
# A region is a range of start points and a range of end points.
# $regs{$chr} = list of [lolo,lohi,hilo,hihi,count] records.
# Switches:
#   -p min percentage identity for a match [90]
#   -f min frequency of a repeat to output [2]
#   -d max difference in start/end points to cluster [10]
#   -s min size of a difference-region to create [30] -- NOT USED

# Output to stdout: lines of the form
# chr lo hi count
# representing repeated regions which may therefore be deleted;

use Utility;
use Getopt::Std;
use strict;

my %OPT = qw(p 90 d 10 f 2 s 30);
getopts('d:f:p:s:',\%OPT);

my $MDRSZ = 30; # min size for a difference-region

clusterAlignedRegions(\%OPT,@ARGV);

sub clusterAlignedRegions {
    my ($pOpt,@files) = @_;
    my $pRegs = readRegions(\@files,$pOpt->{p},$pOpt->{f});
    foreach my $chr (sort keys %$pRegs) {
	clusterAndPrint($chr,$pRegs->{$chr},$pOpt->{d},$pOpt->{f},$pOpt->{s});
    }
}

sub readRegions {
    my ($pFiles,$minPerc,$minF) = @_;
    my %regs;
    foreach my $f (@$pFiles) {
	my $fh = openToRead($f);
	while (<$fh>) {
	    chomp;
	    my @a = split;
	    if ($a[10] >= $minPerc) {
		my $chr = $a[2];
		my $lo = min($a[4],$a[5]);
		my $hi = max($a[4],$a[5]);
		push @{$regs{$chr}},[$lo,$hi,1];
		# Look for near-tandem repeat (to within 7k, i.e. to
		# include transposons)
		push @{$regs{$chr}},nearTandemRepeatTrios($minF,@a);
	    }
	}
    }
    return \%regs;
}

sub nearTandemRepeatTrios {
    my ($minF,undef,undef,$chr1,$chr2,$lo1,$hi1,$lo2,$hi2,$fc) = @_;
    if ($fc eq 'F' && $chr1 eq $chr2) {
	my $d = $lo2-$hi1;
	if ($d > 0 && $d <= 7000) {
	    # Give "frequency" of $minF to ensure always used
	    return ([$lo1,$lo2-1,$minF],[$hi1+1,$hi2,$minF]);
	}
    }
    return ();
}    

sub clusterAndPrint {
    my ($chr,$pRegs,$minD,$minF,$minS) = @_;
    my $pCRegs = clusterRegions($pRegs,$minD,$minF);
    #my $pDRegs1 = differenceRegions($pCRegs,$minD,$minS);
    #my $pDRegs = clusterRegions($pDRegs1,$minD,$minF);
    # Print results
    foreach my $pR (sort {$a->[0] <=> $b->[0] || $a->[1] <=> $b->[1]} @$pCRegs) { # ,@$pDRegs) {
	printf("%s %7d %7d %3d %s\n",$chr,@$pR);
    }
}

sub clusterRegions {
    my ($pRegs,$minD,$minF) = @_;
    my @regs;
    foreach my $pA (sort {$a->[0] <=> $b->[0]} @$pRegs) {
	push @regs,[$pA->[0],$pA->[0],$pA->[1],$pA->[1],$pA->[2]];
    }
    while (1) {
	my ($dMin,$lolo,$lohi,$hilo,$hihi,$n,$iDel,$jDel) = ($minD+0.5);
	foreach my $i (0 .. $#regs-1) {
	    next unless (defined $regs[$i]);
	    my ($lolo0,$lohi0,$hilo0,$hihi0,$n0) = @{$regs[$i]};
	    foreach my $j ($i+1 .. $#regs) {
		next unless (defined $regs[$j]);
		my ($lolo1,$lohi1,$hilo1,$hihi1,$n1) = @{$regs[$j]};
		my $dLo = $lohi1-$lolo0;
		last if ($dLo >= $dMin);
		my $dHi = max(abs($hihi1-$hilo0),abs($hihi0-$hilo1));
		if ($dHi < $dMin) {
		    $dMin = max($dLo,$dHi);
		    $lolo = $lolo0; # must be = to: min($lolo0,$lolo1);
		    $lohi = max($lohi0,$lohi1);
		    $hilo = min($hilo0,$hilo1);
		    $hihi = max($hihi0,$hihi1);
		    $n = $n0+$n1;
		    $iDel = $i;
		    $jDel = $j;
		}
	    }
	}
	last unless (defined $lolo);
	$regs[$iDel] = [$lolo,$lohi,$hilo,$hihi,$n];
	$regs[$jDel] = undef;
    }
    my @finalRegs;
    foreach my $pR (@regs) {
	if (defined $pR && $pR->[4] >= $minF) {
	    my $lo = int(($pR->[0]+$pR->[1])/2);
	    my $hi = int(($pR->[2]+$pR->[3])/2);
	    push @finalRegs,[$lo,$hi,$pR->[4]];
	}
    }
    @finalRegs = sort {$a->[0] <=> $b->[0] || $a->[1] <=> $b->[1]} @finalRegs;
    return \@finalRegs;
}

# We want difference regions to allow for a delta being left behind after
# the original transposon is deleted.

sub differenceRegions {
    my ($pRegs,$minD,$minS) = @_;
    my @dRegs;
    for (my $i=0; $i<$#$pRegs; $i++) {
	my ($lo0,$hi0,$n0) = @{$pRegs->[$i]};
	for (my $j=$i+1; $j<=$#$pRegs && $pRegs->[$j][0] < $hi0; $j++) {
	    my ($lo1,$hi1,$n1) = @{$pRegs->[$j]};
	    my ($lo2,$hi2);
	    if (abs($lo1-$lo0) <= $minD) {
		$lo2 = min($hi0,$hi1)+1;
		$hi2 = max($hi0,$hi1);
	    } elsif (abs($hi1-$hi0) <= $minD) {
		$lo2 = min($lo0,$lo1);
		$hi2 = max($lo0,$lo1)-1;
	    }
	    if (defined $hi2 && $hi2-$lo2+1 >= $minS) {
		push @dRegs,[$lo2,$hi2,min($n0,$n1)];
	    }
	}
    }
    @dRegs = sort {$a->[0] <=> $b->[0] || $a->[1] <=> $b->[1]} @dRegs;
    return \@dRegs;
}
