#!/usr/bin/env perl

use Utility;
use strict;

# Creates these files in ~/writing/imputing/figs12:
# 1) rates_method_[ic]
#   where method = mf, mfsol, win, star
#       [ic] = i for imputation, c for conf-corr
#   Each line in the file is of the form
#     af errorRate
#   where af is mid point of allele-freq range and
#   errorRate is in [0,1].
# 2) rates_null, rates_nullsol
#   containing "0 r" and "1 r", where r is mean
#   error rate for null (in abi and solexa resp)
# 3) cumul_method_[ic]
#   cumulative error count at increasing af values
# 4) permillion
#   LaTeX table, same structure as tables 5 and 6.

my $TGT_DIR = "$ENV{HOME}/writing/imputing/figs12";

my @RANGE_TOPS = qw(0.001 0.005 0.01 0.02 0.03 0.05 0.10 0.20 0.30
                    0.40 0.50 0.60 0.70 0.80 0.90 0.95 0.97 0.98
                    0.99 0.995 0.999 1);

my @RANGES;
my $prev = 0;
foreach my $top (@RANGE_TOPS) {
    push @RANGES,"$prev:$top";
    $prev = $top;
}
push @RANGES,1;

my %METHODS = (abi => ['null', 'star', 'win', 'mf'], sol => ['null', 'mf']);

my %data;
$data{abi} = gatherData("$ENV{SGR}/match/FREG200_1/freg",0);
$data{sol} = gatherData("$ENV{SGR}/match/FREG200_1S/freg",1);
writeFiles(\%data);
writeTable(\%data); # permillion

sub gatherData {
    my ($d,$isSol) = @_;
    my %data;
    foreach my $r (@RANGES) {
	my $sf = "$d/cia_chrs_errs_complete_mac$r/sum.txt";
	unless (-s $sf) {
	    print "% WARNING: missing $sf\n";
	    next;
	}
	my $af = rangeMidpoint($r);
	my $fh = openToRead($sf);
	my (%den,%num);
	while (<$fh>) {
	    chomp;
	    my @a = split;
	    if ($a[0] eq 'N.m') {
		my ($count,$star,$window,$mf);
		if ($isSol) {
		    (undef,$count,undef,$mf) = @a;
		} else {
		    (undef,$count,undef,$star,$window,$mf) = @a;
		}
		$den{m} += $count;
		$num{m}{star} += $star*$count;
		$num{m}{win} += $window*$count;
		$num{m}{mf} += $mf*$count;
	    } elsif ($a[0] eq 'N.p') {
		my ($count,$null,$star,$window,$mf);
		if ($isSol) {
		    (undef,$count,undef,$null,$mf) = @a;
		} else {
		    (undef,$count,undef,$null,$star,$window,$mf) = @a;
		}
		$den{p} += $count;
		$num{p}{star} += $star*$count;
		$num{p}{win} += $window*$count;
		$num{p}{mf} += $mf*$count;
		$num{p}{null} += $null*$count;
	    }
	}
	$data{$af}{num} = \%num;
	$data{$af}{den} = \%den;
    }
    return \%data;
}

sub rangeMidpoint {
    my ($r) = @_;
    my ($lo,$hi) = split(/:/,$r);
    return defined $hi ? ($lo+$hi)/2 : $r;
}

sub writeFiles {
    my ($pData) = @_;
    foreach my $type qw(abi sol) {
	my $pD = $pData->{$type};
	my $suf = $type eq 'abi' ? '' : $type;
	foreach my $meth (@{$METHODS{$type}}) {
	    foreach my $ic (qw(i c)) {
		next if ($ic eq 'i' && $meth eq 'null');
		my $mp = $ic eq 'i' ? 'm' : 'p';
		my $g = sprintf("%s/rates_%s%s_%s",$TGT_DIR,$meth,$suf,$ic);
		my $gh = openToWrite($g);
		my $hh;
		unless ($meth eq 'null') {
		    my $h = sprintf("%s/cumul_%s%s_%s",$TGT_DIR,$meth,$suf,$ic);
		    $hh = openToWrite($h);
		}
		my ($sumN,$sumD);
		my $cumul = 0;
		#my %cumulH;
		foreach my $af (sort {$a <=> $b} keys %$pD) {
		    my $num = $pD->{$af}{num}{$mp}{$meth};
		    my $den = $pD->{$af}{den}{$mp};
		    if ($meth eq 'null') {
			$sumN += $num;
			$sumD += $den;
		    } else {
			unless ($af == 1) {
			    printf($gh "%5.3f %8.6f\n",$af,($num+0.5)/($den+1));
			}
			$cumul += $num;
			printf($hh "%5.3f %d\n",$af,$cumul);
			#$cumulH{$af} = $cumul;
		    }
		}
		if ($meth eq 'null') {
		    foreach my $x (0,1) {
			printf($gh "%5.3f %8.6f\n",$x,$sumN/$sumD);
		    }
		#} else {
		#    foreach my $af (0,sort {$a <=> $b} keys %cumulH) {
		#	printf($hh "%5.3f %d\n",$af,$cumul-$cumulH{$af});
		}
	    }
	}
    }
}

sub writeTable {
    my ($pData) = @_;
    my $gh = openToWrite("$TGT_DIR/permillion");
    print $gh "Operation &   Type   &  Cases  &   Star   & Window   &    MF    & MF Solexa\\\\ \\hline\n";
    my %tags = qw(i Impute c Conf/Corr b Both);
    foreach my $ic (qw(c i b)) {
	foreach my $mpb (qw(mono poly both)) {
	    printf($gh "%9s ",$mpb eq 'poly' ? $tags{$ic} : "");
	    printf($gh "&   %s   ",capitalize($mpb));
	    my $count = countCases($pData,'abi',$ic,$mpb);
	    printf($gh "& %6.2fM ",$count/1000000);
	    foreach my $type (qw(abi sol)) {
		foreach my $meth (@{$METHODS{$type}}) {
		    unless ($meth eq 'null') {
			my $epm = errorsPerMillion($pData,$type,$ic,$mpb,$meth);
			printf($gh "& %7d  ",$epm);
		    }
		}
	    }
	    print $gh "\\\\ \\hline";
	    print $gh " \\hline" if ($mpb eq 'both');
	    print $gh "\n";
	}
    }
}

sub capitalize {
    my ($w) = @_;
    substr($w,0,1) =~ tr/a-z/A-Z/;
    return $w;
}

sub countCases {
    my ($pData,$type,$ic,$mpb) = @_;
    my ($num,$den) = sumNumAndDen($pData,$type,$ic,$mpb);
    return $den;
}

sub sumNumAndDen {
    my ($pData,$type,$ic,$mpb,$meth) = @_;
    my $pD = $pData->{$type};
    my @mpList;
    if ($ic eq 'i') {
	@mpList = qw(m);
    } elsif ($ic eq 'c') {
	@mpList = qw(p);
    } else {
	@mpList = qw(m p);
    }
    my ($num,$den);
    foreach my $af (keys %$pD) {
	unless (($mpb eq 'mono' && $af < 1) ||
		($mpb eq 'poly' && $af == 1)) {
	    foreach my $mp (@mpList) {
		$den += $pD->{$af}{den}{$mp};
		if (defined $meth) {
		    $num += $pD->{$af}{num}{$mp}{$meth};
		}
	    }
	}
    }
    return ($num,$den);
}

sub errorsPerMillion {
    my ($pData,$type,$ic,$mpb,$meth) = @_;
    my ($num,$den) = sumNumAndDen($pData,$type,$ic,$mpb,$meth);
    return 1000000*$num/$den;
}
    
