#!/usr/bin/env perl

=head1 NAME

compareImputedAcross.pl

=head1 SYNOPSIS

compareImputedAcross.pl mainDir heldOutDir, e.g. methc09 methc91

=head1 DESCRIPTION

=head1 AUTHOR

David Carter, C<dmc@sanger.ac.uk>

=head1 SEE ALSO

       

=cut

use Utility;
use File::Basename;
use Getopt::Std;
use Cwd;
use strict;

autoflush STDOUT 1;

# TYPES of evaluation: see th/2006/1128.

my %OPT;
getopts('bm:t:u:',\%OPT);
# If we have "-m file" we should have "-t lo:hi" as well; and there should be only one @chr.
my $MAF_FILE = $OPT{m};
# Meaning of -t range: if simply 0, then MAF=0 only; else if lo:hi, lo < maf <= hi.
my ($POLY_LO,$POLY_HI) = split(/:/,$OPT{t}); # allele freq range of minor allele
my ($OFFN,$OFFD);
# -u value should be N:D; only lines with off%D == N are examined
if ($OPT{u}) {
    ($OFFN,$OFFD) = split(/:/,$OPT{u});
}

my @MAF_LIMS = qw(undef 0.001 0.005 0.01 0.02 0.03 0.05 0.1 0.2 0.3 0.4 0.51);

my $MAX_Q = 94;

my $QFAC = 5; # size for each quality bucket for reporting

# INDICES used for input files and prediction types:
# 0 held-out sorted (we eval against these; no predictions!)
# 1 sorted (no imputation)
# 2 imputed_d (dumb, i.e. star-topology from genomic sorted.bl)
# 3 imputed_g (imputed by strain distance, static across genome, i.e. sorted.bl)
# 4 imputed_c (imputed by strain distance, static across chromosome)
# 5 imputed_v (imputed by strain distance, 500 kb windows)
# 6 imputed_w (imputed by strain distance, 200 kb windows)
# 7 imputed_h (imputed by strain distance, 100 kb windows)
# 8 imputed_f (imputed by strain distance,  50 kb windows)
# 9 imputed_e (imputed by strain distance,  20 kb windows)
# 10 imputed_t (imputed by strain distance,  10 kb windows)
# 11 imputed_m (Margarita)
# 12 imputed_u.gz (union; created with combineAlignmentFiles.pl)
# 13 oracle (no input file, predict here on fly)

my @SUFFIXES = (undef,undef,'d','g','c','v','w','h','f','e','t','m','u',undef);
my @TAGS     = qw(Heldout Null Star Static Chrom W500k W200k W100k W50k W20k W10k Margarita Union Oracle);

compareImputedAcross(@ARGV);

sub compareImputedAcross {
    my ($mainDir,$hoDir,@chrList) = @_;
    if (@_ < 2) {
	die("Usage: $0 [-m mafFile -t lo:hi] mainDir heldOutDir [chr1 chr2 ...]");
    }
    checkSwitchCombinations(scalar(@chrList));
    if ($OPT{b}) {
	bsubCompareImputedAcross($mainDir,$hoDir,@chrList);
    } else {
	reallyCompareImputedAcross($mainDir,$hoDir,@chrList);
    }
}

sub checkSwitchCombinations {
    my ($nChrs) = @_;
    if (!$OPT{b} && $nChrs > 1 && defined $MAF_FILE) {
	die("$0: if -m is specified, there must be only one chr given");
    }
    if (defined $MAF_FILE) {
	if (!defined $POLY_LO) {
	    die("$0: if -m is specified, -t lo:hi must also be specified");
	}
	if (! -s $MAF_FILE) {
	    die("$0: missing or empty: $MAF_FILE");
	}
    }
    if (defined $POLY_LO) {
	if (($POLY_LO != 1 && !defined $POLY_HI) || $POLY_LO < 0 || $POLY_HI > 1 || (defined $POLY_HI && $POLY_LO > $POLY_HI)) {
	    print "LO $POLY_LO HI $POLY_HI\n";
	    die("$0: value of -t must be of form lo:hi where 0 <= lo <= hi <= 1");
	}
	if (!defined $MAF_FILE) {
	    die("$0: if -t is given, -m must also be given");
	}
    }
}

sub bsubCompareImputedAcross {
    my ($mainDir,$hoDir,@chrList) = @_;
    my $swExpr;
    foreach my $sw (qw(m p t)) {
	if (defined $OPT{$sw}) {
	    $swExpr .= " -$sw -$OPT{$sw}";
	}
    }
    if ($hoDir eq 'complete' && ! -d $hoDir) {
	$hoDir = "$ENV{SGR}/match/FREG_COMPLETE/" . extractNHaps($mainDir);
    }
    foreach my $dir ($mainDir,$hoDir) {
	unless (-d $dir) {
	    die("$0: missing directory $dir");
	}
    }
    my $ciaDir = sprintf("%s/cia_%s_%s",dirname($mainDir),dirname($hoDir));
    # MORE HERE...
    die("MORE NEEDED HERE");
}

sub extractNHaps {
    my ($dir) = @_;
    foreach my $d ($dir,cwd()) {
	if ($d =~ /FREG(\d+)_/) {
	    return $1;
	}
    }
    die("$0: hoDir = complete but cannot parse $dir in " . cwd());
}

sub reallyCompareImputedAcross {
    my ($mainDir,$hoDir,@chrList) = @_;
    # $count{$type}{$alg}[$v] where $type is [NG][012][pm], $alg is [nrusw],
    # and $v is 0 for wrong, 1 for right.
    my %count;
    # SNP discovery: $nPoly[$i][$k] is apparent minor allele count for method $i.
    my @nPoly;
    my $pMaf = readMinorAlleleFrequencies($MAF_FILE); # may be undef
    my $pSuffixes = collectData($mainDir,$hoDir,\@chrList,$pMaf,\%count,\@nPoly);
    printPoly($pSuffixes,\@nPoly);
    unless (@chrList) { # in which case we'll have just printed the table ...
	print "Overall results:\n";
	printResults($pSuffixes,\%count);
    }
}

sub readMinorAlleleFrequencies {
    my ($f) = @_;
    my %maf;
    return undef unless (-s $f);
    my $fh = openToRead($f);
    while (<$fh>) {
	chomp;
	my ($off,$freq) = split;
	$maf{$off} = $freq;
    }
    return \%maf;
}

sub collectData {
    my ($mnTopDir,$hoTopDir,$pChrList,$pMaf,$pCount,$pNPoly) = @_;
    my $lhtd = length($hoTopDir);
    my %doChr = map {$_ => 1} @$pChrList;
    # Get all the held-out sorted.gz files for that part, with
    # smallest first so we spot problems early.
    my $pSuffixes;
    foreach my $hoSor (filesFromSmallest(glob("$hoTopDir/p[0-9]/*/sorted.gz"))) {
	my $hoDir = dirname($hoSor);
	my $hChr = basename($hoDir); # chr of held-out dir
	next if (%doChr && !$doChr{$hChr});
	# Dir under mnTopDir that is equivalent to hoDir:
	my $mnDir = $mnTopDir . substr($hoDir,$lhtd);
	$pSuffixes = collectDataForChr($mnDir,$hoDir,$hChr,$pMaf,$pCount,$pNPoly);
    }
    return $pSuffixes;
}

# Typical values:
#    mnDir  methc09/p1/chr01
#    hoDir  methc91/p1/chr01

sub collectDataForChr {
    my ($mnDir,$hoDir,$chr,$pMaf,$pCount,$pNPoly) = @_;

    # Work out the corresponding main-directory sorted and
    # imputed files.
    my @files = ("$hoDir/sorted.gz",
		 "$mnDir/sorted.gz"); 
    my @suffixes = (undef,undef); # of the files actually present

    # Check everything is present; return if not.
    foreach my $f (@files) {
	if (! -s $f) {
	    report("not doing $mnDir: missing $f",1);
	    return undef;
	} elsif (-f "$f.touch") {
	    report("not doing $mnDir: found $f.touch",1);
	    return undef;
	}
    }

    foreach my $i (0 .. $#SUFFIXES) {
	if ($SUFFIXES[$i]) {
	    my $f = sprintf("%s/imputed_%s.gz",
			    $mnDir,$SUFFIXES[$i]);
	    if (checkReady($f)) {
		push @files,$f;
		push @suffixes,$SUFFIXES[$i];
	    }
	}
    }
    push @suffixes,undef; # to represent oracle

    # If any of these are missing or incomplete, we don't try and
    # process them, but we do carry on (even if just to evaluate the oracle)
    # as long as sorted and held-out are present.
    if (@files >= 2) {
	reallyCollectDataForChr($chr,$pMaf,$pCount,$pNPoly,\@files,\@suffixes);
    }
    return \@suffixes;
}

sub checkReady {
    my ($f) = @_;
    if (-f "$f.touch" || ! -s $f) {
	report("not doing $f: missing or being built",1);
	return undef;
    } else {
	return $f;
    }
}

sub reallyCollectDataForChr {
    my ($chr,$pMaf,$pCount,$pNPoly,$pFiles,$pSuffixes) = @_;
    my %count;
    my $ok = compareImputedAcross1($chr,$pMaf,\%count,$pNPoly,$pFiles,$pSuffixes);
    if ($ok) {
	print "Results for $chr:\n";
	printResults($pSuffixes,\%count);
	while (my ($k1,$pA1) = each %count) {
	    foreach my $k2 (0 .. $#$pA1) {
		my $pA2 = $pA1->[$k2];
		foreach my $k3 (0 .. $#$pA2) {
		    foreach my $k4 (0,1) {
			$pCount->{$k1}[$k2][$k3][$k4] += $pA2->[$k3][$k4];
		    }
		}
	    }
	}
    }
}

sub compareImputedAcross1 {
    my ($chr,$pMaf,$pCount,$pNPoly,$pFiles,$pSuffixes) = @_;

    my %mcNemar;
    # Input streams when files are present:
    my @in = openToReadWherePresent($pFiles);
    # Check headers identical and find number of (non-ref) strains
    my ($nStr,$pHMMap) = strainCount(\@in,$chr);
    return 0 unless ($nStr);

    my $off=0; # offset in chromosome (from 1)
    my @saved = oneLineFromEachStream(@in); # Non-insert lines saved till next time

    while (1) {
	# Stream numbers:
	my @lines = readLineGroups(\@in,$pHMMap,\@saved);
	last unless (@lines && $lines[0]);
	$off ++;
	report("at offset $off",1) if ($off%100000 == 0);
	#last if ($off > 10000);
	next if (defined $OFFD && $off%$OFFD != $OFFN); # apply -u filter
        # $pV->[$j][$i][$k] is nucleotide for: stream $i, offset $j in group, strain $k
	# $pQ->[$j][$i][$k] is quality of that.
	my ($pV,$pQ) = parseLineGroups($nStr,$pFiles,$off,@lines);
	# For each position in group ...
	foreach my $j (0 .. $#$pV) {
	    my $pVJ = $pV->[$j];
	    my $pQJ = $pQ->[$j];
	    # Sorted (main-dir) data:
	    my $pVJ1 = $pVJ->[1];
	    my $pQJ1 = $pQJ->[1];
	    # Find basic type (second char of type) from main-sorted line
	    my $ty2 = basicType($pVJ1);
	    my ($mostAvail,$pAvail) = findAvailability($pVJ1,$pQJ1);
	    # Filter out inappropriate MAFs
	    my $alleleWanted;
	    if (defined $pMaf) {
		if (defined $pMaf->{$off}) {
		    # Frequency of the C and A alleles
		    my %af = (C => $pMaf->{$off}, A => 1-$pMaf->{$off});
		    if (defined $POLY_HI) {
			foreach my $v (qw(A C)) {
			    if ($af{$v} >= $POLY_LO && $af{$v} < $POLY_HI) {
				$alleleWanted = $v;
				last;
			    }
			}
			next unless ($alleleWanted);
		    } else {
			next; # no poly-hi, so want mono only
		    }
		} else {
		    next if (defined $POLY_HI);
		}
	    }
	    augmentNPoly($pV->[$j],$pNPoly);
	    # Evaluate for STRAINS (not ref) only, so from $k = 1 ...
	    for (my $k=1; $k<=$#{$pVJ->[0]}; $k++) {
		if ($pQJ->[0][$k] >= 40) { # only of interest if present and high-qual in held-out
		    compareImputedAcross2($pVJ,$pQJ,$off,$k,$ty2,$alleleWanted,$mostAvail,$pAvail,$pFiles,$pSuffixes,$pCount,\%mcNemar,
					  $j,\@lines);
		}
	    }
	}
    }
    reportMcNemarData($chr,\%mcNemar,$pSuffixes);
    return 1; # success
}

sub openToReadWherePresent {
    my ($pFiles) = @_;
    my @in;
    foreach my $i (0 .. $#$pFiles) {
	if ($pFiles->[$i]) {
	    #$in[$i] = openToRead($pFiles->[$i]);
	    $in[$i] = openFromPipe("gunzip -c $pFiles->[$i] | grep '^[>A-Z]'");
	}
    }
    return @in;
}

sub strainCount {
    my ($pIn,$chr) = @_;
    my $hdr;
    my (@strainsH,@strainsM);
    my $line1;
    my $n=0;
    foreach my $fh (@$pIn) {
	my @strList;
	if ($fh) {
	    my $line = <$fh>;
	    chomp $line;
	    $line =~ s/^>//;
	    $line =~ s/ MRCA$//; # ignore MRCA
	    @strList = split(" ",$line);
	    if ($n == 0) {
		@strainsH = @strList;
	    } elsif ($n == 1) {
		@strainsM = @strList;
		$line1 = $line;
	    } else {
		unless ($line eq $line1) {
		    warn("Header mismatch for $chr:\n   $line1\n   $line\n");
		    return undef;
		}
	    }
	}
	$n++;
    }
    my %strPosM; # position of strain in $strTab[1], from 1; if empty, does not map.
    for (my $k=0; $k<=$#strainsM; $k++) {
	$strPosM{$strainsM[$k]} = $k+1;
    }
    my @hmMap;
    for (my $k=0; $k<=$#strainsH; $k++) {
	$hmMap[$k+1] = $strPosM{$strainsH[$k]};
    }
    return (scalar(@strainsM),\@hmMap);
}

sub oneLineFromEachStream {
    my @in = @_;
    my @saved;
    foreach my $fh (@in) {
	my $line;
	if ($fh) {
	    chomp ($line = <$fh>);
	}
	push @saved,$line;
    }
    return @saved;
}

# For each handle in hList, return list of lines, of which
# only the first (from $pSaved) starts with a non-gap. If we end up with
# lengths 1 and some L>1, extend the 1's to L's; else discard
# all length>1 as we can't easily match things up.

sub readLineGroups {
    my ($pHList,$pHMMap,$pSaved) = @_;
    my @ans;
    my %seenLen;
    foreach my $k (0 .. $#$pHList) {
	my $fh = $pHList->[$k];
	my @lines = ($pSaved->[$k]);
	my $going = $pSaved->[$k];
	$pSaved->[$k] = undef;
	if ($fh) {
	    while (<$fh>) {
		chomp;
		$_ = applyHMMap($pHMMap,$_) if ($k == 0);
		if (/^[-0-9]/) {
		    push @lines,$_;
		} else {
		    if (/^[a-z]/) {
                      substr($_,0,2) = '= ';
                    }
		    $pSaved->[$k] = $_;
		    $going = 1;
		    last;
		}
	    }
	    return undef unless ($going);
	    push @ans,\@lines;
	    $seenLen{scalar(@lines)} = 1;
	} else {
	    push @ans,[]; # nothing, since stream is absent
	}
	#return undef unless ($going);
    }
    my @keys = sort {$a <=> $b} keys %seenLen;
    if (@keys == 1) {
	1; # easy case
    } elsif (@keys == 2 && $keys[0] == 1) {
	# Extend the length-one items
	my $nExt = $keys[1]-$keys[0];
	# Find default gap character
	my $dft;
	foreach my $pLine (@ans) {
	    if (@$pLine > 1) {
		$dft = substr($pLine->[1],0,2);
		last;
	    }
	}
	foreach my $i (0 .. $#ans) {
	    my $pLine = $ans[$i];
	    if (@$pLine == 1) {
		my $n = length($pLine->[0])/2;
		my $dftLine;
		for (my $k=0; $k<2*$n; $k+=2) {
		    if (substr($pLine->[0],$k,1) ne ' ' &&
			substr($pSaved->[$i],$k,1) ne ' ') {
			$dftLine .= $dft;
		    } else {
			$dftLine .= '  ';
		    }
		}
		foreach my $j (1 .. $nExt) {
		    # All gaps:
		    push @$pLine,$dftLine;
		}
	    }
	}
    } else {
	# Discard all the gap lines
	foreach my $pLine (@ans) {
	    if (@$pLine > 1) {
		@$pLine = ($pLine->[0]);
	    }
	}
    }
    return @ans;
}

sub applyHMMap {
    my ($pHMMap,$hLine) = @_;
    my $expanded = substr($hLine,0,2); # start with ref
    my $len = 2;
    for (my $k=1; $k<=$#$pHMMap; $k++) {
	my $m = $pHMMap->[$k];
	if ($m) {
	    my $mm = $m+$m;
	    my $kk = $k+$k;
	    if ($len < $mm) {
		my $diff = $mm-$len;
		$expanded .= ' ' x $diff;
		$len = $mm;
	    }
	    substr($expanded,$mm,2) = substr($hLine,$kk,2);
	}
    }
    return $expanded;
}
# @lines is a list of refs of line groups, one for each file being compared.

# A line group is a line where the reference has a nucleotide, followed
# by 0 or more lines where it has a gap.

# Return \@v and \@q, where:
#   v[j][i][k] = upper cased nucleotide for line j in group, 
#                file i in file list (see above compareImputedAcross1 above)
#                strain k (k=0 for reference)
#   q[j][i][k] = quality score for that nucleotide (always 40 for reference)

sub parseLineGroups {
    my ($nStr,$pFiles,$off,@lines) = @_;
    my (@v,@q);
    my $len = 2*($nStr+1);
    my $ref = substr($lines[0][0],0,1);
    foreach my $i (0 .. $#lines) {
	my $pLines = $lines[$i];
	next unless (@$pLines); # could be missing file
	unless (substr($pLines->[0],0,1) eq $ref) {
	    printf(STDERR "REF MISMATCH at offset %d\n",$off);
	    printf(STDERR "  %s: %s\n",$pFiles->[0],$lines[0][0]);
	    printf(STDERR "  %s: %s\n",$pFiles->[$i],$pLines->[0]);
	    die();
	}
	foreach my $j (0 .. $#$pLines) {
	    my $line = $pLines->[$j];
	    # Just in case:
	    if (length($line) < $len) {
		$line .= ' ' x ($len-length($line));
	    }
	    foreach my $k (0 .. $nStr) {
		my $k2 = $k*2;
		my $val = substr($line,$k2,1);
		if ($val && $val ne ' ') {
		    # Regularise nucleotide values:
		    if ($val =~ /[1-9_]/) {
			$val = '-';
		    } elsif ($val eq '=') {
			return ([],[]); # GIVE UP if any hard gaps found!
		    } else {
			$val =~ tr/acgt/ACGT/;
		    }
		    # NOTE: we swap i and j here...
		    $v[$j][$i][$k] = $val;
		    my $qc = substr($line,$k2+1,1);
		    # Ref quals should all be 40, else ' ' means 127-33
		    $q[$j][$i][$k] = ($k == 0) ? 40 : ($qc eq ' ' ? 127 : ord($qc)) - 33;
		} elsif ($i == 0) {
		    1;
		}
	    }
	}
    }
    my $pV0 = $v[0];
    my $ref = $pV0->[0][0];
    foreach my $i (1 .. $#$pV0) {
	if ($pV0->[$i]) {
	    unless ($pV0->[$i][0] eq $ref) {
		die("MISMATCH: $ref vs $pV0->[$i][0]");
	    }
	}
    }
    return (\@v,\@q);
}

sub basicType {
    my ($pV) = @_;
    my %seen;
    foreach my $v (@$pV) {
	$seen{$v} = 1 if ($v && $v ne ' ');
    }
    if (scalar(keys %seen) == 1) {
	return 0; # only one value
    } elsif ($seen{'-'}) {
	return 2; # gap and a nucleotide
    } else {
	return 1; # two nucleotides
    }
}

sub findAvailability {
    my ($pVJ1,$pQJ1) = @_;
    my ($mostAvail,%available);
    # "Availability" of each nucleotide is sum of its qualities
    # over all the strains it appears in.
    foreach my $k (0 .. $#$pVJ1) {
	my $v = $pVJ1->[$k];
	if ($v) {
	    $available{$v} += $pQJ1->[$k];
	}
    }
    my @keys = sort {$available{$b} <=> $available{$a}} keys %available;
    $mostAvail = $keys[0];
    return ($mostAvail,\%available);
}

sub augmentNPoly {
    my ($pVJ,$pNPoly) = @_;
    my $iMax = $#$pVJ;
    for (my $i=1; $i<=$iMax; $i++) {
	$pNPoly->[$i][apparentMAFRange($pVJ->[$i])] ++;
    }
}

sub apparentMAFRange {
    my ($pA) = @_;
    my $ref = $pA->[0];
    my $nVals = scalar(@$pA);
    # nSame starts as 1 because of k=0
    my ($nSame,$nDiff) = (1,0);
    for (my $k=1; $k<=$#$pA; $k++) {
	my $v = $pA->[$k];
	if ($v) {
	    if ($v eq $ref) {
		$nSame ++;
	    } else {
		$nDiff ++;
	    }
	}
    }
    return 0 unless ($nDiff);
    my $frac = min($nSame,$nDiff)/($nSame+$nDiff);
    foreach my $i (1 .. $#MAF_LIMS) {
	return $i if ($frac <= $MAF_LIMS[$i]);
    }
}

# $pVJ and $pQJ are nucleotide and quality matrices for some offset $j in a line group.
# We are interested in strain $k on this call.

sub compareImputedAcross2 {
    my ($pVJ,$pQJ,$off,$k,$ty2,$alleleWanted,$mostAvail,$pAvail,$pFiles,$pSuffixes,$pCount,$pMcNemar,$j,$pLines) = @_;

    my $ans = $pVJ->[0][$k];
    return if (defined $alleleWanted && $ans ne $alleleWanted);
    my $type = findPositionType($ans,$ty2,$pVJ->[1][$k]);
    my $isPresent = substr($type,2) eq 'p';

    foreach my $i (($isPresent ? 1 : 2) .. $#$pFiles) {
	if ($pFiles->[$i]) {
	    # Round to nearest; will multiply up on printing.
	    my $q = int(0.5+$pQJ->[$i][$k]/$QFAC);
	    my $v = $pVJ->[$i][$k];
	    $pCount->{$type}[$i][$q][$v eq $ans] ++;
	}
    }

    # Index $#$pSuffixes is "o" ("oracle") imputation algorithm. If the sequenced value
    # is wrong and the right answer has higher SUM of qualities overall than the
    # sequenced value has HERE, it's conceivable that, with the right shaped ARG
    # and a following wind, the right answer would be chosen, so use it.
    my $oGuess = $pVJ->[1][$k];
    if ($oGuess ne $ans && $pAvail->{$ans} > $pQJ->[1][$k]) {
	$oGuess = $ans;
    } elsif (!$oGuess) {
	$oGuess = $pVJ->[2][$k]; # go with dumb; won't change the scores...
    }
    $pCount->{$type}[$#$pSuffixes][0][$oGuess eq $ans] ++;

    # For debugging:
    reportSuperiority($type,'N1m',$ans,$oGuess,$pVJ->[11][$k],'Oracle','Margarita',$off,$j,$k,$pLines,$pVJ->[0],$pVJ->[1]);
    reportSuperiority($type,'N1p',$ans,$oGuess,$pVJ->[11][$k],'Oracle','Margarita',$off,$j,$k,$pLines,$pVJ->[0],$pVJ->[1]);

    # Note disagreement for McNemar counts
    countForMcNemar($type,$pVJ,$pQJ,$off,$k,$ans,$pSuffixes,$pMcNemar);
}

sub findPositionType {
    my ($ans,$ty2,$sorVal) = @_;
    # Answer is gap or nucleotide:
    my $ty1 = ($ans eq '-') ? 'G' : 'N';
    # Present or missing in sorted:
    my $ty3 = $sorVal ? 'p' : 'm';
    return "$ty1$ty2$ty3";
}

sub countForMcNemar {
    my ($type,$pVJ,$pQJ,$off,$k,$ans,$pSuffixes,$pMcNemar) = @_;
    foreach my $i (1 .. $#$pSuffixes-2) {
	foreach my $j ($i+1 .. $#$pSuffixes-1) {
	    my $vi = $pVJ->[$i][$k];
	    my $vj = $pVJ->[$j][$k];
	    if ($vi eq $vj || !defined $vi || !defined $vj) {
		# do nothing
	    } else {
		my $qi = $pQJ->[$i][$k];
		my $qj = $pQJ->[$j][$k];
		my $si = $pSuffixes->[$i] || 'n';
		my $sj = $pSuffixes->[$j];
		if ($vi eq $ans) {
		    if ($type eq 'N1m') {
			print "$si beats $sj at $off and $k with quals $qi $qj\n";
		    }
		    $pMcNemar->{$type}[$i][$j] ++;
		} elsif ($vj eq $ans) {
		    if ($type eq 'N1m') {
			print "$sj beats $si at $off and $k with quals $qj $qi\n";
		    }
		    $pMcNemar->{$type}[$j][$i] ++;
		}
	    }
	}
    }
}

sub reportMcNemarData {
    my ($chr,$pMcNemar,$pSuffixes) = @_;
    my $nf = $#$pSuffixes;
    foreach my $type (sort keys %$pMcNemar) {
	my $pA = $pMcNemar->{$type};
	foreach my $i (($type =~ /p$/ ? 1 : 2) .. $nf-1) {
	    foreach my $j ($i+1 .. $nf) {
		my $bi = $pA->[$i][$j];
		my $bj = $pA->[$j][$i];
		my $z = mcNemarZValue($bi,$bj);
		printf("Better %3s %5s %s:%s = %3d %3d with z = %6.3f\n",
		       $type,$chr,$pSuffixes->[$i] || 'n',$pSuffixes->[$j] || 'n',$bi,$bj,$z);
	    }
	}
    }
}

sub mcNemarZValue {
    my ($n1,$n2) = @_;
    my $n = $n1+$n2;
    return $n ? ($n1-$n2)/sqrt($n1+$n2) : 0;
}    

# Each output line is:
# Type Count Proportion pe(n) [pe(@$pSuffixes)...] pe(o)

# Type is:
#  [NG]:  answer is nucleotide or gap
#  [012]: "basicType": whether main sorted line has only one value (0), two or more nucleotides (1),
#         or gap and one or more nucleotides (2).
#  [pm]:  present or missing in main sorted.

# pe(.) indices are as in @$pSuffixes:

sub printResults {
    my ($pSuffixes,$pCount) = @_;
    my $sum;
    foreach my $ty1 (qw(N G)) {
	foreach my $ty2 (qw(0 1 2)) {
	    foreach my $ty3 (qw(p m)) {
		my $type = "$ty1$ty2$ty3";
		my $pAT = $pCount->{$type};
		my $pA = $pAT->[2];
		foreach my $q (0 .. $#$pA) {
		    $sum += $pA->[$q][0]+$pA->[$q][1];
		}
	    }
	}
    }
    print "Type      Count Proportion      pe(n)";
    foreach my $suf (@$pSuffixes) {
	print "      pe($suf)" if ($suf);
    }
    print "      pe(o)\n";
    foreach my $ty1 (qw(N G)) {
	foreach my $ty2 (qw(0 1 2)) {
	    foreach my $ty3 (qw(p m)) {
		my $type = "$ty1$ty2$ty3";
		my $pA = $pCount->{$type}[2];
		next unless (defined @$pA);
		my @sum01;
		foreach my $q (0 .. $#$pA) {
		    my $pAQ = $pA->[$q];
		    if (defined @$pAQ) {
			$sum01[0] += $pAQ->[0];
			$sum01[1] += $pAQ->[1];
		    }
		}
		my $n = $sum01[0]+$sum01[1];
		next unless ($n > 0);
		# Print line for all qualities together:
		printLineForQuality($pSuffixes,$type,$n,$n/$sum,undef,$pCount->{$type});
		# Now print for each quality separately:
		foreach my $q (0 .. $MAX_Q) {
		    printLineForQuality($pSuffixes,$type,undef,undef,$q,$pCount->{$type});
		}
	    }
	}
    }
}

sub printLineForQuality {
    my ($pSuffixes,$type,$n,$prop,$q,$pCT) = @_;
    my $line;
    if (defined $q) {
	my $q2 = sprintf("%2d",$q*$QFAC);
	$q2 =~ tr/ /0/;
	$line = sprintf("%3s.%2s %8s %10s",$type,$q2,'.','.');
    } else {
	$line = sprintf("%3s    %8d %10.8f",$type,$n,$prop);
    }
    my $useful;
    foreach my $alg (1 .. $#$pSuffixes) {
	my $pA = $pCT->[$alg];
	if (defined @$pA && defined $q) {
	    $pA = $pA->[$q];
	}
	if (!defined $pA) {
	    $line .= sprintf(" %4s %4s ",'.','.');
	} elsif (defined $q) {
	    my $den = (defined @$pA) ? ($pA->[0]+$pA->[1]) : undef;
	    if ($den) {
		$line .= sprintf(" %4s %4s ",encodeKMG($pA->[0]),encodeKMG($den));
		$useful = 1;
	    } else {
		$line .= sprintf(" %4s %4s ",'.','.');
	    }
	} else {
	    # need to add up the values over all q's
	    my ($n0,$n1);
	    foreach my $qq (0 .. $#$pA) {
		if (defined $pA->[$qq]) {
		    $n0 += $pA->[$qq][0];
		    $n1 += $pA->[$qq][1];
		}
	    }
	    my $den = $n0+$n1;
	    if ($den) {
		my $pe = $n0/$den;
		$line .= sprintf(" %10.8f",$pe);
		$useful = 1;
	    } else {
		$line .= sprintf(" %10s",'.');
	    }
	}
    }
    print "$line\n" if ($useful);
}

sub encodeKMG {
    my ($n) = @_;
    my $s = encodeKMG1($n);
    return (' ' x (4-length($s))) . $s;
}

sub encodeKMG1 {
    my ($n) = @_;
    if ($n < 10000) { # allow 4 digits ...
	return $n || 0;
    } elsif ($n < 1000000) {
	return int(0.5+$n/1000) . 'k';
    } elsif ($n < 1000000000) {
	return int(0.5+$n/1000000) . 'm';
    } else {
	return int(0.5+$n/1000000000) . 'g';
    }
}

sub printPoly {
    my ($pSuffixes,$pNPoly) = @_;
    print "\nSNP discovery counts\nMethod";
    foreach my $i (0 .. $#MAF_LIMS) {
	printf(" %6.3f",min(0.5,$MAF_LIMS[$i]));
    }
    print "\n";
    for (my $i=1; $i<=$#$pNPoly; $i++) {
	my $suf = $pSuffixes->[$i] || 'n';
	my @counts = @{$pNPoly->[$i]};
	printf("   %s  ",$suf);
	foreach my $i (0 .. $#MAF_LIMS) {
	    printf(" %6d",$counts[$i]);
	}
	print "\n";
    }
}

sub reportSuperiority {
    my ($type,$tyRep,$ans,$v1,$v2,$name1,$name2,$off,$j,$k,$pLines,@pAnsSeqLines) = @_;
    return unless ($type eq $tyRep);
    unless (($v1 eq $ans) == ($v2 eq $ans)) {
	my @nWithAns;
	foreach my $i (0,1) {
	    foreach my $v (@{$pAnsSeqLines[$i]}) {
		if ($v eq $ans) {
		    $nWithAns[$i] ++;
		} elsif ($v ne ' ' && $v ne '' && $i == 1) {
		    $nWithAns[2] ++; # seq'd NOT with ans
		}
	    }
	}
	printf("Answer %s, #ans %d, #seq %d:%d, %s %s, %s %s at offset %d.%d, strain %d, type %s ...\n",
	       $ans,$nWithAns[0],$nWithAns[1],$nWithAns[2],$name1,$v1,$name2,$v2,$off,$j,$k,$type);
	if (0) {
	    printf("%s**\n", ' ' x (14+$k*2));
	    foreach my $i (0 .. $#$pLines) {
		if (defined $pLines->[$i][$j]) {
		    printf("   %9s: %s\n",$TAGS[$i],$pLines->[$i][$j]);
		}
	    }
	}
    }
}
