#!/usr/bin/env perl

use FileHandle;
use Carp;
use strict;

# Args (all optional): inFile outFile argseqOutFile

# When debugging this, may need to set $DB::deep (from within debugger)
# to something >> 100.

# Representation of branches: see checkValidBranch below.

# Output format:
#
# Each ARG is printed in turn, beginning with "ARG N" where N is the
# ARG identifier (0,1,2,...).
#
# Then we have "ar" (active region) lines, each describing a node in
# the compressed ARG, in the form
#   ar Ident Lo Hi CreationTime OldIdent
# where:
#   Ident is the identifier of the node in this (compressed) version
#     of the ARG; Idents should form a complete set 0,1,2,...,some N
#   OldIdent is its identifier in the input (i.e. the Margarita output)
#   Lo and Hi are the minimum and maximum marker identifiers of the
#     range over which this node is active, i.e. over which it contributes
#     some genetic material to some leaf node. The first marker identifier
#     is 0.
#   CreationTime is the creation time of this node. A non-terminal node
#     node has a creation time 1 more than the one given in
#     input, i.e. the first field of the input line where the OldIdent
#     of this node appears as a mother. Leaf nodes have a creation time
#     of zero. (The metaphor here is that time runs backwards: time 0
#     is the present, and as time increass, mothers get created from
#     daughters. Recommended reading: "An Age", by Brian Aldiss).
#
# Then we have lines for the ARG itself. These are in an extended
# version of the input format. The possibilities are:
#
# Time co LDtr RDtr Mthr : LMuts : RMuts
# Time re Dtr LMthr RMthr SwitchAt : LMuts : RMuts
# Time mu Dtr Mthr : Muts
# Time er Dtr Mthr : Muts
#
# These lines represent coalescences, recombinations, mutations and
# sequencing errors, shown by the second field, as in Margarita. The 
# other items are:
#
#   Time: creation time of this event; this should agree with the
#         CreationTime shown for the "Mthr" node(s) on the line
#   Dtr, LDtr, RDtr: daughter node identifiers (single, left, right).
#   Mthr, LMthr, RMthr: mother node identifiers (single, left, right)
#   Muts, LMuts, RMuts: zero or more locations (marker identifiers) of mutations
#         occurring on the (single, left, right) branch of this event
#   SwitchAt in a 're' event is the location of the recombination: the
#         mother is the LMthr at positions up to and including SwitchAt,
#         and the RMthr for positions greater than SwitchAt.
#
# In the compression process, nodes that are _both_ the mother of one
# mu (or er) branch _and_ the daughter of another are deleted. This means
# that most surviving events are 're' and 'co'. However, a 'mu' branch can
# survive if its daughter is a coalescence mother and its mother is a 
# recombination daughter. 'mu' and 'er' branches can also occur when their
# daughter is a leaf node and their mother is a recombination daughter.

# If any ARGSEQ line is found, the rest of the file is copied to
# outFile.argseqs (shifting any .gz to the end of the name)

if (@ARGV <= 3) {
    compressARG(@ARGV);
} else {
    die("Usage: $0 [inFile [outFile [argseqOutFile]]");
}

sub compressARG {
    my ($f,$g,$a) = @_;
    if (! -s $f && -s "$f.gz") {
	$f = "$f.gz";
    }
    my $fh = openToRead($f);
    my $gh = openToWrite($g);
    readARGPreamble($fh,$gh);
    my $n = 0;
    my $ah;
    $ah = openToWrite($a) if ($a);
    while (my $pArg = readARG($fh,$ah)) {
	compressARG1($pArg);
	print $gh "ARG $n\n";
	writeARG($pArg,$gh);
	$n ++;
    }
    $ah->close() if ($ah);
    unlink($a) unless (-s $a);
}

sub openToRead {
    my ($f) = @_;
    if (! defined $f) {
	return *STDIN;
    } elsif ($f =~ /gz$/) {
	return new FileHandle("gunzip -c $f |");
    } elsif (-s "$f.gz") {
	openToRead("$f.gz");
    } else {
	return new FileHandle($f);
    }
}

sub openToWrite {
    my ($g) = @_;
    if (! defined $g) {
	return *STDOUT;
    } elsif ($g =~ /gz$/) {
	return new FileHandle("| gzip -c > $g");
    } else {
	return new FileHandle($g,'w');
    }
}


# Read and copy all material up to and including the first "ARG"
# line (except for %ARGS line produced by MargaritaLite).

sub readARGPreamble {
    my ($fh,$gh) = @_;
    while (<$fh>) {
	return if (/^ARG /);
	print $gh $_ unless (/^%ARGS/);
    }
}

sub readARG {
    my ($fh,$ah) = @_;
    # $arg[$i] will be the node with index $i.
    my @arg;
    # Read each line in the ARG, stopping when
    # we get to the next ARG (or EOF).
  LOOP:
    while (<$fh>) {
	if (/^ARG /) {
	    last;
	} elsif (/^ARGSEQ/) {
	    unless ($ah) {
		die("Found ARGSEQ line, but no argseqOutputFile specified");
	    }
	    # Copy stuff to $aFile until next ARG line if any
	    print $ah $_;
	    while (<$fh>) {
		last LOOP if (/^ARG /);
		print $ah $_;
	    }
	} else {
	    chomp;
	    incrementARG($_,\@arg);
	}
    }
    # If we got some stuff, set the type of each node
    # (see below) and return \@arg.
    if (@arg) {
	# Change "er" to "mu" on all branches that dominate
	# a non-"er" branch.
	detectBackMutationBranches(\@arg);
	foreach my $pN (@arg) {
	    setType($pN,\@arg);
	}
	return \@arg;
    } else {
	return undef;
    }
}

sub incrementARG {
    my ($line,$pARG) = @_;
    my @b = split(/\s+/,$line);
    my ($time,$type,@rest) = @b;
    $time ++; # so that terminal nodes, which have 0 time, are OK.
    # %bh is "branch hash" -- our representation, for
    # now, of a branch. Defined fully with checkValidBranch.
    if ($type eq 'mu' || $type eq 'er') {
	my ($d,$m,$loc) = @rest;
	my %bh = (time => $time, type => $type, 
		  dtr => $d, mthr => $m, muts => [$loc]);
	# This branch is the "up" of the daughter node and the
	# "down" of the mother node.
	$pARG->[$d]{up} = $pARG->[$m]{down} = \%bh;
    } elsif ($type eq 're') {
	my ($d,$m1,$m2,$loc) = @rest;
	my %bh1 = (time => $time, type => $type, 
		   dtr => $d, mthr => $m1, muts => []);
	my %bh2 = (time => $time, type => $type, 
		   dtr => $d, mthr => $m2, muts => []);
	$pARG->[$d]{up1} = $pARG->[$m1]{down} = \%bh1;
	$pARG->[$d]{up2} = $pARG->[$m2]{down} = \%bh2;
	$pARG->[$d]{loc} = $loc;
    } elsif ($type eq 'co') {
	my ($d1,$d2,$m) = @rest;
	my %bh1 = (time => $time, type => $type, 
		   dtr => $d1, mthr => $m, muts => []);
	my %bh2 = (time => $time, type => $type, 
		   dtr => $d2, mthr => $m, muts => []);
	$pARG->[$d1]{up} = $pARG->[$m]{down1} = \%bh1;
	$pARG->[$d2]{up} = $pARG->[$m]{down2} = \%bh2;
    }
}

# Change branch types from "er" to "mu" when they dominate
# anything other than "er".

sub detectBackMutationBranches {
    my ($pArg) = @_;
    my @done;
    foreach my $i (0 .. $#$pArg) {
	detectBackMutationBranches1($pArg,\@done,$i,0);
    }
}

# "DNE" is "dominates non-error".

sub detectBackMutationBranches1 {
    my ($pArg,$pDone,$i,$dne) = @_;
    return if ($pDone->[$i]);
    $pDone->[$i] = 1;
    my $pN = $pArg->[$i];
    if (defined $pN->{up}) {
	my $pB = $pN->{up};
	if ($pB->{type} eq 'er') {
	    $pB->{type} = 'mu' if ($dne); # the point of this
	} else {
	    $dne = 1;
	}
	detectBackMutationBranches1($pArg,$pDone,$pB->{mthr},$dne);
    } elsif (defined $pN->{up1}) {
	foreach my $pN ($pN->{up1}, $pN->{up2}) {
	    detectBackMutationBranches1($pArg,$pDone,$pN->{mthr},1);
	}
    }
}

# Set the type of a node. If none of the conditions holds, there's no
# type, and this node will be discarded during compression.

sub setType {
    my ($pN,$pArg) = @_;
    if (defined $pN->{down} && $pN->{down}{mthr} == 15) {
	1;
    }
    if (!defined $pN->{down} && !defined $pN->{down1}) {
	# If no down branch, it's a terminal node.
	$pN->{type} = 'term';
    } elsif (!defined $pN->{up} && !defined $pN->{up1}) {
	# If no up branch, it's a root node.
	$pN->{type} = 'root';
    } elsif (defined $pN->{up1}) {
	# If up1 is defined, this node is a recombination daughter.
	$pN->{type} = 'recdau';
    } elsif (defined $pN->{down1}) {
	# If down1 is defined, this node is a coalescent mother.
	$pN->{type} = 'coamot';
    } elsif (isIntermediateNode($pN,$pArg)) {
	$pN->{type} = 'inter'; # daughter of coalescent, mother of recomb
    } elsif ($pN->{down}{type} eq 'er' && $pN->{up}{type} ne 'er') {
	# If down is error branch and up is not, it's a preterminal,
	# i.e. it represents the true sequence of one of our strains.
	$pN->{type} = 'preterm';
    }
}

sub isIntermediateNode {
    my ($pN,$pArg) = @_;
    if (defined $pN->{up} && defined $pN->{down}) {
	my $m = $pN->{up}{mthr};
	my $d = $pN->{down}{dtr};
	return (defined $pArg->[$m]{down1} &&
		defined $pArg->[$d]{up1});
    } else {
	return 0;
    }
}

sub compressARG1 {
    my ($pArg) = @_;
    my @done;
    for (my $k=0; $k <= $#$pArg; $k++) {
	unless (defined $pArg->[$k]{type}) {
	    eliminateNode($pArg,$k);
	}
    }
    setActiveRegions($pArg);
}

# Node must have an "up" and a "down", i.e. it's a simple one.

sub eliminateNode {
    my ($pArg,$k) = @_;
    my $pK = $pArg->[$k];    # this node, to be eliminated
    my $pUp = $pK->{up};     # branch up from this node
    my $m = $pUp->{mthr};    # mother node index (to keep)
    my $pM = $pArg->[$m];    # mother node itself
    my $pDown = $pK->{down}; # branch down from this node
    my $d = $pDown->{dtr};   # daughter node index (to keep)
    my $pD = $pArg->[$d];    # daughter node itself
    my $downKey;
    foreach my $key (qw(down down1 down2)) {
	if (defined $pM->{$key} && $pM->{$key} == $pUp) {
	    $downKey = $key;
	    last;
	}
    }
    my $upKey;
    foreach my $key (qw(up up1 up2)) {
	if (defined $pD->{$key} && $pD->{$key} == $pDown) {
	    $upKey = $key;
	    last;
	}
    }
    if ($downKey && $upKey) {
	# Construct new branch to connect $pM and $pD, leaving out $k
	my %bh = (time => max($pDown->{time},$pUp->{time}),
		  type => combineBranchTypes($pUp->{type},$pDown->{type}),
		  mthr => $m,
		  dtr => $d,
		  muts => [@{$pDown->{muts}},@{$pUp->{muts}}]);
	$pM->{$downKey} = $pD->{$upKey} = \%bh;
    } else {
	1;
    }
    foreach my $pNode ($pM,$pD) {
	if (isIntermediateNode($pNode,$pArg)) {
	   $pNode->{type} = 'inter'; # daughter of coalescent, mother of recomb
	}
    }
}

sub combineBranchTypes {
    my ($ut,$dt) = @_;
    if ($ut eq $dt) {
	return $ut;
    } elsif ($ut eq 'er' || $ut eq 'mu') {
	return $dt;
    } elsif ($dt eq 'er' || $dt eq 'mu') {
	return $ut;
    } else {
	1;
    }
}

sub setActiveRegions {
    my ($pArg) = @_;
    my $max = maxOffsetInArg($pArg);
    foreach my $k (0 .. $#$pArg) {
	last unless ($pArg->[$k] && $pArg->[$k]{type} eq 'term');
	$pArg->[$k]{lo} = 0;
	$pArg->[$k]{hi} = $max;
    }
    foreach my $k (0 .. $#$pArg) {
	if (defined $pArg->[$k] && defined $pArg->[$k]{lo}) {
	    setActiveRegion1($pArg,$k);
	}
    }
}

# We could get this from the preamble, but ...
sub maxOffsetInArg {
    my ($pArg) = @_;
    my $max;
    foreach my $pN (@$pArg) {
	if ($pN) {
	    # We can go 1 to the right of a recomb:
	    $max = max($max,$pN->{loc}+1) if (defined $pN->{loc});
	    foreach my $key (qw(up up1 up2)) {
		if (defined $pN->{$key} && defined $pN->{$key}{muts}) {
		    foreach my $mut (@{$pN->{$key}{muts}}) {
			$max = max($max,$mut);
		    }
		}
	    }
	}
    }
    return $max;
}

sub setActiveRegion1 {
    my ($pArg,$k) = @_;
    my $pN = $pArg->[$k];
    my $lo = $pN->{lo};
    my $hi = $pN->{hi};
    if (defined $pN->{up}) {
	setActiveRegion2($pArg,$pN->{up}{mthr},$lo,$hi);
    } elsif (defined $pN->{up1}) {
	my $mid = $pN->{loc}; # recomb point
	setActiveRegion2($pArg,$pN->{up1}{mthr},$lo,$mid);
	setActiveRegion2($pArg,$pN->{up2}{mthr},$mid+1,$hi);
    }
}

sub setActiveRegion2 {
    my ($pArg,$k,$lo,$hi) = @_;
    my $pN = $pArg->[$k];
    my ($lo2,$hi2) = regionUnion($pN->{lo},$pN->{hi},$lo,$hi);
    $pN->{lo} = $lo2;
    $pN->{hi} = $hi2;
}

sub regionUnion {
    my ($lo1,$hi1,$lo2,$hi2,$k) = @_;
    if (defined $lo1 && defined $hi1) {
	if ($hi1+1 < $lo2 || $hi2+1 < $lo1) {
	    croak("Bad region union: @_");
	}
    }
    return (min($lo1,$lo2),max($hi1,$hi2));
}

sub min {
    my ($x,$y) = @_;
    if (defined $y && $x > $y) {
	return $y;
    } elsif (defined $x) {
	return $x;
    } else {
	return $y;
    }
}

sub max {
    my ($x,$y) = @_;
    if (defined $y && $y > $x) {
	return $y;
    } elsif (defined $x) {
	return $x;
    } else {
	return $y;
    }
}

sub writeARG {
    my ($pCA,$gh) = @_;
    # $pMap specifies new IDs for the nodes we're going to keep.
    my $pMap = nodeIDMap($pCA);
    # We "write" the events to @event, where we can have at most
    # one event at each time.
    my @event;
    foreach my $k (0 .. $#$pCA) {
	if (defined $pMap->[$k]) { # surviving node ...
	    my $pN = $pCA->[$k];
	    my $ct = creationTime($pN);
	    printf($gh "ar %d %d %d %d %d\n",
		   $pMap->[$k],$pN->{lo},$pN->{hi},$ct,$k);
	    if (defined $pN->{down1}) {
		# pN is mother of coalescence
		writeCoalescence($pCA,$pN,$k,\@event,$pMap);
	    } 
	    if (defined $pN->{up1}) {
		# pN is daughter of recombination
		writeRecombination($pCA,$pN,$k,\@event,$pMap);
	    }
	    # Will only write if actually simple, i.e. if $pN->{up}
	    # exists and has type 'er' or 'mu':
	    writeSimple($pCA,$pN,$k,\@event,$pMap);
	}
    }
    # Actually write the events, in time order:
    foreach my $t (0 .. $#event) {
	if ($event[$t]) {
	    printf($gh "%d %s\n",$t,$event[$t]);
	}
    }
}

sub nodeIDMap {
    my ($pCA) = @_;
    my $idx = 0;
    my @map;
    foreach my $k (0 .. $#$pCA) {
	if (defined $pCA->[$k] && defined $pCA->[$k]{type}) {
	    $map[$k] = $idx;
	    $idx ++;
	}
    }
    return \@map;
}

sub creationTime {
    my ($pN) = @_;
    foreach my $key qw(down down1 down2) {
	if (defined $pN->{$key}) {
	    return $pN->{$key}{time};
	}
    }
    # must be terminal ...
    return 0;
}

sub writeCoalescence {
    my ($pArg,$pN,$k,$pEv,$pMap) = @_;
    my $t = $pN->{down1}{time};
    die("Event clash at $t") if ($pEv->[$t]);
    $pEv->[$t] =
	sprintf("co %d %d %d : %s : %s",
		$pMap->[$pN->{down1}{dtr}],
		$pMap->[$pN->{down2}{dtr}],
		$pMap->[$k],
		join(" ",@{$pN->{down1}{muts}}),
		join(" ",@{$pN->{down2}{muts}}));
}

# "mu" and "er" output events are created here.
# They arise in two situations:
#   mu or er: upwards from a terminal node to
#             a recombination daughter
#   mu only:  daughter node is a coalescence mother,
#             mother is a recombination daughter
# This is because we can only compress "mu" events
# away when either the mother is a coalescence or
# the daughter is a recombination.

sub writeSimple {
    my ($pArg,$pN,$k,$pEv,$pMap) = @_;
    my $pUp = $pN->{up};
    my $type = $pUp->{type};
    if ($type eq 'er' || $type eq 'mu') {
	my $t = $pUp->{time};
	die("Event clash at $t") if ($pEv->[$t]);
	$pEv->[$t] =
	    sprintf("%s %d %d : %s",
		    $type,
		    $pMap->[$pUp->{dtr}],
		    $pMap->[$pUp->{mthr}],
		    join(" ",@{$pUp->{muts}}));
    }
}
    
sub writeRecombination {
    my ($pArg,$pN,$k,$pEv,$pMap) = @_;
    my $pUp1 = $pN->{up1};
    my $pUp2 = $pN->{up2};
    my $t = $pUp1->{time};
    die("Event clash at $t") if ($pEv->[$t]);
    $pEv->[$t] =
	sprintf("re %d %d %d %d : %s : %s",
		$pMap->[$k],
		$pMap->[$pUp1->{mthr}],
		$pMap->[$pUp2->{mthr}],
		$pN->{loc},
		join(" ",@{$pUp1->{muts}}),
		join(" ",@{$pUp2->{muts}}));
}
    
	
