#!/usr/bin/env perl

use Utility;
use File::Basename;
use Getopt::Std;

use strict;

autoflush STDOUT 1;

my %OPT;
getopts('q',\%OPT);

my $REL = '/lustre/work1/sanger/dmc/sgroot/release3';

my $SGR = $ENV{SGR};

my @loSp = qw(cere para baya);

if ($OPT{q} && -d $REL) {
    report("$REL: already exists",1);
} else {
    constructRelease();
}
constructTarFiles();

sub constructRelease {
    my @mDirs = qw(match/CERE/cere/methc match/PARA/para/methc);

    chdir($SGR);

    my $pSp = decideSpecies();

    # Miscellaneous
    foreach my $b (qw(alicat.pl applyNQS.pl plotStrainSimilarities.pl tabulateSNPsByWindows.pl
                      Utility.pm OffsetMap.pm strainDivergenceValues.pl)) {
	toRel("$ENV{HOME}/perl/$b","code/$b");
    }
    mkdir_p("$REL/trees");
    foreach my $suf (qw(q40 tree ps)) {
	foreach my $f (glob("$ENV{HOME}/strainDistances/*.$suf")) {
	    my $b = basename($f);
	    if ($suf eq 'q40') {
		(my $f2 = $f) =~ s/q40$/strains/;
		$b =~ s/q40$/matrix/;
		`paste $f2 $f > $REL/trees/$b`;
	    } else {
		toRel($f,"trees/$b");
	    }
	}
    }
    toRel("$ENV{HOME}/writing/sgrp_manual/sgrp_manual.pdf","sgrp_manual.pdf");

    # Read fates
    foreach my $i (0,1) {
	my $d = $mDirs[$i];
	my $sp = $loSp[$i];
	distributeReadFates("$d/rft.txt",$sp);
    }

    foreach my $sp (@loSp) {
	report("copying $sp reference genomes");
	foreach my $b (qw(genome.fa genome.sz genome.gff)) {
	    toRel("ref/$sp/$b","$sp/ref/$b");
	}
	report("copying self and other-species alignments");
	foreach my $f (glob("ref/self/*.ali")) {
	    toRel($f,"$sp/ref/self/" . basename($f));
	}
	foreach my $f (glob("misc/multiway/chr??/*.gz")) {
	    my $g = $f;
	    $g =~ s/misc.multiway/multipleSpeciesAlignments/;
	    toRel($f,$g);
	}
	report();
    }

    # Solexa consensi.
    # Bear in mind these are aligned to 025/assembly.orig !!
    my %wantAssemblyOrig;
    foreach my $d (@mDirs) {
	my $fiDir = dirname(glob("$d/???/isFinalIteration"));
	my $sp = (split(/\//,$d))[2];
	report("copying $sp Solexa consensus files");
	foreach my $f (glob("$fiDir/solexa/*.fastq.gz")) {
	    my $str = basename($f,'.fastq.gz');
	    # Cere Solexa data is for strains we already have ...
	    $wantAssemblyOrig{$str} = 1 if ($sp eq 'cere');
	    toRel($f,"$sp/strains/$str/solexa.fastq.gz");
	}
	report();
    }

    # Sequenced and imputed files, etc.
    # Dangles are in assembly dirs ...
    foreach my $d (@mDirs) {
	my $sp = (split(/\//,$d))[2];
	my $fiDir = dirname(glob("$d/0??/isFinalIteration"));
	foreach my $sd (glob("$fiDir/p?/*")) {
	    my $chr = basename($sd);
	    report("copying $sp sequenced and imputed files for $chr");
	    toRel("$sd/sorted_z.gz","$sp/match/$chr/sequenced.gz");
	    toRel("$sd/imputed_z.gz","$sp/match/$chr/imputed.gz");
	    toRel("$sd/dangles_z.txt","$sp/match/$chr/dangles.txt");
	    #toRel("$sd/layout_y.gz","$sp/match/$chr/layout.gz");
	    # coverage*.gz are created by coverageByStrain.pl
	    toRel("$sd/coverageABI.gz","$sp/match/$chr/coverageABI.gz");
	    toRel("$sd/coverageIllumina.gz","$sp/match/$chr/coverageIllumina.gz");
	    toRel("$sd/sbw.txt","$sp/match/$chr/sbw.txt");
	    copySNPReport("$sd/ssnp_report.gz","$REL/$sp/match/$chr/sequencedSNPs.txt");
	    copySNPReport("$sd/isnp_report.gz","$REL/$sp/match/$chr/imputedSNPs.txt");
	    toRel("$sd/translations_z_fwd","$sp/match/$chr/translations_fwd");
	    toRel("$sd/translations_z_rev","$sp/match/$chr/translations_rev");
	    report();
	}
	my $exDir = dirname(glob("$d/???/isExtraIteration"));
	foreach my $sd (glob("$exDir/p?/*")) {
	    my $chr = basename($sd);
	    report("copying $sp layout files for $chr");
	    copyLayoutFile("$sd/layout.gz","$REL/$sp/match/$chr/layout.gz");
	    toRel("$sd/acr.txt","$sp/match/$chr/alignmentClashRates.txt");
	    report();
	}
	my $soDir = dirname(glob("$d/0??_solexa/isFinalIteration"));
	foreach my $sd (glob("$soDir/p?/*")) {
	    my $chr = basename($sd);
	    report("copying $sp recombs files for $chr");
	    copyMgrFiles($sd,"$sp/match/$chr/recombs");
	    report();
	}
    }

    # Assembly dirs
    foreach my $i (0,1) {
	my $d = $mDirs[$i];
	my $sp = $loSp[$i];
	my $exDir = dirname(glob("$d/???/isExtraIteration"));
	report("copying $sp assembly files");
	toRel("$exDir/assembly/genome.fa","$sp/ref/genomeAligned.gff");
	foreach my $as (qw(assembly assembly.orig)) {
	    foreach my $sd (glob("$exDir/$as/*")) {
		if (-d $sd) {
		    my $str = basename($sd);
		    my $tgtAs;
		    my @sufs = qw(ali.gz dangles fa gff map seqr);
		    if ($as eq 'assembly') {
			$tgtAs = $as;
		    } elsif ($wantAssemblyOrig{$str}) {
			$tgtAs = 'assembly_for_solexa';
		    } else {
			next;
		    }
		    foreach my $suf (@sufs) {
			my $toSub = $suf eq 'ali.gz' ? 'refalign.gz' : $suf;
			my $from = "$sd/genome.$suf";
			my $to = "$sp/strains/$str/$tgtAs/genome.$toSub";
			if ($suf eq 'fa') {
			    copyFAChromosomesOnly($from,"$REL/$to");
			} elsif ($suf eq 'map') {
			    copyMapChromosomesOnly($from,"$REL/$to");
			} else {
			    toRel($from,$to);
			}
		    }
		}
	    }
	}
	report();
    }

    foreach my $sd (glob("read/*")) {
	if (-d $sd) {
	    my $str = basename($sd);
	    report("copying reads for $str");
	    my $sp = $pSp->{$str};
	    foreach my $b (qw(name_corrections.txt renamed.fastq renamed.wm)) {
		toRel("$sd/$b","$sp/strains/$str/$b");
	    }
	    report();
	}
    }
}

sub copySNPReport {
    my ($f,$g) = @_;
    unless (-s $f) {
	print "WARNING: missing $f\n";
	return;
    }
    return if (-s $g);
    my $fh = openToRead($f);
    my $gh = openToWrite($g);
    while (<$fh>) {
	last if (/^MatchCounts/);
	print $gh $_;
    }
}

sub copyLayoutFile {
    my ($f,$g) = @_;
    unless (-s $f) {
	print "WARNING: missing $f\n";
	return;
    }
    return if (-s $g);
    my $fh = openToRead($f);
    my $gh = openToWrite($g);
    while (<$fh>) {
	next if (/^ / || /^>REF /);
	s/^[a-z]./= /;
	print $gh $_;
    }
}

sub copyFAChromosomesOnly {
    my ($f,$g) = @_;
    return if (-s $g);
    my $fh = openToRead($f);
    my $gh = openToWrite($g);
    my $active;
    while (<$fh>) {
	if (/^>/) {
	    $active = !(/\.\d\d\d$/ || /^>vc_/);
	}
	print $gh $_ if ($active);
    }
}

sub copyMapChromosomesOnly {
    my ($f,$g) = @_;
    return if (-s $g);
    my $fh = openToRead($f);
    my $gh = openToWrite($g);
    my $active;
    while (<$fh>) {
	print $gh $_ unless (/\.\d\d\d /);
    }
}

sub copyMgrFiles {
    my ($srcDir,$tgtDir) = @_;
    foreach my $b (qw(mgr_in args_bl recombinationFrequencies.txt)) {
	toRel("$srcDir/$b","$tgtDir/$b");
    }
    foreach my $sd (glob("$srcDir/m0??")) {
	my $mm = basename($sd);
	toRel("$sd/mgr_in","$tgtDir/$mm/mgr_in");
	my $out = "$REL/$tgtDir/$mm/mgr_out";
	`cat $sd/arg???.mgr > $out`;
    }
}	

sub constructTarFiles {
    chdir($REL);
    mkdir_p('tgz');
    my $tMisc = "tgz/misc.tgz";
    mayConstructTar($tMisc,"*/ref code trees multipleSpeciesAlignments sgrp_manual.pdf");
    foreach my $i (0,1) {
	my $sp = $loSp[$i];
	mayConstructTar("tgz/${sp}_recombs.tgz","$sp/match/*/recombs");
	mayConstructTar("tgz/${sp}_matches.tgz","--exclude '*recombs*' $sp/match");
	mayConstructTar("tgz/${sp}_assemblies.tgz","$sp/strains/*/assembly");
	mayConstructTar("tgz/${sp}_reads.tgz","--exclude '*assembly*' $sp/strains");
    }
}

sub mayConstructTar {
    my ($tgt,$args) = @_;
    if (-s $tgt) {
	report("$tgt: already exists",1);
    } else {
	report("$tgt: creating");
	`tar zcf $tgt $args`;
	report();
    }
}

sub distributeReadFates {
    my ($f,$sp) = @_;
    unless (-s $f) {
	print "WARNING: missing $f\n";
	return;
    }
    my $fh = openToRead($f);
    my ($str,$gh);
    report("distributing $f");
    while (<$fh>) {
	my $s = substr($_,0,index($_,'-'));
	unless ($s eq $str) {
	    $str = $s;
	    my $g = "$REL/$sp/strains/$str/fates.txt";
	    if (-s $g) {
		#print "Already exists: $g\n";
		report();
		return;
	    }
	    $gh = openToWrite($g);
	}
	print $gh $_;
    }
    report();
}

sub decideSpecies {
    my %sp;
    foreach my $d (glob("read/*")) {
	if (-d $d) {
	    my $str = basename($d);
	    my $c = -d "$d/cere";
	    my $p = -d "$d/para";
	    if ($p) {
		# Even if both para and cere, it should be para, as
		# we may align para strains to cere ref but not vv.
		$sp{$str} = 'para';
	    } elsif ($c) {
		$sp{$str} = 'cere';
	    } else {
		die("Cannot decide species for $d");
	    }
	}
    }
    return \%sp;
}

sub toRel {
    my ($from,$to) = @_;
    $to = "$REL/$to" unless (substr($to,0,1) eq '/');
    if (-s $to || -d $to) {
	#print "WARNING: not overwriting $to\n";
    } elsif (-f $from || -d $from) {
	mkdir_p(dirname($to));
        `cp -pr $from $to`;
    } elsif ($from =~ /REF.*ali/ || 
	     $from =~ /REF.*dangles/ || 
	     $from =~ /translations/) {
	1; # don't worry
    } else {
	print "WARNING: missing: $from\n";
    }
}
