#!/usr/bin/env perl

# Run in (final) NNN dir

use Utility;
use SGRP::OffsetMap;
use File::Basename;
use strict;

# For Solexa:
my $eDir = dirname(glob("../???/isExtraIteration"));
my @strains;
my @solFiles = glob("solexa/*/all.map");
my %oMaps;
foreach my $f (@solFiles) {
    my $str = basename(dirname($f));
    push @strains,$str;
    $oMaps{$str} = new OffsetMap('?',"$eDir/assembly/$str/genome.map");
}
# Will be needed for Solexa:
$oMaps{'REF'} = new OffsetMap('?',"$eDir/assembly/REF/genome.map");

foreach my $d (glob("p?/*")) {
    my $f = "$d/layout_z.gz";
    my $g = "$d/coverageABI.gz";
    if (-s $f && stamp($g)) {
	report("$g: creating");
	my $sh = openToRead("$d/sorted_z.gz");
	my $hdr = <$sh>;
	chomp $hdr;
	$sh->close();
	coverageByStrain($hdr,openToRead($f),openToWrite($g));
	unstamp($g);
	report();
    }
    $g = "$d/coverageIllumina.gz";
    if (stamp($g)) {
	my $chr = basename($d);
	my %scov;
	foreach my $i (0 .. $#solFiles) {
	    my $str = $strains[$i];
	    report("reading for $str.$chr");
	    augmentIlluminaCoverage($solFiles[$i],$str,$chr,$oMaps{$str},$oMaps{'REF'},\%scov);
	    report();
	}
	reportIlluminaCoverage($d,$chr,\%scov,\@strains,$g);
	unstamp($g);
    }
}    

sub coverageByStrain {
    my ($hdr,$fh,$gh) = @_;
    my $off = 0;
    my @posToStr;
    my @strains = split(" ",substr($hdr,1));
    my %strIdx;
    my @stats;
    foreach my $i (0 .. $#strains) {
	$strIdx{$strains[$i]} = $i;
	$stats[$i] = 0;
    }
    my $rep;
    my $start = 1;
    print $gh substr($hdr,1) . "\n";
    while (<$fh>) {
	chomp;
	if (/^>(\S+) (\d+) /) {
	    my ($read,$pos) = ($1,$2);
	    my ($str) = split(/-/,$read);
	    unless ($str eq $posToStr[$pos]) {
		$posToStr[$pos] = $str;
		$rep = 1;
	    }
	} elsif (/^[^>]/) {
	    $off ++;
	    my $pMax = (length($_)-4)/2;
	    foreach my $p (0 .. $#posToStr) {
		if (defined $posToStr[$p]) {
		    if ($p > $pMax || substr($_,2*($p+1),1) eq ' ') {
			# alignment has finished
			$posToStr[$p] = undef;
			$rep = 1;
		    }
		# This does happen occasionally, though it shouldn't ...
		#} elsif ($p <= $pMax && substr($_,2*($p+1),1) ne ' ') {
		#    1;
		}
	    }
	    if ($rep) {
		reportChange($start,$off,\@posToStr,\%strIdx,\@stats,$gh);
		$rep = undef;
		$start = $off;
	    }
	}
    }
    reportChange($start,$off+1,undef,undef,\@stats,$gh);
}
		
sub reportChange {
    my ($start,$off,$pPosToStr,$pStrIdx,$pStats,$gh) = @_;
    if (@$pStats) {
	my $end = $off-1;
	print $gh "$start $end @$pStats\n";
    }
    return unless (defined @$pPosToStr);
    foreach my $i (0 .. $#$pStats) {
	$pStats->[$i] = 0; # ensure all defined
    }
    foreach my $str (@$pPosToStr) {
	if (defined $str) {
	    my $idx = $pStrIdx->{$str};
	    die("No index found for $str") unless (defined $idx);
	    $pStats->[$idx] ++;
	}
    }
    while (@$pPosToStr && !defined $pPosToStr->[-1]) {
	pop @$pPosToStr; # economy ...
    }
}

sub augmentIlluminaCoverage {
    my ($f,$str,$chr,$pOM,$pOMRef,$pCov) = @_;
    my $fh = openFromPipe("maq mapview $f");
    my %delta;
    my $key = "$str.$chr";
    my $c = 0;
    my $isRef;
    while (<$fh>) {
	chomp;
	my @a = split;
	my (undef,$schr,$lo) = @a;
	if ($schr eq $key || ($schr eq $chr && ($isRef = 1))) {
	    my $len = length($a[-1]);
	    $delta{$lo} ++;
	    $delta{$lo+$len}--;
	    $c ++;
	    #last if ($c == 1000);
	} elsif (%delta) {
	    last; # gone far enough
	}
    }
    print "$c reads ... ";
    if ($isRef) {
	$pOM = $pOMRef; # matched against ref genome, apparently ...
	$key = "REF.$chr";
    }
    my $lev = 0;
    my @offs = sort {$a <=> $b} keys %delta;
    my @aOffs;
    foreach my $off (@offs) {
	push @aOffs,$pOM->translation($key,$off);
    }
    foreach my $i (0 .. $#offs) {
	my $lo = $offs[$i];
	my $aLo = $aOffs[$i];
	$lev += $delta{$lo};
	$pCov->{$aLo}{$str} = $lev;
	if ($lev > 0) {
	    my $hi = $offs[$i+1]-1;
	    my $aHi = $aOffs[$i+1]-1;
	    if ($aHi-$aLo > $hi-$lo) {
		# Allow for deletions:
		my $aPrev = $aLo;
		for (my $k=$lo+1; $k<=$hi+1; $k++) {
		    my $ak = $pOM->translation($key,$k);
		    # We don't react to a 10-base or smaller deletion...
		    if ($ak-$aPrev > 11) {
			$pCov->{$aPrev+1}{$str} = 0;
			$pCov->{$ak}{$str} = $lev if ($k <= $hi);
		    }
		    $aPrev = $ak;
		}
	    }
	}
    }
}

sub reportIlluminaCoverage {
    my ($d,$chr,$pCov,$pStrains,$g) = @_;
    my $fh = openToRead("$d/delete_yz.txt");
    chomp (my @delLocs = <$fh>);
    $fh->close();
    my @strains = sort @$pStrains;
    my %strIdx;
    foreach my $i (0 .. $#strains) {
	$strIdx{$strains[$i]} = $i;
    }
    report("writing $g");
    my $gh = openToWrite($g);
    print $gh "@strains\n";
    my @lev;
    foreach my $i (0 .. scalar(keys %strIdx)-1) {
	$lev[$i] = 0;
    }
    my @offsY = sort {$a <=> $b} keys %$pCov;
    my @offsZ;
    if (@delLocs) {
	my $delta = 0;
	foreach my $i (0 .. $#offsY) {
	    my $off = $offsY[$i];
	    while (@delLocs && $off >= $delLocs[0]) {
		$delta ++;
		shift @delLocs;
	    }
	    $offsZ[$i] = $off-$delta;
	}
    } else {
	@offsZ = @offsY;
    }
    if ($offsZ[0] > 1) {
	printf($gh "%d %d %s\n",1,$offsZ[0]-1,join(" ",@lev));
    }
    foreach my $i (0 .. $#offsY-1) {
	my $pH = $pCov->{$offsY[$i]};
	foreach my $str (keys %$pH) {
	    $lev[$strIdx{$str}] = $pH->{$str};
	}
	if ($offsZ[$i+1]>$offsZ[$i]) {
	    printf($gh "%d %d %s\n",$offsZ[$i],$offsZ[$i+1]-1,join(" ",@lev));
	}
    }
    report();
}

      
