#!/usr/bin/env perl

# Args: central genome.fa then trios of
#  speciesName *.out *.acc (or acc2); all for a single central chr.
#
# Example (from $SGR/ref):
#
# createMultiwayAlignment.pl cere/split/chr01.fa \
#   para para_new/chr01.out para_new/chr01.acc \
#   baya baya/to_cere/chr01.out baya/to_cere/all.acc

use Utility;
use File::Basename;
use strict;

# Custom calling sequence:

createMultiwayAlignment();

sub createMultiwayAlignment {
    my $gffFile = "cere/genome.gff";
    report("$gffFile: reading");
    my $pGFF = readGFF($gffFile);
    report();
    my $paraDir = "para_new";
    my $bayaDir = "baya/to_cere";
    foreach my $f (glob("cere/split/*.fa")) {
	my $chr = basename($f,".fa");
	my $pGFFPoints = gffPoints($pGFF->{$chr});
	my $out = "multi/$chr.gz";
	if (stamp($out)) {
	    my @rest = ('para',
			"$paraDir/$chr.out",
			"$paraDir/$chr.acc",
			'baya',
			"$bayaDir/$chr.out",
			"$bayaDir/all.acc");
	    my $ok = 1;
	    foreach my $f (@rest) {
		if (index($f,'/') > 0) {
		    unless (-s $f) {
			print "Missing $f\n";
			$ok = 0;
		    }
		}
	    }
	    if ($ok) {
		report("$out: creating");
		createMultiwayAlignment1($out,$f,$pGFFPoints->{$chr},@rest) if ($ok);
		report();
		my $outGFF = "$paraDir/$chr.gff";
		report("$outGFF: creating");
		printLiftedGFF($pGFF->{$chr},$pGFFPoints->{$chr},$outGFF);
		report();
	    } else {
		report("$out: cannot create",1);
	    }
	    unstamp($out);
	    report();
	} else {
	    report("$out: already exists",1);
	}
    }
}

sub readGFF {
    my ($f) = @_;
    my %lines;
    my $fh = openToRead($f);
    while (<$fh>) {
	chomp;
	my @a = split(/\t/,$_);
	push @{$lines{$a[0]}},\@a;
    }
    return \%lines;
}

sub gffPoints {
    my ($pLines) = @_;
    my %points;
    foreach my $pLine (@$pLines) {
	my ($chr,$lo,$hi) = @{$pLine}[0,3,4];
	$points{$chr}{$lo} = 1;
	$points{$chr}{$hi} = 1;
    }
    return \%points;
}

sub createMultiwayAlignment1 {
    my ($out,$fa0,$pGFFPoints,@trios) = @_;

    my (@spList,@bfList,@afList); # species, blast files, acceptance files
    
    for (my $i=0; $i<$#trios; $i+=3) {
	push @spList,$trios[$i];
	push @bfList,$trios[$i+1];
	push @afList,$trios[$i+2];
    }
    
    my $gh = openToWrite($out);

    print $gh ">@spList\n";
    
    my $pSeq = readFastaFile($fa0);
    
    my ($chr) = keys %$pSeq;
    my $seq0 = $pSeq->{$chr};
    
    my @mat;
    $mat[0] = [undef,split("",$seq0)];
    
    for (my $i=0; $i<=$#spList; $i++) {
	# We only pass GFFPoints through for the FIRST species, para (since bayanus
	# isn't going to get annotations).
	$mat[$i+1] = fillAlignmentColumn($bfList[$i],$chr,$mat[0],$afList[$i],$i==0 ? $pGFFPoints : undef);
    }
    foreach my $k (1 .. $#{$mat[0]}) {
	#if ($k%100 == 0) {
	#    print $gh ">REF $k\n";
	#}
	my @seqs;
	my $maxL = 1;
	foreach my $i (0 .. $#mat) {
	    $seqs[$i] = $mat[$i][$k];
	    $maxL = max($maxL,length($seqs[$i]));
	}
	foreach my $l (0 .. $maxL-1) {
	    my $line;
	    foreach my $i (0 .. $#seqs) {
		if ($seqs[$i]) {
		    $line .= substr($seqs[$i],$l,1) || '-';
		} else {
		    $line .= " ";
		}
		$line .= " "; # quality column, for consistency with normal file format
	    }
	    print $gh "$line\n" if ($line =~ /[A-Z]/); # occasionally just gaps...
	}
    }
}

sub fillAlignmentColumn {
    my ($bFile,$chr,$refSeq,$aFile,$pGFFPoints) = @_;
    my %wanted;
    my $ah = openToRead($aFile);
    while (<$ah>) {
	next if (/^\#/);
	chomp;
	my @a = split;
	next unless (@a == 15 && $a[4] eq $chr);
	my ($qchr,$fc,$idx,$qmin,$qmax,$smin,$smax) = @a[4,7,9,11..14];
	$wanted{$idx} = [$smin,$smax,$fc,$qchr,$qmin,$qmax];
    }
    my @col; # result
    my $bh = openToRead($bFile);
    my ($qchr,$idx,$pArr,$qrow);
    while (<$bh>) {
	chomp;
	if (/^ Score = /) {
	    $idx ++;
	    $pArr = $wanted{$idx}; # may or may not be defined
	} elsif (!$pArr) {
	    # nothing
	} elsif (/^Query:/) {
	    $qrow = $_;
	} elsif (/^Sbjct:/) {
	    my $srow = $_;
	    my (undef,$q1,$qseq,$q2) = split(" ",$qrow);
	    my (undef,$s1,$sseq,$s2) = split(" ",$srow);
	    my ($smin,$smax,$fc,$qchr,$qmin,$qmax) = @$pArr;
	    next if ($s1 > $smax || $s2 < $smin);
	    if ($fc eq 'F') {
		next if ($q1 > $qmax || $q2 < $qmin);
	    } else {
		next if ($q2 > $qmax || $q1 < $qmin);
	    }
	    # At least some of the row is wanted.
	    my $dq = $fc eq 'F' ? 1 : 0;
	    my $qoff = $q1;
	    my $soff = $s1;
	    my $iLim = length($sseq);
	    for (my $i=0; $i<$iLim; $i++) {
		my $qv = substr($qseq,$i,1);
		my $sv = substr($sseq,$i,1);
		if ($qoff >= $qmin && $qoff <= $qmax &&
		    $soff >= $smin && $soff <= $smax) {
		    $col[$soff] .= $qv;
		}
		if (defined $pGFFPoints && defined $pGFFPoints->{$soff}) {
		    $pGFFPoints->{$soff} = [$qoff,$fc];
		}
		$qoff += $dq unless ($qv eq '-');
		$soff ++ unless ($sv eq '-');
	    }
	}
    }
    return \@col;
}

sub printLiftedGFF {
    my ($pLines,$pPoints,$file) = @_;
    my $gh = openToWrite($file);
    my $eh = openToWrite("$file.comments");
    foreach my $pLine (@$pLines) {
	my ($tLo,$fcLo) = liftoverPoint($pLine->[3],$pPoints);
	my ($tHi,$fcHi) = liftoverPoint($pLine->[4],$pPoints);
	next unless (defined $tLo && defined $tHi);
	if ($fcLo eq $fcHi) {
	    my $fromLen = $pLine->[4]-$pLine->[3]+1;
	    if ($fcLo eq 'C') {
		my $tmp = $tHi;
		$tHi = $tLo;
		$tLo = $tmp;
	    }
	    my $toLen = $tHi-$tLo+1;
	    # Extreme difference in length -> don't believe it
	    unless ($toLen >= $fromLen/2 && $toLen < $fromLen*2) {
		print $eh "# Extreme change in length, from $fromLen to $toLen, for:\n";
		print $eh join("\t",@$pLine) . "\n";
		next;
	    }
	    $pLine->[6] =~ tr/-+/+-/ if ($fcLo eq 'C');
	    $pLine->[3] = $tLo;
	    $pLine->[4] = $tHi;
	    print $gh join("\t",@$pLine) . "\n";
	} elsif ($fcLo eq 'C') {
	    print $eh "# Only low end is inverted, for:\n";
	    print $eh join("\t",@$pLine) . "\n";
	} else {
	    print $eh "# Only high end is inverted, for:\n";
	    print $eh join("\t",@$pLine) . "\n";
	}
    }
}

sub liftoverPoint {
    my ($i,$pPoints) = @_;
    if (defined $pPoints->{$i} && $pPoints->{$i} != 1) {
	return @{$pPoints->{$i}};
    } else {
	return undef;
    }
}
