#!/usr/bin/env perl

# See th/0522

use Utility;
use File::Basename;
use Getopt::Std;
use strict;

autoflush STDOUT 1;

my %OPT = qw(t sequenced);
getopts('t:',\%OPT);

createMultiwayAlignmentFull();

sub createMultiwayAlignmentFull {

    # Directory containing alignment and accepted-alignment files
    my $nWayDir = "$ENV{SGR}/misc/multiway";
    # SGRP release directory
    my $relDir = "$ENV{SGRW}/release3";
    # Sc reference genome
    my $cereRefFile = "$relDir/cere/ref/genome.fa";
    # Non-cere species; para must be first
    my @species    = qw(para  mika  kudr  baya );
    my %speciesIndex = indexArray(@species);
    # Assemblies and aligners to use for each species
    my @assemblies = qw(si1   mit   washu washu);
    my @aligners   = qw(blast ssaha ssaha ssaha);
    # Strain info: sequenced or imputed
    my $type = $OPT{t};
    # Dirs containing $type.gz files, for cere and para
    my $cereMatchDir = "$ENV{SGRW}/release3/cere/match";
    my $paraMatchDir = "$ENV{SGRW}/release3/para/match";

    my %segs = readAndSortSegments($nWayDir,\@species,\@assemblies,\@aligners);

    my $pCere = readFastaFile($cereRefFile);
    my $pFastaSeqs = readFastaSeqs($nWayDir,\@species,\@assemblies);
    
    # NOTE we only do it for nuclear chromosomes at present...
    foreach my $cereFile (glob("$cereMatchDir/chr??/$type.gz")) {
	my $chr = basename(dirname($cereFile));
	my $tgt = "$chr/$type.gz";
	if (stamp($tgt)) {
	    my $paraFile = "$paraMatchDir/$chr/$type.gz";
	    my ($paraStrains,$paraAli) = readAlignmentChunked($paraFile);
	    report("$tgt: creating");
	    createMultiwayAlignmentFull1(\@species,\%speciesIndex,$pFastaSeqs,
					 $cereFile,$paraStrains,$paraAli,$segs{$chr},$tgt);
	    unstamp($tgt);
	    report();
	    my $rtgt = "$chr/refalign.gz";
	    if (stamp($rtgt)) {
		#report("sleeping to ensure gzip has finished");
		#sleep(10); # to ensure gzip on $tgt has finished
		#report();
		report("$rtgt: creating");
		createReferenceAlignment($tgt,$rtgt);
		unstamp($rtgt);
		report();
	    }
	}
    }
}

sub indexArray {
    my @arr = @_;
    my %h;
    foreach my $i (0 .. $#arr) {
	$h{$arr[$i]} = $i;
    }
    return %h;
}

sub readAndSortSegments {
    my ($dir,$pSpecs,$pAssems,$pAligs) = @_;
    report("reading and sorting segments");
    my %segs;
    foreach my $i (0 .. $#$pSpecs) {
	my $accFile = "$dir/$pSpecs->[$i].$pAssems->[$i].$pAligs->[$i].accseg";
	readSegments($pSpecs->[$i],$accFile,\%segs);
    }
    foreach my $chr (keys %segs) {
	$segs{$chr} = [sort {$a->[0] <=> $b->[0]} @{$segs{$chr}}];
    }
    report();
    return %segs;
}

sub readSegments {
    my ($sp,$f,$pSegs) = @_;
    my $fh = openToRead($f);
    while (<$fh>) {
	chomp;
	my (undef,$qchr,$qoff,$qfc,$qlen,$schr,$soff,$sfc,$slen) = split;
	die("Subject direction not F in $_") unless ($sfc eq 'F');
	push @{$pSegs->{$schr}},[$soff,$sp,$qchr,$qoff,$qfc,$qlen,$slen];
    }
}

sub readFastaSeqs {
    my ($nWayDir,$pSpecies,$pAssemblies) = @_;
    report("reading fasta sequences");
    my @fastaSeqs;
    foreach my $i (1 .. $#$pSpecies) { # not para here
	$fastaSeqs[$i] = readFastaFile("$nWayDir/$pSpecies->[$i].$pAssemblies->[$i].fasta");
    }
    report();
    return \@fastaSeqs;
}

sub readAlignmentChunked {
    my ($f) = @_;
    report("reading $f");
    my $fh = openToRead($f);
    my $hdr = <$fh>;
    chomp $hdr;
    $hdr =~ s/^>//;
    my @chunks = ([]);
    while (<$fh>) {
	chomp;
	if (/^>/) {
	    # nothing
	} elsif (/^[^A-Z]/) {
	    push @{$chunks[-1]},$_;
	} else {
	    push @chunks,[$_];
	}
    }
    report();
    return ($hdr,\@chunks);
}

# Cols will be:
#   cere ref, cere strains
#   para ref, para strains
#   other species in @species

sub createMultiwayAlignmentFull1 {
    my ($pSpecies,$pSpeciesIndex,$pFastaSeqs,
	$cereFile,$paraStrains,$paraAli,$pSegs,$tgt) = @_;

    # Header lines
    my $fh = openToRead($cereFile);
    my $gh = openToWrite($tgt);
    chomp (my $cereHdr = <$fh>);
    print $gh "$cereHdr para $paraStrains @$pSpecies[1..$#$pSpecies]\n";

    # Width for cere ref and strains
    my $csw = pairWidth($cereHdr);
    # Width for para ref and strains
    my $psw = pairWidth($paraStrains);
    # Total width of line
    my $wid = $csw + $psw + 2*(scalar(@$pSpecies)-1);
    # Start columns: species 0 (para) starts at $csw
    my @col = $csw;
    # Subsequent species start at csw+psw, csw+psw+2, ...
    foreach my $k (1 .. $#$pSpecies) {
	$col[$k] = $csw+$psw+2*($k-1);
    }

    # As we go through the cere sequenced.gz or imputed.gz file ...
    my $off = 0; # offset (from 1) in cere ref
    my $i   = 0; # index into $pSegs->{$chr}
    my $prev; # previous line
    # $pSegs item is: [$soff,$sp,$qchr,$qoff,$qfc,$qlen,$slen];
    while (<$fh>) {
	chomp;
	next if (/^>/);
        # Pad $_ to correct width
	$_ .= ' ' x ($wid-length($_)); 
	# Cere ref gap; just ensure gaps in correct places, print and continue
	unless (/^[A-Z]/) {
	    for (my $i=$csw; $i<=$wid; $i+=2) {
		if (substr($prev,$i,1) ne ' ') {
		    substr($_,$i,1) = '-';
		}
	    }
	    print $gh "$_\n";
	    $prev = $_;
	    next;
	}
	# Increment offset in cere ref
	$off ++;
	# Skip forward in $pSegs to first relevant item
	while ($i < $#$pSegs && $pSegs->[$i][0]+$pSegs->[$i][6] < $off) {
	    $i++;
	}
	my (@extraLinesPre,@extraLinesPost);
	# $j is $pSegs index currently being considered. Go from $i
	# until we've gone past $off.
	for (my $j=$i;
	     $j <= $#$pSegs && $pSegs->[$j][0] <= $off;
	     $j++) {
	    $_ = applySegment($pSegs->[$j],$_,
			      $wid,$psw,\@col,$off,$paraAli,$pFastaSeqs,$pSpeciesIndex,
			      \@extraLinesPre,\@extraLinesPost);
	}
	foreach my $line (reverse(@extraLinesPre)) {
	    for (my $i=0; $i<$wid; $i+=2) {
		substr($line,$i,1) =~ tr/ACGTacgt/TGCAtgca/;
	    }
	    print $gh "$line\n";
	}
	print $gh "$_\n";
	foreach my $line (@extraLinesPost) {
	    print $gh "$line\n";
	}
	$prev = $_;
    }
    $gh->close();
}

sub applySegment {
    my ($pSeg,$outLine,$wid,$psw,$pCol,$off,$paraAli,$pFastaSeqs,$pSpeciesIndex,
	$pExtraLinesPre,$pExtraLinesPost) = @_;
    my ($soff,$sp,$qchr,$qoff,$qfc,$qlen,$slen) = @$pSeg;
    my @extraLines;
    if ($slen == 0) {
	if ($soff == $off) {
	    # push the inserted stuff to @extraLines with gaps elsewhere.
	    # When $qfc eq 'C', we want or $qoff-$qlen+1 .. $qoff
	    if ($sp eq 'para') {
		my @range;
		if ($qfc eq 'F') {
		    @range = ($qoff .. $qoff+$qlen-1);
		} else {
		    @range = ($qoff-$qlen+1 .. $qoff);
		}
		foreach my $qo (@range) {
		    # $paraAli->[$qo] because zeroth element of paraAli is
		    # undef plus any initial insertions.
		    foreach my $paraLine (@{$paraAli->[$qo]}) {
			my $line = '- ' x ($wid/2);
			substr($line,$pCol->[0],$psw) = $paraLine;
			push @extraLines,$line;
		    }
		}
	    } else {
		my $idx = $pSpeciesIndex->{$sp};
		foreach my $qo ($qoff .. $qoff+$qlen-1) {
		    my $line = '- ' x ($wid/2);
		    substr($line,$pCol->[$idx],1) = substr($pFastaSeqs->[$idx]{$qchr},$qo-1,1);
		    push @extraLines,$line;
		}
	    }
	}
    } elsif ($soff+$slen <= $off) {
	# This segment may fall short (but subsequent ones may not)
	1;
    } elsif ($qlen == 0) {
	# deletion in query sequence
	if ($sp eq 'para') {
	    substr($outLine,$pCol->[0],$psw) = ('- ') x ($psw/2);
	} else {
	    substr($outLine,$pCol->[$pSpeciesIndex->{$sp}],2) = '- ';
	}
    } else {
	# $qoff is offset (from 1) in query sequence
	$qoff += ($qfc eq 'F' ? 1 : -1) * ($off-$soff);
	if ($sp eq 'para') {
	    my ($paraLine,@extra) = @{$paraAli->[$qoff]};
	    foreach my $ex (@extra) {
		my $line = '- ' x ($wid/2);
		substr($line,$pCol->[0],$psw) = $ex;
		push @extraLines,$line;
	    }
	    substr($outLine,$pCol->[0],$psw) = mayRevcompPairs($qfc,$paraLine);
	} else {
	    my $idx = $pSpeciesIndex->{$sp};
	    # $qoff-1 because $qoff is 1-indexed
	    my $v = substr($pFastaSeqs->[$idx]{$qchr},$qoff-1,1);
	    substr($outLine,$pCol->[$idx],1) = mayRevcompPairs($qfc,$v);
	}
    }
    if (@extraLines) {
	# Enforce truncation and spaces in extraLines to match outLine
	foreach my $i (0 .. $#extraLines) {
	    $extraLines[$i] = enforceSpaces($outLine,$extraLines[$i]);
	}
	if ($qfc eq 'F') {
	    push @$pExtraLinesPost,@extraLines;
	} else {
	    push @$pExtraLinesPre,@extraLines;
	}
    }
    return $outLine;
}

sub enforceSpaces {
    my ($outLine,$extraLine) = @_;
    my $leno = length($outLine);
    my $lene = length($extraLine);
    if ($lene < $leno) {
	$extraLine .= ' ' x ($leno-$lene);
    }
    for (my $i=0; $i<$lene; $i+=2) {
	if ($i >= $leno || (substr($outLine,$i,2) eq '  ' && substr($extraLine,$i,1) eq '-')) {
	    substr($extraLine,$i,2) = '  ';
	}
    }
    return $extraLine;
}

sub pairWidth {
    my ($str) = @_;
    my @a = split(" ",$str);
    return 2*(scalar(@a)+1);
}

sub mayRevcompPairs {
    my ($fc,$line) = @_;
    if ($fc eq 'C') {
	for (my $i=0; $i<length($fc); $i+=2) {
	    substr($line,$i,1) =~ tr/ACGTacgt/TGCAtgca/;
	}
    }
    return $line;
}

sub createReferenceAlignment {
    my ($src,$tgt) = @_;
    my $fh = openToRead($src);
    my $gh = openToWrite($tgt);
    chomp (my $hdr = <$fh>);
    my @names = split(" ",substr($hdr,1));
    my @want = (0);
    my $tgtHdr = ">";
    foreach my $i (0 .. $#names) {
	if ($names[$i] =~ /^[a-z]+$/) {
	    push @want,2*($i+1);
	    $tgtHdr .= "$names[$i] ";
	}
    }
    $tgtHdr =~ s/ $//;
    print $gh "$tgtHdr\n";
    while (<$fh>) {
	my $line;
	foreach my $col (@want) {
	    $line .= (substr($_,$col,1) || " ") . " ";
	}
	$line =~ s/[0-9]/-/g;
	print $gh "$line\n";
    }
}
	
