#!/usr/bin/env perl

# Args: tgtDir strain1,strain2,...,strainN span_MMM_NNN span_MMM_NNN 

# They're specified in this rather counterintuitive way to enable lots
# to be done in one run. See example below...

# The "span" files (space separated) should exist, and the strain names
# (comma separated) should be a subset of this in the (e.g.) span_MMM_NNN.strains
# files. tgtDir is appended to the front of each "span" filename to create
# the corresponding target file; this those filenames should be relative not
# absolute. For example:

# createSubtrees.pl foo BC187,SK1,Y55 p1/chr01/smat/span_???_???.matrix

# This should create:

# foo/p1/chr01/smat/span_???_???.{matrix,tree,q40,strains}
#  and then you can do e.g.
# unrooted  foo/p1/chr01/smat/span_005_007.tree

# NOTE: the scale of the output trees (span_MMM_NNN.tree files) is
# 1000 times what it should be, i.e.  where the scale says "1" it
# ought to say "0.001". This is tricky to fix just now. Will do it if
# you need it.

use Utility; # DMC's code; make sure it's installed in your PERL5LIB path
use strict;

# You will need to download and install your own copy of "quicktree" (to make
# createSubtrees.pl work), and "njplot" and "unrooted" (to view the .tree it creates).
#    http://www.sanger.ac.uk/Software/analysis/quicktree/
#    http://pbil.univ-lyon1.fr/software/njplot.html
#    http://pbil.univ-lyon1.fr/software/unrooted.html

# Change this as required:
my $QUICKTREE = "$ENV{HOME}/software/quicktree_1.1/bin/quicktree";

# Coordinates of block, cumul and span files in p?/chr*/smat directories:
#
# With block size B (25k by default), vertex V is at position V*B.
# V=0,1,2,...VMax, where VMax = round(aliSz/B); the final vertex
# is at aliSz-1 (i.e. the last position) not V*B. That is, the
# final block is of length between B/2 and 3B/2, so that the blocks
# cover the whole chromosome and the final one is a sensible length.
#
# blockVVV is portion of sorted.gz between (V-1)B and B (V=1,2,...).
#
# countsVVV has counts between vertices V-1 and V (V=1,2,...).
#
# cumulVVV has counts between vertices 0 and V (V=0,2,...),
# so that cumul000 is empty (all zeros), and for V>0 and U=V-1,
# cumulVVV = cumulUUU + countsVVV.
#
# span_UUU_VVV.matrix is matrix between vertices U and V,
# calculated from cumulVVV - cumulUUU.
#
# Markers (to delimit where trees apply) are indexed according to
# window size W = 2BR where R must be an integer. Marker
# M is at position M*W/2 + W/4. Between markers M and M+1 we are
# closest to the middle of the window from M*W/2 to (M+2)*W/2 =
# MBR to (M+2)BR, i.e. we use span_MR_(M+2)R. The final window
# span is from nR to VMax, where n+2 = round(VMax/R). Max marker
# index mMax is therefore n+1, i.e. mMax = round(VMax/R)-1.
#
# If B = 25k, for W = 50k we have R=1, so:
#
# 0         25k        50k        5k        100k       125k
# V=0       V=1        V=2        V=3        V=4        V=5
# |----------|----------|----------|----------|----------|
#       |          |          |          |          |
#      M=0        M=1        M=2        M=3        M=4
#  <---- span_0_2 -----> <---- span_2_4 ----->
#             <---- span_1_3 -----> <---- span_3_5 ----->
#
# For W=100k we have R=2, so:
#
# 0         50k        00k      150k       200k       250k
# V=0       V=2        V=4        V=6       V=8       V=10
# |----------|----------|----------|----------|----------|
#       |          |          |          |          |
#      M=0        M=1        M=2        M=3        M=4
#  <---- span_0_20-----> <--- span_20_40 ---->
#             <--- span_10_30 ----> <--- span_30_50 ---->

# Span files created (and then perhaps mostly deleted) are, e.g.:
#
# outputSNPMatrices:
#   span_000_004.strains  strain names, ordered by relatedness
#   span_000_004.q40      rate matrix, organized by strain names
#   span_000_004.matrix   = paste q40 strains
# createQuicktreeInput:
#   span_000_004.alpha    strain count, then q40 rate matrix
#                           multiplied up by 1000, and with
#                           each row STARTING with strain name.
# quicktree program (on alpha file)
#   span_000_004.tree

unless (-x $QUICKTREE) {
    die("$QUICKTREE is missing or not executable");
}

if (@ARGV > 2) {
    createSubmatrix(@ARGV);
} else {
    die("Usage: $0 tgtDir strain1,strain2,strainN span_MMM_NNN span_MMM_NNN ...");
}

sub createSubmatrix {
    my ($tgtDir,$strainList,@srcFiles) = @_;
    my @wanted = split(/,/,$strainList);

    foreach my $src (@srcFiles) {
	my %want = map {$_ => 1} @wanted;
	my $srcB = $src;
	$srcB =~ s/\.[^\.]*$//;
	my $tgtB = "$tgtDir/$srcB";
	my $fh;
	my $q40 = "$srcB.q40";
	my $strains = "$srcB.strains";
	my $matrix = "$srcB.matrix";
	if (-s $matrix) {
	    $fh = openToRead($matrix);
	} elsif (-s $q40 && -s $strains) {
	    $fh = openFromPipe("paste $q40 $strains");
	} else {
	    print("Need either .matrix or .q40 and .strains for $srcB\n");
	    next;
	}
	my (@mat,@strains);
	while (<$fh>) {
	    chomp;
	    my @a = split;
	    push @strains,pop @a;
	    push @mat,\@a;
	}
	my @keep;
	foreach my $s (@strains) {
	    push @keep,$want{$s};
	    delete $want{$s};
	}
	foreach my $s (keys %want) {
	    print STDERR "WARNING: cannot find strain $s\n";
	}
	my $sh = openToWrite("$tgtB.strains");
	my $mh = openToWrite("$tgtB.q40");
	foreach my $i (0 .. $#mat) {
	    if ($keep[$i]) {
		print $sh "$strains[$i]\n";
		my $pRow = $mat[$i];
		foreach my $j (0 .. $#$pRow) {
		    if ($keep[$j]) {
			print $mh " $pRow->[$j]";
		    }
		}
		print $mh "\n";
	    }
	}
	$mh->close();
	$sh->close();
	createQuicktreeInput($tgtB);
	`$QUICKTREE -in m -out t $tgtB.alpha > $tgtB.tree`;
    }
}

# Pinched from SGRP/SNPMatrix.pm ...

sub createQuicktreeInput {
    my ($pfx) = @_;
    my $fh = openToRead("$pfx.strains");
    my @strains = <$fh>;
    my $out = "$pfx.alpha";
    my $gh = openToWrite($out);
    print $gh scalar(@strains) . "\n";
    $fh = openToRead("$pfx.q40");
    while (<$fh>) {
	chomp;
	my @row = split;
	my $str = shift @strains;
	chomp $str;
	$str =~ s/UWOPS/UW/;
	print $gh $str;
	foreach my $v (@row) {
	    printf($gh " %8.6f",$v*1000);
	}
	print $gh "\n";
    }
    $gh->close();
    return $out;
}
