#!/usr/bin/env perl

=head1 NAME

divergenceMatrixAndNJTree.pl

=head1 SYNOPSIS

divergenceMatrixAndNJTree.pl [-m MinComp] outPrefix file1 file2 ...

=head1 DESCRIPTION

outPrefix is a string of your choice; files outPrefix.XYZ will be written
for various XYZ as described below.

Switch -m MinComp: see below.

file1, file2 etc should be output(s) of alicat.pl using only (zero or
more of) the following switches: -a -c -l -h -i -o -r -x -z.

This script reads file1, file2 etc and, for each pair of strains, counts
the number of occasions where they are the same or different at positions
where both have a nucleotide (not a gap) and both are quality 40 or more.
It then creates the following files:

outPrefix.counts: for each strain pair, number of differences and total
number of compared positions (and two zeros, which you can ignore).

outPrefix.alpha: distance matrix. The first field on each line is a strain
name, and the rest are differences per thousand between that strain and
the rest. Only strains that have at least minComp (-m switch) positions
compared with some other strain are represented. The default value of
minComp is 500, below which distance estimates are unlikely to be
meaningful.

outPrefix.tree: neighbour-joining tree constructed from outPrefix.alpha,
using the "quicktree" program pointed to by the $QUICKTREE variable
below. You can download quicktree from 

http://www.sanger.ac.uk/Software/analysis/quicktree/

outPrefix.tree is suitable for viewing with "njplot" and "unrooted",
available from

http://pbil.univ-lyon1.fr/software/

=head1 AUTHOR

David Carter, C<dmc@sanger.ac.uk>. Suggestions for new functionality welcome.

=head1 TERMS AND CONDITIONS

Copyright (c) 2008 Wellcome Trust Sanger Institute (WTSI)

This program is free software; you can redistribute it and/or modify
it under the same terms as Perl itself. See the Artistic License file
in the main Perl distribution for specific terms and conditions of
use.  In addition, the following disclaimers apply:

WTSI makes no representations whatsoever as to the SOFTWARE contained
herein.  It is experimental in nature and is provided WITHOUT WARRANTY
OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE OR ANY OTHER
WARRANTY, EXPRESS OR IMPLIED.  CSHL MAKES NO REPRESENTATION OR
WARRANTY THAT THE USE OF THIS SOFTWARE WILL NOT INFRINGE ANY PATENT OR
OTHER PROPRIETARY RIGHT.

By downloading this SOFTWARE, your Institution hereby indemnifies WTSI
against any loss, claim, damage or liability, of whatsoever kind or
nature, which may arise from your Institution's respective use,
handling or storage of the SOFTWARE.

If publications result from research using this SOFTWARE, we ask that
WTSI be acknowledged and/or credit be given to WTSI scientists, as
scientifically appropriate.

=head1 SEE ALSO

alicat.pl

=cut

use FileHandle;
use Getopt::Std;
use strict;

my $QUICKTREE = "$ENV{HOME}/software/quicktree_1.1/bin/quicktree";

my %OPT;
getopts('m:',\%OPT);

divergenceMatrixAndNJTree(@ARGV);

# $f should be the output of alicat.pl (with no -s, -p or -t).
sub divergenceMatrixAndNJTree {
    my ($outPfx,@files) = @_;
    unless (@files) {
	die("Usage: $0 outPrefix file1 file2 file3 ...");
    }
    my (%num,%den);
    setQualThreshold(); # deliberately no args -> q = 40
    foreach my $f (@files) {
	my $c = "$f.counts";
	mayCreateSmatCountsFile($f,$c,undef,1);
	addToSNPMatrix1($c,\%num,\%den);
    }
    my $pStrains = allKeys(\%den,defined $OPT{m} ? $OPT{m} : 500);
    outputSNPMatrices($outPfx,0,\%num,\%den,$pStrains);
    createNJTree($outPfx); 
    foreach my $suf (qw(strains q40)) {
	unlink("$outPfx.$suf");
    }
}

sub setQualThreshold {
    my ($filtMeth,$strainSet) = @_;
    if ($filtMeth eq 'noerrs') {
	return ('~',0); # actually anything should work for the char
    } else {
	my $qth = ($strainSet =~ /^FREG.*S$/) ? 30 : 40;
	my $pe = exp(-0.1*$qth*log(10));
	return (chr($qth+ord('!')), 2*$pe*(1-$pe));
    }
}

# Lines in ...smat file are of the form
#   strain1 strain2 numAll denAll numSyn denSyn
# We create these files separately to save time.
# The counts are for pairs of quality $QUAL_THRESHOLD or above.

sub mayCreateSmatCountsFile {
    my ($f,$g,$pSynList,$isAlicat) = @_;
    unless (-f $g) {
	my ($pStrains,$pNum,$pDen) = countReliableVariations($f,$pSynList,$isAlicat);
	writeSmatCountsFile($pStrains,$pNum,$pDen,$g);
    }
}

sub countReliableVariations {
    my ($f,$pSynList,$isAlicat) = @_;
    my $fh = openToRead($f);
    $_ = <$fh>;
    chomp;
    s/^>//;
    my @strains = split;
    unless ($isAlicat) {
	@strains = ('REF',@strains);
    }
    my $nLines = 0;
    my $off = 1; # nucleotide offset, from 1
    my $ao = 0; # alignment offset, from 0
    my (@num,@den);
    while (<$fh>) {
	$nLines ++;
	chomp;
	if ($isAlicat) {
	    # Change back to nuc-and-qual format
	    unless ((substr($_,7,1) eq '.' || substr($_,16,1) eq '|') && substr($_,-1,1) eq '|') {
		die("Bad line (not vanilla alicat.pl format) in $f:\n$_\n");
	    }
	    my $idx = index($_,'|');
	    $_ = substr($_,$idx+1);
	    s/\|$//;
	}
	my ($pKList,$pVals) = valuesForGoodQualities($_);
	my $isSyn;
	if (/^[ACGT]/) {
	    if (defined @$pSynList && $pSynList->[0] == $off) {
		$isSyn = 1;
		shift @$pSynList;
	    }
	    $off ++;
	}
	for (my $i=0; $i<$#$pKList; $i++) {
	    my $ki = $pKList->[$i];
	    for (my $j=$i+1; $j<=$#$pKList; $j++) {
		my $kj = $pKList->[$j];
		my $isSNP = ($pVals->[$i] ne $pVals->[$j]);
		$num[$ki][$kj][0] ++ if ($isSNP);
		$den[$ki][$kj][0] ++;
		if ($isSyn) {
		    $num[$ki][$kj][1] ++ if ($isSNP);
		    $den[$ki][$kj][1] ++;
		}
	    }
	}
    }
    return (\@strains,\@num,\@den);
}

sub valuesForGoodQualities {
    my ($line) = @_;
    my $kMax = (length($line)-2)/2;
    my (@kList,@vals,@quals);
    my $qtChar = 'I';
    for (my $k=0; $k<=$kMax; $k++) {
	my $kk = $k+$k;
	my $v = substr($line,$kk,1);
	$v =~ tr/acgt/ACGT/ if ($k>0);
	if ($v =~ /^[ACGT]$/) {
	    my $cq = substr($line,$kk+1,1);
	    if ($cq eq ' ' || $cq ge $qtChar) {
		push @vals,$v;
		push @kList,$k;
	    }
	}
    }
    return (\@kList,\@vals);
}

sub writeSmatCountsFile {
    my ($pStrains,$pNum,$pDen,$g) = @_;
    my $gh = openToWrite($g);
    foreach my $ki (0 .. $#$pDen) {
	my $pNum1 = $pNum->[$ki];
	my $pDen1 = $pDen->[$ki];
	my $si = $pStrains->[$ki];
	foreach my $kj ($ki+1 .. $#$pDen1) {
	    my $pNA;
	    $pNA = $pNum1->[$kj] if (defined $pNum1 && defined $pNum1->[$kj]);
	    my $pND = $pDen1->[$kj];
	    if ($pND) {
		my $sj = $pStrains->[$kj];
		printf($gh "%s %s",$si,$sj);
		foreach my $n01 (0,1) {
		    printf($gh " %d %d",
			   (defined $pNA ? $pNA->[$n01] : 0),
			   $pND->[$n01]);
		}
		print $gh "\n";
	    }
	}
    }
}

sub addToSNPMatrix1 {
    my ($g,$pNum,$pDen) = @_;
    my $fh = openToRead($g);
    while (<$fh>) {
	chomp;
	my ($si,$sj,$num0,$den0,$num1,$den1) = split;
	$pNum->{$si}{$sj}[0] += $num0 if ($num0);
	if ($den0) {
	    $pDen->{$si}{$sj}[0] += $den0;
	} elsif (!defined $pDen->{$si}{$sj}[0]) {
	    # make sure it's defined (this case should be rare)
	    $pDen->{$si}{$sj}[0] = 0;
	}
	$pNum->{$si}{$sj}[1] += $num1 if ($num1);
	$pDen->{$si}{$sj}[1] += $den1 if ($den1);
    }
}

sub allKeys {
    my ($pH,$th) = @_;
    my %isKey;
    foreach my $key1 (keys %$pH) {
	foreach my $key2 (keys %{$pH->{$key1}}) {
	    my $pA = $pH->{$key1}{$key2};
	    if ($pA && $pA->[0] >= $th) {
		$isKey{$key2} = $isKey{$key1} = 1;
	    }
	}
    }
    my @keys = sort keys %isKey;
    return \@keys;
}

sub outputSNPMatrices {
    my ($outPfx,$syn01,$pNum,$pDen,$pStrains,@dList) = @_;
    my $pRates = findSNPRateMatrix($pNum,$pDen,$pStrains,$syn01,$outPfx);
    my @sList = orderKeysByRelatedness($pRates);
    #report("writing to $outPfx files ...");
    my $gh = openToWrite("$outPfx.strains");
    print $gh join("\n",@sList) . "\n";
    $gh->close();
    
    my $gh = openToWrite("$outPfx.q40");
    foreach my $i (0 .. $#sList) {
	my $si = $sList[$i];
	foreach my $j (0 .. $#sList) {
	    my $sj = $sList[$j];
	    print $gh ' ' if ($j);
	    printf($gh "%12.10f",$pRates->{$si}{$sj});
	}
	print $gh "\n";
    }
    #report();
}

sub findSNPRateMatrix {
    my ($pNum,$pDen,$pStrains,$syn01,$outPfx) = @_;
    my %rate;
    my @sorted = sort {$a cmp $b} @$pStrains;
    my %notSeen = map {$_ => 1} @sorted;
    my %keep = map {$_ => 1} @$pStrains;
    foreach my $si (keys %$pDen) {
	next unless ($keep{$si});
	my $pNumI = $pNum->{$si};
	my $pDenI = $pDen->{$si};
	foreach my $sj (keys %$pDenI) {
	    next unless ($keep{$sj});
	    my $r = findSNPRate(defined $pNumI ? $pNumI->{$sj} : undef,
				   $pDenI->{$sj},$syn01);
	    if (defined $r) {
		$rate{$si}{$sj} = $rate{$sj}{$si} = $r;
		delete $notSeen{$si};
		delete $notSeen{$sj};
	    }
	}
    }
    # Complete, for missing ones
    if (keys %notSeen) {
	fillInMissingRates($pStrains,\%rate,$outPfx);
    }
    return \%rate;
}

sub findSNPRate {
    my ($pNum,$pDen,$syn01) = @_;
    my $sumD = $pDen->[$syn01];
    if ($sumD) {
	my $sumN = defined $pNum ? $pNum->[$syn01] : 0;
	my $r = $sumN/$sumD;
	# See th/0615 and (superceding it) th/1022, th/1023
	#my $rate1 = ($r-$QUAL_THRESHOLD_B)/(1-2*$QUAL_THRESHOLD_B);
	#my $rate2 = $r*3/4;
	#return max(0.00001,max($rate1,$rate2));  #4
	#return max(0.0000001,$rate1);            #3
	#return max(0.00001,$r);                 #5
	return max(0.0000001,$r);
    } else {
	return undef;
    }
}

sub fillInMissingRates {
    my ($pStrains,$pRate,$outPfx) = @_;
    my ($count,$sum);
    foreach my $si (keys %$pRate) {
	my $pRI = $pRate->{$si};
	foreach my $sj (keys %$pRI) {
	    $count ++;
	    $sum += $pRI->{$sj};
	}
    }
    #if ($count == 0) {
	#confess("Zero count in fillInMissingRates for $outPfx...");
    #}
    my $meanRate = $count > 0 ? $sum/$count : 0.001; # just a guess...
    foreach my $i (0 .. $#$pStrains-1) {
	my $si = $pStrains->[$i];
	foreach my $j ($i+1 .. $#$pStrains) {
	    my $sj = $pStrains->[$j];
	    unless (defined $pRate->{$si}{$sj}) {
		$pRate->{$si}{$sj} = $pRate->{$sj}{$si} = $meanRate;
	    }
	}
    }
}
     
sub orderKeysByRelatedness {
    my ($pRel) = @_;
    my @todo = keys %$pRel;
    # Find the LEAST related (highest pRel value) pair of strains, and start
    # with one of them.
    my ($max,$maxi);
    foreach my $i (0 .. $#todo-1) {
	foreach my $j ($i+1 .. $#todo) {
	    my $r = $pRel->{$todo[$i]}{$todo[$j]};
	    if ($r > $max) {
		$max = $r;
		$maxi = $todo[$j] eq 'REF' ? $j : $i;
	    }
	}
    }
    my @ans = ($todo[$maxi]);
    @todo = (@todo[0 .. $maxi-1],@todo[$maxi+1 .. $#todo]);
    while (@todo) {
	my $si = $ans[-1];
	@todo = sort {$pRel->{$si}{$a} <=> $pRel->{$si}{$b}} @todo;
	push @ans,shift @todo;
    }
    return @ans;
}

sub createNJTree {
    my ($pfx) = @_;
    #report("creating $pfx.bl");
    # We change UWOPS to UW to avoid exceeding name length limit in quicktree.
    # Also multiply all values by 1000 to avoid underflow in quicktree, and back again afterwards.
    my $alpha = createQuicktreeInput($pfx);
    `$QUICKTREE -in m -out t $pfx.alpha > $pfx.tree`;
    #qtToBlTree("$pfx.tree","$pfx.bl");
    #report();
}

sub createQuicktreeInput {
    my ($pfx) = @_;
    my $fh = openToRead("$pfx.strains");
    my @strains = <$fh>;
    my $out = "$pfx.alpha";
    my $gh = openToWrite($out);
    print $gh scalar(@strains) . "\n";
    $fh = openToRead("$pfx.q40");
    while (<$fh>) {
	chomp;
	my @row = split;
	my $str = shift @strains;
	chomp $str;
	$str =~ s/UWOPS/UW/;
	print $gh $str;
	foreach my $v (@row) {
	    printf($gh " %8.6f",$v*1000);
	}
	print $gh "\n";
    }
    $gh->close();
    return $out;
}

sub openToRead {
    my ($f) = @_;
    return new FileHandle($f);
}


sub openToWrite {
    my ($f) = @_;
    return new FileHandle(">$f");
}
   
sub max {
    my ($x,$y) = @_;
    if (defined $y && $y > $x) {
	return $y;
    } elsif (defined $x) {
	return $x;
    } else {
	return $y;
    }
}
