#!/usr/bin/env perl

use GD;
use GD::Simple;
use Getopt::Std;

use Utility;

use strict;

# Plot layout:
#  1 pixel per nucleotide, so width = width of read
#  $leeway positions each way vertically for each alignment, so
#  2*$leeway+1 positions. $leeway has max 100, min 750/#alis.
#  Vertical offsets 750 to 900 are for quality scores.

my %OPT = qw(s cere w 3 d .);
getopts('d:s:w:',\%OPT);

my $MAX_ALIGNMENTS = 30;

unless (@ARGV) {
    die("Usage: $0 [-d outDir] [-s species] [-w wsz] read1 read2 ...");
}

dotplotAli($OPT{d},$OPT{s},$OPT{w},@ARGV);

sub dotplotAli {
    my ($outDir,$species,$wsz,@reads) = @_;
    foreach my $r (expandReads(@reads)) {
	my $str = $r;
	$str =~ s/-.*//;
	my $sgr = $ENV{'SGR'};
	my $fq = "$sgr/read/$str/renamed.fastq";
	my $al = "$sgr/read/$str/$species/alilines.all.txt";
	my $ref = "$sgr/ref/$species/ALL.fa";
	if (! -s $fq) {
	    report("$r: missing $fq",1);
	} elsif (! -s $al) {
	    report("$r: missing $al",1);
	} elsif (! -s $ref) {
	    report("$r: missing $ref",1);
	} else {
	    dotplotAli1($fq,$r,$al,$ref,$wsz,$outDir);
	}
    }
}

sub expandReads {
    my (@lst) = @_;
    my @ans;
    foreach my $r (@lst) {
	if ($r =~ /[pq]1k$/) {
	    push @ans,$r;
	} else {
	    push @ans,"$r.p1k","$r.q1k";
	}
    }
    return @ans;
}

sub dotplotAli1 {
    my ($fqFile,$read,$alFile,$ref,$wsz,$outDir) = @_;

    report("$read: reading $fqFile");
    my ($seq1,$quals) = readFromFastq($fqFile,$read);
    unless ($seq1) {
	print "cannot find $read\n";
	return;
    }
    report();

    my @aliList = readAlilines($alFile,$read);
    unless (@aliList) {
	print "no alignments for $read\n";
	return;
    }
    my ($xsz,$ysz) = (length($seq1),900);
    my $leeway = int(($ysz-180)/(2*scalar(@aliList))-1);
    my $im = new GD::Simple($xsz,$ysz);
    $im->interlaced('true');
    $im->font('Tahoma');
    $im->fontsize(6);

    for (my $i=0; $i<=$#aliList; $i++) {
	dotplotAli2($read,$seq1,$quals,$aliList[$i],$i+1,$leeway,$ref,$wsz,$im);
    }

    plotQualities($quals,$ysz,$im);
    plotGrid($xsz,$ysz,$im);
    for (my $i=0; $i<=$#aliList; $i++) {
	my $yOff = 2*($leeway+1)*($i+1);
	dotplotAnnotate($aliList[$i],$yOff,$im);
    }

    my $outFile = "$outDir/$read.png";
    report("$read: creating $outFile");
    my $gh = $outFile ? openToWrite($outFile) : openToPipe('display');
    binmode $gh;
    print $gh $im->png;
    report();
}

sub plotGrid {
    my ($xsz,$ysz,$im) = @_;
    my $step = $xsz <= 200 ? 10 : 100;
    $im->fgcolor('red');
    for (my $i=$step; $i<=$xsz; $i+= $step) {
	my $psz = $i%(5*$step) ? 1 : 2;
	$im->penSize($psz,$psz);
	$im->moveTo($i,0);
	$im->lineTo($i,$ysz);
    }
    $im->fgcolor('black');
}

sub plotQualities {
    my ($quals,$ysz,$im) = @_;
    my $len = length($quals);
    $im->moveTo(0,$ysz);
    for (my $i=0; $i<$len; $i++) {
	my $q = ord(substr($quals,$i,1))-33;
	$im->lineTo($i,$ysz-3*$q);
    }
}

sub dotplotAli2 {
    my ($read,$seq1,$quals,$pA,$aliNum,$leeway,$ref,$wsz,$im) = @_;
    my ($chr,$rlo,$rhi,$clo,$chi,$fc) = @{$pA}[3..8];
    if ($rlo > $rhi) {
	my $tmp = $rlo;
	$rlo = $rhi;
	$rhi = $tmp;
    }

    report("$read: reading $ref");
    my $seq2 = readReference($pA,$ref,$leeway);
    report();

    report("$read: creating image");
    my $len1 = length($seq1);

    my $yOff = 2*($leeway+1)*$aliNum;

    # Pink to show the official alignment
    my $pink = $im->colorAllocate(255,230,230);
    $im->bgcolor($pink);
    $im->fgcolor($pink);
    my $wid = min($leeway-4,20);
    $im->rectangle($rlo-1,$yOff-2-$leeway-$wid,$rhi-1,$yOff-1-$leeway+$wid);
    # Parallel pink on quality region
    $im->rectangle($rlo-1,720,$rhi-1,900);
    $im->fgcolor('black');
    $im->bgcolor('white');

    plotMatches($seq1,$seq2,$wsz,$yOff,$leeway,$im);
    report();
}

sub dotplotAnnotate {
    my ($pA,$yOff,$im) = @_;

    my ($chr,undef,undef,$clo,$chi,$fc) = @{$pA}[3..8];
    # Clear space behind annotation, and write it
    $im->fgcolor('white');
    $im->rectangle(0,$yOff-10,160,$yOff-2);
    $im->fgcolor('black');
    $im->moveTo(5,$yOff-2);
    $im->string(sprintf("%s %d %d %s %s %5.2f%%",$chr,$clo,$chi,$fc,$pA->[1],$pA->[-2]));
}

sub readFromFastq {
    my ($fqFile,$read) = @_;
    my $fh = openToRead($fqFile);
    while (1) {
	my @buf;
	for (my $k=0; $k<4; $k++) {
	    push @buf,scalar(<$fh>);
	}
	return unless ($buf[0]);
	my ($r) = split(/\s+/,substr($buf[0],1));
	if ($r eq $read) {
	    chomp @buf;
	    return ($buf[1],$buf[3]);
	}
    }
}

sub readAlilines {
    my ($alFile,$read,$ref) = @_;
    report("$read: reading $alFile");
    my $fh = openToRead($alFile);
    my @ans;
    while (<$fh>) {
	chomp;
	my @a = split;
	if ($a[2] eq $read && @ans < $MAX_ALIGNMENTS) {
	    push @ans,\@a;
	} elsif (@ans) {
	    report();
	    return @ans;
	}
    }
    report();
    return ();
}

sub readReference {
    my ($pA,$ref,$leeway) = @_;
    my ($chr,$rlo,$rhi,$clo,$chi,$fc) = @{$pA}[3..8];
    my $rlen = $pA->[-1];
    my $rsw;
    if ($fc eq 'F') {
	$rsw = '';
	$clo -= $rlo-1;
	$chi += $rlen-$rhi;
    } else {
	$rsw = '-r';
	$clo -= $rlen-$rlo;
	$chi += $rhi-1;
    }
    $clo -= $leeway;
    $chi += $leeway;
    my $clo2 = $clo;
    my $pfx;
    if ($clo < 1) {
	$pfx = 'X' x (1-$clo);
	$clo2 = 1;
    };
    my $cmd = "extractRegion.pl -c $chr -f $clo2 -t $chi $rsw $ref";
    #print "CMD: $cmd\n";
    my $fh = openFromPipe($cmd);
    my $seq = $pfx;
    while (<$fh>) {
	chomp;
	$seq .= $_;
    }
    if (length($seq) < $chi-$clo+1) {
	$seq .= 'X' x ($chi-$clo+1-length($seq));
    }
    return $seq;
}

sub plotSequences {
    my ($seq1,$seq2,$d,$im) = @_;
    my $len1 = length($seq1);
    my $len2 = length($seq2);
    for (my $k=0; $k<$len1; $k++) {
	$im->moveTo(($k+1)*$d,$d);
	$im->string(substr($seq1,$k,1));
	$im->moveTo(($k+1)*$d,($len2+2)*$d);
	$im->string(substr($seq1,$k,1));
    }
    for (my $k=0; $k<$len2; $k++) {
	my $y = ($len2+1-$k)*$d;
	$im->moveTo(0,$y);
	$im->string(substr($seq2,$k,1));
	$im->moveTo(($len1+1)*$d,$y);
	$im->string(substr($seq2,$k,1));
    }
}

sub plotMatches {
    my ($seq1,$seq2,$wsz,$yOff,$leeway,$im) = @_;
    my $len1 = length($seq1);
    my $len2 = length($seq2);
    for (my $i=0; $i<=$len1-$wsz; $i++) {
	my $ci = substr($seq1,$i,$wsz);
	my $jMax = $i+2*$leeway;
	for (my $j=$i; $j<=$jMax; $j++) {
	    my $cj = substr($seq2,$j,$wsz);
	    if ($ci eq $cj) {
		for (my $k=0; $k<$wsz; $k++) {
		    $im->moveTo($i+$k,$yOff-$j+$i);
		    $im->ellipse(1,1);
		    #$im->lineTo(($i+1.5+$wsz-1)*$d,($len2+0.5-$j-$wsz+1)*$d);
		}
	    }
	}
    }
    $im->moveTo(0,$yOff);
    $im->fgcolor($im->colorAllocate(0,255,0));
    $im->penSize(2,2);
    $im->lineTo($len1,$yOff);
    $im->penSize(1,1);
    $im->fgcolor('black');
}
