#!/usr/bin/env perl

use Utility;
use strict;

# Args should be e.g. $SGR/ref/cere/self/chr12.cere.ali, then ref genome.fa
# containing only nuclear chromosomes, then read file(s) for same strain.

estimateRDNACoverage(@ARGV);

sub estimateRDNACoverage {
    my ($selfChr12,$refFa,@fastqFiles) = @_;
    my ($rdnaLo,$rdnaHi,$rdnaLen) = getRDNARefCoords($selfChr12);
    my $fh = openToRead($refFa);
    my $sz = 0;
    while (<$fh>) {
	chomp;
	$sz += length($_) unless (/^>/);
    }
    my ($inside,$total);
    foreach my $f (@fastqFiles) {
	my $fh = openFromPipe("ssaha2 $f $refFa -tags 1 -align 0");
	my ($prevRead,$prevScore,@mat);
	while (<$fh>) {
	    chomp;
	    if (/^ALIGNMENT/) {
		my (undef,$score,$read,$chr,$rlo,$rhi,$clo,$chi) = split;
		unless ($read eq $prevRead) {
		    if (@mat) {
			($inside,$total) = augmentCounts(\@mat,$rdnaLo,$rdnaHi,$inside,$total);
			@mat = ();
		    }
		    $prevRead = $read;
		    $prevScore = $score;
		}
		push @mat,[$chr,$rlo,$rhi,$clo,$chi] if ($score == $prevScore);
	    }
	}
	($inside,$total) = augmentCounts(\@mat,$rdnaLo,$rdnaHi,$inside,$total);
    }
    #my $rdnaLen = $rdnaHi-$rdnaLo+1;
    my $covInside = $inside/$rdnaLen;
    my $covOutside = ($total-$inside)/($sz-$rdnaLen);
    printf("Inside %7.3f outside %7.3f ratio %7.3f\n",
	   $covInside,$covOutside,$covInside/$covOutside);
}

sub augmentCounts {
    my ($pMat,$rdnaLo,$rdnaHi,$inside,$total) = @_;
    my @depth;
    foreach my $pA (@$pMat) {
	my (undef,$rlo,$rhi) = @$pA;
	foreach my $k ($rlo .. $rhi) {
	    $depth[$k] ++;
	}
    }
    foreach my $pA (@$pMat) {
	my ($chr,$rlo,$rhi,$clo,$chi) = @$pA;
	my $sumD;
	foreach my $k ($rlo .. $rhi) {
	    my $ratio = ($k-$rlo)/($rhi-$rlo);
	    my $j = $ratio*$chi + (1-$ratio)*$clo;
	    my $w = 1/$depth[$k];
	    $total += $w;
	    if ($chr eq 'chr12' && $j >= $rdnaLo && $j <= $rdnaHi) {
		$inside += $w;
	    }
	}
    }
    return ($inside,$total);
}

sub getRDNARefCoords {
    my ($f) = @_;
    my $fh = openFromPipe("sort +1 -nr $f");
    while (<$fh>) {
	chomp;
	my @a = split;
	if ($a[3] eq 'chr12' && $a[4] < $a[6] && $a[10] >= 99.9) {
	    # rDNA reference region is $a[4] to $a[7], but the length
	    # of one copy is start of first to start of second.
	    return ($a[4],$a[7],$a[6]-$a[4]);
	}
    }
    die("Cannot find rDNA coords in $f");
}
