#!/usr/bin/env perl

# -c: chromosome/contig name
# -z: zero-offset instead of one-offset
# -f,-t: from, to (inclusive)
# -g: geneName
# -G: GFF file name (needed iff we have -g)
# -r: reverse-complement result
# -n: number of bases
# -l: line length (default 60)
# -o: output (default STDOUT)

use Utility;
use Getopt::Std;
use strict;

my %OPT = qw(l 60);
getopts('c:zf:g:G:t:l:n:o:r',\%OPT);

if ($OPT{g}) {
    die("Need -G if we have -g") unless ($OPT{G} && -s $OPT{G});
    ($OPT{c},$OPT{f},$OPT{t},$OPT{r}) = findGeneRegion($OPT{g},$OPT{G});
}

# Not sure why we need this ...
# my @arr = split(/:/,$ARGV[0]);
# if (@arr == 3) {
#     shift @ARGV;
#     if ($arr[1] > $arr[2]) {
# 	($OPT{c},$OPT{t},$OPT{f}) = @arr;
# 	$OPT{r} = 1;
#     } else {
# 	($OPT{c},$OPT{f},$OPT{t}) = @arr;
#     }
#     $OPT{z} = 1;
# }

my $chr = $OPT{c};
my ($from,$to) = setRegion();

sub setRegion {
    my ($from,$to);
    if (defined $OPT{f} && defined $OPT{t} && $OPT{f} > $OPT{t}) {
	my $tmp = $OPT{f};
	$OPT{f} = $OPT{t};
	$OPT{t} = $tmp;
    }
    unless (defined $OPT{z}) {
	$OPT{f} -- if (defined $OPT{f});
	$OPT{t} -- if (defined $OPT{t});
    }
    if (defined $OPT{n}) {
	if (defined $OPT{f}) {
	    $from = $OPT{f};
	    if (defined $OPT{t}) {
		unless ($OPT{t} == $OPT{f}+$OPT{n}-1) {
		    die("-f, -t and -n must satisfy T = F+N-1");
		}
		$to = $OPT{t};
	    } else {
		$to = $OPT{f}+$OPT{n}-1;
	    }
	} elsif (defined $OPT{t}) {
	    $to = $OPT{t};
	    $from = $to-$OPT{n}+1;
	} else {
	    $from = 0;
	    $to = $from+$OPT{n}-1;
	}
    } else {
	$from = defined $OPT{f} ? $OPT{f} : 0;
	$to = defined $OPT{t} ? $OPT{t} : 1000000000;
    }
    return ($from,$to);
}

sub findGeneRegion {
    my ($geneName,$gffFile) = @_;
    my $fh = openToRead($gffFile);
    my $pat = "gene=$geneName;";
    while (<$fh>) {
	chomp;
	if (index($_,$pat) > 0) {
	    my ($chr,undef,$type,$lo,$hi,undef,$strand) = split;
	    if ($type eq 'gene') {
		return ($chr,$lo,$hi,$strand eq '-');
	    }
	}
    }
    die("Cannot find $geneName in $gffFile");
}

my $gh = defined $OPT{o} ? openToWrite($OPT{o}) : *STDOUT;

if ($OPT{g}) {
    print $gh ">$OPT{g}\n";
}

my ($i,$c) = (0);
my $active = !defined $OPT{c};
my $buf = "";
 LOOP:
    while (<>) {
	chomp;
	if (/^>(\S+)/) {
	    $i = 0;
	    my $c = $1;
	    $c =~ s/^ref\|(.*)\|$/\1/;
	    $c =~ s/\[.*//;
	    if (defined $OPT{c}) {
		last if ($active); # done it
		(my $cc = $c) =~ s/\[.*//;
		$active = ($c eq $OPT{c} || $cc eq $OPT{c});
	    }
	} elsif ($active) {
	    my $j = $i+length($_)-1;
	    if ($j < $from) {
		# line before reg: discard
	    } elsif ($i < $from) {
		# line starts before reg but overlaps
		if ($j < $to) {
		    # i < from < j < to
		    $buf = substr($_,$from-$i);
		} else {
		    # i < from < to <= j
		    $buf = substr($_,$from-$i,$to-$from+1);
		    last LOOP;
		}
	    } elsif ($j < $to) {
		# from <= i < j < to
		$buf .= $_;
	    } else {
		# from <= i <= to <= j
		$buf .= substr($_,0,$to-$i+1);
		last LOOP;
	    }
	    $i = $j+1;
	}
    }

if ($OPT{r}) {
    $buf = reverse($buf);
    $buf =~ tr/ACGT/TGCA/;
}
my $lb = length($buf);
my $ll = $OPT{l};
while ($lb > 0) {
    if ($lb > $ll) {
	print $gh substr($buf,0,$ll) . "\n";
	$buf = substr($buf,$ll);
	$lb -= $ll;
    } else {
	print $gh $buf . "\n";
	$lb = 0;
    }
}
$gh->close();
