#!/usr/bin/env perl

# Called from e.g. SGRP/RunFregeneExperiment.pm to make equal-sized "chromosomes".

use SGRP::Variables;

use Utility;
use Getopt::Std;
use strict;

my %OPT = qw(l 1 h 10000000);
getopts('l:h:e',\%OPT);

# Each line in fregeneQualities.gz is of the form
# offset individualID qualString
# where offset is from 1 in the chromosome, and individualID is also from 1.

my $REGION_LO = $OPT{l};
my $REGION_HI = $OPT{h};

my %ERROR_PROB;

if (@ARGV == 3) {
    setVariables();
    fregeneCreateLayout(@ARGV);
} else {
    die("Usage: fregeneCreateLayout.pl qualFile fregFile outFile");
}

sub fregeneCreateLayout {
    my ($qualFile,$fregFile,$outFile) = @_;
    my $qh = openToRead($qualFile);
    my $fh = openToRead($fregFile);
    # Read up to and including <DATA> line.
    my @snpOffsets;
    while (<$fh>) {
	chomp;
	if (/^<DATA>/) {
	    # Minor allele positions for the reference sequence:
	    push @snpOffsets,snpHash(substr($_,6));
	    last;
	}
    }
    my $gh = openToWrite($outFile);
    fregeneCreateLayout1($qh,$fh,$gh,\@snpOffsets);
}

sub fregeneCreateLayout1 {
    my ($qh,$fh,$gh,$pSNPOffs) = @_;
    my $off = $REGION_LO;
    my @buf;
    while (<$qh>) {
	chomp;
	# $qtOff is offset (from 1, in chromosome) at which read starts;
	# $qtIndiv is the individual (from 1; ref is zero) that it's for.
	my ($qtOff,$qtIndiv,$qtQuals) = split(" ",$_,3);
	unless (defined $qtQuals) {
	    die("No quality string in $_");
	}
	my $hi = $qtOff+length($qtQuals)-1;
	# Ignore if we haven't reached our region yet.
	next if ($hi < $REGION_LO);
	# Finish if we've gone past our region.
	last if ($qtOff > $REGION_HI);
	# If read spans start of our region, truncate beginning.
	# (Equivalent situation at end won't hurt us).
	if ($qtOff < $REGION_LO) {
	    $qtQuals = substr($qtQuals,$REGION_LO-$qtOff);
	    $qtOff = $REGION_LO;
	}
	# Make sure we've read enough of $fh to have the SNP line
	# for qtIndiv in $pSNPOffs->[$qtIndiv].
	ensureSNPOffsets($qtIndiv,$fh,$pSNPOffs);
	extendBuffer($qtOff,$off,$qtIndiv,$qtQuals,$pSNPOffs,\@buf,$gh);
	while ($off < $qtOff) {
	    outputLine(\@buf,$off,$gh);
	    $off ++;
	}
    }
    while ($off <= $REGION_HI) {
	outputLine(\@buf,$off,$gh);
	$off ++;
    }
}

sub ensureSNPOffsets {
    my ($ind,$fh,$pSNPOffs) = @_;
    while ($#$pSNPOffs < $ind) {
	chomp ($_ = <$fh>);
	push @$pSNPOffs,snpHash($_);
    }
}

sub snpHash {
    my ($line) = @_;
    my %h;
    foreach my $off (split(" ",$line)) {
	last if ($off > $REGION_HI);
	$h{$off} = 1 if ($off >= $REGION_LO);
    }
    return \%h;
}

sub extendBuffer {
    my ($qtOff,$off,$ind,$quals,$pSNPOffs,$pBuf,$gh) = @_;
    my $delta = $qtOff-$off;
    # Find a free channel in $pBuf
    my $i=0;
    while ($i <= $#$pBuf && scalar(@{$pBuf->[$i]}) > $delta) {
	$i++;
    }
    my $jMax = length($quals)-1;
    for (my $j=0; $j<=$jMax; $j++) {
	my $refIsSNP = defined $pSNPOffs->[0]{$qtOff+$j};
	my $indIsSNP = defined $pSNPOffs->[$ind]{$qtOff+$j};
	my $val = ($refIsSNP eq $indIsSNP) ? 'A' : 'C';
	my $qc;
	if ($OPT{e}) {
	    $qc = substr($quals,$j,1);
	    my $ep = errorProb($qc);
	    # Swap over if error
	    if (rand() < $ep) {
		$val = $val eq 'A' ? 'C' : 'A';
	    }
	} else {
	    $qc = '~'; # i.e. very high quality; and we never swap
	}
	$pBuf->[$i][$j+$delta] = "$val$qc";
    }
    # To trigger start-of-read line:
    my $bid = $pBuf->[$i][$delta];
    if (length($bid) == 2) {
	$pBuf->[$i][$delta] .= $ind;
    } else {
	die("Bad pBuf $i $delta at $off, jMax = $jMax: $bid");
    }
    #printf($gh "Channel for %d (from %d) is %d; filled from %d to %d\n",
	#   $ind,$qtOff,$i,$delta,$delta+$jMax);
}
	
sub outputLine {
    my ($pBuf,$off,$gh) = @_;
    if ($off%100 == 0) {
	print $gh ">REF $off\n";
	if ($off % 10000 == 0) {
	    report("reached offset $off",1);
	}
    }
    # ">" lines for starts of reads here
    for (my $i=0; $i<=$#$pBuf; $i++) {
	if (@{$pBuf->[$i]}) {
	    my $pair = $pBuf->[$i][0];
	    if ($pair && substr($pair,2)) {
		printf($gh ">%d %d\n",substr($pair,2),$i);
	    }
	}
    }
    # Sequence line.
    my $line = 'A ';
    #my $debug = ($off == 18009);
    for (my $i=0; $i<=$#$pBuf; $i++) {
	if (@{$pBuf->[$i]}) {
	   my $pair = shift @{$pBuf->[$i]};
	   #print "PAIR <$pair>\n" if ($debug);
	   # $pair can have third and subsequent chars as above.
	   # It can also be empty, if there was a gap in the use
	   # of a channel! (This was the cause of the Easter 2007 bug...)
	   $line .= $pair ? substr($pair,0,2) : '  ';
       } else {
	   $line .= '  ';
       }
    }
    #exit() if ($debug);
    $line =~ s/  *$//; # chop trailing spaces
    print $gh "$line\n";
}

sub errorProb {
    my ($qc) = @_;
    unless (defined $ERROR_PROB{$qc}) {
	my $q = ord($qc) - 33;
	# Normalize according to our experiments
	$q = 55*(1-exp(-0.029*$q));
	# Error prob above 0.75 makes no sense!
	$ERROR_PROB{$qc} = min(0.75,exp(-log(10)*0.1*$q));
    }
    return $ERROR_PROB{$qc};
}

