#!/usr/bin/env perl

use SGRP::Variables;
use File::Basename;
use Utility;

use Getopt::Std;
use strict;

autoflush STDOUT 1;

my %OPT = qw(r 10 n 200 c 1 m 1);
getopts('c:m:n:r:',\%OPT);

# Each line output to fregeneQualities*.gz is of the form
# offset individualID qualString

# In Solexa case, input file could be /lustre/scratch1/sanger/lh3/lh3/work/yeast/SK1.fastq

my $REGION_SIZE = $OPT{r}*1000000;
my $COVERAGE = $OPT{c};
my $NUM_INDIVS = $OPT{n};

my ($QUAL_DIFF_MAIN,$QUAL_DIFF_HELD);

if (@ARGV == 3) {
    setVariables();
    fregeneQualityStrings(0,@ARGV);
} elsif (@ARGV == 1) {
    setVariables();
    fregeneQualityStrings(1,$ARGV[0]);
} else {
    print STDERR "Usage: one of\n";
    print STDERR "   fregeneQualityStrings.pl [-c cov (1)] [-n nIndivs (200)] [-r regionMB (10)] species filtMeth strainSetName\n";
    die          "   fregeneQualityStrings.pl [-c cov (1)] [-n nIndivs (200)] [-r regionMB (10)] solexaFastqFile\n";
}

sub fregeneQualityStrings {
    my ($solexa,$speciesOrFastq,$filtMeth,$strainSet) = @_;
    my $topDir;
    if ($solexa) {
	my $base = basename($speciesOrFastq,'.fastq');
	$topDir = "$ROOT/match/FQS/$base";
    } else {
	$topDir = "$ROOT/match/$strainSet/$speciesOrFastq/$filtMeth/000";
    }
    my $fregDir = sprintf("%s/freg%d_%d",$topDir,$NUM_INDIVS,$COVERAGE);
    mkdir_p($fregDir);
    # Split version:
    #my $gMain = "$fregDir/fregeneQualities09.gz";
    #my $gHeld = "$fregDir/fregeneQualities91.gz";
    # Unsplit version:
    my $gMain = "$fregDir/fregeneQualities.gz";
    my $gHeld;
    if (stamp($gMain)) {
	report("writing $gMain",1);
	my $gh = openToPipe("sort -n | gzip -c > $gMain");
	my $hh = openToPipe("sort -n | gzip -c > $gHeld") if ($gHeld);
	# Number of bases to cover:
	my $needed = $REGION_SIZE*$COVERAGE*$NUM_INDIVS;
	my $rev = 0; # alternate between forward and reverse application
	while ($needed > 0) {
	    report(sprintf("running %s with %4.2f Mb needed",$rev ? 'backwards' : 'forwards',
			   $needed/1000000));
	    if ($solexa) {
		my $qh = openToRead($speciesOrFastq);
		$needed = fregeneQualityStrings1($needed,$rev,undef,$qh,$gh,$hh);
	    } else {
		# Allow both old and new structures:
		foreach my $d (glob("$topDir/p?/*"),glob("$topDir/00[01]/p?/*")) {
		    my $mf = "$d/matches.gz";
		    my $qf = "$d/qualities.gz";
		    if (-s $mf && -s $qf) {
			my $mh = openFromPipe("gunzip -c $mf | grep '^ALIGNMENT'");
			my $qh = openToRead($qf);
			#report("processing $d");
			$needed = fregeneQualityStrings1($needed,$rev,$mh,$qh,$gh,$hh);
			#report();
		    }
		}
	    }
	    report();
	    $rev = 1-$rev;
	}
	report("waiting for sorting to finish");
	$gh->close();
	$hh->close() if ($hh);
	report();
	if (0) {
	    # Wait for sorting to finish
	    my ($szMain,$szHeld);
	    while (1) {
		my $newMain = -s $szMain;
		my $newHeld = -s $szHeld;
		last if ($newMain == $szMain && $szMain > 0
			 && $newHeld == $szHeld && $szHeld > 0);
		$szMain = $newMain;
		$szHeld = $newHeld;
		sleep(10);
	    }
	    sleep(10); # to make sure it can finish
	}
	unstamp($gMain);
    } else {
	die("Cannot stamp $gMain");
    }
}

# $mh (ALIGNMENT lines) will be undef when Solexa.

sub fregeneQualityStrings1 {
    my ($needed,$rev,$mh,$qh,$gh,$hh) = @_;
    while (1) {
	my $qstr;
	if ($mh) {
	    $_ = <$mh>;
	    last unless ($_);
	    chomp;
	    my @a = split;
	    my ($rlo,$rhi,$fc) = @a[4,5,8];
	    $_ = <$qh>;
	    unless ($_) {
		die("No quality line to go with @a");
	    }
	    chomp;
	    (undef,$qstr) = split(" ",$_,2);
	    if ($fc eq 'F') {
		$qstr = substr($qstr,$rlo-1,$rhi-$rlo+1);
		if ($rev) {
		    $qstr = reverse($qstr);
		}
	    } else {
		$qstr = substr($qstr,$rhi-1,$rlo-$rhi+1);
		if (!$rev) {
		    $qstr = reverse($qstr);
		}
	    }
	} else {
	    <$qh>;
	    <$qh>;
	    <$qh>;
	    $qstr = <$qh>; # fourth one is quality string
	    chomp $qstr;
	    if (rand() < 0.5) {
		$qstr = reverse($qstr);
	    }
	}
	if ($qstr) { # allow for indels etc ...
	    # Send 90% to $gh and 10% to $hh; or all to $gh.
	    my $oh = $gh;
	    my $oh = ($hh && rand() >= 0.9) ? $hh : $gh;
	    my $len = length($qstr);
	    my $start = 1+int(rand()*$REGION_SIZE);
	    my $end = $start+$len-1;
	    my $ind = int($NUM_INDIVS*rand())+1;
	    if ($end <= $REGION_SIZE) {
		printf($oh "%d %d %s\n",$start,$ind,$qstr);
	    } else {
		printf($oh "%d %d %s\n",$start,$ind,substr($qstr,$REGION_SIZE-$start+1));
		printf($oh "%d %d %s\n",1,$ind,substr($qstr,$REGION_SIZE-$start+1));
	    }
	    $needed -= $len;
	    last unless ($needed > 0);
	}
    }
    return $needed;
}
