#!/usr/bin/perl

use Utility;
use Getopt::Std;
use File::Basename;
use strict;

# Args: rin*mafs.gz file (from fregeneMAFs.pl) then one or more sorted.gz or imputed.gz files,
# assumed to go "end to end" to make up the offsets in rin*mafs.gz

# Bin 0 is freq == 0
my @FLOOR = qw(undef 0 0.001 0.005 0.01 0.02 0.03 0.05 0.10 0.20 0.30 0.40 0.5001);

# -m: maximum offset to consider, e.g. 1000000
# -n: number of discrepancies we need to call a SNP (default: 1)
# -o: output file (default STDOUT)
# -q: min quality for calling a discrepancy (default: any)

my %OPT;
my @ORIG_ARGS = @ARGV;

getopt('m:n:o:q:',\%OPT);
my $OFFMAX = $OPT{m};
my $NDISC = $OPT{n} || 1;
my $QDISC = $OPT{q} ? chr($OPT{q}+33) : '!';
my $OUTFILE = $OPT{o};

if (@ARGV < 2) {
    die("Usage: $0 [-m offMax] [-n nDisc] [-q qDisc] rin*mafs.gz sorted|imputed1.gz sorted|imputed2.gz ...");
} else {
    fregeneSNPEval(@ARGV);
}

sub fregeneSNPEval {
    my ($mafFile,@aliFiles) = @_;
    my $mh = openToRead($mafFile);
    my %freq;
    while (<$mh>) {
	chomp;
	my ($off,$fr) = split;
	$fr = 1-$fr if ($fr > 0.5);
	$freq{$off} = $fr;
	last if (defined $OFFMAX && $off > $OFFMAX);
    }
    my $off = 1;
    my ($fpCount,%seen);
    foreach my $f (@aliFiles) {
	my $fh = openToRead($f);
	<$fh>; # header
	while (<$fh>) {
	    chomp;
	    my $len = length($_);
	    my %count;
	    for (my $k=0; $k<$len; $k+=2) {
		my $v = substr($_,$k,1);
		if ($v =~ /[A-Za-z]/) {
		    my $q = substr($_,$k+1,1);
		    unless (defined $QDISC && $q ne ' ' && $QDISC gt $q) {
			$v =~ tr/a-z/A-Z/;
			$count{$v} ++;
		    }
		}
	    }
	    my $nVals=0;
	    foreach my $c (values %count) {
		if ($c >= $NDISC) {
		    $nVals++;
		    if ($nVals == 2) {
			# then it's claimed to be a SNP
			if (defined $freq{$off}) {
			    $seen{$off} = 1;
			} else {
			    $fpCount ++;
			}
			last;
		    }
		}
	    }
	    $off ++;
	    last if (defined $OFFMAX && $off > $OFFMAX);
	}
    }
    # Max offset we actually saw
    my $maxOff = $off-1;
    my @offs;
    foreach my $off (keys %freq) {
	push @offs,$off if ($off <= $maxOff);
    }
    # No. of monomorphic locations.
    my $nMono = $maxOff - scalar(@offs);
    my (@nPos,@nNeg);
    $nPos[0] = $fpCount;
    $nNeg[0] = $nMono-$fpCount;
    foreach my $off (@offs) {
	my $fr = $freq{$off};
	my $b = frequencyBin($fr);
	if (defined $seen{$off}) {
	    $nPos[$b] ++;
	} else {
	    $nNeg[$b] ++; 
	}
    }
    my $cmd = basename($0);
    my $gh = openToWrite($OUTFILE);
    print $gh "$cmd @ORIG_ARGS\n";
    print $gh "       Range               Negative           Positive\n";
    foreach my $b (0 .. $#FLOOR-1) {
	if ($b == 0) {
	    print $gh "     MAF == 0       ";
	} else {
	    printf($gh "%4.1f%% < MAF <= %4.1f%%",100*$FLOOR[$b],100*$FLOOR[$b+1]);
	}
	my $sum = $nPos[$b]+$nNeg[$b];
	if ($sum > 0) {
	    printf($gh " %8d = %6.2f%%",$nNeg[$b],100*$nNeg[$b]/$sum);
	    printf($gh " %8d = %6.2f%%\n",$nPos[$b],100*$nPos[$b]/$sum);
	} else {
	    print $gh "                 none\n";
	}
    }
    print $gh "\n";
}

sub frequencyBin {
    my ($fr) = @_;
    return 0 if ($fr == 0);
    foreach my $k (1 .. $#FLOOR-1) {
	return $k if ($fr < $FLOOR[$k+1]); # strictly ">"
    }
    die("Cannot find bin for $fr");
}
