#!/usr/bin/perl

# Args: [-l lo] [-h hi] alicoords.gz chrName refGenome.map out.png

use GD::Simple;
use Utility;
use Getopt::Std;

use strict;

my @COLOURS = qw(black red green blue orange purple brown);

my %OPT;
getopts('l:h:',\%OPT);

if (@ARGV == 4) {
    gdLayout(@ARGV);
} else {
    die("Usage: $0 [-l lo] [-h hi] alicoords.gz chrName refGenome.map out.png\n");
}

sub gdLayout {
    my ($inFile,$chr,$rgmFile,$outFile) = @_;

    my $pRGMap = readMapFile($chr,$rgmFile);

    my ($regLo,$regHi,$pLines) = readSpans($inFile,$OPT{l},$OPT{h});

    my ($pBars,$barLo,$pLevel) = decideBarsAndLevels($pLines);
    plotBars($pRGMap,$pBars,$pLevel,$regLo,$regHi,$barLo,$outFile);
}

# Return [from,to,len] triple for every alignment for the
# specified chromosome. from is offset in ref coords, to is 
# offset in alignment coords.

sub readMapFile {
    my ($chr,$f) = @_;
    my @map;
    my $fh = openToRead($f);
    while (<$fh>) {
	chomp;
	my ($name,$from,$to,$len) = split;
	foreach my $comp (split(/\./,$name)) {
	    if ($comp eq $chr) {
		push @map,[$from,$to,$len];
		next;
	    }
	}
    }
    return \@map;
}

sub readSpans {
    my ($inFile,$regLo,$regHi) = @_;
    my $fh = openToRead($inFile);
    my ($foundLo,$foundHi,@lines);
    while (<$fh>) {
	chomp;
	my ($read,$chr,$dir,$iter,@spans) = split;
	# Ignore if completely out of range of interest:
	next unless (overlapsRangeOfInterest($regLo,$regHi,\@spans));
	$foundLo = min($spans[0],$foundLo);
	$foundHi = max($spans[-1],$foundLo);
	(my $str = $read) =~ s/-.*//;
	$str = 0 if ($str eq 'REF');
	push @{$lines[$iter]{$str}},[$read,$chr,$dir,@spans];
    }
    return (defined $regLo ? $regLo : $foundLo,
	    defined $regHi ? $regHi : $foundHi,
	    \@lines);
}

sub overlapsRangeOfInterest {
    my ($regLo,$regHi,$pSpans) = @_;
    return 1 unless (defined $regLo || defined $regHi);
    return 0 if (defined $regLo && $regLo > $pSpans->[-1]);
    return 0 if (defined $regHi && $regHi < $pSpans->[0]);
    for (my $i=0; $i<$#$pSpans; $i+=2) {
	if ((!defined $regHi || $pSpans->[$i] < $regHi) && 
	    (!defined $regLo || $pSpans->[$i+1] > $regLo)) {
	    return 1;
	}
    }
    return 0;
}

# Each $pLines->[$iter]{$str} is: [read,chr,direction,@spans], as in
# alicoords.gz.

# Each $bars[$i] is [level,iter,str,read,chr,direction,@spans]

sub decideBarsAndLevels {
    my ($pLines) = @_;
    my @bars;
    my $barLo = 0; # Low bar index for current strain.
    # [barHi,strain] for each strain:
    my @level;
    # Sort by iteration first
    foreach my $iter (0 .. $#$pLines) {
	my $pLinesI = $pLines->[$iter];
	next unless (defined %$pLinesI); # maybe we have nothing for some
	foreach my $str (sort keys %$pLinesI) {
	    my @usedUpTo;
	    my @strBars;
	    # Sort by starting point of alignment, and form @strBars.
	    foreach my $pLine (sort {$a->[3] <=> $b->[3]} @{$pLinesI->{$str}}) {
		my $kUse = levelToPlaceAt($pLine,\@usedUpTo);
		push @strBars,[$kUse,@$pLine];
		$usedUpTo[$kUse] = $pLine->[-1];
	    }
	    # Add contents of @strBars to @bars, adding on $barLo to each level.
	    foreach my $pBar (@strBars) {
		my ($k,@rest) = @$pBar;
		push @bars,[$barLo+$k,$iter,$str,@rest];
	    }
	    $barLo += scalar(@usedUpTo);
	    push @level,[$barLo,$iter,$str]; # we've increased $barLo, so it's now barHi.
	}
    }
    # Sort by bar height.
    @bars = sort {$a->[0] <=> $b->[0]} @bars;
    return (\@bars,$barLo,\@level);
}

# Return the first level that is not occupied from $pLine->[3] (low
# end of spans) forwards, or create a new level if none such.

sub levelToPlaceAt {
    my ($pLine,$pUsedUpTo) = @_;
    my $loEnd = $pLine->[3];
    foreach my $k (0 .. $#$pUsedUpTo) {
	return $k if ($pUsedUpTo->[$k] < $loEnd);
    }
    return $#$pUsedUpTo+1;
}

sub plotBars {
    my ($pRGMap,$pBars,$pLevel,$xlo,$xhi,$yhi,$outFile) = @_;
    
    # X and Y sizes of (main part of) image, in pixels
    my $dimX = max(1000,int(($xhi-$xlo+1)/20)); # was: max(1000,int(.../200))
    my $dimY = max(900,($yhi+1)*14);

    # Scale factors for X and Y
    my $sx = $dimX/($xhi-$xlo+1);
    my $sy = $dimY/($yhi+1);

    # The actual image extends further to right and downwards
    my $im = new GD::Simple(100+$dimX,100+$dimY);

    # Strain colour index: give an index to each strain by its first
    # occurrence in $pBars. We will recycle colours later by taking mod.
    my $pStrColIdx = decideStrainColourIndices($pBars);

    # Also do vertical line at every break.
    drawVerticalLinesAtBreaks($pRGMap,$pBars,$dimY,$xlo,$xhi,$sx,$im);

    $im->fontsize(3);

    # Plot each bar in the image
    foreach my $pB (@$pBars) {
	plotBar($pB,$pStrColIdx,$sx,$sy,$xlo,$xhi,$im);
    }

    drawStrainLabels($pLevel,$sy,$pStrColIdx,$im);

    writeImage($im,$outFile);

}

# Each $bars[$i] is [level,iter,str,read,chr,direction,@spans]

sub decideStrainColourIndices {
    my ($pBars) = @_;
    my ($nSCI,%strColIdx)=(0);
    foreach my $pB (@$pBars) {
	my $str = $pB->[2];
	unless (defined $strColIdx{$str}) {
	    $strColIdx{$str} = $nSCI;
	    $nSCI++;
	}
    }
    return \%strColIdx;
}

sub drawVerticalLinesAtBreaks {
    my ($pRGMap,$pBars,$dimY,$xlo,$xhi,$sx,$im) = @_;
    my $yOff = 27;
    my $prevDXX;
    foreach my $pB (@$pBars) {
	next if ($pB->[2] ne 0); # only do at ref breaks
	my @spans = @{$pB}[6..$#$pB];
	for (my $i=1; $i<$#spans; $i++) {
	    my $x = $spans[$i];
	    if ($x > $xlo && $x < $xhi) {
		my $cr = looksLikeBreakpoint($pBars,$x) ? 'red' : 'gray';
		$im->fgcolor($cr);
		my $xx = ($x-$xlo)*$sx+100;
		$im->moveTo($xx,0);
		$im->lineTo($xx,$dimY);
		if ($cr eq 'red' && $xx > $prevDXX) {
		    $im->moveTo($xx-2,$dimY+15);
		    $im->string('D');
		    if ($prevDXX > 0 && $xx - $prevDXX < 30) {
			$yOff = 39;
		    } else {
			$yOff = 27;
		    }
		    $prevDXX = $xx;
		    $im->moveTo($xx-10,$dimY+$yOff);
		    $im->string(referencePosition($x,$pRGMap));
		}
	    }
	}
    }
}

# $pB is [level,iter,str,read,chr,direction,@spans]

sub plotBar {
    my ($pB,$pStrColIdx,$sx,$sy,$xlo,$xhi,$im) = @_;
    my ($y,$iter,$str,$name,undef,$dir,@spans) = @$pB;
    my $suf = substr($name,length($str)+1);
    $suf =~ s/1k$//;
    my $cr = $COLOURS[$pStrColIdx->{$str}%scalar(@COLOURS)];
    my $solid = $str eq 0;
    $im->bgcolor($solid ? $cr : 'white');
    $im->fgcolor($cr);
    my $yy = $y*$sy;
    my $iAnno = spanToAnnotate(\@spans,$sx);
    my $iMax = $#spans-1;
    for (my $i=0; $i<=$iMax; $i+=2) {
	plotSpan($i,$iAnno,$iMax,$solid,$cr,$dir,$iter,$suf,
		 @spans[$i..$i+2],$sx,$sy,$xlo,$xhi,$yy,$im);
    }
}

sub plotSpan {
    my ($i,$iAnno,$iMax,$solid,$cr,$dir,$iter,$suf,
	$lo,$hi,$next,$sx,$sy,$xlo,$xhi,$yy,$im) = @_;
    my $xx1 = max(0,$lo-$xlo)*$sx+101;
    my $xx2 = (min($xhi,$hi)-$xlo)*$sx+99;
    $xx2 = max($xx1,$xx2);
    # For $i=0 and $dir<0, we want this rectangle to
    # point to left; for $i==$#spans-1 and $dir>0,
    # to point to the right.
    my $dir0 = 0; # instead of $dir, to save time
    addPointingRectangle($xx1,$yy,$xx2,$yy+$sy*0.8,$i,$iMax,$dir0,$im);
    if ($i == $iAnno) {
	printSuffixInsideRead($i,$solid,$cr,$dir,$suf,$xx1,$yy,$sy,$im);
    }
    if ($next > $xlo && $next < $xhi+30000) { # 30k is max break size...
	$im->moveTo(max(0,($hi-$xlo)*$sx)+100,$yy+$sy*0.4);
	$im->lineTo(($next-$xlo)*$sx+100,$yy+$sy*0.4);
    }
}

sub drawStrainLabels {
    my ($pLevel,$sy,$pStrColIdx,$im) = @_;
    $im->fgcolor('black');
    $im->bgcolor('black');
    foreach my $k (0 .. $#$pLevel) {
	my ($top,$iter,$str) = @{$pLevel->[$k]};
	my $mid = ($top + ($k == 0 ? 0 : $pLevel->[$k-1][0]))/2;
	my $cr = $COLOURS[$pStrColIdx->{$str}%scalar(@COLOURS)];
	$im->bgcolor($cr);
	$im->fgcolor($cr);
	#$im->rectangle(5,$mid*$sy-7,10,$mid*$sy-2);
	$im->moveTo(5,$mid*$sy+2); # 5 shd be 15 with rectangle
	$im->string($str ? "$iter:$str" : 'REF');
    }
}

sub printSuffixInsideRead {
    my ($i,$solid,$cr,$dir,$suf,$xx1,$yy,$sy,$im) = @_;
    # Print suf part of name inside read
    $im->fgcolor($solid ? 'white' : $cr);
    my $xa = $xx1 + ($i == 0 && $dir < 0 ? 5 : 2);
    $im->moveTo($xx1+2,$yy+$sy*0.9);
    my $anno = $suf;
    if ($suf =~ /p$/) {
	$anno =~ tr/a-z/A-Z/;
    }
    $anno =~ s/\.[Pq]$//;
    if ($dir > 0) {
	$anno .= '>';
    } elsif ($dir < 0) {
	$anno = "<$anno";
    }
    $im->string($anno);
    $im->fgcolor($cr);
}

sub writeImage {
    my ($im,$outFile) = @_;
    my $gh = openToWrite($outFile);
    binmode $gh;
    print $gh $im->png;
    $gh->close();
}

sub referencePosition {
    my ($x,$pRGMap) = @_;
    $x--; # to switch to zero-offset
    foreach my $i (0 .. $#$pRGMap) {
	my ($from,$to,$len) = @{$pRGMap->[$i]};
	if ($to <= $x && $to+$len >= $x) {
	    return 1+$from+$x-$to;
	} elsif ($to > $x) {
	    return 1+$from;
	}
    }
    die("Cannot resolve position $x");
}
	
sub looksLikeBreakpoint {
    my ($pBars,$x) = @_;
    foreach my $pB (@$pBars) {
	my @spans = @{$pB}[6..$#$pB];
	for (my $i=0; $i<$#spans; $i+=2) {
	    return 0 if ($spans[$i] < $x && $spans[$i+1] > $x);
	}
    }
    return 1;
}


sub addPointingRectangle {
    my ($xx1,$yy1,$xx2,$yy2,$i,$iMax,$dir,$im) = @_;
    if ($i == 0 && $dir < 0) {
	# point to left
	my $poly = new GD::Polygon;
	my $h2 = int(($yy2-$yy1)/2);
	$poly->addPt($xx1+$h2,$yy1);
	$poly->addPt($xx2,$yy1);
	$poly->addPt($xx2,$yy2);
	$poly->addPt($xx1+$h2,$yy2);
	$poly->addPt($xx1,$yy1+$h2);
	$im->polygon($poly);
    } elsif ($i == $iMax && $dir > 0) {
	# point to right
	my $poly = new GD::Polygon;
	my $h2 = int(($yy2-$yy1)/2);
	$poly->addPt($xx1,$yy1);
	$poly->addPt($xx2-$h2,$yy1);
	$poly->addPt($xx2,$yy1+$h2);
	$poly->addPt($xx2-$h2,$yy2);
	$poly->addPt($xx1,$yy2);
	$im->polygon($poly);
    } else {
	$im->rectangle($xx1,$yy1,$xx2,$yy2);
    }
}

sub spanToAnnotate {
    my ($pSpans,$sx) = @_;
    my ($iAnno,$maxLen) = (-1,5/$sx);
    for (my $i=0; $i<$#$pSpans; $i+=2) {
	my $len = $pSpans->[$i+1] - $pSpans->[$i] + 1;
	if ($len > $maxLen) {
	    $maxLen = $len;
	    $iAnno = $i;
	}
    }
    return $iAnno;
}
