#!/usr/bin/perl

# Input: strain name, genome/supercontig/contig/read fasta file(s)
# Output: clip.fastq, in current directory.
#
# Use on external genomes. For example:
#
# YJM789: genome.fa contains everything (the other .fa files are
# the individual contigs). Copy genome.fa to $SGR/read/YJM789, then
# call genomeToReads.pl YJM789 genome.fa
#
# RM11_1A: use both of
# saccharomyces_cerevisiae_rm11-1a_1_excluded_reads.fasta.gz
# saccharomyces_cerevisiae_rm11-1a_1_supercontigs.fasta.gz

use Utility;
use Getopt::Std;
use strict;

my $READ_LENGTH = 1000;
my $INSERT_SIZE = 4500;
my $STAGGER = 1000; # we will get every 500k because of the q reads

if (@ARGV >= 2) {
    genomeToReads(@ARGV);
} else {
    die("Usage: $0 strName *.fa");
}

sub genomeToReads {
    my ($str,@files) = @_;

    my ($cname,$buf);
    my $loc = "1a01";

    my $fq = "clip.fastq";
    if (-s $fq) {
	report("$fq: already exists",1);
	return;
    }

    my $gh = openToWrite($fq);

    foreach my $f (@files) {
	my $fh = openToRead($f);
	while (<$fh>) {
	    chomp;
	    if (/^>(\S+)/) {
		if ($buf) {
		    $loc = processBuffer($cname,$buf,$loc,$str,$gh);
		    $buf = "";
		}
		$cname = $1;
	    } else {
		$buf .= $_;
	    }
	}
	processBuffer($cname,$buf,$loc,$str,$gh) if ($buf);
    }
    $gh->close();
}

sub processBuffer {
    my ($cname,$buf,$loc,$str,$gh) = @_;
    my $len = length($buf);
    for (my $i=0; $i<$len; $i+=$STAGGER) {
	my $pseq = substr($buf,$i,$READ_LENGTH);
	my $iq = $i+$INSERT_SIZE-$READ_LENGTH;
	my $qseq = reverse(substr($buf,$iq,$READ_LENGTH));
	$qseq =~ tr/ACGT/TGCA/;
	printRead($loc,'p',$cname,$i,$pseq,$str,$gh);
	printRead($loc,'q',$cname,$iq,$qseq,$str,$gh);
	$loc = successorLocation($loc);
    }
    return $loc;
}

sub printRead {
    my ($loc,$pq,$cname,$i,$seq,$str,$gh) = @_;
    return unless ($seq);
    my $len = length($seq);
    my $name = sprintf("%s-%s.%s1k",$str,$loc,$pq);
    printf($gh "@%s bases 1 to %d %s:%d:%d\n",
	   $name,$len,$cname,$i+1,$i+$len);
    print $gh $seq;
    print $gh "\n";
    printf($gh "+%s bases 1 to %d %s:%d:%d\n",
	   $name,$len,$cname,$i+1,$i+$len);
    print $gh ('I' x $len);
    print $gh "\n";
}

sub successorLocation {
    my ($loc) = @_;
    my ($plate,$x,$y) = ($loc =~ /^(\d+)([a-z])(\d+)/);
    if ($y < 24) {
	$y++;
    } else {
	$y=1;
	if ($x lt 'p') {
	    $x = chr(ord($x)+1);
	} else {
	    $x = 'a';
	    $plate ++;
	}
    }
    my $ans = sprintf("%d%s%2d",$plate,$x,$y);
    $ans =~ tr/ /0/;
    return $ans;
}
