#!/usr/bin/env perl

# Run in "final" dir, ONCE only (not too slow, so not worth parallelizing).

# Expect runMaq.pl (no longer solexaAssembly.pl, which is obsolete) to
# have written solexa/STR/all.fastq.gz for each strain to be
# incorporated. Formerly I ran runMaq.pl, and then solexaAssembly.pl
# on each of the all.mapview.gz files.

# See ~/th/2007/1127 for full instructions.

use Utility;
use File::Basename;
use strict;

autoflush STDOUT 1;

foreach my $f (glob("solexa/*.fastq.gz")) {
    createSolexaAlignments($f);
}
foreach my $type (qw(sorted)) { # could do imputed too
    foreach my $d (glob("p?/*/solexa")) {
	incorporateSolexa($d,$type);
    }
}

# Create files that line up nicely beside each imputed_y.gz and therefore sorted_y.gz.

sub createSolexaAlignments {
    my ($f) = @_;
    my $str = basename($f,'.fastq.gz');
    $str =~ tr/-./_/; # Solexa strain
    my $need = 0;
    foreach my $d (glob("p?/*")) {
	unless (-s "$d/solexa/$str.gz") {
	    $need = 1;
	    last;
	}
    }
    unless ($need) {
	report("$str: already created per-chromosome files",1);
	return;
    }
    my $fromStr; # strain against which $str was matched
    my $fh = openToRead($f);
    while (1) {
	my $pQuad = readQuad($fh);
	last unless ($pQuad);
	my @arr = split(/\./,substr($pQuad->[0],1));
	my ($fromStr,$chr);
	if (@arr == 1) {
	    $fromStr = 'REF';
	    $chr = $arr[0];
	} elsif (@arr == 2) {
	    ($fromStr,$chr) = @arr;
	} else {
	    die("Bad chromosome name, not 1 or 2 components: @arr");
	}
	next if ($chr =~ /^vc/); # vector -- ignore
	my ($chrDir) = glob("p?/$chr");
	die("Cannot find dir for $chr") unless ($chrDir);
	my $ih = openToRead("$chrDir/imputed_y.gz");
	my $col = decideColumn($fromStr,$ih);
	my $g = "$chrDir/solexa/$str.gz";
	my $gh = openToWrite($g);
	report("$g: writing");
	print $gh ">$str\n"; # header just so that lines match
	# Process the line
	my $j=0;
	my $nucs = $pQuad->[1];
	my $quals = $pQuad->[3];
	my $jLim = length($nucs);
	while ($j < $jLim) {
	    my $line = <$ih> || die("Expected line to match $pQuad->[0]");
	    my $v = substr($line,$col,1);
	    my $pair;
	    if ($v =~ /[A-Za-z]/) {
		# then use a nucleotide
		$pair = substr($nucs,$j,1) . substr($quals,$j,1);
		$j++;
	    } else {
		$pair = substr($line,$col,2);
	    }
	    print $gh "$pair\n";
	}
	# Check nothing left over:
	while (<$ih>) {
	    my $v = substr($_,$col,1);
	    if ($v =~ /[A-Za-z]/) {
		die("Extra line in $chrDir/imputed_y.gz: $_");
	    }
	}
	$gh->close();
	report();
    }
}

sub decideColumn {
    my ($str,$fh) = @_;
    my $hdr = <$fh>;
    return 0 if ($str eq 'REF');
    chomp $hdr;
    my @sList = split(" ",$hdr);
    foreach my $i (0 .. $#sList) {
	return 2*($i+1) if ($sList[$i] eq $str);
    }
    die("Cannot find strain $str in $hdr");
}

sub incorporateSolexa {
    my ($solDir,$type) = @_;
    my $dir = dirname($solDir);
    my @solFiles = glob("$solDir/*.gz");
    my $fh = openToRead("$dir/${type}_y.gz");
    my $g = "$dir/${type}_z.gz";
    unless (stamp($g)) {
	report("$g: cannot stamp",1);
	return;
    }
    report("$g: creating");
    my $gh = openToWrite($g);
    chomp (my $hdr = <$fh>);
    my @strains1 = split(" ",substr($hdr,1));
    my $hasMRCA = $strains1[-1] eq 'MRCA';
    pop @strains1 if ($hasMRCA);
    my %isOrigStrain = map {$_ => 1} @strains1;
    my @strains2 = @strains1;
    my @in;
    foreach my $f (@solFiles) {
	my $s = basename($f,'.gz');
	push @strains2,$s unless ($isOrigStrain{$s});
	my $ih = openToRead($f);
	<$ih>; # for (redundant) header
	push @in,$ih;
    }
    @strains2 = sort @strains2;
    push @strains2,'MRCA' if ($hasMRCA);
    print $gh ">@strains2\n";
    my %strains2Index;
    foreach my $i (0 .. $#strains2) {
	$strains2Index{$strains2[$i]} = $i;
    }
    my @colCopy = (0); # to copy ref
    foreach my $i (0 .. $#strains1) {
	$colCopy[2*($i+1)] = 2*($strains2Index{$strains1[$i]}+1);
    }
    my @colMerge;
    foreach my $i (0 .. $#solFiles) {
	my $s = basename($solFiles[$i],'.gz');
	$colMerge[$i] = 2*($strains2Index{$s}+1);
    }
    my $lineLen = 2*(1+scalar(@strains2));
    while (<$fh>) {
	chomp;
	my $line = ' ' x $lineLen; # output line
	# Copy from ${type}_y.gz file
	for (my $i=0; $i<=$#colCopy; $i+=2) {
	    substr($line,$colCopy[$i],2) = substr($_,$i,2);
	}
	# Merge in Solexa info
	foreach my $i (0 .. $#in) {
	    my $ih = $in[$i];
	    chomp (my $pair = <$ih>);
	    if ($pair =~ /^[ACGT]/ && substr($pair,1,1) ne '!') { # i.e. not gap or N, or worthless
		my $j = $colMerge[$i];
		substr($line,$j,2) = mergePairs(substr($line,$j,2),$pair);
	    }
	}
	# Change lower case and "!" to "= " for reference, to avoid confusing
	# people with "ghost" transposons.
	$line =~ s/^[a-z]./= /;
	print $gh "$line\n";
    }
    $gh->close();
    unstamp($g);
    report();
}

sub mergePairs {
    my ($pair1,$pair2) = @_;
    if ($pair1 eq '  ') {
	return $pair2;
    } elsif ($pair2 eq '  ') {
	return $pair1;
    } else {
	my $v1 = substr($pair1,0,1);
	my $v2 = substr($pair2,0,1);
	my $q1 = substr($pair1,1,1);
	my $q2 = substr($pair2,1,1);
	my $u1 = $v1;
	my $u2 = $v2;
	$u1 =~ tr/a-z/A-Z/;
	$u2 =~ tr/a-z/A-Z/;
	if (geQual($q1,$q2)) {
	    my $q = ($u1 eq $u2) ? $q1 : diffQual($q1,$q2);
	    return "$v1$q";
	} else {
	    my $q = ($u1 eq $u2) ? $q2 : diffQual($q2,$q1);
	    return "$v2$q"; # could be upper case
	}
    }
}

sub maxQual {
    my ($q1,$q2) = @_;
    return geQual($q1,$q2) ? $q1 : $q2;
}

sub diffQual {
    my ($q1,$q2) = @_;
    my $n1 = $q1 eq ' ' ? 127 : ord($q1);
    my $n2 = $q2 eq ' ' ? 127 : ord($q2);
    my $code = $n1-$n2+33;
    return $code == 127 ? ' ' : chr($code);
}

sub geQual {
    my ($q1,$q2) = @_;
    return $q1 eq ' ' ? 1 : $q2 eq ' ' ? 0 : $q1 ge $q2;
}
