#!/usr/bin/env perl

use Utility;
use File::Basename;
use strict;

# Arg: genome.gff file, then sorted_y.gz file(s) created with
# $WIPE_UNLIKELY_INDELS = 0 in Imputer.pm

# Output (to stdout): LOCI T N, where T is 0 (intergenic)
# D (dubious), U (unverified) or V (verified); then lines of form
#
# chr offset avScore len [i|d] T
#
# where avScore is average score of minority value (nuc or gap)
# over region of length n, inserting/deleting, with gene type = T.

my $gff = shift @ARGV;
my %wanted;
foreach my $f (@ARGV) {
    $wanted{basename(dirname($f))} = 1;
}
my %cds;
my $fh = openToRead($gff);
while (<$fh>) {
    chomp;
    my @a = split;
    if ($a[2] eq 'CDS') {
	my $str = $a[-1];
	$str =~ s/.*=//;
	my $type = substr($str,0,1);
	# Mark as dubious unless it has a classification
	$type = 'D' unless ($type eq 'V' || $type eq 'U' || $type eq 'D');
	$a[0] = 'chrm' if ($a[0] eq 'chrM');
	if ($wanted{$a[0]}) {
	    push @{$cds{$a[0]}},[$a[3],$a[4],$type];
	}
    }
}
my %sumLen;
foreach my $pA (values %cds) {
    my $prevHi = 0;
    foreach my $pT (sort {$a->[0] <=> $b->[0]} @$pA) {
	my ($lo,$hi,$type) = @$pT;
	my $effLo = max($lo,$prevHi+1);
	if ($effLo <= $hi) {
	    $sumLen{$type} += $hi-$effLo+1;
	}
	if ($lo > $prevHi) {
	    $sumLen{0} += $lo-$prevHi-1;
	}
	$prevHi = $hi;
    }
}
foreach my $type (sort keys %sumLen) {
    printf("LOCI %s %7d\n",$type,$sumLen{$type});
}

foreach my $f (@ARGV) {
    indelEvidence($f);
}

sub indelEvidence {
    my ($f) = @_;
    my $fh = openToRead($f);
    my $ln = 0;
    my $locs = 0;
    my $chr = basename(dirname($f));
    my $pCDS = $cds{$chr};
    <$fh>; # header
    my ($oldID,$sumEv,$len);
    while (<$fh>) {
	next if (/^=/);
	$locs ++;
	$ln ++ if (/^[ACGT]/);
	my ($id,$ev) = indelEvidence1($_);
	if ($id eq $oldID) {
	    if ($oldID) {
		$sumEv += $ev;
		$len ++;
	    }
	} else {
	    if ($oldID) {
		printf("%s %7d %6.3f %d %s %s\n",$chr,$ln-1,$sumEv/$len,$len,$oldID,inCDS($ln-1,$pCDS));
	    }
	    $oldID = $id;
	    if ($ev) {
		$sumEv = $ev;
		$len = 1;
	    } else {
		$sumEv = 0;
		$len = 0;
	    }
	}
    }
}

sub inCDS {
    my ($ln,$pCDS) = @_;
    foreach my $pA (@$pCDS) {
	my ($lo,$hi,$type) = @$pA;
	last if ($lo > $ln);
	next if ($hi < $ln);
	return $type;
    }
    return 0;
}

sub indelEvidence1 {
    my ($line) = @_;
    my $iMax = length($line)-2;
    my $ev = 0;
    if (/^=/) {
	# nothing
    } elsif (/^[-0-9_]/) {
	my $nSeq = 0;
	for (my $i=2; $i<=$iMax; $i+=2) {
	    my $v = substr($line,$i,1);
	    if ($v =~ /[ACGTacgt]/) {
		$ev += ord(substr($line,$i+1,1))-33;
	    }
	    $nSeq ++ unless ($v eq ' ' || $v eq '=');
	}
	return undef unless ($ev);
	return ('i',$ev/$nSeq);
    } else {
	my $nSeq = 0;
	for (my $i=2; $i<=$iMax; $i+=2) {
	    my $v = substr($line,$i,1);
	    if ($v =~ /[-_0-9]/) {
		my $qc =  substr($line,$i+1,1);
		$ev += ($qc eq ' ' ? 127 : ord($qc))-33;
	    }
	    $nSeq ++ unless ($v eq ' ' || $v eq '=');
	}
	return undef unless ($ev);
	return ('d',$ev/$nSeq);
    }
}
