#!/usr/bin/env perl

# Run on output of indelEvidence.pl.

use Utility;
use strict;

# indelEvidencePost.pl 1 15 0.2 indelEvidence.txt > iep1.txt
# indelEvidencePost.pl 2 10 0.2 indelEvidence.txt > iep2.txt
# indelEvidencePost.pl 3 10 0.2 indelEvidence.txt > iep3.txt

# for n in 1 2 3
#   do
#   awk '{if (NF == 24) print ($1+$3)/2,$22}' iep$n.txt > iep${n}_0.xy
#   awk '{if (NF == 24) print ($1+$3)/2,$24}' iep$n.txt > iep${n}_V.xy
#   done
# indelEvidencePlotFit.pl

# where we previously did

# indelEvidence.pl /lustre/scratch1/sanger/dmc/sgroot/ref/cere/genome.gff \
#   p?/*/sorted_y.gz > indelEvidence.txt
# This gives posterior prob (true indel) in Post0 and PostV columns for
# intergenic and verified respectively.

# Suggests p(true) is roughly linear (in fact a bit asymptotic) in these ranges:
# Len   Intergenic     Verified    Below intergenic lo  At or above
#  1    1.2 to 7.0   1.2 to 18.0        740474             85327
#  2    0.6 to 4.0   0.8 to 10.0         85622             31520
#  3    0.0 to 3.0   0.0 to 10.0             0             24729
#                                Sum    826096            141486
# So we can weed out 85.4% of apparent indel locations (counts above
# are for locations not indels, i.e. multiplied by Len; they are a slight
# underestimate as "3" includes > 3.

collectConfidentCounts(@ARGV);

# Suppose true intergenic indel rate is pi and true verified-CDS rate
# is pv, and false positive rate in current bucket is pf. Then we expect
# to see pi+pf in intergenic and pv+pf in verified. Assume pf[15] = 0.
# Then we have pi/pv = ratio[15]{0} = rt (known). Given rb = ratio[b]{0},
# we have (pi+pf)/(pv+pf) = rb = (rt+pf/pv)/(1+pf/pv). Setting rf = pf/pv,
# we have rb = (rt+rf)/(1+rf), so rb + rb*rf = rt+rf, so
# rt-rb = (rb-1)rf, so rf = (rt-rb)/(rb-1).
#
# Posterior prob (true pos) in verified region is then pv/(pv+pf) = 1/(1+rf),
# and in intergenic region, pi/(pi+pf) = pv*rt/(pv*rt+pf) = rt/(rt+rf).

# This suggests there are virtually no true indels with q<30, and that 
# q>=100 are pretty trustworthy. Over the non-chrm genome as a whole, 13.3% of
# indel locations have q >= 30, and 3.7% have q >= 100. We should create
# {sorted,imputed,layout}_z.gz with only q>=30 indels.

sub collectConfidentCounts {
    my ($indelSize,$qMax,$bSz,@files) = @_;
    my (@c,%locusC);
    foreach my $f (@files) {
	my $fh = openToRead($f);
	while (<$fh>) {
	    chomp;
	    my @a = split;
	    if ($a[0] =~ /^chr/) {
		my ($chr,undef,$ev,$size,$id,$regType) = @a;
		if (min($size,3) == $indelSize) {
		    my $b = min(int($ev/$bSz),$qMax/$bSz); # quality bucket
		    $c[$b]{$regType} ++;
		}
	    } elsif ($a[0] eq 'LOCI') {
		$locusC{$a[1]} += $a[2];
	    }
	}
    }
    # $ratio[b]{t} is ratio of density of indels in given bucket
    # over regions of type t vs regions of type V, allowing for
    # overall locus counts of those types. We believe it for b=15.
    my @ratio;
    foreach my $b (reverse(0 .. $qMax/$bSz)) {
	my $pC = $c[$b];
	printf("%5.1f to %5.1f",$b*$bSz,($b+1)*$bSz);
	foreach my $type (qw(V U D 0)) {
	    my $n = $pC->{$type};
	    my $nv = $pC->{V};
	    printf(" | %s %6d ",$type,$n);
	    if ($nv && $locusC{$type}) {
		my $r = ($n*$locusC{V})/($nv*$locusC{$type});
		$ratio[$b]{$type} = $r;
		printf("%8.4f",$r);
	    } else {
		printf("%8s","");
	    }
	}

	my $rb = $ratio[$b]{0};
	my $rt = $ratio[$qMax/$bSz]{0};
	if (defined $rt && defined $rb && $rb > 1) {
	    my $rf = ($rt-$rb)/($rb-1);
	    my $postv = 1/(1+$rf);
	    my $post0 = $rt/($rt+$rf);
	    printf(" | Post0 %8.6f PostV %8.6f\n",
		   max(0,min(1,$post0)),
		   max(0,min(1,$postv)));
	} else {
	    print "\n";
	}
    }
}
