#!/usr/bin/env perl

# Run in release "match" dir. Args:
# [-g GFFFile] [-o outFile] inFiles.
# Default outFile: indelSpectrum.txt
# Default inFiles: */imputed.gz (but works on sequenced.gz files too)
# GFFFile, if present, should be in alignment coordinates,
# e.g. ../ref/genomeAligned.gff

use Utility;
use Getopt::Std;
use File::Basename;
use strict;

my %OPT;
getopts('g:o:',\%OPT);

unless ($OPT{g}) {
    print STDERR "WARNING: -g gffFile not specified\n";
}

indelSpectrum(\%OPT,@ARGV);

sub indelSpectrum {
    my ($pOpt,@inFiles) = @_;
    my $gffFile = $pOpt->{g};
    my $outFile = $pOpt->{o} || "indelSpectrum.txt";
    @inFiles = glob("*/imputed.gz") unless (@inFiles);

    my $pOmit;
    $pOmit = regionsToOmit($gffFile) if ($gffFile);
    my $gh = openToWrite($outFile);
    my %count;
    foreach my $f (@inFiles) {
	report("$f: processing");
	my $chr = basename(dirname($f));
	my $pOmitChr;
	$pOmitChr = $pOmit->{$chr} if ($pOmit);
	augmentIndelSpectrum(openToRead($f),$chr,$pOmitChr,\%count,$gh);
	report();
    }
    my @stateNames = qw(mch ins del);
    report("$outFile: writing");
    foreach my $str (sort keys %count) {
	foreach my $s (0 .. 2) {
	    my $pH = $count{$str}[$s];
	    my @lens = sort {$a <=> $b} keys %$pH;
	    my ($sum1,$sum2);
	    foreach my $len (keys %$pH) {
		my $v = $pH->{$len};
		$sum1 += $v;
		$sum2 += $len*$v;
	    }
	    my ($prop1,$prop2);
	    foreach my $len (@lens) {
		$prop1 += $pH->{$len}/$sum1;
		$prop2 += $len*$pH->{$len}/$sum2;
		printf($gh "%s %s %5d %5d %8.6f %7d %8.6f\n",
		       $str,$stateNames[$s],$len,$pH->{$len},$prop1,
		       $pH->{$len}*$len,$prop2);
	    }
	}
    }
    report();
}

sub augmentIndelSpectrum {
    my ($fh,$chr,$pOmit,$pCount,$gh) = @_;
    chomp (my $hdr = <$fh>);
    my @strains = split(" ",substr($hdr,1));
    pop @strains if ($strains[-1] eq 'MRCA'); # not of interest
    my $colMax = 2*scalar(@strains);
    # $state[$col] is:
    #   -- -1 if in middle of space (inter-read) sequence
    #   -- 0 if both nucs ("mch" = match)
    #   -- 1 if ref only is gap ("ins"; and then $insertionSequence[$col] is filled)
    #   -- 2 if col only is gap ("del")
    # $len[$col] is length in current state
    my ($prev,$rPrev,@state,@len,@insertionSequence);
    my %isGap = map {$_ => 1} qw(- = _ 1 2 3 4 5);
    my @count;
    my $off = 0; # offset in alignment, from 1
    my $roff = 1; # offset in reference, from 1
    while (<$fh>) {
	chomp;
	$off ++;
	while ($pOmit && @$pOmit && $pOmit->[0][1] < $off) {
	    shift @$pOmit;
	}
	next if ($pOmit && @$pOmit && $pOmit->[0][0] <= $off && $pOmit->[0][1] >= $off);
	my $rGap = $isGap{substr($_,0,1)};
	for (my $col=2; $col<=$colMax; $col+=2) {
	    my $v = substr($_,$col,1);
	    my $vGap = $isGap{$v};
	    next if ($rGap && $vGap); # suspended
	    my $s = $state[$col];
	    my $l = $len[$col];
	    if ($v eq ' ' or $v eq 'N') { # end of read, for sequenced.gz data; or unknown, for imputed.gz
		if ($s >= 0) {
		    $count[$col][$s]{$l} ++;
		    $state[$col] = -1;
		    purgeInsertionSequence(\@insertionSequence,$col,\@strains,$chr,$roff,$gh);
		}
	    } elsif ($rGap) {
		# then can't be vGap
		$insertionSequence[$col] .= $v;
		if ($s == 1) {
		    $len[$col] ++;
		} else {
		    $count[$col][$s]{$l} ++ if ($s >= 0);
		    $state[$col] = 1;
		    $len[$col] = 1;
		}
	    } elsif ($vGap) {
		purgeInsertionSequence(\@insertionSequence,$col,\@strains,$chr,$roff,$gh);
		# then can't be rGap
		if ($s == 2) {
		    $len[$col] ++;
		} else {
		    $count[$col][$s]{$l} ++ if ($s >= 0);
		    $state[$col] = 2;
		    $len[$col] = 1;
		}
	    } elsif ($s == 0) {
		$len[$col] ++;
	    } else {
		$count[$col][$s]{$l} ++ if ($s >= 0);
		purgeInsertionSequence(\@insertionSequence,$col,\@strains,$chr,$roff,$gh);
		$state[$col] = 0;
		$len[$col] = 1;
	    }
	}    
	$roff ++ unless ($rGap);
    }
    # Add final counts and transfer to hash
    for (my $col=2; $col<=$colMax; $col+=2) {
	$count[$col][$state[$col]]{$len[$col]} ++;
	purgeInsertionSequence(\@insertionSequence,$col,\@strains,$chr,$roff,$gh);
	my $str = $strains[$col/2-1];
	my $pC = $count[$col];
	foreach my $s (0 .. $#$pC) {
	    my $pH = $pC->[$s];
	    delete $pH->{0};
	    delete $pH->{''};
	    foreach my $l (keys %$pH) {
		$pCount->{$str}[$s]{$l} += $pH->{$l};
		$pCount->{'*ALL*'}[$s]{$l} += $pH->{$l};
	    }
	}
    }
}

sub purgeInsertionSequence {
    my ($pA,$col,$pStrains,$chr,$roff,$gh) = @_;
    my $seq = $pA->[$col];
    my $len = length($seq);
    if ($len >= 256) {
	my $str = $pStrains->[int($col/2-1)];
	if ($str eq 'Y55' || $str eq 'SK1') {
	    print $gh "Insertion $str $chr $roff $len $seq\n";
	}
    }
    $pA->[$col] = undef;
}

sub regionsToOmit {
    my ($f) = @_;
    my $fh = openToRead($f);
    my %omit;
    while (<$fh>) {
	my ($chr,undef,$type,$lo,$hi) = split;
	if ($type eq 'chromosome') {
	    push @{$omit{$chr}},([$lo,$lo+30000-1],[$hi-30000+1,$hi]);
	} elsif ($type eq 'repeat_region' || $type eq 'transposable_element') {
	    push @{$omit{$chr}},[$lo,$hi];
	}
    }
    foreach my $chr (keys %omit) {
	$omit{$chr} = sortAndMergeRegions($omit{$chr});
    }
    return \%omit;
}

sub sortAndMergeRegions {
    my ($pA) = @_;
    my @ans;
    foreach my $pReg (sort {$a->[0] <=> $b->[0]} @$pA) {
	if (@ans && $ans[-1][1] >= $pReg->[0]-1) {
	    $ans[-1][1] = max($ans[-1][1],$pReg->[1]);
	} else {
	    push @ans,$pReg;
	}
    }
    return \@ans;
}
