#!/usr/bin/env perl

# Run in 025/align dir

use SGRP::OffsetMap;
use SGRP::SequencedRegionSet;
use Utility;
use File::Basename;
use strict;

insertedSequences();

sub insertedSequences {
    my $gh = openToWrite("insertions.fa");
    my $pat = '*'; # 'BC187';
    my %ins;
    my @strains = qw(REF);
    foreach my $f (glob("$pat/alilines.pass.txt")) {
	my $str = dirname($f);
	report("$str: processing");
	my $pOM = new OffsetMap('?',"../assembly/$str/genome.map");
	my $pSRS = new SequencedRegionSet("../assembly/$str/genome.seqr");
	my $fh = openToRead($f);
	insertedRegionSpecs($fh,$pOM,$pSRS,\%ins);
	report();
	push @strains,$str;
    }
    push @strains,'MRCA';
    my ($pPresent,$pByRead) = keepBestRecords(\%ins);
    my %occ;
    foreach my $str (@strains) {
	my $pSRS = new SequencedRegionSet("../assembly/$str/genome.seqr");
	foreach my $chr (keys %$pPresent) {
	    my $pH = $pPresent->{$chr};
	    foreach my $off (keys %$pH) {
		my $v = defined $pH->{$off}{$str} ? '1' : 
		    ($pSRS && $pSRS->inSequencedRegion($chr,$off,10)) ? '0' : 'U';
		$occ{$chr}{$off} .= $v;
	    }
	}
    }
    my %insLen;
    foreach my $str (@strains) {
	foreach my $rf (glob("$str/renamed.?.fastq")) {
	    writeSequences($str,openToRead($rf),$pByRead,\%occ,$gh,\%insLen);
	}
    }
    foreach my $r (keys %$pByRead) {
	print STDERR "WARNING: could not find fastq entry for $r\n";
    }
    findInsertionLengths(\%insLen);
}

# Result is a hash from read names to lists of insertion specs, where
# an insertion spec is
#   [rlo, rhi, L|R|B, read]
# and off is in alignment coordinates.

sub insertedRegionSpecs {
    my ($fh,$pOM,$pSRS,$pIns) = @_;
    while (<$fh>) {
	chomp;
	my (undef,$score,$read,$chr,$rlo,$rhi,$clo,$chi,$fc,undef,undef,$anno) = split;
	if ($score == 0 && $clo == $chi+1 && $anno =~ /late/) {
	    my $off = $pOM->translation($chr,$chi);
	    (undef,$chr) = split(/\./,$chr); # real chr name only
	    unless ($pSRS->inSequencedRegion($chr,$off,10)) {
		my ($lrb) = ($anno =~ /([LR])/);
		$lrb ||= 'B';
		my $len = $fc eq 'F' ? ($rhi-$rlo+1) : ($rlo-$rhi+1);
		if ($len >= 50) {
		    push @{$pIns->{$chr}{$off}},[$rlo,$rhi,$lrb,$read];
		}
	    }
	}
    }
}

sub keepBestRecords {
    my ($pIns) = @_;
    my (%present,%byRead);
    foreach my $chr (keys %$pIns) {
	my $pH = $pIns->{$chr};
	# %$pH is hash from chr offsets to in insertion specs
	my @offs = sort {$a <=> $b} keys %$pH;
	# First merge values within 10 of each other
	my $i=0;
	while ($i < $#offs) {
	    my $th = $offs[$i]+10;
	    my $j=$i;
	    while ($j+1 < $#offs && $offs[$j+1]<=$th) {
		$j++;
	    }
	    if ($j > $i) {
		# Merge from i to j inclusive
		my $off = int(($offs[$i]+$offs[$j])/2);
		my @list;
		foreach my $k ($i .. $j) {
		    push @list,@{$pH->{$offs[$k]}};
		    delete $pH->{$offs[$k]};
		}
		$pH->{$off} = \@list;
	    }
	    $i = $j+1;
	}
	# Then prune each @list
	foreach my $off (keys %$pH) {
	    my %maxLen;
	    foreach my $pRec (@{$pH->{$off}}) {
		my ($rlo,$rhi,$lrb,$read) = @$pRec;
		($rlo,$rhi) = sort {$a <=> $b} ($rlo,$rhi);
		my $len = $rhi-$rlo+1;
		$maxLen{$lrb} = max($maxLen{$lrb},$len);
		my ($str) = split(/-/,$read);
		$present{$chr}{$off}{$str} = 1;
	    }
	    if ($maxLen{L} > $maxLen{B} || $maxLen{R} > $maxLen{B}) {
		delete $maxLen{B};
	    } elsif ($maxLen{L} <= $maxLen{B} && $maxLen{R} <= $maxLen{B}) {
		delete $maxLen{L};
		delete $maxLen{R};
	    }
	    my @ans;
	    my %seenLRB;
	    foreach my $pRec (@{$pH->{$off}}) {
		my ($rlo,$rhi,$lrb,$read) = @$pRec;
		if (defined $maxLen{$lrb}) {
		    ($rlo,$rhi) = sort {$a <=> $b} ($rlo,$rhi);
		    my $len = $rhi-$rlo+1;
		    if ($len == $maxLen{$lrb} && !$seenLRB{$lrb}) {
			$seenLRB{$lrb} = 1;
			push @ans,$pRec;
			push @{$byRead{$read}},[$rlo,$rhi,$lrb,$chr,$off];
		    }
		}
	    }
	    $pH->{$off} = \@ans;
	}
    }
    return (\%present,\%byRead);
}
		
sub writeSequences {
    my ($str,$fh,$pByRead,$pOcc,$gh,$pInsLen) = @_;
    while (1) {
	my @quad;
	foreach my $i (0 .. 3) {
	    my $line = <$fh>;
	    return unless ($line);
	    push @quad,$line;
	}
	my ($r) = split(" ",substr($quad[0],1));
	next unless (defined $pByRead->{$r});
	foreach my $pRec (@{$pByRead->{$r}}) {
	    my ($rlo,$rhi,$lrb,$chr,$off) = @$pRec;
	    my $seq;
	    if ($rlo < $rhi) {
		$seq = substr($quad[1],$rlo,$rhi-$rlo+1);
	    } else {
		$seq = substr($quad[1],$rhi,$rlo-$rhi+1);
		$seq =~ tr/ACGT/TGCA/;
		$seq = reverse($seq);
	    }
	    print $gh ">MAX.$chr.$off.$lrb $r:$rlo..$rhi $pOcc->{$chr}{$off}\n";
	    for (my $i=0; $i<length($seq); $i+=60) {
		print $gh substr($seq,$i,60) . "\n";
	    }
	    $pInsLen->{$chr}{$off} = max($pInsLen->{$chr}{$off},length($seq));
	}
	delete $pByRead->{$r};
    }
}

sub findInsertionLengths {
    my ($pInsLen) = @_;
    my ($n,$sum);
    foreach my $pH (values %$pInsLen) {
	foreach my $v (values %$pH) {
	    if ($v > 0) {
		$n++;
		$sum += $v;
	    }
	}
    }
    print "Total of $n distinct insertions, total length $sum\n";
}
