#!/usr/bin/env perl

use Utility;
use File::Basename;
use strict;

# Run in iter dir. Creates inseqs/inseqs.fa etc

insertionSequences();

sub insertionSequences {

    my $out1 = 'inseqs/inseqs1.fa';
    my $ss1 = 'inseqs/inseqs1.ss';
    my $out2 = 'inseqs/inseqs2.fa';
    my $ss2 = 'inseqs/inseqs2.ss';
    my $out3 = 'inseqs/inseqs3.fa';
    my $ss3 = 'inseqs/inseqs3.ss';
    my $out4 = 'inseqs/inseqs4.fa';
    my $ss4 = 'inseqs/inseqs4.ss'; # we hope there are no alignments here!
    my $dir = 'inseqs';

    # Find all insertion sequences of length >= 100 in p?/*/imputed_y.gz
    if (-s $out1) {
	report("$out1: already exists",1);
    } else {
	report("$out1: creating");
	unlink($ss1) if (-s $ss1);
	my $gh = openToWrite($out1);
	foreach my $d (glob("p?/*")) {
	    printInsertionSequences($d,$gh);
	}
	report();
    }
    
    # Ssaha2 sequences against each other
    runSsaha2($out1,$out1,$ss1,$out2);

    # Prune out similar but shorter sequences
    if (-s $out2) {
	report("$out2: already exists",1);
    } else {
	report("$out2: creating");
	unlink($ss2) if (-s $ss2);
	pruneInseqsFile($out1,$ss1,$out2);
	report();
    }

    # Ssaha2 surviving sequences against reference
    my $sgr = $ENV{SGR};
    runSsaha2($out2,"$sgr/ref/cere/genome.fa",$ss2,$out3);
    
    # Mask out matched regions
    if (-s $out3) {
	report("$out3: already exists",1);
    } else {
	report("$out3: creating");
	maskOutMatched($out2,$ss2,$out3);
	report();
    }

    # Ssaha2 surviving sequences against reference
    runSsaha2($out3,"$sgr/ref/cere/genome.fa",$ss3,$out4);
    
    # Mask out matched regions
    if (-s $out4) {
	report("$out4: already exists",1);
    } else {
	report("$out4: creating");
	maskOutMatched($out3,$ss3,$out4);
	report();
    }

    # Ssaha2 surviving sequences against reference
    runSsaha2($out4,"$sgr/ref/cere/genome.fa",$ss4);

    if (-d "$dir/00") {
	report("$dir/00: already exists",1);
    } else {
	report("$dir/NN: creating");
	splitFA($out4,$dir);
	report();
    }
}

sub runSsaha2 {
    my ($f1,$f2,$g,$h) = @_;
    if (-s $g) {
	report("$g: already exists",1);
    } else {
	unlink($h) if ($h && -s $h);
	report("$g: creating");
	`ssaha2 $f1 $f2 -tags 1 -align 0 -memory 6000 > $g`;
	report();
    }
}

sub printInsertionSequences {
    my ($dir,$gh) = @_;
    my $chr = basename($dir);
    my $fh = openToRead("$dir/imputed_y.gz");
    my $hdr = <$fh>;
    my $sh = openToRead("$dir/sorted_y.gz"); # so we can spot the Ty insertions
    <$sh>;
    chomp $hdr;
    my @strains = split(" ",substr($hdr,1));
    my $ln = 1;
    my $off = 0; # ref genome offset
    my @ins; # inserted sequence in given position $i for $strains[$i]
    my @extra; # extra sequence after ref resumes; allows us to bridge up to 10 ref positions
    my $since; # no. of posns since ref has resumed
    while (<$fh>) {
	$ln++; # LINE number
	my $sLine = <$sh>;
	if ($sLine =~ /^[-_=0-9]/) {
	    if ($since > 0 && $since < 10) {
		foreach my $i (0 .. $#extra) {
		    $ins[$i] .= $extra[$i] if ($ins[$i] && $extra[$i]);
		}
	    }
	    @extra = ();
	    # ref is gap
	    foreach my $i (0 .. $#strains) {
		my $v = substr($_,($i+1)*2,1);
		$ins[$i] .= $v unless ($v =~ /[-_=]/);
	    }
	    $since = 0;
	} else {
	    $off ++;
	    $since ++;
	    if ($since == 10 && @ins) {
		# then we discharge all the insertions WITHOUT the extras
		foreach my $i (0 .. $#ins) {
		    if (defined $ins[$i]) {
			my $seq = $ins[$i];
			my $len = length($seq);
			if ($len >= 100 && $seq =~ /[ACGT]/) {
			    my $nUpper = 0;
			    for (my $i=0; $i<$len; $i++) {
				$nUpper ++ if (substr($seq,$i,1) =~ /[ACGT]/);
			    }
			    printf($gh ">%s.%d.%d.%s %d %d\n",$chr,$off-10,$ln-10,$strains[$i],$len,$nUpper);
			    for (my $j=0; $j<$len; $j+=60) {
				print $gh substr($seq,$j,60) . "\n";
			    }
			}
		    }
		}
		@ins = ();
	    } elsif ($since < 10) {
		# add to extras ...
		foreach my $i (0 .. $#strains) {
		    my $v = substr($_,($i+1)*2,1);
		    $extra[$i] .= $v unless ($v =~ /[-_=]/);
		}
	    }
	}
    }
    # Discharge at end too
    foreach my $i (0 .. $#ins) {
	if (defined $ins[$i]) {
	    my $seq = $ins[$i];
	    my $len = length($seq);
	    if ($len >= 100 && $seq =~ /[ACGT]/) {
		my $nUpper = 0;
		for (my $i=0; $i<$len; $i++) {
		    $nUpper ++ if (substr($seq,$i,1) =~ /[ACGT]/);
		}
		printf($gh ">%s.%d.%d.%s %d %d\n",$chr,$off-$since,$ln-$since,$strains[$i],$len,$nUpper);
		for (my $j=0; $j<$len; $j+=60) {
		    print $gh substr($seq,$j,60) . "\n";
		}
	    }
	}
    }

}

sub pruneInseqsFile {
    my ($faIn,$ssFile,$faOut) = @_;

    my $fh = openToRead($faIn);
    my $gh = openToWrite($faOut);
    my (%sz,%seq,$name,$len,$uclen);
    while (<$fh>) {
	if (/^>/) {
	    chomp;
	    ($name,$len,$uclen) = split(" ",substr($_,1));
	    $sz{$name} = [$len,$uclen];
	} else {
	    $seq{$name} .= $_;
	}
    }

    $fh = openToRead($ssFile);
    my %remove;
    while (<$fh>) {
	chomp;
	my @a = split;
	if ($a[0] eq 'ALIGNMENT' && $a[9] >= $a[11]*0.9 && $a[10] >= 95) {
	    my ($name1,$name2) = ($a[2],$a[3]);
	    if ($name1 eq $name2) {
		# nothing
	    } elsif (effectivelyLonger($name1,$name2,\%sz)) {
		$remove{$name2} = 1;
	    } else {
		$remove{$name1} = 1;
	    }
	}
    }
    foreach my $name (sort {$sz{$b}[0] >= $sz{$a}[0]} (keys %sz)) {
	unless ($remove{$name}) {
	    printf($gh ">%s %d %d\n%s",$name,@{$sz{$name}},$seq{$name});
	}
    }   
}

sub effectivelyLonger {
    my ($x,$y,$pSz) = @_;
    foreach my $i (0,1) {
	my $d = $pSz->{$x}[$i] - $pSz->{$y}[$i];
	return $d > 0 unless ($d == 0);
    }
    return $x lt $y;
}

sub maskOutMatched {
    my ($faIn,$ss,$faOut) = @_;
    my $fh = openToRead($ss);
    my %mask;
    while (<$fh>) {
	chomp;
	my ($key,undef,$name,undef,$lo,$hi) = split;
	if ($key eq 'ALIGNMENT') {
	    push @{$mask{$name}},[$lo,$hi];
	}
    }
    $fh = openToRead($faIn);
    my $gh = openToWrite($faOut);
    my @maskA;
    my $off;
    my ($hdr,$seq);
    while (<$fh>) {
	chomp;
	if (/^>(\S+)/) {
	    maskAndPrint($hdr,$seq,\@maskA,$gh);
	    @maskA = ();
	    my $name = $1;
	    if ($mask{$name}) {
		foreach my $pPair (sort {$a->[0] <=> $b->[0]} @{$mask{$name}}) {
		    if (@maskA && $maskA[-1][1] >= $pPair->[0]-1) {
			$maskA[-1][1] = max($maskA[-1][1],$pPair->[1]);
		    } else {
			push @maskA,$pPair;
		    }
		}
	    }
	    $hdr = $_;
	    $seq = '';
	} else {
	    $seq .= $_;
	}
    }
    maskAndPrint($hdr,$seq,\@maskA,$gh);
}

sub maskAndPrint {
    my ($hdr,$seq,$pMaskA,$gh) = @_;
    my ($name) = split(" ",substr($hdr,1));
    my $oldHi = 0;
    push @$pMaskA,[length($seq)+1,0];
    foreach my $pA (@$pMaskA) {
	my ($lo,$hi) = @$pA;
	my $len = $lo-$oldHi-1;
	if ($len >= 100) {
	    my $oh1 = $oldHi+1;
	    my $sub = substr($seq,$oldHi,$len);
	    my $sublen = length($sub);
	    print $gh ">$name.$oh1 $sublen\n";
	    for (my $i=0; $i<$sublen; $i+=60) {
		print $gh substr($sub,$i,60) . "\n";
	    }
	}
	$oldHi = $hi;
    }
}

sub splitFA {
    my ($f,$d) = @_;
    my $n = 0;
    my $fh = openToRead($f);
    my $gh;
    while (<$fh>) {
	if (/^>(\S+) /) {
	    my $n2 = sprintf("%2d",$n);
	    $n2 =~ tr/ /0/;
	    $gh = openToWrite("$d/$n2/$1");
	    $n ++;
	    $n = 0 if ($n == 20);
	}
	print $gh $_;
    }
}
