#!/usr/bin/env perl

use Utility;
use strict;

my $ROOT = $ENV{SGR};

autoflush STDOUT 1;

liftoverGFF(@ARGV);

# Run this after running stra, so assembly/REF/genome.gff exists.
# OR run with iter 0.

sub liftoverGFF {
    my ($sp,$filt,$set,$iter) = @_;
    my $srcGFF = "$ROOT/ref/$sp/genome.gff";
    my $ptisFile = "$ROOT/ref/$sp/te/$set.ptis2";
    my ($tgtDir,$base);
    if ($iter == 0) {
	$base = 'extended';
	$tgtDir = "$ROOT/ref/$sp";
    } else {
	$base = 'genome';
	my $i3 = sprintf("%3d",$iter);
	$i3 =~ tr/ /0/;
	$tgtDir = "$ROOT/match/$set/$sp/$filt/$i3/assembly/REF";
    }
    my $mapFile = "$tgtDir/genome.map"; # NOT $base.map!
    my $tgtGFF = "$tgtDir/$base.gff";
    my $ok = 1;
    foreach my $f ($srcGFF,$ptisFile,$mapFile) {
	unless (-s $f) {
	    print STDERR "Missing or empty: $f\n";
	    $ok = 0;
	}
    }
    if (-s $tgtGFF) {
	print STDERR "Already exists: $tgtGFF\n";
	$ok = 0;
    }
    liftoverGFF1($srcGFF,$ptisFile,$mapFile,$tgtGFF) if ($ok);
}

sub liftoverGFF1 {
    my ($srcGFF,$ptisFile,$mapFile,$tgtGFF) = @_;
    report("$ptisFile: reading");
    my $pTI = readTransposonInsertions($ptisFile,$srcGFF);
    report();
    report("$mapFile: reading");
    my $pMap = readMap($mapFile);
    report();
    report("$tgtGFF: creating");
    my $fh = openToRead($srcGFF);
    my $gh = openToWrite($tgtGFF);
    while (<$fh>) {
	chomp;
	my @a = split;
	my ($chr,$db,undef,$lo,$hi) = @a;
	# Is it time to print an added TE?
	my $pTIC = $pTI->{$chr};
	while (defined $pTIC && @$pTIC && $pTIC->[0][0] < $lo) {
	    my $pElt = shift @$pTIC;
	    my ($off,$teFC,@teDef) = @$pElt;
	    my @te = ($chr,$db,'transposable_element',
		      # Make sure we get both sides!
		      mapLocation($chr,$off-1,$pMap)+2,
		      mapLocation($chr,$off,$pMap)-1,
		      '.',
		      ($teFC eq 'F') eq ($teDef[6] eq '+') ? '+' : '-',
		      '.',
		      $teDef[-1]);
	    print $gh join("\t",@te) . "\n";
	    #print "DONE @te\n";
	    #printf("pTIC has size %d\n",scalar(@$pTIC));
	}
	$a[3] = mapLocation($chr,$lo,$pMap);
	$a[4] = mapLocation($chr,$hi,$pMap);
	print $gh join("\t",@a) . "\n";
    }
    report();
}

sub readTransposonInsertions {
    my ($f,$srcGFF) = @_;
    my $fh = openToRead($srcGFF);
    my %teRef;
    while (<$fh>) {
	chomp;
	my @a = split;
	if ($a[2] eq 'transposable_element') {
	    $teRef{$a[0]}{$a[3]}{$a[4]} = \@a;
	}
    }
    my $fh = openToRead($f);
    my %ti;
    while (<$fh>) {
	chomp;
	my ($chr,$off,$te,$fc) = split;
	my ($teChr,$teLo,$teHi) = split(/\./,$te);
	my $pA = $teRef{$teChr}{$teLo+0}{$teHi+0};
	$pA->[-1] =~ s/ID=Y/ID=COPY_Y/;
	$pA->[-1] =~ s/Name=Y/Name=COPY_Y/;
	if (defined @$pA) {
	    push @{$ti{$chr}},[$off,$fc,@$pA];
	} else {
	    die("Cannot resolve $te");
	}
    }
    return \%ti;
}

sub readMap {
    my ($f) = @_;
    my $fh = openToRead($f);
    my %map;
    while (<$fh>) {
	chomp;
	my ($chr,$from,$to,$len) = split;
	if ($chr =~ /^REF\./) {
	    push @{$map{substr($chr,4)}},[$from,$to,$len];
	}
    }
    return \%map;
}

sub mapLocation {
    my ($chr,$off,$pMap) = @_;
    $chr =~ tr/A-Z/a-z/; # for chrm
    my $pA = $pMap->{$chr};
    my $kLo = 0; # lowest possible $k
    my $kHi = $#$pA+1; # ONE MORE THAN highest possible $k
    while (1) {
	my $k = int(($kLo+$kHi)/2);
	my $pAK = $pA->[$k];
	my $lo = $pAK->[0];
	my $hi = $lo + $pAK->[2] - 1;
	if ($lo > $off) {
	    $kHi = $k;
	} elsif ($hi < $off && $k < $#$pA) {
	    $kLo = $k+1;
	} else {
	    return $pAK->[1]+$off-$lo;
	}
    }
}
