#!/usr/bin/env perl

use Utility;
use strict;

my $MAX_READ_LENGTH = 1200;

mainAndAuxes(@ARGV);

# mapFile lines are of the form
#   seqName seqLo otherLo length
# where seqName is the sequence being defined
#       seqLo   is the starting point of a chunk (offset 0) in seqName
#       otherLo is the starting point of the corresponding chunk in "the other" sequence
#       length  is the length of the chunk.
# NOTE that when seqName is the main sequence for the strain (strainName.chrName),
# the "other" sequence is the alignment in sorted.gz, imputed.ga etc. But when
# seqName is an auxiliary sequence (strainName.chrName.NNN) the "other" sequence
# is the one for the strain, NOT the alignment. This is possible because the main
# sequence is the maximal one, i.e. we never impute across hard gaps from other strains.

sub mainAndAuxes {
    my ($chr,$idx,$impFile,$faFile,$qualFile,$mapFile) = @_;
    my $fh = openToRead($impFile);
    my $col = $idx*2;
    chomp (my $hdr = <$fh>);
    my @strains = ('REF',split(" ",substr($hdr,1)));
    my $ns = scalar(@strains);
    my $str = $strains[$idx];
    my $gh = openToWrite($faFile);
    my $qh = openToWrite($qualFile);
    my $mh = openToWrite($mapFile);
    my ($pHGRList,$mainSeq,$mainQuals) = deriveMainSequence($str,$ns,$chr,$col,$fh,$gh,$qh,$mh);
    # A valid set of HGRs for an aux sequence is any of length 1 or more
    # such that the sum of internal distances between them is <= $AUX_REGION_LENGTH.
    # We represent an HGRList as a list: [sumIntDist,HGR1,HGR2,...]
    my $pSets = findAllHGRSets($pHGRList);
    $gh->close();
    $qh->close();
    $mh->close();
    $gh = openToAppend($faFile);
    $qh = openToAppend($qualFile);
    $mh = openToAppend($mapFile);
    writeAuxiliarySequences("$str.$chr",$pSets,$mainSeq,$mainQuals,$gh,$qh,$mh);
}

sub deriveMainSequence {
    my ($str,$ns,$chr,$col,$fh,$gh,$qh,$mh) = @_;
    my %seenHardGapRegion;
    print $gh ">$str.$chr\n";
    print $qh ">$str.$chr\n";
    my $offStr = 0;
    my $offAli = 0;
    print $mh "$str.$chr $offStr $offAli ";
    my $active = 1;
    my $runLength = 0;
    my @hardGapStart;
    my ($seq,$quals);
    while (<$fh>) {
	chomp;
	my $nuc = substr($_,$col,1);
	my $isReal = ($nuc =~ /[-=ACGT]/);
	# Search for hard gap starts/ends:
	if ($isReal) {
	    @hardGapStart = ();
	} else {
	    for (my $i=0; $i<$ns; $i++) {
		my $ii = $i*2;
		next if ($ii == $col); # ignore own hard gap
		my $hgs = $hardGapStart[$i];
		if (substr($_,$ii,1) eq '=') {
		    $hardGapStart[$i] = $offStr unless (defined $hgs);
		} elsif (defined $hgs) {
		    $seenHardGapRegion{$hgs}{$offStr-1} = 1;
		    $hardGapStart[$i] = undef;
		}
	    }
	}
	if ($nuc eq '-' || $nuc eq '=') {
	    if ($active) {
		print $mh "$runLength\n";
		$runLength = 0;
		$active = 0;
	    }
	} else {
	    if (!$active) {
		print $mh "$str.$chr $offStr $offAli ";
		$active = 1;
	    }
	    $nuc =~ tr/a-z/A-Z/;
	    print $gh $nuc;
	    my $qual = substr($_,$col+1,1);
	    print $qh $qual;
	    $seq .= $nuc;
	    $quals .= $qual;
	    $offStr ++;
	    $runLength ++;
	    if ($offStr%60 == 0) {
		print $gh "\n";
		print $qh "\n";
	    }
	}
	$offAli ++;
    }
    if ($active) {
	print $mh "$runLength\n";
    }
    if ($offStr%60) {
	print $gh "\n";
	print $qh "\n";
    }
    for (my $i=0; $i<$ns; $i++) {
	my $ii = $i*2;
	next if ($ii == $col); # ignore own hard gap
	my $hgs = $hardGapStart[$i];
	if (defined $hgs) {
	    $seenHardGapRegion{$hgs}{$offAli-1} = 1;
	}
    }
    my @hardGapRegions;
    foreach my $lo (sort {$a <=> $b} keys %seenHardGapRegion) {
	foreach my $hi (sort {$a <=> $b} keys %{$seenHardGapRegion{$lo}}) {
	    push @hardGapRegions,[$lo,$hi];
	}
    }
    return (\@hardGapRegions,$seq,$quals);
}

# For each existing set in $pBuilding:
#  -- find its SID if we add $pHGRList->[$i];
#       if OK, add extended version to $pBuilding unless overlap,
#       else add current version to $pHGRSets and
#       delete it from $pBuilding.
# Also form a new set from $pHGRList->[$i] and add it
#   to pBuilding.
# At end: add everything in $pBuilding to $pHGRSets

sub hgrSetOrderp {
    my $iMax = min($#$a,$#$b);
    foreach my $i (1 .. $iMax) {
	foreach my $j (0,1) {
	    my $cmp = $a->[$i][$j] <=> $b->[$i][$j];
	    return $cmp if ($cmp);
	}
    }
    return $#$a <=> $#$b;
}

sub findAllHGRSets {
    my ($pHGRList) = @_;
    my (@building,@done);
    foreach my $pHGR (@$pHGRList) {
	my @newBuilding = ([0,$pHGR]);
	foreach my $pSet (@building) {
	    my $sid = $pSet->[0] + ($pHGR->[0] - $pSet->[-1][1] - 1);
	    if ($sid < $MAX_READ_LENGTH) {
		# Keep pSet on the go in both forms
		push @newBuilding,$pSet;
		if ($sid >= 0) {
		    my @newSet = @$pSet;
		    push @newSet,$pHGR;
		    $newSet[0] = $sid;
		    push @newBuilding,\@newSet;
		}
	    } else {
		push @done,$pSet;
	    }
	}
	@building = @newBuilding;
    }
    push @done,@building;
    @done = sort {hgrSetOrderp($a,$b)} @done;
    return \@done;
}

sub writeAuxiliarySequences {
    my ($pfx,$pSets,$mainSeq,$mainQuals,$gh,$qh,$mh) = @_;
    foreach my $i (0 .. $#$pSets) {
	writeAuxiliarySequence($pfx,$i+1,$pSets->[$i],$mainSeq,$mainQuals,$gh,$qh,$mh);
    }
}

sub writeAuxiliarySequence {
    my ($pfx,$idx,$pSet,$mainSeq,$mainQuals,$gh,$qh,$mh) = @_;
    my ($sid,@pairs) = @$pSet;
    my $pad = $MAX_READ_LENGTH - $sid;
    my $preLen = min($pairs[0][0],$pad);
    my ($seq,$quals);
    my $auxId = sprintf("%s.%3d",$pfx,$idx);
    $auxId =~ tr/ /0/;
    my $auxLen = 0;
    addChunk($auxId,$mainSeq,$mainQuals,
	     $pairs[0][0]-$preLen,
	     $preLen,
	     \$seq,\$quals,\$auxLen,$mh);
    for (my $i=0; $i<$#pairs; $i++) {
	my $pPair1 = $pairs[$i];
	my $pPair2 = $pairs[$i+1];
	addChunk($auxId,$mainSeq,$mainQuals,
		 $pPair1->[1]+1,
		 $pPair2->[0]-$pPair1->[1]-1,
		 \$seq,\$quals,\$auxLen,$mh);
    }
    addChunk($auxId,$mainSeq,$mainQuals,
	     $pairs[-1][1]+1,
	     min(length($mainSeq)-$pairs[-1][1]-1,$pad),
	     \$seq,\$quals,\$auxLen,$mh);
    printFormatted($gh,$auxId,$seq);
    printFormatted($qh,$auxId,$quals);
}

sub addChunk {
    my ($id,$mainSeq,$mainQuals,$chunkStart,$chunkLen,$pSeq,$pQuals,$pAuxLen,$mh) = @_;
    $$pSeq .= substr($mainSeq,$chunkStart,$chunkLen);
    $$pQuals .= substr($mainQuals,$chunkStart,$chunkLen);
    print $mh "$id $$pAuxLen $chunkStart $chunkLen\n";
    $$pAuxLen += $chunkLen;
}

sub printFormatted {
    my ($gh,$id,$seq) = @_;
    print $gh ">$id\n";
    my $len = length($seq);
    for (my $i=0; $i<$len; $i+=60) {
	print $gh substr($seq,$i,60) . "\n";
    }
}

