#!/usr/bin/env perl

# Data for this is created with submitTransposonJobs.

use Utility;
use SGRP::Variables;
use SGRP::OffsetMap;
use SGRP::SequencedRegionSet;
use File::Basename;
use strict;

autoflush STDOUT 1;

my $MAX_SPACE = 10;

makeTransposonTable();

sub makeTransposonTable {

    setVariables();

    my $topDir = "$ROOT/match/CERE/cere/methc";
    #my $topDir = "$ROOT/match/PARA/para/methc";
    
    # See th/1206:
    my $nnn1 = '030'; # 036 for para (Feb 2008)
    my $alFile1 = "$topDir/$nnn1/repeat_alilines_pruned.txt";
    # Created using tRNAscan-SE on maximal.fa; need to use it!
    #   ~/bin/tRNAscan-SE -o trna.txt maximal.fa
    my $trFile = "$topDir/$nnn1/trna.txt";
    (my $nnn2 = sprintf("%3d",$nnn1+1)) =~ tr/ /0/;
    my $asDir = "$topDir/$nnn2/assembly";
    my $outFile = "$topDir/$nnn1/transposonTable.txt";
    
    report("reading REF offset map");
    my $pOMRef = new OffsetMap('?',"$asDir/REF/genome.map");
    report();
    
    my $pAnnos = readRepeatAlilines($alFile1,$pOMRef);

    readTRNAs($trFile,$pAnnos,$pOMRef);

    collapseCloseAnnotations($pAnnos);

    my @strains = decideStrainList($asDir);
    
    my $gh = openToWrite($outFile);
    foreach my $i (0 .. $#strains) {
	printf($gh "# %2d %s\n",$i,$strains[$i]);
	decideStrainPresence($strains[$i],$asDir,$pAnnos);
    }
    
    printTransposonTable($pAnnos,$gh);
}

sub readRepeatAlilines {
    my ($f,$pOMRef) = @_;
    my $fh = openToRead($f);
    my %annos;
    while (<$fh>) {
	chomp;
	my (undef,undef,$elt,$chr,undef,undef,$alo,$ahi,$fc) = split;
	($alo,$ahi) = sort {$a <=> $b} ($alo,$ahi);
	my ($fam,$iltr) = ($elt =~ /^TY([0-9][^-]*)-(I|LTR)/);
	#my $type = $elt =~ /^TY.-LTR/ ? 'SOLO' : $elt =~ /^TY.-I/ ? 'TY' : undef;
	if (defined $fam) {
	    my $type = $iltr eq 'I' ? "TY$fam" : "SOLO$fam";
	    $chr =~ s/^MAX\.//;
	    push @{$annos{$chr}},{alo => $alo,
				  ahi => $ahi,
				  type => $type,
				  fc => $fc,
				  rlo => $pOMRef->backTranslation("REF.$chr",$alo,2),
				  rhi => $pOMRef->backTranslation("REF.$chr",$ahi,1)};
	}
    }
    return \%annos;
}

sub readTRNAs {
    my ($f,$pAnnos,$pOMRef) = @_;
    my $fh = openToRead($f);
    # Three header lines:
    <$fh>;
    <$fh>;
    <$fh>;
    while (<$fh>) {
	chomp;
	my ($chr,undef,$alo,$ahi) = split;
	$chr =~ s/^MAX\.//;
	my $fc = 'F';
	if ($alo > $ahi) {
	    $fc = 'C';
	    my $tmp = $alo;
	    $alo = $ahi;
	    $ahi = $tmp;
	}
	push @{$pAnnos->{$chr}},{alo => $alo,
			      ahi => $ahi,
			      type => 'tRNA',
			      fc => $fc,
			      rlo => $pOMRef->backTranslation("REF.$chr",$alo,2),
			      rhi => $pOMRef->backTranslation("REF.$chr",$ahi,1)};
    }
}

sub collapseCloseAnnotations {
    my ($pAnnos) = @_;
    foreach my $chr (sort keys %$pAnnos) {
	my @list;
	my @aList = sortAnnoOrderp($pAnnos->{$chr});
	foreach my $pA (@aList) {
	    if (@list
		&& $list[-1]{type} eq $pA->{type} 
		&& $list[-1]{fc} eq $pA->{fc}
		&& sufficientlyClose($list[-1],$pA)) {
		$list[-1]{ahi} = max($list[-1]{ahi},$pA->{ahi}); # extend list[-1] to cover all of $pA
		$list[-1]{rhi} = max($list[-1]{rhi},$pA->{rhi}); # extend list[-1] to cover all of $pA
	    } else {
		push @list,$pA;
	    }
	}
	$pAnnos->{$chr} = \@list;
    }
}

# Need to use bubble sort because of inconsistent sorting criterion
sub sortAnnoOrderp {
    my ($pList) = @_;
    my @list = sort {$a->{alo} <=> $b->{alo}} @$pList;
    for (my $j=1; $j<=$#list; $j++) {
	# Push $list[$j] backwards
	my $i=$j;
	unless ($i == 0 || (annoCmp($list[$i-1],$list[$j]) >= 0)) {
	    $i--;
	}
	# list[j] should be in position i ...
	if ($i < $j) {
	    my $pAJ = $list[$j];
	    for (my $k=$j-1; $k>=$i; $k--) {
		$list[$k+1] = $list[$k];
	    }
	    $list[$i] = $pAJ;
	}
    }
    return @list;
}

sub annoCmp {
    if ($a->{type} eq $b->{type}) {
	return 
	    $b->{fc} cmp $a->{fc} || # F before C
	    $a->{rlo} <=> $b->{rlo} ||
	    $a->{rhi} <=> $b->{rhi} ||
	    $a->{alo} <=> $b->{alo} ||
	    $a->{ahi} <=> $b->{ahi};
    } else {
	return 
	    $a->{alo} <=> $b->{alo} ||
	    $a->{ahi} <=> $b->{ahi};
    }
}

sub decideStrainList {
    my ($asDir) = @_;
    my @strains;
    foreach my $f (glob("$asDir/*/genome.map")) {
	my $str = basename(dirname($f));
	if ($str eq 'REF') {
	    unshift @strains,$str;
	} elsif ($str ne 'MRCA') {
	    push @strains,$str;
	}
    }
    return @strains;
}

# Presence characters:
#   Y,y if >= 1/4 present (sequenced or imputed)
#   X,x if >= 1/10 present (sequenced or imputed)
#   N,n if any at all present
#   0 otherwise
# We use upper case if at least half of the material that is present
# is actually sequenced.

sub decideStrainPresence {
    my ($str,$asDir,$pAnnos) = @_;
    report("processing $str");
    my $gmf = "$asDir/$str/genome.map";
    my $pOM = new OffsetMap('?',$gmf) || die("Cannot find $gmf");
    my $srf = "$asDir/$str/genome.seqr";
    my $pSRS = new SequencedRegionSet($srf);
    foreach my $chr (sort keys %$pAnnos) {
	foreach my $pA (@{$pAnnos->{$chr}}) {
	    my ($lo,$hi,$type,$fc) = map {$pA->{$_}} qw(alo ahi type fc);
	    my @pair = sort {$a <=> $b} ($lo,$hi);
	    my $pSegs = $pOM->alignedSegments("$str.$chr",0,@pair);
	    my $present;
	    if ($lo > $hi) {
		# then it's an insertion ... shouldn't happen
		die("CHR $chr LO $lo HI $hi\n");
	    } else {
		# $sum1 is length of sequenced or imputed material;
		# $sum2 is length of sequenced only.
		my ($sum1,$sum2);
		foreach my $pPair (@$pSegs) {
		    my ($lo,$hi) = @$pPair;
		    $sum1 += $hi-$lo+1;
		    $sum2 += $pSRS->overlapsSequencedRegion($chr,$lo,$hi,0,1);
		}
		# Notional length
		my $nl = ($type =~ /^TY/) ? 5000 : ($type eq 'tRNA') ? 80 : 400;
		$present = (4*$sum1 >= $nl) ? 'y' : (10*$sum1 >= $nl) ? 'x' : $sum1 ? 'n' : 0;
		# At least half sequenced -> upper case:
		$present =~ tr/a-z/A-Z/ if ($sum1 && $sum2*2 >= $sum1);
	    }
	    $pA->{pres} .= $present;
	}
    }
    report();
}

sub sufficientlyClose {
    my ($pA1,$pA2) = @_;
    # Allow up to $MAX_SPACE bp in between:
    return ($pA1->{ahi} >= $pA2->{alo}-1-$MAX_SPACE
	    ||
	    ($pA1->{rlo} == $pA2->{rlo} && $pA1->{rhi} == $pA2->{rhi}));
}    

sub printTransposonTable {
    my ($pAnnos,$gh) = @_;
    printf($gh "%5s %7s %7s %7s %7s %5s %3s %s\n",
	   'CHR','ALI_LO','ALI_HI','REF_LO','REF_HI','TYPE','F/C','PRESENCE');
    foreach my $chr (sort keys %$pAnnos) {
	my @list = @{$pAnnos->{$chr}};
	my $i=0;
	my $prevHi;
	while ($i <= $#list) {
	    my $pAI = $list[$i];
	    my $j=$i+1;
	    my $pAJ = $list[$j];
	    my $union = $pAI->{pres};
	    while ($pAI->{type} eq $pAJ->{type} &&
		   $pAI->{fc} eq $pAJ->{fc} &&
		   $pAI->{rlo}+$MAX_SPACE >= $pAJ->{rhi}) {
		$union = stringUnion($union,$pAJ->{pres});
		$j++;
		last if ($j > $#list);
		$pAJ = $list[$j];
	    }
	    $j--;
	    $pAJ = $list[$j];
	    if ($union =~ /[^0]/) {
		print $gh "\n" if ($pAI->{rlo} >= $prevHi+$MAX_SPACE);
		$prevHi = $pAJ->{rhi};
		printf($gh "%5s %7d %7d %7d %7d %5s %3s %s\n",
		       $chr,
		       $pAI->{alo},$pAJ->{ahi},
		       $pAI->{rlo},$pAJ->{rhi},
		       $pAI->{type},$pAI->{fc},$union);
		if ($j > $i) {
		    foreach my $k ($i .. $j) {
			my ($lo,$hi,$rlo,$rhi,$type,$fc,$pres) = 
			    map {$list[$k]{$_}} qw(alo ahi rlo rhi type fc pres);
			if ($pres =~ /[^0]/) {
			    printf($gh "%5s %7d %7d %7d %7d %4s %4s %s\n",'',$lo,$hi,$rlo,$rhi,$type,$fc,$pres);
			}
		    }
		}
	    }
	    $i = $j+1;
	}
    }
}

sub stringUnion {
    my ($s1,$s2) = @_;
    my $s;
    my $iLim = length($s1);
    for (my $i=0; $i<$iLim; $i++) {
	my $c1 = substr($s1,$i,1);
	my $c2 = substr($s2,$i,1);
	foreach my $v (qw(1 x 0)) {
	    if ($c1 eq $v || $c2 eq $v) {
		$s .= $v;
		last;
	    }
	}
    }
    return $s;
}

