#!/usr/bin/env perl

=head1 NAME

pasteLayoutRegions.pl

=head1 SYNOPSIS

pasteLayoutRegions.pl layout1 layout2

=head1 DESCRIPTION

Given two layout files, output a pasted combination of the two, omitting
REF lines but otherwise adhering to layout format. Written for NCYC people.

=head1 AUTHOR

David Carter, dmc@sanger.ac.uk , August 2006

=head1 SEE ALSO

    extractLayoutRegion.pl
    splitLayoutByStrain.pl
    bin/createRDNAFiles

=cut

use Utility;
use Carp;

use strict;

my ($f1,$f2) = @ARGV;

my $fh1 = openFromPipe("grep -v '^>REF' $f1");
my $fh2 = openFromPipe("grep -v '^>REF' $f2");

my (@map1,@map2,@taken,%activeRead);

while (1) {
    my $line1 = <$fh1>;
    my $line2 = <$fh2>;
    last unless ($line1 || $line2);
    chomp $line1;
    chomp $line2;
    # Deal with read-start lines and mismatched gaps:
    while (1) {
	if ($line1 =~ /^>(\S+) (\d+)/) {
	    my ($read,$pos) = ($1,$2);
	    unless ($activeRead{$read}) {
		my $tgt = firstFreePosition(\@taken);
		$taken[$tgt] = $read;
		$activeRead{$read} = 1;
		$map1[$pos] = $tgt;
		printf(">%s %d\n",$read,$tgt);
	    }
	    $line1 = <$fh1>;
	    chomp $line1;
	} elsif ($line2 =~ /^>(\S+) (\d+)/) {
	    my ($read,$pos) = ($1,$2);
	    unless ($activeRead{$read}) {
		my $tgt = firstFreePosition(\@taken);
		$taken[$tgt] = $read;
		$activeRead{$read} = 1;
		$map2[$pos] = $tgt;
		printf(">%s %d\n",$read,$tgt);
	    }
	    $line2 = <$fh2>;
	    chomp $line2;
	} elsif ($line1 =~ /^[-=]/ && $line2 !~ /^[-=]/) {
	    # Deal with mismatched gaps:
	    my $line = substr($line1,0,2);
	    $line = mapToLine($line1,\@map1,\@taken,\%activeRead,$line);
	    $line = mapToLine(undef,\@map2,\@taken,\%activeRead,$line);
	    print "$line\n";
	    $line1 = <$fh1>;
	    chomp $line1;
	} elsif ($line2 =~ /^[-=]/ && $line1 !~ /^[-=]/) {
	    my $line = substr($line1,0,2);
	    $line = mapToLine($line2,\@map2,\@taken,\%activeRead,$line);
	    $line = mapToLine(undef,\@map1,\@taken,\%activeRead,$line);
	    print "$line\n";
	    $line2 = <$fh2>;
	    chomp $line2;
	} else {
	    last;
	}	
    }
    # Else the two lines should be for nucs or gaps and
    # have the same initial character, counting "-" and "=" as 
    # equivalent:
    my $c1 = substr($line1,0,1);
    $c1 =~ s/=/-/;
    my $c2 = substr($line2,0,1);
    $c2 =~ s/=/-/;
    unless ($c1 eq $c2) {
	die("MISMATCH TYPE 2:\n$line1\n$line2\n");
    }
    # Align these lines:
    my $line = substr($line1,0,2);
    $line = mapToLine($line1,\@map1,\@taken,\%activeRead,$line);
    $line = mapToLine($line2,\@map2,\@taken,\%activeRead,$line);
    print "$line\n";
}

sub firstFreePosition {
    my ($pA) = @_;
    for (my $k=0; $k<=$#$pA; $k++) {
	return $k unless (defined $pA->[$k]);
    }
    return scalar(@$pA);
}

sub mapToLine {
    my ($lineIn,$pMap,$pTaken,$pActive,$lineOut) = @_;
    my $short = scalar(@$pTaken) - (length($lineOut)/2);
    $lineOut .= '  ' x $short if ($short > 0);
    foreach my $k (0 .. $#$pMap) {
	my $tgt = $pMap->[$k];
	if (defined $tgt) {
	    my $pair = defined $lineIn ? substr($lineIn,2*($k+1),2) : '- ';
	    if ($pair =~ /^\S/) {
		substr($lineOut,2*($tgt+1),2) = $pair;
	    } else {
		delete $pActive->{$pTaken->[$tgt]};
		$pTaken->[$tgt] = $pMap->[$k] = undef;
		while (! defined $pTaken->[-1]) {
		    pop @$pTaken;
		    croak("Emptied pTaken") unless (@$pTaken);
		}
	    }
	}
    }
    return $lineOut;
}
