#!/usr/bin/perl

# Input: two forms allowed:
#
#   [-m mgr_in] [-n namesFile] args_bl [loMarker hiMarker]
#   [-m mgr_in] [-n namesFile] [-h id] [-q] args_bl marker line1 line2 ...
#
# where line(s) (quoted) are from sorted/imputed files. We distinguish between the two according
# to whether $ARGV[2] is either absent or > 0.
#
# If mgr_in is given, the "marker" numbers are interpreted as offsets (from 1) in the alignment instead.
# (Note that marker numbers proper go from 0).
#
# If "-n namesFile" is given, strain names (INCLUDING the reference) are read from it, and are used
# instead of 0,1,2... in the plots (so they should be short!).
#
# In form (2), "-h id" means to highlight in red the parts of the "lines" for the strain numbered "id".
#
# Output from form (1): blt.mLo-mHi.NN.png for marker regions mLo-mHi; (2) blt.marker.NN.png.

use Utility;
use GD::Simple;
use Getopt::Std;
use strict;

my %OPT;
getopts('h:m:n:q',\%OPT);

# OPT{q} means interpret line1 ... to have interpolated quality scores

if ($#ARGV == 0 || $ARGV[2] > 0) {
    plotBLTrees($ARGV[0],$ARGV[1],$ARGV[2]);
} else {
    plotBLTrees($ARGV[0],$ARGV[1],$ARGV[1],@ARGV[2..$#ARGV]);
}

sub plotBLTrees {
    my ($f,$mMin,$mMax,@lines) = @_;
    my $pNames = readNames($OPT{n});
    my @markerMap;
    if ($OPT{m}) {
	readMarkerMap($OPT{m},\@markerMap);
    }
    $mMin = 0 unless (defined $mMin);
    unless (defined $mMax) {
	chomp (my $line = `tail -1 $f`);
	(undef,undef,$mMax) = split(" ",$line);
    }
    my $type;
    if (@markerMap) {
	$type = "alignment positions";
	$mMin = $markerMap[$mMin];
	$mMax = $markerMap[$mMax];
    } else {
	$type = "markers";
    }
    my $nd = length($mMax);
    my $fh = openToRead($f); # args_bl
    while (<$fh>) {
	chomp;
	my ($mLo,$k,$mHi,$scale,@rest) = split;
	if (@markerMap) {
	    $mLo = $markerMap[$mLo];
	    $mHi = $markerMap[$mHi];
	}
	if ($mLo <= $mMax && $mHi >= $mMin) {
	    my ($g,$key);
	    if ($mMin == $mMax) {
		$g = sprintf("blt.%d.%2d.png",$mMin,$k);
		$key = sprintf("ARG %2d, %s %2d",$k,$type,$mMin);
	    } else {
		$g = sprintf("arg%2d/blt.%${nd}d-%${nd}d.png",$k,max($mLo,$mMin),min($mHi,$mMax));
		$key = sprintf("ARG %2d, %s %${nd}d to %${nd}d",$k,$type,max($mLo,$mMin),min($mHi,$mMax));
	    }
	    $g =~ tr/ /0/;
	    plotBLTree($scale,$key,\@rest,\@lines,$pNames,$g);
	} elsif ($mLo > $mMax) {
	    last;
	}
    }
}

sub readNames {
    my ($f) = @_;
    return undef unless (defined $f);
    my $fh = openToRead($f);
    my @names;
    while (<$fh>) {
	chomp;
	push @names,split;
    }
    return \@names;
}

sub readMarkerMap {
    my ($mgrIn,$pMM) = @_;
    my $fh = openToRead($mgrIn);
    chomp (my $hdr = <$fh>);
    my (undef,$nm) = split(" ",$hdr);
    foreach my $i (0 .. $nm-1) {
	chomp ($pMM->[$i] = <$fh>);
    }
}
    
sub plotBLTree {
    my ($scale,$key,$pRest,$pLines,$pNames,$g) = @_;
    my (@mthr,@len,@dtrs,@nTerms);
    for (my $k=0; 2*$k<$#$pRest; $k++) {
	$len[$k] = $pRest->[2*$k];
	$mthr[$k] = $pRest->[2*$k+1];
    }
    my $nt = 1+$#mthr/2;
    my @y;
    # Number of terminal descendants (including self):
    for (my $i=0; $i<$nt; $i++) {
	$nTerms[$i] = 1;
    }
    for (my $i=0; $i<=$#mthr; $i++) {
	my $m = $mthr[$i];
	# Don't assume molecular clock holds. Set y[m] to the
	# average of what it would be going up from its two
	# daughters. We may still get some mothers below
	# their daughters!
	$y[$m] += 0.5*($y[$i] + $len[$i]);
	push @{$dtrs[$m]},$i;
	$nTerms[$m] += $nTerms[$i];
    }
    my $yMax = $y[-1];
    my @x;
    setXValues(0,0.5,\@mthr,\@dtrs,\@nTerms,\@x);
    my $xSz = 800;
    my $xScale = $xSz/($nt+1);
    my $ySz = 500;
    my $yScale = $ySz/$yMax;
    my $im = new GD::Simple($xSz,$ySz+25+15*scalar(@$pLines));
    $im->bgcolor('white');
    $im->moveTo(15,15);
    $im->string($key);
    $im->moveTo(15,30);
    $im->string(sprintf("Total branch length %8.6f",$scale));
    for (my $i=0; $i<=$#x; $i++) {
	if (defined $OPT{h} && $OPT{h} == $i) {
	    $im->fgcolor('red');
	} else {
	    $im->fgcolor('black');
	}
	# Move to point i
	my $xi = $x[$i]*$xScale;
	my $yi = ($yMax-$y[$i])*$yScale;
	if ($i < $nt) {
	    # Annotate
	    my $ii = $OPT{q} ? $i*2 : $i;
	    my $dy = 30;
	    foreach my $line (@$pLines) {
		if (length($line) > $ii) {
		    my $v = substr($line,$ii,1);
		    my $qc = $OPT{q} ? substr($line,$ii+1,1) : ' ';
		    my $q = $qc eq ' ' ? ' ' : ord($qc)-33;
		    $im->moveTo($xi-5,$yi+$dy);
		    $im->string("$v$q");
		}
		$dy += 15;
	    }
	    $im->moveTo($xi-($i>=10 ? 5 : 0),$yi+15);
	    $im->string(defined $pNames->[$i] ? $pNames->[$i] : $i);
	    $im->fgcolor('black');
	}
	$im->moveTo($xi,$yi);
	# Draw to mother
	if (defined $mthr[$i]) {
	    my $m = $mthr[$i];
	    my $xm = $x[$m]*$xScale;
	    my $ym = ($yMax-$y[$m])*$yScale;
	    #print "$i to $m is $xi $yi to $xm $ym\n";
	    $im->lineTo($xm,$ym);
	}
    }
    my $gh = openToWrite($g);
    binmode $gh;
    print $gh $im->png;
    $gh->close();
}

sub setXValues {
    my ($i,$xl,$pMthr,$pDtrs,$pNTerms,$pX,$prev) = @_;
    unless (defined $pX->[$i]) {
	my $xr = $xl + $pNTerms->[$i] - 1;
	$pX->[$i] = ($xl+$xr)/2;
	#print "$i between $xl and $xr (from $prev)\n";
	if (defined $pDtrs->[$i]) {
	    my ($dl,$dr) = @{$pDtrs->[$i]};
	    if ($prev == $dl) {
		setXValues($dr,$xl+$pNTerms->[$dl],$pMthr,$pDtrs,$pNTerms,$pX,$i);
	    } elsif ($prev == $dr) {
		setXValues($dl,$xl+$pNTerms->[$dr],$pMthr,$pDtrs,$pNTerms,$pX,$i);
	    } else {
		setXValues($dl,$xl,$pMthr,$pDtrs,$pNTerms,$pX,$i);
		setXValues($dr,$xl+$pNTerms->[$dl],$pMthr,$pDtrs,$pNTerms,$pX,$i);
	    }
	}
	if (defined $pMthr->[$i]) {
	    my $m = $pMthr->[$i];
	    #print "$i has mother $m\n";
	    setXValues($m,$xl,$pMthr,$pDtrs,$pNTerms,$pX,$i);
	}
    }
}

