#!/usr/bin/perl

# Args: strain names; first is main one, rest are ones to compare against.
# Run this script in a directory where files [p?/]*/sbw.txt exist. They
# should have been created by snpsByWindows.pl. It will create files
# plots/STR/CHR.png, where STR is the main strain and CHR is a chromosome
# name (corresponding to the "*" in the sbw.txt file pattern).

# Example, in a "cere" directory:
#
# plotStrainSimilarities.pl SK1 REF NCYC110 RM11_1A 

# -l: use a log scale for the similarity (vertical) axis. Default is linear.
#     Value after -l is the minimum similarity to appear; thus if you say
#     -l 100 -h 10000, similarities below 100 will be off the bottom of the
#     plot and not appear.
# -w: width in pixels of 100k blocks (as shown by vertical lines). Default:
#     whatever value is needed to fit image in 1000 pixels.
# -W: width of whole image in pixels (cannot be combined with -w)
# -h: max similarity height of a plot (default: max value actually found).
#     10000 is recommended, or some bars may go off the top.
# -H: height of whole image in pixels (OK to combine with -h).
# -s: STRUCTURE-like display (not implemented yet)

# TO DO:
#  -- multiple chromosomes/regions; cover whole genome for paper, in multiple rows
#  -- also have "structure-like" output: multi-coloured bars, total
#     height should vary (optionally) 
#  -- think about doing same thing with distances in Margarita trees
#  -- do sim plot with best of each of the 4 pure clusters, check
#     for any odd patterns

use GD::Simple; # you may need to grab this from CPAN
use File::Basename;
use Getopt::Std;
use strict;

autoflush STDOUT 1;

my %OPT;
getopt('h:w:sH:W:l:',\%OPT);

my @COLOURS = qw(black red green blue orange purple brown);

# sbw.txt files are created by snpsByWindows.pl
my @sbwFiles = (glob("p?/chr??/sbw.txt"),glob("chr??/sbw.txt"));

if (!@ARGV) {
    die("Usage: $0 mainStrain compStrain1 compStrain2 ...\n");
} elsif (!@sbwFiles) {
    die("$0: cannot find [p?/]chr??/sbw.txt\n");
} else {
    plotStrainSimilarities(\@sbwFiles,\%OPT,@ARGV);
}

sub plotStrainSimilarities {
    my ($pSBWFiles,$pOpt,@strains) = @_;
    my %fromIdx; # strain(s) FROM which we plot sims (one line per plot)
    my %toIdx; # strains(s) TO which we plot sims (one plot per strain)
    my $pH = \%fromIdx;
    my $k=0;
    my ($fromMax,$toMax);
    foreach my $str (@strains) {
	if ($str eq ':') {
	    $fromMax = $k-1;
	    $toMax = undef;
	    $k=0;
	    $pH = \%toIdx;
	} else {
	    $pH->{$str} = $k;
	    $toMax = $k;
	    $k++;
	}
    }
    my $fixedHeight = $pOpt->{h};
    my $fixedWidth = $pOpt->{w};
    my $structureLike = $pOpt->{s};
    my $simMin = defined $pOpt->{l} ? log($pOpt->{l}) : undef;
    my $key = join("_",@strains[0..$fromMax]);
    foreach my $f (@$pSBWFiles) {
	my $chr = basename(dirname($f));
	my $g = "plots/$key/$chr.png";
	print "Creating $g ... ";
	# $bMax is maximum block number (blocks are 10k), so
	#   $bMax/10 is number of 100k segments.
	# $simMax is maximum similarity value.
	my ($bMax,$simMax,$pSim) = readSimilarities($f,\%fromIdx,\%toIdx);
	# If -w (fixedWidth) is given, dimX will have to be
	# fixedWidth*bMax/10 = facX*1000
	my $facX = defined $fixedWidth ? $bMax*$fixedWidth/10000 : 1;
	# If -h is given, just set simMax to it.
	$simMax = $fixedHeight if defined ($fixedHeight);
	# Round $simMax up to nearest 1000
	$simMax = 1000*int(1+($simMax-1)/1000);
	# Take log if we're doing logs
	$simMax = log($simMax) if (defined $simMin);
	my $simDiff = $simMax-$simMin;
	my $spaceX = 80;
	my $spaceY = 100;
	my $dimX = defined $pOpt->{W} ? $pOpt->{W}-$spaceX : $facX*1000;
	my $sx = $dimX/($bMax+3);
	my $dimY = defined $pOpt->{H} ? $pOpt->{H}-$spaceY : 500;
	my $sy = $dimY/($simDiff*($toMax+1));
	my $im = new GD::Simple(max(700,$dimX+$spaceX+20),$dimY+$spaceY);
	# Vertical lines
	for (my $b=0; $b<=$bMax+10; $b+=10) {
	    my $xx = $b <= $bMax ? $spaceX+$sx*$b : $spaceX+$dimX;
	    $im->fgcolor($b > 0 && $b <= $bMax ? 'lightgrey' : 'black');
	    foreach my $ti (0 .. $toMax) {
		my $y0 = $dimY-$sy*$ti*$simDiff;
		$im->moveTo($xx,$y0);
		$im->lineTo($xx,$y0-$sy*0.9*$simDiff);
	    }
	}
	# Horizontal lines
	my $hdy = defined $simMin ? log(2) : 1000;
	foreach my $idx (0 .. $toMax) {
	    for (my $y=$simMax; $y>$simMin; $y -= $hdy) {
		$im->fgcolor($y < $simMax ? 'lightgrey' : 'black');
		my $yy = $dimY-$sy*(0.9*($y-$simMin)+$idx*$simDiff);
		$im->moveTo($spaceX,$yy);
		$im->lineTo($dimX+$spaceX,$yy);
	    }
	    # Lowest black line in a plot:
	    $im->fgcolor('black');
	    my $lowY = $dimY-$sy*$idx*$simDiff;
	    $im->moveTo($spaceX,$lowY);
	    $im->lineTo($dimX+$spaceX,$lowY);
	}
	$im->fgcolor('black');
	$im->moveTo(10+$spaceX,$dimY+30);
	my $legend = "Similarity in $chr";
	$legend .= ". Vertical bars every 100kbp.";
	$im->string($legend);
	$im->moveTo(10+$spaceX,$dimY+45);
	$legend = defined $simMin
	    ? sprintf("Horizontal bars every factor of 2; range is %d to %d.",
		      int(exp($simMin)),int(exp($simMax)))
	    : sprintf("Horizontal bars every 1000; top is %d.",int($simMax));
	$im->string($legend);
	$im->moveTo(10+$spaceX,$dimY+60);
	$im->string("Lines are for: ");
	my $xOff = 100+$spaceX;
	foreach my $fi (0 .. $fromMax) {
	    $im->fgcolor($COLOURS[$fi%scalar(@COLOURS)]);
	    $im->moveTo($xOff,$dimY+60);
	    $im->string($strains[$fi]);
	    $xOff += length($strains[$fi])*10;
	}
	foreach my $idx (0 .. $toMax) {
	    $im->fgcolor('black');
	    $im->moveTo(10,$dimY-$sy*$simDiff*($idx+0.45));
	    $im->string($strains[$idx+$fromMax+2]);
	}
	foreach my $ti (0 .. $toMax) {
	    my $pA = $pSim->[$ti];
	    next unless ($pA);
	    foreach my $fi (0 .. $#$pA) {
		my $pB = $pA->[$fi];
		next unless ($pB);
		$im->fgcolor($COLOURS[$fi%scalar(@COLOURS)]);
		my $init = 1;
		my $wasNeg;
		foreach my $block (0 .. $#$pB) {
		    my $sim = $pB->[$block];
		    if (defined $sim) {
			$sim = log($sim)-$simMin if (defined $simMin);
			my $xx = $spaceX+$sx*$block;
			my $yy2 = $dimY-$sy*$simDiff*$ti;
			my $yy1 = $yy2-0.9*$sy*$sim;
			if ($init || $wasNeg || $sim < 0) {
			    $im->moveTo($xx,$yy1);
			    $init = 0;
			    $wasNeg = $sim<0;
			} else {
			    $im->lineTo($xx,$yy1);
			}
		    }
		}
	    }
	}
	`mkdir -p plots/$key`;
	my $gh = new FileHandle(">$g");
	binmode $gh;
	print $gh $im->png;
	print "done\n";
    }
}

sub readSimilarities {
    my ($f,$pFromIdx,$pToIdx) = @_;
    my $fh = new FileHandle($f);
    my ($bMax,$simMax,@sim);
    while (<$fh>) {
	chomp;
	my ($s1,$s2,undef,$block,undef,undef,undef,$n,undef,$d) = split;
	next if ($d < 50); # uncertainty
	my @pairs;
	if (defined $pFromIdx->{$s1} && defined $pToIdx->{$s2}) {
	    push @pairs,[$pFromIdx->{$s1},$pToIdx->{$s2}];
	}
	if (defined $pFromIdx->{$s2} && defined $pToIdx->{$s1}) {
	    push @pairs,[$pFromIdx->{$s2},$pToIdx->{$s1}];
	}
	foreach my $pPair (@pairs) {
	    my ($fi,$ti) = @$pPair;
	    my $sib = $d/($n+1); # was: $n || 1
	    $sim[$ti][$fi][$block] = $sib;
	    $bMax = max($block,$bMax);
	    $simMax = max($sib,$simMax);
	}
    }
    return ($bMax,$simMax,\@sim);
}
	    
sub min {
    my ($x,$y) = @_;
    if (defined $y && $x > $y) {
	return $y;
    } elsif (defined $x) {
	return $x;
    } else {
	return $y;
    }
}
   
sub max {
    my ($x,$y) = @_;
    if (defined $y && $y > $x) {
	return $y;
    } elsif (defined $x) {
	return $x;
    } else {

	return $y;
    }
}

