#!/usr/bin/env perl

# Polymorphism rate calculator using monomorphic tract lengths. dirs containing sorted/imputed files.

use Utility;
use Getopt::Std;
use strict;

my %OPT;
# Output file, quality threshold, file type
getopts('o:q:t:',\%OPT);

my $QTH = defined $OPT{q} ? chr(ord('!')+$OPT{q}) : 'I'; # Quality threshold: I = 40 by default

polymorphismRates($OPT{t} || 'm',@ARGV);

sub polymorphismRates {
    my ($suf,@dirs) = @_;
    my (%sCount,%iCount);
    foreach my $d (@dirs) {
	report("$d: processing");
	augmentTractCounts($d,$suf,\%sCount,\%iCount);
	report();
    }
    printTractMeans(\%sCount,\%iCount);
}

sub augmentTractCounts {
    my ($d,$suf,$pSCount,$pICount) = @_;
    my $sFile = sprintf("%s/sorted%s.gz",$d,$suf eq 'm' ? '' : "_$suf");
    my $iFile = sprintf("%s/imputed%s.gz",$d,"_$suf");
    return unless (-s $sFile && -s $iFile);
    my $sh = openToRead($sFile);
    my $ih = openToRead($iFile);
    chomp (my $hdr = <$sh>);
    my @strains = ('REF',split(" ",substr($hdr,1)));
    <$ih>; # consume header line
    # Tract length. Each tract finishes either in a polymorphism
    # or in the end of the sequence.
    my (@sLen,@iLen);
    # Counts, indexed by column position
    my (@sCount,@iCount);
    while (<$sh>) {
	chomp;
	my $iLim = length($_);
	chomp (my $iLine = <$ih>);
	# Find positions to use
	my @trips;
	for (my $i=0; $i<$iLim; $i+=2) {
	    my $vi = substr($_,$i,1);
	    next if ($vi eq ' ' || $vi eq 'N');
	    my $qi = substr($_,$i+1,1);
	    next if ($qi ne ' ' && $qi lt $QTH);
	    $vi = normalizeSymbol($vi,$i);
	    my $vii = normalizeSymbol(substr($iLine,$i,1),$i);
	    push @trips,[$i/2,$vi,$vii];
	}
	foreach my $ki (0 .. $#trips-1) {
	    my ($i,$vis,$vii) = @{$trips[$ki]};
	    foreach my $kj ($ki+1 .. $#trips) {
		my ($j,$vjs,$vji) = @{$trips[$kj]};
		if ($vis =~ /[ACGT]/ || $vjs =~ /[ACGT]/) {
		    $sLen[$i][$j] ++;
		    unless ($vis eq $vjs) {
			$sCount[$i][$j]{$sLen[$i][$j]} ++;
			$sLen[$i][$j] = 0;
		    }
		}
		if ($vii =~ /[ACGT]/ || $vji =~ /[ACGT]/) {
		    $iLen[$i][$j] ++;
		    unless ($vii eq $vji) {
			$iCount[$i][$j]{$iLen[$i][$j]} ++;
			$iLen[$i][$j] = 0;
		    }
		}
	    }
	}
    }
    foreach my $i (0 .. $#iLen) {
	my $si = $strains[$i];
	foreach my $j ($i+1 .. $#{$iLen[$i]}) {
	    my $sj = $strains[$j];
	    # Add final tract
	    $sCount[$i][$j]{$sLen[$i][$j]} ++ if ($sLen[$i][$j]);
	    $iCount[$i][$j]{$iLen[$i][$j]} ++ if ($iLen[$i][$j]);
	    # Add into pSCount and $pICount
	    my ($s1,$s2) = sort ($si,$sj);
	    my $pH = $sCount[$i][$j];
	    if ($pH) {
		foreach my $l (keys %$pH) {
		    $pSCount->{$s1}{$s2}{$l} += $pH->{$l};
		}
	    }
	    my $pH = $iCount[$i][$j];
	    if ($pH) {
		foreach my $l (keys %$pH) {
		    $pICount->{$s1}{$s2}{$l} += $pH->{$l};
		}
	    }
	}
    }
}

sub normalizeSymbol {
    my ($vi,$i) = @_;
    if ($i == 0) {
	$vi =~ tr/acgt/----/;
    } else {
	$vi =~ tr/acgt/ACGT/;
    }
    $vi =~ tr/_=/--/;
    return $vi;
}

sub printTractMeans {
    my ($pSCount,$pICount) = @_;
    my $gh = defined $OPT{o} ? openToWrite($OPT{o}) : *STDOUT;
    foreach my $s1 (sort keys %$pICount) {
	my $pSH1 = $pSCount->{$s1};
	my $pIH1 = $pICount->{$s1};
	foreach my $s2 (sort keys %$pIH1) {
	    my ($sMed,$sTot) = medianFromHash($pSH1->{$s2});
	    my ($iMed,$iTot) = medianFromHash($pIH1->{$s2});
	    printf($gh "%13s %13s %12.10f %12.10f %7d %7d\n",
		   $s1,$s2,1/$sMed,1/$iMed,$sTot,$iTot);
	}
    }
}

sub medianFromHash {
    my ($pH) = @_;
    my @lens = sort {$a <=> $b} keys %$pH;
    my $sum = 0;
    foreach my $len (@lens) {
	$sum += $len*$pH->{$len};
    }
    my $want = $sum/2;
    foreach my $len (@lens) {
	$want -= $len*$pH->{$len};
	return ($len,$sum) if ($want <= 0);
    }
    return ($lens[-1],$sum);
}

	
	    
