#!/usr/bin/env perl

# Input: ssaha2 output of reads against TE database; see th/0710

use Utility;
use strict;

my $MIN_SCORE = 100;
my $MAX_SLACK = 10;
my $MIN_CHUNK = 50;

my $TEL_SZ = 20000;

# First get chr sizes
my %CHRSZ;
my $fh = openToRead("$ENV{SGR}/ref/cere/genome.sz");
while (<$fh>) {
    chomp;
    my ($chr,$sz) = split;
    $CHRSZ{$chr} = $sz;
}

my (@regs,$oldRead,$oldRLen);
while (<>) {
    chomp;
    next unless (/^ALIGNMENT/);
    my (undef,$score,$read,$elt,$rlo,$rhi,$clo,$chi,$fc,$mlen,$perci,$rlen) = split;
    # Ignore if telomeric element
    my ($teChr,$teLo,$teHi) = split(/\./,$elt);
    next if ($teLo <= $TEL_SZ || $teHi >= $CHRSZ{$teChr}-$TEL_SZ);
    unless ($read eq $oldRead) {
	processRegions($oldRead,$oldRLen,@regs);
	$oldRead = $read;
	$oldRLen = $rlen;
	@regs = ();
    }
    next unless ($score >= $MIN_SCORE);
    ($rlo,$rhi) = sort {$a <=> $b} ($rlo,$rhi);
    ($clo,$chi) = sort {$a <=> $b} ($clo,$chi);
    my $eLen = elementLength($elt);
    my $sb;
    # Test is to stop us suggesting repeated transposons in the aug sequence
    if ($eLen <= 7000) {
	$sb = suggestsBorder($rlo,$rhi,$rlen,$eLen,$clo,$chi,$fc);
    }
    if ($sb) {
	push @regs,[$rlo,$rhi,$sb,$elt,$eLen,$fc];
    } else {
	push @regs,[$rlo,$rhi];
    }
}
processRegions($oldRead,$oldRLen,@regs);

sub suggestsBorder {
    my ($rlo,$rhi,$rlen,$eLen,$clo,$chi,$fc) = @_;
    if ($fc eq 'F') {
	# Border to LEFT of this match if rlo and clo both small
	return 'L' if ($rlo > $MIN_CHUNK && $clo <= $MAX_SLACK);
	return 'R' if ($rhi <= $rlen-$MIN_CHUNK && $chi >= $eLen-$MAX_SLACK);
    } else {
	return 'L' if ($rlo > $MIN_CHUNK && $chi >= $eLen-$MAX_SLACK);
	return 'R' if ($rhi <= $rlen-$MIN_CHUNK && $clo <= $MAX_SLACK);
    }
}

sub elementLength {
    my ($elt) = @_;
    my (undef,$lo,$hi) = split(/\./,$elt);
    return $hi-$lo+1;
}

# Find the union of the "r" parts of the regions.
sub processRegions {
    my ($read,$rlen,@regs) = @_;
    return unless (@regs);
    # Longer elements first, for printing choice
    @regs = sort {$b->[4] <=> $a->[4]} @regs;
    my ($lo,$hi) = @{$regs[0]};
    foreach my $i (1 .. $#regs) {
	my $pR = $regs[$i];
	$lo = min($lo,$pR->[0]);
	$hi = max($hi,$pR->[1]);
    }
    foreach my $pR (@regs) {
	if ($pR->[0] == $lo && $pR->[2] eq 'L') {
	    printf("%s %4d %4d %s %s\n",$read,1,$lo-1,$pR->[3],$pR->[5]);
	    # We return rather than trying for an "R" as well because if we
	    # get an "L", it's overwhelmingly likely that any apparently
	    # non-te at the R end is poor sequence.
	    return;
	}
    }
    foreach my $pR (@regs) {
	if ($pR->[1] == $hi && $pR->[2] eq 'R') {
	    printf("%s %4d %4d %s %s\n",$read,$hi+1,$oldRLen,$pR->[3],$pR->[5]);
	    return;
	}
    }
}
