#!/usr/bin/env perl

# Args: qMap file(s). For each one, report entropy wrt 
#  (1) uncal qualities, 
#  (2) cal quals from first file
#  (3) cal quals from first file for same run
#  (4) cal quals from own file.

use Utility;
use File::Basename;
use Getopt::Std;
use strict;

my %OPT;
# -r is lowest ratio to use for subsampling; can be $r or 1/$r
getopts('r:',\%OPT);

my @prop;
if ($OPT{r}) {
    my $ratio = $OPT{r} > 1 ? 1/$OPT{r} : $OPT{r};
    $ratio = exp(log($ratio)/3);
    my $pr = 1;
    foreach my $i (0 .. 2) {
	$pr *= $ratio;
	unshift @prop,$pr;
    }
}

my @data;
foreach my $i (0 .. $#ARGV) {
    my $fh = openToRead($ARGV[$i]);
    while (<$fh>) {
	chomp;
	my ($r,$pos,$q1,$q2,$n,$m) = split;
	$data[$i][$r][$pos][$q1] = [$q2,$n,$m];
    }
}

my @names;
if (@prop) {
    foreach my $i (0 .. $#prop) {
	$names[$i] = sprintf("P=%6.4f",$prop[$i]);
    }
    $names[$#prop+1] = sprintf("P=%6.4f",1);
} else {
    @names = qw(Uncal CalFirst CalRun CalSelf);
}
printf("%8s | %8s %7s | %8s %7s %7s | %8s %7s %7s %7s | %3s %s\n",
       $names[0],$names[1],"%Save0",$names[2],"%Save0","%Save1",$names[3],"%Save0","%Save1","%Save2","Mbp","File");
foreach my $i (0 .. $#ARGV) {
    my $f = $ARGV[$i];
    my (@h,$sum);
    # $j is to be index of first file for same run as this one
    my $j=0;
    my $len2 = length(dirname($f))-2;
    my $key = substr(dirname($f),0,$len2);
    while ($j<=$i) {
	last if (substr(dirname($ARGV[$j]),0,$len2) eq $key);
	$j++;
    }
    foreach my $r (0 .. $#{$data[$i]}) {
	my $pA1 = $data[$i][$r];
	if (defined $pA1) {
	    foreach my $pos (1 .. $#$pA1) {
		my $pA2 = $pA1->[$pos];
		if (defined $pA2) {
		    my $maxQ = $#$pA2;
		    foreach my $q1 (0 .. $maxQ) {
			my $pA3 = $pA2->[$q1];
			if (defined $pA3) {
			    my ($q2,$nOccs,$nErrs) = @{$pA3};
			    if ($nOccs > 0) {
				$sum += $nOccs;
				my $nCorr = $nOccs-$nErrs;
				if (@prop) {
				    foreach my $i (0 .. $#prop) {
					$h[$i] += entropy($nErrs,$nCorr,subsamplingQuality($nErrs,$nCorr,$prop[$i],$maxQ));
				    }
				} else {
				    $h[0] += entropy($nErrs,$nCorr,$q1);
				    $h[1] += entropy($nErrs,$nCorr,$data[0][$r][$pos][$q1][0]);
				    $h[2] += entropy($nErrs,$nCorr,$data[$j][$r][$pos][$q1][0]);
				}
				$h[3] += entropy($nErrs,$nCorr,$q2);
			    }
			}
		    }
		}
	    }
	}
    }
    foreach my $i (0 .. $#h) {
	printf("%8.6f",$h[$i]/($sum*log(2)));
	foreach my $j (0 .. $i-1) {
	    printf(" %7.2f",100*(1-$h[$i]/$h[$j]));
	}
	print " | ";
    }
    printf("%3d %s\n",$sum/1000000,$f);
}

sub entropy {
    my ($nErrs,$nCorr,$q) = @_;
    my $p1 = min(0.5,exp(-log(10)*0.1*$q));
    return -log($p1)*$nErrs -log(1-$p1)*$nCorr;
}

sub subsamplingQuality {
    my ($nErrs,$nCorr,$pr,$maxQ) = @_;
    my $ne = subsampleCount($nErrs,$pr);
    my $nc = subsampleCount($nCorr,$pr);
    # Heng's formula for estimated error prob:
    my $pe = ($ne+0.1/($maxQ+1))/($nc+0.1);
    return int(0.5-10*log($pe)/log(10));
}

sub subsampleCount {
    my ($n,$p) = @_;
    my $c = 0;
    if ($n < 10) {
	# Do exactly:
	foreach my $i (1 .. $n) {
	    $c++ if (rand() < $p);
	}
    } else {
	# Use normal approximation:
	my $mu = $n*$p;
	my $var = $n*$p*(1-$p);
	my $r = -log(rand());
	my $theta = rand()*6.28319;
	my $z = $r*cos($theta);
	my $v = $mu+$z*sqrt($var);
	$c = max(0,min(int($v+0.5),$n));
    }
    return $c;
}
