#!/usr/bin/env perl

# Switches: -o outFile; -r ref positions only; -s wrt strains;
#           -t causes qual -> int qual/10
# Args: alignment files (layout, sorted or imputed)
# Output: lines of form "type qual count", where type
# is one of hg, sg (hard and soft gaps), un, ln
# (upper and lower case nucleotide). Or input file
# can be a (Maq) fastq.gz file, in which case we
# only print the results for "un" (and ignore N's).

use Utility;
use Getopt::Std;
use strict;

autoflush STDOUT 1;

my %SYMBOL_TYPE = qw(A un C un G un T un N na
                     a ln c ln g ln t ln n ln
                     = hg _ lg
                     - ug 0 ug 1 ug 2 ug 3 ug 4 ug 5 ug 6 ug 7 ug 8 ug 9 ug);

$SYMBOL_TYPE{' '} = 'na';

my %SYMBOL_INDEX = qw(un 0 ln 1 ug 2 lg 3 hg 4 na 5);

my %OPT;
getopts('o:rst',\%OPT);

my %count;
foreach my $f (@ARGV) {
    report("$f: reading");
    if ($f =~ /fastq/) {
	augmentQualityDistributionFastq(openToRead($f),\%count);
    } else {
	augmentQualityDistribution(openToRead($f),\%count);
    }
    report();
}
printQualityDistribution(\%count,defined $OPT{o} ? openToWrite($OPT{o}) : *STDOUT);

sub printQualityDistribution {
    my ($pCount,$gh) = @_;
    foreach my $str (sort keys %$pCount) {
	printQualityDistribution1($str,$pCount->{$str},$gh);
    }
}

sub printQualityDistribution1 {
    my ($str,$pCount,$gh) = @_;
    print $gh "\nStrain $str\n" if ($str);
    my @type;
    if ($OPT{r}) {
	@type = qw(eq df gp nn na);
    } else {
	@type = qw(un ln ug lg hg na);
    }
    my @sum = 0;
    my $qMax = $#$pCount; # $OPT{s} ? 9 : 94;
    foreach my $q (0 .. $qMax) {
	foreach my $i (0 .. $#type) {
	    $sum[$i] += $pCount->[$q][$i];
	}
    }
    my $tot;
    foreach my $v (@sum) {
	$tot += $v;
    }
    # Final column is sum along the row:
    push @type,'all';
    push @sum,$tot;
    print $gh "type";
    foreach my $i (0 .. $#type) {
	printf($gh " %12s",$type[$i]) if ($sum[$i]);
    }
    print $gh "\n";
    print $gh " sum";
    foreach my $i (0 .. $#type) {
	printf($gh " %12d",$sum[$i]) if ($sum[$i]);
    }
    print $gh "\n";
    #print $gh "prps";
    my @left;
    foreach my $i (0 .. $#type) {
	$left[$i] = $sum[$i]/$tot;
	#printf($gh " %12.10f",$left[$i]) if ($sum[$i]);
    }
    #print $gh "\n";
    foreach my $q (0 .. $qMax) {
	printf($gh "%4d",$OPT{t} ? $q*10 : $q);
	my $t = 0;
	foreach my $i (0 .. $#type-1) {
	    if ($sum[$i]) {
		my $r = $pCount->[$q][$i]/$tot;
		printf($gh " %12.10f",max(0,$left[$i]));
		$t += $r;
		$left[$i] -= $r;
	    }
	}
	printf($gh " %12.10f %s\n",max(0,$left[$#type]),$str);
	$left[$#type] -= $t;
    }
}

sub augmentQualityDistributionFastq {
    my ($fh,$pCount) = @_;
    my $pQuad;
    my @count;
    while ($pQuad = readQuad($fh)) {
	my (undef,$nucs,undef,$quals) = @$pQuad;
	my $iLim = length($nucs);
	for (my $i=0; $i<$iLim; $i++) {
	    if (substr($nucs,$i,1) =~ /[ACGT]/) {
		$count[ord(substr($quals,$i,1))-33] ++;
	    }
	}
    }
    foreach my $q (0 .. $#count) {
	$pCount->[$q][0] += $count[$q]; # 0 is index for "un"
    }
}

sub augmentQualityDistribution {
    my ($fh,$pCount) = @_;
    my $wrtRef = $OPT{r};
    my $byStrain = $OPT{s};
    chomp (my $hdr = <$fh>);
    my @strains = split(" ",substr($hdr,1));
    while (<$fh>) {
	chomp;
	my $iLim = length($_);
	my $r = substr($_,0,1);
	next if ($wrtRef && $r !~ /[ACGT]/);
	for (my $i=2; $i<$iLim; $i+=2) {
	    my $v = substr($_,$i,1);
	    #if ($v ne ' ') {
	    my $qc = substr($_,$i+1,1);
	    my $idx;
	    if ($wrtRef) {
		$v =~ tr/a-z/A-Z/;
		if ($v eq $r) {
		    $idx = 0; # same
		} elsif ($v =~ /[ACGT]/) {
		    $idx = 1; # SNP
		} elsif ($v eq 'N') {
		    $idx = 3; # N
		} elsif ($v ne ' ') {
		    $idx = 2; # gap
		} else {
		    $idx = 4; # not sequenced
		}
	    } else {
		my $ty = $SYMBOL_TYPE{$v} || die("No type for $v at $i of $_");
		$idx = $SYMBOL_INDEX{$ty};
	    }
	    # Treat all qualities on N as zero ...
	    my $q = ($v eq 'N' || $v eq ' ') ? 0 : $qc eq ' ' ? 94 : ord($qc)-33;
	    my $str;
	    if ($byStrain) {
		$str = $strains[$i/2-1];
	    }
	    if ($OPT{t}) {
		$q = int($q/10);
	    }
	    $pCount->{$str}[$q][$idx] ++;
	}
    }
}
