#!/usr/bin/env perl

# Args: 
#  -o outFile (else stdout)
#  -m minScore (else zero)
#  -i lastIter (else use all)
#  -s subDir (to add onto root dir, if any)
#  -x extension (to e.g. cere dir under read dir, e.g. .sv2)
#  -d comma-separated list of "discard" files (from readsToDiscard.pl).
#     Class these as "discarded" if they come out as placed.
# cere methc CERE

# Can hand the output of this to pairFates.pl

use Utility;
use Getopt::Std;
use File::Basename;
use strict;

autoflush STDOUT 1;

my %OPT;
getopts('d:o:m:i:s:x:',\%OPT);

my $MIN_SCORE = $OPT{m};
my $LAST_ITER = $OPT{i};
my $OUT_FILE = $OPT{o}; # or undef
my $SUB_DIR = $OPT{s};
my $EXTN = $OPT{x};
my $DISCFILES = $OPT{d};

my %FATE_TRANSLATION = 
  qw(accept           placed
     accept_bisr      clashed
     accept_chi_late  clashed
     accept_close     ambiguous
     accept_insertion unsafe_insertion
     accept_late      clashed
     manual           ambiguous
     never_aligned    never_aligned
     reject_bisr      clashed
     reject_chi       clashed
     reject_close     ambiguous
     reject_insertion unsafe_insertion
     reject_low       no_good_mate
     reject_mlt_bisr  clashed
     reject_poor      no_good_cluster);

my %DISCARD;
foreach my $f (split(",",$OPT{d})) {
    my $fh = openToRead($f);
    while (<$fh>) {
	chomp;
	$DISCARD{$_} = 1;
    }
}

readFatesTable(@ARGV);

sub readFatesTable {
    my ($sp,$meth,$set) = @_;
    my $sgr = $ENV{SGR} || die("No value for ENV SGR");
    my $list = $ENV{$set} || die("No value for ENV $set");
    my @strains = sort split(" ",$list);
    my $dir = "$sgr/match/$set/$sp/$meth";
    $dir .= "/$SUB_DIR" if (defined $SUB_DIR);
    my (%isRead,%firstAli,%firstPass,%firstAcc,%bestAli);
    my ($firstFinal,$firstExtra,$maxIter);
    my @subs = glob("$dir/[0-9][0-9][0-9]");
    foreach my $subd (@subs) {
	my $iter = basename($subd);
	#next if ($iter > 0 && $iter < 24);
	$maxIter = $iter;
	my $type = iterationType($subd);
	if ($type eq 'EXTRA') {
	    readExtraFates($subd,$iter,\%firstAcc,\%bestAli);
	    $firstExtra = $iter;
	    last;
	} else {
	    $firstFinal = $iter if ($type eq 'FINAL');
	    readAlilinesFromSubdir(\@strains,$sgr,$sp,$subd,$meth,
				   $iter,\%isRead,\%firstAli,\%firstPass,\%firstAcc,\%bestAli);
	}
	last if (defined $LAST_ITER && $iter == $LAST_ITER);
    }
    unless (defined $firstFinal && defined $firstExtra) {
	#die("Missing firstFinal and/or firstExtra");
	# Do the best we can...
	print STDERR "WARNING: final and extra iterations not found\n";
	$firstFinal = $maxIter+1;
	$firstExtra = $maxIter+2;
    }
    my $gh;
    if (defined $OUT_FILE) {
	report("$OUT_FILE: writing");
	$gh = openToWrite($OUT_FILE);
    } else {
	$gh = *STDOUT;
    }
    # Output is:
    # readName aliIter passIter accIter aliWhen passWhen accWhen aScore aLo aHi percI rLen chr [reason]
    #   "Iter" is first iteration on which (aligned|passed|accepted),
    #          with firstExtra for "never"
    #   "When" is 0 for iter 0, 1 for later but non-final iter, 
    #          2 for final iter, 3 for extra (after final), 4 for never.
    #   "aScore..." are properties of highest-scoring alignment
    foreach my $r (sort keys %isRead) {
	print $gh $r;
	my @nums2;
	foreach my $pH (\%firstAli,\%firstPass,\%firstAcc) {
	    my $iter = $pH->{$r};
	    $iter = $firstExtra unless (defined $iter);
	    printf($gh " %3d",$iter);
	    my $i2 = 
		$iter == 0 ? 0 
		: $iter < $firstFinal ? 1 
		: $iter == $firstFinal ? 2
		: $iter == $firstExtra ? 3 : 4;
	    push @nums2,$i2;
	}
	print $gh " " . join(" ",@nums2);
	my $pRec = $bestAli{$r};
	my ($rlen,$good) = @{$isRead{$r}};
	my ($score,$lo,$hi,$perci,$chr,$reason);
	if (defined $pRec) {
	    ($score,$lo,$hi,$perci,$chr,$reason) = @$pRec;
	    $reason = 'accept' unless ($reason);
	    $reason = 'discarded' if ($DISCARD{$r});
	} elsif ($good < 100) {
	    $reason = 'bad_read';
	} elsif ($good < 400) {
	    $reason = 'mediocre_read';
	} else {
	    $reason = 'never_aligned';
	}
	printf($gh " %4d %6.2f %4d %4d %4d %4d %8s %s\n",
	       $score,$perci,$lo,$hi,$rlen,$good,$chr || 'NULL',$FATE_TRANSLATION{$reason} || $reason);
    }
    $gh->close();
    report() if ($OUT_FILE);
}

sub readAlilinesFromSubdir {
    my ($pStrains,$sgr,$sp,$subd,$meth,
	$iter,$pIsRead,$pFirstAli,$pFirstPass,$pFirstAcc,$pBestAli) = @_;
    my $type;
    report("Processing iteration $iter");
    foreach my $str (@$pStrains) {
	my $aliFile  = $iter == 0 ? "$sgr/read/$str/$sp$EXTN/alilines.all.txt" : "$subd/align/$str/alilines.all.txt";
	my $passFile = $iter == 0 ? "$sgr/read/$str/$sp$EXTN/$meth/alilines.pass.txt" : "$subd/align/$str/alilines.pass.txt";
	readAlignmentFile($aliFile,$iter,$pFirstAli,undef);
	readAlignmentFile($passFile,$iter,$pFirstPass,undef);
	if ($iter == 0) {
	    my $wmFile = "$sgr/read/$str/renamed.wm";
	    readWMFile($wmFile,$pIsRead);
	}
    }
    report();
    # We could use alilines.nola.txt for the final iter too, but this enables earlier processing...
    #my $pat = (defined $firstFinal) ? "$subd/align/*/alilines.pass.txt" : "$subd/p?/*/alilines.nola.txt";
    my $pat = "$subd/p?/*/alilines.nola.txt";
    report("$pat: reading");
    foreach my $f (glob($pat)) {
	# Pass pBestAli through here because these are ACCEPTED alignments.
	readAlignmentFile($f,$iter,$pFirstAcc,$pBestAli);
    }
    report();
    return $type;
}

sub iterationType {
    my ($d) = @_;
    return 'FINAL' if (-f "$d/isFinalIteration");
    return 'EXTRA' if (-f "$d/isExtraIteration");
    return 'NORMAL';
}

sub readWMFile {
    my ($f,$pRead) = @_;
    #report("$f: reading");
    my $fh = openToRead($f);
    while (<$fh>) {
	chomp;
	my @a = split;
	# Keep it if it has at least 400 bp of good quality material.
	# (Lots of e.g. DBVPG6044 and YPS128 are only about 200 bp).
	my $good;
	for (my $i=2; $i<$#a; $i+=2) {
	    $good += $a[$i+1]-$a[$i]+1;
	}
	#if ($good >= 400) {
	$pRead->{$a[0]} = [$a[1],$good]; # Keep read length and #good bases.
	#}
    }
    #report();
}

sub readExtraFates {
    my ($d,$iter,$pFirst,$pBest) = @_;
    report("Processing $d/align");
    foreach my $f (glob("$d/align/*/alilines.all.status.txt")) {
	readAlignmentFile($f,$iter,$pFirst,$pBest);
    }
    report();
}

sub readAlignmentFile {
    my ($f,$iter,$pFirst,$pBest) = @_;
    #report("$f: reading");
    my $fh = openToRead($f);
    while (<$fh>) {
	chomp;
	my @a = split;
	my (undef,$score,$read,$fullChr,$r1,$r2,undef,undef,undef,undef,$perci,undef,$status) = @a;
	# If status is ftsb, there ought to be an equivalent match to main chromosome.
	if ($score >= $MIN_SCORE && $status ne 'reject_ftsb') {
	    $pFirst->{$read} = $iter unless (defined $pFirst->{$read});
	    if (defined $pBest) {
		$score += 0.1 if ($status !~ /reject/); # no status, or accept
		unless (defined $pBest->{$read} && $pBest->{$read}[0] > $score) {
		    ($r1,$r2) = sort {$a <=> $b} ($r1,$r2);
		    my (undef,$chr) = split(/\./,$fullChr);
		    my @rec = ($score,$r1,$r2,$perci,$chr);
		    if (defined $status) {
			my ($ar,$reason) = split(/_/,$status);
			push @rec,$status;
		    }
		    $pBest->{$read} = \@rec;
		}
	    }
	}
    }
    #report();
}
