#!/usr/bin/env perl

# Input: output of extractPolymorphicRegions.pl

use Utility;
use FileHandle;
use strict;

my $DEBUG = 0;

# Copied from alicat.pl:

# Bit positions are:
#  A 1
#  C 2
#  G 4
#  T 8
#  - 16

my @DNA_SYM;
my %DNA_BITS;

setDNABitsAndSymbols();

sub setDNABitsAndSymbols {
    %DNA_BITS = qw(A  1 C  2 G  4 T  8
                   M  3 R  5 W  9
                   S  6 Y 10 K 12
                   B 14 D 13 H 11 V  7
                   N 15 - 16 = 16
                   a 17 c 18 g 20 t 24
                   m 19 r 21 w 25
                   s 22 y 26 k 28
                   b 30 d 29 h 27 v 23
                   n 31);
    @DNA_SYM = ();
    while (my ($k,$v) = each %DNA_BITS) {
        # Make sure it's been "used as a number" so
        # bitwise "|" works numerically!
        $DNA_BITS{$k} += 0;
        $DNA_SYM[$v] = $k;
    }
}

sub consensusValue { # simplified from alicat.pl
    my ($v1,$v2) = @_;
    if ($v1 eq '-') {
	return $v2;
    } elsif ($v2 eq '-') {
	return $v1;
    } else {
	return $DNA_SYM[$DNA_BITS{$v1} | $DNA_BITS{$v2}];
    }
}

my $inter = 1;
autoflush STDOUT 1;
while (<>) {
    chomp;
    if (/^(\d+) (\d+)$/) {
	my ($off,$len) = ($1,$2);
	my @buf;
	foreach my $i (1 .. $len) {
	    my $line = <>;
	    chomp $line;
	    push @buf,$line;
	}
	realignPolymorphicRegion($_,@buf);
 	<>; # blank line
    } else {
	die("Expecting offset and length; got: $_");
    }
}

sub realignPolymorphicRegion {
    my ($hdr,@buf) = @_;
    my @orig;
    my %s;
    my @q;
    my $maxLen;
    foreach my $i (0 .. $#buf) {
	my $line = $buf[$i];
	my $len = length($line);
	$maxLen = max($len,$maxLen);
	for (my $j=0; $j<$len; $j+=2) {
	    my $k = $j/2;
	    my $val = substr($line,$j,1);
	    if ($val ne ' ') {
		$orig[$k][$i] = $val;
		$q[$k][$i] = substr($line,$j+1,1);
		$val =~ tr/acgt/ACGT/;
		if ($val =~ /[ACGT]/) {
		    $s{$k} .= $val; # normalized
		}
	    }
	}
    }
    my @order = sort {length($s{$b}) <=> length($s{$a})} (keys %s);
    my %ali; # mapping from each distinct sequence to gapped version
    my %seen;
    my $cons = $s{$order[0]};
    $seen{$cons} = 1;
    print "  $cons -> $cons\n" if ($DEBUG);
    $ali{$s{$order[0]}} = $cons;
    shift @order;
    foreach my $k (@order) {
	next if ($seen{$s{$k}});
	$seen{$s{$k}} = 1;
	print "  $s{$k} -> " if ($DEBUG);
	$cons = addToAlignment($s{$k},\%ali,$cons);
    }
    printRealignment($hdr,\%ali,\%s,\@orig,\@q,$maxLen);
}    
	
sub addToAlignment {
    my ($seq,$pAli,$cons) = @_;
    return $cons if (defined $pAli->{$seq});
    my @cigar = align($seq,$cons);
    # Apply the cigar to everything currently in $pAli,
    # _if_ cigar has any 'i' in it (else no change)
    my $ci = containsInsertions(\@cigar);
    if ($ci) {
	foreach my $seq1 (keys %$pAli) {
	    $pAli->{$seq1} = applyCigar('i',\@cigar,$pAli->{$seq1});
	}
    }
    # Apply the cigar to $seq
    $pAli->{$seq} = applyCigar('d',\@cigar,$seq);
    $cons = merge($pAli->{$seq},applyCigar('i',\@cigar,$cons));
    print "$pAli->{$seq} -> $cons\n" if ($DEBUG);
    return $cons;
}

sub containsInsertions {
    my ($pCigar) = @_;
    foreach my $pA (@$pCigar) {
	return 1 if ($pA->[0] eq 'i');
    }
    return 0;
}

sub applyCigar {
    my ($insOp,$pCigar,$seq) = @_;
    my $ali;
    my $off = 0;
    foreach my $pA (@$pCigar) {
	my ($op,$len) = @$pA;
	if ($op eq $insOp) {
	    $ali .= '-' x $len;
	} else {
	    $ali .= substr($seq,$off,$len);
	    $off += $len;
	}
    }
    return $ali;
}
    
sub align {
    my ($seq,$cons) = @_;
    my $ls = length($seq);
    my $lc = length($cons);
    my @score; # no. of nuc matches back to (0,0) from here
    my @op; # to go backwards
    $score[0][0] = 0;
    my $indelPenalty = 0.6;
    my $indelPenaltyExtra = 0.2;
    foreach my $j (1 .. $lc) {
	$score[0][$j] = -$indelPenalty*$j;
	$op[0][$j] = 'd';
    }
    foreach my $i (1 .. $ls) {
	$score[$i][0] = -$indelPenalty*$i;
	$op[$i][0] = 'i';
	my $vi = substr($seq,$i-1,1);
	foreach my $j (1 .. $lc) {
	    my $ms = matchScore($vi,substr($cons,$j-1,1));
	    my $mScore = $score[$i-1][$j-1] + $ms;
	    my $dScore = $score[$i][$j-1]-$indelPenalty;
	    # Cheating affine-gap adjustment...
	    $dScore -= $indelPenaltyExtra unless ($op[$i][$j-1] eq 'd');
	    my $iScore = $score[$i-1][$j]-$indelPenalty;
	    $iScore -= $indelPenaltyExtra unless ($op[$i-1][$j] eq 'i');
	    if ($mScore > $dScore && $mScore > $iScore) {
		$op[$i][$j] = 'm';
		$score[$i][$j] = $mScore;
	    } elsif ($dScore >= $iScore) {
		$op[$i][$j] = 'd';
		$score[$i][$j] = $dScore;
	    } else {
		$op[$i][$j] = 'i';
		$score[$i][$j] = $iScore;
	    }
	}
    }
    my @ops;
    my $i = $ls;
    my $j = $lc;
    while ($i > 0 || $j > 0) {
	my $opij = $op[$i][$j];
	push @ops,$opij;
	$i -- unless ($opij eq 'd');
	$j -- unless ($opij eq 'i');
    }
    my @cigar;
    while (@ops) {
	my $opn = $ops[-1];
	my $len = 0;
	while (@ops && $ops[-1] eq $opn) {
	    $len ++;
	    pop @ops;
	}
	push @cigar,[$opn,$len];
    }
    return @cigar;    
}

sub matchScore {
    my ($vi,$vj) = @_;
    if ($vi eq $vj) {
	return 1.2; # bonus for no mutations
    } else {
	my $v = consensusValue($vi,$vj);
	return ($v eq $vj) ? 1 : 0;
    }
}

sub merge {
    my ($s1,$s2) = @_;
    my $len = length($s1);
    my $s;
    foreach my $i (0 .. $len-1) {
	$s .= consensusValue(substr($s1,$i,1),substr($s2,$i,1));
    }
    return $s;
}

sub printRealignment {
    my ($hdr,$pAli,$pS,$pOrig,$pQ,$maxLen) = @_;
    my (@val,@qual); # TRANSPOSED for printing
    foreach my $k (keys %$pS) {
	my $seq = $pS->{$k};
	my @ali = split("",$pAli->{$seq});
	my $ii = 0; # in orig
	my ($vGap,$qGap); # calculate when needed
	foreach my $i (0 .. $#ali) {
	    my $v = $ali[$i];
	    if ($v eq '-') {
		if ($pOrig->[$k][$ii] =~ /[-_=0-9]/) {
		    # use it
		    $val[$i][$k] = $pOrig->[$k][$ii];
		    $qual[$i][$k] = $pQ->[$k][$ii];
		    $ii++;
		} else {
		    unless (defined $vGap) {
			($vGap,$qGap) = decideVQGap($pOrig,$pQ,$ii,$k);
		    }
		    $val[$i][$k] = $vGap;
		    $qual[$i][$k] = $qGap;
		}
	    } else {
		while ($pOrig->[$k][$ii] !~ /[A-Za-z]/) {
		    $ii++; # discard
		}
		$val[$i][$k] = $pOrig->[$k][$ii];
		$qual[$i][$k] = $pQ->[$k][$ii];
		$ii++;
	    }
	}
    }
    foreach my $k (0 .. $#$pOrig) {
	next unless ($pOrig->[$k] && !$pS->{$k}); # gaps only...
	my $iMax = min($#val,$#{$pOrig->[$k]});
	foreach my $i (0 .. $iMax) {
	    $val[$i][$k] = $pOrig->[$k][$i];
	    $qual[$i][$k] = $pQ->[$k][$i];
	}
    }
    my $isDiff = 0;
  DIFF:
    foreach my $i (0 .. $#val) {
	my $pVI = $val[$i];
	my $pQI = $qual[$i];
	foreach my $k (0 .. $#$pVI) {
	    unless ($pVI->[$k] eq $pOrig->[$k][$i]) {
		$isDiff = 1;
		last DIFF;
	    }
	}
    }
    return unless ($isDiff);
    # Actually print
    print "$hdr\n";
    my $kMax = $#$pOrig;
    foreach my $i (0 .. $#val) {
	my $pVI = $val[$i];
	my $pQI = $qual[$i];
	foreach my $k (0 .. $#$pVI) {
	    my $v = $pVI->[$k];
	    my $q = $pQI->[$k];
	    if ($v =~ /[^ ]/) {
		print "$v$q";
	    } else {
		print '  ';
	    }
	}
	my $extra = $maxLen - 2*scalar(@$pVI);
	print ' ' x $extra if ($extra);
	print "\n";
    }
    print "\n";
}

sub decideVQGap {
    my ($pV,$pQ,$ii,$kk) = @_;
    # Same strain:
    foreach my $i (0 .. $#{$pV->[$kk]}) {
	my $vi = $pV->[$kk][$i];
	if ($vi eq '-' || $vi > 0) {
	    return ($vi,$pQ->[$kk][$i] || ' ');
	}
    }
    # Same position:
    foreach my $k (1 .. $#$pV) {
	my $vi = $pV->[$k][$ii];
	if ($vi eq '-' || $vi > 0) {
	    return ($vi,$kk == 0 ? ' ' : $pQ->[$k][$ii] || ' ');
	}
    }
    # Anywhere:
    foreach my $k (1 .. $#$pV) {
	foreach my $i (0 .. $#{$pV->[$k]}) {
	    my $vi = $pV->[$k][$i];
	    if ($vi eq '-' || $vi > 0) {
		my $qi = ' ';
		if ($kk > 0) {
		    $qi = $pQ->[$k][$i];
		    $qi = ':' unless (defined $qi && $qi ne ' ');
		}
		return ($vi,$qi);
	    }
	}
    }
    # Arbitrary:
    return ('5',':');
}
