#!/usr/bin/env perl

# Args: layout.gz files
# Outputs: raq_layout.gz files

use File::Basename;
use Utility;
use strict;

autoflush STDOUT 1;

realignQualities(@ARGV);

sub realignQualities {
    my (@files) = @_;
    foreach my $f (@files) {
	my $g = sprintf("%s/raq_%s",dirname($f),basename($f));
	realignQualities1($f,$g);
    }
}

sub realignQualities1 {
    my ($f,$g) = @_;
    if (-s $g) {
	report("$g: already exists",1);
    } else {
	my $qf = "$f.tmpquals";
	$qf =~ s/\.gz//;
	if (-s $qf) {
	    report("$qf: already exists",1);
	} else {
	    report("$f: reading segments");
	    my ($pSegs,$pMultiple) = readSegments($f);
	    report();
	    report("$qf: creating");
	    createTmpquals($pSegs,$pMultiple,$qf);
	    report();
	}
	report("$g: creating");
	realignQualities2($f,$qf,$g);
	report();
    }
}

sub readSegments {
    my ($f) = @_;
    my $fh = openToRead($f);
    my (%segId,%isMultiple);
    my $id = 0;
    while (<$fh>) {
	chomp;
	if (substr($_,0,1) eq '>' && substr($_,0,5) ne '>REF ') {
	    my ($seg) = split(" ",substr($_,1),2);
	    my $read = substr($seg,0,index($seg,':'));
	    my $str = substr($read,0,index($read,'-'));
	    if (defined $segId{$str}{$read}{$seg}) {
		$isMultiple{$seg} = 1;
	    } else {
		$segId{$str}{$read}{$seg} = $id;
		$id++;
	    }
	}
    }
    return (\%segId,\%isMultiple);
}

sub createTmpquals {
    my ($pSegId,$pMultiple,$g) = @_;
    my ($str,$fh);
    my $gh = openToPipe("sort -n > $g");
    my $root = $ENV{SGR} || '/lustre/scratch1/sanger/dmc/sgroot';
    foreach my $str (sort keys %$pSegId) {
	my $pH = $pSegId->{$str};
	my $fh = openToRead("$ENV{SGR}/read/$str/renamed.fastq");
	print substr($str,0,1); # progress ...
	while (<$fh>) {
	    my $read = substr($_,1,index($_,' ')-1);
	    next unless (defined $pH->{$read});
	    # skip middle 2 lines
	    <$fh>;
	    <$fh>;
	    chomp (my $qLine = <$fh>);
	    my $pHH = $pH->{$read};
	    foreach my $seg (keys %$pHH) {
		printQualitySubstring($seg,$pHH->{$seg},$pMultiple->{$seg},$qLine,$gh);
	    }
	}
    }
    $gh->close();
}

sub printQualitySubstring {
    my ($seg,$id,$multi,$qLine,$gh) = @_;
    my (undef,$i,$j) = split(/[:|]/,$seg);
    my ($lo,$hi) = sort {$a <=> $b} ($i,$j);
    my $part = substr(substr($qLine,0,$hi),$lo-1);
    if ($i > $j) {
	$part = reverse($part);
    }
    printf($gh "%d %s %d %s\n",$id,$seg,$multi,$part);
}

sub realignQualities2 {
    my ($f,$qf,$g) = @_;
    my $fh = openToRead($f);
    my $qh = openToRead($qf);
    my $gh = openToWrite($g);
    my (%qualString,%qualPos,@segment,@gapQual,%isMultiple);
    my $nzapN = 0;
    my $nzapG = 0;
    while (<$fh>) {
	if (/^>(\S+) (\d+) /) {
	    my ($seg,$pos) = ($1,$2);
	    unless ($isMultiple{$seg}) {
		chomp (my $qLine = <$qh>);
		my (undef,$qSeg,$multi,$quals) = split(" ",$qLine);
		unless ($qSeg eq $seg) {
		    die("Expecting $seg, found $qSeg");
		}
		$segment[$pos] = $seg;
		$isMultiple{$seg} = 1 if ($multi);
		$qualString{$seg} = $quals;
		$qualPos{$seg} = 0;
	    }
	}
	if (/^>/) {
	    print $gh $_;
	} else {
	    chomp;
	    my $iLim = length($_);
	    my $pos = 0;
	    my $line = substr($_,0,2);
	    for (my $i=2; $i<$iLim; $i+=2) {
		my $v = substr($_,$i,1);
		$line .= $v;
		if ($v eq ' ') {
		    # nothing
		    $line .= ' ';
		} elsif ($v =~ /[-=_]/) {
		    my $q = $gapQual[$pos];
		    if ($q eq "" || !defined $q) {
			$line =~ s/.$/  /; # zap this bit of sequence
			$nzapG ++;
		    } else {
			$line .= $q;		
		    }
		} else {
		    my $seg = $segment[$pos];
		    my $qs = $qualString{$seg};
		    my $q = substr($qs,$qualPos{$seg},1);
		    if ($q eq "" || !defined $q) {
			#die "No qualString for $pos $i $seg at $_";
			$line =~ s/.$/  /; # zap this bit of sequence
			$nzapN ++;
		    } else {
			$line .= $q;
			$gapQual[$pos] = minQualChar($q,substr($qs,0,1));
			if ($qualPos{$seg} < length($qs)-1) {
			    $qualPos{$seg} ++;
			} elsif ($isMultiple{$seg}) {
			    $qualPos{$seg} = 0; # reset for next time
			} else {
			    delete $qualString{$seg};
			    delete $qualPos{$seg};
			}
		    }
		}
		$pos ++;
	    }
	    print $gh "$line\n";
	}
    }
    if ($nzapG || $nzapN) {
	print "zapped $nzapN orphan nucleotides and $nzapG orphan gaps ... ";
    }
}
		    
sub minQualChar {
    my ($q1,$q2) = @_;
    if (!defined $q2) {
	return $q1;
    } elsif (!defined $q1) {
	return $q2;
    } else {
	return ($q1 lt $q2) ? $q1 : $q2; # no spaces ...
    }
}
