#!/usr/bin/env perl

# Args: fasta file "queries.fa" and output file from
# ssaha2 queries.fa subject.fa -tags 1 -align 1 -best 2

use Utility;
use strict;

reconstructFastaQueries(@ARGV);

sub reconstructFastaQueries {
    my ($faFile,$ssFile) = @_;
    my $fh = openToRead($faFile);
    
    my ($hdr,$seq,%cdsRegion);
    while (<$fh>) {
	chomp;
	if (/^>/) {
	    rememberRegion($hdr,$seq,\%cdsRegion);
	    $hdr = $_;
	    $seq = "";
	} else {
	    $seq .= "$_";
	}
    }
    rememberRegion($hdr,$seq,\%cdsRegion);

    my %isStop = map {$_ => 1} qw(TAA TAG TGA);
    my %bestValues;
    my %seenSequence;
    $fh = openToRead($ssFile);
  ALIG:
    while (<$fh>) {
	chomp;
	if (/^ALIGNMENT/) {
	    my @problems;
	    my (undef,undef,$name,$chr,$q1,$q2,$s1,$s2,$fc,$mLen,$perci,$qLen) = split;
	    if (defined $bestValues{$name}) {
		my ($bestMLen,$percd) = @{$bestValues{$name}};
		next if ($mLen < $bestMLen-5 || 100-$perci > $percd*1.5);
	    } else {
		$bestValues{$name} = [$mLen,100-$perci];
	    }
	    my ($qStart,$qSeq,$sSeq) = readQandS($fh,$fc eq 'F' ? $q2 : 1);
	    #next unless (defined $cdsRegion{$name});
	    my ($gene,$lPre,$lCDS,$lPost) = @{$cdsRegion{$name}};
	    my $offLim = length($qSeq);
	    my $off = 0;
	    # No. of chars from "pre" that got aligned
	    my $qlcLen = $lPre+1-$qStart;
	    if ($qlcLen < 0) {
		push @problems,'missedStart';
	    } else {
		while (1) {
		    $qlcLen -- if (substr($qSeq,$off,1) =~ /[A-Z]/);
		    last if ($qlcLen < 0);
		    $off ++;
		    if ($off >= $offLim) {
			push @problems,'missedStart';
			last;
		    }
		}
	    }
	    # Shift alignment of gap character...
	    unless (substr($sSeq,$off,3) eq 'ATG') {
		if (substr($sSeq,$off+1,3) eq 'ATG') {
		    $off ++;
		} elsif ($off > 0 && substr($sSeq,$off-1,3) eq 'ATG') {
		    $off --;
		} else {
		    push @problems,'badStartCodon';
		}
	    }
	    substr($sSeq,0,$off) =~ tr/A-Z/a-z/;
	    while ($lCDS > 0) {
		$lCDS -- if (substr($qSeq,$off,1) =~ /[A-Z]/);
		$off ++;
		if ($off >= $offLim) {
		    push @problems,'missedEnd';
		    last;
		}
	    }
	    # Shift alignment of gap character...
	    unless ($isStop{substr($sSeq,$off-3,3)}) {
		if ($off >= 2 && $isStop{substr($sSeq,$off-2,3)}) {
		    $off ++;
		} elsif ($off >= 4 && $isStop{substr($sSeq,$off-4,3)}) {
		    $off --;
		} else {
		    push @problems,'badEndCodon';
		}
	    }
	    substr($sSeq,$off) =~ tr/A-Z/a-z/ if ($off < length($sSeq));
	    # Now we can remove the gaps
	    $sSeq =~ s/-//g;
	    # Now count the coding characters
	    my $coding = $sSeq;
	    $coding =~ s/[a-z]//g;
	    if (@problems) {
		unless (length($coding)%3 == 0) {
		    push @problems,'frameShift';
		}
	    } else {
		# Look for premature stops
		my $cMax = length($coding)-6;
		for (my $i=0; $i<=$cMax; $i+=3) {
		    if ($isStop{substr($coding,$i,3)}) {
			push @problems,'prematureStopCodon';
			last;
		    }
		}
	    }
	    # If already seen this sequence, go no further
	    unless (defined $seenSequence{$name}{$sSeq}) {
		$seenSequence{$name}{$sSeq} = 1;
		# And now print the new header ...
		print ">$name $gene Chromosome $chr ";
		if ($fc eq 'F') {
		    print "$s1-$s2 forward";
		} else {
		    print "$s2-$s1 reverse";
		}
		# Print problems and newline
		if (@problems) {
		    printf(" %s\n",join(",",@problems));
		} else {
		    print " OK\n";
		}
		# ... and the new sequence
		my $off = 0;
		my $offLim = length($sSeq);
		while ($off < $offLim) {
		    print substr($sSeq,$off,60) . "\n";
		    $off += 60;
		}
	    }
	}
    }
}

sub rememberRegion {
    my ($hdr,$seq,$pCdsRegion) = @_;
    if ($hdr) {
	my ($name,$gene) = split(" ",substr($hdr,1));
	my ($pre,$cds,$post) = ($seq =~ /^([a-z]*)([A-Z]*)([a-z]*)$/);
	die "Bad sequence for $name: $seq" unless ($cds);
	$pCdsRegion->{$name} = 
	    [$gene,length($pre),length($cds),length($post)];
    }
}
    
# Return query and subject sequences with gaps ...

sub readQandS {
    my ($fh,$qStop) = @_;
    my ($qStart,$qEnd,$qSeq,$sSeq);
    while (<$fh>) {
	my @quad = ($_);
	foreach my $i (1 .. 3) {
	    $quad[$i] = <$fh>;
	    if ($i == 1) {
		unless ($quad[$i] =~ /^Query/) {
		    if ($qStart > $qEnd) {
			$qSeq =~ tr/ACGT/TGCA/;
			$qSeq = reverse($qSeq);
			$sSeq =~ tr/ACGT/TGCA/;
			$sSeq = reverse($sSeq);
			$qStart = $qEnd;
		    }
		    return ($qStart,$qSeq,$sSeq);
		}
	    }
	}
	chomp @quad;
	my (undef,$off,$qMore,$off2) = split(" ",$quad[1]);
	$qStart = $off unless (defined $qStart);
	$qEnd = $off2;
	$qSeq .= $qMore;
	my (undef,undef,$sMore) = split(" ",$quad[3]);
	$sSeq .= $sMore;
	last if ($quad[1] =~ / $qStop$/);
    }
    if ($qStart > $qEnd) {
	$qSeq =~ tr/ACGT/TGCA/;
	$qSeq = reverse($qSeq);
	$sSeq =~ tr/ACGT/TGCA/;
	$sSeq = reverse($sSeq);
	$qStart = $qEnd;
    }
    return ($qStart,$qSeq,$sSeq);
}
