#!/usr/bin/env perl

# Args: sorted.gz or imputed.gz files

use Utility;
use strict;

my @count;

# Types:
#   0 match: both nuc
#   1 match: both gap
#   2 mismatch: different nucs
#   3 del: ref nuc, str "-" or "_"
#   4 del: ref nuc, str "="
#   5 ins: ref "-", str nuc
#   6 ins: ref "=", str nuc

# Monomorphic region size count
my @MRSC;

my $MMTH = 30; # quality threshold for printing polymorphism
my @MMTHLIST = (0,20,30,40,50,60,70,80,90);

foreach my $f (@ARGV) {
    my $fh = openToRead($f);
    chomp (my $hdr = <$fh>);
    my @strains = split(" ",substr($hdr,1));
    my $col;
    foreach my $i (0 .. $#strains) {
	if ($strains[$i] eq 'S288c') {
	    $col = 2*($i+1);
	    last;
	}
    }
    if ($col) {
	my $line=1;
	my $off=0;
	my @mrs = 0; # monomorphic region sizes for various thresholds
	while (<$fh>) {
	    chomp;
	    $line ++;
	    $off ++ if (/^[A-Z]/);
	    next if (length($_) <= $col);
	    my $r = substr($_,0,1);
	    my $v = substr($_,$col,1);
	    next if ($v eq ' ' || $v eq 'N');
	    my $qc = substr($_,$col+1,1);
	    $r =~ tr/a-z/=/; # yes!
	    $r =~ s/[0-9_]/-/;
	    $v =~ tr/a-z/A-Z/;
	    $v =~ s/[0-9_]/-/;
	    my $t; # type
	    if ($r =~ /[A-Z]/) {
		if ($v =~ /[A-Z]/) {
		    $t = ($r eq $v) ? 0 : 2;
		} else {
		    $t = ($v eq '=') ? 4 : 3;
		}
	    } elsif ($r =~ /[A-Z]/ || $v =~ /[A-Z]/) {
		$t = ($r eq '=') ? 6 : 5;
	    } else {
		$t = 1;
	    }
	    my $q = ($qc eq ' ' ? 127 : ord($qc))-33;
	    $count[$q][$t] ++;
	    if ($q >= $MMTH && $t > 1) {
		printf("%s %7d %s %s %3d",
		       $f,$off,$r,$v,$q);
		# Print the other values ...
		my $iMax = length($_)-2;
		my ($nRef,$nS288c,$nOther)=(0,0,0);
		for (my $i=2; $i<=$iMax; $i+=2) {
		    if ($i != $col) {
			my $qci = substr($_,$i+1,1);
			if ($qci ne ' ' && ord($qc)-33 >= $MMTH) {
			    my $vi = substr($_,$i,1);
			    $vi =~ tr/a-z/A-Z/;
			    $vi =~ s/[0-9_]/-/;
			    if ($vi eq $r) {
				$nRef ++;
			    } elsif ($vi eq $v) {
				$nS288c ++;
			    } else {
				$nOther ++;
			    }
			}
		    }
		}
		printf(" %2d %2d %2d\n",$nRef,$nS288c,$nOther);
	    }
	    if (/^[A-Z]/) {
		foreach my $qth (@MMTHLIST) {
		    if ($q < $qth) {
			last;
		    } elsif ($t > 1) {
			$MRSC[$qth]{$mrs[$qth]} ++;
			$mrs[$qth] = 0;
		    } else {
			$mrs[$qth] ++;
		    }
		}
	    }
	}
	foreach my $qth (@MMTHLIST) {
	    $MRSC[$qth]{$mrs[$qth]} ++;
	}
    }
}

# Monomorphic region size analysis
foreach my $qth (@MMTHLIST) {
    my $pH = $MRSC[$qth];
    my @sizes = sort {$a <=> $b} keys %$pH;
    my ($num,$den);
    foreach my $sz (@sizes) {
	$num += $pH->{$sz}*($sz+1)*$sz;
	$den += $pH->{$sz}*$sz+1;
    }
    my $want = $den/2;
    my $med;
    foreach my $sz (@sizes) {
	$want -= $pH->{$sz}*($sz+1);
	if ($want <= 0) {
	    $med = $sz;
	    last;
	}
    }
    printf("Monomorphic region sizes, quality threshold %3d: count %7d, mean %6d, median %6d, max %6d\n",
	   $qth,$den,$num/$den,$med,$sizes[-1]);
}

printf("%3s %10s",'Qu1','Pe1');
foreach my $t (0 .. 6) {
    printf(" %7s","Ty$t");
}
printf(" %3s %10s\n",'Qu2','Pe2');
for (my $q0=0; $q0<=$#count; $q0+=10) {
    my @sum;
    foreach my $u (0 .. 9) {
	my $q = $q0+$u;
	last if ($q > $#count);
	next unless ($count[$q]);
	my @lst = @{$count[$q]};
	foreach my $i (0 .. $#lst) {
	    $sum[$i] += $lst[$i];
	}
	printLine($q,\@lst);
    }
    my $qx = sprintf("%2s",$q0);
    $qx =~ s/0$/X/;
    printLine($qx,\@sum);
}

sub printLine {
    my ($q,$pLst) = @_;
    my $g = $pLst->[0]+$pLst->[1];
    my $pe1 = exp(-$q*log(10)*0.1);
    my $b = 0;
    foreach my $i (2 .. $#$pLst) {
	$b += $pLst->[$i];
    }
    my $n = $g+$b;
    my $pe2 = ($b+$pe1)/($n+1);
    my $q2 = int(0.5-log($pe2)*10/log(10));
    printf("%3s %10.8f",$q,$pe1);
    foreach my $t (0 .. 6) {
	printf(" %7d",$pLst->[$t]);
    }
    printf(" %3d %10.8f\n",$q2,$pe2);
}
    
