#!/usr/bin/env perl

use Utility;
use strict;

my $DEBUG = 1;

my $MIN_TO_KEEP = 10;

my $DEFAULT_ALIGNER = 'SSAHA'; # or could be 'BLAST'

# A match has the fields
#    0 score
#    1 qchr 
#    2 qmin
#    3 qmax
#    4 schr
#    5 smin
#    6 smax
#    7 fc
#    8 percid
#    9 idx

selectMatches(@ARGV);

sub selectMatches {
    my (@files) = @_;

    my $aligner;
    if ($files[0] =~ /blast/i) {
	$aligner = 'BLAST';
    } elsif ($files[0] =~ /ssaha/i) {
	$aligner = 'SSAHA';
    } else {
	$aligner = $DEFAULT_ALIGNER;
    }

    my $pMatches = $aligner eq 'BLAST' ? readMatchesBlast(@files) : readMatchesSsaha(@files);
    
    printf("%5s %10s %7s %7s %10s %7s %7s %2s %6s %4s %s\n",
	   "Score","QSeq","QMin","QMax","SSeq","SMin","SMax","FC","PercId","Idx","A");
    my (%boxes,%tgt);
    foreach my $pA (@$pMatches) {
	my ($score,$qchr,$qmin,$qmax,$schr,$smin,$smax,$fc,$percid,$idx) = @$pA;
	# subrange of qmin..qmax and smin..smax that we accept; or undef if reject.
	my $pAcc = decideAcceptance(\%tgt,\%boxes,$pA);
	printf("%5d %10s %7d %7d %10s %7d %7d %2s %6.2f %4d ",
	       $score,$qchr,$qmin,$qmax,$schr,$smin,$smax,$fc,$percid,$idx);
	if ($pAcc) {
	    printf("1 %7d %7d %7d %7d\n",@$pAcc);
	} else {
	    print("0\n");
	}
    }
}

sub readMatchesBlast {
    my (@files) = @_;
    my @arr;
    foreach my $f (@files) {
	# We reset the index for each file, as we assume we can determine the file
	# from the subject chromosome name.
	my $idx = 0;
	my ($qchr,$schr,$score,$qmin,$qmax,$smin,$smax,$fc,$percid);
	my $fh = openToRead($f);
	while (<$fh>) {
	    chomp;
	    if (/^Query=\s+(\S+)/) {
		$qchr = $1;
	    } elsif (/^>(\S+)/) {
		$schr = $1;
	    } elsif (/Plus Strand HSPs/) {
		$fc = 'F';
	    } elsif (/Minus Strand HSPs/) {
		$fc = 'C';
	    } elsif (/ Score = (\d+) /) {
		if (defined $score) {
		    $idx ++;
		    push @arr,[$score,$qchr,$qmin,$qmax,$schr,$smin,$smax,$fc,$percid,$idx];
		    $qmin=$qmax=$smin=$smax=$percid=undef;
		}
		$score = $1;
		#$score /= 2 if ($fc eq 'C'); # best option?
	    } elsif (/ Identities = (\d+)\/(\d+) /) {
		$percid = 100*$1/$2;
	    } elsif (/^Query:/) {
		my (undef,$q1,undef,$q2) = split;
		if ($fc eq 'F') {
		    $qmin = $q1 unless (defined $qmin);
		    $qmax = $q2;
		} else {
		    $qmax = $q1 unless (defined $qmax);
		    $qmin = $q2;
		}
	    } elsif (/^Sbjct:/) {
		my (undef,$s1,undef,$s2) = split;
		$smin = $s1 unless (defined $smin);
		$smax = $s2;
	    }
	}
	if (defined $score) {
	    $idx ++;
	    push @arr,[$score,$qchr,$qmin,$qmax,$schr,$smin,$smax,$fc,$percid,$idx];
	}
    }
    return [sort {$b->[0] <=> $a->[0]} @arr];
}

sub readMatchesSsaha {
    my (@files) = @_;
    my @arr;
    foreach my $f (@files) {
	# We reset the index for each file, as we assume we can determine the file
	# from the subject chromosome name.
	my $idx = 0;
	my ($qchr,$schr,$score,$qmin,$qmax,$smin,$smax,$fc,$percid);
	my $fh = openToRead($f);
	while (<$fh>) {
	    chomp;
	    if (/^ALIGNMENT/) {
		(undef,$score,$qchr,$schr,$qmin,$qmax,$smin,$smax,$fc,undef,$percid) = split;
		($qmin,$qmax) = sort {$a <=> $b} ($qmin,$qmax);
		($smin,$smax) = sort {$a <=> $b} ($smin,$smax);
		$idx ++;
		push @arr,[$score,$qchr,$qmin,$qmax,$schr,$smin,$smax,$fc,$percid,$idx];
	    }
	}
    }
    return [sort {$b->[0] <=> $a->[0]} @arr];
}

sub printReject {
    my ($fmt,@vals) = @_;
    if ($DEBUG) {
	printf("# Rejecting because $fmt\n",@vals);
    }
}

# $pTgt->{$qchr} = [$schr,$fcO]

# We pass members of $pM individually as we want a copy here.

sub decideAcceptance {
    my ($pTgt,$pBoxes,$pM) = @_;

    my ($score,$qchr,$qmin,$qmax,$schr,$smin,$smax,$fc) = @$pM;

    my $fcOuter;
    if (defined $pTgt->{$qchr}) {
	if ($pTgt->{$qchr}[0] eq $schr) {
	    $fcOuter = $pTgt->{$qchr}[1];
	} else {
	    printReject("%s maps to %s not %s",$qchr,$pTgt->{$qchr}[0],$schr);
	    return undef; # qchr goes to another schr
	}
    } else {
	$pTgt->{$qchr} = [$schr,undef]; # we don't know fcOuter yet
	if (!defined $pBoxes->{$schr}) {
	    # First match to this schr; easy.
	    $pBoxes->{$schr} = [$pM];
	    return [$qmin,$qmax,$smin,$smax];
	}
    }

    # It's not the first match to this schr.
    my $pBoxList = $pBoxes->{$schr};
    # Index before which we should place a match from smin to smax:
    my $iM = indexForMatch($smin,$smax,$pBoxList);
    unless (defined $iM) {
	# undef -> subsumed already
	printReject("subsumed in subject coords");
	return undef;
    }
    # Indices of box range for this qchr:
    my ($iLoQ,$iHiQ) = boxIndices($qchr,$pBoxList);
    my $pKeep;
    if (!defined $iHiQ) {
	# First match for this qchr.
	if ($iM > 0 && 
	    $iM <= $#$pBoxList) {
	    my $qcL = $pBoxList->[$iM-1][1];
	    my $qcR = $pBoxList->[$iM][1];
	    # Attempting to put it between two matches for the same
	    # other query sequence -- no good.
	    if ($qchr ne $qcL && $qcL eq $qcR) {
		printReject("$qchr cannot be inserted into $qcL region in $schr");
		return undef;
	    }
	}
	($pKeep,$pBoxes->{$schr}) = addToBoxList($pM,$iM,$pBoxList,$fcOuter);
	return $pKeep;
    }
    if ($iM < $iLoQ || $iM > $iHiQ+1) {
	# then new alignment is out of range
	printReject("%s region in %s is %d to %d, not at %d",$qchr,$schr,$iLoQ,$iHiQ,$iM);
	return undef;
    }
    # Most complex case: already some alignments for qchr; fit
    # this new one in if we can.
    ($pKeep,$pBoxes->{$schr},$pTgt->{$qchr}[1]) = addToBoxList($pM,$iM,$pBoxList,$fcOuter,$iLoQ,$iHiQ);
    #my $pBA = $pBoxes->{$schr};
    #foreach my $i (1 .. $#$pBA) {
#	if ($pBA->[$i][5] < $pBA->[$i-1][5]) {
#	    1;
#	}
#    }
    return $pKeep;
    
}

sub indexForMatch {
    my ($smin,$smax,$pBoxList) = @_;
    foreach my $i (0 .. $#$pBoxList) {
	my ($sminI,$smaxI) = @{$pBoxList->[$i]}[5,6];
	return $i if ($sminI > $smin); # some stuff to left of I
	return undef if ($smaxI >= $smax); # whole s range is taken
    }
    return $#$pBoxList + 1; # off right hand end
}

sub boxIndices {
    my ($qchr,$pBoxList) = @_;
    my $iLoQ;
    foreach my $i (0 .. $#$pBoxList) {
	if (defined $iLoQ) {
	    return ($iLoQ,$i-1) if ($pBoxList->[$i][1] ne $qchr);
	} else {
	    $iLoQ = $i if ($pBoxList->[$i][1] eq $qchr);
	}
    }
    if (defined $iLoQ) {
	return ($iLoQ,$#$pBoxList);
    } else {
	return undef;
    }
}

sub addToBoxList {
    my ($pM,$iM,$pBoxList,$fcOuter,$iLoQ,$iHiQ) = @_;
    
    my (undef,undef,$qmin,$qmax,undef,$smin,$smax,$fc) = @$pM;

    my ($pLeft,$pRight);
    $pRight = $pBoxList->[$iM] if ($iM <= $#$pBoxList);
    $pLeft = $pBoxList->[$iM-1] if ($iM > 0);
    
    $pM->[5] = $smin = max($smin,$pLeft->[6]+1) if (defined $pLeft);
    $pM->[6] = $smax = min($smax,$pRight->[5]-1) if (defined $pRight);

    if ($smax-$smin+1 < $MIN_TO_KEEP) {
	printReject("region is already taken in subject coords");
	return (undef,$pBoxList,$fcOuter); # reject this match as no space in subj sequence
    }

    my $fcOuter1;
    if (defined $iLoQ) {
	($qmin,$qmax,$fcOuter1) = pruneQueryRegion($qmin,$qmax,$smin,$smax,$fc,
						   $iM,$pBoxList,$iLoQ,$iHiQ,$fcOuter);
	if ($qmax-$qmin+1 < $MIN_TO_KEEP) {
	    printReject("region is already taken in query coords");
	    return (undef,$pBoxList,$fcOuter1 || $fcOuter); # reject this match as no space in query sequence
	}
    }

    $fcOuter = $fcOuter1 if (defined $fcOuter1);

    my $ratio = ($smax-$smin+1)/($qmax-$qmin+1);
    if ($ratio > 2 || $ratio < 0.5) {
	printReject("region has unrealistic shape after pruning: %d (q) by %d (s)",
		    $qmax-$qmin+1,$smax-$smin+1);
	return (undef,$pBoxList,$fcOuter);
    }

    $pM->[2] = $qmin;
    $pM->[3] = $qmax;
	
    if ($iM == 0) {
	unshift @$pBoxList,$pM;
    } elsif ($iM > $#$pBoxList) {
	push @$pBoxList,$pM;
    } else {
	$pBoxList = [@{$pBoxList}[0..$iM-1],$pM,@{$pBoxList}[$iM..$#$pBoxList]];
    }
    return ([$qmin,$qmax,$smin,$smax],$pBoxList,$fcOuter);
}

sub pruneQueryRegion {
    my ($qmin,$qmax,$smin,$smax,$fc,$iM,$pBoxList,$iLoQ,$iHiQ,$fcOuter) = @_;
    # This match is to go into $pBoxList between positions $iM-1 and $iM, as long
    # as the query region is consistent with what's already there.
    
    if ($iHiQ == $iLoQ) {
	# Special case: exactly one match already there for this qchr.
	return pruneQueryRegionSecond($qmin,$qmax,$smin,$smax,$fc,$iM == $iLoQ,$pBoxList->[$iLoQ]);
    } else {
	return pruneQueryRegionLater($qmin,$qmax,$smin,$smax,$fc,$iM,$pBoxList,$iLoQ,$iHiQ,$fcOuter);
    }
}

sub pruneQueryRegionSecond {
    my ($qmin,$qmax,$smin,$smax,$fc,$newToLeft,$pMFirst) = @_;
    my (undef,undef,$qmin1,$qmax1,undef,$smin1,$smax1,$fc1) = @$pMFirst;
    my $newToLeftQ = $qmin < $qmin1;
    my $fcOuter;
    if ($fc eq $fc1) {
	$fcOuter = $fc; # assume first two will be outer if same (should be safe)
    } elsif ($newToLeft eq $newToLeftQ) {
	$fcOuter = 'F';
    } else {
	$fcOuter = 'C';
    }
    if ($newToLeft eq ($fcOuter eq 'F')) {
	$qmax = min($qmax,$qmin1-1);
    } else {
	$qmin = max($qmin,$qmax1+1);
    }
    if ($qmax-$qmin+1 < $MIN_TO_KEEP) {
	return undef;
    } else {
	return ($qmin,$qmax,$fcOuter);
    }
}

sub pruneQueryRegionLater {
    my ($qmin,$qmax,$smin,$smax,$fc,$iM,$pBoxList,$iLoQ,$iHiQ,$fcOuter) = @_;

    # If this match is inverted and there is an inversion to its left or
    # right (shouldn't have both!) then it must fit into that inversion box,
    # and the extended box must fit with its own (non-inverted) neighbours.
    my $doAsNormal = 1;
    if ($fc ne $fcOuter) {
	if ($iM > 0 && $pBoxList->[$iM-1][7] eq $fc) {
	    # Must fit with inversion to left. $iMInv will be start
	    # of the inversion, so inversion is iMInv to iM-1 inclusive.
	    my $iMInv = $iM-1;
	    while ($iMInv > 0 && $pBoxList->[$iMInv-1][7] eq $fc) {
		$iMInv --;
	    }
	    if ($fc eq 'C') {
		# Inversion has C orientation
		$qmax = min($qmax,$pBoxList->[$iM-1][2]-1);
		$qmin = max($qmin,$pBoxList->[$iMInv-1][3]+1) if ($iMInv > 0);
	    } else {
		# Inversion has F orientation
		$qmax = min($qmax,$pBoxList->[$iMInv-1][2]-1) if ($iMInv > 0);
		$qmin = max($qmin,$pBoxList->[$iM-1][3]+1);
	    }
	    $doAsNormal = 0;
	} elsif ($iM <= $#$pBoxList && $pBoxList->[$iM][7] eq $fc) {
	    # Must fit with inversion to right.
	    my $iMInv = $iM;
	    while ($iMInv < $#$pBoxList && $pBoxList->[$iMInv+1][7] eq $fc) {
		$iMInv ++;
	    }
	    if ($fc eq 'F') {
		# Inversion has F orientation
		$qmax = min($qmax,$pBoxList->[$iM][2]-1);
		$qmin = max($qmin,$pBoxList->[$iMInv+1][3]+1) if ($iMInv < $#$pBoxList);
	    } else {
		# Inversion has C orientation
		$qmax = min($qmax,$pBoxList->[$iMInv+1][2]-1) if ($iMInv < $#$pBoxList);
		$qmin = max($qmin,$pBoxList->[$iM][3]+1);
	    }
	    $doAsNormal = 0;
	}
    }
    if ($doAsNormal) {
	# This match is normal, or is an inversion with normals on both side
	# so is to be treated as normal.
	# Check boxes to either side, one of which could be an inversion.
	# Get iMLo to bottom end of box to left.
	my $iMLo = $iM-1;
	while ($iMLo > 0 && 
	       $pBoxList->[$iMLo][7] ne $fcOuter &&
	       $pBoxList->[$iMLo-1][7] ne $fcOuter) {
	    $iMLo --;
	}
	my $iMHi = $iM;
	while ($iMHi < $#$pBoxList && 
	       $pBoxList->[$iMHi][7] ne $fcOuter &&
	       $pBoxList->[$iMHi+1][7] ne $fcOuter) {
	    $iMHi ++;
	}
	if ($fcOuter eq 'F') {
	    $qmax = min($qmax,$pBoxList->[$iMHi][2]-1) if ($iMHi <= $#$pBoxList);
	    $qmin = max($qmin,$pBoxList->[$iMLo][3]+1) if ($iMLo >= 0);
	} else {
	    $qmax = min($qmax,$pBoxList->[$iMLo][2]-1) if ($iMLo >= 0);
	    $qmin = max($qmin,$pBoxList->[$iMHi][3]+1) if ($iMHi <= $#$pBoxList);
	}
    }
    if ($qmax-$qmin+1 < $MIN_TO_KEEP) {
	return undef;
    } else {
	return ($qmin,$qmax,$fcOuter);
    }

}

sub addToBoxListFirst {
    my ($pM,$iM,$pBoxList) = @_;

    my (undef,undef,$qmin,$qmax,undef,$smin,$smax,$fc) = @$pM;
    my ($pLeft,$pRight);
    $pRight = $pBoxList->[$iM] if ($iM <= $#$pBoxList);
    $pLeft = $pBoxList->[$iM-1] if ($iM > 0);
    
    $pM->[5] = $smin = max($smin,$pLeft->[6]) if (defined $pLeft);
    $pM->[6] = $smax = min($smax,$pLeft->[5]) if (defined $pRight);

    if ($smax-$smin+1 < $MIN_TO_KEEP) {
	return ($pBoxList,undef); # reject this match
    }

    if ($iM == 0) {
	unshift @$pBoxList,$pM;
    } elsif ($iM == $#$pBoxList) {
	push @$pBoxList,$pM;
    } else {
	$pBoxList = [@{$pBoxList}[0..$iM-1],$pM,@{$pBoxList}[$iM..$#$pBoxList]];
    }
    return ([$qmin,$qmax,$smin,$smax],$pBoxList);
}

    
