#!/usr/bin/env perl

use Utility;
use SGRP::SGRP;
use SGRP::Variables;
use SGRP::BsubSGRP;
use Bsub;
use Getopt::Std;
use Sys::Hostname;
use strict;

my @CMD_LIST = qw(ssnp alip merg prc1 marg impu stra);
my %MULTI = map {$_ => 1} qw(prc2 prc3 marg impu smat smim);
#my @CMD_LIST = qw(ssnp alip merg prc1s smat smim stra);

my %OPT;
getopts('w:',\%OPT); # -w Pat means bwait -p Pat

sgrpCycle(@ARGV);

sub sgrpCycle {
    my ($spec,$meth,$set,$first,$last) = @_;

    if (@_ < 3 || @_ > 5) {
	die(sprintf("Usage: %s [-w Pat] species method SET [%s]\n",
		    $0,join("|",@CMD_LIST)));
    }

    my $host = hostname();
    my $process = $$;
    my $logBase = "$ENV{HOME}/cycle/$host.$process";
    print "Command: sgrpCycle.pl @_\n";
    print "Files: $logBase.{log,err,break}\n";
    my $breakFile = "$logBase.break";
    open STDOUT, ">$logBase.log";
    autoflush STDOUT 1;
    open STDERR, ">$logBase.err";
    autoflush STDERR 1;
    
    exitIfBreak($breakFile);

    setVariables();
    my $pStrains = getStrains($set);
    if ($OPT{w}) {
	report("awaiting completion of $OPT{w}");
	`bwait -p $OPT{w}`;
	report();
    }
  LOOP:
    while (1) {
	exitIfBreak($breakFile);
	my @cmds = @CMD_LIST;
	if ($first) {
	    shift @cmds while (@cmds && $cmds[0] ne $first);
	    if (@cmds) {
		$first = undef;
	    } else {
		die("Cannot find first command '$first' in @CMD_LIST\n");
	    }
	}
	foreach my $cmd (@cmds) {
	    last LOOP unless runAction($cmd,$spec,$meth,$set,$pStrains);
	    last LOOP if (defined $last && $cmd eq $last);
	}
    }
}

sub exitIfBreak {
    my ($f) = @_;
    if (-f $f) {
	print "Exiting: found $f\n";
	exit;
    }
}

sub runAction {
    my ($cmd,$spec,$meth,$set,$pStrains) = @_;
    my $dir = "$ROOT/match/$set/$spec/$meth";
    my ($iter,$iDir) = getCurrentIteration($spec,$meth,$set);
    if ($cmd eq 'marg' && $IS_FINAL_ITERATION eq 'EXTRA') {
	report("exiting at marg: extra iteration in $iDir",1);
	return 0;
    } elsif ($iter > 0) {
	my $gfa = "$iDir/assembly/REF/genome.fa";
	my $sz = -s $gfa || 0;
	if ($cmd eq 'ssnp' && $sz < 9000000) {
	    report("exiting at ssnp: size $sz for $gfa",1);
	    return 0;
	}
    }
    report("running $cmd");
    my @ids = runActionNow($cmd,$spec,$meth,$set,$pStrains);
    report();
    report("awaiting completion");
    waitUntilDone(@ids);
    report();
    # Check it worked...
    my $tf = findTouchFile($dir);
    if ($tf) {
	report("exiting because of $tf",1);
	return 0; # failure
    } else {
	return 1; # success
    }
}

sub runActionNow {
    my ($cmd,$spec,$meth,$set,$pStrains) = @_;
    my $nSubs;
    if (!$MULTI{$cmd}) {
	$nSubs = 1;
    #} elsif ($set ne 'CERQ') {
	#$nSubs = 15;
    #} elsif ($cmd eq 'marg') {
	#$nSubs = 5;
    } else {
	$nSubs = 10;
    }
    my $nFixed = 3; # $cmd eq 'merg' ? 2 : 3;
    my @args = ($spec,$meth,$set);
    if (grep {$_ eq $cmd} qw(ssnp alip stra)) {
	push @args,'REF' if ($cmd eq 'stra');
	push @args,@$pStrains;
    }
    if ($cmd =~ /^prc/ || $MULTI{$cmd}) {
	my @parts = (4,5,2,7,0,1,3,6,8,9,0 .. 7);
	unshift @parts,(4,5) unless ($cmd eq 'prc1');
	unshift @parts,(2,4) if ($cmd eq 'prc3' || $cmd eq 'smim' || $cmd eq 'impu');
	push @args,@parts;
    } elsif ($cmd eq 'merg') {
	push @args,(0..9); # only want ONE job per part!
    }
    if ($cmd eq 'ssnp') {
	$nFixed ++;
	my @ids;
	foreach my $p (0 .. 9) {
	    push @ids,bsubSGRP(1,$nSubs,$cmd,$nFixed,
			       @args[0..2],$p,@args[3..$#args]);
	}
	return @ids;
    } else {
	return bsubSGRP(1,$nSubs,$cmd,$nFixed,@args);
    }
}

sub findTouchFile {
    my ($dir) = @_;
    my ($sub) = reverse(glob("$dir/[0-9][0-9][0-9]"));
    return unless ($sub); # could be iter 000 ...
    #die("Cannot find any directory $dir/[0-9][0-9][0-9]");
    return findTouchFile1($sub);
}

sub findTouchFile1 {
    my ($dir) = @_;
    foreach my $f (glob("$dir/*")) {
	if ($f =~ /\.touch$/) {
	    return $f;
	} elsif (-d $f) {
	    my $tf = findTouchFile1($f);
	    return $tf if ($tf);
	}
    }
    return undef;
}
