#!/usr/bin/env perl

# Input dirs: a target dir of runMaq.pl, e.g. 024/solexa/S288c
# Outputs: file of same name but with .fastq.gz suffix

use Utility;
use strict;

my $CHUNK_LEN = 100;

my @ACGT = qw(A C G T);
my %ACGT = qw(A 0 C 1 G 2 T 3);

foreach my $d (@ARGV) {
    my $f = "$d/all.mapview.gz";
    my $g = "$d.fastq.gz";
    solexaAssembly($f,$g);
}

# Assumes each line of input file has the form
# foo ChrName Offset foo* Nucleotides Quals

sub solexaAssembly {
    my ($f,$g) = @_;
    if (! -s $f) {
	report("$f: missing or empty",1);
	return;
    } elsif (! stamp($g)) {
	report("$g: cannot stamp",1);
	return;
    } else {
	report("$g: creating");
    }
    my $fh = openToRead($f);
    my $gh = openToWrite($g) || die("Cannot write to $g");
    my $off = 1; # offset of instart of $seq
    my @buf; # each element is Aqual Cqual Gqual Tqual
    my $chr;
    while (<$fh>) {
	#print "off = $off\n";
	#print;
	chomp;
	my (undef,$chr1,$off1,@rest) = split;
	my $seq1 = $rest[-2];
	$seq1 =~ tr/a-z/A-Z/;
	my $quals1 = $rest[-1];
	if ($chr1 ne $chr) {
	    outputChunk(\@buf,scalar(@buf),"$chr.$off",$gh) if ($chr);
	    @buf = ();
	    $off=1;
            $chr=$chr1;
	} elsif ($off1 >= $off+$CHUNK_LEN) {
	    outputChunk(\@buf,$CHUNK_LEN,"$chr.$off",$gh);
	    @buf = @buf[$CHUNK_LEN .. $#buf];
	    $off += $CHUNK_LEN;
	}
	my $d = $off1-$off;
	#print "off1 = $off1, off = $off\n";
	integrateRead($seq1,$quals1,$d,\@buf);
    }
    outputChunk(\@buf,scalar(@buf),"$chr.$off",$gh);
    $gh->close();
    unstamp($g);
    report();
}
	    
sub outputChunk {
    my ($pBuf,$len,$name,$gh) = @_;
    $name = "REF.$name" if ($name =~ /^chr/ || $name =~ /^scplas/);
    my @comps = split(/\./,$name);
    if (@comps == 3) { # check format is right: str.chr.off
	my ($seq,$quals);
	for (my $i=0; $i<$len; $i++) {
	    my ($v,$q) = valueAndQuality($pBuf->[$i]);
	    $seq .= $v;
	    $quals .= $q;
	}
	print $gh "@" . $name . "\n";
	print $gh substr($seq,0,$len) . "\n";
	print $gh "+$name\n";
	print $gh substr($quals,0,$len) . "\n";
    }
}

sub valueAndQuality {
    my ($quad) = @_;
    my @a = split("",$quad);
    my ($q1,$q2) = sort {$b cmp $a} @a;
    my $q = chr(33+ord($q1)-ord($q2));
    if ($q eq '!') {
	return ('N','!');
    } else {
	foreach my $i (0 .. 3) {
	    if ($a[$i] eq $q1) {
		return ($ACGT[$i],$q);
	    }
	}
    }
}

sub integrateRead {
    my ($seq1,$quals1,$d,$pBuf) = @_;
    # Extend seq and quals if needed
    my $sf = $d-scalar(@$pBuf);
    if ($sf > 0) {
	for (my $i=0; $i<$sf; $i++) {
	    push @$pBuf,'!!!!';
	}
    }
    # Deal with the intersection
    my $iLim = min(length($seq1)+$d,scalar(@$pBuf));
    for (my $i=$d; $i<$iLim; $i++) {
	my $q1 = substr($quals1,$i-$d,1);
	my $v1 = substr($seq1,$i-$d,1);
	my $j = $ACGT{$v1};
	my $q = substr($pBuf->[$i],$j,1);
	if ($q1 gt $q) {
	    substr($pBuf->[$i],$j,1) = $q;
	}
    }   
    # Add stuff on the end
    my $iLim2 = length($seq1)+$d;
    for (my $i=$iLim; $i<$iLim2; $i++) {
	my $v1 = substr($seq1,$i-$d,1);
	my $j = $ACGT{$v1};
	$pBuf->[$i] = '!!!!';
	substr($pBuf->[$i],$j,1) = substr($quals1,$i-$d,1);
    }
}
