#!/usr/bin/env perl

=head1 NAME

splitLayoutByStrain.pl

=head1 SYNOPSIS

splitLayoutByStrain.pl -l layout.gz strain1 strain2 ...
splitLayoutByStrain.pl -s sorted.gz [strain1 strain2 ...]
splitLayoutByStrain.pl -i imputed.gz [strain1 strain2 ...]

=head1 DESCRIPTION

Given a layout file (-l switch), output str/layout.gz for each str in
the strain list, containing only reads for that strain.

Given a sorted or imputed file (-s or -i switch), do the same, but
expect to find the strain list on the first line of the file.

Written for NCYC people.

=head1 AUTHOR

David Carter, dmc@sanger.ac.uk , August 2006

=head1 SEE ALSO

    extractLayoutRegion.pl
    pasteLayoutRegions.pl
    bin/createRDNAFiles

=cut

use Utility;
use Getopt::Std;
use File::Basename;
use strict;

my %OPT;
getopts('ils',\%OPT);
dieWithUsage() unless (scalar(keys %OPT) == 1);
if ($OPT{l}) {
    dieWithUsage() unless (@ARGV > 1);
}
if ($OPT{s} || $OPT{i}) {
    dieWithUsage() if (@ARGV < 1);
}
my $STRAINS_IN_FILE = !$OPT{l};

sub dieWithUsage {
    die("Usage: $0 [-l layout.gz strains | -s sorted.gz [strains] | -i imputed.gz [strains]]\n");
}

my $f = shift @ARGV;
my $base = basename($f);
my $fh = openToRead($f);

# Output streams:
my %out;

# For each output stream, a list of the currently
# active input positions.
my %map;
if ($STRAINS_IN_FILE) {
    chomp (my $line = <$fh>);
    my @strains = split(" ",substr($line,1));
    my %use = map {$_ => 1} @ARGV;
    foreach my $k (0 .. $#strains) {
	my $str = $strains[$k];
	unless (%use && !defined $use{$str}) {
	    my $g = "$str/$base";
	    unless (-s $g) {
		my $gh = $out{$str} = openToWrite($g);
		print $gh ">$str\n";
		$map{$str} = [$k];
	    }
	}
    }
} else {
    foreach my $str (@ARGV) {
	my $g = "$str/$base";
	unless (-s $g) {
	    $out{$str} = openToWrite($g);
	    $map{$str} = [];
	}
    }
}

exit unless (%out);

while (<$fh>) {
    chomp;
    if (/^>REF/) {
	foreach my $gh (values %out) {
	    print $gh "$_\n";
	}
	if (/0000$/) {
	    report($_,1);
	}
    } elsif (/^>(\S+) (\d+) /) {
	my ($read,$inPos) = ($1,$2);
	my ($str) = split(/-/,$read);
	next unless ($out{$str}); # maybe strain is not of interest
	# Place $inPos in the first available slot in $map{$str}.
	my $outPos;
	my $pA = $map{$str};
	foreach my $k (0 .. $#$pA) {
	    unless (defined $pA->[$k]) {
		$outPos = $k;
		$pA->[$k] = $inPos;
		last;
	    }
	}
	unless (defined $outPos) {
	    push @$pA,$inPos;
	    $outPos = $#$pA;
	}
	my $gh = $out{$str};
	print $gh ">$read $outPos\n";
    } else {
	while (my ($str,$gh) = each %out) {
	    my $pA = $map{$str};
	    my $line =  '  ' x scalar(@$pA);
	    my $kMax;
	    my $notAllGaps = (substr($_,0,1) =~ /[ACGTacgt]/);
	    foreach my $k (0 .. $#$pA) {
		if (defined $pA->[$k]) {
		    my $pair = substr($_,2+2*$pA->[$k],2);
		    if ($pair =~ /^\S/) {
			substr($line,2*$k,2) = $pair;
			$notAllGaps ||= (substr($pair,0,1) =~ /[ACGTacgt]/);
			$kMax = $k;
		    } else {
			$pA->[$k] = undef;
		    }
		}
	    }
	    # Shrink $pA if possible
	    while ($#$pA > $kMax) {
		pop @$pA;
	    }
	    # Prepend the reference onto the front of $line
	    if ($notAllGaps) {
		printf($gh "%s %s\n",substr($_,0,1),$line);
	    }
	}
    }
}

