#!/usr/bin/env perl

# Arg: base $b for files created by sraFilesToXml.pl, such that
# $b.{experiment,run,submission}.xml exist; and then any extra experiment.xml files
# need to define previously submitted experiments.

use XML::Simple;
use Utility;
use strict;

my $XS = XML::Simple->new();
my %OPTS_IN = (KeepRoot => 1, ForceArray => 1, KeyAttr => []);

my ($base,@extraExptFiles) = @ARGV;
unless (-s "$base.submission.header.csv") {
    sraSummarizeXML($base,@extraExptFiles);
}
	 
sub sraSummarizeXML {
    my ($base,@extraExptFiles) = @_;
    # Make sure we get experiment BEFORE run.
    my %exptData;
    foreach my $key (qw(submission experiment run)) {
	my $f = "$base.$key.xml";
	if (-s $f) {
	    my $pXML = $XS->XMLin($f,%OPTS_IN);
	    my ($key) = keys %$pXML;
	    if ($key eq 'SUBMISSION') {
		sraSummarizeSubmission($pXML,outputStreams($base,qw(header meta data)));
	    } elsif ($key eq 'EXPERIMENT_SET') {
		sraSummarizeExperiment($pXML,outputStreams($base,qw(experiments)),\%exptData);
		foreach my $eef (@extraExptFiles) {
		    my $pXMLEEF = $XS->XMLin($eef,%OPTS_IN);
		    sraSummarizeExperiment($pXMLEEF,undef,\%exptData);
		}
	    } elsif ($key eq 'RUN_SET') {
		sraSummarizeRuns($pXML,\%exptData,outputStreams($base,qw(runs lanes samples)));
	    } else {
		die("Cannot summarize $key in $f");
	    }
	} else {
	    die("$f: missing or empty");
	}
    }
}

# 1: header.csv: header
# 2: meta.csv: runs and experiments (XML files)
# 3: data.csv: data files manifest
sub sraSummarizeSubmission {
    my ($pXML,$gh1,$gh2,$gh3) = @_;
    sraSummarizeSubmission1($pXML,$gh1);
    sraSummarizeSubmission2($pXML,$gh2);
    sraSummarizeSubmission3($pXML,$gh3);
}

sub sraSummarizeSubmission1 {
    my ($pXML,$gh) = @_;
    my $pH = $pXML->{SUBMISSION}[0];
    my @keys;
    foreach my $key (sort keys %$pH) {
	push @keys,$key unless ($key =~ /^[A-Z]/ || $key eq 'handle' || $key eq 'xmlns:xsi');
    }
    printUnquoted($gh,@keys);
    my @vals;
    foreach my $key (@keys) {
	push @vals,$pH->{$key};
    }
    printQuoted($gh,@vals);
}
    
sub printUnquoted {
    my ($gh,@lst) = @_;
    print $gh join(",",@lst) . "\n" if ($gh);
}
   
sub printQuoted {
    my ($gh,@lst) = @_;
    printf($gh "\"%s\"\n",join('","',@lst)) if ($gh);
}

# Actions, i.e. XML files to be added.

sub sraSummarizeSubmission2 {
    my ($pXML,$gh) = @_;
    my @keys = qw(schema source);
    printUnquoted($gh,@keys);
    foreach my $pH (@{$pXML->{SUBMISSION}[0]{ACTIONS}[0]{ACTION}}) {
	my $pHA = $pH->{ADD}[0];
	my @vals = map {$pHA->{$_}} @keys;
	printQuoted($gh,@vals);
    }
}

sub sraSummarizeExperiment {
    my ($pXML,$gh,$pExptData) = @_;
    my @keys = qw(alias project expected_number_runs expected_number_spots expected_number_reads);
    printUnquoted($gh,@keys);
    my @phList = (@{$pXML->{EXPERIMENT_SET}[0]{EXPERIMENT}});
    foreach my $pH (@phList) {
	my @vals = map {$_ eq 'project' ? $pH->{STUDY_REF}[0]{refname} : $pH->{$_}} @keys;
	printQuoted($gh,@vals);
	my (undef,undef,$indiv,undef,$libNum) = split(/-/,$vals[0]);
	my $proj = $vals[1];
	$pExptData->{$vals[0]} = {sample => $indiv, library => "lib$libNum", project => $proj};
    }
}

# File manifest
sub sraSummarizeSubmission3 {
    my ($pXML,$gh) = @_;
    my @keys = qw(filename checksum);
    printUnquoted($gh,@keys);
    foreach my $pH (@{$pXML->{SUBMISSION}[0]{FILES}[0]{FILE}}) {
	my @vals;
	foreach my $key (@keys) {
	    push @vals,$pH->{$key};
	}
	printQuoted($gh,@vals);
    }
}

sub outputStreams {
    my ($base,@tags) = @_;
    my @ghList;
    foreach my $tag (@tags) {
	my $g = "$base.$tag.csv";
	report("$g: creating",1);
	push @ghList,new FileHandle(">$g");
    }
    return @ghList;
}

sub sraSummarizeRuns {
    my ($pXML,$pExptData,$gh1,$gh2,$gh3) = @_;
    # Runs
    my $pRunData = sraSummarizeRuns1($pXML,$pExptData,$gh1);
    # Lanes and samples
    sraSummarizeRuns2($pXML,$pRunData,$gh2,$gh3);
}

# Run level
sub sraSummarizeRuns1 {
    my ($pXML,$pExptData,$gh) = @_;
    my @keys1 = qw(alias instrument_model instrument_name run_center run_date total_data_blocks);
    my @keys2 = qw(actual_read_length total_bases);
    my @keys = (@keys1,'experiment_ref',@keys2);
    printUnquoted($gh,@keys);
    my $pRuns = $pXML->{RUN_SET}[0]{RUN};
    my @runData; # to pass back
    foreach my $pRun (@$pRuns) {
	my @vals;
	foreach my $key (@keys1) {
	    push @vals,$pRun->{$key};
	}
	my $eRef = $pRun->{EXPERIMENT_REF}[0]{refname};
	push @vals,$eRef;	    
	my %h;
	foreach my $pRA (@{$pRun->{RUN_ATTRIBUTES}[0]{RUN_ATTRIBUTE}}) {
	    $h{$pRA->{TAG}[0]} = $pRA->{VALUE}[0];
	}
	foreach my $key (@keys2) {
	    push @vals,$h{$key};
	}
	printQuoted($gh,@vals);
	my $pRD = $pExptData->{$eRef} || die("No experiment data for $eRef");
	$pRD->{read_length} = $h{actual_read_length};
	push @runData,$pRD;
    }
    return \@runData;
}

# Lane level ($gh) and sample level ($sh)
sub sraSummarizeRuns2 {
    my ($pXML,$pRunData,$gh,$sh) = @_;
    my @keys1 = qw(name run_alias sector total_reads total_spots filename);
    my @keys2 = qw(project sample library total_bases);
    printUnquoted($gh,@keys1,@keys2);
    my @runData = @$pRunData;
    my @mat;
    my %sampleSums;
    foreach my $pRun (@{$pXML->{RUN_SET}[0]{RUN}}) {
	my $pRunDataHere = shift @runData;
	my %h = (run_alias => $pRun->{alias});
	my %seenSample;
	foreach my $pDB (@{$pRun->{DATA_BLOCK}}) {
	    foreach my $key (@keys1) {
		if ($key eq 'filename') {
		    $h{$key} = $pDB->{FILES}[0]{FILE}[0]{filename};
		} elsif ($key ne 'run_alias') {
		    $h{$key} = $pDB->{$key};
		}
	    }
	    foreach my $key (@keys2) {
		if ($key eq 'total_bases') {
		    $h{$key} = $h{total_reads}*$pRunDataHere->{read_length};
		} else {
		    $h{$key} = $pRunDataHere->{$key};
		}
	    }
	    my @vals = map {$h{$_}} (@keys1,@keys2);
	    push @mat,\@vals;
	    my $sam = $h{sample};
	    $sampleSums{$sam}{lanes} ++;
	    $seenSample{$sam} = 1;
	    $sampleSums{$sam}{total_reads} += $h{total_reads};
	    $sampleSums{$sam}{total_bases} += $h{total_bases};
	}
	foreach my $sam (keys %seenSample) {
	    $sampleSums{$sam}{runs} ++;
	}
    }
    @mat = sort {$a->[5] cmp $b->[5]} @mat;
    foreach my $pRow (@mat) {
	printQuoted($gh,@$pRow);
    }
    my @sKeys = qw(sample runs lanes read_length reads gbases);
    printUnquoted($sh,@sKeys);
    foreach my $sam (sort keys %sampleSums) {
	my $pH = $sampleSums{$sam};
	my $nr = $pH->{total_reads};
	my $bases = $pH->{total_bases};
	my $rlen = $bases/$nr;
	my @row = ($sam,$pH->{runs},$pH->{lanes},$rlen,$nr,$bases/1000000000);
	printQuoted($sh,@row);
    }    
}
