#!/usr/bin/env perl

# Args should be .xls files.
# Switches:
#   -d dir (default "."): directory containing default.{submission,run,experiment}.xml files.

# http://search.cpan.org/dist/Spreadsheet-ParseExcel/lib/Spreadsheet/ParseExcel.pm
use Spreadsheet::ParseExcel;
# http://search.cpan.org/dist/XML-Simple/lib/XML/Simple.pm
use XML::Simple;
use Getopt::Std;
use FileHandle;
use File::Basename;
use strict;

my %OPT = qw(d .);
getopts('d:',\%OPT);
my $DEFAULT_XML_DIR = $OPT{d};
# Check DEFAULT_XML_DIR exists and contains the right files
unless (-d $DEFAULT_XML_DIR) {
    die("Cannot find default XML directory $DEFAULT_XML_DIR\n");
}
my %DEFAULT_XML;
foreach my $key (qw(submission run experiment)) {
    my $f = "$DEFAULT_XML_DIR/default.$key.xml";
    if (-s $f) {
	$DEFAULT_XML{$key} = $f;
    } else {
	die("$f: missing or empty");
    }
}

foreach my $f (@ARGV) {
    sraXlsToXml($f);
}

sub sraXlsToXml {
    my ($file) = @_;
    if ($file !~ /\.xls$/) {
	die("Usage: $0 [-d defaultXmlDir] foo.xls bar.xls ...");
    } elsif (-s $file) {
	sraXlsToXml1($file);
    } else {
	warn("$file: missing or empty");
    }
}

sub sraXlsToXml1 {
    my ($file) = @_;
    my $pSheet = xlsHash($file);
    my $xs = XML::Simple->new();
    my %opts1 = (KeepRoot => 1, ForceArray => 1, KeyAttr => []);
    my %opts2 = (KeepRoot => 1, KeyAttr => []);
    my $pXmlSub = $xs->XMLin($DEFAULT_XML{submission},%opts1);
    my $pXmlRun = $xs->XMLin($DEFAULT_XML{run},%opts1);
    my $pXmlExp = $xs->XMLin($DEFAULT_XML{experiment},%opts1);
    foreach my $pA (applySheetToXml($pSheet,$pXmlSub,$pXmlRun,$pXmlExp)) {
	my ($base,$pSub,$pRun,$pExp) = @$pA;
	writeToFile($xs->XMLout($pSub,%opts2),"$base.submission.xml");
	writeToFile($xs->XMLout($pRun,%opts2),"$base.run.xml");
	writeToFile($xs->XMLout($pExp,%opts2),"$base.experiment.xml");
    }
}

# Return $pXLS such that
# $pXLS->{$sheetName}[$rowIdx]{$colName} = val
# The colNames are the values in the top row.
sub xlsHash {
    my ($file) = @_;
    my $excel = Spreadsheet::ParseExcel::Workbook->Parse($file);
    my $g = "$file.txt";
    my $gh = -s $g ? undef : new FileHandle(">$g");
    my %xls;
    foreach my $sheet (@{$excel->{Worksheet}}) {
        my $name = $sheet->{Name};
        $sheet->{MaxRow} ||= $sheet->{MinRow};
	$sheet->{MaxCol} ||= $sheet->{MinCol};
	my @colName;
	my $hr = $sheet->{MinRow};
	foreach my $col ($sheet->{MinCol} ..  $sheet->{MaxCol}) {
	    my $cell = $sheet->{Cells}[$hr][$col];
	    if ($cell) {
		$colName[$col] = $cell->{Val};
	    }
	}
        foreach my $row ($hr+1 .. $sheet->{MaxRow}) {
	    my $idx = $row-$hr-1;
            foreach my $col ($sheet->{MinCol} ..  $sheet->{MaxCol}) {
                my $cell = $sheet->{Cells}[$row][$col];
                if ($cell && defined $cell->{Val} && $cell->{Val} ne '') {
		    $xls{$name}[$idx]{$colName[$col]} = $cell->{Val};
		    if ($gh) {
			my $nm = $name;
			my $cn = $colName[$col];
			$cn =~ tr/ /_/;
			printf($gh "%11s %2d %30s %s\n",$name,$idx,$cn,$cell->{Val});
		    }
		}
	    }
	}
    }
    return \%xls;
}

# The clever stuff...
sub applySheetToXml {
    my ($pSheet,$pDftSub,$pDftRun,$pDftExp) = @_;
    my @ans;
    foreach my $pRow (@{$pSheet->{Submission}}) {
	my $base = $pRow->{submission_id};
	my $pSub = editSubmissionXML($pRow,$base,$pSheet,copyHash($pDftSub));
	my $pRun = editRunXML($base,$pSheet,copyHash($pDftRun));
	my $pExp = editExperimentXML($base,$pSheet,copyHash($pDftExp));
	push @ans,[$base,$pSub,$pRun,$pExp];
    }
    return @ans;
}

sub editSubmissionXML {
    my ($pRow,$base,$pSheet,$pXML) = @_;
    # SUBMISSION attributes
    # submission_date is "NOW" by default only...
    $pXML->{SUBMISSION}[0]{submission_date} = xmlNowString(); # write this!!
    foreach my $key (qw(submission_id submission_date accession 
                        submission_comment center_name lab_name)) {
	$pXML->{SUBMISSION}[0]{$key} = $pRow->{$key} if (defined $pRow->{$key});
    }
    $pXML->{SUBMISSION}[0]{ACTIONS}[0]{ACTION}[0]{ADD}[0]{source} = "$base.experiment.xml";
    $pXML->{SUBMISSION}[0]{ACTIONS}[0]{ACTION}[1]{ADD}[0]{source} = "$base.run.xml";
    # CONTACTS
    foreach my $key (qw(contact_name inform_on_status inform_on_error)) {
	my $key2 = $key eq 'contact_name' ? 'name' : $key;
	$pXML->{SUBMISSION}[0]{CONTACTS}[0]{CONTACT}[0]{$key2} = $pRow->{$key} if (defined $pRow->{$key});
    }
    # FILES
    my @files;
    foreach my $pRow (@{$pSheet->{Files}}) {
	if ($pRow->{submission_id} eq $base) {
	    push @files,{checksum => $pRow->{checksum},
			 checksum_method => $pRow->{method} || "MD5",
			 filename => $pRow->{filename}};
	}
    }
    $pXML->{SUBMISSION}[0]{FILES}[0]{FILE} = \@files;
    return $pXML;
}
    
sub xmlNowString {
    my ($sec,$min,$hour,$mday,$mon,$year) = localtime(time());
    my $now = sprintf("%4d-%2d-%2dT%2d:%2d:%2d+01:00",
		      $year+1900,$mon+1,$mday,$hour,$min,$sec);
    $now =~ tr/ /0/;
    return $now;
}

sub copyHash {
    my ($pH1) = @_;
    my $pH2 = {};
    foreach my $key (keys %$pH1) {
	my $val = $pH1->{$key};
	$pH2->{$key} = copyValue($val);
    }
    return $pH2;
}

sub copyValue {
    my ($x) = @_;
    my $r = ref($x);
    if ($r eq 'ARRAY') {
	return copyArray($x);
    } elsif ($r eq 'HASH') {
	return copyHash($x);
    } else {
	return $x;
    }
}

sub copyArray {
    my ($pA1) = @_;
    my $pA2 = [];
    foreach my $i (0 .. $#$pA1) {
	my $val = $pA1->[$i];
	$pA2->[$i] = copyValue($val);
    }
    return $pA2;
}

sub editRunXML {
    my ($base,$pSheet,$pXML) = @_;
    my $pRunXMLOrig = $pXML->{RUN_SET}[0]{RUN}[0];
    my @runs;
    my %dataTrans = qw(run_name name 
                       lane sector
                       tile region);
    foreach my $pRow (@{$pSheet->{Runs}}) {
	if ($pRow->{submission_id} eq $base) {
	    my $pRunXML = copyHash($pRunXMLOrig);
	    # RUN attributes
	    foreach my $key (qw(flowcell instrument_model instrument_name run_date run_center
                                run_file total_data_blocks)) {
		# Is this translation right?
		my $key2 = $key eq 'flowcell' ? 'alias' : $key;
		$pRunXML->{$key2} = $pRow->{$key} if (defined $pRow->{$key});
	    }
	    # EXPERIMENT_REF
	    if (defined $pRow->{experiment_name}) {
		$pRunXML->{EXPERIMENT_REF}[0]{refname} = $pRow->{experiment_name};
	    }
	    # DATA_BLOCK
	    foreach my $key (qw(run_name lane tile total_spots total_reads number_channels
                                format_code)) {
		if (defined $pRow->{$key}) {
		    my $key2 = $dataTrans{$key} || $key;
                    $pRunXML->{DATA_BLOCK}[0]{$key2} = $pRow->{$key};
                }
            }
            # RUN_ATTRIBUTES
            my @runAttrs;
            foreach my $key ("run","total_bases (bp)","actual_read_length (bp)") {
                if (defined $pRow->{$key}) {
                   if ($key =~ /^(.*) \((.*)\)$/) {
                      push @runAttrs,{TAG => [$1], UNITS => [$2], VALUE => [$pRow->{$key}]};
                   } else {
                      push @runAttrs,{TAG => [$key], VALUE => [$pRow->{$key}]};
		   }
                }
	    }
            $pRunXML->{RUN_ATTRIBUTES}[0]{RUN_ATTRIBUTE} = \@runAttrs;
            # Store the <RUN ...>.
	    push @runs,$pRunXML;
	}
    }
    $pXML->{RUN_SET}[0]{RUN} = \@runs;
    return $pXML;
}

sub editExperimentXML {
    my ($base,$pSheet,$pXML) = @_;
    my $pExptXMLOrig = $pXML->{EXPERIMENT_SET}[0]{EXPERIMENT}[0];
    my @expts;
    foreach my $pRow (@{$pSheet->{Experiments}}) {
	if ($pRow->{submission_id} eq $base) {
	    my $pExptXML = copyHash($pExptXMLOrig);
	    # EXPERIMENT attributes
	    foreach my $key (qw(experiment_name expected_number_runs expected_number_spots expected_number_reads)) {
		# Is this translation right?
		my $key2 = $key eq 'experiment_name' ? 'alias' : $key;
		$pExptXML->{$key2} = $pRow->{$key} if (defined $pRow->{$key});
	    }
	    # STUDY_REF
	    if (defined $pRow->{study_name}) {
		$pExptXML->{STUDY_REF}[0]{refname} = $pRow->{study_name};
	    }
	    # DESIGN
	    if (defined $pRow->{design_description}) {
		$pExptXML->{DESIGN}[0]{DESIGN_DESCRIPTION}[0] = $pRow->{design_description};
	    }
	    if (defined $pRow->{sample_name}) {
		$pExptXML->{DESIGN}[0]{SAMPLE_DESCRIPTOR}[0]{refname} = $pRow->{design_description};
	    }
	    foreach my $key (qw(library_name library_strategy library_source library_selection library_construction_protocol)) {
		if (defined $pRow->{$key}) {
		    (my $tag = $key) =~ tr/a-z/A-Z/;
		    $pExptXML->{DESIGN}[0]{LIBRARY_DESCRIPTOR}[0]{$tag}[0] = $pRow->{$key};
		}
	    }
	    if ($pRow->{library_type} eq 'PAIRED') {
		$pExptXML->{DESIGN}[0]{LIBRARY_DESCRIPTOR}[0]{LIBRARY_LAYOUT}[0]{PAIRED}[0]{NOMINAL_LENGTH} = $pRow->{"insert_size (bp)"};
	    } else {
		die("library_type not PAIRED -- don't know what to do");
	    }
	    my $rl = $pRow->{"read length (bp)"};
	    $pExptXML->{DESIGN}[0]{SPOT_DESCRIPTOR}[0]{SPOT_DECODE_SPEC}[0]{READ_SPEC}[1]{BASE_COORD}[0] = $rl+1;
	    $pExptXML->{PLATFORM}[0]{ILLUMINA}[0]{CYCLE_COUNT}[0] = $rl;
	    my @eaList;
	    push @eaList,{TAG => 'center_name', VALUE => $pRow->{center_name} || 'SC'};
	    push @eaList,{TAG => 'expected_number_bases', 
			  UNITS => 'MB',
			  VALUE => $pRow->{"expected_number_bases (MB)"}};
	    $pExptXML->{EXPERIMENT_ATTRIBUTES}[0]{EXPERIMENT_ATTRIBUTE} = \@eaList;
	    push @expts,$pExptXML;
	}
    }
    $pXML->{EXPERIMENT_SET}[0]{EXPERIMENT} = \@expts;
    return $pXML;
}

sub writeToFile {
    my ($string,$file) = @_;
    my $gh = new FileHandle(">$file");
    print $gh $string;
    $gh->close();
}

foreach my $f (@ARGV) {
    sraXlsToXml($f);
}
