#!/usr/bin/env perl

use OffsetMap;
use Utility;
use Getopt::Std;
use strict;

# Generate either sets of points for a two-strain distance plot (to visualize
# in R or gnuplot, etc), or a tree for displaying with njplot or unrooted.
#
# Usage:
#  For distance plot:
#    strainDivergenceValues.pl species chr strain1 strain2 releaseDir
#  For tree:
#    strainDivergenceValues.pl [-s abbrev] species chr offset outBase releaseDir
#
# Switches: -s N: shorten strain names (for tree production). 
#                 N=1 to single letters, N=2 to minimal prefixes+suffixes.
#                 Expansions will appear in outBase.abbrevs
#
# In tree mode, "offset" is reference offset into chr, and "outBase" is
# base for output file creation: outBase.matrix, outBase.abbrevs and outBase.tree
#
# If "releaseDir" is omitted it defaults to the current directory.
#
# Note: scale of trees is multiplied by 1000, i.e. unit "1" really means 0.001.
#
# Examples:
#
#   strainDivergenceValues.pl cere chr01 SK1 Y55 
#
#   strainDivergenceValues.pl cere chr01 120000 tmpout 

my %OPT;
getopts('s:',\%OPT);

strainDivergenceValues(@ARGV);

sub strainDivergenceValues {
    my ($sp,$chr,@rest) = @_;

    if (@rest < 2 || @rest > 3) {
	die("Usage: $0 species chr (refOffset base|str1 str2) [releaseDir]\n");
    }
    my $doTree = ($rest[0] =~ /^[0-9]+$/);

    my $topDir = $rest[2] || ".";
    my $dir = "$topDir/$sp/match/$chr";
    die("No such directory: $dir\n") unless (-d $dir);

    my $sFile = "$dir/sequenced.gz";
    my $pStrains = getStrainNames($sFile);
    my $pOM = getOffsetMap(substr($chr,-1),"$topDir/$sp/strains/REF/assembly/genome.map");
    my $pMarkers = getMarkerOffsets("$dir/recombs/mgr_in","REF.$chr",$pOM);

    my $sddFile = "$dir/recombs/sdd.txt.gz";
    die("Missing or empty: $sddFile\n") unless (-s $sddFile);
    
    my $fh = openToRead($sddFile);
    if ($doTree) {
	strainDivergenceTree($fh,$pStrains,$pMarkers,$pOM->translation("REF.$chr",$rest[0]),$rest[1]);
    } else {
	strainDivergenceValues1($fh,$pStrains,$pMarkers,$rest[0],$rest[1],$sFile);
    }
}

sub getStrainNames {
    my ($sFile) = @_;
    my $fh = openToRead($sFile);
    chomp (my $hdr = <$fh>);
    return ['REF',split(" ",substr($hdr,1))];
}

sub getOffsetMap {
    my ($pat,$omFile) = @_;
    if (-s $omFile) {
	return new OffsetMap($pat,$omFile);
    } else {
	die("Missing or empty: $omFile\n");
    }
}

sub getMarkerOffsets {
    my ($f,$chr,$pOM) = @_;
    my $fh = openToRead($f);
    chomp (my $hdr = <$fh>);
    my (undef,$nm) = split(" ",$hdr);
    my @offs;
    foreach my $i (0 .. $nm-1) {
	chomp (my $off = <$fh>);
	$offs[$i] = $pOM->backTranslation($chr,$off,1);
    }
    return \@offs;
}

sub strainDivergenceValues1 {
    my ($fh,$pStrains,$pMarkers,$str1,$str2,$sFile) = @_;

    my ($idx1,$idx2);
    foreach my $i (0 .. $#$pStrains) {
	if ($pStrains->[$i] eq $str1) {
	    $idx1 = $i;
	} elsif ($pStrains->[$i] eq $str2) {
	    $idx2 = $i;
	}
    }
    die("Cannot find $str1 in header of $sFile") unless (defined $idx1);
    die("Cannot find $str2 in header of $sFile") unless (defined $idx2);
    ($idx1,$idx2) = sort {$a <=> $b} ($idx1,$idx2);
    
    my @dists;
    while (<$fh>) {
	chomp;
	my ($mkr,$s1,$s2,undef,$dist) = split;
	if ($s1 == $idx1 && $s2 == $idx2) {
	    $dists[$mkr] = $dist;
	}
    }

    foreach my $mkr (0 .. $#dists) {
	unless ($mkr > 0 && $mkr < $#dists && $dists[$mkr-1] eq $dists[$mkr]
		&& $dists[$mkr+1] eq $dists[$mkr]) {
	    print "$pMarkers->[$mkr] $dists[$mkr]\n";
	}
    }
}

sub strainDivergenceTree {
    my ($fh,$pStrains,$pMarkers,$off,$base) = @_;

    my $mkr;
    foreach my $i (0 .. $#$pMarkers) {
	last if ($pMarkers->[$i] > $off);
	$mkr = $i; # marker to use...
    }
    
    my @mat;
    while (<$fh>) {
	chomp;
	my ($m,$s1,$s2,undef,$d) = split;
	last if ($m > $mkr);
	if ($m == $mkr) {
	    $mat[$s1][$s2] = $mat[$s2][$s1] = $d;
	}
    }
    
    my $ns = scalar(@$pStrains);
    my $matFile = "$base.matrix";
    my $gh = openToWrite($matFile);
    print $gh "$ns\n";
    my $pShort = shortenNames($pStrains,$base);
    foreach my $i (0 .. $#$pShort) {
	my $str = $pShort->[$i];
	print $gh $str;
	foreach my $j (0 .. $#$pShort) {
	    printf($gh " %10.8f",1000*$mat[$i][$j]);
	}
	print $gh "\n";
    }
    $gh->close();
    `quicktree -in m -out t $matFile > $base.tree`;
    unless (-s "$base.tree") {
	print STDERR "Empty: $base.tree; did you install quicktree?\n";
    }
}

sub shortenNames {
    my ($pNames,$base) = @_;
    my $gh = openToPipe("sort > $base.abbrevs");
    my $pAns;
    if ($OPT{s} == 1) {
	$pAns = shortenNamesToLetters($pNames);
    } elsif ($OPT{s} == 2) {
	$pAns = shortenNamesToPrefSufs($pNames);
    } else {
	my @ans;
	foreach my $name (@$pNames) {
	    $name =~ s/^UWOPS/UW/;
	    push @ans,$name;
	}
	$pAns = \@ans;
    }
    foreach my $i (0 .. $#$pNames) {
	print $gh "$pAns->[$i] $pNames->[$i]\n";
    }
    return $pAns;
}

sub shortenNamesToLetters {
    my ($pNames) = @_;
    my @ans;
    foreach my $i (0 .. $#$pNames) {
	push @ans,chr($i+($i>=26 ? 71 : 65));
    }
    return \@ans;
}

sub shortenNamesToPrefSufs {
    my ($pNames) = @_;
    my (%count,%versions,%isName);
    foreach my $name (@$pNames) {
	$isName{$name} = 1;
	my $len = length($name);
	my %short;
	foreach my $i (1 .. $len-1) {
	    my $pfx = substr($name,0,$i);
	    foreach my $j ($i .. $len-1) {
		my $sfx = substr($name,$j);
		$short{"$pfx$sfx"} = 1;
	    }
	}
	$versions{$name} = [keys %short];
	foreach my $ver (keys %short) {
	    $count{$ver} ++;
	}
    }
    my @ans;
  NAME:
    foreach my $name (@$pNames) {
	foreach my $ver (sort {length($a) <=> length($b)} @{$versions{$name}}) {
	    if ($count{$ver} == 1 && !$isName{$ver}) {
		push @ans,$ver;
		next NAME;
	    }
	}
	push @ans,$name;
    }
    return \@ans;
}
