#!/usr/bin/env perl

# Outliers in Heng's summary_lib.txt (etc) files

# See RD's message of 20080606: first runs, then lanes, then libraries.
#
#   org              SC, BGI, BI etc
#   run              run ID (a number when SC)
#   lane             lane ID in lane file; NUMBER of lanes in other files
#   sample_id        NAnnnnn; in summary_run.txt, no. of different samples in the run
#   lib              library name; in summary_run.txt, no. of different libraries in the run,
#                    and ditto in summary_sample.txt
#   l_read1          length of first read
#   l_read2          length of second read
#   ins_mean         mean length of insert
#   ins_sd           sd length of insert
#   n_reads          read count, in millions
#   n_mapped         mapped read count, in millions
#   n_paired         mapped-in-pairs read count, in millions
#   err_rate         difference rate (fraction) for mapped reads
#   dup_rate         duplication rate (fraction)
#   gc               GC proportion

use Utility;
use Getopt::Std;
use strict;

# -d: directory containing files (default: ".")
# -s: Sanger (SC) data only.

my %OPT = qw(d .);
getopts('d:s',\%OPT);

my %libsPerRun;
recordLibsPerRun($OPT{d},\%libsPerRun);

my (%badRun,%badLaneCount);
screen($OPT{d},'run',\%badRun,\%badLaneCount,\%libsPerRun);
screen($OPT{d},'lane',\%badRun,\%badLaneCount);
screen($OPT{d},'lib',\%badRun,\%badLaneCount);

sub recordLibsPerRun {
    my ($d,$pLPR) = @_;
    my $fh = openToRead("$d/summary_lane.txt");
    <$fh>;
    while (<$fh>) {
	chomp;
	my ($org,$run,undef,undef,$lib) = split;
	next if ($OPT{s} && $org ne 'SC');
	$pLPR->{join("_",$org,$run)}{$lib} ++;
    }
}

sub screen {
    my ($dir,$type,$pBadRun,$pBadLaneCount,$pLPR) = @_;

    my $fh = openToRead("$dir/summary_$type.txt");
    my @names;
    chomp ($_ = <$fh>);
    @names = split;
    my @mat;
    while (<$fh>) {
	chomp;
	my @vals = split;
	my %h;
	foreach my $i (0 .. $#vals) {
	    $h{$names[$i]} = $vals[$i];
	}
	if ($h{org} =~ /^\d+$/) {
	    ($h{org}) = ($h{lib} =~ /^lib([^_]+)_/);
	}
	next if ($OPT{s} && $h{org} ne 'SC');
	if ($type eq 'lane' && $pBadRun->{join('_',$h{org},$h{run})}) {
	    $pBadLaneCount->{join('_',$h{org},$h{lib})} ++;
	    #next;
	}
	$h{id} = $type eq 'lib' ? join('_',$h{org},$h{lib}) :
	    $type eq 'run' ? join("_",$h{org},$h{run}) :
	    join("_",$h{org},$h{run},$h{lane});
	push @mat,\%h unless ($OPT{s} && $h{org} ne 'SC');
    }

    foreach my $pH (sort {$a->{id} cmp $b->{id}} @mat) {
	my $ol;
	my $flib = join('_',$pH->{org},$pH->{lib});
	my $pm = $pH->{n_mapped}/max($pH->{n_reads},0.0001);
	my $id = $pH->{id};
	my $diags;
	if ($pm < 0.9) {
	    my $rel = $pm < 0.8 ? '<<' : '<';
	    $diags .= sprintf("# n_mapped / n_reads = %5.3f %s 0.9\n",$pm,$rel);
	    $ol = max($ol,length($rel));
	}
	my $pp = $pH->{n_paired}/max($pH->{n_mapped},0.0001);
	if (($pp > 0 || $pH->{lib} =~ /^libSC/) && $pp < 0.9) {
	    my $rel = $pp < 0.8 ? '<<' : '<';
	    $diags .= sprintf("# n_paired / n_mapped = %5.3f %s 0.9\n",$pp,$rel);
	    $ol = max($ol,length($rel));
	}
	#if ($pH->{gc} < 0.35 || $pH->{gc} > 0.44) {
	#    printf("# gc = %5.3f not in range 0.35 to 0.44\n",$pH->{gc});
	#    $ol = 1;
	#}
	if ($pH->{err_rate} > 0.05) {
	    my $rel = $pH->{err_rate} > 0.075 ? '>>' : '>';
	    $diags .= sprintf("# err_rate = %5.3f %s 0.05\n",$pH->{err_rate},$rel);
	    $ol = max($ol,length($rel));
	}
	if ($type eq 'lib' && $pH->{dup_rate} > 0.2) { # my figure was 0.07
	    my $rel = $pH->{dup_rate} > 0.4 ? '>>' : '>';
	    $diags .= sprintf("# dup_rate = %5.3f %s 0.07\n",$pH->{dup_rate},$rel);
	    $ol = max($ol,length($rel));
	}
	my $rpl = $pH->{n_reads}/($type eq 'lane' ? 1 : $pH->{lane});
	if ($rpl < 7.5) {
	    my $rel = $rpl < 5 ? '<<' : '<';
	    $diags .= sprintf("# Reads per lane = %5.2f %s 7.5\n",$rpl,$rel);
	    $ol = max($ol,length($rel));
	}
	#if ($pH->{ins_mean} > 0 && $pH->{ins_mean} < 90) {
	#    printf("# ins_mean = %5.1f < 90\n",$pH->{ins_mean});
	#    $ol = 1;
	#}
	#if ($pH->{ins_sd} > 0 && ($pH->{ins_sd} < 10 || $pH->{ins_sd} > 40)) {
	#    printf("# ins_sd = %5.1f not in range 10 to 40\n",$pH->{ins_sd});
	#    $ol = 1;
	#}
	if ($ol) {
	    printf("# %s problem(s) with %s %s\n",$ol == 2 ? 'SERIOUS' : 'Minor',$type,$id);
	    if ($type eq 'lib') {
		my $nb = $pBadLaneCount->{$flib};
		printf("# This library has %d/%d lanes identified as problematic\n",
		       $nb,$pH->{lane});
	    } elsif ($type eq 'run') {
		$pBadRun->{$id} = 1;
		my @lst = sort keys %{$pLPR->{$id}};
		printf("# Librar%s for this run:",
		       @lst > 1 ? 'ies' : 'y');
		foreach my $lib (@lst) {
		    printf(" %s(%d)",$lib,$pLPR->{$id}{$lib});
		}
		print "\n";
	    } elsif ($type eq 'lane') {
		$pBadLaneCount->{$flib} += 1;
	    }
	    print $diags;
	    foreach my $i (0 .. $#names) {
		my $key = $names[$i];
		my $val = $pH->{$key};
		my $len = max(length($key),length($val));
		print " " if ($i);
		printf("%${len}s",$key);
	    }
	    print "\n";
	    foreach my $i (0 .. $#names) {
		my $key = $names[$i];
		my $val = $pH->{$key};
		my $len = max(length($key),length($val));
		print " " if ($i);
		printf("%${len}s",$val);
	    }
	    print "\n\n";
	}
    }
}
