#!/usr/bin/env perl

use FileHandle;

use Getopt::Std;
use strict;

# Usage:

# tabulateSNPsByWindows.pl [-h] [-p] [-c] strainSpecList
#   -h: high-qual only (40 or more; probably want this with -c or -v)
#   -p: polymorphic sites only
#   -c: chi-square test (see (3) below)
#   -v: variability (see (4) below)
#   -d: number of differences per 10k to count as "not close". Default 5.
# strainSpecList: space-separated list, three formats:
#   (1) when one member is ":", should have one or more strains before
#       colon and one or more after. First strain before is the one from
#       which we report distances to each of the others. Other strains
#       before may or may not be prefixed by '^' (with no space in between).
#       If they have a '^', we only report blocks where the first strain
#       is "far from" those (at least $DIFF differences per 10k). Without a
#       '^', we only report when "close to".
#   (2) When no colon: everything in strainSpecList is a strain name,
#       and we print table of percentages of blocks where difference
#       rate is less than -d.
#   (3) When -c given: look at the blocks where the first and second
#       strains in the list are not close, count blocks that are close
#       to each of the N other strains, and do an Nx2 chi-squared test.
#       No colon should be given.
#   (4) When -v given: again no colon. For each strain pair, print
#       number of blocks where both strains had (high-quality) nucleotides
#       in at least 2000 positions out of the 10000 in a block; 
#       mean similarity over those blocks;
#       sd of similarity; and variance/mean, which appears to be the best
#       indicator of one of the strains being mosaic. Lines are sorted by 
#       decreasing variance/mean. Only strain pairs with at least 100 
#       qualifying blocks are shown.
# If a strain occurs more than once in strainSpecList, only the first
# occurrence is counted.

my %OPT = qw(d 5);
getopts('hpcd:v',\%OPT);

autoflush STDOUT 1;

tabulateSNPsByWindows($OPT{h},$OPT{p},$OPT{c},$OPT{v},$OPT{d},@ARGV);

sub tabulateSNPsByWindows {
    my ($isHQ,$isPoly,$isCS,$isVar,$dFar,@strains) = @_;
    my ($pStrains1,$pIsNeg,$pStrains2) = splitStrains($isCS,uniquify(@strains));
    tabulateSNPsByWindows1($isHQ,$isPoly,$isCS,$isVar,$dFar,$pStrains1,$pIsNeg,$pStrains2);
}

sub uniquify {
    my @lst = @_;
    my (@ans,%seen);
    foreach my $x (@lst) {
	unless ($seen{$x}) {
	    $seen{$x} = 1;
	    push @ans,$x;
	}
    }
    return @ans;
}

sub splitStrains {
    my ($isCS,@strains) = @_;
    if ($isCS) {
	if (@strains < 3) {
	    die("Usage: $0 -c str1a str1b str2a str2b ...\n");
	} else {
	    my @str1 = @strains[0..1];
	    my @str2 = @strains[2..$#strains];
	    my %isNeg = ($strains[1],1);
	    return (\@str1,\%isNeg,\@str2);
	}
    }
    if (@strains < 2) {
	die("Usage: $0 [-h] [-p] [-c] str1a str1b ... [:] str2a str2b ...\n");
    }
    foreach my $i (0 .. $#strains) {
	if ($strains[$i] eq ':') {
	    if ($i == 0 || $i == $#strains) {
		die("Usage: $0 [-h] [-p] str1a str1b ... : str2a str2b ...\n");
	    } else {
		my (%isNeg,@str1);
		foreach my $str (@strains[0..$i-1]) {
		    if (substr($str,0,1) eq '^') {
			$isNeg{substr($str,1)} = 1;
			push @str1,substr($str,1);
		    } else {
			push @str1,$str;
		    }
		}
		my @str2 = @strains[$i+1..$#strains];
		return (\@str1,\%isNeg,\@str2);
	    }
	}
    }
    return (\@strains,undef,undef);
}    

sub tabulateSNPsByWindows1 {
    my ($isHQ,$isPoly,$isCS,$isVar,$dFar,$pStrains1,$pIsNeg,$pStrains2) = @_;
    my (%strIdx1,%strIdx2);
    foreach my $k (0 .. $#$pStrains1) {
	$strIdx1{$pStrains1->[$k]} = $k;
    }
    if (defined $pStrains2) {
	foreach my $k (0 .. $#$pStrains2) {
	    $strIdx2{$pStrains2->[$k]} = $k;
	}
    }
    my (@hist,@snpSum);
    my $nIdx = $isHQ ? 3 : 0;
    my $dIdx = $nIdx + ($isPoly ? 1 : 2);
    my %simList;
    foreach my $f (glob("p?/chr??/sbw.txt"),glob("chr??/sbw.txt")) {
	if (defined $pStrains2) {
	    my ($pBad,$pLines) = findBadBlocks($dFar,$nIdx,$dIdx,$f,$pStrains1->[0],\%strIdx1,$pIsNeg,\%strIdx2);
	    collectCounts($isPoly,$isCS,$nIdx,$dIdx,$pStrains1->[0],$pStrains1->[1],\%strIdx2,$pLines,$pBad,\@hist,\@snpSum);
	} elsif ($isVar) {
	    collectSimilarities($f,$nIdx,$dIdx,\%strIdx1,\%simList);
	} else {
	    my $fh = new FileHandle($f);
	    while (<$fh>) {
		chomp;
		my ($s1,$s2,undef,$block,@counts) = split;
		my $idx1 = $strIdx1{$s1};
		my $idx2 = $strIdx1{$s2};
		if (defined $idx1 && defined $idx2) {
		    my $n = $counts[$nIdx];
		    my $d = $counts[$dIdx];
		    if ($d >= 50) {
			my $far = ($n*10000 >= $d*$dFar);
			$hist[min($idx1,$idx2)][max($idx1,$idx2)][$far] ++;
		    }
		}
	    }
	}
	#print ".";
    }
    #print "\n";
    if ($isCS) {
	printCSTable($isHQ,$pStrains1,$pStrains2,\@hist);
    } elsif ($isVar) {
	printVarTable(\%simList);
    } elsif ($pStrains2) {
	printResults($isHQ,$isPoly,$pStrains1,$pIsNeg,$pStrains2,\@hist,\@snpSum);
    } else {
	printCloseTable($pStrains1,\%strIdx1,\@hist);
    }
}

sub collectSimilarities {
    my ($f,$nIdx,$dIdx,$pWanted,$pSimList) = @_;
    my $fh = new FileHandle($f);
    while (<$fh>) {
	chomp;
	my ($str1,$str2,undef,undef,@counts) = split;
	if (defined $pWanted->{$str1} && defined $pWanted->{$str2}) {
	    my $d = $counts[$dIdx];
	    if ($d >= 2000) {
		my $n = $counts[$nIdx];
		my $sim = $d/($n+1);
		push @{$pSimList->{$str1}{$str2}},$sim;
	    }
	}
    }
}

sub min {
    my ($x,$y) = @_;
    if (defined $y && $x > $y) {
	return $y;
    } elsif (defined $x) {
	return $x;
    } else {
	return $y;
    }
}
   
sub max {
    my ($x,$y) = @_;
    if (defined $y && $y > $x) {
	return $y;
    } elsif (defined $x) {
	return $x;
    } else {

	return $y;
    }
}

sub findBadBlocks {
    my ($dFar,$nIdx,$dIdx,$f,$str1,$pStrIdx1,$pIsNeg,$pStrIdx2) = @_;
    my (@bad,@lines);
    my $fh = new FileHandle($f);
    my %use = (%$pStrIdx1,%$pStrIdx2);
    while (<$fh>) {
	chomp;
	my ($s1,$s2,undef,$block,@counts) = split;
	my $close;
	if ($s1 eq $str1 && defined $pStrIdx1->{$s2}) {
	    $close = $pIsNeg->{$s2} ? -1 : 1;
	} elsif ($s2 eq $str1 && defined $pStrIdx1->{$s1}) {
	    $close = $pIsNeg->{$s1} ? -1 : 1;
	}
	if ($close) {
	    my $n = $counts[$nIdx];
	    my $d = $counts[$dIdx];
	    # $close > 0 means they have to be close for a good block
	    if ($d < 50 || ($n*10000 >= $d*$dFar) == ($close > 0)) {
		$bad[$block] = 1;
	    }
	}
	if (defined $use{$s1} && defined $use{$s2}) {
	    push @lines,$_;
	}
    }
    return (\@bad,\@lines);
}

sub collectCounts {
    my ($isPoly,$isCS,$nIdx,$dIdx,$str1,$str2,$pStrIdx2,$pLines,$pBad,$pHist,$pSNPSum) = @_;
    foreach my $line (@$pLines) {
	my ($s1,$s2,undef,$block,@counts) = split(" ",$line);
	next if ($pBad->[$block]);
	my ($idx,$csIdx);
	if ($isCS) {
	    if ($s1 eq $str1) {
		$idx = $pStrIdx2->{$s2};
		$csIdx = 0;
	    } elsif ($s1 eq $str2) {
		$idx = $pStrIdx2->{$s2};
		$csIdx = 1;
	    } elsif ($s2 eq $str1) {
		$idx = $pStrIdx2->{$s1};
		$csIdx = 0;
	    } elsif ($s2 eq $str2) {
		$idx = $pStrIdx2->{$s1};
		$csIdx = 1;
	    }
	} else {
	    if ($s1 eq $str1) {
		$idx = $pStrIdx2->{$s2};
	    } elsif ($s2 eq $str1) {
		$idx = $pStrIdx2->{$s1};
	    }
	}
	if (defined $idx) {
	    my $n = $counts[$nIdx];
	    my $d = $counts[$dIdx];
	    if ($d >= 50) {
		my $b;
		my $nn = $n*10000/$d;
		if ($isPoly) {
		    $b = int(20*$n/$d);
		} else {
		    $b = int(min($nn,100)/5);
		}
		if (!$isCS) {
		    $pSNPSum->[$idx] += $nn;
		    $pHist->[$b][$idx] ++;
		} elsif ($b == 0) {
		    $pHist->[$csIdx][$idx] ++;
		}
	    }
	}
    }
}

sub printResults {
    my ($isHQ,$isPoly,$pStrains1,$pIsNeg,$pStrains2,$pHist,$pSNPSum) = @_;
    my ($h,@t) = @$pStrains1;
    print "Difference counts with respect to $h";
    if (@t) {
	printf(" when %s is ",$h);
	my @conj;
	foreach my $str (@t) {
	    push @conj,sprintf("%s %s",$pIsNeg->{$str} ? 'far from' : 'close to',$str);
	}
	print join(" and ",@conj);
    }
    print "\n";
    if ($isHQ) {
	print "Quality score threshold: 40\n";
    } else {
	print "Quality score threshold: none\n";
    }
    if ($isPoly) {
	print "Proportions of polymorphic sites\n";
	print "             ";
    } else {
	print "Number of sites per 10,000\n";
	print "               ";
    }
    my $iMax = $#$pStrains2;
    #unlink(glob("tmpxy.*"));
    my @out;
    my $bMax = $isPoly ? 19 : 20;
    my $infix = join('_',@$pStrains1);
    foreach my $i (0 .. $iMax) {
	printf(" %5s",$pStrains2->[$i]);
	$out[$i] = new FileHandle(">hist.$infix.$pStrains2->[$i]");
    }
    print "\n";
    my @sum;
    foreach my $b (0 .. $bMax) {
	my $pA = $pHist->[$b];
	foreach my $i (0 .. $iMax) {
	    $sum[$i] += $pA->[$i];
	}
    }
    foreach my $b (0 .. $bMax) {
	my $pA = $pHist->[$b];
	if ($OPT{p}) {
	    printf("%4.2f to %4.2f ",$b*0.05,($b+1)*0.05);
	} elsif ($b < 20) {
	    printf("%3d to %3d.999 ",$b*5,$b*5+4);
	} else {
	    printf("%3d or more    ",$b*5);
	}
	foreach my $i (0 .. $iMax) {
	    my $w = max(0,length($pStrains2->[$i])-5);
	    my $prop = $pA && $sum[$i] ? $pA->[$i]/$sum[$i] : 0;
	    printf(" %${w}s%5.3f","",$prop);
	    my $gh = $out[$i];
	    printf($gh "%4.2f %8.6f\n",0.025+$b/20,$prop);
	}
	print "\n";
    }
    printf("Totals       ");
    print "  " unless ($OPT{p});
    foreach my $i (0 .. $iMax) {
	my $w = max(5,length($pStrains2->[$i]));
	printf(" %${w}d",$sum[$i]);
    }
    print "\n";
    printf("Means        ");
    print "  " unless ($OPT{p});
    foreach my $i (0 .. $iMax) {
	my $w = max(5,length($pStrains2->[$i]));
	my $mu = $pSNPSum->[$i]/($sum[$i] || 1);
	if ($isPoly) {
	    my $dw = $w-5;
	    printf(" %${dw}s%5.3f","",$mu/10000);
	} else {
	    printf(" %${w}.1f",$mu);
	}
    }
    print "\n";
}

sub printCloseTable {
    my ($pStrains1,$pStrIdx1,$pHist) = @_;
    printf("%13s","");
    foreach my $str (@$pStrains1) {
	printf(" %s%s",substr($str,0,1),substr($str,-1,1));
    }
    print "\n";
    foreach my $str1 (@$pStrains1) {
	printf("%13s",$str1);
	my $idx1 = $pStrIdx1->{$str1};
	foreach my $str2 (@$pStrains1) {
	    my $idx2 = $pStrIdx1->{$str2};
	    if ($idx1 == $idx2) {
		print "   ";
	    } else {
		my $idx2 = $pStrIdx1->{$str2};
		my $pA = $pHist->[min($idx1,$idx2)][max($idx1,$idx2)];
		my $prop = $pA->[0]/($pA->[0]+$pA->[1]);
		printf(" %2d",int($prop*100));
	    }
	}
	print "\n";
    }
}

sub printCSTable {
    my ($isHQ,$pStrains1,$pStrains2,$pHist) = @_;
    printf("Chi-squared test for when %s and %s are not close\n",@$pStrains1);
    if ($isHQ) {
	print "Quality score threshold: 40\n";
    } else {
	print "Quality score threshold: none\n";
    }
    my (@hSum,@vSum);
    my $jMax = $#$pStrains2;
    foreach my $i (0,1) {
	foreach my $j (0 .. $jMax) {
	    $hSum[$i] += $pHist->[$i][$j];
	    $vSum[$j] += $pHist->[$i][$j];
	}
    }
    my $sum = $hSum[0]+$hSum[1];
    my $cs = 0;
    my @exp;
    foreach my $j (0 .. $jMax) {
	foreach my $i (0,1) {
	    $exp[$i][$j] = $hSum[$i]*$vSum[$j]/$sum;
	}
    }
    my ($cs,@csa);
    foreach my $j (0 .. $jMax) {
	foreach my $i (0,1) {
	    my $o = $pHist->[$i][$j];
	    my $e = $exp[$i][$j];
	    if ($e >= 5) {
		$csa[$j] += ($o-$e)*($o-$e)/$e;
	    }
	}
	$cs += $csa[$j];
    }
    printf("Chi-squared (%d degrees of freedom) = %5.3f\n",
	   $jMax+1,$cs);
    print "Values in brackets do not contribute to chi-squared value\n\n";

    my %strIdx2 = map {$pStrains2->[$_] => $_} (0 .. $#$pStrains2);
    my @sorted = sort {$csa[$strIdx2{$a}] <=> $csa[$strIdx2{$b}]} @$pStrains2;

    printf("       Strain  Sim to %5s  Sim to %5s\n",
	   substr($pStrains1->[0],0,5),
	   substr($pStrains1->[1],0,5));
    print "                 Obs    Exp    Obs    Exp      CS      Cumul\n";

    my $cumul = 0;
    foreach my $str (@sorted) {
	printf("%13s",$str);
	my $j = $strIdx2{$str};
	foreach my $i (0,1) {
	    printCell($exp[$i][$j] >= 5,6,$pHist->[$i][$j] || 0);
	    printCell($exp[$i][$j] >= 5,6,int($exp[$i][$j]));
	}
	$cumul += $csa[$j];
	printf(" %8.2f %9.2f\n",$csa[$j],$cumul);
    }
}
    
sub printCell {
    my ($good,$w,$n) = @_;
    if ($good) {
	$w --;
	printf(" %${w}d ",$n);
    } else {
	$w -= 2+length($n);
	printf(" %${w}s(%d)","",$n);
    }
}

sub printVarTable {
    my ($pSimList) = @_;
    my @stats;
    foreach my $str1 (keys %$pSimList) {
	my $pH = $pSimList->{$str1};
	foreach my $str2 (keys %$pH) {
	    my $key = sprintf("%13s %13s",$str1,$str2);
	    my $pA = $pH->{$str2};
	    my $n0 = scalar(@$pA);
	    if ($n0 >= 100) {
		my ($n1,$n2);
		foreach my $v (@$pA) {
		    $n1 += $v;
		    $n2 += $v*$v;
		}
		my $mu = $n1/$n0;
		my $var = $n2/$n0-$mu*$mu;
		my $sd = $var > 0 ? sqrt($var) : 0;
		my $r = $var/$mu;
		push @stats,[$key,$n0,$mu,$sd,$r];
	    }
	}
    }
    printf("%13s %13s %6s %5s %5s %7s\n",
	   "Strain1","Strain2","Blocks","Mean","SD","Var/Mean");
    foreach my $pA (sort {$b->[3] <=> $a->[3]} @stats) {
	printf("%s %6d %5d %5d %7.2f\n",@$pA);
    }
}
