#!/usr/bin/env perl

use GD::Simple;
use File::Basename;
use Getopt::Std;
use Utility;
use strict;

my %PATTERN_TOPOLOGY =
    qw(011 1
       101 2
       110 3
       111 4
       100 5
       010 6
       001 7);       

# The number of double mutations we'll tolerate rather than
# switch the consensus:

my $TRANSITION_COST = 3.5;

my $DATA_WIDTH = 1000;
my $ANNO_WIDTH = 50;
# Height of a single track, e.g. for "A = B"
my $TRACK_HEIGHT = 80;
my $MARGARITA_COUNT = 2; # one for ARGs, one for error count histogram
my $TRACK_COUNT = 7 + $MARGARITA_COUNT;
my $DATA_HEIGHT = $TRACK_HEIGHT*$TRACK_COUNT;
my $AXIS_HEIGHT = $DATA_HEIGHT+10; # axis is 10 below data
# Height of the legend, consensus topology bar and ticks
my $ANNO_HEIGHT = 40;

my @COLOURS = qw(0 red green blue orange brown purple black);

# 0 for MMM/argNNN.mgr.gz, 1 for MMM/argNNN_ca (not implemented), 2 for args_bl
# 2 currently broken (bad topologies...?)
my $TOPOLOGY_SOURCE = 0;

# Must be consistent with the one in Margarita.pm ...
my $MARKER_OVERLAP_SIZE = 20;

my $ER_BUCKET_SIZE = 25;

my $MGR_TYPE = 'mgr'; # could be 'mgr' or 'mgrLite'

# Args: -q for no ARG processing;
# -s is for subdir (relative to $dir) in which NNN dirs are found and into which
# we write the .png files
# offset positions for our 4 strains (first = reference = 0), separated by commas,
# then dir(s) to process; or ranges separated by hyphens.

my %OPT;
getopts('qs:',\%OPT);

my $spec = shift @ARGV;
unless (@ARGV) {
    die("Usage: e.g. $0 [-q] n1,n2,n3,n4 dir1 dir2 ...");
}

my @nsList;
foreach my $item (split(/,/,$spec)) {
    if ($item =~ /^(\d+)-(\d+)$/) {
	push @nsList,($1 .. $2);
    } else {
	push @nsList,$item;
    }
}

foreach my $d (@ARGV) {
    for (my $k=0; $k+3 <= $#nsList; $k+=4) {
	topologyGraphs($d,@nsList[$k .. $k+3]);
    }
}

sub topologyGraphs {
    my ($d,@quartet) = @_;
    my $mf = "$d/mgr_in";
    my $af = $TOPOLOGY_SOURCE ? "$d/args_bl" : "$d/args.gz"; # not essential
    if (-s $mf) {
	topologyGraphs1($mf,$af,$d,@quartet);
    }
}

sub topologyGraphs1 {
    my ($mf,$af,$d,@quartet) = @_;
    my ($mh,$ns,$nm) = openMarkerFile($mf,@quartet);
    return unless ($mh);

    # Get the marker locations:
    my $pLocation = getMarkerLocations($mh,$nm);
    my ($pStrains,$pAbsent) = getStrainNames($d,$pLocation,@quartet);

    my $type = qw(mg ca bl)[$TOPOLOGY_SOURCE];
    my $dir = $OPT{s} ? "$d/$OPT{s}" : $d;
    my $imFile = sprintf("%s/topol.%s.%s.png",$dir,join('_',@quartet),$type);
    report("will create $imFile for strains @quartet = @$pStrains",1);

    my $im = createGraph(basename(absoluteFileName($d)),$pStrains);
    my $pTopList = findTopologies($mh,$nm,@quartet);
    #truncateTopologies($pLocation,$pTopList);
    my ($pTopList22,$pIList22) = 
	drawDataPoints($pLocation,$pTopList,$pAbsent,$im);

    # Find chi-square statistic and report for 0.05 level
    my $cs = chiSquared($pTopList22);
    printf("CS = %4.2f from %d transitions\n",$cs,$#$pTopList22) if ($cs > 9.488);

    # Plot ARG average topologies (unless -q)
    plotARGTopologies($af,$d,$im,$ns,$#$pLocation,@quartet) unless ($OPT{q});
    # Draw horizontal separators AFTER drawing the data
    drawHorizontalSeparators($im);
    # Figure out consensus.
    decideConsensus($pIList22,$pTopList22,$DATA_WIDTH/$#$pTopList,$im);
    # Write the file
    report("$imFile: writing");
    writeGraphicsFile($im,$imFile);
    report();
}

sub openMarkerFile {
    my ($mf,@quartet) = @_;
    my $mh = openToRead($mf);
    chomp ($_ = <$mh>);
    my ($ns,$nm) = split;
    foreach my $q (@quartet) {
	if ($ns < $q) {
	    printf(STDERR "%s has %s strains, need at least %d\n",
		   $mf,$ns,$quartet[3]+1);
	    return;
	}
    }
    return ($mh,$ns,$nm);
}

sub getStrainNames {
    my ($d,$pLocation,@quartet) = @_;
    my $fh = openToRead("$d/sorted.gz");
    my $line = <$fh>;
    chomp $line;
    my @names = ('REF',split(' ',substr($line,1)));
    my @strains;
    foreach my $q (@quartet) {
	push @strains,$names[$q];
    }
    # Find the "missing" regions for each strain in the quartet,
    # relative to $pLocation.
    my $off = 0; # offset in sorted.gz file
    my $iOff = 0; # index in $pLocation
    my @absentSince;
    my @missing;
  LOOP:
    while (<$fh>) {
	chomp;
	$off++;
	while ($off > $pLocation->[$iOff]) {
	    $iOff ++;
	    last LOOP if ($iOff > $#$pLocation);
	}
	my $len = length($_);
	foreach my $q (0 .. $#quartet) {
	    my $k2 = $quartet[$q]*2;
	    my $v = ($k2 >= $len) ? ' ' : substr($_,$k2,1);
	    if ($v eq ' ') {
		unless (defined $absentSince[$q]) {
		    $absentSince[$q] = $iOff;
		}
	    } elsif (defined $absentSince[$q]) {
		if ($absentSince[$q] < $iOff) {
		    push @{$missing[$q]},[$absentSince[$q],$iOff-1];
		}
		$absentSince[$q] = undef;
	    }
	}
    }
    foreach my $q (0 .. $#quartet) {
	if ($absentSince[$q]) {
	    push @{$missing[$q]},[$absentSince[$q],$iOff-1];
	}
    }
    return (\@strains,\@missing);
}

sub getMarkerLocations {
    my ($mh,$nm) = @_;
    my @location;
    for (my $i=0; $i<$nm; $i++) {
	push @location,scalar(<$mh>);
    }
    chomp @location;
    return \@location;
}

# Graph layout:
# x = 0 to $DATA_WIDTH: data
#     $DATA_WIDTH to $DATA_WIDTH+$ANNO_WIDTH: annotations

sub createGraph {
    my ($chr,$pStrains) = @_;
    my $fullHeight = $DATA_HEIGHT+$ANNO_HEIGHT;
    my $im = new GD::Simple($DATA_WIDTH+$ANNO_WIDTH,$fullHeight);
    # Main annotation for plot
    $im->bgcolor('white');
    $im->moveTo(25,$fullHeight-5);
    my $title = "Topology for $chr, strains";
    foreach my $q (0 .. 3) {
	$title .= sprintf(" %s=%s",chr(65+$q),$pStrains->[$q]);
	$title .= "," if ($q < 3);
    }
    $im->string($title);
    # ARG consensus
    $im->moveTo($DATA_WIDTH+5,$TRACK_HEIGHT*0.5);
    $im->string("ARG tops");
    $im->moveTo($DATA_WIDTH+5,$TRACK_HEIGHT*1.5);
    $im->string("Errors");
    # Topology 1 (red) at top of display: 0 with 1
    # Topology 2 (green): 0 with 2
    # Topology 3 (blue): 0 with 3
    for (my $n=1; $n<=3; $n++) {
	$im->moveTo($DATA_WIDTH+5,$TRACK_HEIGHT*($n+$MARGARITA_COUNT-0.5));
	$im->string(sprintf("A = %s",chr($n+ord('A'))));
    }
    # Topology 4 (orange): 0 alone
    # Topology 5 (brown): 1 alone
    # Topology 6 (purple): 2 alone
    # Topology 7 (black): 3 alone
    for (my $n=0; $n<=3; $n++) {
	$im->moveTo($DATA_WIDTH+5,$TRACK_HEIGHT*($n+$MARGARITA_COUNT+3.5));
	$im->string(sprintf("%s alone",chr($n+ord('A'))));
    }
    return $im;
}

sub drawHorizontalSeparators {
    my ($im) = @_;
    $im->fgcolor('black');
    $im->moveTo(0,$AXIS_HEIGHT);
    $im->lineTo($DATA_WIDTH,$AXIS_HEIGHT);
    for (my $i=1; $i<$TRACK_COUNT; $i++) {
	my $y = $i*$TRACK_HEIGHT;
	$im->moveTo(0,$y);
	$im->lineTo($DATA_WIDTH,$y);
    }
}

# Returns list of topologies
sub findTopologies {
    my ($mh,$nm,@quartet) = @_;
    my (@locList,@topList);
    for (my $i=0; $i<$nm; $i++) {
	chomp ($_ = <$mh>);
	my $t = findTopology($_,@quartet);
	push @topList,$t;
    }
    return \@topList;
}

sub drawDataPoints {
    my ($pLocation,$pTopList,$pAbsent,$im) = @_;

    # Add shading for absent regions
    my $xScale = $DATA_WIDTH/$#$pTopList;
    my $grey = $im->colorAllocate(230,230,230);
    my $white = $im->colorAllocate(255,255,255);
    $im->fgcolor($white);
    $im->bgcolor($grey);
    foreach my $k (0 .. 3) {
	foreach my $pA (@{$pAbsent->[$k]}) {
	    my ($lo,$hi) = @$pA;
	    my $y = ($k+3+$MARGARITA_COUNT)/$TRACK_COUNT;
	    my $yp = $y*$DATA_HEIGHT;
	    $im->rectangle($lo*$xScale,$yp,
			   $hi*$xScale,$yp+$DATA_HEIGHT/$TRACK_COUNT);
	}
    }
    # Also draw union regions over the 2x2 tracks
    my $yp22Lo = $DATA_HEIGHT*$MARGARITA_COUNT/$TRACK_COUNT;
    my $yp22Hi = $DATA_HEIGHT*(3+$MARGARITA_COUNT)/$TRACK_COUNT;
    foreach my $pA (mergeOverlappingRegions($pAbsent)) {
	my ($lo,$hi) = @$pA;
	$im->rectangle($lo*$xScale,$yp22Lo,
		       $hi*$xScale,$yp22Hi);
    }
    my @tList22; # list of interesting topologies
    my @iList22; # and their offsets
    for (my $i=0; $i<=$#$pTopList; $i++) {
	my $t = $pTopList->[$i];
	my $xp = $i*$xScale;
	mayDrawVertical($i,$pLocation,$xp,$im);
	next if ($t == 0);
	if ($t <= 3) {
	    push @tList22,$t;
	    push @iList22,$i;
	}
	my $y = ($t-1+$MARGARITA_COUNT+rand())/$TRACK_COUNT;
	my $yp = $y*$DATA_HEIGHT;
	$im->moveTo($xp,$yp);
	$im->bgcolor($COLOURS[$t]);
	$im->fgcolor($COLOURS[$t]);
	my $sz = 2; # could be: $t <= 3 ? 4 : 2;
	$im->ellipse($sz,$sz);
    }
    return (\@tList22,\@iList22);
}

sub mergeOverlappingRegions {
    my ($pAbsent) = @_;
    my @regs;
    foreach my $k (0 .. 3) {
	push @regs,@{$pAbsent->[$k]};
    }
    @regs = sort {$a->[0] <=> $b->[0] || $a->[1] <=> $b->[1]} @regs;
    my @union = shift @regs;
    foreach my $pA (@regs) {
	if ($pA->[0] <= $union[-1][1]+1) {
	    if ($pA->[1] > $union[-1][1]) {
		$union[-1][1] = $pA->[1];
	    }
	} else {
	    push @union,$pA;
	}
    }
    return @union;
}

sub mayDrawVertical {
    my ($i,$pLocation,$xp,$im) = @_;
    if ($i > 0) {
	my $rPrev = int($pLocation->[$i-1]/10000);
	my $rHere = int($pLocation->[$i]/10000);
	for (my $r=$rPrev+1; $r<=$rHere; $r++) {
	    # Draw tick for every 10000
	    my $h = $r%5 ? 4 : $r%10 ? 8 : $AXIS_HEIGHT;
	    my $xr = $xp+$r-$rHere; # adjust xp slightly
	    $im->moveTo($xr,$AXIS_HEIGHT);
	    $im->fgcolor('black');
	    $im->lineTo($xr,$AXIS_HEIGHT-$h);
	}
    }
}

sub chiSquared {
    my ($pLst) = @_;
    return 0 unless (@$pLst > 5);
    my (@n,@sumR,@sumC);
    for (my $i=1; $i<=$#$pLst; $i++) {
	$n[$pLst->[$i-1]][$pLst->[$i]] ++;
	$sumR[$pLst->[$i-1]] ++;
	$sumC[$pLst->[$i]] ++;
    }
    my $sum = $#$pLst;
    my $cs = 0;
    foreach my $i (1..3) {
	foreach my $j (1 ..3) {
	    my $ex = $sumR[$i]*$sumC[$j]/$sum;
	    if ($ex >= 5) {
		my $d = $n[$i][$j]-$ex;
		$cs += $d*$d/$ex;
	    }
	}
    }
    return $cs;
}

# Topology is 1,2,3 if 0 goes with 1,2 or 3 in a 2+2;
# or 4,5,6,7 if 0,1,2,3 is 1:3 against the other three.
# Value of 0 means not all are present.

sub findTopology {
    my ($line,@quartet) = @_;
    #$LINES++;
    $line =~ tr/a-z/A-Z/;
    my (@eq,@val);
    foreach my $j (0 .. 3) {
	my $v = substr($line,$quartet[$j]*2,1);
	return 0 if ($v eq ' ' || !$v);
	$v =~ tr/acgt/ACGT/;
	@val[$j] = $v;
    }
    my $pat; # pattern of vals
    my $vDiff; # first DIFFERENT value
    foreach my $j (1 .. 3) {
	if ($val[$j] eq $val[0]) {
	    $pat .= 0;
	} elsif (!$vDiff) {
	    $pat .= 1;
	    $vDiff = $val[$j];
	} elsif ($val[$j] eq $vDiff) {
	    $pat .= 1;
	} else {
	    return 0; # multiway mutation
	}
    }
    my $t = $PATTERN_TOPOLOGY{$pat};
    unless ($t || $pat eq '000') {
	die("Weird pattern $pat for $line");
    }
    #printf("%s %s\n",join("",@val),$pat);
    return $t;
}

sub decideConsensus {
    my ($pIList,$pTList,$scale,$im) = @_;
    my @cost = ([undef,0,0,0]);
    my @ptr = ([]);
    for (my $k=1; $k<=$#$pTList; $k++) {
	my $v = $pTList->[$k];
	foreach my $j (1,2,3) {
	    my ($iMin,$cMin);
	    foreach my $i (1,2,3) {
		my $c = ($j == $v ? 0 : 1) + ($i == $j ? 0 : $TRANSITION_COST) + $cost[$k-1][$i];
		if ($i == 1 || $c < $cMin) {
		    $iMin=$i;
		    $cMin=$c;
		}
	    }
	    $cost[$k][$j] = $cMin;
	    $ptr[$k][$j] = $iMin;
	}
    }
    my ($v,$cMin);
    foreach my $j (1,2,3) {
	my $c = $cost[-1][$j];
	if ($j == 1 || $c < $cMin) {
	    $v = $j;
	    $cMin = $c;
	}
    }
    my $hi = $pIList->[-1];
    for (my $k=$#$pTList; $k>=0; $k--) {
	my $w = $ptr[$k][$v];
	if ($w != $v || $k == 0) {
	    my $lo = $k == 0 ? $pIList->[$k] : ($pIList->[$k-1]+$pIList->[$k])/2;
	    $im->moveTo($lo*$scale,$AXIS_HEIGHT+7);
	    $im->fgcolor($COLOURS[$v]);
	    $im->penSize(5);
	    $im->lineTo($hi*$scale,$AXIS_HEIGHT+7);
	    $hi = $lo;
	    $v = $w;
	}
    }
}

sub writeGraphicsFile {
    my ($im,$f) = @_;
    my $gh = openToWrite($f);
    binmode $gh;
    print $gh $im->png;
    $gh->close();
}

# $af is args_bl file

sub plotARGTopologies {
    my ($af,$d,$im,$ns,$maxMarker,@quartet) = @_;
    my ($pVote,$pErHist);
    if ($TOPOLOGY_SOURCE) {
	$pVote = getARGTopologiesBL($af,$ns,$maxMarker,@quartet);
    } else {
	($pVote,$pErHist) = getARGTopologiesMGR($d,$ns,$maxMarker,@quartet);
    }
    plotARGTopologies1($pVote,$pErHist,$maxMarker,$im) if ($pVote);
}

sub getARGTopologiesMGR {
    my ($d,$ns,$maxMarker,@quartet) = @_;
    # Should be consistent with isCircularChromosome in Margarita.pm ...
    my $circ = (basename($d) !~ /^c.*[0-9]$/);
    my $delta = 0;
    my (@vote,@erHist);
    my ($padBelow,$padAbove);
    my $pat = $OPT{s} ? "$d/$OPT{s}/m[0-9][0-9][0-9]" : "$d/m[0-9][0-9][0-9]";
    my @subs = glob($pat);
    foreach my $i (0 .. $#subs) {
	my $padBelow = $circ || $i>0;
	my $padAbove = $circ || $i<$#subs;
	$delta = getARGTopologiesMGR1($subs[$i],$ns,$padBelow,$padAbove,$delta,\@vote,\@erHist,@quartet);
    }
    return (\@vote,\@erHist);
}

sub getARGTopologiesMGR1 {
    my ($d,$ns,$padBelow,$padAbove,$delta,$pVote,$pErHist,@quartet) = @_;
    my $fh = openToRead("$d/mgr_in");
    chomp ($_ = <$fh>);
    my (undef,$nmHere) = split;
    my $deltaToUse = $padBelow ? $delta-$MARKER_OVERLAP_SIZE : $delta;
    foreach my $f (glob("$d/arg[0-9][0-9][0-9].$MGR_TYPE")) {
	getARGTopologiesMGR2($f,$deltaToUse,$padBelow,$padAbove,$nmHere,$pVote,$pErHist,@quartet);
    }
    return $deltaToUse + $nmHere - ($padAbove ? $MARKER_OVERLAP_SIZE : 0);
}

# Assume $f contains exactly one ARG ...
sub getARGTopologiesMGR2 {
    my ($f,$deltaToUse,$padBelow,$padAbove,$nmHere,$pVote,$pErHist,@quartet) = @_;
    report("processing $f");
    my $fh = openToRead($f);
    # Value in $mthr is [$lMthr,$rMthr,$recombPoint] or [$mthr]
    my @mthr;
    while (<$fh>) {
	chomp;
	my (undef,$op,@vals) = split;
	if ($op eq 're') {
	    $mthr[$vals[0]] = [$vals[1],$vals[2],$vals[3]];
	} elsif ($op eq 'co') {
	    $mthr[$vals[0]] = $mthr[$vals[1]] = [$vals[2]];
	} elsif ($op eq 'mu') {
	    $mthr[$vals[0]] = [$vals[1]];
	} elsif ($op eq 'er') {
	    $mthr[$vals[0]] = [$vals[1]];
	    my $bucket = int(($deltaToUse+$vals[2])/$ER_BUCKET_SIZE);
	    $pErHist->[$bucket] ++;
	}
    }
    # $descs{$m} is a list of descendant records, where each such
    # record is [$q,$lo,$hi], meaning $q (a member of @quartet) is
    # a descendant of this node over the interval $lo to $hi.
    my %descs;
    my @todo;
    foreach my $q (@quartet) {
	push @todo,[$q,$q,0,$nmHere-1];
    }
    while (@todo) {
	my $pA = shift @todo;
	my ($j,$q,$lo,$hi) = @$pA;
	if (defined $mthr[$j] && $lo<=$hi) {
	    my $pM = $mthr[$j];
	    if ($#$pM) { # it's a recomb
		my ($lm,$rm,$loc) = @$pM;
		my $hi2 = min($hi,$loc);
		if ($hi2 >= $lo) {
		    push @{$descs{$lm}},[$q,$lo,$hi2];
		    push @todo,[$lm,$q,$lo,$hi2];
		}
		my $lo2 = max($lo,$loc+1);
		if ($lo2 <= $hi) {
		    push @{$descs{$rm}},[$q,$lo2,$hi];
		    push @todo,[$rm,$q,$lo2,$hi];
		}
	    } else { # it's a single mother
		my ($m) = @$pM;
		push @{$descs{$m}},[$q,$lo,$hi];
		push @todo,[$m,$q,$lo,$hi];
	    }
	}
    }
    # Regions we care about, in coords of this ARG
    my $loCare = $padBelow ? $MARKER_OVERLAP_SIZE : 0;
    my $hiCare = $nmHere - ($padAbove ? $MARKER_OVERLAP_SIZE : 0);
    # Now use the descs records
    my @donePosition;
    foreach my $pTrips (values %descs) {
	my %qSeen;
	my %lowers;
	foreach my $pT (@$pTrips) {
	    my ($q,$lo,$hi) = @$pT;
	    $qSeen{$q} = 1;
	    $lowers{$lo} = 1;
	    $lowers{$hi+1} = 1;
	}
	next if (keys %qSeen < 2); # efficiency: need at least two
	my @loList = sort {$a <=> $b} keys %lowers;
	for (my $k=0; $k<$#loList; $k++) {
	    # Define the interval:
	    my $lo = $loList[$k];
	    my $hi = $loList[$k+1]-1;
	    my %qPres;
	    foreach my $pT (@$pTrips) {
		if ($pT->[1] <= $lo && $pT->[2] >= $hi) {
		    $qPres{$pT->[0]} = 1;
		}
	    }
	    if (keys %qPres == 2) {
		my $t = decideTopology(\%qPres,\@quartet);
		foreach my $k (max($lo,$loCare) .. min($hi,$hiCare)) {
		    unless ($donePosition[$k]) {
			$pVote->[$k+$deltaToUse][$t] ++;
			$donePosition[$k] = $t;
		    }
		}
	    }
	}
    }
    report();
    if (0) {
	my $g = $f;
	$g =~ s/gz$/pats/;
	my $gh = openToWrite($g);
	foreach my $v (@donePosition) {
	    print $gh ($v || '.');
	}
	print $gh "\n";
	$gh->close();
    }
}

sub getARGTopologiesBL {
    my ($af,$ns,@quartet) = @_;
    return unless (-s $af);
    my $fh = openToRead($af);
    my @vote;
    while (<$fh>) {
	chomp;
	my ($lo,undef,$hi,undef,@lengthsAndMothers) = split;
	my @mthr;
	for (my $i=0; 2*$i<$#lengthsAndMothers; $i++) {
	    $mthr[$i] = $lengthsAndMothers[2*$i+1];
	}
	my %descs;
	foreach my $i (@quartet) {
	    my $j=$i;
	    while ($j <= $#mthr) {
		my $m = $mthr[$j];
		push @{$descs{$m}},$i;
		$j=$m;
	    }
	}
	foreach my $pA (values %descs) {
	    if (@$pA == 2) {
		my %got = map {$_ => 1} @$pA;
		my $t = decideTopology(\%got,\@quartet);
		for (my $k=$lo; $k<=$hi; $k++) {
		    $vote[$k][$t] ++;
		}
		last;
	    }
	}
    }
    return \@vote;
}

sub plotARGTopologies1 {
    my ($pVote,$pErHist,$maxMarker,$im) = @_;
    my $sum=0;
    for (my $t=1; $t<=3; $t++) {
	$sum += $pVote->[0][$t]; # assume same ARG count for all markers!
    }
    my $xScale = $DATA_WIDTH/$maxMarker;
    my $yScale = $TRACK_HEIGHT/$sum;
    # Plot as lines:
    $im->penSize(1);
    for (my $i=0; $i<=$#$pVote; $i++) {
	my $xp = $xScale*$i;
	my $yLo = 0;
	for (my $t=1; $t<=3; $t++) {
	    my $v = $pVote->[$i][$t];
	    if ($v > 0) {
		my $ypLo = $yScale*$yLo;
		my $yHi = $yLo+$v;
		my $ypHi = $yScale*$yHi;
		$im->moveTo($xp,$ypLo);
		$im->fgcolor($COLOURS[$t]);
		$im->lineTo($xp,$ypHi);
		$yLo = $yHi;
	    }
	}
    }
    $im->fgcolor('pink'); # for error rectangles
    $im->bgcolor('pink');
    # Theoretical maximum count is one er per ARG (=sum) per place in the bucket.
    my $max = $sum*$ER_BUCKET_SIZE;
    my $yp0 = $TRACK_HEIGHT*2;
    for (my $i=0; $i<$#$pErHist; $i++) {
	my $x1 = $ER_BUCKET_SIZE*$i;
	my $x2 = $x1+$ER_BUCKET_SIZE;
	my $xp1 = $xScale*$x1;
	my $xp2 = $xScale*$x2;
	my $y = $pErHist->[$i];
	my $yp = $TRACK_HEIGHT*(2-$y/$max);
	$im->rectangle($xp1,$yp,$xp2,$yp0);
    }
}

sub isBigJump {
    my ($old,$new,$range) = @_;
    my $d = $old > $new ? $old-$new : $new-$old;
    return 2*$d > $range;
}

sub decideTopology {
    my ($pGot,$pQuartet) = @_;
    my $pat4;
    foreach my $i (@$pQuartet) {
	$pat4 .= $pGot->{$i} ? '1' : '0';
    }
    my $pat = substr($pat4,1);
    $pat =~ tr/01/10/ if ($pat4 =~ /^1/);
    my $t = $PATTERN_TOPOLOGY{$pat};
    if ($t >= 1 && $t <= 3) {
	return $t;
    } else {
	die("Bad topology, should be 2+2");
    }
}

