#!/usr/bin/env perl

# Roger's NPG code ...
use lib qw(/software/solexa/lib/site_perl/5.8.8);
use st::api::project;
use npg::api::run;

use Utility;
use File::Basename;
use Getopt::Std;

use strict;

autoflush STDOUT 1;

# -v: report even when not actually doing anything
# -b, -r and -s: specify batch (integer), run (3 digits) or sample (e.g. NA12891-WGv2); can be comma-separated
# -f: from batch id (for when run falls over)
# -n: only do batches with index finishing in $n (digit); used by updateDataDirectory10
# -p: project (50 for low cov, 10 for trios)
# -S: "staging" directory: contains fastq, fastqcheck and md5 files from /staging, plus SOURCES.txt
#     listing all the source files.
# -L: file containing list of fastq[.gz] files, e.g. from collectSlxFastq.pl
# -o: output (log) file

# -A: call augmentDataDirectory instead. Args should then be summary_lane.txt, org name (e.g. BI),
#     and one or more (basenames of) repository directories. See augmentDataDirectory below for examples.

# If -S is given, we look for .fastq[check] and _srf.md5 files there before trying the relevant repository dir.
# The fastq[check] files there will only have their first four lines, though.
# Staging directories should be populated by populateStagingDirectory.pl.

# Use ~/bin/updateDataDirectory10 for faster throughput. Good way to do it:
#   cd $G1K/DATA
#   collectSlxFastq.pl > slx_fastq.txt
#   updateDataDirectory10 -L slx_fastq.txt

my $DATA_DIR;
foreach my $d (qw(/lustre/thougen/thougen/DATA
                  /lustre/scratch1/dmc/g1k)) {
    if (-d $d) {
	$DATA_DIR = $d;
    }
}
die("Cannot find a valid DATA_DIR") unless ($DATA_DIR);

my %PROJECT_NAMES = qw(10 Trio 50 LowCov);

my $LOG_STREAM;

my %SLX_FASTQ;

my %OPT;
getopts('Avb:f:L:n:o:p:qr:s:S:',\%OPT);

if ($OPT{A}) {
    augmentDataDirectory(@ARGV);
} else {
    updateDataDirectory(\%OPT);
}

sub updateDataDirectory {
    my ($pOpt) = @_;
    if ($pOpt->{o}) {
	$LOG_STREAM = openToWrite($pOpt->{o});
	autoflush $LOG_STREAM 1;
    }

    # Check no args left
    if (@ARGV) {
	die("Usage: $0 [-b batches] [-f fromBatchIndex] [-p project] [-q] [-r runs] [-s samples] [-S stagingDir]\n");
    }

    setSlxFastq($pOpt);

    # Set project, batch, run and sample and restrictions
    my ($pProjectOnly,@hashes) = map {commaListToHash($pOpt->{$_})} qw(p b r s);

    my $pStaging = readStaging($pOpt->{S});
    my $pTrioSubpop = setSubpopulations();

    if ($pOpt->{q} && $pOpt->{L}) {
	foreach my $d (glob("$DATA_DIR/*/*/libSC*/lane*")) {
	    my $b = basename($d);
	    my $fqb = sprintf("%s_s_%s.fastq",substr($b,4,3),substr($b,8,1));
	    my $fq = "$d/$fqb";
	    my $tgt = $SLX_FASTQ{$fqb};
	    if (-l $fq && ! -s readlink($fq)) {
		report2("removing dangling link $fq",1);
		unlink($fq);
	    }
	    if (-l $fq) {
		# OK
	    } elsif (defined $tgt) {
		report2("linking $fq to $tgt");
		symlink($tgt,$fq);
		report2();
		linkOrCreateFastqCheck($fq,$tgt);
	    } else {
		report2("cannot find target for $fq",1);
	    }
	    my $rlFile = sprintf("%s/%s_s_%s.readlengths",$d,substr($b,4,3),substr($b,8,1));
	    createReadLengthsFile($tgt,$rlFile);
	}
    } else {
	# Gorilla = 33
	# 10000Genomes-A1-LowCoveragePilot = 50
	# 10000Genomes-A2-TriosPilot = 10
	foreach my $id (keys %PROJECT_NAMES) {
	    if (passesFilter($id,$pProjectOnly)) {
		applyToLanes($id,\&updateDataDirectory2,
			     $pOpt,@hashes,$pStaging,$pTrioSubpop);
	    }
	}
    }
}

sub linkOrCreateFastqCheck {
    my ($src,$tgt) = @_;
    foreach my $suf (qw(fastq fastq.gz fq fq.gz)) {
	$src =~ s/\.$suf$//;
	$tgt =~ s/\.$suf$//;
    }
    my $srcc = "$tgt.fastqcheck";
    my $tgtc = "$tgt.fastqcheck";
    if (-s $tgtc) {
	return if (-l $srcc && readlink($srcc) eq $tgtc);
	unlink($srcc);
	symlink($tgtc,$srcc);
    } elsif (! -s $srcc) {
	unlink($srcc);
	report2("$srcc: creating as stub");
	createFastqCheckStub($tgt,$srcc);
	report2();
    }
}    

# Just create the first line of a fastqcheck file.
sub createFastqCheckStub {
    my ($fq,$fqc) = @_;
    my ($ns,$nb,$max);
    my $fh = openToRead($fq);
    while (<$fh>) {
	chomp(my $seq = <$fh>);
	<$fh>;
	<$fh>;
	$ns ++;
	my $len = length($seq);
	$nb += $len;
	$max = max($max,$len);
    }
    my $gh = openToWrite($fqc);
    printf($gh "%d sequences, %d total length, %4.2f average, %d max\n",
	   $ns,$nb,$nb/$ns,$max);
    $gh->close();
}

sub setSlxFastq {
    my ($pOpt) = @_;
    my $f = $pOpt->{L};
    if ($f && -s $f) {
	my $fh = openToRead($f);
	while (<$fh>) {
	    chomp;
	    $SLX_FASTQ{basename($_,'.gz')} = $_;
	}
    }
}

sub passesFilter {
    my ($key,$pH) = @_;
    return $pH ? $pH->{$key} : 1;
}

sub setSubpopulations {

    # Might need CHB and/or JPT too...
    my @ceuTrios = qw(NA12878 NA12891 NA12892);
    my @yri_trios = qw();

    my %trioSubpop;
    foreach my $indiv (@ceuTrios) {
	$trioSubpop{$indiv} = 'CEU';
    }
    foreach my $indiv (@yri_trios) {
	$trioSubpop{$indiv} = 'YRI';
    }
    return \%trioSubpop;
}

sub commaListToHash {
    my ($expr) = @_;
    return defined $expr ? {map {$_ => 1} split(",",$expr)} : undef;
}

# $dir/SOURCES.txt should contain a list of the files (from sflogin:/staging/...) 
# used to create the other files in $dir.

sub readStaging {
    my ($dir) = @_;
    return undef unless ($dir);
    unless (-d $dir) {
	die("-S $dir: directory does not exist\n");
    }
    my $f = "$dir/SOURCES.txt";
    unless (-s $f) {
	die("-S $dir: SOURCES.txt is missing or empty\n");
    }
    my %staging;
    my $fh = openToRead($f);
    while (<$fh>) {
	chomp;
	my @a = split("/",$_);
	my ($run,$pos) = ($a[-1] =~ /^(\d+)_s_(\d)/);
	if ($run) {
	    pop @a;
	    $staging{$run}[$pos] = join("/",@a);
	}
    }
    return \%staging;
}

sub updateDataDirectory2 {
    my ($pOpt,$runId,$pos,$hierDir,$reposDir,$missingOK) = @_;

    my $name = "${runId}_s_$pos";
    my $fqBase = "$name.fastq";
    my $fqFile = "$reposDir/$fqBase";
    my $slFile = "$hierDir/$fqBase";
    if (-s $fqFile || $fqFile =~ /^\/staging/) {
	if (-l $slFile) {
	    my $old = readlink($slFile);
	    if ($old eq $fqFile) {
		report2("$slFile: already exists",1) if ($pOpt->{v});
		return $pos > 1;
	    } else {
		report2("$slFile: MOVING from $old",1);
		unlink($slFile);
	    }
	}
	mkdir_p($hierDir);
	symlink($fqFile,$slFile);
	my $rlFile = "$hierDir/$name.readlengths";
	createReadLengthsFile($fqFile,$rlFile);
	my $adv = $missingOK ? ' UNEXPECTEDLY' : '';
	report2("$slFile: linking to $fqFile$adv",1);
	return 1; # try next lane
    } else {
	if ($pOpt->{v}) {
	    if ($missingOK) {
		report2("$fqFile does not exist, but is not expected to",1);
	    } else {
		reportMissingFastq($fqFile,$missingOK);
	    }
	}
	return $pos > 1; # save time if pos==1: assume other lanes won't hold anything either
    }
}

sub reportMissingFastq {
    my ($fqFile) = @_;
    (my $stFile = $fqFile) =~ s/_s_(\d).fastq$/_\1.status/;
    if (-s $stFile) {
	my $date = localtime(modTime($stFile));
	my $fh = openToRead($stFile);
	my $status;
	while (<$fh>) {
	    chomp;
	    if (/^Status:\s+(.*)$/) {
		$status = $1;
		last;
	    }
	}
	report2("$fqFile does not exist yet. Status '$status' from $date",1);
    } else {
	report2("$fqFile does not exist yet. No status file",1);
    }
}

sub decideDataSubdirectory {
    my ($id,$runId,$pos,$sampleName,$pTrioSubpop) = @_;
    my $projName = $PROJECT_NAMES{$id} || die("No project name found for $id");
    my ($indiv,$subpop,$lib) = decideSubdirectoryComponents($projName,$sampleName,$pTrioSubpop);
    my $projAndSubpop = sprintf("%s-%s",$projName,$subpop);
    return join('/',$DATA_DIR,$projAndSubpop,$indiv,"lib$lib","lane${runId}_$pos");
}

sub decideSubdirectoryComponents {
    my ($projName,$sampleName,$pTrioSubpop) = @_;
    my ($indiv,$subpop,$lib);
    if ($projName eq 'LowCov') {
	($indiv,$subpop,$lib) = split(/-/,$sampleName);
	if ($lib eq '1 1') {
	    $lib = 1;
	} elsif ($lib !~ /^\d+$/) {
	    warn "Bad library name $lib in $sampleName\n";
	}
    } elsif ($projName eq 'Trio') {
	my $suf;
	($indiv,$suf) = split(/-/,$sampleName);
	if (defined $pTrioSubpop->{$indiv}) {
	    $subpop = $pTrioSubpop->{$indiv};
	} else {
	    $subpop = 'CEU';
	    print STDERR "WARNING: no TRIO_SUBPOP for $indiv in $sampleName; guessing CEU\n";
	}
	if ($suf =~ /^WGv(\d+)$/) {
	    $lib = $1;
	} elsif ($suf eq 'WG') {
	    $lib = 1;
	} else {
	    $lib = 0;
	    print STDERR "WARNING: guessing lib=0 for $suf in $sampleName\n";
	}
    }
    if ($lib > 0) {
	$lib = sprintf("SC_%s_%s",$indiv,$lib);
    }
    return ($indiv,$subpop,$lib);
}

sub applyToLanes {
    my ($id,$pFunc,$pOpt,$pBatchOnly,$pRunOnly,$pSampleOnly,$pStaging,@rest) = @_;

    # Way in to NPG database
    my $root_run = npg::api::run->new();
    # Handle for my project; $id should be its project number.
    my $proj = st::api::project->new({id => $id});

    # A batch just has an id(), which is a number. It's a set
    # of runs, not lanes, so it can't correspond to a sample.
    # For that, see $sample below.
    report2("retrieving batches for project $id");
    my @batches = sort {$a->id() <=> $b->id()} @{$proj->batches()};
    report2();
    my $nb = scalar(@batches);
    my $thorough = $pBatchOnly || $pRunOnly || $pSampleOnly;
    foreach my $i (0 .. $#batches) {
	next if (defined $pOpt->{n} && $i%10 != $pOpt->{n});
	my $batch = $batches[$i];
	my $batchId = $batch->id();
	next if (defined $pOpt->{f} && $pOpt->{f} > $batchId);
	next unless (passesFilter($batchId,$pBatchOnly));
	report2(sprintf("retrieving runs for batch %d = %d of %d",$batch->id(),$i+1,$nb));
	my @runs = @{$root_run->runs_on_batch($batch->id())};
	report2();
	# A run has: id_run, id_instrument, id_run_pair,
	# name, is_paired and batch_id (among other fields).
	foreach my $run (@runs) {
	    my $runId = $run->id_run();
	    if (passesFilter($runId,$pRunOnly)) {
		applyToRun($id,$run,$runId,$batchId,$thorough,$pFunc,$pOpt,$pRunOnly,$pSampleOnly,$pStaging,@rest);
	    }
	}
    }
}

sub applyToRun {
    my ($id,$run,$runId,$batchId,$thorough,$pFunc,$pOpt,$pRunOnly,$pSampleOnly,$pStaging,$pTrioSubpop) = @_;
    # A run_lane has: id_run_lane, id_run, tile_count,
    # tracks, id_project (the last is not the same as $proj->id
    # so it must be a different type of project. Instead,
    # use $sample->project_id below); and position.
    my $pairId = $run->is_paired() ? $run->run_pair->id_run() : undef;
    my $missingOK = ($pairId && $pairId < $runId);
    report2("retrieving lanes for run $runId in batch $batchId");
    my @lanes;
    eval {@lanes = @{$run->run_lanes()}};
    if ($@) {
	warn $@;
	report2("failed");
    } else {
	report2();
    }
    #@lanes = @lanes[-1]; # for debugging
  LOOP:
    for my $run_lane (@lanes) {
	my $pos = $run_lane->position();
	# A sample has: id, name, count, user, project_id, ended
	my $sample  = $run_lane->sample();
	if ($sample) {
	    my $sName = $sample->name();
	    # HACKs for aberrant sample names:
	    $sName = 'NA07357-CEU-1' if ($sName eq 'NA07357 1');
	    $sName =~ s/ 1$//;
	    my $spid = $sample->project_id();
	    # Necessary hack ...
	    if ($spid == 45 && ($id == 50 || $id == 10)) {
		report2("WARNING: correcting project from $spid to $id for $sName",1);
		$spid = $id;
	    }
	    my $hierDir = decideDataSubdirectory($id,$runId,$pos,$sName,$pTrioSubpop);
	    if ($spid == $id) {
		my @existing = glob("$hierDir/*.fastq");
		if (@existing == 1) {
		    die("Multiple fastq files: @existing") if (@existing > 1);
		    my $tgt = readlink($existing[0]);
		    if (-s $tgt) {
			report2("already got @existing",1);
		    } else {
			report2("removing dangling link: $existing[0]",1);
			unlink(@existing);
			@existing = ();
		    }
		} 
		if (@existing) {
		    # now OK
		} elsif (passesFilter($sName,$pSampleOnly)) {
		    my $reposDir = $SLX_FASTQ{sprintf("%s_%s.fastq",$pStaging->{$runId},$pos)};
		    #my $reposDir = $pStaging->{$runId}[$pos];
		    unless (-d $reposDir) {
			report2("retrieving repository directory for lane ${runId}_$pos");
			my $casproject;
			eval {$casproject = $run_lane->project();
			      $reposDir = $casproject->repository() if ($casproject);
			};
			warn @_ if (@$);
			report2(-d $reposDir ? "" : "failed");
		    }
		    if (-d $reposDir) {
			last LOOP unless ($pFunc->($pOpt,$runId,$pos,$hierDir,$reposDir,$missingOK) || $thorough);
		    }
		}
	    } else { # mixed run...
		report2("WARNING: lane ${runId}_$pos is for sample $sName from project $spid",1);
		if (-d $hierDir && !defined $PROJECT_NAMES{$spid}) {
		    report2("$hierDir: deleting");
		    `rm -rf $hierDir`;
		    report2();
		}
	    }
	}
    }
}
    
sub report2 {
    my ($msg,$alone) = @_;
    report($msg,$alone,$LOG_STREAM);
}

# Run on cbi4:

# Args:
#  * $G1K/misc/spreadsheets/0804/summary_lane.txt or
#    $G1K/misc/spreadsheets/0806/*.csv
#  * org name wanted
#  * one or more repository dirs to try.
# The repos dirs can just be basenames, in which case pfind is called. 
#
# Example repos dirs:
# /nfs/repository/d0037/SLX_1000_EXT_1 BI data, 0804
# /nfs/repository/d0037/SLX_1000_EXT_2 BGI data, 0804 and 0806
# /nfs/repository/d0039/SLX_1000_EXT_3 Non-Illumina data, not of interest here
# /nfs/repository/d0039/SLX_1000_EXT_4 BI data, 0806 (and some non-Illumina)
#
# Example calls:
#   cd $G1K/DATA
#   updateDataDirectory.pl -A ../misc/spreadsheets/0804/summary_lane.txt BI  /nfs/repository/d0037/SLX_1000_EXT_1
#   updateDataDirectory.pl -A ../misc/spreadsheets/0804/summary_lane.txt BGI /nfs/repository/d0037/SLX_1000_EXT_2
#   updateDataDirectory.pl -A ../misc/spreadsheets/0806/bi.csv  BI  /nfs/repository/d0039/SLX_1000_EXT_4
#   updateDataDirectory.pl -A ../misc/spreadsheets/0806/bgi.csv BGI /nfs/repository/d0037/SLX_1000_EXT_2

sub augmentDataDirectory {
    my ($laneFile,$orgWanted,@reposDirs) = @_;

    my $isCSV = ($laneFile =~ /csv$/);

    foreach my $i (0 .. $#reposDirs) {
	$reposDirs[$i] = mayPfind($reposDirs[$i]);
    }

    my %complained;
    my $fh = openToRead($laneFile);
    my $hdr = <$fh> if ($isCSV); # header
    while (<$fh>) {
	chomp;
	my ($org,$run,$lane,$sample,$lib,$len1,$len2);
	if ($isCSV) {
	    s/ *, */,/g;
	    my @a = split(/,/,$_);
	    my ($rps,$rlen);
	    ($org,$run,$sample,$lib,$lane,$rps,$rlen) = @a[2,9,3,7,22,10,11];
	    $lane =~ s/.*[\._]([1-8])\.srf/\1/;
	    if ($org eq 'BI') {
		$run =~ s/\.[1-8]$//;
		$run =~ s/^BI\.//;
	    }
	    if ($rps == 1) {
		($len1,$len2) = ($rlen,0);
	    } else {
		($len1,$len2) = ($rlen,$rlen);
	    }
	} else {
	    # When processing summary_lane.txt ...
	    ($org,$run,$lane,$sample,$lib,$len1,$len2) = split;
	}
	next unless ($org eq $orgWanted);
	my $reposBase = join('_',$org,$run,$lane);
	my $reposSuf;
	# We expect the repository file basename to match this pattern:
	my $reposBasePat = sprintf("%s*%s?%s*",$org,$run,$lane);
	my $reposFile;
      DIRS:
	foreach my $d (@reposDirs) {
	    foreach my $rf (glob("$d/$reposBasePat")) {
		foreach my $suf (qw(fastq fastq.gz fq.gz)) {
		    if ($rf =~ /$suf$/) {
			$reposFile = $rf;
			$reposSuf = $suf;
			last DIRS;
		    }
		}
	    }
	}
	unless ($reposFile) {
	    die("Cannot find $reposBasePat in @reposDirs");
	}
	my ($sampleDir) = glob("$DATA_DIR/*/$sample");
	if (defined $sampleDir) {
	    my $libSub = sprintf("lib%s_%s_%s",$org,$sample,$lib);
	    my $laneDir = join("/",$sampleDir,$libSub,"lane$reposBase");
	    my $linkFile = sprintf("%s/%s.%s",$laneDir,$reposBase,$reposSuf);
	    unless (-e $linkFile && readlink($linkFile) eq $reposFile) {
		mkdir_p($laneDir);
		symlink($reposFile,$linkFile);
	    }
	    my $rlFile = sprintf("%s/%s.readlengths",$laneDir,$reposBase);
	    if ($orgWanted eq 'BI') {
		my ($len1C,$len2C) = mayCorrectReadLengths($reposFile);
		unless ($len1C == $len1 && $len2C == $len2) {
		    unlink($rlFile) if (-s $rlFile);
		    ($len1,$len2) = ($len1C,$len2C);
		}
	    }
	    unless (-s $rlFile) {
		my $gh = openToWrite($rlFile);
		print $gh "$len1 $len2\n";
		$gh->close();
	    }
	} elsif (!$complained{$sample}) {
	    # We don't know what subpop it's for ...
	    $complained{$sample} = 1;
	    warn("Cannot find $DATA_DIR/*/$sample");
	}
    }
}

sub mayCorrectReadLengths {
    my ($f) = @_;
    my ($d,$b) = (dirname($f),basename($f,'.fastq.gz'));
    my $f1 = sprintf("%s/bi_separate/%s_1.fastq.gz",$d,$b);
    my $f2 = sprintf("%s/bi_separate/%s_2.fastq.gz",$d,$b);
    my $len = getReadLength($f);
    if (-s $f1 && -s $f2) {
	return ($len/2,$len/2);
    } elsif (-s $f) {
	return ($len,0);
    } 
}

sub getReadLength {
    my ($f) = @_;
    if (-s $f) {
	my $fh = openToRead($f);
	<$fh>;
	chomp ($_  = <$fh>);
	return length($_);
    } else {
	return -1;
    }
}

sub mayPfind {
    my ($d) = @_;
    if (substr($d,0,1) ne '/') {
	chomp ($d = `pfind $d`);
    }
    return $d;
}

sub createReadLengthsFile {
    my ($reposFile,$rlFile) = @_;
    if (-s $reposFile && ! -s $rlFile) {
	my $fh = openToRead($reposFile);
	<$fh>; #
	chomp (my $seq = <$fh>);
	# Assume Sanger data is always paired-end with equal lengths,
	# except if odd or <50.
	my $len = length($seq);
	my $gh = openToWrite($rlFile);
	if ($len%2 || $len < 50) {
	    print $gh "$len 0\n";
	} else {
	    my $len2 = $len/2;
	    print $gh "$len2 $len2\n";
	}
	$gh->close();
    }
}
