#!/usr/bin/env perl

=head1 NAME

vocab.pl

=head1 SYNOPSIS

vocab.pl [-lctnprz] file1 file2 ...

=head1 DESCRIPTION

Switches are:

=over

=item -l: token is whole line

=item -c: token is non-space character

=item -t: line is "x y"

=item -n: line is "tokenCount token", or "token tokenCount" if -rn, 
          or "x y pairCount" if -tn

=item -p: print proportions not counts

=item -r: print token then count instead of vice versa

=item -s: print sum-so-far of counts (or proportions) at end of each line

=item -z: print zeros for missing values (only if -r)

=back

=head1 AUTHOR

David Carter, C<dmc@sanger.ac.uk>

=head1 SEE ALSO

       

=cut

use Getopt::Std;
use strict;

my %OPT;
getopts('lcnprstz',\%OPT);

vocab(\%OPT);

sub vocab {
    my ($pOpt) = @_;
    my %c;
    while (<>) {
	chomp;
	if ($pOpt->{l}) {
	    $c{$_} ++;
	} elsif ($pOpt->{t}) {
	    my ($x,$y,$n) = split;
	    if ($pOpt->{n}) {
		$c{$x}{$y} += $n;
	    } else {
		$c{$x}{$y} ++;
	    }
	} elsif ($pOpt->{n}) {
	    my ($tok,$n);
	    if ($pOpt->{r}) {
		($tok,$n) = split;
	    } else {
		($n,$tok) = split;
	    }
	    $c{$tok} += $n;
	} elsif ($pOpt->{c}) {
	    s/\s//g;
	    foreach my $w (split(//,$_)) {
		$c{$w} ++;
	    }
	} else {
	    foreach my $w (split) {
		$c{$w} ++;
	    }
	}
    }
    if ($pOpt->{t}) {
	printPairTable(\%c,$pOpt);
    } else {
	printCounts(\%c,$pOpt);
    }
}

sub printPairTable {
    my ($pC,$pOpt) = @_;
    my ($sumxy,%x,%y);
    my ($maxlx,$maxly) = (3,$pOpt->{p} ? 6 : 3);
    foreach my $xi (keys %$pC) {
	my $pH = $pC->{$xi};
	$maxlx = length($xi) if ($maxlx < length($xi));
	foreach my $yi (keys %$pH) {
	    my $len1 = length($yi);
	    my $len2 = length($pH->{$yi});
	    my $len = $len1 > $len2 ? $len1 : $len2;
	    $maxly = $len if ($maxly < $len);
	    $x{$xi}=$y{$yi}=1;
	    $sumxy += $pH->{$yi};
	}
    }
    # Print top row:
    printf("%${maxlx}s");
    foreach my $yi (sort keys %y) {
	printf(" %${maxly}s",$yi);
    }
    printf(" %${maxly}s\n","Sum");
    my (%ysum,%xsum);
    foreach my $xi (sort keys %x) {
	my $pH = $pC->{$xi};
	printf("%${maxlx}s",$xi);
	my $sum;
	foreach my $yi (sort keys %y) {
	    $sum += $pH->{$yi};
	    $ysum{$yi} += $pH->{$yi};
	}
	if ($pOpt->{p}) {
	    foreach my $yi (sort keys %y) {
		printf(" %${maxly}.4f",$pH->{$yi}/$sum);
	    }
	    printf(" %${maxly}.4f",$sum/$sumxy);
	} else {
	    foreach my $yi (sort keys %y) {
		printf(" %${maxly}d",$pH->{$yi});
	    }
	    printf(" %${maxly}d",$sum);
	}
	print "\n";
	$xsum{$xi} = $sum;
    }
    printf("%${maxlx}s","Sum");
    if ($pOpt->{p}) {
        foreach my $yi (sort keys %ysum) {
            printf(" %${maxly}.4f",$ysum{$yi}/$sumxy);
        }
	printf(" %${maxly}.4f\n",1);
    } else {
        foreach my $yi (sort keys %ysum) {
            printf(" %${maxly}d",$ysum{$yi});
        }
	printf(" %${maxly}d\n",$sumxy);
    }
    # Chi squared and entropy
    my $cs = 0;
    my $csok = 1;
    my $init = 1;
    my ($hy,$hygx);
    foreach my $xi (keys %x) {
	my $px = $xsum{$xi}/$sumxy;
	foreach my $yi (keys %y) {
	    if ($init) {
		my $py = $ysum{$yi}/$sumxy;
		$hy -= $py*log($py);
	    }
	    my $o = $pC->{$xi}{$yi};
	    if ($o > 0) {
		my $pygx = $o/$xsum{$xi};
		$hygx -= $px*($pygx*log($pygx));
	    }
	    my $e = $xsum{$xi}*$ysum{$yi}/$sumxy;
	    if ($e < 5) {
		$csok=0;
	    } elsif ($csok) {
		$cs += ($o-$e)*($o-$e)/$e;
	    }
	}
	$init = 0;
    }
    if ($csok) {
	my $nx = scalar(keys %x);
	my $ny = scalar(keys %y);
	printf("Chi-sq = %5.3f, DF = %d\n",$cs,($nx-1)*($ny-1));
    }
    my $q = 1/(log(2));
    my $pie = 1-$hygx/$hy;
    printf("H(Y) %5.3f H(Y|X) %5.3f %%PIE %5.3f\n",$hy*$q,$hygx*$q,100*$pie);
}

sub printCounts {
    my ($pC,$pOpt) = @_;
    my $fmt = $pOpt->{p} ? "%8.8f" : "%d";
    my $cumul = $pOpt->{s};
    my $ssf = 0;
    if ($pOpt->{p}) {
	my $sum;
	foreach my $w (keys %$pC) {
	    $sum += $pC->{$w};
	}
	foreach my $w (keys %$pC) {
	    $pC->{$w} /= $sum;
	}
    }
    if ($pOpt->{r}) {
	if ($pOpt->{z}) {
	    my @keys = sort {$a <=> $b} keys %$pC;
	    for (my $w = $keys[0]; $w <= $keys[-1]; $w++) {
		if ($cumul) {
		    $ssf += $pC->{$w};
		    printf("%d $fmt $fmt\n",$w,$pC->{$w},$ssf);
		} else {
		    printf("%d $fmt\n",$w,$pC->{$w});
		}
	    }
	} elsif ($cumul) {
	    foreach my $w (sort {$a cmp $b} keys %$pC) {
		$ssf += $pC->{$w};
		printf("%s $fmt $fmt\n",$w,$pC->{$w},$ssf);
	    }
	} else {
	    foreach my $w (sort {$a cmp $b} keys %$pC) {
		printf("%s $fmt\n",$w,$pC->{$w});
	    }
	}
    } elsif ($cumul) {
	foreach my $w (sort {$pC->{$b} <=> $pC->{$a} || $a cmp $b} (keys %$pC)) {
	    $ssf += $pC->{$w};
	    printf("$fmt %s $fmt\n",$pC->{$w},$w,$ssf);
	}
    } else {
	foreach my $w (sort {$pC->{$b} <=> $pC->{$a} || $a cmp $b} (keys %$pC)) {
	    printf("$fmt %s\n",$pC->{$w},$w);
	}
    }
}
