#!/usr/bin/env perl

=head1 NAME

blast.pl

=head1 SYNOPSIS

blast.pl [-r fun] [-x] [-v] foo.fa bar.fa ...

=head1 DESCRIPTION

Run BLASTN on specified FASTA files (which should only contain ONE
sequence each) against EMBL databases, processing the results to print
one line on each match. If "-r xyz" is given, we only run against
database names containing "xyz"; specifically, database files named
embl_*_xyz-... Thus possible values of xyz at the time of writing are:
env fun hum inv mam mus pln pro rod unc vrt.  If "-v" is given, we
match against all databases EXCEPT xyz ones.

For each foo.fa, creates an output file foo.fa.blastout, or if
"-r xyz" is given, then foo.fa.xyz.blastout; or if "-v -r xyz",
foo.fa.notxyz.blastout.

If "-x", use blastx, else blastn.

Useful for finding matches for reads that didn't get any matches
from the main SGRP pipeline.

=head1 AUTHOR

David Carter, C<dmc@sanger.ac.uk>

=head1 SEE ALSO

       

=cut


use FileHandle;
use File::Basename;
use Getopt::Std;

use strict;

my %OPT;
getopts('r:vx',\%OPT); # -r fun, etc; -v means "not".

my $BLAST = $OPT{x} ? 'blastx' : 'blastn';
foreach my $d (qw(/usr/local/ensembl/bin /software/pubseq/bin/wu_blast)) {
    if (-x "$d/$BLAST") {
	$BLAST = "$d/$BLAST";
	last;
    }
}

# From Ian Korf's "BLAST theory and practice" PDF. The blastn parameters are
# his "faster and more sensitive" ones; the other set (with hitdist) gobble
# up all available memory.

my $PARAMS = $OPT{x} ? 'W=4 W=20 wordmask=seg' : 'W=9 M=1 N=-1 Q=3 R=2 wordmask=seg';

# Replace this by wherever you have your blast databases. We have
# /data/blastdb/embl_est_env-1, etc.

my $DB_DIR;
foreach my $d (qw(/data/blastdb/Supported /data/blastdb)) {
    if (-d $d) {
	$DB_DIR = $d;
	last;
    }
}

my @dbList = getDatabases();
#my @dbList = qw(/nfs/faculty/jkh1/Mitocheck/Database/external_data/Homo_sapiens.NCBI36.46.cdna.all.fa);

autoflush STDOUT 1;

my $ifile = 0;
foreach my $file (@ARGV) {
    $ifile++;
    my $not = defined $OPT{v} ? 'not' : '';
    my $g = defined $OPT{r} ? "$file.$not$OPT{r}.blastout" : "$file.blastout";
    if (-f $g) {
	print "$g: already exists\n";
	next;
    }
    my $gh = new FileHandle(">$g"); # create it straight away
    my %matches;
    my $idb = 0;
    my %description;
    my %ident;
    foreach my $db (@dbList) {
	$idb ++;
	printf("Matching %s (file %d/%d) against %s (database %d/%d)\n",
	       $file,$ifile,scalar(@ARGV),basename($db),$idb,scalar(@dbList));
	my $base = basename($db);
	my $fh = new FileHandle("$BLAST $db $file $PARAMS |");
	my $idx = 0;
	my $curName;
	my $seenLength;
	while (<$fh>) {
	    chomp;
	    if (/^Sequences producing High-scoring Segment Pairs:/) {
		$idx = 1;
	    } elsif ($idx) {
		$idx ++;
		if ($idx > 2) {
		    next if (/Query:/ || /Sbjct:/);
		    my ($name,$post) = (/^(\S+) .* (\d+\s+\S+\s+\d+)$/);
		    if ($name) {
			my $line = "$base $post $name";
			my (undef,undef,$e) = split(" ",$line);
			if ($e <= 0.01) { # e > 0.01 unlikely to be of interest...
			    $matches{$e}{$name} = "$line";
			    print "$line\n"; # for bpeeking at
			}
		    } elsif (/^>(\S+) (.*)$/) {
			$curName = $1;
			$description{$1} = $2;
			$seenLength = undef;
		    } elsif (/^\s+Length/) {
			$seenLength = 1;
		    } elsif ($curName && !$seenLength) {
			s/^\s+//;
			$description{$curName} .= " $_";
		    } elsif (/Identities = (\d+)\/(\d+)/) {
			$ident{$curName} = [$1,$2];
		    }
		}
	    }
	}
    }
    my $gh = new FileHandle(">$g");
    foreach my $e (sort {$a <=> $b} keys %matches) {
	my $pH = $matches{$e};
	foreach my $name (sort keys %$pH) {
	    if (defined $ident{$name}) {
		my ($num,$den) = @{$ident{$name}};
		printf($gh "%s %4d %4d %3d%% %s\n",
		       $pH->{$name},
		       $num,$den,int(0.5+100*$num/$den),
		       $description{$name});
	    } else {
		printf($gh "%s NO_MORE_DETAILS\n",
		       $pH->{$name});
	    }
	}
    }
    $gh->close();
}

sub getDatabases {
    my @ans;
    my @names;
    if ($OPT{x}) {
	@names = qw(uniprot);
    } else {
	@names = qw(est gss sts htg other vertrna);
    }
    foreach my $name (@names) {
	foreach my $file (glob("$DB_DIR/embl_${name}_*-[0-9].xnt")) {
	    $file =~ s/\.xnt$//;
	    if ($OPT{v}) {
		push @ans,$file unless ($file =~ /_$OPT{r}/);
	    } elsif ($OPT{r}) {
		push @ans,$file if ($file =~ /_$OPT{r}/);
	    } else {
		push @ans,$file;
	    }
	}
    }
    return @ans;
}
	
